Package: hexify
Title: Equal-Area Hex Grids on the 'Snyder' 'ISEA' 'Icosahedron'
Version: 0.6.5
Authors@R: 
    person("Gilles", "Colling", , "gilles.colling051@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-3070-6066"))
Description: Provides functions to build and use hexagonal discrete global grids
    using the 'Snyder' 'ISEA' projection ('Snyder' 1992
    <doi:10.3138/27H7-8K88-4882-1752>) and the 'H3' hierarchical hexagonal system
    ('Uber' Technologies). Implements the 'ISEA' discrete global grid system
    ('Sahr', 'White' and 'Kimerling' 2003 <doi:10.1559/152304003100011090>).
    Includes a fast 'C++' core for 'ISEA' projection and aperture quantization,
    an included 'H3' v4.4.1 C library for native 'H3' grid operations, and
    'sf'/'terra'-compatible R wrappers for grid generation and coordinate assignment.
    Output is compatible with 'dggridR' for interoperability.
License: MIT + file LICENSE
SystemRequirements: C99
Language: en-US
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: testthat (>= 3.0.0), lifecycle, knitr, rmarkdown, terra,
        raster, ggplot2, RColorBrewer, rnaturalearth, tibble, gridExtra
VignetteBuilder: knitr
LinkingTo: Rcpp
Imports: sf, Rcpp, methods, rlang
URL: https://gillescolling.com/hexify/
BugReports: https://github.com/gcol33/hexify/issues
Config/testthat/edition: 3
Depends: R (>= 3.5)
LazyData: true
NeedsCompilation: yes
Packaged: 2026-02-28 21:52:52 UTC; Gilles Colling
Author: Gilles Colling [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-3070-6066>)
Maintainer: Gilles Colling <gilles.colling051@gmail.com>
Repository: CRAN
Date/Publication: 2026-02-28 22:50:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 08:02:38 UTC; unix
Archs: hexify.so.dSYM

Package: hexify
Title: Equal-Area Hex Grids on the Snyder ISEA Icosahedron
Version: 0.8.2
Authors@R: c(
    person("Gilles", "Colling", , "gilles.colling051@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-3070-6066")),
    person("Uber Technologies, Inc.", role = c("ctb", "cph"),
           comment = "H3 C library, src/h3 (Apache-2.0)")
  )
Description: Provides functions to build and use hexagonal discrete global grids
    using the Icosahedral Snyder Equal Area ('ISEA') projection (Snyder 1992
    <doi:10.3138/27H7-8K88-4882-1752>) and the 'H3' hierarchical hexagonal system
    (Uber Technologies). Implements the 'ISEA' discrete global grid system
    (Sahr, White and Kimerling 2003 <doi:10.1559/152304003100011090>).
    Includes a fast C++ core for 'ISEA' projection and aperture quantization,
    an included 'H3' v4.4.1 C library for native 'H3' grid operations, and
    'sf'/'terra'-compatible R wrappers for grid generation and coordinate assignment.
    Output is compatible with 'dggridR' for interoperability.
License: MIT + file LICENSE
SystemRequirements: C99
Language: en-US
Encoding: UTF-8
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, terra, raster,
        ggplot2, RColorBrewer, rnaturalearth, tibble, gridExtra,
        leaflet
VignetteBuilder: knitr
LinkingTo: Rcpp
Imports: sf, Rcpp, methods, rlang
URL: https://gillescolling.com/hexify/
BugReports: https://github.com/gcol33/hexify/issues
Config/testthat/edition: 3
Depends: R (>= 3.5)
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-21 22:29:04 UTC; Gilles Colling
Author: Gilles Colling [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-3070-6066>),
  Uber Technologies, Inc. [ctb, cph] (H3 C library, src/h3 (Apache-2.0))
Maintainer: Gilles Colling <gilles.colling051@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-22 16:00:03 UTC

# Generated by roxygen2: do not edit by hand

S3method(as_sf,HexData)
S3method(as_sf,default)
export(as_dggrid)
export(as_sf)
export(as_tibble.HexData)
export(cell_area)
export(cell_to_index)
export(cell_to_lonlat)
export(cell_to_sf)
export(cells)
export(dg_closest_res_to_area)
export(dgearthstat)
export(dggrid_43h_sequence)
export(dggrid_is_compatible)
export(dgverify)
export(from_dggrid)
export(get_children)
export(get_parent)
export(globe_centers)
export(grid_clip)
export(grid_global)
export(grid_info)
export(grid_rect)
export(h3_crosswalk)
export(hex_corners_to_sf)
export(hex_grid)
export(hexify)
export(hexify_area_to_eff_res)
export(hexify_assign)
export(hexify_build_icosa)
export(hexify_cell_id_to_quad_ij)
export(hexify_cell_to_icosa_tri)
export(hexify_cell_to_index)
export(hexify_cell_to_lonlat)
export(hexify_cell_to_plane)
export(hexify_cell_to_quad_ij)
export(hexify_cell_to_quad_xy)
export(hexify_cell_to_sf)
export(hexify_compare_indices)
export(hexify_compare_resolutions)
export(hexify_default_index_type)
export(hexify_eff_res_to_area)
export(hexify_eff_res_to_resolution)
export(hexify_face_centers)
export(hexify_forward)
export(hexify_forward_to_face)
export(hexify_get_children)
export(hexify_get_parent)
export(hexify_get_precision)
export(hexify_get_resolution)
export(hexify_grid)
export(hexify_grid_cell_to_lonlat)
export(hexify_grid_global)
export(hexify_grid_rect)
export(hexify_grid_to_cell)
export(hexify_heatmap)
export(hexify_icosa_tri_to_plane)
export(hexify_icosa_tri_to_quad_ij)
export(hexify_icosa_tri_to_quad_xy)
export(hexify_index_to_cell)
export(hexify_index_to_lonlat)
export(hexify_inverse)
export(hexify_is_valid_index_type)
export(hexify_lonlat_to_cell)
export(hexify_lonlat_to_index)
export(hexify_lonlat_to_plane)
export(hexify_lonlat_to_quad_ij)
export(hexify_projection_stats)
export(hexify_quad_ij_to_cell)
export(hexify_quad_ij_to_icosa_tri)
export(hexify_quad_ij_to_xy)
export(hexify_quad_xy_to_cell)
export(hexify_quad_xy_to_icosa_tri)
export(hexify_resolution_to_eff_res)
export(hexify_roundtrip_test)
export(hexify_set_precision)
export(hexify_set_verbose)
export(hexify_which_face)
export(hexify_z7_canonical)
export(is_hex_data)
export(is_hex_grid)
export(lonlat_to_cell)
export(n_cells)
export(plot_globe)
export(plot_grid)
export(plot_world)
exportClasses(HexData)
exportClasses(HexGridInfo)
exportMethods("$")
exportMethods("$<-")
exportMethods("[")
exportMethods("[[")
exportMethods("[[<-")
exportMethods(as.data.frame)
exportMethods(as.list)
exportMethods(cells)
exportMethods(dim)
exportMethods(grid_info)
exportMethods(n_cells)
exportMethods(names)
exportMethods(ncol)
exportMethods(nrow)
exportMethods(plot)
exportMethods(show)
import(methods)
importFrom(Rcpp,sourceCpp)
importFrom(grDevices,adjustcolor)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,setValidity)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(methods,validObject)
importFrom(rlang,.data)
importFrom(utils,head)
useDynLib(hexify, .registration = TRUE)

# hexify

[![CRAN status](https://www.r-pkg.org/badges/version/hexify)](https://CRAN.R-project.org/package=hexify)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/grand-total/hexify)](https://cran.r-project.org/package=hexify)
[![Monthly downloads](https://cranlogs.r-pkg.org/badges/hexify)](https://cran.r-project.org/package=hexify)
[![R-CMD-check](https://github.com/gcol33/hexify/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/gcol33/hexify/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/gcol33/hexify/graph/badge.svg)](https://app.codecov.io/gh/gcol33/hexify)
[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)

**Equal-Area Hexagonal Grids for Global Spatial Analysis**

<p align="center">
  <img src="man/figures/hexify-hero.png" alt="Multi-resolution hexagonal grids" width="100%">
</p>

`hexify` assigns geographic coordinates to equal-area hexagonal grid cells using the ISEA (Icosahedral Snyder Equal Area) projection. Every cell has the same area regardless of latitude, eliminating the sampling bias inherent in rectangular lat-lon grids. H3 is supported for compatibility with existing H3 workflows.

## Quick Start

```r
library(hexify)

cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid"),
  lon = c(16.37, 2.35, -3.70),
  lat = c(48.21, 48.86, 40.42)
)

# ISEA equal-area grid (default)
grid <- hex_grid(area_km2 = 10000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)
plot(result)

# H3 grid (Uber's system)
h3_grid <- hex_grid(resolution = 4, type = "h3")
result_h3 <- hexify(cities, lon = "lon", lat = "lat", grid = h3_grid)
plot(result_h3)
```

## Statement of Need

Spatial binning is fundamental to ecological modeling, epidemiology, and geographic analysis. Standard approaches using rectangular lat-lon grids introduce severe area distortions: a 1° cell at the equator covers ~12,300 km², while the same cell near the poles covers a fraction of that area. This violates the equal-sampling assumption underlying most spatial statistics.

Discrete Global Grid Systems (DGGS) solve this by partitioning Earth's surface into cells of uniform area. hexify's primary backend is **ISEA** (Icosahedral Snyder Equal Area): true equal-area hexagonal grids with apertures 3, 4 and 7 in any sequence, implemented in C++ with no external dependencies. For interoperability with industry ecosystems (FCC, Foursquare, DuckDB), hexify also supports **H3** grids via a vendored C library.

Equal-area grids are directly applicable to:

- Species distribution modeling and biodiversity assessments
- Epidemiological surveillance and disease mapping
- Environmental monitoring and remote sensing aggregation
- Any analysis requiring unbiased spatial binning

## Why Hexagonal Grids?

Hexagons tile the sphere with three properties that squares and triangles lack:

1. **Equal area** — every cell covers the same surface area, from equator to pole
2. **Uniform adjacency** — all six neighbors share an edge (no ambiguous diagonal neighbors)
3. **Low shape distortion** — hexagons approximate circles better than any other regular polygon, minimizing edge effects in spatial statistics

These properties make hexagonal grids the natural choice for unbiased spatial binning. Rectangular lat-lon grids, by contrast, shrink toward the poles: a 1° cell at 60°N has half the area of the same cell at the equator.

## Features

### Core Workflow

- **`hex_grid()`**: Define a grid by target cell area (km²) or resolution level
- **`hexify()`**: Assign points to grid cells (data.frame or sf input)
- **`plot()` / `hexify_heatmap()`**: Visualize results with base R or ggplot2
- **Any body**: `hex_grid(area_km2 = 1000, radius_km = "mars")` sizes a grid on Mars, the Moon, Titan, or any radius you give — both backends

### Grid Generation

- **`grid_rect()`**: Generate cell polygons for a bounding box
- **`grid_global()`**: Generate a complete global grid (all cells)
- **`grid_clip()`**: Clip grid to a polygon boundary (country, region, etc.)

### Cell Operations

- **`cell_to_sf()`**: Convert cell IDs to sf polygon geometries
- **`cell_to_lonlat()`**: Get cell center coordinates
- **`get_parent()` / `get_children()`**: Navigate grid hierarchy

### Interoperability

- **`as_dggrid()` / `from_dggrid()`**: Convert to/from dggridR format
- **`as_sf()`**: Export HexData to sf object
- **`as.data.frame()`**: Extract data with cell assignments
- **H3 support**: `hex_grid(resolution = 8, type = "h3")` — vendored H3 C library, no extra install needed

## Installation

```r
# Install from CRAN
install.packages("hexify")

# Or install development version from GitHub
# install.packages("pak")
pak::pak("gcol33/hexify")
```

## Usage Examples

### Basic Point Assignment

```r
library(hexify)

# Define grid: ~10,000 km² cells
grid <- hex_grid(area_km2 = 10000)
grid
#> HexGridInfo Specification
#> -------------------------
#> Aperture:    3
#> Resolution:  8
#> Area:        7773.97 km^2
#> Diagonal:    94.74 km
#> CRS:         EPSG:4326
#> Total Cells: 65612

# Assign coordinates to cells
coords <- data.frame(
  lon = c(-122.4, 2.35, 139.7),
  lat = c(37.8, 48.9, 35.7)
)
result <- hexify(coords, lon = "lon", lat = "lat", grid = grid)

# Access cell IDs
result@cell_id
```

### Working with sf Objects

```r
library(sf)

# Any CRS works - hexify transforms automatically
points_sf <- st_as_sf(coords, coords = c("lon", "lat"), crs = 4326)
result <- hexify(points_sf, area_km2 = 10000)

# Export back to sf
result_sf <- as_sf(result)
```

### Generating Grid Polygons

```r
# Grid for Europe
grid <- hex_grid(area_km2 = 50000)
europe_hexes <- grid_rect(c(-10, 35, 40, 70), grid)
plot(europe_hexes["cell_id"])

# Clip to a country boundary
library(rnaturalearth)
france <- ne_countries(country = "France", returnclass = "sf")
france_grid <- grid_clip(france, grid)
```

### Aggregating Point Data

```r
# Species occurrence data
occurrences <- data.frame(
  species = sample(c("Sp A", "Sp B", "Sp C"), 1000, replace = TRUE),
  lon = runif(1000, -10, 30),
  lat = runif(1000, 35, 60)
)

# Assign to grid
grid <- hex_grid(area_km2 = 20000)
occ_hex <- hexify(occurrences, lon = "lon", lat = "lat", grid = grid)

# Count per cell
occ_df <- as.data.frame(occ_hex)
occ_df$cell_id <- occ_hex@cell_id

cell_counts <- aggregate(species ~ cell_id, data = occ_df, FUN = length)
names(cell_counts)[2] <- "n_records"

# Richness per cell
richness <- aggregate(species ~ cell_id, data = occ_df,
                      FUN = function(x) length(unique(x)))
names(richness)[2] <- "n_species"
```

### Visualization

```r
# Quick plot
plot(result)

# Heatmap with basemap
hexify_heatmap(occ_hex, value = "n_records", basemap = TRUE)

# Custom ggplot
library(ggplot2)
cell_polys <- cell_to_sf(cell_counts$cell_id, grid)
cell_polys <- merge(cell_polys, cell_counts, by = "cell_id")

ggplot(cell_polys) +
  geom_sf(aes(fill = n_records), color = "white", linewidth = 0.2) +
  scale_fill_viridis_c() +
  theme_minimal()
```

## Known Limitations

- **H3 grids**: Fixed aperture 7, maximum resolution 15 (~0.9 m² cells). ISEA grids support apertures 3, 4 and 7 in any sequence up to resolution 30.
- **Pentagons**: Any hexagonal tiling of a sphere requires exactly 12 pentagonal cells (at icosahedron vertices). These cells have 5 neighbors instead of 6. Use `is_pentagon()` to detect them.
- **Projection precision**: The inverse Snyder projection uses iterative Newton-Raphson convergence. Default precision is sufficient for sub-meter accuracy; use `hexify_set_precision()` to adjust the speed/accuracy trade-off.

## Documentation

- [Quick Start](https://gillescolling.com/hexify/articles/quickstart.html) - Basic concepts and workflow
- [Visualization](https://gillescolling.com/hexify/articles/visualization.html) - Plotting with base R and ggplot2
- [Workflows](https://gillescolling.com/hexify/articles/workflows.html) - Grid generation, clipping, multi-resolution analysis

## Support

> "Software is like sex: it's better when it's free." — Linus Torvalds

I'm a PhD student who builds R packages in my free time because I believe good tools should be free and open. I started these projects for my own work and figured others might find them useful too.

If this package saved you some time, buying me a coffee is a nice way to say thanks. It helps with my coffee addiction.

[![Buy Me A Coffee](https://img.shields.io/badge/-Buy%20me%20a%20coffee-FFDD00?logo=buymeacoffee&logoColor=black)](https://buymeacoffee.com/gcol33)

## Citation

```bibtex
@software{hexify,
  author = {Colling, Gilles},
  title = {hexify: Equal-Area Hexagonal Grids for Spatial Analysis},
  year = {2025},
  url = {https://CRAN.R-project.org/package=hexify},
  doi = {10.32614/CRAN.package.hexify}
}
```

## License

MIT (see LICENSE.md)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
library(hexify)
library(sf)
library(ggplot2)

## ----sample-data--------------------------------------------------------------
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome",
           "Warsaw", "Prague", "Brussels", "Amsterdam", "Lisbon"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50,
          21.01, 14.42, 4.35, 4.90, -9.14),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90,
          52.23, 50.08, 50.85, 52.37, 38.72)
)

grid <- hex_grid(area_km2 = 10000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

## ----plot-basic, fig.width=7, fig.height=5------------------------------------
plot(result, main = "European Cities")

## ----plot-custom, fig.width=7, fig.height=5-----------------------------------
plot(result,
     grid_fill = "steelblue",
     grid_border = "darkblue",
     grid_alpha = 0.6,
     basemap_fill = "ivory",
     basemap_border = "gray50",
     main = "Custom Styling")

## ----plot-no-basemap, fig.width=7, fig.height=5-------------------------------
plot(result,
     basemap = FALSE,
     grid_fill = "forestgreen",
     grid_border = "darkgreen",
     main = "No Basemap")

## ----plot-extent, fig.width=7, fig.height=5-----------------------------------
plot(result,
     crop = TRUE,
     crop_expand = 0.2,
     main = "Custom Extent (20% padding)")

## ----plot-points, fig.width=7, fig.height=5-----------------------------------
plot(result,
     show_points = TRUE,
     point_color = "red",
     main = "Cities with Points")

## ----point-sizes, fig.width=7, fig.height=7-----------------------------------
oldpar <- par(mfrow = c(2, 2))

plot(result, show_points = TRUE, point_size = "small",
     point_color = "red", main = "small (~5%)")
plot(result, show_points = TRUE, point_size = "normal",
     point_color = "red", main = "normal (~10%)")
plot(result, show_points = TRUE, point_size = "large",
     point_color = "red", main = "large (~20%)")
plot(result, show_points = TRUE, point_size = "very large",
     point_color = "red", main = "very large (~35%)")

par(oldpar)

## ----plot-points-custom, fig.width=7, fig.height=5----------------------------
plot(result,
     show_points = TRUE,
     point_size = "small",
     point_color = "darkblue",
     point_alpha = 0.8,
     grid_fill = "lightyellow",
     grid_border = "orange",
     main = "Custom Point Styling")

## ----plot-no-jitter, fig.width=7, fig.height=5--------------------------------
plot(result,
     show_points = TRUE,
     jitter = FALSE,
     point_color = "red",
     main = "Points at Cell Centers (No Jitter)")

## ----ggplot-basic, fig.width=7, fig.height=5----------------------------------
hexify_heatmap(result, basemap = "world", title = "European Cities")

## ----ggplot-custom, fig.width=7, fig.height=5---------------------------------
hexify_heatmap(result, basemap = "world") +
  labs(
    title = "Major European Cities",
    subtitle = "Assigned to ISEA hexagonal grid cells",
    caption = "Data: Sample cities"
  ) +
  theme_minimal() +
  theme(
    plot.title = element_text(face = "bold", size = 14),
    panel.grid = element_blank()
  )

## ----ggplot-layers, fig.width=7, fig.height=5---------------------------------
# Get the city coordinates
city_points <- st_as_sf(cities, coords = c("lon", "lat"), crs = 4326)

hexify_heatmap(result, basemap = "world", title = "Cities with Labels") +
  geom_sf(data = city_points, color = "red", size = 2) +
  geom_sf_text(data = city_points, aes(label = name),
               nudge_y = 0.8, size = 3, color = "darkgray") +
  coord_sf(xlim = c(-5, 25), ylim = c(45, 55))

## ----heatmap-setup------------------------------------------------------------
# Simulate observation data with counts
set.seed(42)
n_obs <- 100
obs_data <- data.frame(
  lon = c(rnorm(60, 10, 8), rnorm(40, 0, 10)),
  lat = c(rnorm(60, 48, 5), rnorm(40, 52, 6)),
  count = rpois(n_obs, lambda = 50)
)

# Hexify
grid <- hex_grid(area_km2 = 10000)
obs_hex <- hexify(obs_data, lon = "lon", lat = "lat", grid = grid)

## ----heatmap-basic, fig.width=8, fig.height=6---------------------------------
hexify_heatmap(
  obs_hex,
  value = "count",
  title = "Observation Counts"
)

## ----heatmap-colors, fig.width=8, fig.height=6--------------------------------
hexify_heatmap(
  obs_hex,
  value = "count",
  colors = "YlOrRd",
  title = "Yellow-Orange-Red Palette"
)

## ----heatmap-palettes, fig.width=8, fig.height=4------------------------------
p1 <- hexify_heatmap(obs_hex, value = "count", colors = "viridis",
                     title = "viridis", xlim = c(-20, 35), ylim = c(35, 65))
p2 <- hexify_heatmap(obs_hex, value = "count", colors = "YlGnBu",
                     title = "YlGnBu", xlim = c(-20, 35), ylim = c(35, 65))

gridExtra::grid.arrange(p1, p2, ncol = 2)

## ----heatmap-extent, fig.width=8, fig.height=6--------------------------------
hexify_heatmap(
  obs_hex,
  value = "count",
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "Zoomed to Region",
  legend_title = "Count"
)

## ----heatmap-basemap, fig.width=8, fig.height=6-------------------------------
# With world basemap (default)
hexify_heatmap(
  obs_hex,
  value = "count",
  basemap = "world",
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "With World Basemap"
)

## ----heatmap-no-basemap, fig.width=8, fig.height=6----------------------------
# Without basemap
hexify_heatmap(
  obs_hex,
  value = "count",
  basemap = NULL,
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "No Basemap"
)

## ----plot-world, fig.width=8, fig.height=4------------------------------------
plot_world(fill = "lightgray", border = "gray50")

## ----plot-world-custom, fig.width=8, fig.height=4-----------------------------
plot_world(
  fill = "antiquewhite",
  border = "sienna",
  xlim = c(-30, 50),
  ylim = c(30, 70)
)

## ----pentagon-locations, fig.width=8, fig.height=4----------------------------
# Pentagon locations (icosahedron vertices in standard ISEA orientation)
pentagon_coords <- data.frame(
  type = c("Pole", "Pole", rep("Vertex", 10)),
  lon = c(0, 0, seq(0, 324, by = 36)),
  lat = c(90, -90, rep(c(26.57, -26.57), 5))
)

# Assign to grid and get polygons
grid <- hex_grid(area_km2 = 500000)
pentagon_cells <- lonlat_to_cell(pentagon_coords$lon, pentagon_coords$lat, grid)

pentagon_polys <- cell_to_sf(pentagon_cells, grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = pentagon_polys, fill = alpha("purple", 0.6),
          color = "purple", linewidth = 0.8) +
  labs(
    title = "Pentagon Cell Locations",
    subtitle = "12 pentagonal cells at icosahedron vertices (area = 5/6 of hexagons)"
  ) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----random-sample, fig.width=8, fig.height=4---------------------------------
# Grid parameters (coarse for faster build)
grid <- hex_grid(area_km2 = 200000, aperture = 3)
max_cell <- 10 * (3^grid@resolution) + 2

# Sample random cell IDs
set.seed(123)
N <- 50
random_cells <- sample(1:max_cell, N, replace = FALSE)

# Generate polygons for sampled cells
sample_polys <- cell_to_sf(random_cells, grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = sample_polys, fill = alpha("forestgreen", 0.5),
          color = "darkgreen", linewidth = 0.4) +
  labs(title = sprintf("Random Sample of %d Cells (~%.0f km2 each)", N, grid@area_km2)) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----custom-viz, fig.width=8, fig.height=6------------------------------------
# Create data with a numeric variable
set.seed(456)
stations <- data.frame(
  lon = runif(50, -10, 30),
  lat = runif(50, 35, 60),
  temperature = rnorm(50, mean = 15, sd = 5)
)

# Hexify (coarser grid for faster build)
grid <- hex_grid(area_km2 = 20000)
stations_hex <- hexify(stations, lon = "lon", lat = "lat", grid = grid)

# Aggregate temperature by cell
stations_df <- as.data.frame(stations_hex)
stations_df$cell_id <- stations_hex@cell_id
cell_temps <- aggregate(temperature ~ cell_id, data = stations_df, FUN = mean)

# Generate polygons and merge data
cell_polys <- cell_to_sf(cell_temps$cell_id, grid)
cell_polys <- merge(cell_polys, cell_temps, by = "cell_id")

# Custom visualization
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = temperature),
          color = "white", linewidth = 0.2) +
  scale_fill_gradient2(
    low = "blue", mid = "white", high = "red",
    midpoint = 15, name = "Temp (°C)"
  ) +
  coord_sf(xlim = c(-10, 30), ylim = c(35, 60)) +
  labs(
    title = "Mean Temperature by Grid Cell",
    subtitle = "Diverging color scale centered at 15°C"
  ) +
  theme_minimal() +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )

# inst/scripts/compare_dggrid_hexify.R
suppressPackageStartupMessages({
  library(hexify)
  library(dggridR)
  library(dplyr)
  library(tidyr)
  library(readr)
  library(tibble)
})

# ---- config ---------------------------------------------------------------

# Sample lon/lat points (well inside faces; EU cities)
pts <- tibble(
  id  = paste0("p", 1:10),
  lon = c(16.3738,  6.1296, 14.4378, 11.5820,  8.5417,
          2.3522, 12.4964, -3.7038, 12.5683, 21.0122),
  lat = c(48.2082, 49.6116, 50.0755, 48.1351, 47.3769,
          48.8566, 41.9028, 40.4168, 55.6761, 52.2297)
)

areas_km2 <- c(863, 2591, 7774, 23322)
WRITE_CSV  <- FALSE

# ---- helpers --------------------------------------------------------------

# Haversine distance in meters (inputs in degrees)
hav_m <- function(lon1, lat1, lon2, lat2) {
  to_rad <- pi/180
  φ1 <- lat1 * to_rad; φ2 <- lat2 * to_rad
  dφ <- (lat2 - lat1) * to_rad
  dλ <- (lon2 - lon1) * to_rad
  a <- sin(dφ/2)^2 + cos(φ1) * cos(φ2) * sin(dλ/2)^2
  2 * 6371008.8 * asin(pmin(1, sqrt(a)))  # Earth mean radius
}

# hexify helper: Snyder → quantize Z3 at eff_res → hex center → inverse
# IMPORTANT: flip_classes = TRUE to match ISEA3H (dggridR).
hexify_center_ll <- function(lon_deg, lat_deg, eff_res) {
  fwd  <- cpp_snyder_forward(lon_deg, lat_deg)
  face <- as.integer(fwd[["face"]])
  tx   <- as.numeric(fwd[["tx"]])
  ty   <- as.numeric(fwd[["ty"]])

  digs <- cpp_hex_index_z3_quantize_digits(tx, ty, eff_res,
                                           center_thr = 0.4,
                                           flip_classes = TRUE)$digits
  cen  <- cpp_hex_index_z3_center(digs, flip_classes = TRUE)
  ll   <- hex_face_xy_to_ll(cen[["cx"]], cen[["cy"]], face)

  c(lon = as.numeric(ll[["lon"]]), lat = as.numeric(ll[["lat"]]))
}

# optional: tighten inverse precision (if exposed)
if ("hex_snyder_set_precision" %in% getNamespaceExports("hexify")) {
  hex_snyder_set_precision("ultra")
}

# ---- compare across areas -------------------------------------------------

all_comparisons <- vector("list", length(areas_km2))
names(all_comparisons) <- as.character(areas_km2)

for (area in areas_km2) {
  grid <- dgconstruct(area = area, topology = "HEXAGON", metric = TRUE, resround = "nearest")
  eff_res <- grid$res

  cat(sprintf(
    "Resolution: %d, Area (km^2): %.12f, Spacing (km): %.10f, CLS (km): %.10f\n",
    eff_res, grid$Area, grid$Spacing, grid$CLS
  ))

  # dggridR centers (coerce list -> tibble)
  seqnum <- dgGEO_to_SEQNUM(grid, pts$lon, pts$lat)$seqnum
  centers_list <- dgSEQNUM_to_GEO(grid, seqnum)  # list with lon_deg, lat_deg
  dgg_df <- tibble(
    id      = pts$id,
    dgg_lon = as.numeric(centers_list$lon_deg),
    dgg_lat = as.numeric(centers_list$lat_deg)
  )

  # hexify centers at same effective resolution
  hex_mat <- t(mapply(function(lo, la) hexify_center_ll(lo, la, eff_res), pts$lon, pts$lat))
  hex_df <- as_tibble(hex_mat) |>
    mutate(id = pts$id, .before = 1L) |>
    rename(hex_lon = lon, hex_lat = lat)

  # join and compute spherical distance error (meters)
  cmp <- dgg_df |>
    left_join(hex_df, by = "id") |>
    mutate(
      area_km2 = area,
      gc_err_m = hav_m(dgg_lon, dgg_lat, hex_lon, hex_lat)
    )

  all_comparisons[[as.character(area)]] <- cmp
}

cmp_all <- bind_rows(all_comparisons)

summary_tbl <- cmp_all |>
  group_by(area_km2) |>
  summarize(
    n           = n(),
    max_err_m   = max(gc_err_m, na.rm = TRUE),
    mean_err_m  = mean(gc_err_m, na.rm = TRUE),
    median_err_m= median(gc_err_m, na.rm = TRUE),
    .groups = "drop"
  )

cat("\nGreat-circle error summary (hexify vs dggridR centers):\n")
print(summary_tbl, n = Inf)

if (isTRUE(WRITE_CSV)) {
  write_csv(cmp_all,    "compare_hexify_dggrid_points.csv")
  write_csv(summary_tbl,"compare_hexify_dggrid_summary.csv")
  cat("\nWrote: compare_hexify_dggrid_points.csv and compare_hexify_dggrid_summary.csv\n")
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
library(hexify)

## ----cover-image, echo=FALSE, message=FALSE, warning=FALSE, fig.width=8, fig.height=4, fig.align='center'----
# Generate cover image: multi-resolution grids over different regions
library(sf)
library(ggplot2)

# Get world basemap
countries <- hexify_world

# Generate grids at different resolutions for specific regions (coarser for speed)
# Coarse grid: South America
grid_sa <- hex_grid(area_km2 = 300000)
sa_hexes <- grid_rect(c(-80, -40, -35, 10), grid_sa)

# Medium grid: Europe/Africa
grid_eu <- hex_grid(area_km2 = 80000)
eu_hexes <- grid_rect(c(-10, -10, 40, 50), grid_eu)

# Fine grid: East Asia
grid_asia <- hex_grid(area_km2 = 50000)
asia_hexes <- grid_rect(c(90, 10, 140, 50), grid_asia)

ggplot() +
  geom_sf(data = countries, fill = "gray95", color = "gray80", linewidth = 0.15) +
  geom_sf(data = sa_hexes, fill = NA, color = "#1B9E77", linewidth = 0.5) +
  geom_sf(data = eu_hexes, fill = NA, color = "#D95F02", linewidth = 0.4) +
  geom_sf(data = asia_hexes, fill = NA, color = "#7570B3", linewidth = 0.3) +
  coord_sf(xlim = c(-100, 150), ylim = c(-50, 60)) +
  theme_void() +
  theme(panel.background = element_rect(fill = "white", color = NA))

## ----basic-usage--------------------------------------------------------------
# Sample data: European cities
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90)
)

# Create a grid specification
grid <- hex_grid(area_km2 = 10000)
grid

# Assign cities to hexagonal cells
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)
result

## ----hexdata-access-----------------------------------------------------------
# Get the grid specification
grid_info(result)

# Get unique cell IDs
cells(result)

# Count unique cells
n_cells(result)

# Access all cell IDs (one per row)
result@cell_id

# Access cell centers
head(result@cell_center)

# Extract original data as data.frame
head(as.data.frame(result))

## ----sf-usage, eval = requireNamespace("sf", quietly = TRUE)------------------
library(sf)

# Create sf object (any CRS works - hexify transforms automatically)
pts <- st_as_sf(cities, coords = c("lon", "lat"), crs = 4326)

# hexify handles CRS transformation automatically
result_sf <- hexify(pts, area_km2 = 10000)
class(result_sf)

## ----bird-setup, message=FALSE, warning=FALSE---------------------------------
library(sf)
library(ggplot2)

# Simulate bird observation data (reduced for faster vignette build)
set.seed(123)
n_obs <- 500

# Generate observations with realistic spatial clustering
birds <- data.frame(
  lon = c(
    rnorm(150, mean = 10, sd = 15),    # Western Europe
    rnorm(100, mean = 25, sd = 10),    # Eastern Europe
    rnorm(100, mean = 20, sd = 20),    # Mediterranean
    rnorm(80, mean = 0, sd = 15),      # West Africa
    rnorm(70, mean = 35, sd = 10)      # East Africa
  ),
  lat = c(
    rnorm(150, mean = 50, sd = 8),     # Western Europe
    rnorm(100, mean = 55, sd = 6),     # Eastern Europe
    rnorm(100, mean = 42, sd = 5),     # Mediterranean
    rnorm(80, mean = 10, sd = 10),     # West Africa
    rnorm(70, mean = -5, sd = 12)      # East Africa
  ),
  species = sample(c("Passer domesticus", "Turdus merula", "Parus major",
                     "Columba palumbus", "Sturnus vulgaris"), n_obs, replace = TRUE)
)

# Clip to valid ranges
birds$lon <- pmax(-30, pmin(60, birds$lon))
birds$lat <- pmax(-35, pmin(70, birds$lat))

## ----bird-assign--------------------------------------------------------------
# Create grid and assign observations (coarser grid for faster build)
grid <- hex_grid(area_km2 = 50000)
birds_hex <- hexify(birds, lon = "lon", lat = "lat", grid = grid)

# Extract data with cell IDs
birds_gridded <- as.data.frame(birds_hex)
birds_gridded$cell_id <- birds_hex@cell_id

# Count observations per cell
obs_counts <- aggregate(
  species ~ cell_id,
  data = birds_gridded,
  FUN = length
)
names(obs_counts)[2] <- "n_observations"

# Species richness per cell
richness <- aggregate(
  species ~ cell_id,
  data = birds_gridded,
  FUN = function(x) length(unique(x))
)
names(richness)[2] <- "n_species"

obs_counts <- merge(obs_counts, richness, by = "cell_id")
head(obs_counts)

## ----bird-plot, fig.width=8, fig.height=6, message=FALSE, warning=FALSE-------
# Generate polygons for cells with data
cell_polys <- cell_to_sf(obs_counts$cell_id, grid)
cell_polys <- merge(cell_polys, obs_counts, by = "cell_id")

# Get relevant countries
region <- hexify_world[hexify_world$continent %in% c("Europe", "Africa"), ]

ggplot() +
  geom_sf(data = region, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = n_observations), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "plasma", name = "Observations", trans = "sqrt") +
  coord_sf(xlim = c(-30, 60), ylim = c(-35, 70)) +
  labs(
    title = "Bird Observations in Equal-Area Hexagonal Cells",
    subtitle = sprintf("ISEA3H grid at resolution %d (~%.0f km² cells)",
                       grid@resolution, grid@area_km2)
  ) +
  theme_minimal() +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )

## ----richness-plot, fig.width=8, fig.height=6, message=FALSE------------------
ggplot() +
  geom_sf(data = region, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = n_species), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "mako", name = "Species\nRichness", direction = -1) +
  coord_sf(xlim = c(-30, 60), ylim = c(-35, 70)) +
  labs(
    title = "Species Richness per Grid Cell",
    subtitle = "Number of unique species observed in each equal-area cell"
  ) +
  theme_minimal() +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )

## ----setup, include = FALSE---------------------------------------------------
CAN_RUN <- requireNamespace("sf", quietly = TRUE)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  eval = CAN_RUN
)
library(hexify)

## ----h3-grid------------------------------------------------------------------
library(sf)
library(ggplot2)

# Create an H3 grid specification
grid_h3 <- hex_grid(resolution = 5, type = "h3")
grid_h3

## ----h3-hexify----------------------------------------------------------------
# Sample cities
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome",
           "London", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50,
          -0.12, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90,
          51.51, 50.08, 52.23, 47.50, 52.37)
)

result <- hexify(cities, lon = "lon", lat = "lat", grid = grid_h3)
result

## ----h3-cell-ids--------------------------------------------------------------
# Cell IDs are hexadecimal strings
result@cell_id

# All standard accessors work
cells(result)
n_cells(result)

## ----h3-by-area---------------------------------------------------------------
grid_area <- hex_grid(area_km2 = 500, type = "h3")
grid_area

## ----h3-grid-rect, fig.width=7, fig.height=5----------------------------------
# Generate H3 hexagons over Western Europe
grid_h3 <- hex_grid(resolution = 3, type = "h3")
europe_h3 <- grid_rect(c(-10, 35, 25, 60), grid_h3)

# Basemap
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_h3, fill = NA, color = "#E6550D", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("H3 Resolution %d Grid (~%.0f km² avg cells)",
                       grid_h3@resolution, grid_h3@area_km2)) +
  theme_minimal()

## ----h3-grid-clip, fig.width=6, fig.height=6----------------------------------
# Clip H3 grid to France
france <- hexify_world[hexify_world$name == "France", ]
grid_h3 <- hex_grid(resolution = 4, type = "h3")
france_h3 <- grid_clip(france, grid_h3)

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_h3, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = sprintf("H3 Grid Clipped to France (res %d)", grid_h3@resolution)) +
  theme_minimal()

## ----h3-parent----------------------------------------------------------------
# Get parent cells (one resolution coarser)
grid_h3 <- hex_grid(resolution = 5, type = "h3")
child_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40),
  lat = c(48.21, 48.86, 52.52),
  grid = grid_h3
)

parent_ids <- get_parent(child_ids, grid_h3, levels = 1)
data.frame(child = child_ids, parent = parent_ids)

## ----h3-children--------------------------------------------------------------
# Get children of a single cell (one resolution finer)
grid_coarse <- hex_grid(resolution = 3, type = "h3")
coarse_id <- lonlat_to_cell(16.37, 48.21, grid_coarse)

children <- get_children(coarse_id, grid_coarse, levels = 1)
cat(length(children[[1]]), "children at resolution", grid_coarse@resolution + 1, "\n")
head(children[[1]])

## ----h3-hierarchy-plot, fig.width=7, fig.height=5-----------------------------
# Parent cell polygon
parent_poly <- cell_to_sf(coarse_id, grid_coarse)

# Children cell polygons
grid_fine <- hex_grid(resolution = 4, type = "h3")
children_poly <- cell_to_sf(children[[1]], grid_fine)

ggplot() +
  geom_sf(data = children_poly, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.5) +
  geom_sf(data = parent_poly, fill = NA, color = "black", linewidth = 1.2) +
  labs(title = sprintf("H3 Hierarchy: 1 parent (res %d) → %d children (res %d)",
                       grid_coarse@resolution,
                       length(children[[1]]),
                       grid_fine@resolution)) +
  theme_minimal()

## ----h3-workflow, fig.width=8, fig.height=6, message=FALSE, warning=FALSE-----
set.seed(42)

# Simulate observations across Europe
obs <- data.frame(
  lon = c(rnorm(200, 10, 12), rnorm(100, 25, 8)),
  lat = c(rnorm(200, 48, 6), rnorm(100, 55, 4)),
  species = sample(c("Sp. A", "Sp. B", "Sp. C"), 300, replace = TRUE)
)
obs$lon <- pmax(-10, pmin(40, obs$lon))
obs$lat <- pmax(35, pmin(65, obs$lat))

# Hexify with H3
grid_h3 <- hex_grid(resolution = 3, type = "h3")
obs_hex <- hexify(obs, lon = "lon", lat = "lat", grid = grid_h3)

# Aggregate: species richness per cell
obs_df <- as.data.frame(obs_hex)
obs_df$cell_id <- obs_hex@cell_id

richness <- aggregate(species ~ cell_id, data = obs_df,
                      FUN = function(x) length(unique(x)))
names(richness)[2] <- "n_species"

# Map it
polys <- cell_to_sf(richness$cell_id, grid_h3)
polys <- merge(polys, richness, by = "cell_id")

europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = polys, aes(fill = n_species), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "plasma", name = "Species\nRichness") +
  coord_sf(xlim = c(-10, 40), ylim = c(35, 65)) +
  labs(title = "Species Richness on H3 Grid",
       subtitle = sprintf("H3 resolution %d (~%.0f km² avg cells)",
                          grid_h3@resolution, grid_h3@area_km2)) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----h3-crosswalk-------------------------------------------------------------
# Start with an ISEA grid and some cells
grid_isea <- hex_grid(resolution = 9, aperture = 3)
isea_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40, -3.70, 12.50),
  lat = c(48.21, 48.86, 52.52, 40.42, 41.90),
  grid = grid_isea
)

# Map ISEA cells to their closest H3 equivalents
xw <- h3_crosswalk(isea_ids, grid_isea)
xw[, c("isea_cell_id", "h3_cell_id", "isea_area_km2", "h3_area_km2")]

## ----h3-resolution-table------------------------------------------------------
h3_res <- hexify_compare_resolutions(type = "h3", res_range = 0:15)
h3_res$n_cells_fmt <- ifelse(
  h3_res$n_cells > 1e9,
  sprintf("%.1fB", h3_res$n_cells / 1e9),
  ifelse(h3_res$n_cells > 1e6,
         sprintf("%.1fM", h3_res$n_cells / 1e6),
         ifelse(h3_res$n_cells > 1e3,
                sprintf("%.1fK", h3_res$n_cells / 1e3),
                as.character(h3_res$n_cells)))
)
knitr::kable(
  h3_res[, c("resolution", "n_cells_fmt", "cell_area_km2", "cell_spacing_km")],
  col.names = c("Resolution", "# Cells", "Avg Area (km²)", "Spacing (km)"),
  digits = 1
)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
library(hexify)
library(sf)
library(ggplot2)

## ----one-grid-many-datasets---------------------------------------------------
# Step 1: Define the grid once
# This is your shared spatial reference system
grid <- hex_grid(area_km2 = 5000)
print(grid)

# Step 2: Create multiple datasets with different structures
set.seed(123)

# Dataset 1: Bird observations (note different column names)
bird_obs <- data.frame(
  species = sample(c("Passer domesticus", "Turdus merula", "Parus major"), 50, replace = TRUE),
  longitude = runif(50, -10, 30),
  latitude = runif(50, 35, 60)
)

# Dataset 2: Mammal records
mammal_obs <- data.frame(
  species = sample(c("Vulpes vulpes", "Meles meles", "Sciurus vulgaris"), 40, replace = TRUE),
  lon = runif(40, -10, 30),
  lat = runif(40, 35, 60)
)

# Dataset 3: Climate stations
climate_data <- data.frame(
  station_id = paste0("WS", 1:20),
  x = runif(20, -10, 30),
  y = runif(20, 35, 60),
  temp_c = rnorm(20, 12, 5)
)

# Step 3: Attach all datasets to the SAME grid
birds <- hexify(bird_obs, lon = "longitude", lat = "latitude", grid = grid)
mammals <- hexify(mammal_obs, lon = "lon", lat = "lat", grid = grid)
climate <- hexify(climate_data, lon = "x", lat = "y", grid = grid)

cat("Birds:  ", nrow(as.data.frame(birds)), "observations in", n_cells(birds), "cells\n")
cat("Mammals:", nrow(as.data.frame(mammals)), "observations in", n_cells(mammals), "cells\n")
cat("Climate:", nrow(as.data.frame(climate)), "stations in", n_cells(climate), "cells\n")

## ----cell-level-analysis------------------------------------------------------
# Extract data frames with cell IDs
birds_df <- as.data.frame(birds)
birds_df$cell_id <- birds@cell_id

mammals_df <- as.data.frame(mammals)
mammals_df$cell_id <- mammals@cell_id

climate_df <- as.data.frame(climate)
climate_df$cell_id <- climate@cell_id

# Aggregate each dataset by cell
bird_richness <- aggregate(
  species ~ cell_id,
  data = birds_df,
  FUN = function(x) length(unique(x))
)
names(bird_richness)[2] <- "bird_species"

mammal_richness <- aggregate(
  species ~ cell_id,
  data = mammals_df,
  FUN = function(x) length(unique(x))
)
names(mammal_richness)[2] <- "mammal_species"

mean_temp <- aggregate(
  temp_c ~ cell_id,
  data = climate_df,
  FUN = mean
)
names(mean_temp)[2] <- "mean_temp"

# Join datasets by cell_id - guaranteed to align because same grid
combined <- merge(bird_richness, mammal_richness, by = "cell_id", all = TRUE)
combined <- merge(combined, mean_temp, by = "cell_id", all = TRUE)

head(combined)

## ----grid-rect, fig.width=7, fig.height=5-------------------------------------
# Create grid specification
grid <- hex_grid(area_km2 = 5000)

# Generate hexagons over Western Europe
europe_hexes <- grid_rect(c(-10, 35, 25, 60), grid)

# Get European countries for context
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_hexes, fill = NA, color = "steelblue", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal()

## ----grid-polygon, fig.width=6, fig.height=6----------------------------------
# Get France boundary
france <- hexify_world[hexify_world$name == "France", ]

# Generate grid covering mainland France
grid <- hex_grid(area_km2 = 2000)
france_grid <- grid_rect(c(-5, 41, 10, 52), grid)

# Clip grid to France boundary
france_grid_clipped <- st_intersection(france_grid, st_geometry(france))

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_grid_clipped, fill = alpha("steelblue", 0.3),
          color = "steelblue", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = sprintf("Hexagonal Grid Clipped to France (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal()

## ----grid-global, fig.width=8, fig.height=4, warning=FALSE--------------------
# Coarse global grid (be careful with fine grids - many cells!)
grid <- hex_grid(area_km2 = 500000)
global_hexes <- grid_global(grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray90", color = "gray70", linewidth = 0.2) +
  geom_sf(data = global_hexes, fill = NA, color = "darkgreen", linewidth = 0.3) +
  labs(title = sprintf("Global Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----multi-resolution---------------------------------------------------------
# Sample data
set.seed(42)
observations <- data.frame(
  species = sample(c("Species A", "Species B", "Species C"), 100, replace = TRUE),
  lon = runif(100, -10, 30),
  lat = runif(100, 35, 60)
)

# Fine resolution (~1000 km² cells)
grid_fine <- hex_grid(area_km2 = 1000)
obs_fine <- hexify(observations, lon = "lon", lat = "lat", grid = grid_fine)

# Coarse resolution (~10000 km² cells)
grid_coarse <- hex_grid(area_km2 = 10000)
obs_coarse <- hexify(observations, lon = "lon", lat = "lat", grid = grid_coarse)

cat(sprintf("Fine resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_fine), grid_fine@area_km2))
cat(sprintf("Coarse resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_coarse), grid_coarse@area_km2))

## ----multi-scale-aggregate----------------------------------------------------
# Extract data with cell IDs
fine_df <- as.data.frame(obs_fine)
fine_df$cell_id <- obs_fine@cell_id

coarse_df <- as.data.frame(obs_coarse)
coarse_df$cell_id <- obs_coarse@cell_id

# Species richness at each scale
richness_fine <- aggregate(species ~ cell_id, data = fine_df,
                           FUN = function(x) length(unique(x)))
richness_coarse <- aggregate(species ~ cell_id, data = coarse_df,
                             FUN = function(x) length(unique(x)))

cat(sprintf("Fine scale: mean %.2f species per cell\n", mean(richness_fine$species)))
cat(sprintf("Coarse scale: mean %.2f species per cell\n", mean(richness_coarse$species)))

## ----spatial-joins------------------------------------------------------------
# Dataset 1: Weather stations
stations <- data.frame(
  station_id = paste0("ST", 1:50),
  lon = runif(50, -10, 30),
  lat = runif(50, 35, 60),
  temperature = rnorm(50, 15, 5)
)

# Dataset 2: Cities
cities <- data.frame(
  city = c("Vienna", "Paris", "London", "Berlin", "Rome",
           "Madrid", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -0.12, 13.40, 12.50,
          -3.70, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 51.51, 52.52, 41.90,
          40.42, 50.08, 52.23, 47.50, 52.37)
)

# Use a coarse grid for joining disparate points
grid <- hex_grid(area_km2 = 50000)

# Hexify both datasets with the same grid
stations_hex <- hexify(stations, lon = "lon", lat = "lat", grid = grid)
cities_hex <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Extract with cell IDs
stations_df <- as.data.frame(stations_hex)
stations_df$cell_id <- stations_hex@cell_id

cities_df <- as.data.frame(cities_hex)
cities_df$cell_id <- cities_hex@cell_id

# Join by cell_id
city_weather <- merge(
  cities_df[, c("city", "cell_id")],
  aggregate(temperature ~ cell_id, data = stations_df, FUN = mean),
  by = "cell_id",
  all.x = TRUE
)

city_weather

## ----choosing-resolution------------------------------------------------------
# Target: 100 km² cells
grid_100 <- hex_grid(area_km2 = 100, aperture = 3)
cat(sprintf("Target ~100 km²: resolution %d (actual: %.1f km²)\n",
            grid_100@resolution, grid_100@area_km2))

# Target: 1000 km² cells
grid_1000 <- hex_grid(area_km2 = 1000, aperture = 3)
cat(sprintf("Target ~1000 km²: resolution %d (actual: %.1f km²)\n",
            grid_1000@resolution, grid_1000@area_km2))

# Target: 10000 km² cells
grid_10000 <- hex_grid(area_km2 = 10000, aperture = 3)
cat(sprintf("Target ~10000 km²: resolution %d (actual: %.1f km²)\n",
            grid_10000@resolution, grid_10000@area_km2))

## ----resolution-table, echo=FALSE---------------------------------------------
comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:15)
comparison$n_cells_fmt <- ifelse(
  comparison$n_cells > 1e6,
  sprintf("%.1fM", comparison$n_cells / 1e6),
  ifelse(comparison$n_cells > 1e3,
         sprintf("%.1fK", comparison$n_cells / 1e3),
         as.character(comparison$n_cells))
)
knitr::kable(
  comparison[, c("resolution", "n_cells_fmt", "cell_area_km2", "cell_spacing_km")],
  col.names = c("Resolution", "# Cells", "Cell Area (km²)", "Spacing (km)"),
  digits = 1
)

## ----compare-apertures--------------------------------------------------------
target_area <- 1000  # km²

cat(sprintf("Target: ~%d km² cells\n\n", target_area))
for (ap in c(3, 4, 7)) {
  grid <- hex_grid(area_km2 = target_area, aperture = ap)
  n_cells <- 10 * (ap^grid@resolution) + 2
  cat(sprintf("Aperture %d: resolution %d -> %.1f km² (%s cells globally)\n",
              ap, grid@resolution, grid@area_km2,
              format(n_cells, big.mark = ",")))
}

## ----sf-workflow--------------------------------------------------------------
# Hexify some data
grid <- hex_grid(area_km2 = 20000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Convert to sf points (uses cell centers)
sf_points <- as_sf(result, geometry = "point")
class(sf_points)

# Convert to sf polygons (for choropleth maps)
sf_polys <- as_sf(result, geometry = "polygon")
class(sf_polys)

# Or generate polygons directly from cell IDs
unique_cells <- cells(result)
cell_polys <- cell_to_sf(unique_cells, grid)

## ----sf-plot, fig.width=7, fig.height=5---------------------------------------
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "ivory", color = "gray70") +
  geom_sf(data = cell_polys, fill = "steelblue", alpha = 0.5, color = "darkblue") +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 58)) +
  labs(title = "European Cities - Hexagonal Grid") +
  theme_minimal()

## ----export-formats, eval=FALSE-----------------------------------------------
# # Generate a grid
# grid <- hex_grid(area_km2 = 10000)
# europe <- grid_rect(c(-10, 35, 25, 60), grid)
# 
# # Export to various formats
# st_write(europe, "europe_grid.gpkg", layer = "hexgrid")     # GeoPackage
# st_write(europe, "europe_grid.shp")                         # Shapefile
# st_write(europe, "europe_grid.geojson")                     # GeoJSON
# st_write(europe, "europe_grid.kml", layer = "hexgrid")      # KML (Google Earth)

## ----edge-cases---------------------------------------------------------------
data_with_na <- data.frame(
  lon = c(16.37, NA, 2.35, 13.40),
  lat = c(48.21, 48.86, NA, 52.52)
)

grid <- hex_grid(area_km2 = 1000)
result <- hexify(data_with_na, lon = "lon", lat = "lat", grid = grid)

# Check which rows have valid cell assignments
cat("Cell IDs:", result@cell_id, "\n")
cat("NA indicates invalid coordinates\n")

# Run this script to regenerate test cache
# Usage: Rscript generate_cache.R

library(hexify)

cache <- list()

# Basic hexify results
df <- data.frame(lon = c(0, 10, 16.37, 2.35, -3.70), lat = c(45, 50, 48.21, 48.86, 40.42))
result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
cache$hexify_basic <- list(
  cell_id = result@cell_id,
  cell_cen_lon = result@cell_center[, "lon"],
  cell_cen_lat = result@cell_center[, "lat"],
  resolution = result@grid@resolution,
  aperture = result@grid@aperture
)

# Grid info
grid <- hex_grid(area_km2 = 1000)
cache$grid_1000 <- list(
  resolution = grid@resolution,
  aperture = grid@aperture,
  area_km2 = grid@area_km2,
  diagonal_km = grid@diagonal_km
)

# Cell conversions for aperture 3
cells3 <- hexify_lonlat_to_cell(c(0, 10, -30), c(45, 50, -15), resolution = 5, aperture = 3)
coords3 <- hexify_cell_to_lonlat(cells3, resolution = 5, aperture = 3)
cache$cells_ap3 <- list(cells = cells3, lon = coords3$lon_deg, lat = coords3$lat_deg)

# Cell conversions for aperture 4
cells4 <- hexify_lonlat_to_cell(c(0, 10, -30), c(45, 50, -15), resolution = 5, aperture = 4)
coords4 <- hexify_cell_to_lonlat(cells4, resolution = 5, aperture = 4)
cache$cells_ap4 <- list(cells = cells4, lon = coords4$lon_deg, lat = coords4$lat_deg)

# Cell conversions for aperture 7
cells7 <- hexify_lonlat_to_cell(c(0, 10, -30), c(45, 50, -15), resolution = 3, aperture = 7)
coords7 <- hexify_cell_to_lonlat(cells7, resolution = 3, aperture = 7)
cache$cells_ap7 <- list(cells = cells7, lon = coords7$lon_deg, lat = coords7$lat_deg)

# Projection forward
proj <- hexify:::cpp_snyder_forward(10, 45)
cache$proj_forward <- as.list(proj)

# Face centers
cache$face_centers <- hexify_face_centers()

# Index conversions z3
cache$index_z3_sample <- hexify_cell_to_index(0, 5, 4, 3, 3, "z3")

# Quad IJ round-trip
qij <- hexify_cell_to_quad_ij(c(100, 1000, 5000), resolution = 10, aperture = 3)
cache$quad_ij <- qij

# Save
saveRDS(cache, "tests/testthat/fixtures/test_cache.rds")
cat("Cache saved with", length(cache), "entries\n")

#' hexify
#'
#' Core icosahedron and 'Snyder' projection helpers.
#'
#' @useDynLib hexify, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @importFrom methods new setClass setMethod setGeneric setValidity is as
#' @importFrom grDevices adjustcolor
#' @importFrom graphics points polygon
#' @importFrom utils head
#' @importFrom rlang .data
"_PACKAGE"

# Global variables to avoid R CMD check notes
utils::globalVariables(c(
  "hexify_world", "cpp_decode_z3", "cpp_decode_zorder",
  "cpp_h3_latLngToCell", "cpp_h3_cellToLatLng", "cpp_h3_isValidCell",
  "cpp_h3_cellToParent", "cpp_h3_cellToChildren", "cpp_h3_cellToBoundary",
  "cpp_h3_polygonToCells", "cpp_h3_cellAreaKm2", "cpp_h3_getResolution",
  "cpp_h3_gridDisk", "cpp_h3_gridDiskDistances", "cpp_h3_gridRingUnsafe",
  "cpp_h3_compactCells", "cpp_h3_uncompactCells",
  "cpp_h3_isPentagon", "cpp_h3_gridDistance",
  "cpp_get_neighbors_isea", "cpp_get_neighbors_z7"
))

#' @importFrom sf st_crs
#' @export
sf::st_crs

# =============================================================================
# POINT ASSIGNMENT (ISEA3H, APERTURE 3)
# =============================================================================

#' Assign hex cells ('ISEA3H', aperture 3) for lon/lat
#'
#' Forward projection -> quantize -> cell centre -> inverse to lon/lat, on the
#' aperture-3 pipeline used by [hexify()]: points fold into the non-negative
#' quad frame ([lonlat_to_cell()]), and the centre comes back from the cell
#' itself ([cell_to_lonlat()]), so the returned centre is the centre of the
#' assigned cell.
#'
#' @param lon,lat numeric vectors (same length), degrees.
#' @param effective_res integer effective resolution (>= 1).
#' @param make_polygons logical; if TRUE, return an sf with hex polygons.
#' @return data.frame with id (Z3 index string), face (quad, 0-11),
#'         effective_res, center_lon, center_lat; if make_polygons=TRUE,
#'         an sf with geometry column.
#'
#' @keywords internal
#' @export
hexify_assign <- function(lon, lat, effective_res, make_polygons = FALSE) {

  stopifnot(
    length(lon) == length(lat),
    length(effective_res) == 1L,
    effective_res >= 1L
  )

  lon <- as.numeric(lon)
  lat <- as.numeric(lat)
  res <- as.integer(effective_res)

  cell_id <- cpp_lonlat_to_cell(lon, lat, res, 3L)
  center <- cpp_cell_to_lonlat(cell_id, res, 3L)
  quad_ij <- cpp_cell_to_quad_ij(cell_id, res, 3L)

  df <- data.frame(
    id = vapply(seq_along(cell_id), function(k) {
      cpp_cell_to_index(quad_ij$quad[k], quad_ij$i[k], quad_ij$j[k],
                        res, 3L, "z3")
    }, character(1)),
    face = quad_ij$quad,
    effective_res = res,
    center_lon = center$lon_deg,
    center_lat = center$lat_deg,
    stringsAsFactors = FALSE
  )

  if (!isTRUE(make_polygons)) return(df)

  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("make_polygons=TRUE requires the 'sf' package.")
  }

  sf::st_sf(df, geometry = isea_cells_to_sfc(cell_id, res, 3L))
}

# hexify_conversions.R
# Coordinate system conversion functions
#
# This file contains functions for converting between geographic coordinates
# (longitude/latitude) and hexagonal cell indices.

#' @title Coordinate Conversions
#' @description Functions for converting between coordinate systems
#' @name hexify-conversions
NULL

# Validation: use validate_aperture() / validate_resolution() from constants.R

#' Convert longitude/latitude to hexagonal cell hierarchical index
#'
#' Converts geographic coordinates (longitude, latitude) to hexagonal cell
#' hierarchical index strings. These strings encode the face, resolution, and
#' cell location in a Z-order (Morton code) format.
#'
#' @param grid Grid specification from hexify_grid()
#' @param lon Longitude vector in degrees (numeric, -180 to 180)
#' @param lat Latitude vector in degrees (numeric, -90 to 90)
#'
#' @return Data frame with columns:
#'   \item{h_index}{Hierarchical index (character string)}
#'   \item{face}{Icosahedron face number (integer, 0-19)}
#'
#' @details
#' Most users should use \code{\link{hexify_lonlat_to_cell}} or
#' \code{\link{hexify_grid_to_cell}} which return DGGRID-compatible
#' integer cell IDs.
#'
#' This function returns hierarchical index strings useful for:
#' - Understanding the cell's position in the hierarchy
#' - Prefix-based spatial queries
#' - Parent/child cell relationships
#'
#' @keywords internal
hexify_lonlat_to_h_index <- function(grid, lon, lat) {
  
  # Validate grid object
  if (!inherits(grid, "hexify_grid")) {
    stop("grid must be a hexify_grid object from hexify_grid()")
  }
  
  # Validate inputs
  if (length(lon) != length(lat)) {
    stop("lon and lat must have the same length")
  }
  
  if (!is.numeric(lon) || !is.numeric(lat)) {
    stop("lon and lat must be numeric vectors")
  }
  
  # Check for out-of-range coordinates
  if (any(!is.na(lon) & (lon < -180 | lon > 180))) {
    warning("Some longitude values are outside valid range [-180, 180]")
  }
  
  if (any(!is.na(lat) & (lat < -90 | lat > 90))) {
    warning("Some latitude values are outside valid range [-90, 90]")
  }
  
  n <- length(lon)
  
  # Preallocate result vectors
  cell_indices <- character(n)
  faces <- integer(n)
  
  # Process each point
  for (i in seq_along(lon)) {
    # Handle NA values
    if (is.na(lon[i]) || is.na(lat[i])) {
      cell_indices[i] <- NA_character_
      faces[i] <- NA_integer_
      next
    }
    
    # Call unified C++ conversion function
    # This function:
    # 1. Projects lon/lat to icosahedron face
    # 2. Converts to face plane coordinates
    # 3. Finds hex cell containing the point
    # 4. Encodes cell as index string
    cell_indices[i] <- cpp_lonlat_to_index(
      lon[i], lat[i], 
      grid$resolution,
      grid$aperture,
      grid$index_type
    )
    
    # The leading field of an index is not always the face on its own -- the
    # aperture-7 Z7 index packs its hierarchy seed in there too -- so read the
    # face back through the decoder that owns the format.
    faces[i] <- as.integer(
      cpp_index_to_cell(cell_indices[i], grid$aperture, grid$index_type)$face
    )
  }
  
  return(data.frame(
    h_index = cell_indices,
    face = faces
  ))
}

#' Convert hierarchical index strings to longitude/latitude centers
#'
#' Converts hierarchical cell index strings back to geographic coordinates,
#' returning the center point of each cell. This is the inverse operation
#' of hexify_lonlat_to_h_index().
#'
#' @param grid Grid specification from hexify_grid()
#' @param h_index Hierarchical index strings (character vector)
#'
#' @return Data frame with columns:
#'   \item{lon}{Longitude in degrees}
#'   \item{lat}{Latitude in degrees}
#'
#' @details
#' Most users should use \code{\link{hexify_cell_to_lonlat}} or
#' \code{\link{hexify_grid_cell_to_lonlat}} which work with DGGRID-compatible
#' integer cell IDs.
#'
#' @keywords internal
hexify_h_index_to_lonlat <- function(grid, h_index) {

  # Validate grid object
  if (!inherits(grid, "hexify_grid")) {
    stop("grid must be a hexify_grid object from hexify_grid()")
  }

  # Validate input
  if (!is.character(h_index)) {
    stop("h_index must be a character vector")
  }

  n <- length(h_index)

  # Preallocate result vectors
  lon <- numeric(n)
  lat <- numeric(n)
  
  # Process each cell index
  for (i in seq_along(h_index)) {
    # Handle NA values
    if (is.na(h_index[i])) {
      lon[i] <- NA_real_
      lat[i] <- NA_real_
      next
    }

    # Call C++ conversion function
    # This function:
    # 1. Decodes the index string to face, i, j, resolution
    # 2. Calculates hex center in face plane coordinates
    # 3. Projects back to lon/lat using inverse ISEA projection
    result <- cpp_index_to_lonlat(
      h_index[i],
      grid$aperture,
      grid$index_type
    )

    # Extract lon/lat from named vector
    lon[i] <- result["lon"]
    lat[i] <- result["lat"]
  }

  return(data.frame(
    lon = lon,
    lat = lat
  ))
}

# =============================================================================
# GRID-BASED WRAPPERS FOR CELL IDs (DGGRID-COMPATIBLE)
# =============================================================================

#' Convert longitude/latitude to cell ID using a grid object
#'
#' Grid-based wrapper for \code{\link{hexify_lonlat_to_cell}}. Converts
#' geographic coordinates to DGGRID-compatible cell IDs using
#' the resolution and aperture from a grid object.
#'
#' @param grid Grid specification from hexify_grid()
#' @param lon Numeric vector of longitudes in degrees
#' @param lat Numeric vector of latitudes in degrees
#'
#' @return Numeric vector of cell IDs (1-based)
#'
#' @family coordinate conversion
#' @seealso \code{\link{lonlat_to_cell}} for the recommended S4 interface,
#'   \code{\link{hexify_grid_cell_to_lonlat}} for the inverse operation
#'
#' @keywords internal
#' @export
#' @examples
#' grid <- hexify_grid(area = 1000, aperture = 3)
#' cell_ids <- hexify_grid_to_cell(grid, lon = c(0, 10), lat = c(45, 50))
hexify_grid_to_cell <- function(grid, lon, lat) {
  # Validate grid object

  if (!inherits(grid, "hexify_grid")) {
    stop("grid must be a hexify_grid object from hexify_grid()")
  }

  hexify_lonlat_to_cell(lon, lat, grid$resolution, grid$aperture)
}

#' Convert cell ID to longitude/latitude using a grid object
#'
#' Grid-based wrapper for \code{\link{hexify_cell_to_lonlat}}. Converts
#' DGGRID-compatible cell IDs back to cell center coordinates
#' using the resolution and aperture from a grid object.
#'
#' @param grid Grid specification from hexify_grid()
#' @param cell_id Numeric vector of cell IDs (1-based)
#'
#' @return Data frame with lon_deg and lat_deg columns
#'
#' @family coordinate conversion
#' @seealso \code{\link{cell_to_lonlat}} for the recommended S4 interface,
#'   \code{\link{hexify_grid_to_cell}} for the forward operation
#'
#' @keywords internal
#' @export
#' @examples
#' grid <- hexify_grid(area = 1000, aperture = 3)
#' cell_ids <- hexify_grid_to_cell(grid, lon = 5, lat = 45)
#' coords <- hexify_grid_cell_to_lonlat(grid, cell_ids)
hexify_grid_cell_to_lonlat <- function(grid, cell_id) {
  # Validate grid object
  if (!inherits(grid, "hexify_grid")) {
    stop("grid must be a hexify_grid object from hexify_grid()")
  }

  hexify_cell_to_lonlat(cell_id, grid$resolution, grid$aperture)
}

# =============================================================================
# INTERNAL HELPERS
# =============================================================================


#' Round-trip accuracy test
#' 
#' Tests the accuracy of the coordinate conversion functions by converting
#' coordinates to cells and back, measuring the distance between original
#' and reconstructed coordinates.
#' 
#' @param grid Grid specification
#' @param lon Longitude to test
#' @param lat Latitude to test
#' @param units Distance units ("km" or "degrees")
#' 
#' @return List with:
#'   \item{original}{Original coordinates}
#'   \item{cell}{Cell index}
#'   \item{reconstructed}{Reconstructed coordinates}
#'   \item{error}{Distance between original and reconstructed}
#'
#' @family coordinate conversion
#' @export
hexify_roundtrip_test <- function(grid, lon, lat, units = "km") {
  # Convert to hierarchical index
  result <- hexify_lonlat_to_h_index(grid, lon, lat)

  # Convert back to coordinates
  coords <- hexify_h_index_to_lonlat(grid, result$h_index)

  # Calculate distance
  if (units == "km") {
    # Haversine distance
    R <- grid_radius_km(grid)
    dlat <- (coords$lat - lat) * pi / 180
    dlon <- (coords$lon - lon) * pi / 180
    lat_rad <- lat * pi / 180
    cen_lat_rad <- coords$lat * pi / 180
    a <- sin(dlat/2)^2 + cos(lat_rad) * cos(cen_lat_rad) * sin(dlon/2)^2
    c <- 2 * atan2(sqrt(a), sqrt(1-a))
    distance <- R * c
  } else {
    # Euclidean distance in degrees
    distance <- sqrt((coords$lon - lon)^2 + (coords$lat - lat)^2)
  }

  return(list(
    original = c(lon = lon, lat = lat),
    h_index = result$h_index,
    reconstructed = c(lon = coords$lon, lat = coords$lat),
    error = distance,
    units = units
  ))
}

# =============================================================================
# QUAD COORDINATE ACCESS FUNCTIONS
# =============================================================================
#
# These functions expose the intermediate Quad IJ coordinate system used
# internally by DGGS. The coordinate pipeline is:
#
#   lon/lat → Icosa Triangle (face, tx, ty) → Quad IJ (quad, i, j) → Cell ID
#
# Most users should use the high-level hexify() function or cell ID-based
# functions. These Quad IJ functions are useful for:
# - Understanding the DGGS internals
# - Debugging coordinate transformations
# - Advanced applications requiring direct grid access
# =============================================================================

#' Convert longitude/latitude to Quad IJ coordinates
#'
#' Converts geographic coordinates to the intermediate Quad IJ representation
#' used internally by ISEA DGGS. Returns the quad number (0-11) and integer
#' cell indices (i, j) within that quad.
#'
#' @param lon Longitude in degrees (-180 to 180)
#' @param lat Latitude in degrees (-90 to 90)
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return List with components:
#'   \item{quad}{Quad number (0-11)}
#'   \item{i}{Integer cell index along first axis}
#'   \item{j}{Integer cell index along second axis}
#'   \item{icosa_triangle_face}{Source icosahedral face (0-19)}
#'   \item{icosa_triangle_x}{X coordinate on triangle face}
#'   \item{icosa_triangle_y}{Y coordinate on triangle face}
#'
#' @details
#' The 20 icosahedral triangle faces are grouped into 12 quads:
#' - Quad 0: North polar region
#' - Quads 1-5: Upper hemisphere rhombi
#' - Quads 6-10: Lower hemisphere rhombi
#' - Quad 11: South polar region
#'
#' @family coordinate conversion
#' @keywords internal
#' @export
#' @examples
#' # Get Quad IJ coordinates for Paris
#' result <- hexify_lonlat_to_quad_ij(lon = 2.35, lat = 48.86,
#'                                     resolution = 10, aperture = 3)
#' print(result)
hexify_lonlat_to_quad_ij <- function(lon, lat, resolution, aperture = 3L) {

  validate_aperture(aperture)
  validate_resolution(resolution)

  cpp_lonlat_to_quad_ij(
    lon_deg = as.numeric(lon),
    lat_deg = as.numeric(lat),
    aperture = as.integer(aperture),
    resolution = as.integer(resolution)
  )
}

#' Convert Quad IJ coordinates to cell ID
#'
#' Converts Quad IJ coordinates to a global cell identifier.
#' This is the final step in the coordinate pipeline.
#'
#' @param quad Quad number (0-11), integer or vector
#' @param i Cell index along first axis, integer or vector
#' @param j Cell index along second axis, integer or vector
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return Numeric vector of cell IDs
#'
#' @family coordinate conversion
#' @keywords internal
#' @export
#' @examples
#' # Convert Quad IJ to cell ID
#' cell_id <- hexify_quad_ij_to_cell(quad = 1, i = 100, j = 50,
#'                                    resolution = 10, aperture = 3)
#' print(cell_id)
hexify_quad_ij_to_cell <- function(quad, i, j, resolution, aperture = 3L) {

  validate_aperture(aperture)
  validate_resolution(resolution)

  cpp_quad_ij_to_cell(
    quad = as.integer(quad),
    i = as.numeric(i),
    j = as.numeric(j),
    resolution = as.integer(resolution),
    aperture = as.integer(aperture)
  )
}

#' Convert Quad IJ to Quad XY (continuous coordinates)
#'
#' Converts discrete cell indices to continuous quad coordinates.
#' Useful for computing cell centers or understanding the cell geometry.
#'
#' @param quad Quad number (0-11)
#' @param i Cell index along first axis
#' @param j Cell index along second axis
#' @param resolution Grid resolution level
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return List with components:
#'   \item{quad_x}{Continuous X coordinate in quad space}
#'   \item{quad_y}{Continuous Y coordinate in quad space}
#'
#' @family coordinate conversion
#' @keywords internal
#' @export
#' @examples
#' # Get continuous quad coordinates for a cell
#' xy <- hexify_quad_ij_to_xy(quad = 1, i = 100, j = 50,
#'                            resolution = 10, aperture = 3)
#' print(xy)
hexify_quad_ij_to_xy <- function(quad, i, j, resolution, aperture = 3L) {

  validate_aperture(aperture)

  cpp_quad_ij_to_xy(
    quad = as.integer(quad),
    i = as.numeric(i),
    j = as.numeric(j),
    aperture = as.integer(aperture),
    resolution = as.integer(resolution)
  )
}

#' Convert Icosa Triangle to Quad XY coordinates
#'
#' Converts icosahedral triangle coordinates (from Snyder projection)
#' to quad XY coordinates. This is an intermediate step in the pipeline.
#'
#' @param icosa_triangle_face Triangle face number (0-19)
#' @param icosa_triangle_x X coordinate on triangle face
#' @param icosa_triangle_y Y coordinate on triangle face
#'
#' @return List with components:
#'   \item{quad}{Quad number (0-11)}
#'   \item{quad_x}{Continuous X coordinate in quad space}
#'   \item{quad_y}{Continuous Y coordinate in quad space}
#'
#' @family coordinate conversion
#' @keywords internal
#' @export
#' @examples
#' # First get triangle coordinates from lon/lat
#' fwd <- hexify_forward(lon = 2.35, lat = 48.86)
#'
#' # Then convert to quad XY
#' quad_xy <- hexify_icosa_tri_to_quad_xy(
#'   icosa_triangle_face = fwd["face"],
#'   icosa_triangle_x = fwd["icosa_triangle_x"],
#'   icosa_triangle_y = fwd["icosa_triangle_y"]
#' )
#' print(quad_xy)
hexify_icosa_tri_to_quad_xy <- function(icosa_triangle_face,
                                         icosa_triangle_x,
                                         icosa_triangle_y) {
  cpp_icosa_tri_to_quad_xy(
    icosa_triangle_face = as.integer(icosa_triangle_face),
    icosa_triangle_x = as.numeric(icosa_triangle_x),
    icosa_triangle_y = as.numeric(icosa_triangle_y)
  )
}

#' Convert Icosa Triangle to Quad IJ coordinates
#'
#' Converts icosahedral triangle coordinates directly to Quad IJ,
#' combining the transformation and quantization steps.
#'
#' @param icosa_triangle_face Triangle face number (0-19)
#' @param icosa_triangle_x X coordinate on triangle face
#' @param icosa_triangle_y Y coordinate on triangle face
#' @param resolution Grid resolution level
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return List with components:
#'   \item{quad}{Quad number (0-11)}
#'   \item{i}{Integer cell index along first axis}
#'   \item{j}{Integer cell index along second axis}
#'
#' @family coordinate conversion
#' @keywords internal
#' @export
#' @examples
#' # First get triangle coordinates from lon/lat
#' fwd <- hexify_forward(lon = 2.35, lat = 48.86)
#'
#' # Then convert to quad IJ
#' quad_ij <- hexify_icosa_tri_to_quad_ij(
#'   icosa_triangle_face = fwd["face"],
#'   icosa_triangle_x = fwd["icosa_triangle_x"],
#'   icosa_triangle_y = fwd["icosa_triangle_y"],
#'   resolution = 10,
#'   aperture = 3
#' )
#' print(quad_ij)
hexify_icosa_tri_to_quad_ij <- function(icosa_triangle_face,
                                         icosa_triangle_x,
                                         icosa_triangle_y,
                                         resolution,
                                         aperture = 3L) {

  validate_aperture(aperture)

  cpp_icosa_tri_to_quad_ij(
    icosa_triangle_face = as.integer(icosa_triangle_face),
    icosa_triangle_x = as.numeric(icosa_triangle_x),
    icosa_triangle_y = as.numeric(icosa_triangle_y),
    aperture = as.integer(aperture),
    resolution = as.integer(resolution)
  )
}

#' Convert Quad XY to Icosa Triangle coordinates
#'
#' Inverse transformation from quad coordinates back to icosahedral
#' triangle coordinates. Useful for projecting cell centers back to lon/lat.
#'
#' @param quad Quad number (0-11)
#' @param quad_x Continuous X coordinate in quad space
#' @param quad_y Continuous Y coordinate in quad space
#'
#' @return List with components:
#'   \item{icosa_triangle_face}{Triangle face number (0-19)}
#'   \item{icosa_triangle_x}{X coordinate on triangle face}
#'   \item{icosa_triangle_y}{Y coordinate on triangle face}
#'
#' @family coordinate conversion
#' @keywords internal
#' @export
#' @examples
#' # Convert quad XY back to triangle coordinates
#' tri <- hexify_quad_xy_to_icosa_tri(quad = 1, quad_x = 0.5, quad_y = 0.3)
#' print(tri)
hexify_quad_xy_to_icosa_tri <- function(quad, quad_x, quad_y) {
  cpp_quad_xy_to_icosa_tri(
    quad = as.integer(quad),
    quad_x = as.numeric(quad_x),
    quad_y = as.numeric(quad_y)
  )
}

# =============================================================================
# CELL ID TO COORDINATE SYSTEM CONVERSIONS
# =============================================================================
#
# These functions convert from Cell IDs back to intermediate
# coordinate representations, compatible with 'dggridR' conversion functions.
# =============================================================================

#' Convert Cell ID to Quad IJ coordinates
#'
#' Converts DGGRID-compatible cell IDs to Quad IJ coordinates.
#' This is the inverse of hexify_quad_ij_to_cell().
#'
#' Compatible with 'dggridR' dgSEQNUM_to_Q2DI().
#'
#' @param cell_id Numeric vector of cell IDs (1-based)
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return Data frame with columns:
#'   \item{quad}{Quad number (0-11)}
#'   \item{i}{Integer cell index along first axis}
#'   \item{j}{Integer cell index along second axis}
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_quad_ij_to_cell}} for the forward operation,
#'   \code{\link{hexify_cell_to_icosa_tri}} for conversion to triangle coords
#'
#' @keywords internal
#' @export
#' @examples
#' # Get Quad IJ coordinates for a cell
#' result <- hexify_cell_to_quad_ij(cell_id = 1000, resolution = 10, aperture = 3)
#' print(result)
#'
#' # Round-trip test
#' cell_id <- hexify_quad_ij_to_cell(result$quad, result$i, result$j,
#'                                    resolution = 10, aperture = 3)
#' # Should equal original cell_id
hexify_cell_to_quad_ij <- function(cell_id, resolution, aperture = 3L) {
  hexify_cell_id_to_quad_ij(
    cell_id = as.numeric(cell_id),
    resolution = as.integer(resolution),
    aperture = as.integer(aperture)
  )
}

#' Convert Cell ID to Icosa Triangle coordinates
#'
#' Converts DGGRID-compatible cell IDs to icosahedral triangle
#' coordinates (face, x, y). These are the coordinates produced by the
#' 'Snyder' 'ISEA' forward projection.
#'
#' Compatible with 'dggridR' dgSEQNUM_to_PROJTRI().
#'
#' @param cell_id Numeric vector of cell IDs (1-based)
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return Data frame with columns:
#'   \item{icosa_triangle_face}{Triangle face number (0-19)}
#'   \item{icosa_triangle_x}{X coordinate on triangle face}
#'   \item{icosa_triangle_y}{Y coordinate on triangle face}
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_cell_to_quad_ij}} for conversion to Quad IJ,
#'   \code{\link{hexify_cell_to_lonlat}} for conversion to lon/lat
#'
#' @keywords internal
#' @export
#' @examples
#' # Get triangle coordinates for a cell
#' result <- hexify_cell_to_icosa_tri(cell_id = 1000, resolution = 10, aperture = 3)
#' print(result)
#'
#' # Convert back to lon/lat using inverse projection
#' coords <- hexify_inverse(result$icosa_triangle_face,
#'                          result$icosa_triangle_x,
#'                          result$icosa_triangle_y)
hexify_cell_to_icosa_tri <- function(cell_id, resolution, aperture = 3L) {

  validate_aperture(aperture)
  validate_resolution(resolution)

  cpp_cell_to_icosa_tri(
    cell_id = as.numeric(cell_id),
    resolution = as.integer(resolution),
    aperture = as.integer(aperture)
  )
}

#' Convert Quad IJ to Icosa Triangle coordinates
#'
#' Converts Quad IJ coordinates to icosahedral triangle coordinates.
#' This is useful for understanding where a cell is located on the
#' icosahedral projection.
#'
#' Equivalent to 'dggridR' dgQ2DI_to_PROJTRI().
#'
#' @param quad Quad number (0-11), integer or vector
#' @param i Cell index along first axis, integer or vector
#' @param j Cell index along second axis, integer or vector
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return Data frame with columns:
#'   \item{icosa_triangle_face}{Triangle face number (0-19)}
#'   \item{icosa_triangle_x}{X coordinate on triangle face}
#'   \item{icosa_triangle_y}{Y coordinate on triangle face}
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_icosa_tri_to_quad_ij}} for the inverse,
#'   \code{\link{hexify_cell_to_icosa_tri}} for conversion from cell ID
#'
#' @keywords internal
#' @export
#' @examples
#' # Get triangle coordinates for a Quad IJ position
#' result <- hexify_quad_ij_to_icosa_tri(quad = 1, i = 100, j = 50,
#'                                        resolution = 10, aperture = 3)
#' print(result)
hexify_quad_ij_to_icosa_tri <- function(quad, i, j, resolution, aperture = 3L) {

  validate_aperture(aperture)
  validate_resolution(resolution)

  cpp_quad_ij_to_icosa_tri(
    quad = as.integer(quad),
    i = as.numeric(i),
    j = as.numeric(j),
    resolution = as.integer(resolution),
    aperture = as.integer(aperture)
  )
}

# =============================================================================
# QUAD XY (CONTINUOUS) COORDINATE CONVERSIONS
# =============================================================================
#
# These functions work with Quad XY coordinates - the continuous (non-quantized)
# representation in quad space. Equivalent to 'dggridR' Q2DD coordinate system.
# =============================================================================

#' Convert Cell ID to Quad XY coordinates
#'
#' Converts DGGRID-compatible cell IDs to Quad XY coordinates
#' (continuous quad space). This is the cell center in quad coordinates.
#'
#' Compatible with 'dggridR' dgSEQNUM_to_Q2DD().
#'
#' @param cell_id Numeric vector of cell IDs (1-based)
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return Data frame with columns:
#'   \item{quad}{Quad number (0-11)}
#'   \item{quad_x}{Continuous X coordinate in quad space}
#'   \item{quad_y}{Continuous Y coordinate in quad space}
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_quad_xy_to_cell}} for the inverse operation,
#'   \code{\link{hexify_cell_to_quad_ij}} for integer grid coordinates
#'
#' @keywords internal
#' @export
#' @examples
#' # Get Quad XY coordinates for a cell
#' result <- hexify_cell_to_quad_xy(cell_id = 1000, resolution = 10, aperture = 3)
#' print(result)
#'
#' # Round-trip test
#' cell_id <- hexify_quad_xy_to_cell(result$quad, result$quad_x, result$quad_y,
#'                                    resolution = 10, aperture = 3)
#' # Should equal original cell_id
hexify_cell_to_quad_xy <- function(cell_id, resolution, aperture = 3L) {

  validate_aperture(aperture)
  validate_resolution(resolution)

  cpp_cell_to_quad_xy(
    cell_id = as.numeric(cell_id),
    resolution = as.integer(resolution),
    aperture = as.integer(aperture)
  )
}

#' Convert Quad XY coordinates to Cell ID
#'
#' Converts Quad XY coordinates (continuous quad space) to DGGRID-compatible
#' cell IDs. The coordinates are quantized to the nearest cell.
#'
#' Compatible with 'dggridR' dgQ2DD_to_SEQNUM().
#'
#' @param quad Quad number (0-11), integer or vector
#' @param quad_x Continuous X coordinate in quad space
#' @param quad_y Continuous Y coordinate in quad space
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return Numeric vector of cell IDs
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_cell_to_quad_xy}} for the inverse operation,
#'   \code{\link{hexify_quad_ij_to_cell}} for integer grid coordinates
#'
#' @keywords internal
#' @export
#' @examples
#' # Convert Quad XY to cell ID
#' cell_id <- hexify_quad_xy_to_cell(quad = 1, quad_x = 0.5, quad_y = 0.3,
#'                                    resolution = 10, aperture = 3)
#' print(cell_id)
hexify_quad_xy_to_cell <- function(quad, quad_x, quad_y, resolution,
                                   aperture = 3L) {

  validate_aperture(aperture)
  validate_resolution(resolution)

  cpp_quad_xy_to_cell(
    quad = as.integer(quad),
    quad_x = as.numeric(quad_x),
    quad_y = as.numeric(quad_y),
    resolution = as.integer(resolution),
    aperture = as.integer(aperture)
  )
}

# =============================================================================
# PLANE COORDINATE CONVERSIONS
# =============================================================================
#
# PLANE coordinates represent the unfolded icosahedron in 2D space.
# Each triangle face is rotated and translated to form a contiguous layout
# covering approximately 5.5 x 1.73 units.
#
# These functions are useful for:
# - Visualizing the entire DGGS on a flat surface
# - Understanding spatial relationships between triangles
# - Creating custom map projections
# =============================================================================

#' Convert Icosa Triangle coordinates to PLANE coordinates
#'
#' Transforms icosahedral triangle coordinates to the 2D PLANE representation
#' (unfolded icosahedron). Each triangle is rotated and translated to its
#' position in the unfolded layout.
#'
#' Equivalent to 'dggridR' dgPROJTRI_to_PLANE().
#'
#' @param icosa_triangle_face Triangle face number (0-19), integer or vector
#' @param icosa_triangle_x X coordinate on triangle face
#' @param icosa_triangle_y Y coordinate on triangle face
#'
#' @return Data frame with columns:
#'   \item{plane_x}{X coordinate in PLANE space (range ~0 to 5.5)}
#'   \item{plane_y}{Y coordinate in PLANE space (range ~0 to 1.73)}
#'
#' @details
#' The PLANE layout arranges all 20 icosahedral faces into a roughly
#' rectangular region. Faces 0-4 and 5-9 form the upper row, while
#' faces 10-14 and 15-19 form the lower row. Adjacent faces share
#' edges in this representation.
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_cell_to_plane}} for direct cell ID conversion,
#'   \code{\link{hexify_lonlat_to_plane}} for lon/lat to PLANE
#'
#' @keywords internal
#' @export
#' @examples
#' # Get PLANE coordinates from triangle coordinates
#' fwd <- hexify_forward(lon = 2.35, lat = 48.86)
#' plane <- hexify_icosa_tri_to_plane(
#'   icosa_triangle_face = fwd["face"],
#'   icosa_triangle_x = fwd["icosa_triangle_x"],
#'   icosa_triangle_y = fwd["icosa_triangle_y"]
#' )
#' print(plane)
hexify_icosa_tri_to_plane <- function(icosa_triangle_face,
                                       icosa_triangle_x,
                                       icosa_triangle_y) {
  cpp_icosa_tri_to_plane(
    icosa_triangle_face = as.integer(icosa_triangle_face),
    icosa_triangle_x = as.numeric(icosa_triangle_x),
    icosa_triangle_y = as.numeric(icosa_triangle_y)
  )
}

#' Convert Cell ID to PLANE coordinates
#'
#' Converts DGGRID-compatible cell IDs directly to PLANE coordinates.
#' Returns the cell center in the unfolded icosahedron layout.
#'
#' Compatible with 'dggridR' dgSEQNUM_to_PLANE().
#'
#' @param cell_id Numeric vector of cell IDs (1-based)
#' @param resolution Grid resolution level (0-30)
#' @param aperture Grid aperture: 3, 4, or 7
#'
#' @return Data frame with columns:
#'   \item{plane_x}{X coordinate in PLANE space (range ~0 to 5.5)}
#'   \item{plane_y}{Y coordinate in PLANE space (range ~0 to 1.73)}
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_icosa_tri_to_plane}} for triangle conversion,
#'   \code{\link{hexify_lonlat_to_plane}} for lon/lat conversion
#'
#' @keywords internal
#' @export
#' @examples
#' # Get PLANE coordinates for cells
#' plane <- hexify_cell_to_plane(cell_id = c(100, 200, 300),
#'                                resolution = 5, aperture = 3)
#' plot(plane$plane_x, plane$plane_y)
hexify_cell_to_plane <- function(cell_id, resolution, aperture = 3L) {

  validate_aperture(aperture)
  validate_resolution(resolution)

  cpp_cell_to_plane(
    cell_id = as.numeric(cell_id),
    resolution = as.integer(resolution),
    aperture = as.integer(aperture)
  )
}

#' Convert longitude/latitude to PLANE coordinates
#'
#' Converts geographic coordinates directly to PLANE coordinates
#' (unfolded icosahedron). Combines forward 'Snyder' projection with
#' the PLANE transformation.
#'
#' Equivalent to 'dggridR' dgGEO_to_PLANE().
#'
#' @param lon Longitude in degrees (-180 to 180)
#' @param lat Latitude in degrees (-90 to 90)
#'
#' @return Data frame with columns:
#'   \item{plane_x}{X coordinate in PLANE space (range ~0 to 5.5)}
#'   \item{plane_y}{Y coordinate in PLANE space (range ~0 to 1.73)}
#'
#' @family coordinate conversion
#' @seealso \code{\link{hexify_cell_to_plane}} for cell ID conversion,
#'   \code{\link{hexify_icosa_tri_to_plane}} for triangle conversion
#'
#' @keywords internal
#' @export
#' @examples
#' # Plot world cities in PLANE coordinates
#' cities <- data.frame(
#'   lon = c(2.35, -74.00, 139.69, 151.21),
#'   lat = c(48.86, 40.71, 35.69, -33.87)
#' )
#' plane <- hexify_lonlat_to_plane(cities$lon, cities$lat)
#' plot(plane$plane_x, plane$plane_y)
hexify_lonlat_to_plane <- function(lon, lat) {
  cpp_lonlat_to_plane(
    lon = as.numeric(lon),
    lat = as.numeric(lat)
  )
}

# tests/testthat/test-grid-wrappers.R
# Tests for grid-based coordinate conversion wrappers

setup_icosa <- function() {
  cpp_build_icosa()
}

# =============================================================================
# GRID-BASED CELL CONVERSIONS
# =============================================================================

test_that("hexify_grid_to_cell works with grid object", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  cell_ids <- hexify_grid_to_cell(grid, lon = c(0, 10), lat = c(45, 50))

  expect_true(all(cell_ids > 0))
  expect_equal(length(cell_ids), 2)
})

test_that("hexify_grid_to_cell validates grid object", {
  expect_error(hexify_grid_to_cell("not a grid", lon = 0, lat = 0))
  expect_error(hexify_grid_to_cell(list(x = 1), lon = 0, lat = 0))
})

test_that("hexify_grid_cell_to_lonlat works with grid object", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  cell_ids <- hexify_grid_to_cell(grid, lon = c(0, 10), lat = c(45, 50))
  coords <- hexify_grid_cell_to_lonlat(grid, cell_ids)

  expect_true("lon_deg" %in% names(coords))
  expect_true("lat_deg" %in% names(coords))
  expect_equal(length(coords$lon_deg), 2)
})

test_that("hexify_grid_cell_to_lonlat validates grid object", {
  expect_error(hexify_grid_cell_to_lonlat("not a grid", cell_id = 1))
})

test_that("grid-based round-trip is consistent", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  lon <- c(0, 45, -120)
  lat <- c(0, 30, -45)

  cell_ids <- hexify_grid_to_cell(grid, lon = lon, lat = lat)
  coords <- hexify_grid_cell_to_lonlat(grid, cell_ids)
  cell_ids2 <- hexify_grid_to_cell(grid, lon = coords$lon_deg, lat = coords$lat_deg)

  expect_equal(cell_ids, cell_ids2)
})

# =============================================================================
# HIERARCHICAL INDEX CONVERSIONS
# =============================================================================

test_that("hexify_lonlat_to_h_index returns correct structure", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  result <- hexify_lonlat_to_h_index(grid, lon = 0, lat = 45)

  expect_s3_class(result, "data.frame")
  expect_true("h_index" %in% names(result))
  expect_true("face" %in% names(result))
  expect_equal(nrow(result), 1)
})

test_that("hexify_lonlat_to_h_index validates grid object", {
  expect_error(
    hexify_lonlat_to_h_index("not a grid", lon = 0, lat = 0),
    "hexify_grid"
  )
})

test_that("hexify_lonlat_to_h_index validates input lengths", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  expect_error(
    hexify_lonlat_to_h_index(grid, lon = c(0, 1), lat = c(0)),
    "same length"
  )
})

test_that("hexify_lonlat_to_h_index validates numeric input", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  expect_error(
    hexify_lonlat_to_h_index(grid, lon = "0", lat = "45"),
    "numeric"
  )
})

test_that("hexify_lonlat_to_h_index warns on invalid coordinates", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  expect_warning(
    hexify_lonlat_to_h_index(grid, lon = 200, lat = 45),
    "longitude"
  )
  expect_warning(
    hexify_lonlat_to_h_index(grid, lon = 0, lat = 100),
    "latitude"
  )
})

test_that("hexify_lonlat_to_h_index handles NA values", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  result <- hexify_lonlat_to_h_index(grid, lon = c(0, NA), lat = c(45, NA))

  expect_equal(nrow(result), 2)
  expect_true(is.na(result$h_index[2]))
  expect_true(is.na(result$face[2]))
})

test_that("hexify_h_index_to_lonlat returns correct structure", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  h_result <- hexify_lonlat_to_h_index(grid, lon = 0, lat = 45)
  coords <- hexify_h_index_to_lonlat(grid, h_result$h_index)

  expect_s3_class(coords, "data.frame")
  expect_true("lon" %in% names(coords))
  expect_true("lat" %in% names(coords))
})

test_that("hexify_h_index_to_lonlat validates grid object", {
  expect_error(
    hexify_h_index_to_lonlat("not a grid", h_index = "01"),
    "hexify_grid"
  )
})

test_that("hexify_h_index_to_lonlat validates character input", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  expect_error(
    hexify_h_index_to_lonlat(grid, h_index = 123),
    "character"
  )
})

test_that("hexify_h_index_to_lonlat handles NA values", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  h_result <- hexify_lonlat_to_h_index(grid, lon = 0, lat = 45)
  coords <- hexify_h_index_to_lonlat(grid, c(h_result$h_index, NA))

  expect_equal(nrow(coords), 2)
  expect_true(is.na(coords$lon[2]))
  expect_true(is.na(coords$lat[2]))
})

test_that("h_index round-trip is consistent", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  lon <- 10
  lat <- 50

  h_result <- hexify_lonlat_to_h_index(grid, lon = lon, lat = lat)
  coords <- hexify_h_index_to_lonlat(grid, h_result$h_index)
  h_result2 <- hexify_lonlat_to_h_index(grid, lon = coords$lon, lat = coords$lat)

  expect_equal(h_result$h_index, h_result2$h_index)
})

# =============================================================================
# ROUND-TRIP TEST FUNCTION
# =============================================================================

test_that("hexify_roundtrip_test returns correct structure", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  result <- hexify_roundtrip_test(grid, lon = 10, lat = 50)

  expect_type(result, "list")
  expect_true("original" %in% names(result))
  expect_true("h_index" %in% names(result))
  expect_true("reconstructed" %in% names(result))
  expect_true("error" %in% names(result))
  expect_true("units" %in% names(result))
})

test_that("hexify_roundtrip_test default units are km", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  result <- hexify_roundtrip_test(grid, lon = 10, lat = 50)

  expect_equal(result$units, "km")
})

test_that("hexify_roundtrip_test supports degrees units", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  result <- hexify_roundtrip_test(grid, lon = 10, lat = 50, units = "degrees")

  expect_equal(result$units, "degrees")
})

test_that("hexify_roundtrip_test error is reasonable", {
  setup_icosa()

  grid <- hexify_grid(area = 1000, aperture = 3)
  result <- hexify_roundtrip_test(grid, lon = 10, lat = 50)

  # Error should be less than cell spacing (roughly sqrt(area))
  expect_lt(result$error, sqrt(1000))
})

# =============================================================================
# DIFFERENT APERTURES
# =============================================================================

test_that("grid wrappers work for all apertures", {
  skip_on_cran()  # Loop test across apertures
  setup_icosa()

  for (ap in c(3, 4, 7)) {
    grid <- hexify_grid(area = 10000, aperture = ap)

    cell_ids <- hexify_grid_to_cell(grid, lon = 0, lat = 45)
    coords <- hexify_grid_cell_to_lonlat(grid, cell_ids)

    expect_true(cell_ids > 0)
    expect_true(is.finite(coords$lon_deg))
    expect_true(is.finite(coords$lat_deg))
  }
})

test_that("h_index works for different apertures", {
  skip_on_cran()  # Loop test across apertures
  setup_icosa()

  for (ap in c(3, 4, 7)) {
    grid <- hexify_grid(area = 10000, aperture = ap)

    result <- hexify_lonlat_to_h_index(grid, lon = 0, lat = 45)
    coords <- hexify_h_index_to_lonlat(grid, result$h_index)

    expect_true(nchar(result$h_index) > 0)
    expect_true(is.finite(coords$lon))
    expect_true(is.finite(coords$lat))
  }
})

# tests/testthat/test-cell-id.R
# Tests for cell ID functions
#
# Functions tested:
# - hexify_lonlat_to_cell()
# - hexify_cell_to_lonlat()
# - hexify_cell_to_quad_ij()
# - hexify_cell_to_icosa_tri()
# - hexify_cell_id_to_quad_ij()
# - hexify_grid_to_cell()
# - hexify_grid_cell_to_lonlat()

# =============================================================================
# LONLAT TO CELL (CELL ID)
# =============================================================================

test_that("lonlat_to_cell returns positive integers for aperture 3", {
  lons <- c(0, 10, 20, -30, 45)
  lats <- c(0, 45, -30, 15, -60)

  cell_ids <- hexify_lonlat_to_cell(lons, lats, resolution = 10, aperture = 3)

  expect_true(all(cell_ids > 0))
  expect_true(is.numeric(cell_ids))
  expect_true(all(is.finite(cell_ids)))
})

test_that("lonlat_to_cell returns positive integers for aperture 4", {
  lons <- c(0, 10, 20, -30, 45)
  lats <- c(0, 45, -30, 15, -60)

  cell_ids <- hexify_lonlat_to_cell(lons, lats, resolution = 10, aperture = 4)

  expect_true(all(cell_ids > 0))
  expect_true(is.numeric(cell_ids))
  expect_true(all(is.finite(cell_ids)))
})

test_that("lonlat_to_cell returns positive integers for aperture 7", {
  lons <- c(0, 10, 20, -30, 45)
  lats <- c(0, 45, -30, 15, -60)

  cell_ids <- hexify_lonlat_to_cell(lons, lats, resolution = 5, aperture = 7)

  expect_true(all(cell_ids > 0))
  expect_true(is.numeric(cell_ids))
  expect_true(all(is.finite(cell_ids)))
})

test_that("lonlat_to_cell handles extreme coordinates", {
  extreme_coords <- data.frame(
    lon = c(-180, 180, 0, 0),
    lat = c(0, 0, -89, 89)
  )

  for (i in seq_len(nrow(extreme_coords))) {
    cell_id <- hexify_lonlat_to_cell(
      extreme_coords$lon[i],
      extreme_coords$lat[i],
      resolution = 10,
      aperture = 3
    )

    expect_true(is.finite(cell_id))
    expect_true(cell_id > 0)
  }
})

test_that("resolution 0 returns face numbers (1-20)", {
  lons <- c(0, 60, 120, -60, -120, 0)
  lats <- c(45, 45, 45, 45, 45, -45)

  cell_ids <- hexify_lonlat_to_cell(lons, lats, resolution = 0, aperture = 3)

  expect_true(all(cell_ids >= 1 & cell_ids <= 20))
})

# =============================================================================
# CELL TO LONLAT
# =============================================================================

test_that("cell_to_lonlat returns valid coordinates", {
  cell_ids <- c(100, 1000, 10000)

  coords <- hexify_cell_to_lonlat(cell_ids, resolution = 10, aperture = 3)

  expect_true("lon_deg" %in% names(coords))
  expect_true("lat_deg" %in% names(coords))
  expect_true(all(is.finite(coords$lon_deg)))
  expect_true(all(is.finite(coords$lat_deg)))
  expect_true(all(coords$lon_deg >= -180 & coords$lon_deg <= 180))
  expect_true(all(coords$lat_deg >= -90 & coords$lat_deg <= 90))
})

# =============================================================================
# ROUND-TRIP CONVERSION
# =============================================================================

test_that("round-trip conversion works for aperture 3", {
  test_lons <- c(0, 5, 10, -30, 45, 90, -90)
  test_lats <- c(0, 45, -30, 15, -60, 30, -15)
  resolution <- 10
  tolerance <- 1.0  # degrees

  cell_ids <- hexify_lonlat_to_cell(test_lons, test_lats, resolution, aperture = 3)
  coords <- hexify_cell_to_lonlat(cell_ids, resolution, aperture = 3)

  for (i in seq_along(test_lons)) {
    lon_error <- abs(coords$lon_deg[i] - test_lons[i])
    lat_error <- abs(coords$lat_deg[i] - test_lats[i])

    # Account for longitude wrapping
    if (lon_error > 180) lon_error <- 360 - lon_error

    expect_lt(lon_error, tolerance)
    expect_lt(lat_error, tolerance)
  }
})

test_that("round-trip conversion works for aperture 4", {
  test_lons <- c(0, 5, 10, -30, 45)
  test_lats <- c(0, 45, -30, 15, -60)
  resolution <- 10
  tolerance <- 1.0

  cell_ids <- hexify_lonlat_to_cell(test_lons, test_lats, resolution, aperture = 4)
  coords <- hexify_cell_to_lonlat(cell_ids, resolution, aperture = 4)

  for (i in seq_along(test_lons)) {
    lon_error <- abs(coords$lon_deg[i] - test_lons[i])
    lat_error <- abs(coords$lat_deg[i] - test_lats[i])
    if (lon_error > 180) lon_error <- 360 - lon_error

    expect_lt(lon_error, tolerance)
    expect_lt(lat_error, tolerance)
  }
})

test_that("round-trip conversion works for aperture 7", {
  test_lons <- c(0, 5, 10, -30, 45)
  test_lats <- c(0, 45, -30, 15, -60)
  resolution <- 5
  # Aperture 7 uses surrogate-substrate conversion with rounding
  tolerance <- 4.0

  cell_ids <- hexify_lonlat_to_cell(test_lons, test_lats, resolution, aperture = 7)
  coords <- hexify_cell_to_lonlat(cell_ids, resolution, aperture = 7)

  for (i in seq_along(test_lons)) {
    lon_error <- abs(coords$lon_deg[i] - test_lons[i])
    lat_error <- abs(coords$lat_deg[i] - test_lats[i])
    if (lon_error > 180) lon_error <- 360 - lon_error

    expect_lt(lon_error, tolerance)
    expect_lt(lat_error, tolerance)
  }
})

test_that("vertex-quad pentagons report their own centre", {
  # Quads 0 and 11 hold the single pentagon at icosahedron vertex 0 and at its
  # antipode. Under the ISEA orientation neither is a geographic pole.
  vert_lon <- c(ISEA_VERT0_LON_DEG, ISEA_VERT0_LON_DEG - 180)
  vert_lat <- c(ISEA_VERT0_LAT_DEG, -ISEA_VERT0_LAT_DEG)

  for (ap in c(3, 4, 7)) {
    for (res in c(2, 5)) {
      cells <- hexify_lonlat_to_cell(vert_lon, vert_lat, res, aperture = ap)
      qij <- hexify_cell_to_quad_ij(cells, res, aperture = ap)
      coords <- hexify_cell_to_lonlat(cells, res, aperture = ap)

      expect_equal(qij$quad, c(0L, 11L))
      expect_equal(coords$lon_deg, vert_lon, tolerance = 1e-6)
      expect_equal(coords$lat_deg, vert_lat, tolerance = 1e-6)
    }
  }
})

# =============================================================================
# CELL TO QUAD IJ
# =============================================================================

test_that("cell_to_quad_ij returns valid structure", {
  cell_ids <- c(100, 1000, 10000)

  result <- hexify_cell_to_quad_ij(cell_ids, resolution = 10, aperture = 3)

  expect_true("quad" %in% names(result))
  expect_true("i" %in% names(result))
  expect_true("j" %in% names(result))
  expect_true(all(result$quad >= 0 & result$quad <= 11))
  expect_true(all(result$i >= 0))
  expect_true(all(result$j >= 0))
})

test_that("cell_to_quad_ij round-trips with quad_ij_to_cell", {
  cell_ids <- c(100, 1000, 5000, 10000)
  resolution <- 10

  for (ap in c(3, 4, 7)) {
    result <- hexify_cell_to_quad_ij(cell_ids, resolution, aperture = ap)
    recovered <- hexify_quad_ij_to_cell(result$quad, result$i, result$j,
                                         resolution, aperture = ap)

    expect_equal(recovered, cell_ids,
                 info = sprintf("aperture %d quad_ij round-trip", ap))
  }
})

test_that("every cell ID in [1, 10 * aperture^res + 2] names one cell", {
  for (ap in c(3, 4, 7)) {
    for (res in 0:3) {
      n_cells <- 10 * ap^res + 2
      ids <- seq_len(n_cells)

      qij <- hexify_cell_to_quad_ij(ids, res, aperture = ap)
      recovered <- hexify_quad_ij_to_cell(qij$quad, qij$i, qij$j, res,
                                          aperture = ap)

      expect_equal(recovered, as.numeric(ids),
                   info = sprintf("aperture %d res %d id round-trip", ap, res))
      expect_equal(length(unique(paste(qij$quad, qij$i, qij$j))), n_cells,
                   info = sprintf("aperture %d res %d distinct coords", ap, res))
    }
  }
})

test_that("assigned cell IDs stay inside the grid's ID range", {
  set.seed(42)
  n <- 500
  lon <- runif(n, -180, 180)
  lat <- asin(runif(n, -1, 1)) * 180 / pi

  for (ap in c(3, 4, 7)) {
    for (res in c(1, 2, 3, 6)) {
      cells <- hexify_lonlat_to_cell(lon, lat, res, aperture = ap)

      expect_true(all(cells >= 1 & cells <= 10 * ap^res + 2),
                  info = sprintf("aperture %d res %d cell ID range", ap, res))
    }
  }
})

# =============================================================================
# CELL TO ICOSA TRIANGLE
# =============================================================================

test_that("cell_to_icosa_tri returns valid structure", {
  cell_ids <- c(100, 1000, 10000)

  result <- hexify_cell_to_icosa_tri(cell_ids, resolution = 10, aperture = 3)

  expect_true("icosa_triangle_face" %in% names(result))
  expect_true("icosa_triangle_x" %in% names(result))
  expect_true("icosa_triangle_y" %in% names(result))
  expect_true(all(result$icosa_triangle_face >= 0 & result$icosa_triangle_face <= 19))
  expect_true(all(is.finite(result$icosa_triangle_x)))
  expect_true(all(is.finite(result$icosa_triangle_y)))
})

# =============================================================================
# CELL ID TO QUAD IJ INFO
# =============================================================================

test_that("cell_id_to_quad_ij returns valid cell info", {
  cell_ids <- c(1702, 1954, 100)

  info <- hexify_cell_id_to_quad_ij(cell_ids, resolution = 5, aperture = 3)

  expect_true(is.data.frame(info))
  expect_true("quad" %in% names(info))
  expect_true("i" %in% names(info))
  expect_true("j" %in% names(info))

  # Quads should be 0-11
  expect_true(all(info$quad >= 0 & info$quad <= 11))

  # i, j should be non-negative
  expect_true(all(info$i >= 0))
  expect_true(all(info$j >= 0))
})

# =============================================================================
# GRID-BASED WRAPPERS
# =============================================================================

test_that("grid_to_cell works with hexify_grid object", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  cell_ids <- hexify_grid_to_cell(grid, lon = c(0, 10), lat = c(45, 50))

  expect_true(all(cell_ids > 0))
  expect_length(cell_ids, 2)
})

test_that("grid_cell_to_lonlat works with hexify_grid object", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  cell_ids <- hexify_grid_to_cell(grid, lon = 5, lat = 45)
  coords <- hexify_grid_cell_to_lonlat(grid, cell_ids)

  expect_true("lon_deg" %in% names(coords))
  expect_true("lat_deg" %in% names(coords))
})

# =============================================================================
# BATCH PROCESSING
# =============================================================================

test_that("batch processing handles large datasets", {
  set.seed(123)
  n <- 1000
  lons <- runif(n, -180, 180)
  lats <- runif(n, -90, 90)

  cell_ids <- hexify_lonlat_to_cell(lons, lats, resolution = 10, aperture = 3)

  expect_length(cell_ids, n)
  expect_true(all(cell_ids > 0))

  # Check reasonable uniqueness
  uniqueness <- length(unique(cell_ids)) / n
  expect_gt(uniqueness, 0.8)
})

# tests/testthat/test-conversions.R
# Tests for coordinate system conversion functions
#
# Functions tested:
# - hexify_lonlat_to_quad_ij()
# - hexify_quad_ij_to_cell()
# - hexify_quad_ij_to_xy()
# - hexify_icosa_tri_to_quad_xy()
# - hexify_icosa_tri_to_quad_ij()
# - hexify_quad_xy_to_icosa_tri()
# - hexify_quad_ij_to_icosa_tri()

# =============================================================================
# LON/LAT TO QUAD IJ
# =============================================================================

test_that("lonlat_to_quad_ij returns valid structure", {
  result <- hexify_lonlat_to_quad_ij(lon = 2.35, lat = 48.86,
                                      resolution = 10, aperture = 3)

  expect_true("quad" %in% names(result))
  expect_true("i" %in% names(result))
  expect_true("j" %in% names(result))
  expect_true("icosa_triangle_face" %in% names(result))
  expect_true("icosa_triangle_x" %in% names(result))
  expect_true("icosa_triangle_y" %in% names(result))
})

test_that("lonlat_to_quad_ij quad is in valid range", {
  result <- hexify_lonlat_to_quad_ij(lon = 0, lat = 0,
                                      resolution = 10, aperture = 3)

  expect_true(result$quad >= 0 && result$quad <= 11)
})

test_that("lonlat_to_quad_ij i,j are non-negative", {
  result <- hexify_lonlat_to_quad_ij(lon = 16.37, lat = 48.21,
                                      resolution = 10, aperture = 3)

  expect_true(result$i >= 0)
  expect_true(result$j >= 0)
})

test_that("lonlat_to_quad_ij works for all apertures", {
  for (ap in c(3, 4, 7)) {
    res <- if (ap == 7) 5 else 10

    result <- hexify_lonlat_to_quad_ij(lon = 0, lat = 45,
                                        resolution = res, aperture = ap)

    expect_true(result$quad >= 0 && result$quad <= 11,
                info = sprintf("aperture %d", ap))
  }
})

# =============================================================================
# QUAD IJ TO CELL
# =============================================================================

test_that("quad_ij_to_cell returns positive cell IDs", {
  cell_id <- hexify_quad_ij_to_cell(quad = 1, i = 100, j = 50,
                                     resolution = 10, aperture = 3)

  expect_true(is.numeric(cell_id))
  expect_equal(cell_id, 24352)
})

test_that("quad_ij_to_cell is consistent with lonlat_to_cell", {
  lon <- 16.37
  lat <- 48.21
  resolution <- 10

  for (ap in c(3, 4)) {
    # Get cell via lon/lat
    cell_direct <- hexify_lonlat_to_cell(lon, lat, resolution, aperture = ap)

    # Get cell via quad IJ
    quad_ij <- hexify_lonlat_to_quad_ij(lon, lat, resolution, aperture = ap)
    cell_indirect <- hexify_quad_ij_to_cell(quad_ij$quad, quad_ij$i, quad_ij$j,
                                             resolution, aperture = ap)

    expect_equal(cell_direct, cell_indirect,
                 info = sprintf("aperture %d", ap))
  }
})

# =============================================================================
# QUAD IJ TO XY
# =============================================================================

test_that("quad_ij_to_xy returns valid structure", {
  result <- hexify_quad_ij_to_xy(quad = 1, i = 100, j = 50,
                                  resolution = 10, aperture = 3)

  expect_true("quad_x" %in% names(result))
  expect_true("quad_y" %in% names(result))
  expect_equal(result$quad_x, 0.30864197530864212, tolerance = 1e-12)
  expect_equal(result$quad_y, 0.17819452752766235, tolerance = 1e-12)
})

# =============================================================================
# ICOSA TRI TO QUAD XY
# =============================================================================

test_that("icosa_tri_to_quad_xy returns valid structure", {
  hexify_build_icosa()

  fwd <- hexify_forward(lon = 2.35, lat = 48.86)

  result <- hexify_icosa_tri_to_quad_xy(
    icosa_triangle_face = fwd["face"],
    icosa_triangle_x = fwd["icosa_triangle_x"],
    icosa_triangle_y = fwd["icosa_triangle_y"]
  )

  expect_true("quad" %in% names(result))
  expect_true("quad_x" %in% names(result))
  expect_true("quad_y" %in% names(result))
  expect_true(result$quad >= 0 && result$quad <= 11)
})

# =============================================================================
# ICOSA TRI TO QUAD IJ
# =============================================================================

test_that("icosa_tri_to_quad_ij returns valid structure", {
  hexify_build_icosa()

  fwd <- hexify_forward(lon = 2.35, lat = 48.86)

  result <- hexify_icosa_tri_to_quad_ij(
    icosa_triangle_face = fwd["face"],
    icosa_triangle_x = fwd["icosa_triangle_x"],
    icosa_triangle_y = fwd["icosa_triangle_y"],
    resolution = 10,
    aperture = 3
  )

  expect_true("quad" %in% names(result))
  expect_true("i" %in% names(result))
  expect_true("j" %in% names(result))
  expect_true(result$quad >= 0 && result$quad <= 11)
})

# =============================================================================
# QUAD XY TO ICOSA TRI (INVERSE)
# =============================================================================

test_that("quad_xy_to_icosa_tri returns valid structure", {
  result <- hexify_quad_xy_to_icosa_tri(quad = 1, quad_x = 0.5, quad_y = 0.3)

  expect_true("icosa_triangle_face" %in% names(result))
  expect_true("icosa_triangle_x" %in% names(result))
  expect_true("icosa_triangle_y" %in% names(result))
  expect_true(result$icosa_triangle_face >= 0 && result$icosa_triangle_face <= 19)
})

test_that("icosa_tri <-> quad_xy round-trip is consistent", {
  hexify_build_icosa()

  # Forward: lon/lat -> icosa tri -> quad xy
  fwd <- hexify_forward(lon = 10, lat = 45)

  quad_xy <- hexify_icosa_tri_to_quad_xy(
    icosa_triangle_face = fwd["face"],
    icosa_triangle_x = fwd["icosa_triangle_x"],
    icosa_triangle_y = fwd["icosa_triangle_y"]
  )

  # Inverse: quad xy -> icosa tri
  back <- hexify_quad_xy_to_icosa_tri(
    quad = quad_xy$quad,
    quad_x = quad_xy$quad_x,
    quad_y = quad_xy$quad_y
  )

  # Should recover original icosa tri coordinates
  expect_equal(as.numeric(fwd["face"]), back$icosa_triangle_face)
  expect_equal(as.numeric(fwd["icosa_triangle_x"]), back$icosa_triangle_x,
               tolerance = 1e-10)
  expect_equal(as.numeric(fwd["icosa_triangle_y"]), back$icosa_triangle_y,
               tolerance = 1e-10)
})

# =============================================================================
# QUAD IJ TO ICOSA TRI
# =============================================================================

test_that("quad_ij_to_icosa_tri returns valid structure", {
  result <- hexify_quad_ij_to_icosa_tri(quad = 1, i = 100, j = 50,
                                         resolution = 10, aperture = 3)

  expect_true("icosa_triangle_face" %in% names(result))
  expect_true("icosa_triangle_x" %in% names(result))
  expect_true("icosa_triangle_y" %in% names(result))
  expect_true(result$icosa_triangle_face >= 0 && result$icosa_triangle_face <= 19)
})

test_that("aperture-7 quad IJ -> XY lands on the cell centre", {
  # quad_ij_to_xy() used to scale aperture-7 coordinates by a sqrt(7)/sqrt(21)
  # substrate while the forward quantization used the exact 7^ceil(res/2) one,
  # so anything reaching it without the aperture-7 guard landed in the wrong
  # place (or outside the quad entirely).
  hexify_build_icosa()

  set.seed(57)
  lon <- runif(40, -170, 170)
  lat <- runif(40, -75, 75)

  for (res in 0:5) {
    g <- hex_grid(resolution = res, aperture = 7)
    cells <- lonlat_to_cell(lon, lat, g)
    ctr <- cell_to_lonlat(cells, g)
    qij <- cpp_cell_to_quad_ij(cells, res, 7L)

    # Polar quads are pentagons that cell_to_lonlat() answers directly.
    keep <- which(qij$quad >= 1 & qij$quad <= 10 & !is.na(ctr$lon))
    expect_gt(length(keep), 0)

    for (k in keep) {
      xy <- cpp_quad_ij_to_xy(qij$quad[k], qij$i[k], qij$j[k], 7L, res)
      tri <- cpp_quad_xy_to_icosa_tri(qij$quad[k], xy$quad_x, xy$quad_y)
      ll <- cpp_face_xy_to_ll(tri[["icosa_triangle_x"]],
                              tri[["icosa_triangle_y"]],
                              tri[["icosa_triangle_face"]])

      expect_equal(as.numeric(ll[["lon"]]), ctr$lon[k], tolerance = 1e-9,
                   info = sprintf("res=%d quad=%d", res, qij$quad[k]))
      expect_equal(as.numeric(ll[["lat"]]), ctr$lat[k], tolerance = 1e-9,
                   info = sprintf("res=%d quad=%d", res, qij$quad[k]))
    }
  }
})

# =============================================================================
# PIPELINE CONSISTENCY
# =============================================================================

test_that("coordinate pipeline is consistent end-to-end", {
  hexify_build_icosa()

  lon <- 16.37
  lat <- 48.21
  resolution <- 10
  aperture <- 3

  # Full pipeline: lon/lat -> quad IJ -> cell ID -> quad IJ (back)
  quad_ij <- hexify_lonlat_to_quad_ij(lon, lat, resolution, aperture)
  cell_id <- hexify_quad_ij_to_cell(quad_ij$quad, quad_ij$i, quad_ij$j,
                                     resolution, aperture)

  quad_ij_back <- hexify_cell_to_quad_ij(cell_id, resolution, aperture)

  expect_equal(quad_ij$quad, quad_ij_back$quad)
  expect_equal(quad_ij$i, quad_ij_back$i)
  expect_equal(quad_ij$j, quad_ij_back$j)
})

test_that("all apertures have consistent coordinate pipeline", {
  skip_on_cran()  # Detailed loop test
  hexify_build_icosa()

  lon <- 0
  lat <- 45

  for (ap in c(3, 4, 7)) {
    res <- if (ap == 7) 5 else 10

    # lon/lat -> cell -> lonlat
    cell_id <- hexify_lonlat_to_cell(lon, lat, res, aperture = ap)
    coords <- hexify_cell_to_lonlat(cell_id, res, aperture = ap)

    # Should be reasonably close
    lon_error <- abs(coords$lon_deg - lon)
    if (lon_error > 180) lon_error <- 360 - lon_error
    lat_error <- abs(coords$lat_deg - lat)

    max_error <- if (ap == 7) 5.0 else 2.0
    expect_true(lon_error < max_error,
                info = sprintf("aperture %d lon error", ap))
    expect_true(lat_error < max_error,
                info = sprintf("aperture %d lat error", ap))
  }
})

# =============================================================================
# INPUT VALIDATION
# =============================================================================

test_that("hexify_lonlat_to_quad_ij validates aperture", {
  expect_error(
    hexify_lonlat_to_quad_ij(lon = 0, lat = 0, resolution = 5, aperture = 5),
    "3, 4, 7"
  )
})

test_that("hexify_lonlat_to_quad_ij validates resolution", {
  expect_error(
    hexify_lonlat_to_quad_ij(lon = 0, lat = 0, resolution = -1, aperture = 3),
    "between 0 and 30"
  )
  expect_error(
    hexify_lonlat_to_quad_ij(lon = 0, lat = 0, resolution = 31, aperture = 3),
    "between 0 and 30"
  )
})

test_that("hexify_quad_ij_to_cell validates aperture", {
  expect_error(
    hexify_quad_ij_to_cell(quad = 0, i = 1, j = 1, resolution = 5, aperture = 5),
    "3, 4, 7"
  )
})

test_that("hexify_quad_ij_to_cell validates resolution", {
  expect_error(
    hexify_quad_ij_to_cell(quad = 0, i = 1, j = 1, resolution = -1, aperture = 3),
    "between 0 and 30"
  )
})

test_that("hexify_quad_ij_to_xy validates aperture", {
  expect_error(
    hexify_quad_ij_to_xy(quad = 1, i = 10, j = 5, resolution = 5, aperture = 5),
    "3, 4, 7"
  )
})

test_that("hexify_icosa_tri_to_quad_ij validates aperture", {
  expect_error(
    hexify_icosa_tri_to_quad_ij(
      icosa_triangle_face = 0,
      icosa_triangle_x = 0.5,
      icosa_triangle_y = 0.5,
      resolution = 5,
      aperture = 5
    ),
    "3, 4, 7"
  )
})

# =============================================================================
# CELL TO QUAD IJ
# =============================================================================

test_that("hexify_cell_to_quad_ij returns correct structure", {
  hexify_build_icosa()

  result <- hexify_cell_to_quad_ij(cell_id = 1000, resolution = 5, aperture = 3)

  expect_s3_class(result, "data.frame")
  expect_true("quad" %in% names(result))
  expect_true("i" %in% names(result))
  expect_true("j" %in% names(result))
})

test_that("hexify_cell_to_quad_ij validates aperture", {
  expect_error(
    hexify_cell_to_quad_ij(cell_id = 100, resolution = 5, aperture = 5),
    "3, 4, 7"
  )
})

test_that("hexify_cell_to_quad_ij validates resolution", {
  expect_error(
    hexify_cell_to_quad_ij(cell_id = 100, resolution = -1, aperture = 3),
    "between 0 and 30"
  )
})

# =============================================================================
# CELL TO ICOSA TRIANGLE
# =============================================================================

test_that("hexify_cell_to_icosa_tri returns correct structure", {
  hexify_build_icosa()

  result <- hexify_cell_to_icosa_tri(cell_id = 1000, resolution = 5, aperture = 3)

  expect_s3_class(result, "data.frame")
  expect_true("icosa_triangle_face" %in% names(result))
  expect_true("icosa_triangle_x" %in% names(result))
  expect_true("icosa_triangle_y" %in% names(result))
})

test_that("hexify_cell_to_icosa_tri validates aperture", {
  expect_error(
    hexify_cell_to_icosa_tri(cell_id = 100, resolution = 5, aperture = 5),
    "3, 4, 7"
  )
})

test_that("hexify_cell_to_icosa_tri validates resolution", {
  expect_error(
    hexify_cell_to_icosa_tri(cell_id = 100, resolution = -1, aperture = 3),
    "between 0 and 30"
  )
})

test_that("hexify_quad_ij_to_icosa_tri validates aperture", {
  expect_error(
    hexify_quad_ij_to_icosa_tri(
      quad = 1, i = 10, j = 5, resolution = 5, aperture = 5
    ),
    "3, 4, 7"
  )
})

test_that("hexify_quad_ij_to_icosa_tri validates resolution", {
  expect_error(
    hexify_quad_ij_to_icosa_tri(
      quad = 1, i = 10, j = 5, resolution = -1, aperture = 3
    ),
    "between 0 and 30"
  )
})

# =============================================================================
# CELL TO QUAD XY (CONTINUOUS COORDINATES)
# =============================================================================

test_that("hexify_cell_to_quad_xy returns correct structure", {
  hexify_build_icosa()

  result <- hexify_cell_to_quad_xy(cell_id = 1000, resolution = 5, aperture = 3)

  expect_s3_class(result, "data.frame")
  expect_true("quad" %in% names(result))
  expect_true("quad_x" %in% names(result))
  expect_true("quad_y" %in% names(result))
})

test_that("hexify_cell_to_quad_xy validates aperture", {
  expect_error(
    hexify_cell_to_quad_xy(cell_id = 100, resolution = 5, aperture = 5),
    "3, 4, 7"
  )
})

test_that("hexify_cell_to_quad_xy validates resolution", {
  expect_error(
    hexify_cell_to_quad_xy(cell_id = 100, resolution = -1, aperture = 3),
    "between 0 and 30"
  )
})

test_that("hexify_quad_xy_to_cell returns valid cell ID", {
  hexify_build_icosa()

  cell_id <- hexify_quad_xy_to_cell(
    quad = 1, quad_x = 0.5, quad_y = 0.3,
    resolution = 5, aperture = 3
  )

  expect_true(is.numeric(cell_id))
  expect_equal(cell_id, 167)
})

test_that("hexify_quad_xy_to_cell validates aperture", {
  expect_error(
    hexify_quad_xy_to_cell(
      quad = 1, quad_x = 0.5, quad_y = 0.3, resolution = 5, aperture = 5
    ),
    "3, 4, 7"
  )
})

test_that("hexify_quad_xy_to_cell validates resolution", {
  expect_error(
    hexify_quad_xy_to_cell(
      quad = 1, quad_x = 0.5, quad_y = 0.3, resolution = -1, aperture = 3
    ),
    "between 0 and 30"
  )
})

test_that("quad_xy round-trip is consistent", {
  hexify_build_icosa()

  cell_id <- 1000

  # cell -> quad_xy -> cell
  result1 <- hexify_cell_to_quad_xy(cell_id = cell_id, resolution = 5, aperture = 3)
  cell_id2 <- hexify_quad_xy_to_cell(
    quad = result1$quad, quad_x = result1$quad_x, quad_y = result1$quad_y,
    resolution = 5, aperture = 3
  )

  expect_equal(cell_id, cell_id2)
})

# =============================================================================
# PLANE COORDINATE CONVERSIONS
# =============================================================================

test_that("hexify_icosa_tri_to_plane returns correct structure", {
  hexify_build_icosa()

  fwd <- hexify_forward(lon = 10, lat = 50)

  result <- hexify_icosa_tri_to_plane(
    icosa_triangle_face = fwd["face"],
    icosa_triangle_x = fwd["icosa_triangle_x"],
    icosa_triangle_y = fwd["icosa_triangle_y"]
  )

  expect_s3_class(result, "data.frame")
  expect_true("plane_x" %in% names(result))
  expect_true("plane_y" %in% names(result))
})

test_that("hexify_cell_to_plane returns correct structure", {
  hexify_build_icosa()

  result <- hexify_cell_to_plane(cell_id = 1000, resolution = 5, aperture = 3)

  expect_s3_class(result, "data.frame")
  expect_true("plane_x" %in% names(result))
  expect_true("plane_y" %in% names(result))
})

test_that("hexify_cell_to_plane validates aperture", {
  expect_error(
    hexify_cell_to_plane(cell_id = 100, resolution = 5, aperture = 5),
    "3, 4, 7"
  )
})

test_that("hexify_cell_to_plane validates resolution", {
  expect_error(
    hexify_cell_to_plane(cell_id = 100, resolution = -1, aperture = 3),
    "between 0 and 30"
  )
})

test_that("hexify_lonlat_to_plane returns correct structure", {
  hexify_build_icosa()

  result <- hexify_lonlat_to_plane(lon = 10, lat = 50)

  expect_s3_class(result, "data.frame")
  expect_true("plane_x" %in% names(result))
  expect_true("plane_y" %in% names(result))
})

test_that("hexify_lonlat_to_plane works for multiple points", {
  hexify_build_icosa()

  lons <- c(0, 45, -120, 180, -180)
  lats <- c(0, 30, -45, 60, -60)

  result <- hexify_lonlat_to_plane(lon = lons, lat = lats)

  expect_equal(nrow(result), 5)
  expect_true(all(is.finite(result$plane_x)))
  expect_true(all(is.finite(result$plane_y)))
})

test_that("hexify_cell_to_plane handles multiple cells", {
  hexify_build_icosa()

  result <- hexify_cell_to_plane(
    cell_id = c(100, 200, 300), resolution = 5, aperture = 3
  )

  expect_equal(nrow(result), 3)
  expect_true(all(is.finite(result$plane_x)))
  expect_true(all(is.finite(result$plane_y)))
})

# =============================================================================
# CROSS-APERTURE PLANE TESTS
# =============================================================================

test_that("plane conversions work for all apertures", {
  skip_on_cran()  # Loop test across apertures
  hexify_build_icosa()

  for (ap in c(3, 4, 7)) {
    # cell -> plane
    plane <- hexify_cell_to_plane(cell_id = 100, resolution = 5, aperture = ap)
    expect_true(is.finite(plane$plane_x))
    expect_true(is.finite(plane$plane_y))
  }
})

# =============================================================================
# HIERARCHICAL INDEX FUNCTIONS
# =============================================================================

test_that("hexify_lonlat_to_h_index returns correct structure", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  result <- hexify_lonlat_to_h_index(grid, lon = 0, lat = 45)

  expect_s3_class(result, "data.frame")
  expect_true("h_index" %in% names(result))
  expect_true("face" %in% names(result))
  expect_type(result$h_index, "character")
  expect_type(result$face, "integer")
})

test_that("hexify_lonlat_to_h_index handles multiple points", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  lons <- c(0, 10, -5)
  lats <- c(45, 30, -20)

  result <- hexify_lonlat_to_h_index(grid, lon = lons, lat = lats)

  expect_equal(nrow(result), 3)
  expect_true(all(nchar(result$h_index) > 0))
})

test_that("hexify_lonlat_to_h_index handles NA values", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  result <- hexify_lonlat_to_h_index(grid, lon = c(0, NA), lat = c(45, NA))

  expect_equal(nrow(result), 2)
  expect_true(is.na(result$h_index[2]))
  expect_true(is.na(result$face[2]))
})

test_that("hexify_lonlat_to_h_index validates grid object", {
  expect_error(
    hexify_lonlat_to_h_index(list(x = 1), lon = 0, lat = 0),
    "hexify_grid object"
  )
})

test_that("hexify_lonlat_to_h_index validates input lengths", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  expect_error(
    hexify_lonlat_to_h_index(grid, lon = c(0, 1), lat = 0),
    "same length"
  )
})

test_that("hexify_lonlat_to_h_index validates numeric input", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  expect_error(
    hexify_lonlat_to_h_index(grid, lon = "a", lat = "b"),
    "must be numeric"
  )
})

test_that("hexify_lonlat_to_h_index warns on out-of-range coordinates", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  expect_warning(
    hexify_lonlat_to_h_index(grid, lon = 200, lat = 45),
    "outside valid range"
  )

  expect_warning(
    hexify_lonlat_to_h_index(grid, lon = 0, lat = 100),
    "outside valid range"
  )
})

test_that("hexify_h_index_to_lonlat returns correct structure", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  # First get an index
  h_result <- hexify_lonlat_to_h_index(grid, lon = 5, lat = 45)

  # Then convert back
  result <- hexify_h_index_to_lonlat(grid, h_result$h_index)

  expect_s3_class(result, "data.frame")
  expect_true("lon" %in% names(result))
  expect_true("lat" %in% names(result))
  expect_type(result$lon, "double")
  expect_type(result$lat, "double")
})

test_that("hexify_h_index_to_lonlat handles NA values", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  # Get a valid h_index first
  valid_index <- hexify_lonlat_to_h_index(grid, lon = 5, lat = 45)$h_index

  result <- hexify_h_index_to_lonlat(grid, c(valid_index, NA_character_))

  expect_equal(nrow(result), 2)
  expect_true(!is.na(result$lon[1]))
  expect_true(is.na(result$lon[2]))
  expect_true(is.na(result$lat[2]))
})

test_that("hexify_h_index_to_lonlat validates grid object", {
  expect_error(
    hexify_h_index_to_lonlat(list(x = 1), "0100000"),
    "hexify_grid object"
  )
})

test_that("hexify_h_index_to_lonlat validates character input", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  expect_error(
    hexify_h_index_to_lonlat(grid, 12345),
    "must be a character"
  )
})

test_that("h_index round-trip is consistent", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  lon <- 10
  lat <- 45

  # Forward
  h_result <- hexify_lonlat_to_h_index(grid, lon = lon, lat = lat)

  # Inverse
  coords <- hexify_h_index_to_lonlat(grid, h_result$h_index)

  # Should be close (within cell distance)
  expect_true(abs(coords$lon - lon) < 5)
  expect_true(abs(coords$lat - lat) < 5)
})

# =============================================================================
# GRID-BASED WRAPPERS
# =============================================================================

test_that("hexify_grid_to_cell returns cell IDs", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  result <- hexify_grid_to_cell(grid, lon = c(0, 10), lat = c(45, 50))

  expect_type(result, "double")
  expect_length(result, 2)
  expect_true(all(result > 0))
})

test_that("hexify_grid_to_cell validates grid object", {
  expect_error(
    hexify_grid_to_cell(list(x = 1), lon = 0, lat = 0),
    "hexify_grid object"
  )
})

test_that("hexify_grid_cell_to_lonlat returns coordinates", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  # Get some cell IDs first
  cell_ids <- hexify_grid_to_cell(grid, lon = c(0, 10), lat = c(45, 50))

  # Convert back
  result <- hexify_grid_cell_to_lonlat(grid, cell_ids)

  expect_s3_class(result, "data.frame")
  expect_true("lon_deg" %in% names(result))
  expect_true("lat_deg" %in% names(result))
  expect_equal(nrow(result), 2)
})

test_that("hexify_grid_cell_to_lonlat validates grid object", {
  expect_error(
    hexify_grid_cell_to_lonlat(list(x = 1), cell_id = 1000),
    "hexify_grid object"
  )
})

test_that("grid wrapper round-trip is consistent", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  lon <- 5
  lat <- 45

  # Forward
  cell_id <- hexify_grid_to_cell(grid, lon = lon, lat = lat)

  # Inverse
  coords <- hexify_grid_cell_to_lonlat(grid, cell_id)

  # Should be close
  expect_true(abs(coords$lon_deg - lon) < 5)
  expect_true(abs(coords$lat_deg - lat) < 5)
})

# =============================================================================
# ROUNDTRIP TEST FUNCTION
# =============================================================================

test_that("hexify_roundtrip_test returns correct structure", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  result <- hexify_roundtrip_test(grid, lon = 5, lat = 45)

  expect_type(result, "list")
  expect_true("original" %in% names(result))
  expect_true("h_index" %in% names(result))
  expect_true("reconstructed" %in% names(result))
  expect_true("error" %in% names(result))
  expect_true("units" %in% names(result))
})

test_that("hexify_roundtrip_test reports error in km", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  result <- hexify_roundtrip_test(grid, lon = 5, lat = 45, units = "km")

  expect_equal(result$units, "km")
  expect_true(is.numeric(result$error))
  expect_true(result$error >= 0)
})

test_that("hexify_roundtrip_test reports error in degrees", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  result <- hexify_roundtrip_test(grid, lon = 5, lat = 45, units = "degrees")

  expect_equal(result$units, "degrees")
  expect_true(is.numeric(result$error))
  expect_true(result$error >= 0)
})

test_that("hexify_roundtrip_test error is reasonable", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  result <- hexify_roundtrip_test(grid, lon = 5, lat = 45, units = "km")

  # Error should be less than the cell diagonal (for ~1000 km2 area, diagonal ~ 35-40 km)
  expect_true(result$error < 50)
})

# tests/testthat/test-dggrid-compat.R
# Tests verifying hexify produces identical output to dggridR
#
# These tests compare hexify results against dggridR reference data to ensure
# compatibility. The reference data was generated using dggridR.

# =============================================================================
# SETUP
# =============================================================================

setup_icosa <- function() {
  cpp_build_icosa()
}

load_validation_data <- function(aperture, resolution) {
  file <- sprintf("data/dggrid_validation_ap%d_res%d.csv", aperture, resolution)
  read.csv(test_path(file))
}

# =============================================================================
# DGGRID VALIDATION TESTS - APERTURE 3
# =============================================================================

test_that("hexify matches dggridR for aperture 3, resolution 3", {
  setup_icosa()

  ref <- load_validation_data(3, 3)

  for (i in seq_len(nrow(ref))) {
    result <- hexify_lonlat_to_quad_ij(
      lon = ref$lon[i],
      lat = ref$lat[i],
      resolution = 3,
      aperture = 3
    )

    expect_equal(result$quad, ref$quad[i],
                 info = sprintf("Row %d: quad mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
    expect_equal(result$i, ref$i[i],
                 info = sprintf("Row %d: i mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
    expect_equal(result$j, ref$j[i],
                 info = sprintf("Row %d: j mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
  }
})

test_that("hexify matches dggridR for aperture 3, resolution 5", {
  setup_icosa()

  ref <- load_validation_data(3, 5)

  for (i in seq_len(nrow(ref))) {
    result <- hexify_lonlat_to_quad_ij(
      lon = ref$lon[i],
      lat = ref$lat[i],
      resolution = 5,
      aperture = 3
    )

    expect_equal(result$quad, ref$quad[i],
                 info = sprintf("Row %d: quad mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
    expect_equal(result$i, ref$i[i],
                 info = sprintf("Row %d: i mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
    expect_equal(result$j, ref$j[i],
                 info = sprintf("Row %d: j mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
  }
})

test_that("hexify matches dggridR for aperture 3, resolution 7", {
  setup_icosa()

  ref <- load_validation_data(3, 7)

  for (i in seq_len(nrow(ref))) {
    result <- hexify_lonlat_to_quad_ij(
      lon = ref$lon[i],
      lat = ref$lat[i],
      resolution = 7,
      aperture = 3
    )

    expect_equal(result$quad, ref$quad[i],
                 info = sprintf("Row %d: quad mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
    expect_equal(result$i, ref$i[i],
                 info = sprintf("Row %d: i mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
    expect_equal(result$j, ref$j[i],
                 info = sprintf("Row %d: j mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
  }
})

# =============================================================================
# VECTORIZED CELL ID TESTS
# =============================================================================

test_that("hexify batch cell_id matches dggridR for aperture 3, resolution 3", {
  setup_icosa()

  ref <- load_validation_data(3, 3)

  # Get hexify cell IDs (vectorized)
  hexify_cells <- hexify_lonlat_to_cell(
    lon = ref$lon,
    lat = ref$lat,
    resolution = 3,
    aperture = 3
  )

  # Convert dggridR quad/i/j to cell ID for comparison
  dggrid_cells <- hexify_quad_ij_to_cell(
    quad = ref$quad,
    i = ref$i,
    j = ref$j,
    resolution = 3,
    aperture = 3
  )

  expect_equal(hexify_cells, dggrid_cells)
})

test_that("hexify batch cell_id matches dggridR for aperture 3, resolution 5", {
  setup_icosa()

  ref <- load_validation_data(3, 5)

  hexify_cells <- hexify_lonlat_to_cell(
    lon = ref$lon,
    lat = ref$lat,
    resolution = 5,
    aperture = 3
  )

  dggrid_cells <- hexify_quad_ij_to_cell(
    quad = ref$quad,
    i = ref$i,
    j = ref$j,
    resolution = 5,
    aperture = 3
  )

  expect_equal(hexify_cells, dggrid_cells)
})

test_that("hexify batch cell_id matches dggridR for aperture 3, resolution 7", {
  setup_icosa()

  ref <- load_validation_data(3, 7)

  hexify_cells <- hexify_lonlat_to_cell(
    lon = ref$lon,
    lat = ref$lat,
    resolution = 7,
    aperture = 3
  )

  dggrid_cells <- hexify_quad_ij_to_cell(
    quad = ref$quad,
    i = ref$i,
    j = ref$j,
    resolution = 7,
    aperture = 3
  )

  expect_equal(hexify_cells, dggrid_cells)
})

# =============================================================================
# CELL ID VALIDATION TESTS
# =============================================================================

test_that("hexify cell_id matches dggridR cell IDs for aperture 3", {
  setup_icosa()

  for (res in c(3, 5, 7)) {
    ref <- load_validation_data(3, res)

    # Get hexify cell IDs
    hexify_cells <- hexify_lonlat_to_cell(
      lon = ref$lon,
      lat = ref$lat,
      resolution = res,
      aperture = 3
    )

    # Convert dggridR quad/i/j to cell ID for comparison
    dggrid_cells <- hexify_quad_ij_to_cell(
      quad = ref$quad,
      i = ref$i,
      j = ref$j,
      resolution = res,
      aperture = 3
    )

    expect_equal(hexify_cells, dggrid_cells,
                 info = sprintf("Cell ID mismatch for aperture 3, resolution %d", res))
  }
})

# =============================================================================
# ICOSAHEDRON FACE VALIDATION
# =============================================================================

test_that("hexify face assignment matches dggridR tnum", {
  setup_icosa()

  for (res in c(3, 5, 7)) {
    ref <- load_validation_data(3, res)

    for (i in seq_len(nrow(ref))) {
      face <- cpp_which_face(ref$lon[i], ref$lat[i])

      expect_equal(face, ref$tnum[i],
                   info = sprintf("Res %d, row %d: face mismatch for lon=%.4f, lat=%.4f (got %d, expected %d)",
                                  res, i, ref$lon[i], ref$lat[i], face, ref$tnum[i]))
    }
  }
})

# =============================================================================
# PROJECTION VALIDATION (tx, ty)
# =============================================================================

test_that("hexify projection matches dggridR tx/ty", {
  setup_icosa()

  ref <- load_validation_data(3, 5)  # Use res 5 data

  for (i in seq_len(nrow(ref))) {
    proj <- cpp_snyder_forward(ref$lon[i], ref$lat[i])
    face <- unname(proj["face"])
    tx <- unname(proj["icosa_triangle_x"])
    ty <- unname(proj["icosa_triangle_y"])

    # Face should match
    expect_equal(face, ref$tnum[i],
                 info = sprintf("Row %d: face mismatch", i))

    # tx/ty should be close (allow tolerance for floating point differences)
    expect_equal(tx, ref$tx[i], tolerance = 1e-6,
                 info = sprintf("Row %d: tx mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
    expect_equal(ty, ref$ty[i], tolerance = 1e-6,
                 info = sprintf("Row %d: ty mismatch for lon=%.4f, lat=%.4f",
                                i, ref$lon[i], ref$lat[i]))
  }
})

# =============================================================================
# ROUND-TRIP CONSISTENCY TESTS
# =============================================================================

test_that("lon/lat -> cell -> lon/lat round-trip lands in same cell", {
  setup_icosa()

  ref <- load_validation_data(3, 5)

  for (res in c(3, 5, 7)) {
    # Get cell IDs
    cells <- hexify_lonlat_to_cell(
      lon = ref$lon,
      lat = ref$lat,
      resolution = res,
      aperture = 3
    )

    # Get cell centers
    centers <- hexify_cell_to_lonlat(
      cell_id = cells,
      resolution = res,
      aperture = 3
    )

    # Convert centers back to cells
    cells2 <- hexify_lonlat_to_cell(
      lon = centers$lon_deg,
      lat = centers$lat_deg,
      resolution = res,
      aperture = 3
    )

    expect_equal(cells, cells2,
                 info = sprintf("Round-trip failed for resolution %d", res))
  }
})

test_that("quad/i/j -> cell -> quad/i/j round-trip is exact", {
  setup_icosa()

  ref <- load_validation_data(3, 5)

  for (res in c(3, 5, 7)) {
    ref_res <- load_validation_data(3, res)

    # quad/i/j -> cell
    cells <- hexify_quad_ij_to_cell(
      quad = ref_res$quad,
      i = ref_res$i,
      j = ref_res$j,
      resolution = res,
      aperture = 3
    )

    # cell -> quad/i/j
    result <- hexify_cell_to_quad_ij(
      cell_id = cells,
      resolution = res,
      aperture = 3
    )

    expect_equal(result$quad, ref_res$quad,
                 info = sprintf("Quad mismatch for resolution %d", res))
    expect_equal(result$i, ref_res$i,
                 info = sprintf("i mismatch for resolution %d", res))
    expect_equal(result$j, ref_res$j,
                 info = sprintf("j mismatch for resolution %d", res))
  }
})

## ----setup, include = FALSE---------------------------------------------------
source("_common.R")
library(hexify)
library(sf)
library(ggplot2)

## ----one-grid-many-datasets---------------------------------------------------
# Step 1: Define the grid once
# This is your shared spatial reference system
grid <- hex_grid(area_km2 = 5000)
print(grid)

# Step 2: Create multiple datasets with different structures
set.seed(123)

# Dataset 1: Bird observations (note different column names)
bird_obs <- data.frame(
  species = sample(c("Passer domesticus", "Turdus merula", "Parus major"), 50, replace = TRUE),
  longitude = runif(50, -10, 30),
  latitude = runif(50, 35, 60)
)

# Dataset 2: Mammal records
mammal_obs <- data.frame(
  species = sample(c("Vulpes vulpes", "Meles meles", "Sciurus vulgaris"), 40, replace = TRUE),
  lon = runif(40, -10, 30),
  lat = runif(40, 35, 60)
)

# Dataset 3: Climate stations
climate_data <- data.frame(
  station_id = paste0("WS", 1:20),
  x = runif(20, -10, 30),
  y = runif(20, 35, 60),
  temp_c = rnorm(20, 12, 5)
)

# Step 3: Attach all datasets to the SAME grid
birds <- hexify(bird_obs, lon = "longitude", lat = "latitude", grid = grid)
mammals <- hexify(mammal_obs, lon = "lon", lat = "lat", grid = grid)
climate <- hexify(climate_data, lon = "x", lat = "y", grid = grid)

cat("Birds:  ", nrow(as.data.frame(birds)), "observations in", n_cells(birds), "cells\n")
cat("Mammals:", nrow(as.data.frame(mammals)), "observations in", n_cells(mammals), "cells\n")
cat("Climate:", nrow(as.data.frame(climate)), "stations in", n_cells(climate), "cells\n")

## ----cell-level-analysis------------------------------------------------------
# Extract data frames with cell IDs
birds_df <- as.data.frame(birds)
birds_df$cell_id <- birds@cell_id

mammals_df <- as.data.frame(mammals)
mammals_df$cell_id <- mammals@cell_id

climate_df <- as.data.frame(climate)
climate_df$cell_id <- climate@cell_id

# Aggregate each dataset by cell
bird_richness <- aggregate(
  species ~ cell_id,
  data = birds_df,
  FUN = function(x) length(unique(x))
)
names(bird_richness)[2] <- "bird_species"

mammal_richness <- aggregate(
  species ~ cell_id,
  data = mammals_df,
  FUN = function(x) length(unique(x))
)
names(mammal_richness)[2] <- "mammal_species"

mean_temp <- aggregate(
  temp_c ~ cell_id,
  data = climate_df,
  FUN = mean
)
names(mean_temp)[2] <- "mean_temp"

# Join datasets by cell_id - guaranteed to align because same grid
combined <- merge(bird_richness, mammal_richness, by = "cell_id", all = TRUE)
combined <- merge(combined, mean_temp, by = "cell_id", all = TRUE)

head(combined)

## ----grid-rect, fig.width=7, fig.height=5-------------------------------------
# Create grid specification
grid <- hex_grid(area_km2 = 5000)

# Generate hexagons over Western Europe
europe_hexes <- grid_rect(c(-10, 35, 25, 60), grid)

# Get European countries for context
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_hexes, fill = NA, color = "steelblue", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE)

## ----grid-polygon, fig.width=7, fig.height=7----------------------------------
# Get France boundary
france <- hexify_world[hexify_world$name == "France", ]

# Generate grid covering mainland France
grid <- hex_grid(area_km2 = 2000)
france_grid <- grid_rect(c(-5, 41, 10, 52), grid)

# Clip grid to France boundary
france_grid_clipped <- st_intersection(france_grid, st_geometry(france))

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_grid_clipped, fill = alpha("steelblue", 0.3),
          color = "steelblue", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = "Hexagonal Grid Clipped to France",
       subtitle = sprintf("~%.0f km² cells", grid@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE)

## ----grid-global, fig.width=7, fig.height=3.5, warning=FALSE------------------
# Coarse global grid (be careful with fine grids - many cells!)
grid <- hex_grid(area_km2 = 500000)
global_hexes <- grid_global(grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray90", color = "gray70", linewidth = 0.2) +
  geom_sf(data = global_hexes, fill = NA, color = "darkgreen", linewidth = 0.3) +
  labs(title = sprintf("Global Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----multi-resolution---------------------------------------------------------
# Sample data
set.seed(42)
observations <- data.frame(
  species = sample(c("Species A", "Species B", "Species C"), 100, replace = TRUE),
  lon = runif(100, -10, 30),
  lat = runif(100, 35, 60)
)

# Fine resolution (~1000 km² cells)
grid_fine <- hex_grid(area_km2 = 1000)
obs_fine <- hexify(observations, lon = "lon", lat = "lat", grid = grid_fine)

# Coarse resolution (~10000 km² cells)
grid_coarse <- hex_grid(area_km2 = 10000)
obs_coarse <- hexify(observations, lon = "lon", lat = "lat", grid = grid_coarse)

cat(sprintf("Fine resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_fine), grid_fine@area_km2))
cat(sprintf("Coarse resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_coarse), grid_coarse@area_km2))

## ----multi-scale-aggregate----------------------------------------------------
# Extract data with cell IDs
fine_df <- as.data.frame(obs_fine)
fine_df$cell_id <- obs_fine@cell_id

coarse_df <- as.data.frame(obs_coarse)
coarse_df$cell_id <- obs_coarse@cell_id

# Species richness at each scale
richness_fine <- aggregate(species ~ cell_id, data = fine_df,
                           FUN = function(x) length(unique(x)))
richness_coarse <- aggregate(species ~ cell_id, data = coarse_df,
                             FUN = function(x) length(unique(x)))

cat(sprintf("Fine scale: mean %.2f species per cell\n", mean(richness_fine$species)))
cat(sprintf("Coarse scale: mean %.2f species per cell\n", mean(richness_coarse$species)))

## ----spatial-joins------------------------------------------------------------
# Dataset 1: Weather stations
stations <- data.frame(
  station_id = paste0("ST", 1:50),
  lon = runif(50, -10, 30),
  lat = runif(50, 35, 60),
  temperature = rnorm(50, 15, 5)
)

# Dataset 2: Cities
cities <- data.frame(
  city = c("Vienna", "Paris", "London", "Berlin", "Rome",
           "Madrid", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -0.12, 13.40, 12.50,
          -3.70, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 51.51, 52.52, 41.90,
          40.42, 50.08, 52.23, 47.50, 52.37)
)

# Use a coarse grid for joining disparate points
grid <- hex_grid(area_km2 = 50000)

# Hexify both datasets with the same grid
stations_hex <- hexify(stations, lon = "lon", lat = "lat", grid = grid)
cities_hex <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Extract with cell IDs
stations_df <- as.data.frame(stations_hex)
stations_df$cell_id <- stations_hex@cell_id

cities_df <- as.data.frame(cities_hex)
cities_df$cell_id <- cities_hex@cell_id

# Join by cell_id
city_weather <- merge(
  cities_df[, c("city", "cell_id")],
  aggregate(temperature ~ cell_id, data = stations_df, FUN = mean),
  by = "cell_id",
  all.x = TRUE
)

city_weather

## ----cell-summarize-----------------------------------------------------------
# Aggregate the station data straight from the HexData object
hex_summarize(stations_hex, mean_temp = mean(temperature), n_stations = length(temperature))

## ----cell-neighbors-----------------------------------------------------------
# The 6 cells surrounding a given cell (k = 1), or a wider ring with k > 1
some_cell <- stations_hex@cell_id[1]
get_neighbors(some_cell, grid)
get_neighbors(some_cell, grid, k = 2)

## ----choosing-resolution------------------------------------------------------
# Target: 100 km² cells
grid_100 <- hex_grid(area_km2 = 100, aperture = 3)
cat(sprintf("Target ~100 km²: resolution %d (actual: %.1f km²)\n",
            grid_100@resolution, grid_100@area_km2))

# Target: 1000 km² cells
grid_1000 <- hex_grid(area_km2 = 1000, aperture = 3)
cat(sprintf("Target ~1000 km²: resolution %d (actual: %.1f km²)\n",
            grid_1000@resolution, grid_1000@area_km2))

# Target: 10000 km² cells
grid_10000 <- hex_grid(area_km2 = 10000, aperture = 3)
cat(sprintf("Target ~10000 km²: resolution %d (actual: %.1f km²)\n",
            grid_10000@resolution, grid_10000@area_km2))

## ----resolution-table, echo=FALSE---------------------------------------------
comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:15)
comparison$n_cells_fmt <- ifelse(
  comparison$n_cells > 1e6,
  sprintf("%.1fM", comparison$n_cells / 1e6),
  ifelse(comparison$n_cells > 1e3,
         sprintf("%.1fK", comparison$n_cells / 1e3),
         as.character(comparison$n_cells))
)
comparison$area_fmt <- ifelse(
  comparison$cell_area_km2 < 0.1,
  formatC(comparison$cell_area_km2, format = "e", digits = 1),
  formatC(comparison$cell_area_km2, format = "f", digits = 1, big.mark = ",")
)
comparison$spacing_fmt <- ifelse(
  comparison$cell_spacing_km < 0.1,
  formatC(comparison$cell_spacing_km, format = "e", digits = 1),
  formatC(comparison$cell_spacing_km, format = "f", digits = 1)
)
knitr::kable(
  comparison[, c("resolution", "n_cells_fmt", "area_fmt", "spacing_fmt")],
  col.names = c("Resolution", "# Cells", "Cell Area (km\u00b2)", "Spacing (km)"),
  align = c("r", "r", "r", "r")
)

## ----compare-apertures--------------------------------------------------------
target_area <- 1000  # km²

cat(sprintf("Target: ~%d km² cells\n\n", target_area))
for (ap in c(3, 4, 7)) {
  grid <- hex_grid(area_km2 = target_area, aperture = ap)
  n_cells <- 10 * (ap^grid@resolution) + 2
  cat(sprintf("Aperture %d: resolution %d -> %.1f km² (%s cells globally)\n",
              ap, grid@resolution, grid@area_km2,
              format(n_cells, big.mark = ",")))
}

## ----other-bodies-------------------------------------------------------------
mars <- hex_grid(area_km2 = 1000, radius_km = "mars")
mars

# The same grid, by radius
identical(mars@area_km2, hex_grid(resolution = mars@resolution, radius_km = 3389.5)@area_km2)

## ----other-bodies-geometry----------------------------------------------------
earth <- hex_grid(resolution = 5)
moon  <- hex_grid(resolution = 5, radius_km = "moon")

lon <- c(0, 16.37, -70.5)
lat <- c(0, 48.21, -33.4)

identical(lonlat_to_cell(lon, lat, moon), lonlat_to_cell(lon, lat, earth))
c(earth = earth@area_km2, moon = moon@area_km2)

## ----other-bodies-crs---------------------------------------------------------
mars@crs

## ----other-bodies-crs-custom--------------------------------------------------
world <- hex_grid(area_km2 = 5000, radius_km = 4200)
world@crs

## ----other-bodies-crs-explicit------------------------------------------------
hex_grid(area_km2 = 5000, radius_km = 4200,
         crs = "+proj=laea +lat_0=0 +lon_0=0 +R=4200000 +no_defs")@crs

## ----other-bodies-h3----------------------------------------------------------
h3_mars <- hex_grid(resolution = 5, type = "h3", radius_km = "mars")
h3_mars@area_km2 / hex_grid(resolution = 5, type = "h3")@area_km2

# H3 measures against the WGS84 authalic radius, so that is what divides out
(3389.5 / 6371.007180918475)^2

## ----sf-workflow--------------------------------------------------------------
# Hexify some data
grid <- hex_grid(area_km2 = 20000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Convert to sf points (uses cell centers)
sf_points <- as_sf(result, geometry = "point")
class(sf_points)

# Convert to sf polygons (for choropleth maps)
sf_polys <- as_sf(result, geometry = "polygon")
class(sf_polys)

# Or generate polygons directly from cell IDs
unique_cells <- cells(result)
cell_polys <- cell_to_sf(unique_cells, grid)

## ----sf-plot, fig.width=7, fig.height=5---------------------------------------
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "ivory", color = "gray70") +
  geom_sf(data = cell_polys, fill = "steelblue", alpha = 0.5, color = "darkblue") +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 58)) +
  labs(title = "European Cities - Hexagonal Grid") +
  theme_minimal(base_size = FIG_BASE_SIZE)

## ----export-formats, eval=FALSE-----------------------------------------------
# # Generate a grid
# grid <- hex_grid(area_km2 = 10000)
# europe <- grid_rect(c(-10, 35, 25, 60), grid)
# 
# # Export to various formats
# st_write(europe, "europe_grid.gpkg", layer = "hexgrid")     # GeoPackage
# st_write(europe, "europe_grid.shp")                         # Shapefile
# st_write(europe, "europe_grid.geojson")                     # GeoJSON
# st_write(europe, "europe_grid.kml", layer = "hexgrid")      # KML (Google Earth)

## ----edge-cases---------------------------------------------------------------
data_with_na <- data.frame(
  lon = c(16.37, NA, 2.35, 13.40),
  lat = c(48.21, 48.86, NA, 52.52)
)

grid <- hex_grid(area_km2 = 1000)
result <- hexify(data_with_na, lon = "lon", lat = "lat", grid = grid)

# Check which rows have valid cell assignments
cat("Cell IDs:", result@cell_id, "\n")
cat("NA indicates invalid coordinates\n")

# hexify_map.R
# Visualization functions with basemap support

# =============================================================================
# INTERNAL HELPER FUNCTIONS
# =============================================================================

#' Calculate view buffer for bounding box
#' @noRd
calculate_view_buffer <- function(bbox, factor = 0.1, min_buffer = 1) {
  x_range <- bbox["xmax"] - bbox["xmin"]
  y_range <- bbox["ymax"] - bbox["ymin"]
  list(
    x = max(x_range * factor, min_buffer),
    y = max(y_range * factor, min_buffer)
  )
}

#' Check if string is a color palette name (not a hex color or named color)
#' @noRd
is_palette_name <- function(x) {
  length(x) == 1 &&
    is.character(x) &&
    !grepl("^#", x) &&
    !x %in% grDevices::colors()
}

#' Check if palette is a valid RColorBrewer palette
#' @noRd
is_brewer_palette <- function(palette_name) {
  requireNamespace("RColorBrewer", quietly = TRUE) &&
    palette_name %in% row.names(RColorBrewer::brewer.pal.info)
}

#' Apply discrete color scale to ggplot
#' @noRd
apply_discrete_scale <- function(p, colors, legend_title, na_color, n_levels) {
  if (is.null(colors)) {
    return(p + ggplot2::scale_fill_viridis_d(
      name = legend_title, na.value = na_color
    ))
  }

  if (!is_palette_name(colors)) {
    return(p + ggplot2::scale_fill_manual(
      values = colors, name = legend_title, na.value = na_color
    ))
  }

  if (is_brewer_palette(colors)) {
    max_colors <- RColorBrewer::brewer.pal.info[colors, "maxcolors"]
    pal_colors <- RColorBrewer::brewer.pal(min(n_levels, max_colors), colors)
    return(p + ggplot2::scale_fill_manual(
      values = pal_colors, name = legend_title, na.value = na_color
    ))
  }

  # Fallback: treat as viridis option name
  p + ggplot2::scale_fill_viridis_d(
    option = tolower(colors), name = legend_title, na.value = na_color
  )
}

#' Apply continuous color scale to ggplot
#' @noRd
apply_continuous_scale <- function(p, colors, legend_title, na_color) {
  if (is.null(colors)) {
    return(p + ggplot2::scale_fill_viridis_c(
      name = legend_title, na.value = na_color
    ))
  }

  if (!is_palette_name(colors)) {
    return(p + ggplot2::scale_fill_gradientn(
      colors = colors, name = legend_title, na.value = na_color
    ))
  }

  if (is_brewer_palette(colors)) {
    return(p + ggplot2::scale_fill_distiller(
      palette = colors, direction = 1,
      name = legend_title, na.value = na_color
    ))
  }

  # Fallback: treat as viridis option name
  p + ggplot2::scale_fill_viridis_c(
    option = tolower(colors), name = legend_title, na.value = na_color
  )
}

#' Convert hexify data to sf polygons
#' @noRd
prepare_hex_sf <- function(data, aperture) {
  # Handle HexData objects
  if (is_hex_data(data)) {
    g <- data@grid
    underlying_data <- data@data
    unique_cells <- unique(data@cell_id)
    hex_sf <- cell_to_sf(unique_cells, g)

    # Join extra columns from original data
    extra_cols <- setdiff(names(underlying_data), c("cell_id", "geometry"))
    if (length(extra_cols) > 0) {
      # Build data frame with cell_id for merge
      data_with_id <- cbind(underlying_data, cell_id = data@cell_id)
      cols <- c("cell_id", extra_cols)
      data_unique <- data_with_id[!duplicated(data_with_id$cell_id), cols, drop = FALSE]
      hex_sf <- merge(hex_sf, data_unique, by = "cell_id", all.x = TRUE)
    }
    return(hex_sf)
  }

  if (inherits(data, "sf")) return(data)

  if (!is.data.frame(data) || !"cell_id" %in% names(data)) {
    stop("data must be a HexData object or an sf object")
  }

  # Handle legacy data frames with cell_area column
  if ("cell_area" %in% names(data) || "cell_area_km2" %in% names(data)) {
    # Get area to determine resolution
    area <- if ("cell_area_km2" %in% names(data)) data$cell_area_km2[1]
            else if ("cell_area" %in% names(data)) data$cell_area[1]

    # Create temporary grid
    grid <- hex_grid(area_km2 = area, aperture = aperture)

    # Generate polygons
    unique_cells <- unique(data$cell_id)
    hex_sf <- cell_to_sf(unique_cells, grid)

    # Join extra columns from original data
    extra_cols <- setdiff(names(data), c("cell_id", "geometry"))
    if (length(extra_cols) > 0) {
      cols <- c("cell_id", extra_cols)
      data_unique <- data[!duplicated(data$cell_id), cols, drop = FALSE]
      hex_sf <- merge(hex_sf, data_unique, by = "cell_id", all.x = TRUE)
    }
    return(hex_sf)
  }

  stop("data must contain 'cell_area' or 'cell_area_km2' column (output from hexify()).")
}

#' Resolve basemap specification to sf object
#' @noRd
resolve_basemap <- function(basemap) {
  if (is.null(basemap)) return(NULL)

  if (is.character(basemap) && basemap == "world") {
    return(hexify_world)
  }

  if (is.character(basemap) && basemap == "world_hires") {
    if (!requireNamespace("rnaturalearth", quietly = TRUE)) {
      stop("Package 'rnaturalearth' is required for basemap = 'world_hires'. ",
           "Install with: install.packages('rnaturalearth')")
    }
    return(rnaturalearth::ne_countries(scale = "medium", returnclass = "sf"))
  }

  if (inherits(basemap, "sf") || inherits(basemap, "sfc")) {
    return(basemap)
  }

  stop("basemap must be 'world', 'world_hires', or an sf object")
}

#' Create mask for areas outside basemap
#' @noRd
create_outside_mask <- function(hex_sf, basemap_sf, xlim, ylim) {
  if (!is.null(xlim) && !is.null(ylim)) {
    bbox_coords <- c(
      xmin = xlim[1], xmax = xlim[2], ymin = ylim[1], ymax = ylim[2]
    )
  } else {
    hex_bbox <- sf::st_bbox(hex_sf)
    buffer <- calculate_view_buffer(hex_bbox)
    bbox_coords <- c(
      xmin = unname(hex_bbox["xmin"] - buffer$x),
      xmax = unname(hex_bbox["xmax"] + buffer$x),
      ymin = unname(hex_bbox["ymin"] - buffer$y),
      ymax = unname(hex_bbox["ymax"] + buffer$y)
    )
  }

  bbox_poly <- sf::st_as_sfc(sf::st_bbox(bbox_coords, crs = sf::st_crs(hex_sf)))
  basemap_union <- sf::st_union(sf::st_make_valid(basemap_sf))
  suppressWarnings(sf::st_difference(bbox_poly, basemap_union))
}

#' Build heatmap layers with masking
#' @noRd
build_masked_layers <- function(p, hex_sf, fill_col, hex_border, hex_lwd,
                                hex_alpha, basemap_sf, basemap_border,
                                basemap_lwd, mask_sf, uniform_fill = "#E69F00") {
 # Hexagons first
  if (is.null(fill_col)) {
    p <- p + ggplot2::geom_sf(
      data = hex_sf,
      fill = uniform_fill,
      color = hex_border,
      linewidth = hex_lwd,
      alpha = hex_alpha
    )
  } else {
    p <- p + ggplot2::geom_sf(
      data = hex_sf,
      ggplot2::aes(fill = .data[[fill_col]]),
      color = hex_border,
      linewidth = hex_lwd,
      alpha = hex_alpha
    )
  }

  # Mask layer to hide hexes outside land
  if (!is.null(mask_sf)) {
    p <- p + ggplot2::geom_sf(data = mask_sf, fill = "white", color = NA)
  }

  # Basemap borders on top (no fill)
  p + ggplot2::geom_sf(
    data = basemap_sf,
    fill = NA,
    color = basemap_border,
    linewidth = basemap_lwd
  )
}

#' Build standard heatmap layers (basemap under hexes)
#' @noRd
build_standard_layers <- function(p, hex_sf, fill_col, hex_border, hex_lwd,
                                  hex_alpha, basemap_sf, basemap_fill,
                                  basemap_border, basemap_lwd,
                                  uniform_fill = "#E69F00") {
  if (!is.null(basemap_sf)) {
    p <- p + ggplot2::geom_sf(
      data = basemap_sf,
      fill = basemap_fill,
      color = basemap_border,
      linewidth = basemap_lwd
    )
  }

  if (is.null(fill_col)) {
    p + ggplot2::geom_sf(
      data = hex_sf,
      fill = uniform_fill,
      color = hex_border,
      linewidth = hex_lwd,
      alpha = hex_alpha
    )
  } else {
    p + ggplot2::geom_sf(
      data = hex_sf,
      ggplot2::aes(fill = .data[[fill_col]]),
      color = hex_border,
      linewidth = hex_lwd,
      alpha = hex_alpha
    )
  }
}

#' Simple sf preparation for hexify_map (no extra column merging)
#' @noRd
prepare_hex_sf_simple <- function(data, aperture) {
  # Handle HexData objects
  if (is_hex_data(data)) {
    g <- data@grid
    unique_cells <- unique(data@cell_id)
    return(cell_to_sf(unique_cells, g))
  }

  if (inherits(data, "sf")) return(data)

  if (!is.data.frame(data) || !"cell_id" %in% names(data)) {
    stop("data must be a HexData object or an sf object")
  }

  # Handle legacy data frames with cell_area column
  if ("cell_area" %in% names(data) || "cell_area_km2" %in% names(data)) {
    # Get area to determine resolution
    area <- if ("cell_area_km2" %in% names(data)) data$cell_area_km2[1]
            else if ("cell_area" %in% names(data)) data$cell_area[1]

    # Create temporary grid
    grid <- hex_grid(area_km2 = area, aperture = aperture)

    # Generate polygons
    unique_cells <- unique(data$cell_id)
    return(cell_to_sf(unique_cells, grid))
  }

  stop("data must contain 'cell_area' or 'cell_area_km2' column (output from hexify()).")
}

#' Resolve value column name (auto-detect if NULL)
#' @noRd
resolve_value_column <- function(hex_sf, value, require = FALSE) {
  if (!is.null(value)) {
    if (!value %in% names(hex_sf)) {
      stop("Column '", value, "' not found in data. ",
           "Available columns: ", paste(names(hex_sf), collapse = ", "))
    }
    return(value)
  }

  # Auto-detect common value columns

  if ("count" %in% names(hex_sf)) return("count")
  if ("n" %in% names(hex_sf)) return("n")


  # No value column found - return NULL for uniform fill
  if (require) {
    stop("No 'value' column specified and no 'count' or 'n' column found in data")
  }
  NULL
}

#' Prepare fill column with optional binning
#' @noRd
prepare_fill_column <- function(hex_sf, value, breaks, labels) {
  value_data <- hex_sf[[value]]
  is_discrete <- is.factor(value_data) || is.character(value_data)

  if (!is.null(breaks) && is_discrete) {
    warning(sprintf(
      "'breaks' is ignored because '%s' is discrete (factor/character)", value
    ))
  }

  if (is.null(breaks) || is_discrete) {
    return(list(data = hex_sf, fill_col = value, is_discrete = is_discrete))
  }

  # Generate labels if not provided
  if (is.null(labels)) {
    labels <- generate_bin_labels(breaks)
  }

  bin_col <- paste0(value, "_bin")
  hex_sf[[bin_col]] <- cut(
    value_data, breaks = breaks, labels = labels, include.lowest = TRUE
  )

  list(data = hex_sf, fill_col = bin_col, is_discrete = TRUE)
}

#' Generate bin labels from break points
#' @noRd
generate_bin_labels <- function(breaks) {
  n_bins <- length(breaks) - 1
  labels <- character(n_bins)

  for (i in seq_len(n_bins)) {
    low <- breaks[i]
    high <- breaks[i + 1]
    labels[i] <- if (is.infinite(low) && low < 0) {
      paste0("<", high)
    } else if (is.infinite(high)) {
      paste0(">", low)
    } else {
      paste0(low, "-", high)
    }
  }

  labels
}

#' Resolve basemap specification including raster support
#' @noRd
resolve_basemap_with_raster <- function(basemap) {
  result <- list(sf = NULL, raster = NULL)
  if (is.null(basemap)) return(result)

  if (is.character(basemap) && basemap == "world") {
    result$sf <- hexify_world
    return(result)
  }

  if (inherits(basemap, "sf") || inherits(basemap, "sfc")) {
    result$sf <- basemap
    return(result)
  }

  if (inherits(basemap, "SpatRaster")) {
    if (!requireNamespace("terra", quietly = TRUE)) {
      stop("Package 'terra' is required for SpatRaster basemaps")
    }
    result$raster <- basemap
    return(result)
  }

  if (inherits(basemap, c("RasterLayer", "RasterBrick", "RasterStack"))) {
    if (!requireNamespace("raster", quietly = TRUE)) {
      stop("Package 'raster' is required for Raster* basemaps")
    }
    result$raster <- basemap
    return(result)
  }

stop("basemap must be 'world', an sf object, or a SpatRaster (terra)")
}

# =============================================================================
# PUBLIC FUNCTIONS
# =============================================================================

# hexify_map() removed - use plot() method on HexData objects instead


#' Quick world map plot
#'
#' Simple wrapper to plot the built-in world map.
#'
#' @param fill Fill color for countries
#' @param border Border color for countries
#' @param ... Additional arguments passed to plot()
#'
#' @return NULL invisibly. Creates a plot as side effect.
#'
#' @family visualization
#' @export
#' @examples
#' # Quick world map
#' plot_world()
#'
#' # Custom colors
#' plot_world(fill = "lightblue", border = "darkblue")
plot_world <- function(fill = "gray90", border = "gray50", ...) {

  plot(sf::st_geometry(hexify_world),
       col = fill, border = border, ...)
  invisible(NULL)
}


#' Create a ggplot2 visualization of hexagonal grid cells
#'
#' Creates a ggplot2-based visualization of hexagonal grid cells, optionally
#' colored by a value column. Supports continuous and discrete color scales,
#' projection transformation, and customizable styling.
#'
#' @param data A HexData object from \code{hexify()}, a data frame with cell_id
#'   and cell_area columns, or an sf object with hexagon polygons.
#' @param value Column name (as string) to use for fill color. If NULL, cells
#'   are drawn with a uniform fill color. If not specified but data has a
#'   'count' or 'n' column, that will be used automatically.
#' @param basemap Optional basemap. Can be:
#'   \itemize{
#'     \item \code{NULL}: No basemap (default)
#'     \item \code{"world"}: Use built-in \code{hexify_world} map (low resolution)
#'     \item \code{"world_hires"}: Use high-resolution map from rnaturalearth (requires package)
#'     \item An sf object: User-supplied vector map, from any format
#'       \code{sf::st_read()} reads, such as a shapefile, 'GeoJSON' or 'GeoPackage'
#'     \item A SpatRaster: User-supplied raster, from any format
#'       \code{terra::rast()} reads, such as a 'GeoTIFF', drawn in grayscale
#'   }
#' @param crs Target CRS for the map projection. Can be:
#'   \itemize{
#'     \item A numeric EPSG code (e.g., 4326 for 'WGS84', 3035 for LAEA Europe)
#'     \item A proj4 string
#'     \item An sf crs object
#'     \item NULL to use 'WGS84' (EPSG:4326)
#'   }
#' @param colors Color palette for the heatmap. Can be:
#'   \itemize{
#'     \item A character vector of colors (for manual scale)
#'     \item A single RColorBrewer palette name (e.g., "YlOrRd", "Greens")
#'     \item NULL to use viridis
#'   }
#' @param breaks Numeric vector of break points for binning continuous values,
#'   or NULL for continuous scale. Use \code{Inf} and \code{-Inf} for open-ended bins.
#' @param labels Labels for the breaks (length should be one less than breaks).
#'   If NULL, labels are auto-generated.
#' @param hex_border Border color for hexagons
#' @param hex_lwd Line width for hexagon borders
#' @param hex_alpha Transparency for hexagon fill (0-1)
#' @param basemap_fill Fill color for basemap polygons
#' @param basemap_border Border color for basemap polygons
#' @param basemap_lwd Line width for basemap borders
#' @param mask_outside Logical. If TRUE and basemap is provided, mask hexagon
#'   portions that fall outside the basemap polygons.
#' @param aperture Grid aperture (default 3), used if data is from hexify()
#' @param xlim Optional x-axis limits (in target CRS units) as c(min, max)
#' @param ylim Optional y-axis limits (in target CRS units) as c(min, max)
#' @param title Plot title
#' @param legend_title Title for the color legend
#' @param na_color Color for NA values
#' @param theme_void Logical. If TRUE (default), use a minimal theme without
#'   axes, gridlines, or background.
#'
#' @return A ggplot2 object that can be further customized or saved.
#'
#' @details
#' This function provides publication-quality heatmap visualizations of
#' hexagonal grids using ggplot2. It returns a ggplot object that can be
#' further customized with standard ggplot2 functions.
#'
#' @section Color Scales:
#' The function supports three types of color scales:
#' \describe{
#'   \item{Continuous}{Set \code{breaks = NULL} for a continuous gradient}
#'   \item{Binned}{Provide \code{breaks} vector to bin values into categories}
#'   \item{Discrete}{If \code{value} column is a factor, discrete colors are used}
#' }
#'
#' @section Projections:
#' Common projections:
#' \describe{
#'   \item{4326}{'WGS84' (unprojected lat/lon)}
#'   \item{3035}{LAEA Europe}
#'   \item{3857}{Web Mercator}
#'   \item{"+proj=robin"}{Robinson (world maps)}
#'   \item{"+proj=moll"}{Mollweide (equal-area world maps)}
#' }
#'
#' @family visualization
#' @seealso \code{\link{plot_grid}} for base R plotting,
#'   \code{\link{cell_to_sf}} to generate polygons manually
#' @export
#' @examples
#' library(hexify)
#'
#' # Sample data with counts
#' cities <- data.frame(
#'   lon = c(16.37, 2.35, -3.70, 12.5, 4.9),
#'   lat = c(48.21, 48.86, 40.42, 41.9, 52.4),
#'   count = c(100, 250, 75, 180, 300)
#' )
#' result <- hexify(cities, lon = "lon", lat = "lat", area_km2 = 5000)
#'
#' # Simple plot (uniform fill, no value mapping)
#' hexify_heatmap(result)
#'
#' \donttest{
#' library(ggplot2)
#'
#' # With world basemap
#' hexify_heatmap(result, basemap = "world")
#'
#' # Heatmap with value mapping
#' hexify_heatmap(result, value = "count")
#'
#' # With world basemap and custom colors
#' hexify_heatmap(result, value = "count",
#'                basemap = "world",
#'                colors = "YlOrRd",
#'                title = "City Density")
#'
#' # Binned values with custom breaks
#' hexify_heatmap(result, value = "count",
#'                basemap = "world",
#'                breaks = c(-Inf, 100, 200, Inf),
#'                labels = c("Low", "Medium", "High"),
#'                colors = c("#fee8c8", "#fdbb84", "#e34a33"))
#'
#' # Different projection (LAEA Europe)
#' hexify_heatmap(result, value = "count",
#'                basemap = "world",
#'                crs = 3035,
#'                xlim = c(2500000, 6500000),
#'                ylim = c(1500000, 5500000))
#'
#' # Customize further with ggplot2
#' hexify_heatmap(result, value = "count", basemap = "world") +
#'   labs(caption = "Data source: Example") +
#'   theme(legend.position = "bottom")
#' }
hexify_heatmap <- function(data,
                        value = NULL,
                        basemap = NULL,
                        crs = NULL,
                        colors = NULL,
                        breaks = NULL,
                        labels = NULL,
                        hex_border = "#5D4E37",
                        hex_lwd = 0.3,
                        hex_alpha = 0.7,
                        basemap_fill = "gray90",
                        basemap_border = "gray50",
                        basemap_lwd = 0.5,
                        mask_outside = FALSE,
                        aperture = 3L,
                        xlim = NULL,
                        ylim = NULL,
                        title = NULL,
                        legend_title = NULL,
                        na_color = "gray90",
                        theme_void = TRUE) {

  # Check ggplot2 availability
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for hexify_heatmap(). ",
         "Install it with: install.packages('ggplot2')")
  }

  # Prepare hex sf with extra columns merged
  hex_sf <- prepare_hex_sf(data, aperture)

  # Resolve value column (NULL means uniform fill)
  value <- resolve_value_column(hex_sf, value, require = FALSE)

  # Setup CRS: a map with no projection asked for stays on the grid's own body
  if (is.na(sf::st_crs(hex_sf))) sf::st_crs(hex_sf) <- 4326
  crs <- if (is.null(crs)) sf::st_crs(hex_sf) else crs
  hex_sf <- sf::st_transform(hex_sf, crs)

  # Resolve and transform basemap (supports sf and SpatRaster)
  basemap_info <- resolve_basemap_with_raster(basemap)
  basemap_sf <- basemap_info$sf
  basemap_raster <- basemap_info$raster
  if (!is.null(basemap_sf)) {
    basemap_sf <- sf::st_transform(basemap_sf, crs)
  }

  # Disable S2 for geometry operations (avoids edge crossing errors)
  s2_was_used <- sf::sf_use_s2()
  sf::sf_use_s2(FALSE)
  on.exit(sf::sf_use_s2(s2_was_used), add = TRUE)

  # Create mask for areas outside basemap (if requested)
  mask_sf <- NULL
  if (mask_outside && !is.null(basemap_sf)) {
    mask_sf <- create_outside_mask(hex_sf, basemap_sf, xlim, ylim)
  }

  # Prepare fill column (apply breaks if needed) - skip if no value
 fill_col <- NULL
  is_discrete <- FALSE
  if (!is.null(value)) {
    fill_info <- prepare_fill_column(hex_sf, value, breaks, labels)
    hex_sf <- fill_info$data
    fill_col <- fill_info$fill_col
    is_discrete <- fill_info$is_discrete
  }

  # Set legend title
  legend_title <- if (is.null(legend_title)) value else legend_title

  # Build ggplot with layers
  p <- ggplot2::ggplot()

  # Add raster basemap if provided (rendered as grayscale annotation)
  if (!is.null(basemap_raster)) {
    if (requireNamespace("terra", quietly = TRUE)) {
      rast_proj <- terra::project(basemap_raster, parse_crs(crs)$wkt)
      ext <- as.vector(terra::ext(rast_proj))
      # Convert to matrix for annotation_raster
      vals <- terra::values(rast_proj[[1]])
      nr <- terra::nrow(rast_proj)
      nc <- terra::ncol(rast_proj)
      # Normalize to grayscale
      vmin <- min(vals, na.rm = TRUE)
      vmax <- max(vals, na.rm = TRUE)
      if (vmax > vmin) {
        gray_vals <- (vals - vmin) / (vmax - vmin)
      } else {
        gray_vals <- rep(0.5, length(vals))
      }
      gray_vals[is.na(gray_vals)] <- 1
      gray_mat <- matrix(grDevices::gray(gray_vals), nrow = nr, ncol = nc,
                          byrow = TRUE)
      p <- p + ggplot2::annotation_raster(
        gray_mat,
        xmin = ext[1], xmax = ext[2],
        ymin = ext[3], ymax = ext[4]
      )
    }
  }

  if (mask_outside && !is.null(basemap_sf)) {
    p <- build_masked_layers(
      p, hex_sf, fill_col, hex_border, hex_lwd, hex_alpha,
      basemap_sf, basemap_border, basemap_lwd, mask_sf
    )
  } else {
    p <- build_standard_layers(
      p, hex_sf, fill_col, hex_border, hex_lwd, hex_alpha,
      basemap_sf, basemap_fill, basemap_border, basemap_lwd
    )
  }

  # Apply color scale (only if we have a value column)
  if (!is.null(fill_col)) {
    n_levels <- length(unique(hex_sf[[fill_col]]))
    if (is_discrete) {
      p <- apply_discrete_scale(p, colors, legend_title, na_color, n_levels)
    } else {
      p <- apply_continuous_scale(p, colors, legend_title, na_color)
    }
  }

  # Set coordinate system with limits
  p <- if (!is.null(xlim) || !is.null(ylim)) {
    p + ggplot2::coord_sf(xlim = xlim, ylim = ylim, expand = FALSE)
  } else {
    p + ggplot2::coord_sf(expand = FALSE)
  }

  # Add title
  if (!is.null(title)) {
    p <- p + ggplot2::labs(title = title)
  }

  # Apply theme
  if (theme_void) {
    p <- p + .theme_clean() +
      ggplot2::theme(
        axis.text = ggplot2::element_blank(),
        axis.title = ggplot2::element_blank(),
        axis.ticks = ggplot2::element_blank()
      )
  }

  p
}

# tests/testthat/test-hexify-heatmap.R
# Tests for hexify_heatmap function

test_that("hexify_heatmap creates ggplot object", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    value = c(10, 20, 30)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result)

  expect_s3_class(p, "ggplot")
  expect_s3_class(p$layers[[1]]$geom, "GeomSf")
  expect_true(nrow(p$layers[[1]]$data) > 0)
})

test_that("hexify_heatmap works with value column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    count = c(10, 20, 30)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result, value = "count")

  expect_s3_class(p, "ggplot")
  # The fill aesthetic must actually reference the requested column, not just
  # produce *a* plot.
  expect_identical(rlang::as_label(p$layers[[1]]$mapping$fill), "count")
})

test_that("hexify_heatmap works with basemap = 'world'", {
  skip_on_cran()  # Slow sf intersection operations
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- suppressMessages(hexify_heatmap(result, basemap = "world"))

  expect_s3_class(p, "ggplot")
})

test_that("hexify_heatmap works with custom sf basemap", {
  skip_on_cran()  # Slow sf operations
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  custom_basemap <- hexify_world[hexify_world$continent == "Europe", ]
  p <- suppressMessages(hexify_heatmap(result, basemap = custom_basemap))

  expect_s3_class(p, "ggplot")
})

test_that("hexify_heatmap works with discrete values", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    category = c("A", "B", "A")
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result, value = "category")

  expect_s3_class(p, "ggplot")
  expect_identical(rlang::as_label(p$layers[[1]]$mapping$fill), "category")
  # A discrete value column must produce a discrete (not continuous) scale.
  fill_scale <- Filter(function(s) identical(s$aesthetics, "fill"), p$scales$scales)[[1]]
  expect_s3_class(fill_scale, "ScaleDiscrete")
})

test_that("hexify_heatmap works with custom colors vector", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    value = c(10, 20, 30)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result, value = "value",
                      colors = c("blue", "white", "red"))

  expect_s3_class(p, "ggplot")
  fill_scale <- Filter(function(s) identical(s$aesthetics, "fill"), p$scales$scales)[[1]]
  # The gradient's endpoints must actually be the requested colors, not the
  # default viridis palette.
  ends <- fill_scale$palette(c(0, 1))
  expect_equal(toupper(ends), c("#0000FF", "#FF0000"))
})

test_that("hexify_heatmap works with breaks", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    value = c(10, 20, 30)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result, value = "value",
                      breaks = c(-Inf, 15, 25, Inf),
                      labels = c("Low", "Medium", "High"))

  expect_s3_class(p, "ggplot")
  expect_identical(rlang::as_label(p$layers[[1]]$mapping$fill), "value_bin")
  expect_identical(levels(p$layers[[1]]$data[["value_bin"]]), c("Low", "Medium", "High"))
  # Bin membership must match the supplied breaks, not just have 3 levels.
  expect_identical(
    as.character(p$layers[[1]]$data[["value_bin"]][order(p$layers[[1]]$data$value)]),
    c("Low", "Medium", "High")
  )
})

test_that("hexify_heatmap works with breaks (auto labels)", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    value = c(10, 20, 30)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  # Auto-generated labels
  p <- hexify_heatmap(result, value = "value",
                      breaks = c(-Inf, 15, 25, Inf))

  expect_s3_class(p, "ggplot")
  bin_levels <- levels(p$layers[[1]]$data[["value_bin"]])
  expect_length(bin_levels, 3)
  # Auto-labels must be distinct and non-empty, not e.g. all "NA" or blank.
  expect_false(anyNA(bin_levels))
  expect_length(unique(bin_levels), 3)
})

test_that("hexify_heatmap works with xlim and ylim", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result,
                      xlim = c(-10, 20),
                      ylim = c(40, 60))

  expect_s3_class(p, "ggplot")
  expect_equal(p$coordinates$limits$x, c(-10, 20))
  expect_equal(p$coordinates$limits$y, c(40, 60))
})

test_that("hexify_heatmap works with title", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(lon = c(2.35), lat = c(48.86))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result, title = "Test Title")

  expect_s3_class(p, "ggplot")
  expect_identical(p$labels$title, "Test Title")
})

test_that("hexify_heatmap works with legend_title", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37),
    value = c(10, 20)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result, value = "value", legend_title = "Custom Legend")

  expect_s3_class(p, "ggplot")
  fill_scale <- Filter(function(s) identical(s$aesthetics, "fill"), p$scales$scales)[[1]]
  expect_identical(fill_scale$name, "Custom Legend")

  # Default (no legend_title) falls back to the value column name.
  p_default <- hexify_heatmap(result, value = "value")
  fill_scale_default <- Filter(function(s) identical(s$aesthetics, "fill"),
                                p_default$scales$scales)[[1]]
  expect_identical(fill_scale_default$name, "value")
})

test_that("hexify_heatmap works with mask_outside", {
  skip_on_cran()  # Slow sf intersection operations
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- suppressMessages(hexify_heatmap(result,
                                        basemap = "world",
                                        mask_outside = TRUE))
  p_unmasked <- suppressMessages(hexify_heatmap(result, basemap = "world"))

  expect_s3_class(p, "ggplot")
  # Masking to land must actually change the hex layer's geometry, not just
  # produce a plot with the same unmasked data.
  expect_false(isTRUE(all.equal(
    sf::st_area(p$layers[[1]]$data),
    sf::st_area(p_unmasked$layers[[1]]$data)
  )))
})

test_that("hexify_heatmap works with theme_void = FALSE", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(lon = c(2.35), lat = c(48.86))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result, theme_void = FALSE)
  p_void <- hexify_heatmap(result, theme_void = TRUE)

  expect_s3_class(p, "ggplot")
  # The two themes must actually differ, not both fall back to one default.
  expect_false(identical(p$theme$axis.text.x, p_void$theme$axis.text.x))
})

test_that("hexify_heatmap errors on invalid value column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(lon = c(2.35), lat = c(48.86))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_error(hexify_heatmap(result, value = "nonexistent"), "not found")
})

test_that("hexify_heatmap errors on invalid basemap", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(lon = c(2.35), lat = c(48.86))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_error(hexify_heatmap(result, basemap = "invalid"), "must be")
})

test_that("hexify_heatmap works with custom styling", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(lon = c(2.35, 4.90), lat = c(48.86, 52.37))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p <- hexify_heatmap(result,
                      hex_border = "blue",
                      hex_lwd = 1.5,
                      hex_alpha = 0.5)

  expect_s3_class(p, "ggplot")
  params <- p$layers[[1]]$aes_params
  expect_identical(params$colour, "blue")
  expect_equal(params$linewidth, 1.5)
  expect_equal(params$alpha, 0.5)
})

test_that("hexify_heatmap works with CRS transformation", {
  skip_on_cran()  # Slow CRS transformation
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  # LAEA Europe projection
  p <- suppressWarnings(hexify_heatmap(result, crs = 3035))

  expect_s3_class(p, "ggplot")
  expect_equal(sf::st_crs(p$layers[[1]]$data)$epsg, 3035)
})

test_that("hexify_heatmap auto-detects count column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37),
    count = c(10, 20)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  # Should auto-detect 'count' column
  p <- hexify_heatmap(result)

  expect_s3_class(p, "ggplot")
  expect_identical(rlang::as_label(p$layers[[1]]$mapping$fill), "count")
})

test_that("hexify_heatmap auto-detects n column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37),
    n = c(10, 20)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  # Should auto-detect 'n' column
  p <- hexify_heatmap(result)

  expect_s3_class(p, "ggplot")
  expect_identical(rlang::as_label(p$layers[[1]]$mapping$fill), "n")
})

# =============================================================================
# plot_world Tests
# =============================================================================

test_that("plot_world creates plot", {
  skip_if_not_installed("sf")

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    result <- plot_world()
  })
  expect_null(result)
})

test_that("plot_world accepts custom colors", {
  skip_if_not_installed("sf")

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot_world(fill = "lightblue", border = "darkblue")
  })
})

# tests/testthat/test-visualization.R
# Tests for visualization functions
#
# Functions tested:
# - plot() method for HexData
# - hexify_heatmap()
# - plot_world()
# - Internal helper functions

# =============================================================================
# INTERNAL HELPER FUNCTIONS
# =============================================================================

test_that("calculate_view_buffer returns valid buffer", {
  bbox <- c(xmin = 0, xmax = 10, ymin = 40, ymax = 50)

  buffer <- hexify:::calculate_view_buffer(bbox)

  expect_type(buffer, "list")
  expect_true("x" %in% names(buffer))
  expect_true("y" %in% names(buffer))
  expect_true(buffer$x > 0)
  expect_true(buffer$y > 0)
})

test_that("calculate_view_buffer respects factor parameter", {
  bbox <- c(xmin = 0, xmax = 100, ymin = 0, ymax = 100)

  buffer_10 <- hexify:::calculate_view_buffer(bbox, factor = 0.1)
  buffer_20 <- hexify:::calculate_view_buffer(bbox, factor = 0.2)

  expect_true(buffer_20$x > buffer_10$x)
  expect_true(buffer_20$y > buffer_10$y)
})

test_that("calculate_view_buffer respects min_buffer", {
  bbox <- c(xmin = 0, xmax = 1, ymin = 0, ymax = 1)

  buffer <- hexify:::calculate_view_buffer(bbox, factor = 0.1, min_buffer = 5)

  expect_gte(buffer$x, 5)
  expect_gte(buffer$y, 5)
})

test_that("is_palette_name identifies palette names correctly", {
  expect_true(hexify:::is_palette_name("YlOrRd"))
  expect_true(hexify:::is_palette_name("viridis"))

  expect_false(hexify:::is_palette_name("#FF0000"))
  expect_false(hexify:::is_palette_name("red"))
  expect_false(hexify:::is_palette_name(c("red", "blue")))
})

test_that("is_brewer_palette identifies valid palettes", {
  skip_if_not_installed("RColorBrewer")

  expect_true(hexify:::is_brewer_palette("YlOrRd"))
  expect_true(hexify:::is_brewer_palette("Blues"))
  expect_true(hexify:::is_brewer_palette("Set1"))

  expect_false(hexify:::is_brewer_palette("NotAPalette"))
  expect_false(hexify:::is_brewer_palette("viridis"))
})

test_that("generate_bin_labels creates correct labels", {
  breaks <- c(0, 10, 20, 30)
  labels <- hexify:::generate_bin_labels(breaks)

  expect_length(labels, 3)
  expect_equal(labels, c("0-10", "10-20", "20-30"))
})

test_that("generate_bin_labels handles infinite bounds", {
  breaks <- c(-Inf, 10, 100, Inf)
  labels <- hexify:::generate_bin_labels(breaks)

  expect_length(labels, 3)
  expect_equal(labels[1], "<10")
  expect_equal(labels[3], ">100")
})

test_that("resolve_value_column auto-detects count column", {
  skip_if_not_installed("sf")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$count <- c(10, 20)

  col <- hexify:::resolve_value_column(hex_sf, NULL)
  expect_equal(col, "count")
})

test_that("resolve_value_column auto-detects n column", {
  skip_if_not_installed("sf")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$n <- c(10, 20)

  col <- hexify:::resolve_value_column(hex_sf, NULL)
  expect_equal(col, "n")
})

test_that("resolve_value_column returns specified column", {
  skip_if_not_installed("sf")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$value <- c(10, 20)

  col <- hexify:::resolve_value_column(hex_sf, "value")
  expect_equal(col, "value")
})

test_that("resolve_value_column errors on missing column", {
  skip_if_not_installed("sf")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)

  expect_error(
    hexify:::resolve_value_column(hex_sf, "nonexistent"),
    "not found in data"
  )
})

test_that("resolve_value_column returns NULL when no suitable column found", {
  skip_if_not_installed("sf")


  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)

  # With require = FALSE (default), returns NULL for uniform fill
  result <- hexify:::resolve_value_column(hex_sf, NULL)
  expect_null(result)

  # With require = TRUE, throws error
  expect_error(
    hexify:::resolve_value_column(hex_sf, NULL, require = TRUE),
    "No 'value' column specified"
  )
})

test_that("prepare_fill_column handles discrete data", {
  skip_if_not_installed("sf")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$category <- factor(c("A", "B"))

  result <- hexify:::prepare_fill_column(hex_sf, "category", NULL, NULL)

  expect_equal(result$fill_col, "category")
  expect_true(result$is_discrete)
})

test_that("prepare_fill_column handles continuous data without breaks", {
  skip_if_not_installed("sf")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$value <- c(10, 20)

  result <- hexify:::prepare_fill_column(hex_sf, "value", NULL, NULL)

  expect_equal(result$fill_col, "value")
  expect_false(result$is_discrete)
})

test_that("prepare_fill_column applies breaks to continuous data", {
  skip_if_not_installed("sf")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$value <- c(5, 15)

  result <- hexify:::prepare_fill_column(
    hex_sf, "value", c(0, 10, 20), c("Low", "High")
  )

  expect_true("value_bin" %in% names(result$data))
  expect_equal(result$fill_col, "value_bin")
  expect_true(result$is_discrete)
})

test_that("resolve_basemap returns NULL for NULL input", {
  result <- hexify:::resolve_basemap(NULL)
  expect_null(result)
})

test_that("resolve_basemap returns hexify_world for 'world'", {
  skip_if_not_installed("sf")

  result <- hexify:::resolve_basemap("world")
  expect_s3_class(result, "sf")
})

test_that("resolve_basemap returns sf for sf input", {
  skip_if_not_installed("sf")

  result <- hexify:::resolve_basemap(hexify_world)
  expect_s3_class(result, "sf")
})

test_that("resolve_basemap errors on invalid input", {
  expect_error(hexify:::resolve_basemap(123), "basemap must be")
  expect_error(hexify:::resolve_basemap("invalid"), "basemap must be")
})

test_that("resolve_basemap_with_raster returns correct structure", {
  result <- hexify:::resolve_basemap_with_raster(NULL)

  expect_type(result, "list")
  expect_true("sf" %in% names(result))
  expect_true("raster" %in% names(result))
  expect_null(result$sf)
  expect_null(result$raster)
})

test_that("resolve_basemap_with_raster handles world basemap", {
  result <- hexify:::resolve_basemap_with_raster("world")

  expect_s3_class(result$sf, "sf")
  expect_null(result$raster)
})

test_that("resolve_basemap_with_raster handles sf input", {
  skip_if_not_installed("sf")

  result <- hexify:::resolve_basemap_with_raster(hexify_world)

  expect_s3_class(result$sf, "sf")
  expect_null(result$raster)
})

test_that("resolve_basemap_with_raster errors on invalid input", {
  expect_error(hexify:::resolve_basemap_with_raster(123), "basemap must be")
})

test_that("resolve_basemap_with_raster handles SpatRaster input", {
  skip_if_not_installed("terra")

  r <- terra::rast(nrows = 5, ncols = 5, xmin = -10, xmax = 10, ymin = 40, ymax = 55)
  terra::values(r) <- seq_len(25)

  result <- hexify:::resolve_basemap_with_raster(r)

  expect_null(result$sf)
  expect_s4_class(result$raster, "SpatRaster")
  expect_identical(result$raster, r)
})

test_that("resolve_basemap_with_raster handles RasterLayer input", {
  skip_if_not_installed("raster")
  skip_if_not_installed("terra")

  r <- terra::rast(nrows = 5, ncols = 5, xmin = -10, xmax = 10, ymin = 40, ymax = 55)
  terra::values(r) <- seq_len(25)
  r_legacy <- raster::raster(r)

  result <- hexify:::resolve_basemap_with_raster(r_legacy)

  expect_null(result$sf)
  expect_s4_class(result$raster, "RasterLayer")
})

test_that("hexify_heatmap renders a SpatRaster basemap as an annotation layer", {
  skip_if_not_installed("terra")
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  r <- terra::rast(nrows = 10, ncols = 10, xmin = -10, xmax = 10, ymin = 40, ymax = 55,
                    crs = "EPSG:4326")
  terra::values(r) <- runif(100)

  df <- data.frame(lon = c(2.35, 4.90), lat = c(48.86, 52.37))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  p_with_raster <- hexify_heatmap(result, basemap = r)
  p_without <- hexify_heatmap(result)

  expect_s3_class(p_with_raster, "ggplot")
  # The raster basemap must add a real layer, not just be silently accepted
  # and ignored.
  expect_gt(length(p_with_raster$layers), length(p_without$layers))
  raster_layer_geoms <- vapply(p_with_raster$layers, function(l) class(l$geom)[1], character(1))
  expect_true(any(grepl("Custom|Annotation|Raster", raster_layer_geoms)))
})

test_that("prepare_hex_sf_simple validates input", {
  expect_error(
    hexify:::prepare_hex_sf_simple(data.frame(x = 1), aperture = 3),
    "HexData object or an sf object"
  )

  expect_error(
    hexify:::prepare_hex_sf_simple(data.frame(cell_id = 1), aperture = 3),
    "cell_area"
  )
})

test_that("prepare_hex_sf_simple passes through sf objects", {
  skip_if_not_installed("sf")

  hex_sf <- hexify_cell_to_sf(c(12847), resolution = 10, aperture = 3)
  result <- hexify:::prepare_hex_sf_simple(hex_sf, aperture = 3)

  expect_s3_class(result, "sf")
})

test_that("prepare_hex_sf validates input", {
  expect_error(
    hexify:::prepare_hex_sf(data.frame(x = 1), aperture = 3),
    "HexData object or an sf object"
  )

  expect_error(
    hexify:::prepare_hex_sf(data.frame(cell_id = 1), aperture = 3),
    "cell_area"
  )
})

test_that("prepare_hex_sf passes through sf objects", {
  skip_if_not_installed("sf")

  hex_sf <- hexify_cell_to_sf(c(12847), resolution = 10, aperture = 3)
  result <- hexify:::prepare_hex_sf(hex_sf, aperture = 3)

  expect_s3_class(result, "sf")
})

test_that("prepare_hex_sf merges extra columns", {
  skip_if_not_installed("sf")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    extra_col = c("A", "B")
  )

  result <- hexify:::prepare_hex_sf(df, aperture = 3)

  expect_s3_class(result, "sf")
  expect_true("extra_col" %in% names(result))
})

# =============================================================================
# HEXIFY_WORLD DATA
# =============================================================================

test_that("hexify_world data is available and valid", {
  expect_true(exists("hexify_world"))
  expect_s3_class(hexify_world, "sf")
  expect_true(nrow(hexify_world) > 100)
  expect_true("name" %in% names(hexify_world))
  expect_true("continent" %in% names(hexify_world))
  expect_equal(sf::st_crs(hexify_world)$epsg, 4326)
})

# =============================================================================
# PLOT_WORLD
# =============================================================================

test_that("plot_world works", {
  skip_if_not_installed("sf")

  expect_silent(plot_world())
  expect_silent(plot_world(fill = "lightblue", border = "navy"))
})

# =============================================================================
# HEXIFY_HEATMAP
# =============================================================================

test_that("hexify_heatmap requires ggplot2", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap auto-detects count column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df)
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap auto-detects n column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    n = c(10, 20)
  )

  result <- hexify_heatmap(df)
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works without value column (uniform fill)", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94)
  )

  # Should work with uniform fill when no value column specified
  result <- hexify_heatmap(df)
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with basemap", {
  skip_on_cran()  # Slow sf operations
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count", basemap = "world")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with breaks", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 200)
  )

  result <- hexify_heatmap(df, value = "count",
                            breaks = c(-Inf, 50, 100, Inf),
                            labels = c("Low", "Medium", "High"))
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with custom colors", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count", colors = c("white", "red"))
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap respects styling parameters", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count",
                            hex_border = "darkblue",
                            hex_lwd = 0.5,
                            hex_alpha = 0.8,
                            title = "Test Plot",
                            legend_title = "Count")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with projection", {
  skip_on_cran()  # Slow CRS transformation
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count", crs = 3035)
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap validates input", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  expect_error(hexify_heatmap(list()), "data must be")
  expect_error(hexify_heatmap(data.frame(x = 1)), "data must be")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  expect_error(hexify_heatmap(df, value = "nonexistent"), "not found in data")
  expect_error(hexify_heatmap(df, basemap = 123), "basemap must be")
})

test_that("hexify_heatmap works with mask_outside", {
  skip_on_cran()  # Slow sf intersection operations
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count", basemap = "world",
                           mask_outside = TRUE)
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with discrete factor column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    category = factor(c("A", "B"))
  )

  result <- hexify_heatmap(df, value = "category")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with character column", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    type = c("type_A", "type_B"),
    stringsAsFactors = FALSE
  )

  result <- hexify_heatmap(df, value = "type")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with RColorBrewer palette", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("RColorBrewer")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count", colors = "YlOrRd")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with RColorBrewer discrete scale", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("RColorBrewer")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    category = factor(c("A", "B"))
  )

  result <- hexify_heatmap(df, value = "category", colors = "Set1")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap works with viridis option", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count", colors = "magma")
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap auto-generates break labels", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 200)
  )

  result <- hexify_heatmap(df, value = "count",
                           breaks = c(0, 50, 100, 200))
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap respects theme_void=FALSE", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count", theme_void = FALSE)
  expect_s3_class(result, "ggplot")
})

test_that("hexify_heatmap respects xlim and ylim", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  df <- data.frame(
    cell_id = c(12847, 12532),
    cell_area = c(863.94, 863.94),
    count = c(10, 20)
  )

  result <- hexify_heatmap(df, value = "count",
                           xlim = c(-10, 20), ylim = c(40, 60))
  expect_s3_class(result, "ggplot")
})

# =============================================================================
# COLOR SCALE HELPER FUNCTIONS
# =============================================================================

test_that("apply_discrete_scale works with NULL colors", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$category <- factor(c("A", "B"))

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(data = hex_sf, ggplot2::aes(fill = category))

  result <- hexify:::apply_discrete_scale(p, NULL, "Category", "gray90", 2)
  expect_s3_class(result, "ggplot")
})

test_that("apply_discrete_scale works with manual colors", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$category <- factor(c("A", "B"))

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(data = hex_sf, ggplot2::aes(fill = category))

  result <- hexify:::apply_discrete_scale(
    p, c("red", "blue"), "Category", "gray90", 2
  )
  expect_s3_class(result, "ggplot")
})

test_that("apply_continuous_scale works with NULL colors", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$value <- c(10, 20)

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(data = hex_sf, ggplot2::aes(fill = value))

  result <- hexify:::apply_continuous_scale(p, NULL, "Value", "gray90")
  expect_s3_class(result, "ggplot")
})

test_that("apply_continuous_scale works with manual colors", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$value <- c(10, 20)

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(data = hex_sf, ggplot2::aes(fill = value))

  result <- hexify:::apply_continuous_scale(
    p, c("white", "red"), "Value", "gray90"
  )
  expect_s3_class(result, "ggplot")
})

test_that("apply_continuous_scale works with Brewer palette", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("RColorBrewer")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$value <- c(10, 20)

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(data = hex_sf, ggplot2::aes(fill = value))

  result <- hexify:::apply_continuous_scale(p, "YlOrRd", "Value", "gray90")
  expect_s3_class(result, "ggplot")
})

test_that("apply_discrete_scale falls back to viridis for unknown palette", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$category <- factor(c("A", "B"))

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(data = hex_sf, ggplot2::aes(fill = category))

  # Use an unknown palette name that's not a Brewer palette
  result <- hexify:::apply_discrete_scale(
    p, "inferno", "Category", "gray90", 2
  )
  expect_s3_class(result, "ggplot")
})

test_that("apply_continuous_scale falls back to viridis for unknown palette", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  cell_ids <- c(12847, 12532)
  hex_sf <- hexify_cell_to_sf(cell_ids, resolution = 10, aperture = 3)
  hex_sf$value <- c(10, 20)

  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(data = hex_sf, ggplot2::aes(fill = value))

  # Use an unknown palette name
  result <- hexify:::apply_continuous_scale(p, "cividis", "Value", "gray90")
  expect_s3_class(result, "ggplot")
})

test_that("resolve_basemap handles world_hires without package", {
  # Skip if rnaturalearth is installed - we want to test the error path
  skip_if(requireNamespace("rnaturalearth", quietly = TRUE),
          "rnaturalearth is installed, skipping error test")

  expect_error(
    hexify:::resolve_basemap("world_hires"),
    "rnaturalearth"
  )
})

test_that("resolve_basemap_with_raster handles sfc input", {
  skip_if_not_installed("sf")

  sfc <- sf::st_sfc(sf::st_point(c(0, 0)), crs = 4326)
  result <- hexify:::resolve_basemap_with_raster(sfc)

  expect_s3_class(result$sf, "sfc")
  expect_null(result$raster)
})

test_that("hexify_heatmap handles data with NA CRS", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  # Create sf object and remove CRS
  hex_sf <- hexify_cell_to_sf(c(12847, 12532), resolution = 10, aperture = 3)
  hex_sf$count <- c(10, 20)
  sf::st_crs(hex_sf) <- NA

  result <- hexify_heatmap(hex_sf, value = "count")
  expect_s3_class(result, "ggplot")
})

# crosswalk.R
# Crosswalk between ISEA and H3 grid systems
#
# Maps cell IDs between hexify's native ISEA grids and the H3 hierarchical
# hexagonal system via cell-center coordinate lookup.

#' Crosswalk Between ISEA and H3 Cell IDs
#'
#' Maps cell IDs between ISEA (equal-area) and H3 grid systems by looking up
#' each cell's center coordinate in the target grid. This enables workflows
#' where analysis is done in ISEA (exact equal-area) and reporting in H3
#' (industry-standard).
#'
#' @param cell_id Cell IDs to translate. Numeric for ISEA, character for H3.
#'   When \code{grid} is a HexData object and \code{cell_id} is \code{NULL},
#'   all cell IDs from the data are used.
#' @param grid A HexGridInfo or HexData object. For \code{direction =
#'   "isea_to_h3"}, this must be an ISEA grid. For \code{direction =
#'   "h3_to_isea"}, this must be an H3 grid.
#' @param h3_resolution Target H3 resolution for \code{"isea_to_h3"}. When
#'   \code{NULL} (default), the closest H3 resolution matching the ISEA cell
#'   area is selected automatically. For \code{"h3_to_isea"}, the source H3
#'   resolution is always inferred from \code{grid}; if \code{h3_resolution}
#'   is supplied for that direction it is validated against \code{grid}'s
#'   resolution rather than used to select one.
#' @param isea_grid A HexGridInfo for the target ISEA grid. Required when
#'   \code{direction = "h3_to_isea"}.
#' @param direction One of \code{"isea_to_h3"} (default) or \code{"h3_to_isea"}.
#'
#' @return A data frame with columns:
#' \describe{
#'   \item{isea_cell_id}{ISEA cell ID (numeric)}
#'   \item{h3_cell_id}{H3 cell ID (character)}
#'   \item{isea_area_km2}{Area of the ISEA cell in km2}
#'   \item{h3_area_km2}{Geodesic area of the H3 cell in km2}
#'   \item{area_ratio}{Ratio of ISEA area to H3 area}
#' }
#'
#' @details
#' The crosswalk works by computing the center coordinate of each source cell,
#' then finding which cell in the target grid contains that center. This is a
#' many-to-one mapping: multiple ISEA cells may map to the same H3 cell (or
#' vice versa) depending on the relative resolutions.
#'
#' When \code{h3_resolution} is \code{NULL} and \code{direction = "isea_to_h3"},
#' the H3 resolution whose average cell area is closest to the ISEA cell area
#' is chosen automatically. This gives the best 1:1 correspondence.
#'
#' @seealso \code{\link{cell_area}} for per-cell area computation,
#'   \code{\link{hex_grid}} for creating grids
#'
#' @export
#' @examples
#' \donttest{
#' # ISEA -> H3
#' grid <- hex_grid(area_km2 = 1000)
#' cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
#' xwalk <- h3_crosswalk(cells, grid)
#' head(xwalk)
#'
#' # H3 -> ISEA
#' h3 <- hex_grid(resolution = 5, type = "h3")
#' h3_cells <- lonlat_to_cell(c(0, 10), c(45, 50), h3)
#' xwalk2 <- h3_crosswalk(h3_cells, h3, isea_grid = grid, direction = "h3_to_isea")
#' }
h3_crosswalk <- function(cell_id = NULL,
                         grid,
                         h3_resolution = NULL,
                         isea_grid = NULL,
                         direction = c("isea_to_h3", "h3_to_isea")) {

  direction <- match.arg(direction)

  # -------------------------------------------------------------------------
  # Extract grid and cell IDs
  # -------------------------------------------------------------------------
  if (is_hex_data(grid)) {
    if (is.null(cell_id)) {
      cell_id <- grid@cell_id
    }
    g <- grid@grid
  } else {
    g <- extract_grid(grid)
    if (is.null(cell_id)) {
      stop("cell_id required when grid is not HexData")
    }
  }

  # -------------------------------------------------------------------------
  # Validate direction vs grid type
  # -------------------------------------------------------------------------
  if (direction == "isea_to_h3") {
    if (is_h3_grid(g)) {
      stop("direction = 'isea_to_h3' requires an ISEA grid, got H3")
    }
  } else {
    if (!is_h3_grid(g)) {
      stop("direction = 'h3_to_isea' requires an H3 grid, got ISEA")
    }
    if (is.null(isea_grid)) {
      stop("isea_grid is required when direction = 'h3_to_isea'")
    }
    isea_g <- extract_grid(isea_grid)
    if (is_h3_grid(isea_g)) {
      stop("isea_grid must be an ISEA grid, not H3")
    }
    if (grid_radius_km(isea_g) != grid_radius_km(g)) {
      stop(sprintf(
        "Both grids must cover the same body: 'grid' has radius %s km, 'isea_grid' %s km",
        format(grid_radius_km(g)), format(grid_radius_km(isea_g))
      ))
    }
  }

  # -------------------------------------------------------------------------
  # Deduplicate cell IDs
  # -------------------------------------------------------------------------
  unique_ids <- unique(cell_id)

  # -------------------------------------------------------------------------
  # ISEA -> H3
  # -------------------------------------------------------------------------
  if (direction == "isea_to_h3") {
    # Auto-select H3 resolution if not provided
    if (is.null(h3_resolution)) {
      h3_resolution <- closest_h3_resolution(g@area_km2, grid_radius_km(g))
    } else {
      h3_resolution <- as.integer(h3_resolution)
      if (h3_resolution < H3_MIN_RESOLUTION || h3_resolution > H3_MAX_RESOLUTION) {
        stop(sprintf("h3_resolution must be between %d and %d",
                     H3_MIN_RESOLUTION, H3_MAX_RESOLUTION))
      }
    }

    # Get ISEA cell centers
    coords <- cell_to_lonlat(unique_ids, g)

    # Convert centers to H3 cell IDs
    h3_ids <- cpp_h3_latLngToCell(coords$lon_deg, coords$lat_deg, h3_resolution)

    # Compute areas
    isea_areas <- rep(g@area_km2, length(unique_ids))
    h3_areas <- scale_area_to_body(cpp_h3_cellAreaKm2(h3_ids), grid_radius_km(g))

    data.frame(
      isea_cell_id = unique_ids,
      h3_cell_id = h3_ids,
      isea_area_km2 = isea_areas,
      h3_area_km2 = h3_areas,
      area_ratio = isea_areas / h3_areas,
      stringsAsFactors = FALSE
    )

  } else {
    # -----------------------------------------------------------------------
    # H3 -> ISEA
    # -----------------------------------------------------------------------

    # h3_resolution isn't used to select anything here (the source resolution
    # is always g@resolution) but if supplied, it must agree with the grid.
    if (!is.null(h3_resolution)) {
      h3_resolution <- as.integer(h3_resolution)
      if (h3_resolution != g@resolution) {
        stop(sprintf(
          "h3_resolution (%d) does not match the resolution of 'grid' (%d); h3_to_isea always infers the source resolution from 'grid'",
          h3_resolution, g@resolution
        ))
      }
    }

    # Get H3 cell centers
    center_df <- cpp_h3_cellToLatLng(as.character(unique_ids))

    # Convert centers to ISEA cell IDs
    isea_ids <- lonlat_to_cell(center_df$lon, center_df$lat, isea_g)

    # Compute areas
    h3_areas <- scale_area_to_body(cpp_h3_cellAreaKm2(as.character(unique_ids)),
                                   grid_radius_km(g))
    isea_areas <- rep(isea_g@area_km2, length(unique_ids))

    data.frame(
      isea_cell_id = isea_ids,
      h3_cell_id = as.character(unique_ids),
      isea_area_km2 = isea_areas,
      h3_area_km2 = h3_areas,
      area_ratio = isea_areas / h3_areas,
      stringsAsFactors = FALSE
    )
  }
}

# test-crosswalk.R
# Tests for h3_crosswalk()

skip_if_not_installed("sf")

# =============================================================================
# ISEA -> H3
# =============================================================================

test_that("h3_crosswalk() maps ISEA to H3", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
  xwalk <- h3_crosswalk(cells, grid)

  expect_true(is.data.frame(xwalk))
  expect_true(all(c("isea_cell_id", "h3_cell_id", "isea_area_km2",
                     "h3_area_km2", "area_ratio") %in% names(xwalk)))
  expect_equal(nrow(xwalk), length(unique(cells)))
  expect_true(is.character(xwalk$h3_cell_id))
  expect_true(all(xwalk$area_ratio > 0))
})

test_that("h3_crosswalk() auto-selects H3 resolution", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(c(0), c(45), grid)
  xwalk <- h3_crosswalk(cells, grid)

  # H3 area should be in the same ballpark as ISEA area
  expect_true(xwalk$area_ratio[1] > 0.1)
  expect_true(xwalk$area_ratio[1] < 10)
})

test_that("h3_crosswalk() accepts explicit H3 resolution", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(c(0), c(45), grid)
  xwalk <- h3_crosswalk(cells, grid, h3_resolution = 3)

  expect_equal(nrow(xwalk), 1)
  expect_true(is.character(xwalk$h3_cell_id))
})

test_that("h3_crosswalk() deduplicates cell IDs", {
  grid <- hex_grid(area_km2 = 1000)
  # Two points that map to the same cell
  cells <- lonlat_to_cell(c(0, 0.001), c(45, 45.001), grid)
  xwalk <- h3_crosswalk(cells, grid)

  # Result should have unique ISEA cell IDs
  expect_equal(nrow(xwalk), length(unique(cells)))
})

test_that("h3_crosswalk() works with HexData", {
  grid <- hex_grid(area_km2 = 1000)
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
  xwalk <- h3_crosswalk(grid = result)

  expect_true(is.data.frame(xwalk))
  expect_true(nrow(xwalk) > 0)
})

# =============================================================================
# H3 -> ISEA
# =============================================================================

test_that("h3_crosswalk() maps H3 to ISEA", {
  h3 <- hex_grid(resolution = 5, type = "h3")
  isea <- hex_grid(area_km2 = 1000)
  h3_cells <- lonlat_to_cell(c(0, 10), c(45, 50), h3)

  xwalk <- h3_crosswalk(h3_cells, h3, isea_grid = isea, direction = "h3_to_isea")

  expect_true(is.data.frame(xwalk))
  expect_equal(nrow(xwalk), length(unique(h3_cells)))
  expect_true(is.numeric(xwalk$isea_cell_id))
  expect_true(is.character(xwalk$h3_cell_id))
})

test_that("h3_crosswalk() errors without isea_grid for h3_to_isea", {
  h3 <- hex_grid(resolution = 5, type = "h3")
  h3_cells <- lonlat_to_cell(c(0), c(45), h3)

  expect_error(
    h3_crosswalk(h3_cells, h3, direction = "h3_to_isea"),
    "isea_grid is required"
  )
})

# =============================================================================
# Validation
# =============================================================================

test_that("h3_crosswalk() errors on wrong grid type for direction", {
  h3 <- hex_grid(resolution = 5, type = "h3")
  isea <- hex_grid(area_km2 = 1000)

  h3_cells <- lonlat_to_cell(c(0), c(45), h3)
  isea_cells <- lonlat_to_cell(c(0), c(45), isea)

  expect_error(
    h3_crosswalk(h3_cells, h3, direction = "isea_to_h3"),
    "requires an ISEA grid"
  )
  expect_error(
    h3_crosswalk(isea_cells, isea, direction = "h3_to_isea"),
    "requires an H3 grid"
  )
})

test_that("h3_crosswalk() validates H3 resolution range", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(c(0), c(45), grid)

  expect_error(
    h3_crosswalk(cells, grid, h3_resolution = 16),
    "between 0 and 15"
  )
})

test_that("h3_crosswalk() errors when cell_id is NULL and grid is HexGridInfo", {
  grid <- hex_grid(area_km2 = 1000)
  expect_error(h3_crosswalk(grid = grid), "cell_id required")
})

# R/h3_compat.R
# H3 interoperability layer
#
# import_h3() — ingest external H3 cell IDs into hexify

# =============================================================================
# IMPORT FROM EXTERNAL H3
# =============================================================================

#' Import External H3 Cell IDs into hexify
#'
#' Ingests H3 cell IDs from an external source (another H3 library, a database,
#' or a CSV file) into hexify. Validates cell IDs, infers the H3 resolution,
#' and optionally attaches data to build a HexData object.
#'
#' For converting between grid specs, use
#' \code{\link{hex_grid}(type = "h3")} directly. For cell-level ISEA/H3
#' mapping, use \code{\link{h3_crosswalk}}.
#'
#' @param cell_ids Character vector of H3 cell ID strings
#' @param data Optional data frame to attach. Must have the same number of rows
#'   as \code{cell_ids}. If \code{NULL}, returns a HexGridInfo object.
#' @param validate If \code{TRUE} (default), checks that all cell IDs are valid
#'   H3 cells at the same resolution before proceeding. Set to \code{FALSE} to
#'   skip validation when cell IDs are known to be correct.
#' @param radius_km Radius of the body the cells cover, in kilometers, or a body
#'   name (default Earth). See \code{\link{hex_grid}}.
#'
#' @return If \code{data = NULL}, a HexGridInfo object for the inferred H3
#'   resolution. If \code{data} is provided, a HexData object with data
#'   attached at the specified cells.
#'
#' @details
#' H3 cell IDs encode their resolution in the index itself, so no resolution
#' argument is needed. The resolution is inferred automatically. All cell IDs
#' must share the same resolution; mixed resolutions produce an error.
#'
#' @seealso \code{\link{hex_grid}} for creating grids directly,
#'   \code{\link{h3_crosswalk}} for cell-level ISEA/H3 mapping
#'
#' @export
#' @examples
#' \donttest{
#' # Import external H3 cell IDs (grid spec only)
#' h3_ids <- c("8528342bfffffff", "85283473fffffff", "85283447fffffff")
#' grid <- import_h3(h3_ids)
#' grid
#'
#' # Import with data attached
#' df <- data.frame(species = c("oak", "pine", "birch"), count = c(10, 5, 3))
#' hd <- import_h3(h3_ids, data = df)
#' hd
#' }
import_h3 <- function(cell_ids, data = NULL, validate = TRUE,
                      radius_km = EARTH_RADIUS_KM) {

  if (!is.character(cell_ids)) {
    stop("cell_ids must be a character vector of H3 cell ID strings")
  }
  if (length(cell_ids) == 0) {
    stop("cell_ids must not be empty")
  }

  # Track non-NA positions
  non_na <- !is.na(cell_ids)
  valid_ids <- cell_ids[non_na]

  if (length(valid_ids) == 0) {
    stop("cell_ids contains only NA values")
  }

  # Validate cell IDs
  if (validate) {
    is_valid <- cpp_h3_isValidCell(valid_ids)
    if (any(!is_valid)) {
      n_invalid <- sum(!is_valid)
      first_bad <- valid_ids[which(!is_valid)[1]]
      stop(sprintf(
        "%d invalid H3 cell ID(s). First invalid: '%s'",
        n_invalid, first_bad
      ))
    }
  }

  # Infer resolution (always needed, even when validate = FALSE)
  resolutions <- cpp_h3_getResolution(valid_ids)
  unique_res <- unique(resolutions)

  if (length(unique_res) > 1) {
    stop(sprintf(
      "All cell_ids must share the same H3 resolution. Found resolutions: %s",
      paste(sort(unique_res), collapse = ", ")
    ))
  }

  res <- unique_res[1L]

  # Build HexGridInfo
  grid <- hex_grid(resolution = res, type = "h3", radius_km = radius_km)

  # If no data, return grid only
  if (is.null(data)) {
    return(grid)
  }

  # Validate data dimensions
  if (!inherits(data, "data.frame")) {
    stop(sprintf("data must be a data.frame, got %s", class(data)[1]))
  }
  if (nrow(data) != length(cell_ids)) {
    stop(sprintf(
      "data has %d rows but cell_ids has %d elements; they must match",
      nrow(data), length(cell_ids)
    ))
  }

  # Drop NA cell_id rows, mirroring hexify()'s NA-coordinate handling --
  # otherwise they'd silently survive into the HexData object with
  # cell_id = NA / cell_center = c(NaN, NaN).
  if (any(!non_na)) {
    warning(sprintf("%d cell ID(s) are NA and will be skipped", sum(!non_na)))
    data <- data[non_na, , drop = FALSE]
    cell_ids <- valid_ids
  }

  # Get cell centers for HexData
  centers <- cpp_h3_cellToLatLng(cell_ids)
  cell_center <- cbind(lon = centers$lon, lat = centers$lat)

  new_hex_data(
    data = data,
    grid = grid,
    cell_id = cell_ids,
    cell_center = cell_center
  )
}

# R/hex_zonal.R
# Full zonal statistics: aggregate raster pixels within hex polygons

#' Zonal Statistics for Hex Cells
#'
#' Computes zonal statistics by aggregating all raster pixels falling
#' within each hexagonal cell polygon. More accurate than [hex_extract()]
#' but slower because it requires polygon geometries.
#'
#' @param raster A `terra::SpatRaster` object.
#' @param grid A HexGridInfo or HexData object specifying the grid.
#' @param fun Summary function name: `"mean"` (default), `"sum"`, `"min"`,
#'   `"max"`, `"sd"`, or `"count"`.
#' @param boundary Optional sf polygon to limit the analysis extent.
#' @param cells Optional cell IDs. If provided, only these cells are included.
#'
#' @return A data.frame with columns `cell_id`, plus one column per raster
#'   layer containing the aggregated values.
#'
#' @details
#' Requires the `terra` package (in Suggests). The function:
#' 1. Generates hex polygon geometries via [cell_to_sf()]
#' 2. Calls `terra::extract(raster, polygons, fun = fun)`
#' 3. Joins results back to cell IDs
#'
#' For point-based extraction (faster, at cell centers only), use
#' [hex_extract()] instead.
#'
#' @seealso [hex_extract()] for cell-center extraction,
#'   [cell_to_sf()] for hex polygon generation
#'
#' @export
#' @examples
#' \donttest{
#' if (requireNamespace("terra", quietly = TRUE)) {
#'   r <- terra::rast(nrows = 100, ncols = 100,
#'                    xmin = -10, xmax = 10, ymin = 40, ymax = 55)
#'   terra::values(r) <- runif(10000)
#'   names(r) <- "temperature"
#'
#'   g <- hex_grid(area_km2 = 500)
#'   df <- data.frame(lon = c(0, 5), lat = c(45, 50))
#'   hd <- hexify(df, lon = "lon", lat = "lat", grid = g)
#'   hex_zonal(r, hd, fun = "mean")
#' }
#' }
hex_zonal <- function(raster, grid, fun = "mean", boundary = NULL,
                       cells = NULL) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    stop("Package 'terra' is required for hex_zonal()")
  }

  g <- extract_grid(grid)

  # Get cell IDs
  if (!is.null(cells)) {
    cell_ids <- unique(cells[!is.na(cells)])
  } else if (is_hex_data(grid)) {
    cell_ids <- unique(grid@cell_id)
  } else if (!is.null(boundary)) {
    cell_data <- grid_clip(boundary, g)
    cell_ids <- unique(cell_data$cell_id)
  } else {
    stop("Provide a HexData object, cell IDs via 'cells', or a 'boundary' polygon")
  }

  # Generate hex polygons
  hex_sf <- cell_to_sf(cell_ids, g)

  # Convert to terra SpatVector
  hex_vect <- terra::vect(hex_sf)

  # Extract zonal statistics. terra::extract() has no built-in "count"
  # function, so translate it to a count of non-NA pixels per cell.
  fun_actual <- if (identical(fun, "count")) {
    function(x, ...) sum(!is.na(x))
  } else {
    fun
  }
  extracted <- terra::extract(raster, hex_vect, fun = fun_actual, ID = TRUE)

  # Build result. Key off hex_sf$cell_id (not the original cell_ids) since
  # cell_to_sf() deduplicates internally and hex_vect/extracted follow its
  # row order -- using the pre-dedup cell_ids here would silently misalign
  # rows whenever the input contained duplicates.
  result <- data.frame(cell_id = hex_sf$cell_id, stringsAsFactors = FALSE)
  if ("ID" %in% names(extracted)) {
    extracted$ID <- NULL
  }
  result <- cbind(result, extracted)
  result
}

# tests/testthat/test-raster.R
# Tests for hex_extract() and hex_zonal()

test_that("hex_extract works with synthetic SpatRaster", {
  skip_if_not_installed("terra")

  r <- terra::rast(nrows = 10, ncols = 10,
                    xmin = -10, xmax = 10, ymin = 40, ymax = 55)
  terra::values(r) <- seq_len(100)
  names(r) <- "elevation"

  df <- data.frame(lon = c(0, 5), lat = c(45, 50))
  g <- hex_grid(area_km2 = 500)
  hd <- hexify(df, lon = "lon", lat = "lat", grid = g)

  result <- hex_extract(r, hd)
  expect_true(is.data.frame(result))
  expect_true("cell_id" %in% names(result))
  expect_true("elevation" %in% names(result))
})

test_that("hex_extract errors without terra", {
  skip_if(requireNamespace("terra", quietly = TRUE))
  expect_error(hex_extract(NULL, hex_grid(area_km2 = 1000)), "terra")
})

test_that("hex_zonal works with synthetic SpatRaster", {
  skip_if_not_installed("terra")

  r <- terra::rast(nrows = 100, ncols = 100,
                    xmin = -10, xmax = 10, ymin = 40, ymax = 55)
  terra::values(r) <- runif(10000)
  names(r) <- "temperature"

  df <- data.frame(lon = c(0, 5), lat = c(45, 50))
  g <- hex_grid(area_km2 = 2000)
  hd <- hexify(df, lon = "lon", lat = "lat", grid = g)

  result <- hex_zonal(r, hd, fun = "mean")
  expect_true(is.data.frame(result))
  expect_true("cell_id" %in% names(result))
  expect_true("temperature" %in% names(result))
})

# inst/scripts/compare_dggrid_hexify.R
suppressPackageStartupMessages({
  library(hexify)
  library(dggridR)
  library(dplyr)
  library(tidyr)
  library(readr)
  library(tibble)
})

# ---- config ---------------------------------------------------------------

# Sample lon/lat points (well inside faces; EU cities)
pts <- tibble(
  id  = paste0("p", 1:10),
  lon = c(16.3738,  6.1296, 14.4378, 11.5820,  8.5417,
          2.3522, 12.4964, -3.7038, 12.5683, 21.0122),
  lat = c(48.2082, 49.6116, 50.0755, 48.1351, 47.3769,
          48.8566, 41.9028, 40.4168, 55.6761, 52.2297)
)

areas_km2 <- c(863, 2591, 7774, 23322)
WRITE_CSV  <- FALSE

# ---- helpers --------------------------------------------------------------

# Haversine distance in meters (inputs in degrees)
hav_m <- function(lon1, lat1, lon2, lat2) {
  to_rad <- pi/180
  φ1 <- lat1 * to_rad; φ2 <- lat2 * to_rad
  dφ <- (lat2 - lat1) * to_rad
  dλ <- (lon2 - lon1) * to_rad
  a <- sin(dφ/2)^2 + cos(φ1) * cos(φ2) * sin(dλ/2)^2
  2 * 6371008.8 * asin(pmin(1, sqrt(a)))  # Earth mean radius
}

# hexify cell centers at an effective resolution (ISEA3H, aperture 3)
hexify_center_ll <- function(lon_deg, lat_deg, eff_res) {
  a <- hexify_assign(lon_deg, lat_deg, eff_res)
  cbind(lon = a$center_lon, lat = a$center_lat)
}

# ---- compare across areas -------------------------------------------------

all_comparisons <- vector("list", length(areas_km2))
names(all_comparisons) <- as.character(areas_km2)

for (area in areas_km2) {
  grid <- dgconstruct(area = area, topology = "HEXAGON", metric = TRUE, resround = "nearest")
  eff_res <- grid$res

  cat(sprintf(
    "Resolution: %d, Area (km^2): %.12f, Spacing (km): %.10f, CLS (km): %.10f\n",
    eff_res, grid$Area, grid$Spacing, grid$CLS
  ))

  # dggridR centers (coerce list -> tibble)
  seqnum <- dgGEO_to_SEQNUM(grid, pts$lon, pts$lat)$seqnum
  centers_list <- dgSEQNUM_to_GEO(grid, seqnum)  # list with lon_deg, lat_deg
  dgg_df <- tibble(
    id      = pts$id,
    dgg_lon = as.numeric(centers_list$lon_deg),
    dgg_lat = as.numeric(centers_list$lat_deg)
  )

  # hexify centers at same effective resolution
  hex_mat <- hexify_center_ll(pts$lon, pts$lat, eff_res)
  hex_df <- as_tibble(hex_mat) |>
    mutate(id = pts$id, .before = 1L) |>
    rename(hex_lon = lon, hex_lat = lat)

  # join and compute spherical distance error (meters)
  cmp <- dgg_df |>
    left_join(hex_df, by = "id") |>
    mutate(
      area_km2 = area,
      gc_err_m = hav_m(dgg_lon, dgg_lat, hex_lon, hex_lat)
    )

  all_comparisons[[as.character(area)]] <- cmp
}

cmp_all <- bind_rows(all_comparisons)

summary_tbl <- cmp_all |>
  group_by(area_km2) |>
  summarize(
    n           = n(),
    max_err_m   = max(gc_err_m, na.rm = TRUE),
    mean_err_m  = mean(gc_err_m, na.rm = TRUE),
    median_err_m= median(gc_err_m, na.rm = TRUE),
    .groups = "drop"
  )

cat("\nGreat-circle error summary (hexify vs dggridR centers):\n")
print(summary_tbl, n = Inf)

if (isTRUE(WRITE_CSV)) {
  write_csv(cmp_all,    "compare_hexify_dggrid_points.csv")
  write_csv(summary_tbl,"compare_hexify_dggrid_summary.csv")
  cat("\nWrote: compare_hexify_dggrid_points.csv and compare_hexify_dggrid_summary.csv\n")
}

# hexify_index.R
# Unified cell indexing: space-filling curves (Z3, Z7, Z-order) and sequential numbers
#
# This module provides all cell identification and indexing operations.
# Each cell can be identified by:
#   1. Coordinates: (face, i, j, resolution)
#   2. Index string: hierarchical encoding (Z3, Z7, Z-order)
#   3. Seqnum: integer sequential number
#
# @name hexify-indexing

# =============================================================================
# INDEX STRING OPERATIONS
# =============================================================================

#' Convert cell coordinates to index string
#'
#' Converts cell coordinates (`face`, `i`, `j`) to a hierarchical index string.
#' The index type is selected from `aperture` when `index_type = "auto"`.
#'
#' @param face Face/quad number (0-19)
#' @param i I coordinate
#' @param j J coordinate
#' @param resolution Resolution level
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index encoding: "auto" (default), "z3", "z7", or "zorder"
#'
#' @return A character vector of index strings.
#'
#' @details
#' Default index types by aperture:
#' - Aperture 3: Z3 (optimized digit selection)
#' - Aperture 4: Z-order (Morton curve)
#' - Aperture 7: Z7 (one base-7 child digit per resolution)
#'
#' A Z7 index has the form `BBd1...dr`, where `BB` is the two-digit base
#' cell (00--11), `r` is the resolution, and every child digit is in 0--6.
#' hexify's Z7 encoding is bijective: decoding and re-encoding a valid index
#' returns the same string. It follows DGGRID's Z7 layout for ordinary cells,
#' but retains distinct indices in pentagon regions where DGGRID's encoder can
#' map different cells to the same string.
#'
#' @family hierarchical index
#' @keywords internal
#' @export
#' @examples
#' idx <- hexify_cell_to_index(5, 10, 15, resolution = 3, aperture = 3)
#'
#' # Aperture-7 indices contain a two-digit base cell and one digit per level
#' z7 <- hexify_cell_to_index(5, 1, 1, resolution = 2, aperture = 7)
#' nchar(z7) == 4L
hexify_cell_to_index <- function(face, i, j, resolution, aperture = 3L,
                                  index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  validate_resolution(resolution)
  validate_aperture(aperture)
  cpp_cell_to_index(
    as.integer(face), as.numeric(i), as.numeric(j),
    as.integer(resolution), as.integer(aperture), index_type
  )
}

#' Convert index string to cell coordinates
#'
#' Decodes a hierarchical index string back to its cell coordinates and
#' resolution. For Z7, valid indices round-trip exactly through
#' [hexify_cell_to_index()].
#'
#' @param index Index string
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index encoding used. Default "auto" infers from aperture.
#'
#' @return A list with `face`, `i`, `j`, and `resolution`.
#'
#' @family hierarchical index
#' @keywords internal
#' @export
#' @examples
#' cell <- hexify_index_to_cell("0012012", aperture = 3)
#'
#' z7_cell <- hexify_index_to_cell("110001", aperture = 7)
#' hexify_cell_to_index(z7_cell$face, z7_cell$i, z7_cell$j,
#'   z7_cell$resolution, aperture = 7
#' )
hexify_index_to_cell <- function(index, aperture = 3L,
                                  index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  validate_aperture(aperture)
  cpp_index_to_cell(as.character(index), as.integer(aperture), index_type)
}

#' Convert longitude/latitude to index string
#'
#' Projects geographic coordinates to grid cells and returns their hierarchical
#' index strings. Inputs are vectorized over `lon` and `lat`.
#'
#' @param lon Longitude in degrees
#' @param lat Latitude in degrees
#' @param resolution Resolution level
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index encoding: "auto" (default), "z3", "z7", or "zorder"
#'
#' @return A character vector of index strings.
#'
#' @family hierarchical index
#' @keywords internal
#' @export
#' @examples
#' idx <- hexify_lonlat_to_index(16.37, 48.21, resolution = 5, aperture = 3)
#' idx7 <- hexify_lonlat_to_index(16.37, 48.21,
#'   resolution = 4, aperture = 7
#' )
hexify_lonlat_to_index <- function(lon, lat, resolution, aperture = 3L,
                                    index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  validate_resolution(resolution)
  validate_aperture(aperture)
  cpp_lonlat_to_index(lon, lat, as.integer(resolution),
                      as.integer(aperture), index_type)
}

#' Convert index string to longitude/latitude
#'
#' Returns the geographic coordinates of an indexed cell's center. This is a
#' cell-level inverse of [hexify_lonlat_to_index()]: the returned point is the
#' center, not necessarily the original input point.
#'
#' @param index Index string
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index encoding. Default "auto" infers from aperture.
#'
#' @return A named numeric vector with `lon` and `lat` in degrees.
#'
#' @family hierarchical index
#' @keywords internal
#' @export
#' @examples
#' coords <- hexify_index_to_lonlat("0012012", aperture = 3)
hexify_index_to_lonlat <- function(index, aperture = 3L,
                                    index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  validate_aperture(aperture)
  cpp_index_to_lonlat(as.character(index), as.integer(aperture), index_type)
}

# =============================================================================
# INDEX HIERARCHY OPERATIONS
# =============================================================================

#' Get parent index
#'
#' Returns the parent index (one resolution coarser).
#'
#' @param index Index string
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index encoding. Default "auto".
#'
#' @return Parent index string
#'
#' @family hierarchical index
#' @keywords internal
#' @export
hexify_get_parent <- function(index, aperture = 3L,
                               index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  validate_aperture(aperture)
  cpp_get_parent_index(as.character(index), as.integer(aperture), index_type)
}

#' Get children indices
#'
#' Returns all children indices (one resolution finer).
#'
#' @param index Index string
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index encoding. Default "auto".
#'
#' @return Character vector of child indices
#'
#' @family hierarchical index
#' @keywords internal
#' @export
hexify_get_children <- function(index, aperture = 3L,
                                 index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  validate_aperture(aperture)
  cpp_get_children_indices(as.character(index), as.integer(aperture), index_type)
}

#' Get index resolution
#'
#' Returns the resolution level encoded in an index string.
#'
#' @param index Index string
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index encoding. Default "auto".
#'
#' @return Integer resolution level
#'
#' @family hierarchical index
#' @keywords internal
#' @export
hexify_get_resolution <- function(index, aperture = 3L,
                                   index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  validate_aperture(aperture)
  cpp_get_index_resolution(as.character(index), as.integer(aperture), index_type)
}

#' Compare two indices
#'
#' Lexicographic comparison of two index strings.
#'
#' @param idx1 First index string
#' @param idx2 Second index string
#'
#' @return Integer: -1 if idx1 < idx2, 0 if equal, 1 if idx1 > idx2
#'
#' @family hierarchical index
#' @keywords internal
#' @export
hexify_compare_indices <- function(idx1, idx2) {
  cpp_compare_indices(as.character(idx1), as.character(idx2))
}

# =============================================================================
# INTEGER CELL IDs
# =============================================================================

#' Convert longitude/latitude to cell ID
#'
#' Converts geographic coordinates to DGGRID-compatible cell identifiers.
#' This is the primary function for geocoding points to grid cells.
#'
#' @param lon Numeric vector of longitudes in degrees
#' @param lat Numeric vector of latitudes in degrees
#' @param resolution Grid resolution (integer >= 0)
#' @param aperture Grid aperture (3, 4, or 7)
#'
#' @return Numeric vector of cell IDs (1-based)
#'
#' @details
#' Returns DGGRID-compatible cell identifiers. The cell ID
#' uniquely identifies each hexagonal cell in the global grid.
#'
#' For a grid-object interface, use \code{\link{lonlat_to_cell}}.
#'
#' @family coordinate conversion
#' @seealso \code{\link{lonlat_to_cell}} for the recommended S4 interface,
#'   \code{\link{hexify_cell_to_lonlat}} for the inverse operation
#'
#' @keywords internal
#' @export
#' @examples
#' cell_id <- hexify_lonlat_to_cell(0, 45, resolution = 5, aperture = 3)
hexify_lonlat_to_cell <- function(lon, lat, resolution, aperture) {
  validate_lon(lon)
  validate_lat(lat)
  validate_resolution(resolution)
  validate_aperture(aperture)
  cpp_lonlat_to_cell(lon, lat, resolution, aperture)
}

#' Convert cell ID to longitude/latitude
#'
#' Converts cell identifiers back to cell center coordinates.
#' This is the inverse of \code{\link{hexify_lonlat_to_cell}}.
#'
#' @param cell_id Numeric vector of cell IDs (1-based)
#' @param resolution Grid resolution (integer >= 0)
#' @param aperture Grid aperture (3, 4, or 7)
#'
#' @return Data frame with lon_deg and lat_deg columns
#'
#' @family coordinate conversion
#' @seealso \code{\link{cell_to_lonlat}} for the recommended S4 interface,
#'   \code{\link{hexify_lonlat_to_cell}} for the forward operation
#'
#' @keywords internal
#' @export
#' @examples
#' coords <- hexify_cell_to_lonlat(1702, resolution = 5, aperture = 3)
hexify_cell_to_lonlat <- function(cell_id, resolution, aperture) {
  validate_resolution(resolution)
  validate_aperture(aperture)
  validate_cell_id(cell_id, resolution, aperture)
  cpp_cell_to_lonlat(cell_id, resolution, aperture)
}

#' Get cell info from cell ID
#'
#' Converts cell ID to cell components (quad, i, j).
#'
#' @param cell_id Numeric vector of cell IDs (1-based)
#' @param resolution Grid resolution (integer >= 0)
#' @param aperture Grid aperture (3, 4, or 7)
#'
#' @return Data frame with quad, i, j columns
#'
#' @family coordinate conversion
#' @keywords internal
#' @export
#' @examples
#' info <- hexify_cell_id_to_quad_ij(1702, resolution = 5, aperture = 3)
hexify_cell_id_to_quad_ij <- function(cell_id, resolution, aperture) {
  validate_resolution(resolution)
  validate_aperture(aperture)
  validate_cell_id(cell_id, resolution, aperture)
  result <- cpp_cell_to_quad_ij(cell_id, resolution, aperture)
  as.data.frame(result)
}

# =============================================================================
# Z7 SPECIAL OPERATIONS
# =============================================================================

#' Get canonical form of Z7 index
#'
#' Decodes and re-encodes a Z7 index until it reaches a stable form. Current Z7
#' indices are bijective, so every valid index is already canonical and this
#' function normally returns its input unchanged. It remains available for
#' validating or normalizing indices created by older hexify versions.
#'
#' @param index A length-one Z7 index string.
#' @param max_iterations Maximum number of decode/encode iterations. This is a
#'   safety bound for legacy indices; the default is 128.
#'
#' @return A length-one character string containing the stable index.
#'
#' @family hierarchical index
#' @keywords internal
#' @export
#' @examples
#' # Valid Z7 indices are stable
#' hexify_z7_canonical("110001")
#'
#' cell <- hexify_index_to_cell("110001", aperture = 7)
#' identical(
#'   hexify_cell_to_index(cell$face, cell$i, cell$j,
#'     cell$resolution, aperture = 7
#'   ),
#'   "110001"
#' )
hexify_z7_canonical <- function(index, max_iterations = 128L) {
  cpp_z7_canonical_form(as.character(index), as.integer(max_iterations))
}

# =============================================================================
# INDEX TYPE UTILITIES
# =============================================================================

#' Check if index type is valid for aperture
#'
#' @param aperture Aperture (3, 4, or 7)
#' @param index_type Index type to check
#'
#' @return Logical: TRUE if valid combination
#'
#' @family hierarchical index
#' @keywords internal
#' @export
hexify_is_valid_index_type <- function(aperture,
                                        index_type = c("auto", "z3", "z7", "zorder")) {
  index_type <- match.arg(index_type)
  cpp_is_valid_index_type(as.integer(aperture), index_type)
}

#' Get default index type for aperture
#'
#' Returns the recommended index type:
#' - Aperture 3: "z3"
#' - Aperture 4: "zorder"
#' - Aperture 7: "z7"
#'
#' @param aperture Aperture (3, 4, or 7)
#'
#' @return String: "z3", "z7", or "zorder"
#'
#' @family hierarchical index
#' @keywords internal
#' @export
hexify_default_index_type <- function(aperture) {
  cpp_get_default_index_type(as.integer(aperture))
}

# =============================================================================
# RESOLUTION CONVERSION (ISEA3H-specific)
# =============================================================================

#' Convert effective resolution to area
#'
#' Calculates approximate cell area for ISEA3H (aperture 3) grids.
#' Based on calibration: eff_res 10 = 863.8006 km^2.
#'
#' @param eff_res Effective resolution (0.5 * resolution for aperture 3)
#'
#' @return Area in km^2
#'
#' @family grid statistics
#' @keywords internal
#' @export
hexify_eff_res_to_area <- function(eff_res) {
  ISEA3H_RES10_AREA_KM2 * (3^(10 - eff_res))
}

#' Convert area to effective resolution
#'
#' Finds the effective resolution that gives approximately the target area.
#'
#' @param area_km2 Target area in km^2
#'
#' @return Effective resolution (may be fractional)
#'
#' @family grid statistics
#' @keywords internal
#' @export
hexify_area_to_eff_res <- function(area_km2) {
  10 - log(area_km2 / ISEA3H_RES10_AREA_KM2) / log(3)
}

#' Convert effective resolution to index resolution
#'
#' For aperture 3: res_index = 2 * eff_res - 1 for odd index resolutions.
#'
#' @param eff_res Effective resolution
#'
#' @return Index resolution (integer)
#'
#' @family grid statistics
#' @keywords internal
#' @export
hexify_eff_res_to_resolution <- function(eff_res) {

  as.integer(2 * eff_res - 1)
}

#' Convert index resolution to effective resolution
#'
#' @param resolution Index resolution
#'
#' @return Effective resolution
#'
#' @family grid statistics
#' @keywords internal
#' @export
hexify_resolution_to_eff_res <- function(resolution) {
  (resolution + 1) / 2
}

# tests/testthat/test-index-utils.R
# Tests for hexify_index.R utility functions
#
# Functions tested:
# - hexify_get_parent()
# - hexify_get_children()
# - hexify_get_resolution()
# - hexify_compare_indices()
# - hexify_is_valid_index_type()
# - hexify_default_index_type()
# - hexify_eff_res_to_area()
# - hexify_area_to_eff_res()
# - hexify_eff_res_to_resolution()
# - hexify_resolution_to_eff_res()

setup_icosa <- function() {
  cpp_build_icosa()
}

# =============================================================================
# GET PARENT INDEX
# =============================================================================

test_that("hexify_get_parent returns shorter index for aperture 3", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 5, aperture = 3)
  parent <- hexify_get_parent(index, aperture = 3)

  expect_true(nchar(parent) < nchar(index))
})

test_that("hexify_get_parent returns shorter index for aperture 4", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 5, aperture = 4)
  parent <- hexify_get_parent(index, aperture = 4, index_type = "zorder")

  expect_true(nchar(parent) < nchar(index))
})

test_that("hexify_get_parent returns shorter index for aperture 7", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 5, aperture = 7)
  parent <- hexify_get_parent(index, aperture = 7, index_type = "z7")

  expect_true(nchar(parent) < nchar(index))
})

test_that("hexify_get_parent of child produces shorter index", {
  setup_icosa()

  # Get a parent index
  parent <- hexify_lonlat_to_index(10, 50, resolution = 3, aperture = 3)

  # Get children
  children <- hexify_get_children(parent, aperture = 3)

  # Verify each child's parent is shorter than the child
  for (child in children) {
    recovered_parent <- hexify_get_parent(child, aperture = 3)
    expect_true(nchar(recovered_parent) < nchar(child))
  }
})

# =============================================================================
# GET CHILDREN INDICES
# =============================================================================

test_that("hexify_get_children returns correct number for aperture 3", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 3, aperture = 3)
  children <- hexify_get_children(index, aperture = 3)

  # Aperture 3 should have 3 children
  expect_equal(length(children), 3)
})

test_that("hexify_get_children returns correct number for aperture 4", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 3, aperture = 4)
  children <- hexify_get_children(index, aperture = 4, index_type = "zorder")

  # Aperture 4 should have 4 children
  expect_equal(length(children), 4)
})

test_that("hexify_get_children returns correct number for aperture 7", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 3, aperture = 7)
  children <- hexify_get_children(index, aperture = 7, index_type = "z7")

  # Aperture 7 should have 7 children
  expect_equal(length(children), 7)
})

test_that("hexify_get_children returns longer indices", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 3, aperture = 3)
  children <- hexify_get_children(index, aperture = 3)

  for (child in children) {
    expect_true(nchar(child) > nchar(index))
  }
})

# =============================================================================
# GET RESOLUTION
# =============================================================================

test_that("hexify_get_resolution returns correct value for aperture 3", {
  setup_icosa()

  for (res in c(1, 3, 5, 7)) {
    index <- hexify_lonlat_to_index(10, 50, resolution = res, aperture = 3)
    recovered_res <- hexify_get_resolution(index, aperture = 3)
    expect_equal(recovered_res, res)
  }
})

test_that("hexify_get_resolution returns correct value for aperture 4", {
  setup_icosa()

  for (res in c(1, 3, 5, 7)) {
    index <- hexify_lonlat_to_index(10, 50, resolution = res, aperture = 4)
    recovered_res <- hexify_get_resolution(
      index, aperture = 4, index_type = "zorder"
    )
    expect_equal(recovered_res, res)
  }
})

test_that("hexify_get_resolution returns correct value for aperture 7", {
  setup_icosa()

  for (res in c(1, 3, 5)) {
    index <- hexify_lonlat_to_index(10, 50, resolution = res, aperture = 7)
    recovered_res <- hexify_get_resolution(
      index, aperture = 7, index_type = "z7"
    )
    expect_equal(recovered_res, res)
  }
})

# =============================================================================
# COMPARE INDICES
# =============================================================================

test_that("hexify_compare_indices returns 0 for equal indices", {
  setup_icosa()

  index <- hexify_lonlat_to_index(10, 50, resolution = 5, aperture = 3)
  result <- hexify_compare_indices(index, index)

  expect_equal(result, 0)
})

test_that("hexify_compare_indices returns -1 for smaller first index", {
  result <- hexify_compare_indices("010000", "020000")
  expect_equal(result, -1)
})

test_that("hexify_compare_indices returns 1 for larger first index", {
  result <- hexify_compare_indices("020000", "010000")
  expect_equal(result, 1)
})

test_that("hexify_compare_indices compares by length when equal prefix", {
  # Shorter index should be "less than" longer
  result <- hexify_compare_indices("010", "0100")
  expect_equal(result, -1)
})

# =============================================================================
# VALID INDEX TYPE
# =============================================================================

test_that("hexify_is_valid_index_type returns TRUE for matching types", {
  expect_true(hexify_is_valid_index_type(3, "z3"))
  expect_true(hexify_is_valid_index_type(4, "zorder"))
  expect_true(hexify_is_valid_index_type(7, "z7"))
})

test_that("hexify_is_valid_index_type returns TRUE for auto", {
  expect_true(hexify_is_valid_index_type(3, "auto"))
  expect_true(hexify_is_valid_index_type(4, "auto"))
  expect_true(hexify_is_valid_index_type(7, "auto"))
})

# =============================================================================
# DEFAULT INDEX TYPE
# =============================================================================

test_that("hexify_default_index_type returns correct types", {
  expect_equal(hexify_default_index_type(3), "z3")
  expect_equal(hexify_default_index_type(4), "zorder")
  expect_equal(hexify_default_index_type(7), "z7")
})

# =============================================================================
# EFFECTIVE RESOLUTION CONVERSIONS
# =============================================================================

test_that("hexify_eff_res_to_area returns positive values", {
  for (eff_res in c(1, 5, 10, 15)) {
    area <- hexify_eff_res_to_area(eff_res)
    expect_true(area > 0)
  }
})

test_that("hexify_eff_res_to_area decreases with resolution", {
  areas <- vapply(1:10, hexify_eff_res_to_area, numeric(1))

  # Each resolution should have smaller area than previous
  expect_true(all(diff(areas) < 0))
})

test_that("hexify_area_to_eff_res is inverse of hexify_eff_res_to_area", {
  for (eff_res in c(1, 5, 10, 15)) {
    area <- hexify_eff_res_to_area(eff_res)
    recovered_res <- hexify_area_to_eff_res(area)
    expect_equal(recovered_res, eff_res, tolerance = 1e-10)
  }
})

test_that("hexify_area_to_eff_res returns expected values", {
  # At eff_res 10, area should be ISEA3H_RES10_AREA_KM2 = 863.8006
  eff_res <- hexify_area_to_eff_res(863.8006)
  expect_equal(eff_res, 10, tolerance = 1e-3)
})

test_that("hexify_eff_res_to_resolution returns integer", {
  for (eff_res in c(1, 5, 10)) {
    res <- hexify_eff_res_to_resolution(eff_res)
    expect_true(is.integer(res))
  }
})

test_that("hexify_eff_res_to_resolution formula is correct", {
  # res_index = 2 * eff_res - 1
  expect_equal(hexify_eff_res_to_resolution(1), 1L)
  expect_equal(hexify_eff_res_to_resolution(5), 9L)
  expect_equal(hexify_eff_res_to_resolution(10), 19L)
})

test_that("hexify_resolution_to_eff_res is inverse", {
  for (res in c(1, 5, 9, 19)) {
    eff_res <- hexify_resolution_to_eff_res(res)
    recovered_res <- hexify_eff_res_to_resolution(eff_res)
    expect_equal(recovered_res, res)
  }
})

test_that("hexify_resolution_to_eff_res formula is correct", {
  # eff_res = (resolution + 1) / 2
  expect_equal(hexify_resolution_to_eff_res(1), 1)
  expect_equal(hexify_resolution_to_eff_res(9), 5)
  expect_equal(hexify_resolution_to_eff_res(19), 10)
})

# =============================================================================
# INDEX STRING OPERATIONS
# =============================================================================

test_that("hexify_cell_to_index and hexify_index_to_cell round-trip", {
  setup_icosa()

  face <- 5
  i <- 10
  j <- 15
  resolution <- 5
  aperture <- 3

  # cell -> index -> cell
  index <- hexify_cell_to_index(face, i, j, resolution, aperture)
  cell <- hexify_index_to_cell(index, aperture)

  # Verify structure is returned
  expect_true("face" %in% names(cell))
  expect_true("i" %in% names(cell))
  expect_true("j" %in% names(cell))
  expect_true("resolution" %in% names(cell))
  expect_equal(cell$resolution, resolution)
})

test_that("hexify_lonlat_to_index and hexify_index_to_lonlat are consistent", {
  setup_icosa()

  lon <- 10
  lat <- 50

  # lon/lat -> index -> lon/lat
  index <- hexify_lonlat_to_index(lon, lat, resolution = 5, aperture = 3)
  coords <- hexify_index_to_lonlat(index, aperture = 3)

  # Reconstructed coordinates should be within one cell diameter
  lon_diff <- abs(coords["lon"] - lon)
  lat_diff <- abs(coords["lat"] - lat)

  expect_true(lon_diff < 5)
  expect_true(lat_diff < 5)
})

test_that("hexify_index_to_lonlat agrees with cell_to_lonlat for apertures 3 and 4", {
  setup_icosa()

  # Both are cell-centre lookups for the same cell, so they must agree exactly.
  # They used to differ at the polar pentagons, where folding (0, 0) through the
  # quad frame returns the icosahedron vertex rather than the pole.
  set.seed(311)
  lon <- c(runif(30, -170, 170), 0, 0)
  lat <- c(runif(30, -80, 80), 89.9, -89.9)

  for (ap in c(3, 4)) {
    for (res in c(1, 2, 3, 5)) {
      g <- hex_grid(resolution = res, aperture = ap)
      cells <- lonlat_to_cell(lon, lat, g)
      ctr <- cell_to_lonlat(cells, g)
      idx <- cell_to_index(cells, g)

      for (k in seq_along(idx)) {
        got <- hexify_index_to_lonlat(idx[k], aperture = ap)
        expect_equal(as.numeric(got[["lon"]]), ctr$lon[k], tolerance = 1e-4,
                     info = sprintf("ap=%d res=%d cell %d lon", ap, res, k))
        expect_equal(as.numeric(got[["lat"]]), ctr$lat[k], tolerance = 1e-4,
                     info = sprintf("ap=%d res=%d cell %d lat", ap, res, k))
      }
    }
  }
})

# test-index_z7.R
# Tests for Z7 encoding/decoding
# Tests for hexify's bijective aperture-7 index.

library(testthat)

test_that("Z7: Valid indices round-trip correctly", {
  # Test indices that are known to be valid and should round-trip
  # Avoid pentagon edge cases by using hexagon faces (1-10) and safe indices
  
  test_indices <- list(
    "01",     # Face 1 center
    "020",    # Face 2, digit 0
    "031",    # Face 3, digit 1  
    "044",    # Face 4, digit 4
    "055",    # Face 5, digit 5 (this one works even on pentagons)
    "066",    # Face 6, digit 6
    "0700",   # Face 7, safe multi-digit
    "0811",   # Face 8
    "0922",   # Face 9
    "1033"    # Face 10
  )
  
  for (idx in test_indices) {
    result <- hexify_index_to_cell(idx, 7L, "z7")
    idx2 <- hexify_cell_to_index(result$face, result$i, result$j, 
                          result$resolution, 7L, "z7")
    
    expect_equal(idx2, idx, 
                 info = sprintf("Index %s should round-trip correctly", idx))
  }
})

test_that("Z7: Resolution 0 (base cells) work correctly", {
  # All 12 base cells should round-trip at resolution 0
  for (face in 0:11) {
    idx <- hexify_cell_to_index(face, 0L, 0L, 0L, 7L, "z7")
    expected <- sprintf("%02d", face)
    expect_equal(idx, expected)
    
    result <- hexify_index_to_cell(idx, 7L, "z7")
    expect_equal(as.integer(result$face), face)
    expect_equal(as.integer(result$i), 0L)
    expect_equal(as.integer(result$j), 0L)
    expect_equal(as.integer(result$resolution), 0L)
  }
})

test_that("Z7: Hexagon faces handle adjacency remapping", {
  # Some coordinates on hexagon faces get remapped to other faces
  # This is expected behavior for aperture 7's cross-face structure
  
  # Test that we can at least encode and decode consistently
  test_cases <- list(
    list(face = 4L, i = 0L, j = 0L, res = 1L),  # Should work
    list(face = 4L, i = 1L, j = 0L, res = 1L),  # May remap
    list(face = 4L, i = 0L, j = 1L, res = 1L),  # May remap
    list(face = 4L, i = 1L, j = 1L, res = 1L),  # May remap
    list(face = 4L, i = 2L, j = 1L, res = 1L)   # May remap
  )
  
  for (tc in test_cases) {
    # Just test that encoding and decoding works without errors
    idx <- hexify_cell_to_index(tc$face, tc$i, tc$j, tc$res, 7L, "z7")
    result <- hexify_index_to_cell(idx, 7L, "z7")
    
    # Resolution should be preserved (extract the numeric value)
    expect_equal(as.integer(result$resolution), tc$res,
                 info = sprintf("Resolution preserved for face=%d i=%d j=%d", 
                               tc$face, tc$i, tc$j))
    
    # Face might change due to adjacency remapping - that's OK
    # Coordinates might change too - that's also OK for aperture 7
  }
})

test_that("Z7: Pentagon rotation behavior follows DGGRID", {
  # Test pentagon-specific behavior as implemented in DGGRID
  # Note: The exact encoding depends on coordinate mappings
  
  # Test that pentagon faces can be encoded/decoded
  test_cases <- list(
    list(face = 0L, i = 0L, j = 0L, res = 1L, expected = "000"),
    list(face = 11L, i = 0L, j = 0L, res = 1L, expected = "110")
  )
  
  for (tc in test_cases) {
    idx <- hexify_cell_to_index(tc$face, tc$i, tc$j, tc$res, 7L, "z7")
    expect_equal(idx, tc$expected,
                 info = sprintf("Pentagon encoding for face %d", tc$face))
  }
  
  # Test known pentagon rotation behaviors
  result_002 <- hexify_index_to_cell("002", 7L, "z7")
  idx_002_re <- hexify_cell_to_index(result_002$face, result_002$i, result_002$j,
                              result_002$resolution, 7L, "z7")
  # Exact result depends on DGGRID implementation
  
  result_115 <- hexify_index_to_cell("115", 7L, "z7")
  idx_115_re <- hexify_cell_to_index(result_115$face, result_115$i, result_115$j,
                              result_115$resolution, 7L, "z7")
  # Exact result depends on DGGRID implementation
})

test_that("Z7: Multi-digit indices round-trip", {
  
  test_cases <- list(
    list(idx = "0111111", expected = "0111111"),
    list(idx = "0222222", expected = "0222222"),
    list(idx = "0333333", expected = "0333333"),
    list(idx = "0400000", expected = "0400000"),
    list(idx = "0511111", expected = "0511111"),
    list(idx = "0622222", expected = "0622222"),
    list(idx = "0733333", expected = "0733333"),
    list(idx = "0844444", expected = "0844444"),
    list(idx = "0955555", expected = "0955555"),
    list(idx = "1066666", expected = "1066666")
  )
  
  for (tc in test_cases) {
    result <- hexify_index_to_cell(tc$idx, 7L, "z7")
    idx2 <- hexify_cell_to_index(result$face, result$i, result$j,
                         result$resolution, 7L, "z7")
    expect_equal(idx2, tc$expected,
                 info = sprintf("Index %s should round-trip to %s",
                               tc$idx, tc$expected))
  }
})

test_that("Z7: Parent-child relationships work", {
  # Test that valid parent indices have consistent children
  # Use hexagon faces to avoid pentagon complications
  
  test_parents <- c("01", "02", "03", "04", "06", "07", "08", "09", "10")
  
  for (parent_idx in test_parents) {
    parent_result <- hexify_index_to_cell(parent_idx, 7L, "z7")
    
    # Generate children and verify they decode properly
    for (digit in 0:6) {
      child_idx <- paste0(parent_idx, digit)
      
      # Decode child and verify it's valid
      child_result <- tryCatch({
        hexify_index_to_cell(child_idx, 7L, "z7")
      }, error = function(e) NULL)
      
      if (!is.null(child_result)) {
        # Child should be one resolution finer (need to extract numeric value)
        expect_equal(as.integer(child_result$resolution), 
                     as.integer(parent_result$resolution) + 1L,
                     info = sprintf("Child %s resolution check", child_idx))
        
        # Every child has a unique, stable index.
        child_idx2 <- hexify_cell_to_index(child_result$face, child_result$i,
                                    child_result$j, child_result$resolution, 7L, "z7")
        
        expect_equal(child_idx2, child_idx,
                     info = sprintf("Child %s should round-trip", child_idx))
      }
    }
  }
})

test_that("Z7: Edge cases are handled correctly", {
  # Test various edge cases and error conditions
  
  # Single digit should fail (need at least 2 for face)
  expect_error(hexify_index_to_cell("1", 7L, "z7"))
  
  # Face numbers 12 and 99 may or may not be errors depending on implementation
  # DGGRID might handle them differently than expected
  # Let's just test that they can be processed without crashing
  result_12 <- tryCatch(hexify_index_to_cell("12", 7L, "z7"), error = function(e) NULL)
  result_99 <- tryCatch(hexify_index_to_cell("99", 7L, "z7"), error = function(e) NULL)
  
  # Empty string should fail
  expect_error(hexify_index_to_cell("", 7L, "z7"))
  
  # Arbitrary integer coordinates should round-trip without crashing.
  idx_neg <- hexify_cell_to_index(5L, -1L, -1L, 1L, 7L, "z7")
  result_neg <- hexify_index_to_cell(idx_neg, 7L, "z7")
  expect_equal(c(result_neg$i, result_neg$j), c(-1, -1),
               info = "Negative coords should round-trip")
})

test_that("Z7: Resolution progression works correctly", {
  # The cells covering one point, at three resolutions: the index carries the
  # two-digit leading field and one digit per level.
  for (res in 1:3) {
    cell <- hexify_lonlat_to_cell(10, 45, res, 7L)
    qij <- hexify_cell_to_quad_ij(cell, res, 7L)
    idx <- hexify_cell_to_index(qij$quad, qij$i, qij$j, res, 7L, "z7")
    expect_equal(nchar(idx), 2L + res)
  }
})

test_that("Z7: a coordinate outside its quad is rejected", {
  # (49, 49) at resolution 3 lies outside quad 5's substrate box, so no cell of
  # that quad carries it and the hierarchy walk leaves the quad's own base cell.
  expect_error(hexify_cell_to_index(5L, 49L, 49L, 3L, 7L, "z7"),
               "does not lie in the given quad")
})

test_that("Z7: Former DGGRID problem indices are stable", {
  problem_indices <- c("012", "110001", "110002", "110004", "110006")

  for (idx in problem_indices) {
    result <- hexify_index_to_cell(idx, 7L, "z7")
    re_encoded <- hexify_cell_to_index(result$face, result$i, result$j,
                                result$resolution, 7L, "z7")
    expect_equal(re_encoded, idx,
                 info = sprintf("%s should round-trip", idx))
  }
})

test_that("Z7: Canonical forms provide stability", {
  for (idx in c("012", "110001", "110002", "110004", "110006", "0955555")) {
    expect_equal(hexify_z7_canonical(idx), idx)
    canonical <- hexify_z7_canonical(idx)
    result <- hexify_index_to_cell(canonical, 7L, "z7")
    re_encoded <- hexify_cell_to_index(result$face, result$i, result$j,
                                result$resolution, 7L, "z7")
    canonical2 <- hexify_z7_canonical(re_encoded)
    expect_equal(canonical2, canonical,
                 info = sprintf("Canonical form of %s should be stable", idx))
  }
})

test_that("Z7: every cell round-trips through its index (#53)", {
  set.seed(53)
  for (res in 1:5) {
    n_cells <- 10 * 7^res + 2
    ids <- if (n_cells <= 600) seq_len(n_cells) else sample.int(n_cells, 600)

    qij <- hexify_cell_to_quad_ij(ids, res, 7L)
    idx <- vapply(seq_along(ids), function(k) {
      hexify_cell_to_index(qij$quad[k], qij$i[k], qij$j[k], res, 7L, "z7")
    }, character(1))

    back <- vapply(idx, function(s) {
      r <- hexify_index_to_cell(s, 7L, "z7")
      hexify_quad_ij_to_cell(r$face, r$i, r$j, r$resolution, 7L)
    }, numeric(1), USE.NAMES = FALSE)

    expect_equal(back, as.numeric(ids),
                 info = sprintf("resolution %d round-trip", res))
    expect_equal(length(unique(idx)), length(idx),
                 info = sprintf("resolution %d indices are distinct", res))
    expect_true(all(nchar(idx) == 2L + res),
                info = sprintf("resolution %d index length", res))
  }
})

test_that("Z7: the two cells DGGRID's encoder merges keep distinct indices (#53)", {
  # DgZ7StringRF sends both of these to "0045310"; they are distinct cells in
  # distinct quads, so a bijective index has to separate them.
  for (res in c(5, 6)) {
    cells <- hexify_lonlat_to_cell(c(5, -34.9), c(45, 60.2), res, 7L)
    qij <- hexify_cell_to_quad_ij(cells, res, 7L)
    idx <- vapply(1:2, function(k) {
      hexify_cell_to_index(qij$quad[k], qij$i[k], qij$j[k], res, 7L, "z7")
    }, character(1))

    expect_false(cells[1] == cells[2])
    expect_false(idx[1] == idx[2])
  }
})

# grid_helpers.R
# Helper functions that accept HexGridInfo or HexData objects
#
# These functions wrap the low-level coordinate conversion functions
# to accept grid specifications, eliminating the need to repeat
# aperture/resolution parameters.

#' Normalize antimeridian-crossing polygon coordinates
#'
#' Shifts longitudes of a ring's coordinate matrix so an antimeridian-crossing
#' polygon becomes contiguous instead of spanning nearly the full -180 to 180
#' range. Downstream `sf::st_wrap_dateline()` then splits it correctly for
#' flat-map rendering.
#'
#' @param coords A 2+ column matrix whose first column is longitude
#' @return The same matrix, with the longitude column normalized if the ring
#'   crosses the antimeridian
#' @noRd
normalize_antimeridian_coords <- function(coords) {
  lons <- coords[, 1]
  lon_range <- max(lons, na.rm = TRUE) - min(lons, na.rm = TRUE)

  if (lon_range > 180) {
    # Polygon crosses antimeridian - normalize to be contiguous.
    # Shift negative lons to 0-360 range.
    lons[lons < 0] <- lons[lons < 0] + 360
    coords[, 1] <- lons

    # Now shift back to standard range, but keeping contiguity: if the
    # centroid ends up > 180, shift everything by -360.
    mean_lon <- mean(lons)
    if (mean_lon > 180) {
      coords[, 1] <- coords[, 1] - 360
    }
  }

  coords
}

#' Build hexagon polygons for ISEA cell IDs
#'
#' Antimeridian-crossing rings are normalized so each polygon is contiguous;
#' callers that render on a flat map pass the result through
#' `sf::st_wrap_dateline()` to split them at +/-180.
#'
#' @param cell_id Numeric vector of cell IDs
#' @param resolution Grid resolution level
#' @param aperture Grid aperture: 3, 4, 7, or a mixed sequence spelling
#' @param crs CRS the polygons carry, as sf reads it
#' @return An sfc of POLYGON geometries, one per cell ID, in input order
#' @noRd
isea_cells_to_sfc <- function(cell_id, resolution, aperture, crs = 4326) {
  corners_list <- if (is_mixed_aperture(aperture)) {
    mixed_cell_corners(cell_id, resolution, aperture)
  } else {
    cpp_cell_to_corners(
      as.numeric(cell_id),
      as.integer(resolution),
      as.integer(aperture)
    )
  }

  polygons <- lapply(corners_list, function(coords) {
    sf::st_polygon(list(normalize_antimeridian_coords(coords)))
  })

  # suppressWarnings: antimeridian normalization may temporarily produce
  # out-of-range longitudes that st_wrap_dateline corrects downstream
  suppressWarnings(sf::st_make_valid(sf::st_sfc(polygons, crs = crs)))
}

# =============================================================================
# COORDINATE CONVERSION HELPERS
# =============================================================================

#' Convert longitude/latitude to cell ID
#'
#' Converts geographic coordinates to DGGS cell IDs using a grid specification.
#'
#' @param lon Numeric vector of longitudes in degrees
#' @param lat Numeric vector of latitudes in degrees
#' @param grid A HexGridInfo or HexData object, or legacy hexify_grid
#'
#' @return Numeric vector of cell IDs
#'
#' @details
#' This function accepts either a HexGridInfo object from \code{hex_grid()} or
#' a HexData object from \code{hexify()}. If a HexData object is provided,
#' its grid specification is extracted automatically.
#'
#' @seealso \code{\link{cell_to_lonlat}} for the inverse operation,
#'   \code{\link{hex_grid}} for creating grid specifications
#'
#' @export
#' @examples
#' grid <- hex_grid(area_km2 = 1000)
#' cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)
#'
#' # Or use HexData object
#' df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
#' result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
#' cells <- lonlat_to_cell(lon = 5, lat = 48, grid = result)
lonlat_to_cell <- function(lon, lat, grid) {
  g <- extract_grid(grid)

  if (is_h3_grid(g)) {
    return(cpp_h3_latLngToCell(as.numeric(lon), as.numeric(lat), g@resolution))
  }

  if (is_mixed_aperture(g@aperture)) {
    cpp_lonlat_to_cell_seq(
      as.numeric(lon),
      as.numeric(lat),
      grid_ap_seq(g)
    )
  } else {
    cpp_lonlat_to_cell(
      as.numeric(lon),
      as.numeric(lat),
      g@resolution,
      as.integer(g@aperture)
    )
  }
}

#' Convert cell ID to longitude/latitude
#'
#' Converts DGGS cell IDs back to geographic coordinates (cell centers).
#'
#' @param cell_id Numeric vector of cell IDs
#' @param grid A HexGridInfo or HexData object
#'
#' @return Data frame with lon_deg and lat_deg columns
#'
#' @seealso \code{\link{lonlat_to_cell}} for the forward operation
#'
#' @export
#' @examples
#' grid <- hex_grid(area_km2 = 1000)
#' cells <- lonlat_to_cell(c(0, 10), c(45, 50), grid)
#' coords <- cell_to_lonlat(cells, grid)
cell_to_lonlat <- function(cell_id, grid) {
  g <- extract_grid(grid)

  if (is_h3_grid(g)) {
    result <- cpp_h3_cellToLatLng(as.character(cell_id))
    return(data.frame(lon_deg = result$lon, lat_deg = result$lat))
  }

  if (is_mixed_aperture(g@aperture)) {
    cpp_cell_to_lonlat_seq(
      as.numeric(cell_id),
      grid_ap_seq(g)
    )
  } else {
    cpp_cell_to_lonlat(
      as.numeric(cell_id),
      g@resolution,
      as.integer(g@aperture)
    )
  }
}

#' Convert cell IDs to sf polygons
#'
#' Creates sf polygon geometries for hexagonal grid cells.
#'
#' @param cell_id Numeric vector of cell IDs. If NULL and x is HexData,
#'   uses cells from x.
#' @param grid A HexGridInfo or HexData object. If HexData and cell_id is NULL,
#'   polygons are generated for all cells in the data.
#' @param wrap_dateline Logical. If TRUE (default), calls
#'   \code{sf::st_wrap_dateline()} to split antimeridian-crossing polygons.
#'   Set to FALSE for orthographic/globe projections where wrapping creates gaps.
#'
#' @return sf object with cell_id and geometry columns
#'
#' @details
#' When called with a HexData object and no cell_id argument, this function
#' generates polygons for all unique cells in the data, which is useful for
#' plotting.
#'
#' @seealso \code{\link{hex_grid}} for grid specifications,
#'   \code{\link{as_sf}} for converting HexData to sf
#'
#' @export
#' @examples
#' # From grid specification
#' grid <- hex_grid(area_km2 = 1000)
#' cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
#' polys <- cell_to_sf(cells, grid)
#'
#' # From HexData (all cells)
#' df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
#' result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
#' polys <- cell_to_sf(grid = result)
cell_to_sf <- function(cell_id = NULL, grid, wrap_dateline = TRUE) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required. Install with: install.packages('sf')")
  }

  # Handle HexData input
  if (is_hex_data(grid)) {
    if (is.null(cell_id)) {
      cell_id <- unique(grid@cell_id)
    }
    g <- grid@grid
  } else {
    g <- extract_grid(grid)
    if (is.null(cell_id)) {
      stop("cell_id required when grid is not HexData")
    }
  }

  # Remove NA and duplicates
  cell_id <- unique(cell_id[!is.na(cell_id)])
  if (length(cell_id) == 0) {
    stop("No valid cell_id values")
  }

  # H3 path: use native C backend for boundaries
  if (is_h3_grid(g)) {
    boundaries <- cpp_h3_cellToBoundary(as.character(cell_id))
    polygons <- lapply(boundaries, function(coords) {
      if (nrow(coords) == 0) return(sf::st_polygon())
      coords <- normalize_antimeridian_coords(coords)
      sf::st_polygon(list(coords))
    })
    sfc <- sf::st_sfc(polygons, crs = grid_crs(g))
    sfc <- suppressWarnings(sf::st_make_valid(sfc))
    result_sf <- sf::st_sf(cell_id = as.character(cell_id), geometry = sfc)
    if (wrap_dateline) {
      result_sf <- sf::st_wrap_dateline(result_sf,
        options = c("WRAPDATELINE=YES", "DATELINEOFFSET=180"), quiet = TRUE)
    }
    return(result_sf)
  }

  # ISEA path: generate polygons using C++ function. For globe/orthographic
  # projections, pass wrap_dateline = FALSE to keep cells intact.
  sfc <- isea_cells_to_sfc(cell_id, g@resolution, g@aperture,
                           crs = grid_crs(g))

  result_sf <- sf::st_sf(cell_id = cell_id, geometry = sfc)
  if (wrap_dateline) {
    result_sf <- sf::st_wrap_dateline(result_sf,
      options = c("WRAPDATELINE=YES", "DATELINEOFFSET=180"), quiet = TRUE)
  }
  result_sf
}

# =============================================================================
# GRID GENERATION HELPERS
# =============================================================================

#' Generate a rectangular grid of hexagons
#'
#' Creates hexagon polygons covering a rectangular geographic region.
#' For H3 grids, all cells that overlap the bounding box are included
#' (not just cells whose center falls inside), ensuring full spatial coverage.
#'
#' @param bbox Bounding box as c(xmin, ymin, xmax, ymax), or an sf/sfc object
#' @param grid A HexGridInfo object specifying the grid parameters
#'
#' @return sf object with hexagon polygons
#'
#' @seealso \code{\link{grid_global}} for global grids
#'
#' @export
#' @examples
#' grid <- hex_grid(area_km2 = 5000)
#' europe <- grid_rect(c(-10, 35, 30, 60), grid)
#' plot(europe)
grid_rect <- function(bbox, grid) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required")
  }

  g <- extract_grid(grid)

  # Handle sf/sfc bbox input
  if (inherits(bbox, c("sf", "sfc", "bbox"))) {
    bbox <- as.numeric(sf::st_bbox(bbox))
  }

  # H3 path: fill bbox with cells using native C backend
  if (is_h3_grid(g)) {
    bbox_coords <- matrix(c(
      bbox[1], bbox[2],
      bbox[3], bbox[2],
      bbox[3], bbox[4],
      bbox[1], bbox[4],
      bbox[1], bbox[2]
    ), ncol = 2, byrow = TRUE)
    cell_ids <- cpp_h3_polygonToCells(bbox_coords, g@resolution)
    if (length(cell_ids) == 0) {
      stop("No H3 cells found in the specified bounding box at resolution ", g@resolution)
    }
    return(cell_to_sf(cell_ids, g))
  }

  minlon <- bbox[1]
  minlat <- bbox[2]
  maxlon <- bbox[3]
  maxlat <- bbox[4]

  # Create sampling grid - use diagonal_km from grid if available
  diagonal <- if (!is.na(g@diagonal_km)) g@diagonal_km else sqrt(g@area_km2 * 2 / sqrt(3))
  spacing_deg <- diagonal / km_per_degree(grid_radius_km(g)) * 0.8

  lons <- seq(minlon, maxlon, by = spacing_deg)
  lats <- seq(minlat, maxlat, by = spacing_deg)
  grid_pts <- expand.grid(lon = lons, lat = lats)

  # Get unique cells covering the region
  cell_ids <- lonlat_to_cell(grid_pts$lon, grid_pts$lat, g)
  unique_cells <- unique(cell_ids)

  cell_to_sf(unique_cells, g)
}

#' Generate a global hexagon grid
#'
#' Creates hexagon polygons covering the entire Earth.
#'
#' @param grid A HexGridInfo object specifying the grid parameters
#' @param wrap_dateline Logical. If TRUE (default), antimeridian-crossing
#'   polygons are split at +/-180 degrees. Set to FALSE for orthographic/globe
#'   projections where wrapping creates gaps.
#'
#' @return sf object with hexagon polygons
#'
#' @details
#' This function generates a complete global grid by sampling points
#' densely across the globe. For large grids (many small cells),
#' consider using \code{grid_rect()} to generate regional subsets.
#'
#' @seealso \code{\link{grid_rect}} for regional grids
#'
#' @export
#' @examples
#' # Coarse global grid
#' grid <- hex_grid(area_km2 = 100000)
#' global <- grid_global(grid)
#' plot(global)
grid_global <- function(grid, wrap_dateline = TRUE) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required")
  }

  g <- extract_grid(grid)

  # H3 path: fill globe using native C backend
  if (is_h3_grid(g)) {
    h3_n_cells <- 2 + 120 * 7^g@resolution
    if (h3_n_cells > 2e6) {
      warning(sprintf(
        "H3 global grid at res %d has ~%.0f cells. This may take a while.",
        g@resolution, h3_n_cells
      ))
    }
    # Split globe into quadrants for polygonToCells
    quads <- list(
      matrix(c(-180, 0, 0, 0, 0, 90, -180, 90, -180, 0), ncol = 2, byrow = TRUE),
      matrix(c(0, 0, 180, 0, 180, 90, 0, 90, 0, 0), ncol = 2, byrow = TRUE),
      matrix(c(-180, -90, 0, -90, 0, 0, -180, 0, -180, -90), ncol = 2, byrow = TRUE),
      matrix(c(0, -90, 180, -90, 180, 0, 0, 0, 0, -90), ncol = 2, byrow = TRUE)
    )
    all_cells <- character(0)
    for (q in quads) {
      quad_cells <- cpp_h3_polygonToCells(q, g@resolution)
      all_cells <- c(all_cells, quad_cells)
    }
    cell_ids <- unique(all_cells)
    return(cell_to_sf(cell_ids, g, wrap_dateline = wrap_dateline))
  }

  # Estimate cell count for warning (ISEA)
  n_cells <- aperture_n_cells(g@aperture, g@resolution)
  if (n_cells > 100000) {
    warning(sprintf(
      "This will generate approximately %.0f cells. Consider larger area_km2.",
      n_cells
    ))
  }

  # Dense sampling - use diagonal_km from grid if available
  diagonal <- if (!is.na(g@diagonal_km)) g@diagonal_km else sqrt(g@area_km2 * 2 / sqrt(3))
  spacing_deg <- diagonal / km_per_degree(grid_radius_km(g)) * 0.7

  lons <- seq(-180, 180, by = spacing_deg)
  lats <- seq(-85, 85, by = spacing_deg)
  grid_pts <- expand.grid(lon = lons, lat = lats)

  # Add polar cap sampling (±85 to ±90 degrees)
  # The regular grid misses polar cells because lat stops at ±85
  # Near poles, longitude spacing must be DENSER not coarser - at 89°N,
  # the entire circumference is only ~6.3° of longitude-equivalent distance.
  # Use the same spacing_deg (or denser) to ensure we catch all cells.
  polar_lon_spacing <- min(spacing_deg, 15)  # At most 15°, or cell-based spacing
  polar_lons <- seq(-180, 180, by = polar_lon_spacing)
  polar_lats <- c(seq(85.5, 89.99, by = 0.5), seq(-89.99, -85.5, by = 0.5))
  polar_pts <- expand.grid(lon = polar_lons, lat = polar_lats)

  # Combine main grid with polar samples
  grid_pts <- rbind(grid_pts, polar_pts)

  cell_ids <- lonlat_to_cell(grid_pts$lon, grid_pts$lat, g)
  unique_cells <- unique(cell_ids)

  cell_to_sf(unique_cells, g, wrap_dateline = wrap_dateline)
}

#' Clip hexagon grid to polygon boundary
#'
#' Creates hexagon polygons clipped to a given polygon boundary. This is useful
#' for generating grids that conform to country borders, study areas, or other
#' irregular boundaries.
#'
#' @param boundary An sf/sfc polygon to clip to. Can be a country boundary,
#'   study area, or any polygon geometry.
#' @param grid A HexGridInfo object specifying the grid parameters
#' @param crop If TRUE (default), cells are cropped to the boundary. If FALSE,
#'   only cells whose centroids fall within the boundary are kept (no cropping).
#'
#' @return sf object with hexagon polygons clipped to the boundary
#'
#' @details
#' The function first generates cells covering the boundary polygon, then
#' clips or filters them. For H3 grids, all cells that overlap the boundary
#' are included (not just cells whose center falls inside), ensuring full
#' spatial coverage with no gaps along the boundary edge.
#'
#' When \code{crop = TRUE}, hexagons are geometrically intersected with the
#' boundary, which may produce partial hexagons at the edges. When
#' \code{crop = FALSE}, only complete hexagons whose centroids fall within
#' the boundary are returned.
#'
#' @seealso \code{\link{grid_rect}} for rectangular grids,
#'   \code{\link{grid_global}} for global grids
#'
#' @export
#' @examples
#' \donttest{
#' # Get France boundary from built-in world map
#' france <- hexify_world[hexify_world$name == "France", ]
#'
#' # Create grid clipped to France
#' grid <- hex_grid(area_km2 = 2000)
#' france_grid <- grid_clip(france, grid)
#'
#' # Plot result
#' library(ggplot2)
#' ggplot() +
#'   geom_sf(data = france, fill = "gray95") +
#'   geom_sf(data = france_grid, fill = alpha("steelblue", 0.3),
#'           color = "steelblue") +
#'   theme_minimal()
#'
#' # Keep only complete hexagons (no cropping)
#' france_grid_complete <- grid_clip(france, grid, crop = FALSE)
#' }
grid_clip <- function(boundary, grid, crop = TRUE) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required")
  }

  # Validate boundary
  if (!inherits(boundary, c("sf", "sfc"))) {
    stop("boundary must be an sf or sfc object")
  }

  g <- extract_grid(grid)

  # H3 path: fill boundary polygon using native C backend
  if (is_h3_grid(g)) {
    # Disable S2 for spatial operations
    s2_state <- sf::sf_use_s2()
    sf::sf_use_s2(FALSE)
    on.exit(sf::sf_use_s2(s2_state), add = TRUE)

    boundary_geom <- sf::st_geometry(boundary)
    if (length(boundary_geom) > 1) {
      boundary_geom <- sf::st_union(boundary_geom)
    }
    boundary_geom <- sf::st_make_valid(boundary_geom)

    # Extract polygon rings for cpp_h3_polygonToCells
    polys <- sf::st_cast(boundary_geom, "POLYGON")
    all_cells <- character(0)
    for (p in polys) {
      rings <- unclass(p)
      outer_ring <- rings[[1]]
      hole_rings <- if (length(rings) > 1) rings[-1] else NULL
      pcells <- cpp_h3_polygonToCells(outer_ring, g@resolution, holes = hole_rings)
      all_cells <- c(all_cells, pcells)
    }
    cell_ids <- unique(all_cells)
    hex_sf <- cell_to_sf(cell_ids, g)

    if (crop) {
      hex_sf <- sf::st_make_valid(hex_sf)
      result <- tryCatch({
        suppressWarnings(sf::st_intersection(hex_sf, boundary_geom))
      }, error = function(e) {
        boundary_buf <- sf::st_buffer(boundary_geom, 0)
        hex_buf <- sf::st_buffer(hex_sf, 0)
        suppressWarnings(sf::st_intersection(hex_buf, boundary_buf))
      })
      geom_types <- sf::st_geometry_type(result)
      result <- result[geom_types %in% c("POLYGON", "MULTIPOLYGON"), ]
      return(result)
    }
    return(hex_sf)
  }

  # Get bounding box of boundary
  bbox <- sf::st_bbox(boundary)

  # Disable S2 for all spatial operations (spherical geometry can cause issues)
  s2_state <- sf::sf_use_s2()
  sf::sf_use_s2(FALSE)
  on.exit(sf::sf_use_s2(s2_state), add = TRUE)

  # Generate rectangular grid covering the boundary
  rect_grid <- grid_rect(bbox, g)
  rect_grid <- sf::st_make_valid(rect_grid)

  # Get boundary geometry and ensure it's valid
  boundary_geom <- sf::st_geometry(boundary)
  boundary_geom <- sf::st_make_valid(boundary_geom)
  if (inherits(boundary, "sf") || length(boundary_geom) > 1) {
    boundary_geom <- sf::st_union(boundary_geom)
    boundary_geom <- sf::st_make_valid(boundary_geom)
  }

  if (crop) {
    # Crop hexagons to boundary
    result <- tryCatch({
      suppressWarnings(sf::st_intersection(rect_grid, boundary_geom))
    }, error = function(e) {
      # If intersection fails, try with buffered geometries
      boundary_buf <- sf::st_buffer(boundary_geom, 0)
      rect_buf <- sf::st_buffer(rect_grid, 0)
      suppressWarnings(sf::st_intersection(rect_buf, boundary_buf))
    })
    # Keep only polygon geometries (filter out points/lines from edge cases)
    geom_types <- sf::st_geometry_type(result)
    result <- result[geom_types %in% c("POLYGON", "MULTIPOLYGON"), ]
  } else {
    # Filter to hexagons whose centroids fall within boundary
    centroids <- suppressWarnings(sf::st_centroid(rect_grid))
    within <- suppressWarnings(
      sf::st_within(centroids, boundary_geom, sparse = FALSE)
    )
    result <- rect_grid[apply(within, 1, any), ]
  }

  result
}

# =============================================================================
# CELL AREA COMPUTATION
# =============================================================================

#' Compute per-cell area in km²
#'
#' Returns the area of each cell in square kilometers. For ISEA grids, all
#' cells have the same area (equal-area property). For H3 grids, each cell
#' has a different geodesic area depending on its location.
#'
#' @param cell_id Cell IDs to compute area for. For ISEA grids, these are
#'   numeric; for H3 grids, character strings. When \code{grid} is a HexData
#'   object and \code{cell_id} is \code{NULL}, all cell IDs from the data are
#'   used.
#' @param grid A HexGridInfo or HexData object.
#'
#' @return Named numeric vector of areas in km², one per \code{cell_id}.
#'
#' @details
#' For ISEA grids the area is constant across all cells and is read directly
#' from the grid specification.
#'
#' For H3 grids the area varies by latitude. The vendored 'H3' library computes
#' each cell's spherical polygon area as a solid angle, which this function
#' reads on the grid's body, so a grid built with \code{radius_km} reports that
#' body's areas.
#'
#' @seealso \code{\link{hex_grid}} for grid specifications,
#'   \code{\link{h3_crosswalk}} for ISEA/H3 interoperability
#'
#' @export
#' @examples
#' # ISEA: constant area
#' grid <- hex_grid(area_km2 = 1000)
#' cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
#' cell_area(cells, grid)
#'
#' # H3: area varies by location
#' \donttest{
#' h3 <- hex_grid(resolution = 5, type = "h3")
#' h3_cells <- lonlat_to_cell(c(0, 0), c(0, 80), h3)
#' cell_area(h3_cells, h3)  # equator vs polar — different areas
#' }
cell_area <- function(cell_id = NULL, grid) {

  # Handle HexData input
  if (is_hex_data(grid)) {
    if (is.null(cell_id)) {
      cell_id <- grid@cell_id
    }
    g <- grid@grid
  } else {
    g <- extract_grid(grid)
    if (is.null(cell_id)) {
      stop("cell_id required when grid is not HexData")
    }
  }

  # ISEA: constant equal-area
  if (!is_h3_grid(g)) {
    areas <- rep(g@area_km2, length(cell_id))
    names(areas) <- as.character(cell_id)
    return(areas)
  }

  # H3: per-cell area via native C backend, read on the grid's body
  cell_id <- as.character(cell_id)
  areas <- scale_area_to_body(cpp_h3_cellAreaKm2(cell_id), grid_radius_km(g))
  names(areas) <- cell_id
  areas
}


# =============================================================================
# HIERARCHICAL INDEX HELPERS
# =============================================================================

#' Hierarchical index string of one cell on a pure-aperture grid
#'
#' The index encoders work on (quad, i, j), so a cell ID makes the round trip
#' through `cpp_cell_to_quad_ij()` first. `cell_to_index()`, `get_parent()` and
#' `get_children()` all enter the hierarchy this way.
#'
#' @param cell_id One cell ID
#' @param resolution Resolution the cell ID belongs to
#' @param aperture_int Integer aperture (3, 4 or 7)
#' @param index_type One of "z3", "z7", "zorder"
#' @return Index string
#' @noRd
isea_cell_to_index_one <- function(cell_id, resolution, aperture_int, index_type) {
  qij <- cpp_cell_to_quad_ij(cell_id, resolution, aperture_int)
  cpp_cell_to_index(qij$quad, qij$i, qij$j, resolution, aperture_int, index_type)
}

#' Cell ID of one hierarchical index string on a pure-aperture grid
#'
#' Inverse of [isea_cell_to_index_one()]. The index carries its own resolution,
#' which is the one the returned cell ID belongs to.
#'
#' @param index One index string
#' @param aperture_int Integer aperture (3, 4 or 7)
#' @param index_type One of "z3", "z7", "zorder"
#' @return Cell ID
#' @noRd
isea_index_to_cell_one <- function(index, aperture_int, index_type) {
  cell <- cpp_index_to_cell(index, aperture_int, index_type)
  cpp_quad_ij_to_cell(cell$face, cell$i, cell$j, cell$resolution, aperture_int)
}

#' Convert cell ID to hierarchical index string
#'
#' Advanced function for working with hierarchical index strings.
#' Most users don't need this - use cell IDs directly.
#'
#' @param cell_id Numeric vector of cell IDs
#' @param grid A HexGridInfo or HexData object
#'
#' @return Character vector of hierarchical index strings
#'
#' @keywords internal
#' @export
cell_to_index <- function(cell_id, grid) {
  g <- extract_grid(grid)

  # H3 cell IDs are already hierarchical index strings
  if (is_h3_grid(g)) {
    return(as.character(cell_id))
  }

  # Mixed sequences use a geometric hierarchical index (see
  # R/aperture_mixed_hierarchy.R); pure apertures use the Z7/Z3/zorder encoders.
  if (is_mixed_aperture(g@aperture)) {
    return(vapply(as.numeric(cell_id),
                  function(id) mixed_cell_to_index_one(id, g@resolution, g@aperture),
                  character(1)))
  }

  # Determine index type based on aperture
  index_type <- index_type_for_aperture(g@aperture)
  aperture_int <- aperture_to_int(g@aperture)

  vapply(as.numeric(cell_id), isea_cell_to_index_one, character(1),
         resolution = g@resolution, aperture_int = aperture_int,
         index_type = index_type, USE.NAMES = FALSE)
}

#' Get parent cell
#'
#' Returns the parent cell at a coarser resolution.
#'
#' @param cell_id Numeric vector of cell IDs
#' @param grid A HexGridInfo or HexData object
#' @param levels Number of levels up (default 1)
#'
#' @return Numeric vector of parent cell IDs
#'
#' @keywords internal
#' @export
#' @examples
#' grid <- hex_grid(resolution = 10)
#' child_cells <- lonlat_to_cell(c(0, 10), c(45, 50), grid)
#' parent_cells <- get_parent(child_cells, grid)
get_parent <- function(cell_id, grid, levels = 1L) {
  g <- extract_grid(grid)

  if (g@resolution < levels) {
    stop("Cannot get parent: already at minimum resolution")
  }

  # H3 path
  if (is_h3_grid(g)) {
    parent_res <- g@resolution - as.integer(levels)
    if (parent_res < H3_MIN_RESOLUTION) {
      stop("Cannot get parent: would go below H3 minimum resolution")
    }
    return(cpp_h3_cellToParent(as.character(cell_id), parent_res))
  }

  # Mixed sequences: geometric parent (centre re-quantised at the coarser resolution).
  if (is_mixed_aperture(g@aperture)) {
    return(mixed_get_parent(as.numeric(cell_id), g@resolution, g@aperture,
                            as.integer(levels)))
  }

  index_type <- index_type_for_aperture(g@aperture)
  aperture_int <- aperture_to_int(g@aperture)
  levels <- as.integer(levels)

  # cpp_get_parent_index() strips one level off the index string, so `levels`
  # levels up is that many strips.
  vapply(as.numeric(cell_id), function(id) {
    idx <- isea_cell_to_index_one(id, g@resolution, aperture_int, index_type)
    for (step in seq_len(levels)) {
      idx <- cpp_get_parent_index(idx, aperture_int, index_type)
    }
    isea_index_to_cell_one(idx, aperture_int, index_type)
  }, numeric(1), USE.NAMES = FALSE)
}

#' Get children cells
#'
#' Returns the child cells at a finer resolution.
#'
#' @param cell_id Numeric vector of cell IDs
#' @param grid A HexGridInfo or HexData object
#' @param levels Number of levels down (default 1)
#'
#' @return List of numeric vectors containing child cell IDs
#'
#' @keywords internal
#' @export
get_children <- function(cell_id, grid, levels = 1L) {
  g <- extract_grid(grid)

  # H3 path
  if (is_h3_grid(g)) {
    child_res <- g@resolution + as.integer(levels)
    if (child_res > H3_MAX_RESOLUTION) {
      stop("Cannot get children: would exceed H3 maximum resolution (15)")
    }
    return(cpp_h3_cellToChildren(as.character(cell_id), child_res))
  }

  if (g@resolution + levels > MAX_RESOLUTION) {
    stop("Cannot get children: would exceed maximum resolution")
  }

  # Mixed sequences: geometric children (cells whose geometric parent is this cell).
  if (is_mixed_aperture(g@aperture)) {
    child_res <- g@resolution + as.integer(levels)
    ncc <- aperture_n_cells(g@aperture, child_res)
    return(lapply(as.numeric(cell_id), function(id)
      mixed_get_children_one(id, g@resolution, child_res, g@aperture, ncc)))
  }

  index_type <- index_type_for_aperture(g@aperture)
  aperture_int <- aperture_to_int(g@aperture)
  levels <- as.integer(levels)

  # cpp_get_children_indices() expands one level, so `levels` levels down is
  # that many expansions of the whole front.
  lapply(as.numeric(cell_id), function(id) {
    idx <- isea_cell_to_index_one(id, g@resolution, aperture_int, index_type)
    for (step in seq_len(levels)) {
      idx <- unlist(lapply(idx, cpp_get_children_indices,
                           aperture = aperture_int, index_type = index_type),
                    use.names = FALSE)
    }
    vapply(idx, isea_index_to_cell_one, numeric(1),
           aperture_int = aperture_int, index_type = index_type,
           USE.NAMES = FALSE)
  })
}

# tests/testthat/test-topology.R
# Tests for is_pentagon() and hex_distance()

test_that("is_pentagon detects 12 pentagons for H3", {
  skip_on_cran()

  g <- hex_grid(resolution = 1, type = "h3")
  cells_data <- grid_global(g)
  pent <- is_pentagon(cells_data$cell_id, g)
  expect_equal(sum(pent), 12L)
})

test_that("is_pentagon returns logical vector", {
  g <- hex_grid(resolution = 3, type = "h3")
  cells <- lonlat_to_cell(c(10, 20), c(50, 55), g)
  result <- is_pentagon(cells, g)
  expect_type(result, "logical")
  expect_length(result, 2)
})

test_that("hex_distance returns 0 for same cell (H3)", {
  g <- hex_grid(resolution = 5, type = "h3")
  cell <- lonlat_to_cell(10, 50, g)
  expect_equal(hex_distance(cell, cell, g), 0L)
})

test_that("hex_distance returns 1 for adjacent H3 cells", {
  g <- hex_grid(resolution = 5, type = "h3")
  cell <- lonlat_to_cell(10, 50, g)
  nbrs <- get_neighbors(cell, g)
  dist <- hex_distance(cell, nbrs[[1]][1], g)
  expect_equal(dist, 1L)
})

test_that("hex_distance is symmetric", {
  g <- hex_grid(resolution = 5, type = "h3")
  a <- lonlat_to_cell(10, 50, g)
  b <- lonlat_to_cell(10.5, 50.5, g)
  expect_equal(hex_distance(a, b, g), hex_distance(b, a, g))
})

test_that("hex_distance handles vectorized input", {
  g <- hex_grid(resolution = 5, type = "h3")
  a <- lonlat_to_cell(c(10, 20), c(50, 55), g)
  b <- lonlat_to_cell(c(10.1, 20.1), c(50.1, 55.1), g)
  result <- hex_distance(a, b, g)
  expect_length(result, 2)
  expect_type(result, "integer")
})

test_that("hex_distance works for same-quad ISEA cells", {
  g <- hex_grid(area_km2 = 1000)
  a <- lonlat_to_cell(10, 50, g)
  b <- lonlat_to_cell(10, 50, g)  # Same location = same cell
  expect_equal(hex_distance(a, b, g), 0L)
})

# tests/testthat/test-assign.R
# Tests for hexify_assign function

setup_icosa <- function() {
  cpp_build_icosa()
}

# =============================================================================
# BASIC FUNCTIONALITY
# =============================================================================

test_that("hexify_assign returns correct structure", {
  setup_icosa()

  lon <- c(0, 45, -120)
  lat <- c(0, 30, -45)

  result <- hexify_assign(lon, lat, effective_res = 3)

  expect_s3_class(result, "data.frame")
  expect_named(result, c(
    "id", "face", "effective_res", "center_lon", "center_lat"
  ))
  expect_equal(nrow(result), 3)
})

test_that("hexify_assign produces valid cell IDs", {
  setup_icosa()

  lon <- c(16.37, 2.35, -3.70)
  lat <- c(48.21, 48.86, 40.42)

  result <- hexify_assign(lon, lat, effective_res = 5)

  # Z3 index: 2-digit quad followed by one digit per resolution level
  expect_true(all(grepl("^\\d{2}[0-2]{5}$", result$id)))

  # Face is the quad (0-11)
  expect_true(all(result$face >= 0))
  expect_true(all(result$face <= 11))
})

test_that("hexify_assign agrees with the grid pipeline", {
  setup_icosa()

  lon <- c(16.37, 2.35, -3.70, 151.21, -58.38)
  lat <- c(48.21, 48.86, 40.42, -33.87, -34.60)

  for (res in 1:8) {
    grid <- hex_grid(resolution = res, aperture = 3)
    cells <- lonlat_to_cell(lon, lat, grid)
    centers <- cell_to_lonlat(cells, grid)

    result <- hexify_assign(lon, lat, effective_res = res)

    expect_equal(result$center_lon, centers$lon_deg)
    expect_equal(result$center_lat, centers$lat_deg)
    expect_equal(result$id, unname(cell_to_index(cells, grid)))
  }
})

test_that("hexify_assign centers are valid coordinates", {
  setup_icosa()

  lon <- c(0, 90, -90, 180, -180)
  lat <- c(0, 45, -45, 60, -60)

  result <- hexify_assign(lon, lat, effective_res = 4)

  # Centers should be valid lon/lat
  expect_true(all(result$center_lon >= -180 & result$center_lon <= 180))
  expect_true(all(result$center_lat >= -90 & result$center_lat <= 90))
})

# =============================================================================
# RESOLUTION HANDLING
# =============================================================================

test_that("hexify_assign handles different resolutions", {
  setup_icosa()

  lon <- c(10)
  lat <- c(50)

  for (res in c(1, 3, 5, 7)) {
    result <- hexify_assign(lon, lat, effective_res = res)

    expect_equal(result$effective_res, res)
    expect_equal(nrow(result), 1)
  }
})

test_that("hexify_assign resolution affects cell ID length", {
  setup_icosa()

  lon <- c(10)
  lat <- c(50)

  result_low <- hexify_assign(lon, lat, effective_res = 2)
  result_high <- hexify_assign(lon, lat, effective_res = 6)

  # Higher resolution should produce longer digit sequences
  # Extract digit part of ID
  digits_low <- sub(".*:Z3:", "", result_low$id)
  digits_high <- sub(".*:Z3:", "", result_high$id)

  expect_lt(nchar(digits_low), nchar(digits_high))
})

# =============================================================================
# CENTER ACCURACY
# =============================================================================

test_that("hexify_assign centers lie inside the assigned cell", {
  setup_icosa()

  # Uniform points on the sphere
  set.seed(23)
  n <- 400
  lon <- 360 * runif(n) - 180
  lat <- asin(2 * runif(n) - 1) * 180 / pi

  gc_km <- function(lon1, lat1, lon2, lat2) {
    r <- pi / 180
    EARTH_RADIUS_KM * acos(pmin(1, pmax(-1,
      sin(lat1 * r) * sin(lat2 * r) +
        cos(lat1 * r) * cos(lat2 * r) * cos((lon2 - lon1) * r))))
  }

  for (res in 1:8) {
    result <- hexify_assign(lon, lat, effective_res = res)
    d <- gc_km(lon, lat, result$center_lon, result$center_lat)

    # Center spacing of a grid of 10 * 3^res + 2 cells over the sphere; a point
    # lies at most one circumradius, about 0.6 of the spacing, from its cell
    # center.
    n_cells <- 10 * 3^res + 2
    spacing <- sqrt(2 * EARTH_SURFACE_KM2 / (sqrt(3) * n_cells))
    expect_lt(max(d), 0.7 * spacing)
  }
})

# =============================================================================
# POLYGON GENERATION
# =============================================================================

test_that("hexify_assign with make_polygons=TRUE returns sf", {
  skip_if_not_installed("sf")
  setup_icosa()


  lon <- c(10, 20)
  lat <- c(50, 55)

  result <- hexify_assign(lon, lat, effective_res = 3, make_polygons = TRUE)

  expect_s3_class(result, "sf")
  expect_true("geometry" %in% names(result))
  expect_equal(nrow(result), 2)
})

test_that("hexify_assign polygons have 6 corners", {
  skip_if_not_installed("sf")
  setup_icosa()

  lon <- c(10)
  lat <- c(50)

  result <- hexify_assign(lon, lat, effective_res = 3, make_polygons = TRUE)

  # Get coordinates of polygon
  coords <- sf::st_coordinates(result$geometry[[1]])

  # Polygon should have 7 points (6 corners + closing point)
  expect_equal(nrow(coords), 7)
})

# =============================================================================
# INPUT VALIDATION
# =============================================================================

test_that("hexify_assign validates input lengths", {
  setup_icosa()

  expect_error(
    hexify_assign(c(0, 1), c(0), effective_res = 3),
    "length"
  )
})

test_that("hexify_assign validates resolution", {
  setup_icosa()

  expect_error(
    hexify_assign(0, 0, effective_res = 0),
    "effective_res"
  )

  expect_error(
    hexify_assign(0, 0, effective_res = c(1, 2)),
    "effective_res"
  )
})

# =============================================================================
# EDGE CASES
# =============================================================================

test_that("hexify_assign handles poles", {
  setup_icosa()

  # North pole
  result_north <- hexify_assign(0, 90, effective_res = 3)
  expect_equal(nrow(result_north), 1)
  expect_true(is.finite(result_north$center_lat))

  # South pole
  result_south <- hexify_assign(0, -90, effective_res = 3)
  expect_equal(nrow(result_south), 1)
  expect_true(is.finite(result_south$center_lat))
})

test_that("hexify_assign handles date line", {
  setup_icosa()

  # Points near date line
  lon <- c(179, -179, 180, -180)
  lat <- c(0, 0, 45, -45)

  result <- hexify_assign(lon, lat, effective_res = 3)
  expect_equal(nrow(result), 4)
})

# tests/testthat/test-grid-helpers.R
# Tests for grid helper functions
# Note: grid_global, grid_rect, grid_clip tests removed for CRAN speed

test_that("lonlat_to_cell works with HexGridInfo", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  expect_type(cells, "double")
  expect_length(cells, 2)
  expect_true(all(cells > 0))
})

test_that("lonlat_to_cell works with HexData", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # Use the HexData object as grid source
  cells <- lonlat_to_cell(lon = 5, lat = 48, grid = result)

  expect_type(cells, "double")
  expect_length(cells, 1)
})

test_that("lonlat_to_cell works with mixed aperture", {
  grid <- hex_grid(area_km2 = 1000, aperture = "4/3")
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  expect_type(cells, "double")
  expect_length(cells, 2)
})

test_that("cell_to_lonlat returns cell centers", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)
  coords <- cell_to_lonlat(cells, grid)

  expect_type(coords, "list")
  expect_true("lon_deg" %in% names(coords))
  expect_true("lat_deg" %in% names(coords))
  expect_length(coords$lon_deg, 2)
})

test_that("cell_to_lonlat works with mixed aperture", {
  grid <- hex_grid(area_km2 = 1000, aperture = "4/3")
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)
  coords <- cell_to_lonlat(cells, grid)

  expect_type(coords, "list")
  expect_length(coords$lon_deg, 2)
})

test_that("lonlat_to_cell -> cell_to_lonlat round-trip lands in same cell", {
  grid <- hex_grid(area_km2 = 1000)

  original_lon <- c(0, 10, -5)
  original_lat <- c(45, 50, 48)

  cells1 <- lonlat_to_cell(original_lon, original_lat, grid)
  centers <- cell_to_lonlat(cells1, grid)
  cells2 <- lonlat_to_cell(centers$lon_deg, centers$lat_deg, grid)

  expect_equal(cells1, cells2)
})

test_that("cell_to_sf creates sf polygons", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 10000)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)
  polys <- cell_to_sf(cells, grid)

  expect_s3_class(polys, "sf")
  expect_true("cell_id" %in% names(polys))
  expect_equal(nrow(polys), length(unique(cells)))
})

test_that("cell_to_sf works with HexData (no cell_id)", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  polys <- cell_to_sf(grid = result)

  expect_s3_class(polys, "sf")
  expect_equal(nrow(polys), length(unique(result@cell_id)))
})

test_that("cell_to_sf errors without cell_id for HexGridInfo", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 10000)
  expect_error(cell_to_sf(grid = grid), "cell_id required")
})

test_that("cell_to_sf errors on empty cell_id", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 10000)
  expect_error(cell_to_sf(cell_id = numeric(0), grid = grid), "No valid")
  expect_error(cell_to_sf(cell_id = c(NA, NA), grid = grid), "No valid")
})

test_that("extract_grid works with HexGridInfo", {
  grid <- hex_grid(area_km2 = 1000)
  g <- hexify:::extract_grid(grid)

  expect_s4_class(g, "HexGridInfo")
})

test_that("extract_grid works with HexData", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
  g <- hexify:::extract_grid(result)

  expect_s4_class(g, "HexGridInfo")
})

test_that("extract_grid works with legacy hexify_grid", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  g <- hexify:::extract_grid(grid)

  expect_s4_class(g, "HexGridInfo")
})

test_that("extract_grid errors on invalid input", {
  expect_error(hexify:::extract_grid(list(a = 1)), "Cannot extract grid")
  expect_error(hexify:::extract_grid(data.frame()), "Cannot extract grid")
})

test_that("cell_to_sf returns valid geometries for all cells", {
  skip_if_not_installed("sf")

  # Test that polar cells (at icosahedral vertices) have valid geometries
  # Cell 1 is always quad 0 (north pole), and the last cell is quad 11 (south pole)
  grid <- hex_grid(area_km2 = 100000)
  n_cells <- 10 * (as.integer(grid@aperture)^grid@resolution) + 2

  # Test north and south pole cells
  polar_cells <- c(1, n_cells)
  polys <- cell_to_sf(polar_cells, grid)

  # All geometries must be valid
  validity <- sf::st_is_valid(polys)
  expect_true(all(validity))

  # These pentagon cells should have 5 vertices (6 coords with closing)
  vertex_counts <- sapply(sf::st_geometry(polys), function(g) nrow(sf::st_coordinates(g)))
  expect_true(all(vertex_counts == 6))
})

# =============================================================================
# ANTIMERIDIAN HANDLING
# =============================================================================

test_that("cell_to_sf splits ISEA cells at antimeridian", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 100000)
  # Get a cell near the antimeridian
  cell <- lonlat_to_cell(179.5, 0, grid)
  polys <- cell_to_sf(cell, grid)

  # All coordinates must be within [-180, 180]
  coords <- sf::st_coordinates(polys)
  expect_true(all(coords[, "X"] >= -180 & coords[, "X"] <= 180))

  # Geometry must be valid
  expect_true(all(sf::st_is_valid(polys)))
})

test_that("cell_to_sf splits H3 cells at antimeridian", {
  skip_if_not_installed("sf")

  h3 <- hex_grid(resolution = 2, type = "h3")
  cell <- lonlat_to_cell(179.5, 0, h3)
  polys <- cell_to_sf(cell, h3)

  # All coordinates must be within [-180, 180]
  coords <- sf::st_coordinates(polys)
  expect_true(all(coords[, "X"] >= -180 & coords[, "X"] <= 180))

  # Geometry must be valid
  expect_true(all(sf::st_is_valid(polys)))
})

test_that("cell_to_sf keeps non-crossing cells as POLYGON", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 10000)
  # Cell far from antimeridian
  cell <- lonlat_to_cell(10, 45, grid)
  polys <- cell_to_sf(cell, grid)

  geom_type <- as.character(sf::st_geometry_type(polys))
  expect_equal(geom_type, "POLYGON")
})

# tests/testthat/test-cell-geometry.R
# Geometry of a cell boundary: orientation, pentagons, and the two shapes the
# same boundary is returned in.

gc_distance <- function(a, b) {
  r <- pi / 180
  lon1 <- a[, 1] * r; lat1 <- a[, 2] * r
  lon2 <- b[, 1] * r; lat2 <- b[, 2] * r
  acos(pmin(1, pmax(-1, sin(lat1) * sin(lat2) +
                       cos(lat1) * cos(lat2) * cos(lon2 - lon1))))
}

lonlat_matrix <- function(df) {
  as.matrix(df[, c("lon_deg", "lat_deg")])
}

test_that("the corner list and the polygon table are the same boundary", {
  for (ap in c(3, 4, 7)) {
    cells <- as.numeric(c(1, 5, 40, 10 * ap^3 + 2))

    rings <- cpp_cell_to_corners(cells, 3L, ap)
    table <- cpp_cell_to_polygon(cells, 3L, ap)

    for (k in seq_along(cells)) {
      rows <- table[table$hex_id == cells[k], ]
      expect_equal(nrow(rows), nrow(rings[[k]]))
      expect_equal(rows$lon, unname(rings[[k]][, "lon"]))
      expect_equal(rows$lat, unname(rings[[k]][, "lat"]))
      expect_equal(rows$order, seq_len(nrow(rows)))
    }
  }
})

test_that("corners fall between the neighbouring cell centres", {
  # A corner is one circumradius from its own centre and the same from the two
  # neighbours sharing it. Turning the lattice by 30 degrees would aim the
  # corners at the neighbouring centres instead, which sit sqrt(3) - 1 of a
  # circumradius away.
  for (ap in c(3, 4, 7)) {
    for (res in 2:4) {
      g <- hex_grid(resolution = res, aperture = ap)
      cells <- round(seq(3, 10 * ap^res, length.out = 12))
      cells <- cells[!is_pentagon(cells, g)]

      centres <- lonlat_matrix(cell_to_lonlat(cells, g))
      rings <- cpp_cell_to_corners(as.numeric(cells), as.integer(res),
                                   as.integer(ap))

      ratios <- numeric(0)
      for (k in seq_along(cells)) {
        neighbours <- get_neighbors(cells[k], g)[[1]]
        neighbours <- neighbours[!is.na(neighbours)]
        nb_centres <- lonlat_matrix(cell_to_lonlat(neighbours, g))

        corners <- rings[[k]]
        corners <- corners[seq_len(nrow(corners) - 1), , drop = FALSE]

        for (v in seq_len(nrow(corners))) {
          one <- matrix(corners[v, ], nrow = 1)
          own <- gc_distance(one, matrix(centres[k, ], nrow = 1))
          near <- min(gc_distance(
            matrix(corners[v, ], nrow = nrow(nb_centres), ncol = 2, byrow = TRUE),
            nb_centres
          ))
          ratios <- c(ratios, near / own)
        }
      }

      expect_gt(min(ratios), 0.78)
    }
  }
})

test_that("a cell at an icosahedral vertex is a pentagon", {
  for (ap in c(3, 4, 7)) {
    for (res in 2:3) {
      g <- hex_grid(resolution = res, aperture = ap)
      n <- 10 * ap^res + 2
      pentagons <- seq_len(n)[is_pentagon(seq_len(n), g)]
      expect_length(pentagons, 12)

      rings <- cpp_cell_to_corners(as.numeric(pentagons), as.integer(res),
                                   as.integer(ap))
      for (ring in rings) {
        expect_equal(nrow(ring), 6L)
        expect_equal(ring[1, ], ring[6, ])
      }
    }
  }
})

test_that("every pentagon corner is shared with a neighbour", {
  # Five quads meet at an icosahedral vertex, so one of the six sectors around
  # the cell there is the icosahedron's angular deficit. Dropping the wrong
  # corner leaves one that no neighbour reaches.
  for (ap in c(3, 4, 7)) {
    for (res in 2:3) {
      g <- hex_grid(resolution = res, aperture = ap)
      n <- 10 * ap^res + 2
      pentagons <- seq_len(n)[is_pentagon(seq_len(n), g)]

      centres <- lonlat_matrix(cell_to_lonlat(pentagons, g))
      rings <- cpp_cell_to_corners(as.numeric(pentagons), as.integer(res),
                                   as.integer(ap))

      for (k in seq_along(pentagons)) {
        neighbours <- get_neighbors(pentagons[k], g)[[1]]
        neighbours <- neighbours[!is.na(neighbours)]
        expect_length(neighbours, 5L)

        nb_centres <- lonlat_matrix(cell_to_lonlat(neighbours, g))
        corners <- rings[[k]][1:5, , drop = FALSE]

        ratios <- vapply(1:5, function(v) {
          own <- gc_distance(matrix(corners[v, ], nrow = 1),
                             matrix(centres[k, ], nrow = 1))
          near <- min(gc_distance(
            matrix(corners[v, ], nrow = nrow(nb_centres), ncol = 2, byrow = TRUE),
            nb_centres
          ))
          near / own
        }, numeric(1))

        expect_lt(max(ratios) / min(ratios) - 1, 0.01)
      }
    }
  }
})

test_that("cell_to_sf follows the aperture of each level", {
  skip_if_not_installed("sf")

  lon <- c(0, 30, -100, 140, -20)
  lat <- c(45, -10, 60, -35, 5)

  for (ap in c("3", "4", "7", "4/3", "4/7", "7/4")) {
    g <- hex_grid(resolution = 4, aperture = ap)
    cells <- lonlat_to_cell(lon = lon, lat = lat, grid = g)
    centres <- cell_to_lonlat(cells, g)

    polygons <- cell_to_sf(cells, g, wrap_dateline = FALSE)
    points <- sf::st_as_sf(centres, coords = c("lon_deg", "lat_deg"), crs = 4326)

    hit <- suppressMessages(
      sf::st_intersects(points, sf::st_geometry(polygons), sparse = FALSE)
    )
    expect_true(all(diag(hit)))
  }
})

test_that("a mixed sequence and its first aperture give different polygons", {
  skip_if_not_installed("sf")

  cells <- 100:104

  mixed <- hex_grid(resolution = 4, aperture = "4/3")
  pure <- hex_grid(resolution = 4, aperture = 4)

  mixed_xy <- sf::st_coordinates(cell_to_sf(cells, mixed, wrap_dateline = FALSE))
  pure_xy <- sf::st_coordinates(cell_to_sf(cells, pure, wrap_dateline = FALSE))

  expect_false(isTRUE(all.equal(mixed_xy, pure_xy)))
})

# test-cell-area.R
# Tests for cell_area() and the H3 virtual column fix

# =============================================================================
# ISEA cell_area()
# =============================================================================

test_that("cell_area() returns constant area for ISEA grids", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(c(0, 10, 20), c(0, 45, 80), grid)
  areas <- cell_area(cells, grid)

  expect_true(is.numeric(areas))
  expect_equal(length(areas), 3)
  # All areas should be the same (equal-area property)
  expect_true(all(areas == areas[1]))
  # Area should match grid spec
  expect_equal(unname(areas[1]), grid@area_km2)
})

test_that("cell_area() returns named vector for ISEA", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(c(0, 10), c(45, 50), grid)
  areas <- cell_area(cells, grid)

  expect_true(!is.null(names(areas)))
  expect_equal(length(names(areas)), 2)
})

test_that("cell_area() works with HexData and NULL cell_id", {
  grid <- hex_grid(area_km2 = 1000)
  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
  areas <- cell_area(grid = result)

  expect_equal(length(areas), nrow(df))
})

test_that("cell_area() errors when cell_id is NULL and grid is HexGridInfo", {
  grid <- hex_grid(area_km2 = 1000)
  expect_error(cell_area(grid = grid), "cell_id required")
})

# =============================================================================
# H3 cell_area()
# =============================================================================

skip_if_not_installed("sf")

test_that("cell_area() returns per-cell areas for H3", {
  grid <- hex_grid(resolution = 5, type = "h3")
  # Equator vs high latitude
  cells <- lonlat_to_cell(c(0, 0), c(0, 80), grid)
  areas <- cell_area(cells, grid)

  expect_true(is.numeric(areas))
  expect_equal(length(areas), 2)
  # Areas should differ (H3 is NOT equal-area)
  expect_false(areas[1] == areas[2])
  # Both should be positive
  expect_true(all(areas > 0))
  # Exact geodesic areas (km^2) for these two cells, computed via the H3 C
  # library at resolution 5; both are within known bounds of the published
  # resolution-5 average of 252.904 km^2 (H3_AVG_AREA_KM2[6] in constants.R)
  expect_equal(unname(areas[1]), 187.9593512, tolerance = 1e-6)
  expect_equal(unname(areas[2]), 232.5434903, tolerance = 1e-6)
})

test_that("cell_area() caching works for H3", {
  grid <- hex_grid(resolution = 5, type = "h3")
  cells <- lonlat_to_cell(c(0), c(45), grid)

  # First call
  a1 <- cell_area(cells, grid)
  # Second call (should hit cache)
  a2 <- cell_area(cells, grid)

  expect_equal(a1, a2)
})

test_that("cell_area() deduplicates H3 cell IDs", {
  grid <- hex_grid(resolution = 5, type = "h3")
  # Same cell repeated
  cells <- lonlat_to_cell(c(0, 0), c(45, 45), grid)
  areas <- cell_area(cells, grid)

  # Should still return 2 values (one per input)
  expect_equal(length(areas), 2)
  expect_equal(unname(areas[1]), unname(areas[2]))
})

test_that("cell_area() with HexData for H3", {
  grid <- hex_grid(resolution = 8, type = "h3")
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
  areas <- cell_area(grid = result)

  expect_equal(length(areas), 2)
  expect_true(all(areas > 0))
})

# =============================================================================
# H3 virtual column fix
# =============================================================================

test_that("$cell_area_km2 returns per-cell areas for H3 HexData", {
  grid <- hex_grid(resolution = 5, type = "h3")
  df <- data.frame(lon = c(0, 0), lat = c(0, 80))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  areas <- result$cell_area_km2
  expect_equal(length(areas), 2)
  # Areas should differ for equator vs polar
  expect_false(areas[1] == areas[2])
})

test_that("[[cell_area_km2]] returns per-cell areas for H3 HexData", {
  grid <- hex_grid(resolution = 5, type = "h3")
  df <- data.frame(lon = c(0, 0), lat = c(0, 80))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  areas <- result[["cell_area_km2"]]
  expect_equal(length(areas), 2)
  expect_false(areas[1] == areas[2])
})

test_that("as.data.frame() returns per-cell areas for H3 HexData", {
  grid <- hex_grid(resolution = 5, type = "h3")
  df <- data.frame(lon = c(0, 0), lat = c(0, 80))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
  out <- as.data.frame(result)

  expect_true("cell_area_km2" %in% names(out))
  # Should not be a single repeated value
  expect_false(out$cell_area_km2[1] == out$cell_area_km2[2])
})

test_that("$cell_area_km2 still works for ISEA HexData", {
  grid <- hex_grid(area_km2 = 1000)
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  areas <- result$cell_area_km2
  expect_equal(length(areas), 2)
  # ISEA: all areas should be the same
  expect_equal(areas[1], areas[2])
  expect_equal(areas[1], grid@area_km2)
})

# tests/testthat/test-hierarchical-index.R
# Tests for hierarchical index functions

test_that("cell_to_index converts cell IDs to index strings", {
  grid <- hex_grid(area_km2 = 1000)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  indices <- cell_to_index(cells, grid)

  expect_type(indices, "character")
  expect_length(indices, 2)
  expect_true(all(nchar(indices) > 0))
})

test_that("cell_to_index works with aperture 4", {
  grid <- hex_grid(area_km2 = 1000, aperture = 4)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  indices <- cell_to_index(cells, grid)

  expect_type(indices, "character")
  expect_length(indices, 2)
})

test_that("cell_to_index works with aperture 7", {
  grid <- hex_grid(area_km2 = 10000, aperture = 7)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  indices <- cell_to_index(cells, grid)

  expect_type(indices, "character")
  expect_length(indices, 2)
})

test_that("get_parent returns parent cell for aperture 3", {
  grid <- hex_grid(area_km2 = 1000, aperture = 3)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  parents <- get_parent(cells, grid)

  expect_type(parents, "double")
  expect_length(parents, 2)
  # Parents should be different from children
  expect_false(all(parents == cells))
})

test_that("get_parent returns parent cell for aperture 4", {
  grid <- hex_grid(area_km2 = 1000, aperture = 4)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  parents <- get_parent(cells, grid)

  expect_type(parents, "double")
  expect_length(parents, 2)
})

test_that("get_parent returns parent cell for aperture 7", {
  grid <- hex_grid(area_km2 = 10000, aperture = 7)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  parents <- get_parent(cells, grid)

  expect_type(parents, "double")
  expect_length(parents, 2)
})

test_that("get_children returns child cells for aperture 3", {
  grid <- hex_grid(area_km2 = 100000, aperture = 3)
  cells <- lonlat_to_cell(lon = 0, lat = 45, grid = grid)

  children <- get_children(cells, grid)

  expect_type(children, "list")
  expect_length(children, 1)
  # Aperture 3 should produce 3 children
  expect_length(children[[1]], 3)
})

test_that("get_children returns child cells for aperture 4", {
  grid <- hex_grid(area_km2 = 100000, aperture = 4)
  cells <- lonlat_to_cell(lon = 0, lat = 45, grid = grid)

  children <- get_children(cells, grid)

  expect_type(children, "list")
  expect_length(children, 1)
  # Aperture 4 should produce 4 children
  expect_length(children[[1]], 4)
})

test_that("get_children returns child cells for aperture 7", {
  grid <- hex_grid(area_km2 = 100000, aperture = 7)
  cells <- lonlat_to_cell(lon = 0, lat = 45, grid = grid)

  children <- get_children(cells, grid)

  expect_type(children, "list")
  expect_length(children, 1)
  # Aperture 7 should produce 7 children
  expect_length(children[[1]], 7)
})

test_that("get_parent errors at minimum resolution", {
  grid <- hex_grid(resolution = 0, aperture = 3)
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

  expect_error(get_parent(cells, grid), "Cannot get parent")
})

test_that("get_children errors at maximum resolution", {
  # Resolution 30 is maximum
  expect_error(
    {
      grid <- hex_grid(resolution = 30, aperture = 3)
      cells <- lonlat_to_cell(lon = 0, lat = 45, grid = grid)
      get_children(cells, grid)
    },
    "Cannot get children"
  )
})

test_that("get_parent honours levels on pure ISEA grids", {
  for (ap in c(3, 4, 7)) {
    g <- hex_grid(resolution = 4, aperture = ap)
    coarse <- hex_grid(resolution = 3, aperture = ap)

    cells <- lonlat_to_cell(
      lon = c(0, 10, -60, 120, -140),
      lat = c(45, 50, -20, 5, -65),
      grid = g
    )

    expect_equal(
      get_parent(cells, g, levels = 2),
      get_parent(get_parent(cells, g), coarse)
    )
  }
})

test_that("get_children honours levels on pure ISEA grids", {
  for (ap in c(3, 4, 7)) {
    g <- hex_grid(resolution = 2, aperture = ap)
    mid <- hex_grid(resolution = 3, aperture = ap)

    cells <- lonlat_to_cell(lon = c(0, 120), lat = c(45, -20), grid = g)

    twice <- lapply(get_children(cells, g), function(kids) {
      sort(unique(unlist(get_children(kids, mid), use.names = FALSE)))
    })
    direct <- lapply(get_children(cells, g, levels = 2), sort)

    expect_equal(direct, twice)
    expect_length(direct[[1]], ap * ap)
  }
})

test_that("a cell is among the children of its two-level ancestor", {
  for (ap in c(3, 4, 7)) {
    g <- hex_grid(resolution = 3, aperture = ap)
    coarse <- hex_grid(resolution = 1, aperture = ap)

    cells <- lonlat_to_cell(
      lon = seq(-170, 170, by = 20),
      lat = rep(c(-55, -20, 15, 50), length.out = 18),
      grid = g
    )
    ancestors <- get_parent(cells, g, levels = 2)

    if (ap == 7) {
      # A pentagon has six aperture-7 children, so a Z7 index does not name
      # seven distinct descendants below one.
      keep <- !is_pentagon(ancestors, coarse)
      cells <- cells[keep]
      ancestors <- ancestors[keep]
    }

    expect_gt(length(cells), 0)

    kids <- get_children(ancestors, coarse, levels = 2)
    expect_true(all(mapply(function(cell, k) cell %in% k, cells, kids)))
  }
})

# tests/testthat/test-geometry-validity.R
# Tests for geometry validity at icosahedral vertices (pentagon cells)
#
# The ISEA3H grid has 12 icosahedral vertices where pentagons occur
# instead of hexagons. These tests verify that all geometries are valid
# and renderable.

test_that("grid_global produces all valid geometries", {
  skip_on_cran()
  skip_if_not_installed("sf")

  # Use a coarse grid for faster testing
  grid <- hex_grid(area_km2 = 100000)
  global <- grid_global(grid)

  # All geometries should be valid
  validity <- sf::st_is_valid(global)
  invalid_count <- sum(!validity)


  expect_true(all(validity),
    info = sprintf(
      "%d invalid geometries found out of %d total",
      invalid_count, nrow(global)
    )
  )
})

test_that("cell_to_sf produces valid geometries for polar pentagon cells", {
  skip_on_cran()
  skip_if_not_installed("sf")

  # Test at multiple apertures
  for (aperture in c(3L, 4L)) {
    grid <- hex_grid(area_km2 = 100000, aperture = aperture)

    # Get polar cells (north and south pole)
    n_cells <- 10 * (aperture^grid@resolution) + 2
    polar_cells <- c(1, n_cells)  # Quad 0 and quad 11

    polys <- cell_to_sf(polar_cells, grid)

    validity <- sf::st_is_valid(polys)
    expect_true(all(validity),
      info = sprintf("Aperture %d: polar pentagon cells have invalid geometry", aperture)
    )

    # Both should have 5 vertices (6 coords with closing)
    vertex_counts <- sapply(sf::st_geometry(polys), function(g) nrow(sf::st_coordinates(g)))
    expect_true(all(vertex_counts == 6),
      info = sprintf("Aperture %d: polar pentagon cells should have 6 coordinates", aperture)
    )
  }
})

test_that("grid_global geometries are non-empty", {
  skip_on_cran()
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 100000)
  global <- grid_global(grid)

  # No empty geometries
  empty <- sf::st_is_empty(global)
  expect_false(any(empty),
    info = sprintf("%d empty geometries found", sum(empty))
  )
})

test_that("pentagon cells have approximately 5 unique vertices", {
  skip_on_cran()
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 100000)
  global <- grid_global(grid)

  # Count unique vertices for each cell
  vertex_counts <- sapply(sf::st_geometry(global), function(g) {
    coords <- sf::st_coordinates(g)
    nrow(unique(coords[, 1:2]))
  })

  # Most cells should have 6 vertices (hexagons)
  # 12 cells should have ~5 vertices (pentagons at icosahedral vertices)
  n_pentagons <- sum(vertex_counts <= 5)
  n_hexagons <- sum(vertex_counts >= 6)

  expect_true(n_pentagons >= 1 && n_pentagons <= 20,
    info = sprintf("Expected 1-20 pentagons, found %d", n_pentagons)
  )
  expect_true(n_hexagons > n_pentagons,
    info = "Hexagons should outnumber pentagons"
  )
})

test_that("grid_rect produces all valid geometries", {
  skip_on_cran()
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 50000)

  # Test a region that doesn't include icosahedral vertices
  europe <- grid_rect(c(-10, 35, 30, 60), grid)

  validity <- sf::st_is_valid(europe)
  expect_true(all(validity),
    info = sprintf("%d invalid geometries in Europe grid", sum(!validity))
  )
})

# tests/testthat/test-edge-cases.R
# Edge case and property-based tests

# =============================================================================
# GEOGRAPHIC EDGE CASES
# =============================================================================

test_that("poles assign to valid cells for all apertures", {
  for (ap in c(3, 4, 7)) {
    res <- if (ap == 7) 3 else 5
    g <- hex_grid(resolution = res, aperture = ap)

    # North pole
    np <- lonlat_to_cell(0, 90, g)
    expect_true(np > 0, info = sprintf("north pole, ap %d", ap))

    # South pole
    sp <- lonlat_to_cell(0, -90, g)
    expect_true(sp > 0, info = sprintf("south pole, ap %d", ap))

    # Poles should be different cells
    expect_true(np != sp, info = sprintf("poles differ, ap %d", ap))
  }
})

test_that("antimeridian coordinates assign consistently", {
  g <- hex_grid(resolution = 5, aperture = 3)

  # Points near the antimeridian (180° / -180°)
  cell_pos <- lonlat_to_cell(179.99, 0, g)
  cell_neg <- lonlat_to_cell(-179.99, 0, g)

  # Should be neighboring cells or the same cell (very close points)
  expect_true(cell_pos > 0)
  expect_true(cell_neg > 0)
})

test_that("dateline-straddling points get valid cells", {
  g <- hex_grid(resolution = 5, aperture = 3)

  # Exactly at +180 and -180 should map to the same cell
  cell_a <- lonlat_to_cell(180, 45, g)
  cell_b <- lonlat_to_cell(-180, 45, g)
  expect_equal(cell_a, cell_b)
})

test_that("equator points assign to valid cells", {
  g <- hex_grid(resolution = 5, aperture = 3)

  lons <- seq(-180, 180, by = 30)
  cells <- lonlat_to_cell(lons, rep(0, length(lons)), g)
  expect_true(all(cells > 0))
  expect_length(cells, length(lons))
})

# =============================================================================
# ROUNDTRIP PROPERTY TESTS
# =============================================================================

test_that("encode-decode roundtrip converges for random points", {
  skip_on_cran()

  set.seed(42)
  n <- 50
  lons <- runif(n, -180, 180)
  lats <- runif(n, -90, 90)

  for (ap in c(3, 4)) {
    res <- 5
    g <- hex_grid(resolution = res, aperture = ap)

    cells <- lonlat_to_cell(lons, lats, g)
    coords <- cell_to_lonlat(cells, g)

    # Re-encode should give the same cell
    cells2 <- lonlat_to_cell(coords$lon_deg, coords$lat_deg, g)
    expect_equal(cells, cells2,
                 info = sprintf("roundtrip stability, ap %d", ap))
  }

  # AP7 has known precision issues at quad boundaries —

  # test that the vast majority of cells survive the roundtrip
  g7 <- hex_grid(resolution = 3, aperture = 7)
  cells7 <- lonlat_to_cell(lons, lats, g7)
  coords7 <- cell_to_lonlat(cells7, g7)
  cells7b <- lonlat_to_cell(coords7$lon_deg, coords7$lat_deg, g7)
  pct_match <- mean(cells7 == cells7b)
  expect_true(pct_match > 0.7,
              info = sprintf("ap7 roundtrip: %.0f%% match", pct_match * 100))
})

test_that("cell centers re-encode to the same cell", {
  skip_on_cran()

  g <- hex_grid(resolution = 5, aperture = 3)
  set.seed(123)
  lons <- runif(100, -180, 180)
  lats <- runif(100, -90, 90)

  cells <- lonlat_to_cell(lons, lats, g)
  centers <- cell_to_lonlat(cells, g)
  cells_back <- lonlat_to_cell(centers$lon_deg, centers$lat_deg, g)
  expect_equal(cells, cells_back)
})

# =============================================================================
# PENTAGON PROPERTIES
# =============================================================================

test_that("exactly 12 pentagons exist for all ISEA apertures", {
  # Deriving the 12 pentagon cell IDs via cpp_quad_ij_to_cell(quad=0:11, 0, 0)
  # only holds for apertures 3/4; aperture 7's substrate/surrogate coordinate
  # distinction makes that forward computation unreliable (see #32). Sweep
  # every valid cell ID instead and let is_pentagon() itself (which decodes
  # each ID's own quad/i/j) determine pentagon status -- this exercises the
  # full valid cell-ID range rather than a handful of specific candidates.
  max_cell_id <- function(res, ap) {
    err <- tryCatch({
      hexify:::cpp_cell_to_quad_ij(.Machine$integer.max, res, ap)
      NA_real_
    }, error = function(e) conditionMessage(e))
    as.numeric(sub(".*\\[1, ([0-9]+)\\].*", "\\1", err))
  }

  for (ap in c(3, 4, 7)) {
    res <- if (ap == 7) 3 else 5
    g <- hex_grid(resolution = res, aperture = ap)

    n <- max_cell_id(res, ap)
    pent_flags <- is_pentagon(seq_len(n), g)
    expect_equal(sum(pent_flags), 12L,
                 info = sprintf("12 pentagons, ap %d", ap))
  }
})

test_that("non-pentagon cells are not detected as pentagons", {
  g <- hex_grid(resolution = 5, aperture = 3)

  # A cell far from any icosahedron vertex
  cell <- lonlat_to_cell(10, 45, g)
  expect_false(is_pentagon(cell, g))
})

# =============================================================================
# NEIGHBOR PROPERTIES
# =============================================================================

test_that("interior cells have 6 neighbors", {
  # Use a location well inside a quad to avoid boundary effects
  g <- hex_grid(resolution = 5, aperture = 4)

  cell <- lonlat_to_cell(10, 45, g)
  nbrs <- get_neighbors(cell, g)

  expect_length(nbrs[[1]], 6L)
})

test_that("neighbors are symmetric for interior cells", {
  skip_on_cran()

  # Use aperture 4 and a location well inside a quad
  g <- hex_grid(resolution = 5, aperture = 4)
  cell <- lonlat_to_cell(10, 45, g)
  nbrs <- get_neighbors(cell, g)[[1]]

  for (nbr in nbrs) {
    nbr_nbrs <- get_neighbors(nbr, g)[[1]]
    expect_true(cell %in% nbr_nbrs,
                info = sprintf("cell %s not in neighbors of %s", cell, nbr))
  }
})

# =============================================================================
# GRID CELL COUNTS
# =============================================================================

test_that("resolution 0 produces 12 cells", {
  for (ap in c(3, 4, 7)) {
    g <- hex_grid(resolution = 0L, aperture = ap)
    cells <- grid_global(g)
    expect_equal(nrow(cells), 12L,
                 info = sprintf("res 0 cell count, ap %d", ap))
  }
})

# tests/testthat/test-neighbors.R
# Tests for get_neighbors() / k_ring

test_that("get_neighbors returns 6 neighbors for interior ISEA cell (ap4)", {
  g <- hex_grid(area_km2 = 1000)
  cell <- lonlat_to_cell(10, 50, g)
  expect_false(is_pentagon(cell, g))
  nbrs <- get_neighbors(cell, g)
  expect_type(nbrs, "list")
  expect_length(nbrs, 1)
  # Interior (non-pentagon) hex cells always have exactly 6 neighbors
  expect_equal(length(nbrs[[1]]), 6)
  # Self should not be in neighbors
  expect_false(cell %in% nbrs[[1]])
})

test_that("get_neighbors returns 6 neighbors for interior ISEA cell (ap3)", {
  g <- hex_grid(area_km2 = 1000, aperture = 3)
  cell <- lonlat_to_cell(10, 50, g)
  expect_false(is_pentagon(cell, g))
  nbrs <- get_neighbors(cell, g)
  expect_equal(length(nbrs[[1]]), 6)
})

test_that("get_neighbors returns 6 neighbors for interior ISEA cell (ap7)", {
  g <- hex_grid(area_km2 = 1000, aperture = 7)
  cell <- lonlat_to_cell(10, 50, g)
  expect_false(is_pentagon(cell, g))
  nbrs <- get_neighbors(cell, g)
  expect_equal(length(nbrs[[1]]), 6)
})

test_that("get_neighbors works for H3 cells", {
  g <- hex_grid(resolution = 5, type = "h3")
  cell <- lonlat_to_cell(10, 50, g)
  nbrs <- get_neighbors(cell, g)
  expect_type(nbrs, "list")
  expect_length(nbrs, 1)
  # H3 pentagons (12 per resolution) have 5 neighbors; this point isn't one
  expect_equal(length(nbrs[[1]]), 6)
})

test_that("k_ring with k=2 returns more cells than k=1", {
  g <- hex_grid(resolution = 5, type = "h3")
  cell <- lonlat_to_cell(10, 50, g)
  nbrs_k1 <- get_neighbors(cell, g, k = 1)
  nbrs_k2 <- get_neighbors(cell, g, k = 2)
  expect_true(length(nbrs_k2[[1]]) > length(nbrs_k1[[1]]))
})

test_that("include_self adds origin to result", {
  g <- hex_grid(resolution = 5, type = "h3")
  cell <- lonlat_to_cell(10, 50, g)
  nbrs <- get_neighbors(cell, g, include_self = TRUE)
  expect_true(cell %in% nbrs[[1]])
})

test_that("distances=TRUE returns data.frame", {
  g <- hex_grid(resolution = 5, type = "h3")
  cell <- lonlat_to_cell(10, 50, g)
  result <- get_neighbors(cell, g, k = 2, distances = TRUE)
  expect_true(is.data.frame(result[[1]]))
  expect_true("cell_id" %in% names(result[[1]]))
  expect_true("ring_distance" %in% names(result[[1]]))
  expect_true(all(result[[1]]$ring_distance %in% 1:2))
})

test_that("get_neighbors handles vectorized input", {
  g <- hex_grid(resolution = 5, type = "h3")
  cells <- lonlat_to_cell(c(10, 20), c(50, 55), g)
  nbrs <- get_neighbors(cells, g)
  expect_length(nbrs, 2)
})

test_that("ISEA neighbours stay inside the grid's cell ID range", {
  for (ap in c(3, 4, 7)) {
    for (res in 0:4) {
      n_cells <- 10 * ap^res + 2
      nbrs <- get_neighbors(seq_len(n_cells), hex_grid(resolution = res, aperture = ap))
      ids <- unlist(nbrs)

      expect_true(all(ids >= 1 & ids <= n_cells),
                  info = sprintf("aperture %d res %d neighbour ID range", ap, res))
    }
  }
})

test_that("ISEA grids hold 12 pentagons and hexagons everywhere else", {
  # The icosahedron's 12 vertices carry a pentagon each, so exactly 12 cells
  # have five neighbours and the rest have six.
  for (ap in c(3, 4, 7)) {
    for (res in 0:4) {
      n_cells <- 10 * ap^res + 2
      cells <- seq_len(n_cells)
      nbrs <- get_neighbors(cells, hex_grid(resolution = res, aperture = ap))
      label <- sprintf("aperture %d res %d", ap, res)

      expect_equal(sum(lengths(nbrs) == 5), 12L,
                   info = sprintf("%s pentagon count", label))
      expect_equal(sum(lengths(nbrs) == 6), n_cells - 12L,
                   info = sprintf("%s hexagon count", label))
      expect_true(all(vapply(nbrs, function(x) !anyDuplicated(x), TRUE)),
                  info = sprintf("%s duplicate neighbours", label))
      expect_true(all(mapply(function(c, x) !(c %in% x), cells, nbrs)),
                  info = sprintf("%s self in neighbours", label))
    }
  }
})

test_that("ISEA adjacency is symmetric", {
  for (ap in c(3, 4, 7)) {
    for (res in 0:4) {
      n_cells <- 10 * ap^res + 2
      cells <- seq_len(n_cells)
      nbrs <- get_neighbors(cells, hex_grid(resolution = res, aperture = ap))

      from <- rep(cells, lengths(nbrs))
      to <- unlist(nbrs)

      expect_setequal(paste(from, to), paste(to, from))
    }
  }
})

test_that("ISEA neighbours are the adjacent cells, not the ring beyond", {
  # Every neighbour centre sits about one cell spacing away; a coordinate that
  # stepped into the wrong quad would land near twice that.
  for (ap in c(3, 4, 7)) {
    for (res in c(2, 3)) {
      g <- hex_grid(resolution = res, aperture = ap)
      n_cells <- 10 * ap^res + 2
      cells <- seq_len(n_cells)
      nbrs <- get_neighbors(cells, g)

      ll <- cell_to_lonlat(cells, g)
      from <- rep(cells, lengths(nbrs))
      to <- unlist(nbrs)
      p <- pi / 180
      h <- sin((ll$lat[to] - ll$lat[from]) * p / 2)^2 +
        cos(ll$lat[from] * p) * cos(ll$lat[to] * p) *
        sin((ll$lon[to] - ll$lon[from]) * p / 2)^2
      km <- 6371 * 2 * asin(pmin(1, sqrt(h)))

      spacing <- sqrt(2 * (5.100656e8 / n_cells) / sqrt(3))
      expect_lt(max(km / spacing), 1.3)
    }
  }
})

# test-h3.R
# Tests for H3 grid support (native C backend)

skip_if_not_installed("sf")

# =============================================================================
# Grid creation
# =============================================================================

test_that("hex_grid() creates H3 grid by resolution", {
  grid <- hex_grid(resolution = 8, type = "h3")
  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@grid_type, "h3")
  expect_equal(grid@aperture, "7")
  expect_equal(grid@resolution, 8L)
  expect_true(!is.na(grid@area_km2))
  expect_true(grid@area_km2 > 0)
})

test_that("hex_grid() creates H3 grid by area_km2", {
  expect_warning(
    grid <- hex_grid(area_km2 = 1.0, type = "h3"),
    "not exactly equal-area"
  )
  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@grid_type, "h3")
  # Resolution 8 has area ~0.737 km^2, closest to 1.0
  expect_equal(grid@resolution, 8L)
})

test_that("hex_grid() validates H3 resolution range", {
  expect_error(hex_grid(resolution = 16, type = "h3"), "between 0 and 15")
  expect_error(hex_grid(resolution = -1, type = "h3"), "between 0 and 15")
})

test_that("is_h3_grid() works correctly", {
  h3_grid <- hex_grid(resolution = 5, type = "h3")
  isea_grid <- hex_grid(resolution = 5, aperture = 3)

  expect_true(is_h3_grid(h3_grid))
  expect_false(is_h3_grid(isea_grid))
})

# =============================================================================
# hexify() round-trip
# =============================================================================

test_that("hexify() works with H3 grid", {
  grid <- hex_grid(resolution = 8, type = "h3")
  df <- data.frame(lon = c(-73.9857, 2.3522, 139.6917),
                   lat = c(40.7484, 48.8566, 35.6895))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  expect_s4_class(result, "HexData")
  expect_true(is.character(result@cell_id))
  expect_equal(length(result@cell_id), 3)
  expect_true(all(nchar(result@cell_id) > 0))
})

test_that("hexify() H3 cell centers are near original points", {
  grid <- hex_grid(resolution = 8, type = "h3")
  df <- data.frame(lon = c(0), lat = c(45))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  # Cell center should be within ~1 km (res 8 cells are ~0.7 km^2)
  center_lon <- result@cell_center[1, "lon"]
  center_lat <- result@cell_center[1, "lat"]
  dist_deg <- sqrt((center_lon - 0)^2 + (center_lat - 45)^2)
  expect_true(dist_deg < 1)  # Less than 1 degree
})

# =============================================================================
# lonlat_to_cell / cell_to_lonlat round-trip
# =============================================================================

test_that("lonlat_to_cell() returns character for H3", {
  grid <- hex_grid(resolution = 5, type = "h3")
  cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)
  expect_true(is.character(cells))
  expect_equal(length(cells), 2)
})

test_that("cell_to_lonlat() round-trip for H3", {
  grid <- hex_grid(resolution = 5, type = "h3")
  cells <- lonlat_to_cell(lon = c(0), lat = c(45), grid = grid)
  coords <- cell_to_lonlat(cells, grid)

  expect_true(is.data.frame(coords))
  expect_true("lon_deg" %in% names(coords))
  expect_true("lat_deg" %in% names(coords))
  # Should be close to original
  expect_true(abs(coords$lon_deg[1] - 0) < 5)
  expect_true(abs(coords$lat_deg[1] - 45) < 5)
})

# =============================================================================
# cell_to_sf()
# =============================================================================

test_that("cell_to_sf() returns valid sf for H3", {
  grid <- hex_grid(resolution = 5, type = "h3")
  cells <- lonlat_to_cell(lon = c(0, 10, 20), lat = c(45, 50, 55), grid = grid)
  polys <- cell_to_sf(cells, grid)

  expect_s3_class(polys, "sf")
  expect_true(nrow(polys) > 0)
  expect_true("cell_id" %in% names(polys))
  expect_true(all(sf::st_is_valid(polys)))
})

test_that("cell_to_sf() works with HexData for H3", {
  grid <- hex_grid(resolution = 8, type = "h3")
  df <- data.frame(lon = c(0, 5, 10), lat = c(45, 46, 47))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
  polys <- cell_to_sf(grid = result)

  expect_s3_class(polys, "sf")
  expect_equal(nrow(polys), n_cells(result))
})

# =============================================================================
# Grid generation
# =============================================================================

test_that("grid_rect() works with H3", {
  grid <- hex_grid(resolution = 3, type = "h3")
  rect <- grid_rect(c(-10, 35, 10, 55), grid)

  expect_s3_class(rect, "sf")
  expect_true(nrow(rect) > 0)
  expect_true(all(sf::st_is_valid(rect)))
})

test_that("grid_clip() works with H3", {
  grid <- hex_grid(resolution = 2, type = "h3")
  # Simple test polygon
  poly <- sf::st_as_sfc("POLYGON((-10 40, 10 40, 10 55, -10 55, -10 40))",
                         crs = 4326)
  poly_sf <- sf::st_sf(geometry = poly)

  clipped <- grid_clip(poly_sf, grid)
  expect_s3_class(clipped, "sf")
  expect_true(nrow(clipped) > 0)
})

# =============================================================================
# Hierarchy
# =============================================================================

test_that("get_parent() works with H3", {
  grid <- hex_grid(resolution = 8, type = "h3")
  cells <- lonlat_to_cell(lon = c(0), lat = c(45), grid = grid)
  parent <- get_parent(cells, grid, levels = 1)

  expect_true(is.character(parent))
  expect_equal(length(parent), 1)
  # Parent should be a different (coarser) cell
  expect_false(parent == cells)
})

test_that("get_children() works with H3", {
  grid <- hex_grid(resolution = 5, type = "h3")
  cells <- lonlat_to_cell(lon = c(0), lat = c(45), grid = grid)
  children <- get_children(cells, grid, levels = 1)

  expect_true(is.list(children))
  expect_equal(length(children), 1)
  # H3 aperture 7: each cell has 7 children
  expect_equal(length(children[[1]]), 7)
})

test_that("cell_to_index() returns character for H3", {
  grid <- hex_grid(resolution = 5, type = "h3")
  cells <- lonlat_to_cell(lon = c(0), lat = c(45), grid = grid)
  idx <- cell_to_index(cells, grid)

  expect_true(is.character(idx))
  expect_equal(idx, cells)  # H3 cell IDs are already indices
})

# =============================================================================
# show() methods
# =============================================================================

test_that("show() for H3 HexGridInfo mentions H3", {
  grid <- hex_grid(resolution = 8, type = "h3")
  output <- capture.output(show(grid))
  expect_true(any(grepl("H3", output)))
  expect_true(any(grepl("not exactly equal-area", output, ignore.case = TRUE)))
})

test_that("show() for H3 HexData mentions H3", {
  grid <- hex_grid(resolution = 8, type = "h3")
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
  output <- capture.output(show(result))
  expect_true(any(grepl("H3", output)))
})

# =============================================================================
# as.data.frame() and accessors
# =============================================================================

test_that("as.data.frame() works with H3 HexData", {
  grid <- hex_grid(resolution = 8, type = "h3")
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
  out <- as.data.frame(result)

  expect_true("cell_id" %in% names(out))
  expect_true(is.character(out$cell_id))
  expect_equal(nrow(out), 2)
})

test_that("cells() and n_cells() work for H3", {
  grid <- hex_grid(resolution = 8, type = "h3")
  df <- data.frame(lon = c(0, 10, 0), lat = c(45, 50, 45))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  unique_cells <- cells(result)
  expect_true(is.character(unique_cells))
  expect_true(n_cells(result) <= 3)
  expect_true(n_cells(result) >= 2)
})

# =============================================================================
# Compare resolutions
# =============================================================================

test_that("hexify_compare_resolutions() works for H3", {
  result <- hexify_compare_resolutions(type = "h3", res_range = 0:5)
  expect_true(is.data.frame(result))
  expect_equal(nrow(result), 6)
  expect_true("cell_area_km2" %in% names(result))
  # Areas should decrease with resolution
  expect_true(all(diff(result$cell_area_km2) < 0))
})

# =============================================================================
# Plot methods (no-error tests)
# =============================================================================

test_that("plot() does not error for H3 HexData", {
  grid <- hex_grid(resolution = 5, type = "h3")
  df <- data.frame(lon = c(0, 5, 10), lat = c(45, 46, 47))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  # Should not error
  expect_silent({
    png(tempfile(fileext = ".png"))
    plot(result, basemap = FALSE)
    dev.off()
  })
})

test_that("hexify_heatmap() does not error for H3 HexData", {
  skip_if_not_installed("ggplot2")

  grid <- hex_grid(resolution = 5, type = "h3")
  df <- data.frame(lon = c(0, 5, 10), lat = c(45, 46, 47), count = c(1, 2, 3))
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  p <- hexify_heatmap(result, value = "count")
  expect_s3_class(p, "gg")
})

# =============================================================================
# dgearthstat() with HexGridInfo
# =============================================================================

test_that("dgearthstat() works with H3 HexGridInfo", {
  grid <- hex_grid(resolution = 8, type = "h3")
  stats <- dgearthstat(grid)

  expect_true(is.list(stats))
  expect_equal(stats$resolution, 8L)
  expect_equal(stats$grid_type, "h3")
  expect_true(stats$n_cells > 0)
  expect_true(stats$cell_area_km2 > 0)
})

# =============================================================================
# Legacy conversion guard
# =============================================================================

test_that("HexGridInfo_to_hexify_grid() errors for H3", {
  grid <- hex_grid(resolution = 8, type = "h3")
  expect_error(HexGridInfo_to_hexify_grid(grid), "cannot be converted")
})

# =============================================================================
# Backward compatibility
# =============================================================================

test_that("extract_grid() handles old objects without grid_type", {
  # Simulate old object by creating a grid and removing grid_type
  grid <- hex_grid(resolution = 5, aperture = 3)
  # In real deserialized objects, the slot would be missing
  # Here we just test the normal path works
  g <- extract_grid(grid)
  expect_false(is_h3_grid(g))
})

# hexify_projection.R
# ISEA Snyder Equal-Area Projection
#
# This module handles all projection operations between geographic coordinates
# (longitude/latitude) and the planar icosahedral faces.
#
# Note: The C++ layer handles lazy initialization of the icosahedron via ico().
# No R-level state management is needed.
#
# @name hexify-projection

# =============================================================================
# ICOSAHEDRON INITIALIZATION
# =============================================================================

#' Initialize icosahedron geometry
#'
#' Sets up the icosahedron state for ISEA projection. Uses standard ISEA3H
#' orientation by default (vertex 0 at 11.25E, 58.28N).
#'
#' @param vert0_lon Vertex 0 longitude in degrees (default ISEA_VERT0_LON_DEG)
#' @param vert0_lat Vertex 0 latitude in degrees (default ISEA_VERT0_LAT_DEG)
#' @param azimuth Azimuth rotation in degrees (default ISEA_AZIMUTH_DEG)
#'
#' @return Invisible NULL. Called for side effect.
#'
#' @details
#' The icosahedron is initialized lazily at the C++ level when first needed.
#' Manual call is only required for non-standard orientations.
#'
#' @family projection
#' @export
#' @examples
#' # Use standard ISEA3H orientation
#' hexify_build_icosa()
#'
#' # Custom orientation
#' hexify_build_icosa(vert0_lon = 0, vert0_lat = 90, azimuth = 0)
hexify_build_icosa <- function(vert0_lon = ISEA_VERT0_LON_DEG,
                                vert0_lat = ISEA_VERT0_LAT_DEG,
                                azimuth = ISEA_AZIMUTH_DEG) {
  cpp_build_icosa(
    as.numeric(vert0_lon),
    as.numeric(vert0_lat),
    as.numeric(azimuth)
  )
  invisible(NULL)
}

#' Get icosahedron face centers
#'
#' Returns the center coordinates of all 20 icosahedral faces.
#'
#' @return Data frame with 20 rows and columns lon, lat (degrees)
#'
#' @family projection
#' @export
#' @examples
#' centers <- hexify_face_centers()
#' plot(centers$lon, centers$lat)
hexify_face_centers <- function() {
  cpp_face_centers()
}

# =============================================================================
# FORWARD PROJECTION (lon/lat -> face coordinates)
# =============================================================================

#' Determine which face contains a point
#'
#' Returns the icosahedral face index (0-19) containing the given coordinates.
#'
#' @param lon Longitude in degrees
#' @param lat Latitude in degrees
#'
#' @return Integer face index (0-19)
#'
#' @family projection
#' @export
#' @examples
#' face <- hexify_which_face(16.37, 48.21)
hexify_which_face <- function(lon, lat) {
  cpp_which_face(as.numeric(lon), as.numeric(lat))
}

#' Forward Snyder projection
#'
#' Projects geographic coordinates onto the icosahedron, returning
#' face index and planar coordinates (tx, ty).
#'
#' @param lon Longitude in degrees
#' @param lat Latitude in degrees
#'
#' @return Named numeric vector: c(face, tx, ty)
#'
#' @details
#' tx and ty are normalized coordinates within the triangular face,
#' typically in range \[0, 1\].
#'
#' @family projection
#' @export
#' @examples
#' result <- hexify_forward(16.37, 48.21)
#' # result["face"], result["icosa_triangle_x"], result["icosa_triangle_y"]
hexify_forward <- function(lon, lat) {
  cpp_snyder_forward(as.numeric(lon), as.numeric(lat))
}

#' Forward projection to specific face
#'
#' Projects to a known face (skips face detection).
#'
#' @param face Face index (0-19)
#' @param lon Longitude in degrees
#' @param lat Latitude in degrees
#'
#' @return Named numeric vector: c(icosa_triangle_x, icosa_triangle_y)
#'
#' @family projection
#' @export
hexify_forward_to_face <- function(face, lon, lat) {
  cpp_project_to_icosa_triangle(as.integer(face), as.numeric(lon), as.numeric(lat))
}

# =============================================================================
# INVERSE PROJECTION (face coordinates -> lon/lat)
# =============================================================================

#' Inverse Snyder projection
#'
#' Converts face plane coordinates back to geographic coordinates.
#'
#' @param x X coordinate on face plane
#' @param y Y coordinate on face plane
#' @param face Face index (0-19)
#' @param tol Convergence tolerance (NULL for default)
#' @param max_iters Maximum iterations (NULL for default)
#'
#' @return Named numeric vector: c(lon_deg, lat_deg)
#'
#' @family projection
#' @export
#' @examples
#' coords <- hexify_inverse(0.5, 0.3, face = 2)
hexify_inverse <- function(x, y, face, tol = NULL, max_iters = NULL) {
  stopifnot(length(x) == 1L, length(y) == 1L, length(face) == 1L)
  cpp_face_xy_to_ll(as.numeric(x), as.numeric(y), as.integer(face), tol, max_iters)
}

# =============================================================================
# PRECISION CONTROL
# =============================================================================

#' Set inverse projection precision
#'
#' Controls the accuracy/speed tradeoff for inverse Snyder projection.
#'
#' @param mode Preset mode: "fast", "default", "high", or "ultra"
#' @param tol Custom tolerance (overrides mode if provided)
#' @param max_iters Custom max iterations (overrides mode if provided)
#'
#' @return Invisible NULL
#'
#' @family projection
#' @export
#' @examples
#' hexify_set_precision("high")
#' hexify_set_precision(tol = 1e-12, max_iters = 100)
hexify_set_precision <- function(mode = c("fast", "default", "high", "ultra"),
                                  tol = NULL, max_iters = NULL) {
  mode <- match.arg(mode)
  cpp_snyder_inv_set_precision(mode, tol, max_iters)
  invisible(NULL)
}

#' Get current precision settings
#'
#' @return List with tol and max_iters
#'
#' @family projection
#' @export
hexify_get_precision <- function() {
  cpp_snyder_inv_get_precision()
}

#' Set verbose mode for inverse projection
#'
#' When enabled, prints convergence information.
#'
#' @param verbose Logical
#'
#' @return Invisible NULL
#'
#' @family projection
#' @export
hexify_set_verbose <- function(verbose = TRUE) {
  cpp_snyder_inv_set_verbose(isTRUE(verbose))
  invisible(NULL)
}

#' Get inverse projection statistics
#'
#' Returns and optionally resets convergence statistics.
#'
#' @param reset Whether to reset statistics after retrieval (default TRUE)
#'
#' @return List with statistics (iterations, convergence info, etc.)
#'
#' @family projection
#' @export
hexify_projection_stats <- function(reset = TRUE) {
  cpp_snyder_inv_get_stats_and_reset()
}

# tests/testthat/test-projection-inverse.R
# Tests for Snyder ISEA inverse projection
#
# Functions tested:
# - hexify_inverse()
# - hexify_set_precision()
# - hexify_get_precision()
# - hexify_projection_stats()

# =============================================================================
# BASIC INVERSE PROJECTION
# =============================================================================

test_that("inverse projection returns valid lon/lat", {
  hexify_build_icosa()

  result <- hexify_inverse(0.5, 0.3, 0)

  expect_true("lon" %in% names(result))
  expect_true("lat" %in% names(result))
  expect_true(is.finite(result["lon"]))
  expect_true(is.finite(result["lat"]))
})

test_that("inverse projection returns coordinates in valid range", {
  skip_on_cran()
  hexify_build_icosa()

  for (face in 0:19) {
    result <- hexify_inverse(0.5, 0.3, face)

    expect_true(result["lon"] >= -180 && result["lon"] <= 180)
    expect_true(result["lat"] >= -90 && result["lat"] <= 90)
  }
})

# =============================================================================
# ROUND-TRIP CONSISTENCY
# =============================================================================

test_that("forward-inverse round-trip works near face centers", {
  skip_on_cran()
  hexify_build_icosa()
  centers <- hexify_face_centers()

  for (face in 0:19) {
    # Forward projection
    fwd <- hexify_forward_to_face(face, centers$lon[face + 1], centers$lat[face + 1])

    # Inverse projection
    inv <- hexify_inverse(fwd["icosa_triangle_x"], fwd["icosa_triangle_y"], face)

    expect_true(abs(inv["lat"] - centers$lat[face + 1]) < 1e-6,
                info = sprintf("Face %d lat mismatch", face))
  }
})

test_that("forward-inverse round-trip works for random points", {
  skip_on_cran()
  hexify_build_icosa()
  hexify_set_precision("high")

  set.seed(123)

  for (i in 1:50) {
    lon <- runif(1, -180, 180)
    lat <- runif(1, -85, 85)  # Avoid extreme poles

    fwd <- hexify_forward(lon, lat)
    face <- as.integer(fwd["face"])

    inv <- hexify_inverse(fwd["icosa_triangle_x"], fwd["icosa_triangle_y"], face)

    lon_diff <- abs(inv["lon"] - lon)
    # Handle longitude wrap-around
    if (lon_diff > 180) lon_diff <- 360 - lon_diff

    expect_true(lon_diff < 1e-5,
                info = sprintf("lon error %.8f at (%.2f, %.2f)", lon_diff, lon, lat))
    expect_true(abs(inv["lat"] - lat) < 1e-5,
                info = sprintf("lat error at (%.2f, %.2f)", lon, lat))
  }
})

# =============================================================================
# PRECISION SETTINGS
# =============================================================================

test_that("precision presets are accepted", {
  expect_no_error(hexify_set_precision("fast"))
  expect_no_error(hexify_set_precision("default"))
  expect_no_error(hexify_set_precision("high"))
  expect_no_error(hexify_set_precision("ultra"))
})

test_that("get_precision returns valid values", {
  hexify_set_precision("fast")
  p <- hexify_get_precision()

  expect_true("tol" %in% names(p))
  expect_true("max_iters" %in% names(p))
  expect_true(is.numeric(p["tol"]))
  expect_true(is.numeric(p["max_iters"]))
})

test_that("precision settings affect iteration count", {
  hexify_build_icosa()

  hexify_set_precision("fast")
  fast_precision <- hexify_get_precision()

  hexify_set_precision("ultra")
  ultra_precision <- hexify_get_precision()

  # Ultra should have tighter tolerance or more iterations
  expect_true(ultra_precision["tol"] <= fast_precision["tol"] ||
                ultra_precision["max_iters"] >= fast_precision["max_iters"])
})

# =============================================================================
# PROJECTION STATS
# =============================================================================

test_that("projection_stats returns valid structure", {
  hexify_build_icosa()

  # Perform some projections
  for (i in 1:10) {
    hexify_inverse(0.5, 0.3, 0)
  }

  stats <- hexify_projection_stats()

  expect_true("calls" %in% names(stats))
  expect_true("iters_total" %in% names(stats))
  expect_true("iters_max" %in% names(stats))
})

test_that("projection_stats tracks calls", {
  hexify_build_icosa()

  # Reset stats
  hexify_projection_stats()

  # Perform known number of calls
  n_calls <- 5
  for (i in 1:n_calls) {
    hexify_inverse(0.5, 0.3, 0)
  }

  stats <- hexify_projection_stats()
  expect_equal(as.integer(stats["calls"]), n_calls)
})

# =============================================================================
# HEXIFY_SET_VERBOSE
# =============================================================================

test_that("hexify_set_verbose accepts TRUE and FALSE", {
  expect_no_error(hexify_set_verbose(TRUE))
  expect_no_error(hexify_set_verbose(FALSE))
})

# =============================================================================
# CUSTOM PRECISION SETTINGS
# =============================================================================

test_that("hexify_set_precision accepts custom tol", {
  expect_no_error(hexify_set_precision(tol = 1e-10))
})

test_that("hexify_set_precision accepts custom max_iters", {
  expect_no_error(hexify_set_precision(max_iters = 50))
})

test_that("hexify_set_precision accepts both custom parameters", {
  expect_no_error(hexify_set_precision(tol = 1e-12, max_iters = 100))
})

# =============================================================================
# HEXIFY_BUILD_ICOSA
# =============================================================================

test_that("hexify_build_icosa with custom parameters", {
  # Custom vertex position
  expect_no_error(hexify_build_icosa(vert0_lon = 0, vert0_lat = 90, azimuth = 0))

  # Reset to standard orientation
  hexify_build_icosa()
})

# =============================================================================
# HEXIFY_FACE_CENTERS
# =============================================================================

test_that("hexify_face_centers returns 20 faces", {
  hexify_build_icosa()

  centers <- hexify_face_centers()

  expect_s3_class(centers, "data.frame")
  expect_equal(nrow(centers), 20)
  expect_true(all(c("lon", "lat") %in% names(centers)))
})

test_that("hexify_face_centers returns valid coordinates", {
  hexify_build_icosa()

  centers <- hexify_face_centers()

  expect_true(all(centers$lon >= -180 & centers$lon <= 180))
  expect_true(all(centers$lat >= -90 & centers$lat <= 90))
})

# =============================================================================
# HEXIFY_WHICH_FACE
# =============================================================================

test_that("hexify_which_face returns valid face indices", {
  skip_on_cran()
  hexify_build_icosa()

  set.seed(42)
  for (i in 1:50) {
    lon <- runif(1, -180, 180)
    lat <- runif(1, -89, 89)

    face <- hexify_which_face(lon, lat)
    expect_true(face >= 0 && face <= 19)
  }
})

test_that("hexify_which_face is consistent with hexify_forward", {
  skip_on_cran()
  hexify_build_icosa()

  set.seed(123)
  for (i in 1:30) {
    lon <- runif(1, -180, 180)
    lat <- runif(1, -85, 85)

    face <- hexify_which_face(lon, lat)
    forward_result <- hexify_forward(lon, lat)

    expect_equal(face, as.integer(forward_result["face"]))
  }
})

# =============================================================================
# HEXIFY_INVERSE WITH CUSTOM PARAMETERS
# =============================================================================

test_that("hexify_inverse with custom tol parameter", {
  hexify_build_icosa()

  result <- hexify_inverse(0.5, 0.3, face = 0, tol = 1e-10)

  expect_true(is.finite(result["lon"]))
  expect_true(is.finite(result["lat"]))
})

test_that("hexify_inverse with custom max_iters parameter", {
  hexify_build_icosa()

  result <- hexify_inverse(0.5, 0.3, face = 0, max_iters = 50)

  expect_true(is.finite(result["lon"]))
  expect_true(is.finite(result["lat"]))
})

test_that("hexify_inverse validates input lengths", {
  hexify_build_icosa()

  expect_error(hexify_inverse(c(0.5, 0.6), 0.3, face = 0))
  expect_error(hexify_inverse(0.5, c(0.3, 0.4), face = 0))
  expect_error(hexify_inverse(0.5, 0.3, face = c(0, 1)))
})

# =============================================================================
# INTERNAL CPP FUNCTION TESTS
# =============================================================================

test_that("cpp_icosa_face_params returns valid face parameters", {
  skip_on_cran()
  hexify_build_icosa()

  for (face in 0:19) {
    params <- cpp_icosa_face_params(face)

    expect_true("cen_lat" %in% names(params))
    expect_true("cen_lon" %in% names(params))
    expect_true("face_azimuth_offset" %in% names(params))

    expect_true(params["cen_lat"] >= -90 && params["cen_lat"] <= 90)
    expect_true(params["cen_lon"] >= -180 && params["cen_lon"] <= 180)
  }
})

test_that("cpp_icosa_face_params errors on invalid face", {
  hexify_build_icosa()

  expect_error(cpp_icosa_face_params(-1), "face must be 0..19")
  expect_error(cpp_icosa_face_params(20), "face must be 0..19")
})

test_that("cpp_hex_index_face_to_lonlat works with degrees=TRUE", {
  hexify_build_icosa()

  # Get face 0 parameters
  params <- cpp_icosa_face_params(0)

  result <- cpp_hex_index_face_to_lonlat(
    x = 0.5,
    y = 0.3,
    cen_lat = params["cen_lat"],
    cen_lon = params["cen_lon"],
    face_azimuth_offset = params["face_azimuth_offset"],
    degrees = TRUE
  )

  expect_length(result, 2)
  expect_true(result[1] >= -180 && result[1] <= 180)  # lon
  expect_true(result[2] >= -90 && result[2] <= 90)    # lat
})

test_that("cpp_hex_index_face_to_lonlat works with degrees=FALSE", {
  hexify_build_icosa()

  # Get face 0 parameters
  params <- cpp_icosa_face_params(0)

  result <- cpp_hex_index_face_to_lonlat(
    x = 0.5,
    y = 0.3,
    cen_lat = params["cen_lat"],
    cen_lon = params["cen_lon"],
    face_azimuth_offset = params["face_azimuth_offset"],
    degrees = FALSE
  )

  expect_length(result, 2)
  # Radians: lon in [-pi, pi], lat in [-pi/2, pi/2]
  expect_true(result[1] >= -pi && result[1] <= pi)
  expect_true(result[2] >= -pi / 2 && result[2] <= pi / 2)
})

test_that("cpp_hex_index_face_to_lonlat works with custom tolerance", {
  hexify_build_icosa()

  params <- cpp_icosa_face_params(5)

  result <- cpp_hex_index_face_to_lonlat(
    x = 0.5,
    y = 0.3,
    cen_lat = params["cen_lat"],
    cen_lon = params["cen_lon"],
    face_azimuth_offset = params["face_azimuth_offset"],
    degrees = TRUE,
    tol = 1e-10,
    max_iters = 50
  )

  expect_length(result, 2)
  expect_true(is.finite(result[1]))
  expect_true(is.finite(result[2]))
})

# tests/testthat/test-icosahedron.R
# Tests for icosahedron construction and face detection
#
# Functions tested:
# - hexify_build_icosa()
# - hexify_face_centers()
# - hexify_which_face()

# =============================================================================
# ICOSAHEDRON CONSTRUCTION
# =============================================================================

test_that("icosahedron builds successfully with default orientation", {
  expect_no_error(hexify_build_icosa())
})

test_that("icosahedron builds with custom orientation", {
  # ISEA3H default orientation
  expect_no_error(hexify_build_icosa(11.25, 58.282525588538995, 0))

  # Alternative orientations
  expect_no_error(hexify_build_icosa(0, 58.282525588538995, 0))
  expect_no_error(hexify_build_icosa(11.25, 58.282525588538995, 45))
})

# =============================================================================
# FACE CENTERS
# =============================================================================

test_that("face_centers returns 20 faces with valid coordinates", {
  hexify_build_icosa()

  centers <- hexify_face_centers()

  expect_equal(nrow(centers), 20)
  expect_true("lon" %in% names(centers))
  expect_true("lat" %in% names(centers))

  # All coordinates should be finite
  expect_true(all(is.finite(centers$lon)))
  expect_true(all(is.finite(centers$lat)))

  # hexify_face_centers() returns radians; valid range is [-pi, pi] / [-pi/2, pi/2]
  expect_true(all(centers$lon >= -pi & centers$lon <= pi))
  expect_true(all(centers$lat >= -pi / 2 & centers$lat <= pi / 2))
})

test_that("face_centers returns the exact standard ISEA orientation", {
  # The 20 icosahedron face centers are a fixed, closed-form geometric
  # structure for the default ISEA3H orientation (pole_lon=11.25,
  # pole_lat=58.282525588538995, azimuth=0) - not just "some finite value".
  hexify_build_icosa()
  centers <- hexify_face_centers()

  expected_lon <- c(
    -1.3744467860, -0.5890486226, 0.1963495408, 0.9817477043, 1.7671458677,
    -2.1598449437, -1.0095829578, 0.1963495479, 1.4022820395, 2.5525440367,
    -1.7393106141, -0.5890486168, 0.9817477099, 2.1320096958, -2.9452431057,
    -2.1598449424, -1.3744467859, 1.7671458676, 2.5525440380, -2.9452431041
  )
  expected_lat <- c(
    1.2059324925, 0.6154797087, 0.3648638281, 0.6154797087, 1.2059325048,
    0.6154797060, 0.0000000000, -0.3648638281, 0.0000000000, 0.6154797113,
    -0.0000000000, -0.6154797113, -0.6154797060, -0.0000000000, 0.3648638281,
    -0.6154797054, -1.2059324973, -1.2059325001, -0.6154797120, -0.3648638281
  )

  expect_equal(centers$lon, expected_lon, tolerance = 1e-8)
  expect_equal(centers$lat, expected_lat, tolerance = 1e-8)
})

test_that("face centers span the globe", {
  hexify_build_icosa()
  centers <- hexify_face_centers()

  # Should have faces in both hemispheres
  expect_true(any(centers$lat > 0))
  expect_true(any(centers$lat < 0))

  # Should have faces across longitude range
  # Note: coordinates may be in radians (range ~[-pi, pi]) or degrees
  lon_range <- max(centers$lon) - min(centers$lon)
  expect_true(lon_range > 2.0)  # Works for both radians (~5.5 rad) and degrees (>100°)
})

# =============================================================================
# FACE DETECTION (which_face)
# =============================================================================

test_that("which_face returns valid face numbers", {
  hexify_build_icosa()

  test_points <- data.frame(
    lon = c(0, 90, -90, 180, 0, 0),
    lat = c(0, 0, 0, 0, 45, -45)
  )

  for (i in seq_len(nrow(test_points))) {
    face <- hexify_which_face(test_points$lon[i], test_points$lat[i])
    expect_true(face >= 0 && face <= 19,
                info = sprintf("lon=%.2f, lat=%.2f",
                              test_points$lon[i], test_points$lat[i]))
  }
})

test_that("which_face is deterministic", {
  hexify_build_icosa()

  lon <- 16.37
  lat <- 48.21

  face1 <- hexify_which_face(lon, lat)
  face2 <- hexify_which_face(lon, lat)

  expect_equal(face1, face2)
})

test_that("which_face handles extreme coordinates", {
  hexify_build_icosa()

  # Near poles
  expect_true(hexify_which_face(0, 89.9999) %in% 0:19)
  expect_true(hexify_which_face(0, -89.9999) %in% 0:19)

  # Near antimeridian
  expect_true(hexify_which_face(179.9999, 0) %in% 0:19)
  expect_true(hexify_which_face(-179.9999, 0) %in% 0:19)
})

test_that("which_face matches reference data", {
  hexify_build_icosa()

  truth <- read.csv(testthat::test_path("data/truth_faces.csv"))
  expect_true(all(c("lon", "lat", "tnum") %in% names(truth)))

  mismatches <- 0L
  msgs <- character()

  for (i in seq_len(nrow(truth))) {
    got <- hexify_which_face(truth$lon[i], truth$lat[i])
    if (got != truth$tnum[i]) {
      mismatches <- mismatches + 1L
      if (mismatches <= 5L) {
        msgs <- c(msgs, sprintf(
          "row %d: got %d vs expected %d (lon=%.9f lat=%.9f)",
          i, got, truth$tnum[i], truth$lon[i], truth$lat[i]
        ))
      }
    }
  }

  if (mismatches > 0L) {
    message(paste(msgs, collapse = "\n"))
  }
  expect_equal(mismatches, 0L)
})

# =============================================================================
# FACE CENTER CONSISTENCY
# =============================================================================

test_that("face center detection returns the originating face", {
  # hexify_face_centers() returns radians; hexify_which_face() expects
  # degrees (see its @param docs) - convert before round-tripping.
  hexify_build_icosa()
  centers <- hexify_face_centers()

  for (face in 0:19) {
    lon_deg <- centers$lon[face + 1] * 180 / pi
    lat_deg <- centers$lat[face + 1] * 180 / pi
    detected <- hexify_which_face(lon_deg, lat_deg)
    expect_equal(detected, face,
                info = sprintf("Face center %d was detected as face %d", face, detected))
  }
})

# tests/testthat/test-projection-forward.R
# Tests for Snyder ISEA forward projection
#
# Functions tested:
# - hexify_forward()
# - hexify_forward_to_face()

# =============================================================================
# BASIC FORWARD PROJECTION
# =============================================================================

test_that("forward projection returns valid structure", {
  hexify_build_icosa()

  result <- hexify_forward(16.37, 48.21)

  expect_true("face" %in% names(result))
  expect_true("icosa_triangle_x" %in% names(result))
  expect_true("icosa_triangle_y" %in% names(result))
})

test_that("forward projection returns finite values", {
  skip_on_cran()
  hexify_build_icosa()

  set.seed(123)

  n <- 100
  lon <- runif(n, -180, 180)
  lat <- runif(n, -89.99, 89.99)

  for (i in seq_len(n)) {
    result <- hexify_forward(lon[i], lat[i])

    expect_true(is.finite(result["face"]))
    expect_true(is.finite(result["icosa_triangle_x"]))
    expect_true(is.finite(result["icosa_triangle_y"]))
  }
})

test_that("forward projection face values are in valid range", {
  skip_on_cran()
  hexify_build_icosa()

  set.seed(456)
  n <- 100

  for (i in 1:n) {
    lon <- runif(1, -180, 180)
    lat <- runif(1, -89.99, 89.99)

    result <- hexify_forward(lon, lat)
    expect_true(result["face"] >= 0 && result["face"] <= 19)
  }
})

test_that("forward projection is deterministic", {
  hexify_build_icosa()

  lon <- 16.37
  lat <- 48.21

  result1 <- hexify_forward(lon, lat)
  result2 <- hexify_forward(lon, lat)

  expect_identical(as.numeric(result1["face"]), as.numeric(result2["face"]))
  expect_identical(result1["icosa_triangle_x"], result2["icosa_triangle_x"])
  expect_identical(result1["icosa_triangle_y"], result2["icosa_triangle_y"])
})

# =============================================================================
# FORWARD TO FACE
# =============================================================================

test_that("forward_to_face matches forward for detected face", {
  skip_on_cran()
  hexify_build_icosa()

  set.seed(789)
  n <- 100

  for (i in 1:n) {
    lon <- runif(1, -180, 180)
    lat <- runif(1, -89.99, 89.99)

    full <- hexify_forward(lon, lat)
    face <- as.integer(full["face"])

    fixed <- hexify_forward_to_face(face, lon, lat)

    expect_equal(as.numeric(full["icosa_triangle_x"]),
                 as.numeric(fixed["icosa_triangle_x"]),
                 tolerance = 1e-15)
    expect_equal(as.numeric(full["icosa_triangle_y"]),
                 as.numeric(fixed["icosa_triangle_y"]),
                 tolerance = 1e-15)
  }
})

test_that("forward_to_face returns values in [0, 1] range", {
  skip_on_cran()
  hexify_build_icosa()

  set.seed(111)
  slack <- 1e-10

  for (i in 1:100) {
    lon <- runif(1, -180, 180)
    lat <- runif(1, -89.99, 89.99)

    result <- hexify_forward(lon, lat)
    tx <- result["icosa_triangle_x"]
    ty <- result["icosa_triangle_y"]

    expect_true(tx >= -slack && tx <= 1 + slack,
                info = sprintf("tx=%.10f out of range", tx))
    expect_true(ty >= -slack && ty <= 1 + slack,
                info = sprintf("ty=%.10f out of range", ty))
  }
})

# =============================================================================
# EDGE CASES
# =============================================================================

test_that("forward projection handles near-pole coordinates", {
  skip_on_cran()
  hexify_build_icosa()

  eps <- 1e-9
  lons <- c(-180, -90, 0, 90, 180)

  for (L in lons) {
    north <- hexify_forward(L, 90 - eps)
    south <- hexify_forward(L, -90 + eps)

    expect_true(is.finite(north["icosa_triangle_x"]))
    expect_true(is.finite(north["icosa_triangle_y"]))
    expect_true(is.finite(south["icosa_triangle_x"]))
    expect_true(is.finite(south["icosa_triangle_y"]))
  }
})

test_that("forward projection handles antimeridian", {
  hexify_build_icosa()

  # Points near +180/-180
  eps <- 1e-7

  east <- hexify_forward(180 - eps, 0)
  west <- hexify_forward(-180 + eps, 0)

  expect_true(is.finite(east["icosa_triangle_x"]))
  expect_true(is.finite(west["icosa_triangle_x"]))
})

# =============================================================================
# REFERENCE DATA VALIDATION
# =============================================================================

test_that("forward projection matches reference data", {
  skip_on_cran()
  hexify_build_icosa(11.25, 58.282525588538995, 0)
  hexify_set_precision("ultra")

  truth <- read.csv(testthat::test_path("data/truth_txty.csv"),
                    stringsAsFactors = FALSE)

  expect_true(all(c("lon", "lat", "tnum", "tx", "ty") %in% names(truth)))

  atol <- 2e-8
  mismatches <- 0L
  msgs <- character()

  for (i in seq_len(nrow(truth))) {
    result <- hexify_forward(truth$lon[i], truth$lat[i])
    face <- as.integer(result["face"])
    tx <- as.numeric(result["icosa_triangle_x"])
    ty <- as.numeric(result["icosa_triangle_y"])

    ok <- (face == as.integer(truth$tnum[i])) &&
      (abs(tx - truth$tx[i]) <= atol) &&
      (abs(ty - truth$ty[i]) <= atol)

    if (!ok) {
      mismatches <- mismatches + 1L
      if (mismatches <= 5L) {
        msgs <- c(msgs, sprintf(
          "lon=%.10f lat=%.10f: face %d vs %d, tx %.12g vs %.12g, ty %.12g vs %.12g",
          truth$lon[i], truth$lat[i],
          face, truth$tnum[i], tx, truth$tx[i], ty, truth$ty[i]
        ))
      }
    }
  }

  if (mismatches > 0L) {
    message(paste(msgs, collapse = "\n"))
  }
  expect_equal(mismatches, 0L)
})

# R/hex_browse.R
# Interactive leaflet-based hex maps

#' Interactive Hex Map
#'
#' Opens an interactive leaflet map with hexagonal cell polygons.
#' Cells can be colored by a data column, with popups showing cell
#' information on click.
#'
#' @param hex_data A HexData object, or a data.frame with a `cell_id` column.
#' @param grid A HexGridInfo object. Required if `hex_data` is a data.frame.
#' @param value Optional column name (character) to color cells by.
#'   If `NULL`, cells are colored uniformly.
#' @param palette Color palette name for continuous values. Default `"viridis"`.
#' @param opacity Fill opacity (0-1). Default 0.7.
#'
#' @return A `leaflet` map object (can be printed or embedded in Shiny).
#'
#' @details
#' Requires the `leaflet` package (in Suggests). The map is built using
#' [as_sf()] to generate polygon geometries, then rendered as a leaflet
#' choropleth.
#'
#' @seealso [hexify_heatmap()] for static ggplot2 maps,
#'   [as_sf()] for sf conversion
#'
#' @export
#' @examples
#' \donttest{
#' if (requireNamespace("leaflet", quietly = TRUE)) {
#'   df <- data.frame(
#'     lon = runif(50, -5, 5),
#'     lat = runif(50, 45, 55),
#'     value = rnorm(50)
#'   )
#'   hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 500)
#'   hex_browse(hd, value = "value")
#' }
#' }
hex_browse <- function(hex_data, grid = NULL, value = NULL,
                        palette = "viridis", opacity = 0.7) {
  if (!requireNamespace("leaflet", quietly = TRUE)) {
    stop("Package 'leaflet' is required for hex_browse(). ",
         "Install with: install.packages('leaflet')")
  }

  # Get grid and cell data

  if (is_hex_data(hex_data)) {
    g <- hex_data@grid

    # Aggregate to unique cells
    df <- as.data.frame(hex_data)
    cell_ids <- unique(df$cell_id)

    # Build summary per cell
    cell_df <- data.frame(cell_id = cell_ids, stringsAsFactors = FALSE)
    first_idx <- match(cell_ids, df$cell_id)
    cell_df$cell_cen_lon <- df$cell_cen_lon[first_idx]
    cell_df$cell_cen_lat <- df$cell_cen_lat[first_idx]

    # Count points per cell
    cell_df$n_points <- as.integer(table(df$cell_id)[as.character(cell_ids)])

    # Add value column if specified
    if (!is.null(value)) {
      if (!value %in% names(df)) {
        stop(sprintf("Column '%s' not found in hex_data", value))
      }
      # Mean per cell
      agg <- tapply(df[[value]], df$cell_id, mean, na.rm = TRUE)
      cell_df[[value]] <- as.numeric(agg[as.character(cell_ids)])
    }
  } else if (is.data.frame(hex_data)) {
    if (is.null(grid)) stop("grid is required when hex_data is a data.frame")
    g <- extract_grid(grid)
    if (!"cell_id" %in% names(hex_data)) {
      stop("hex_data must have a 'cell_id' column")
    }
    # Duplicate cell_id rows are collapsed to their first occurrence --
    # cell_to_sf() dedupes internally below, and cell_df's columns are
    # attached to hex_sf by row position afterward, so unresolved duplicates
    # would misalign the attached data or crash on the row-count mismatch.
    cell_df <- hex_data[!duplicated(hex_data$cell_id), , drop = FALSE]
    cell_ids <- cell_df$cell_id
    # The popup text below needs cell centers; the data.frame input mode
    # doesn't require the caller to supply them.
    if (!all(c("cell_cen_lon", "cell_cen_lat") %in% names(cell_df))) {
      ll <- cell_to_lonlat(cell_ids, g)
      cell_df$cell_cen_lon <- ll$lon_deg
      cell_df$cell_cen_lat <- ll$lat_deg
    }
  } else {
    stop("hex_data must be a HexData object or data.frame with cell_id column")
  }

  # Generate polygons
  hex_sf <- cell_to_sf(cell_ids, g)

  # Attach data to sf by cell_id key rather than assuming row-position
  # alignment (cell_to_sf() dedupes/may reorder internally).
  cell_df <- cell_df[match(hex_sf$cell_id, cell_df$cell_id), , drop = FALSE]
  cell_ids <- cell_df$cell_id
  for (col in names(cell_df)) {
    if (col != "cell_id") {
      hex_sf[[col]] <- cell_df[[col]]
    }
  }

  # Build popup text
  popup_text <- paste0(
    "<b>Cell ID:</b> ", cell_ids, "<br>",
    "<b>Center:</b> (",
    round(cell_df$cell_cen_lon, 4), ", ",
    round(cell_df$cell_cen_lat, 4), ")<br>"
  )
  if ("n_points" %in% names(cell_df)) {
    popup_text <- paste0(popup_text, "<b>Points:</b> ", cell_df$n_points, "<br>")
  }
  if (!is.null(value) && value %in% names(cell_df)) {
    popup_text <- paste0(popup_text, "<b>", value, ":</b> ",
                          round(cell_df[[value]], 4), "<br>")
  }

  # Build leaflet map
  m <- leaflet::leaflet(hex_sf)
  m <- leaflet::addTiles(m)

  if (!is.null(value) && value %in% names(cell_df)) {
    vals <- cell_df[[value]]
    pal <- leaflet::colorNumeric(palette, domain = vals, na.color = "#808080")
    fill_colors <- pal(vals)
    m <- leaflet::addPolygons(m,
                               fillColor = fill_colors,
                               fillOpacity = opacity,
                               weight = 1,
                               color = "#333333",
                               opacity = 0.5,
                               popup = popup_text)
    m <- leaflet::addLegend(m, pal = pal, values = vals, title = value)
  } else {
    m <- leaflet::addPolygons(m,
                               fillColor = "#3388ff",
                               fillOpacity = opacity,
                               weight = 1,
                               color = "#333333",
                               opacity = 0.5,
                               popup = popup_text)
  }

  m
}

# tests/testthat/test-interactive.R
# Smoke tests for hex_browse()

test_that("hex_browse creates a leaflet map", {
  skip_if_not_installed("leaflet")

  df <- data.frame(
    lon = runif(10, 5, 15),
    lat = runif(10, 45, 55),
    value = rnorm(10)
  )
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 2000)
  m <- hex_browse(hd)
  expect_true(inherits(m, "leaflet"))
})

test_that("hex_browse with value column works", {
  skip_if_not_installed("leaflet")

  set.seed(1)
  df <- data.frame(
    lon = runif(10, 5, 15),
    lat = runif(10, 45, 55),
    temp = rnorm(10, 15)
  )
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 2000)
  m <- hex_browse(hd, value = "temp")
  expect_true(inherits(m, "leaflet"))

  # The value= mapping must actually reach the widget's polygon fill colors
  # (not just build *a* map): fillColor should vary across cells, and a map
  # built without value= should use the flat default fill instead. The map
  # carries one polygon per cell, while the hexified data carries one row per
  # point, and several points can share a cell.
  poly_call <- Filter(function(c) c$method == "addPolygons", m$x$calls)[[1]]
  fill_colors <- poly_call$args[[4]]$fillColor
  expect_length(fill_colors, length(unique(as.data.frame(hd)$cell_id)))
  expect_gt(length(unique(fill_colors)), 1)

  m_default <- hex_browse(hd)
  poly_call_default <- Filter(function(c) c$method == "addPolygons", m_default$x$calls)[[1]]
  expect_identical(poly_call_default$args[[4]]$fillColor, "#3388ff")
})

test_that("hex_browse errors without leaflet", {
  skip_if(requireNamespace("leaflet", quietly = TRUE))
  df <- data.frame(lon = 10, lat = 50)
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
  expect_error(hex_browse(hd), "leaflet")
})

# constructors.R
# Constructor functions for HexGridInfo and HexData S4 classes
#
# These functions provide user-friendly interfaces for creating
# grid specifications and hexified data objects.

# =============================================================================
# HexGridInfo CONSTRUCTOR
# =============================================================================

#' Create a Hexagonal Grid Specification
#'
#' Creates a HexGridInfo object that stores all parameters needed for hexagonal
#' grid operations. Use this to define the grid once and pass it to all
#' downstream functions.
#'
#' @param area_km2 Target cell area in square kilometers. Mutually exclusive
#'   with \code{resolution}.
#' @param resolution Grid resolution level (0-30 for ISEA, 0-15 for H3).
#'   Mutually exclusive with \code{area_km2}. For H3, typical use cases by
#'   resolution:
#'   \itemize{
#'     \item 0-3: continental/country scale
#'     \item 4-7: regional/city scale
#'     \item 8-10: neighborhood/block scale (FCC uses 8-9)
#'     \item 11-15: building/sub-meter scale
#'   }
#' @param aperture Grid aperture: 3 (default), 4, 7, a mixed family such as
#'   "4/3", "4/7" or "7/4", or one aperture per resolution level as a vector,
#'   e.g. \code{c(4, 4, 7, 3)}. A family name refines by the first aperture for
#'   the first \code{floor(resolution / 2)} levels and by the second for the
#'   rest, which is how DGGRID arranges ISEA43H. A per-level vector needs
#'   \code{resolution} rather than \code{area_km2}. Ignored for H3 grids (fixed
#'   at 7).
#' @param type Grid type: "isea" (default) or "h3".
#' @param resround Resolution rounding when using \code{area_km2}:
#'   "nearest" (default), "up", or "down".
#' @param crs Coordinate reference system: an EPSG code, or a 'PROJ' or 'WKT'
#'   string. Defaults to 'WGS84' on Earth, and to a longlat CRS on the sphere of
#'   \code{radius_km} on any other body, which has no EPSG code to name it.
#' @param radius_km Radius of the body the grid covers, in kilometers, or the
#'   name of a body: "mercury", "venus", "earth" (default), "moon", "mars",
#'   "ceres", "jupiter", "io", "europa", "ganymede", "callisto", "saturn",
#'   "enceladus", "titan", "uranus", "neptune", "pluto".
#'
#' @return A HexGridInfo object containing the grid specification.
#'
#' @details
#' Exactly one of \code{area_km2} or \code{resolution} must be provided.
#'
#' When \code{area_km2} is provided, the resolution is calculated automatically
#' using the cell count formula: N = 10 * aperture^res + 2 (ISEA) or by
#' matching the closest H3 resolution.
#'
#' H3 grids use the Uber H3 hierarchical hexagonal system. Unlike ISEA grids,
#' H3 cells are NOT exactly equal-area (area varies by ~3-5\% depending on
#' location).
#'
#' @section Other Bodies:
#'
#' A grid is a partition of the sphere, and \code{radius_km} sets the sphere it
#' is measured on. Cell geometry -- which cell a coordinate lands in, where cell
#' centres and corners sit, the hierarchy, the neighbours -- is angular and
#' identical on every body; the radius sets the kilometer figures: cell area,
#' diagonal, spacing, and the resolution that \code{area_km2} picks. Earth's
#' area comes from the 'WGS84' ellipsoid, every other radius gives the sphere
#' area 4*pi*r^2.
#'
#' \preformatted{
#' mars <- hex_grid(area_km2 = 1000, radius_km = "mars")
#' hex_grid(resolution = 8, radius_km = 3389.5)   # the same grid
#' }
#'
#' Both backends take a radius. 'H3' reports a cell's area as its solid angle
#' times Earth's radius squared, so another radius scales those areas by the
#' square of the radius ratio, exactly. One caveat carries: an 'H3' cell ID
#' names a position in 'H3''s topology, which 'Uber''s 'H3' reads on Earth, so
#' the IDs of a grid on another body are that topology on that body and are not
#' interchangeable with Earth 'H3' data.
#'
#' @seealso \code{\link{hexify}} for assigning points to cells,
#'   \code{\link{HexGridInfo-class}} for class documentation
#'
#' @section One Grid, Many Datasets:
#'
#' A HexGridInfo acts as a shared spatial reference system - like a CRS, but
#' discrete and equal-area. Define the grid once, then attach multiple
#' datasets without repeating parameters:
#'
#' \preformatted{
#' # Step 1: Define the grid once
#' grid <- hex_grid(area_km2 = 1000)
#'
#' # Step 2: Attach multiple datasets to the same grid
#' birds <- hexify(bird_obs, lon = "longitude", lat = "latitude", grid = grid)
#' mammals <- hexify(mammal_obs, lon = "lon", lat = "lat", grid = grid)
#' climate <- hexify(weather_stations, lon = "x", lat = "y", grid = grid)
#'
#' # No aperture, resolution, or area needed after step 1 - the grid
#' # travels with the data.
#'
#' # Step 3: Work at the cell level
#' # Once hexified, lon/lat no longer matter - cell_id is the shared key
#' bird_counts <- aggregate(species ~ cell_id, data = as.data.frame(birds), length)
#' mammal_richness <- aggregate(species ~ cell_id, data = as.data.frame(mammals),
#'                              function(x) length(unique(x)))
#'
#' # Join datasets by cell_id - guaranteed to align because same grid
#' combined <- merge(bird_counts, mammal_richness, by = "cell_id")
#'
#' # Step 4: Visual confirmation
#' # All datasets produce identical grid overlays
#' plot(birds)   # See the grid
#' plot(mammals) # Same grid, different data
#' }
#'
#' @export
#' @examples
#' # Create grid by target area
#' grid <- hex_grid(area_km2 = 1000)
#' print(grid)
#'
#' # Create grid by resolution
#' grid <- hex_grid(resolution = 8, aperture = 3)
#'
#' # Create grid with different aperture
#' grid4 <- hex_grid(area_km2 = 500, aperture = 4)
#'
#' # Create mixed aperture grid
#' grid43 <- hex_grid(area_km2 = 1000, aperture = "4/3")
#'
#' # Mix in aperture 7, either as a family or level by level
#' grid47 <- hex_grid(area_km2 = 1000, aperture = "4/7")
#' grid_seq <- hex_grid(resolution = 4, aperture = c(4, 4, 7, 3))
#'
#' # Grid on another body, by name or by radius
#' mars <- hex_grid(area_km2 = 1000, radius_km = "mars")
#' titan <- hex_grid(resolution = 6, radius_km = 2574.76)
#'
#' # Use grid in hexify
#' df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
#' result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
hex_grid <- function(area_km2 = NULL,
                     resolution = NULL,
                     aperture = 3,
                     type = c("isea", "h3"),
                     resround = "nearest",
                     crs = NULL,
                     radius_km = EARTH_RADIUS_KM) {

  type <- match.arg(type)

  radius_km <- resolve_radius_km(radius_km)
  crs <- resolve_crs(crs, radius_km)

  # =========================================================================
  # H3 grid path
  # =========================================================================
  if (type == "h3") {
    if (!missing(aperture) && aperture != 3) {
      warning("aperture is ignored for H3 grids (H3 uses fixed aperture 7)")
    }

    # Validate: exactly one of area_km2 or resolution
    if (is.null(area_km2) && is.null(resolution)) {
      stop("Exactly one of 'area_km2' or 'resolution' must be provided")
    }
    if (!is.null(area_km2) && !is.null(resolution)) {
      stop("Provide either 'area_km2' or 'resolution', not both")
    }

    if (!is.null(area_km2)) {
      if (!is.numeric(area_km2) || area_km2 <= 0) {
        stop("area_km2 must be a positive number")
      }
      # Find closest H3 resolution by area
      resolution <- closest_h3_resolution(area_km2, radius_km)
      warning(sprintf(
        "H3 cells are not exactly equal-area. Closest resolution %d has average area ~%.3f km^2 (requested %.3f km^2)",
        resolution, h3_avg_area_km2(resolution, radius_km), area_km2
      ))
    } else {
      if (!is.numeric(resolution) || length(resolution) != 1 || is.na(resolution)) {
        stop("resolution must be a single non-NA number")
      }
      resolution <- as.integer(resolution)
      if (resolution < H3_MIN_RESOLUTION || resolution > H3_MAX_RESOLUTION) {
        stop(sprintf("H3 resolution must be between %d and %d",
                     H3_MIN_RESOLUTION, H3_MAX_RESOLUTION))
      }
      rlang::inform(
        "H3 cells are not exactly equal-area; area varies ~3-5% by latitude.",
        .frequency = "once",
        .frequency_id = "hexify_h3_not_equal_area"
      )
    }

    if (radius_km != EARTH_RADIUS_KM) {
      rlang::inform(
        paste0(
          "H3 cell IDs name a position in H3's topology, which Uber's H3 reads ",
          "on Earth. A grid on another body reuses that topology and its own ",
          "radius for areas; the IDs are not interchangeable with Earth H3 data."
        ),
        .frequency = "once",
        .frequency_id = "hexify_h3_other_body"
      )
    }

    actual_area <- h3_avg_area_km2(resolution, radius_km)
    actual_diagonal <- sqrt(actual_area * 2 / sqrt(3))

    grid <- new("HexGridInfo",
                aperture = "7",
                resolution = as.integer(resolution),
                area_km2 = as.numeric(actual_area),
                diagonal_km = as.numeric(actual_diagonal),
                crs = crs,
                grid_type = "h3",
                radius_km = radius_km)
    return(grid)
  }

  # =========================================================================
  # ISEA grid path (original logic)
  # =========================================================================

  # -------------------------------------------------------------------------
  # Parse aperture: a single aperture, a family such as "4/3", or one aperture
  # per resolution level (see R/aperture_sequence.R)
  # -------------------------------------------------------------------------
  if (length(aperture) > 1L && is.null(resolution)) {
    stop("An aperture given per level needs 'resolution', not 'area_km2'")
  }
  aperture_str <- format_aperture(aperture, resolution)

  if (is_mixed_aperture(aperture_str)) {
    aperture_num <- NA_integer_
  } else if (aperture_str %in% c("3", "4", "7")) {
    aperture_num <- as.integer(aperture_str)
  } else {
    stop("Aperture must be 3, 4, 7, a family such as \"4/3\", or one aperture per level")
  }

  # -------------------------------------------------------------------------
  # Validate area_km2 / resolution (exactly one required)
  # -------------------------------------------------------------------------
  if (is.null(area_km2) && is.null(resolution)) {
    stop("Exactly one of 'area_km2' or 'resolution' must be provided")
  }
  if (!is.null(area_km2) && !is.null(resolution)) {
    stop("Provide either 'area_km2' or 'resolution', not both")
  }

  # -------------------------------------------------------------------------
  # Calculate resolution from area if needed
  # -------------------------------------------------------------------------
  if (!is.null(area_km2)) {
    if (!is.numeric(area_km2) || area_km2 <= 0) {
      stop("area_km2 must be a positive number")
    }

    res_exact <- if (is_mixed_aperture(aperture_str)) {
      calculate_resolution_for_area_mixed(area_km2, aperture_str, radius_km)
    } else {
      calculate_resolution_for_area(area_km2, aperture_num, radius_km)
    }

    # Apply rounding
    resolution <- switch(resround,
      "nearest" = round(res_exact),
      "up" = ceiling(res_exact),
      "down" = floor(res_exact),
      stop("resround must be 'nearest', 'up', or 'down'")
    )

    # Clamp to valid range
    resolution <- max(MIN_RESOLUTION, min(MAX_RESOLUTION, resolution))
  } else {
    if (!is.numeric(resolution) || length(resolution) != 1 || is.na(resolution)) {
      stop("resolution must be a single non-NA number")
    }
    resolution <- as.integer(resolution)
    if (resolution < MIN_RESOLUTION || resolution > MAX_RESOLUTION) {
      stop(sprintf("Resolution must be between %d and %d",
                   MIN_RESOLUTION, MAX_RESOLUTION))
    }
  }

  # -------------------------------------------------------------------------
  # Calculate actual area and diagonal for this resolution
  # -------------------------------------------------------------------------
  n_cells <- aperture_n_cells(aperture_str, resolution)
  actual_area <- body_surface_km2(radius_km) / n_cells
  actual_diagonal <- sqrt(actual_area * 2 / sqrt(3))

  # -------------------------------------------------------------------------
  # Initialize icosahedron (required for C++ functions)
  # -------------------------------------------------------------------------
  cpp_build_icosa()

  # -------------------------------------------------------------------------
  # Create and validate HexGridInfo object
  # -------------------------------------------------------------------------
  grid <- new("HexGridInfo",
              aperture = aperture_str,
              resolution = as.integer(resolution),
              area_km2 = as.numeric(actual_area),
              diagonal_km = as.numeric(actual_diagonal),
              crs = crs,
              grid_type = "isea",
              radius_km = radius_km)

  # Validation happens automatically via setValidity
  grid
}

# =============================================================================
# HexData CONSTRUCTOR (Internal)
# =============================================================================

#' Create a HexData Object (Internal)
#'
#' Internal constructor for HexData objects. Users should use \code{hexify()}
#' instead.
#'
#' @param data Data frame or sf object (original user data, untouched)
#' @param grid HexGridInfo object
#' @param cell_id Numeric vector of cell IDs for each row
#' @param cell_center Matrix with columns lon, lat for cell centers
#'
#' @return A HexData object
#' @keywords internal
new_hex_data <- function(data,
                         grid,
                         cell_id,
                         cell_center) {

  # Validate inputs
  if (!inherits(data, "data.frame") && !inherits(data, "sf")) {
    stop("data must be a data.frame or sf object")
  }

  if (!is_hex_grid(grid)) {
    stop("grid must be a HexGridInfo object")
  }

  # Ensure cell_center is a matrix with correct column names
  if (!is.matrix(cell_center)) {
    cell_center <- as.matrix(cell_center)
  }
  if (is.null(colnames(cell_center))) {
    colnames(cell_center) <- c("lon", "lat")
  }

  # For ISEA grids, coerce to numeric; for H3, keep as character
  if (is_h3_grid(grid)) {
    cell_id <- as.character(cell_id)
  } else {
    cell_id <- as.numeric(cell_id)
  }

  new("HexData",
      data = data,
      grid = grid,
      cell_id = cell_id,
      cell_center = cell_center)
}

# =============================================================================
# TIBBLE COERCION (if tibble is available)
# =============================================================================

#' Convert HexData to tibble
#'
#' @param x A HexData object
#' @param ... Additional arguments (ignored)
#' @return A tibble
#'
#' @export
as_tibble.HexData <- function(x, ...) {
  if (!requireNamespace("tibble", quietly = TRUE)) {
    stop("Package 'tibble' is required for as_tibble(). ",
         "Install with: install.packages('tibble')")
  }

  df <- x@data
  if (inherits(df, "sf")) {
    df <- as.data.frame(sf::st_drop_geometry(df))
  }

  tibble::as_tibble(df)
}

# =============================================================================
# SF COERCION
# =============================================================================

#' Convert HexData to sf Object
#'
#' Converts a HexData object to an sf spatial features object. Can create
#' either point geometries (cell centers) or polygon geometries (cell boundaries).
#'
#' @param x A HexData object
#' @param geometry Type of geometry: "point" (default) or "polygon"
#' @param ... Additional arguments (ignored)
#'
#' @return An sf object
#'
#' @details
#' For point geometry, cell centers (cell_cen_lon, cell_cen_lat) are used.
#' For polygon geometry, cell boundaries are computed using the grid specification.
#'
#' @export
#' @examples
#' df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
#' result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
#'
#' # Get sf points
#' sf_pts <- as_sf(result)
#'
#' # Get sf polygons
#' sf_poly <- as_sf(result, geometry = "polygon")
as_sf <- function(x, geometry = c("point", "polygon"), ...) {
  UseMethod("as_sf")
}

#' @export
as_sf.HexData <- function(x, geometry = c("point", "polygon"), ...) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required. Install with: install.packages('sf')")
  }

  geometry <- match.arg(geometry)
  data <- x@data
  grid <- x@grid

  # If already sf, check if we need to change geometry type
  if (inherits(data, "sf")) {
    if (geometry == "point") {
      return(data)  # Already has geometry
    }
    # For polygon, we need to replace geometry
    data <- sf::st_drop_geometry(data)
  }

  if (geometry == "point") {
    # Point geometry from cell centers (stored in cell_center slot)
    df_with_coords <- cbind(
      data,
      cell_id = x@cell_id,
      cell_cen_lon = x@cell_center[, "lon"],
      cell_cen_lat = x@cell_center[, "lat"]
    )
    sf::st_as_sf(
      df_with_coords,
      coords = c("cell_cen_lon", "cell_cen_lat"),
      crs = grid_crs(grid)
    )

  } else {
    # Polygon geometry from cell boundaries
    unique_ids <- unique(x@cell_id)

    polys_sf <- cell_to_sf(unique_ids, grid)

    # Add cell_id to data for merge
    data_with_id <- cbind(data, cell_id = x@cell_id)

    # Merge with original data
    result <- merge(polys_sf, data_with_id, by = "cell_id", all.y = TRUE)
    sf::st_as_sf(result)
  }
}

#' @export
as_sf.default <- function(x, ...) {
  stop("as_sf() is not defined for objects of class ", class(x)[1])
}

#' Coordinate reference system of a grid
#'
#' Reads the CRS a grid's coordinates are in, as \code{sf::st_crs()} does for
#' any spatial object. An Earth grid returns 'WGS84'; a grid built on another
#' body returns a longlat CRS on the sphere of its radius.
#'
#' @param x A HexGridInfo or HexData object
#' @param ... Passed on to \code{sf::st_crs()}
#' @return An object of class \code{crs}, as \code{sf::st_crs()} returns:
#'   a list carrying the reference system in 'PROJ' and 'WKT' form. It says how
#'   to read the coordinates that the grid's cells, centres and sf exports come
#'   back in.
#'
#' @importFrom sf st_crs
#' @export
#' @examples
#' st_crs(hex_grid(resolution = 5))
#' st_crs(hex_grid(resolution = 5, radius_km = "mars"))
st_crs.HexGridInfo <- function(x, ...) {
  grid_crs(x)
}

#' @rdname st_crs.HexGridInfo
#' @export
st_crs.HexData <- function(x, ...) {
  grid_crs(x@grid)
}

# tests/testthat/test-constructors.R
# Tests for constructor functions

# =============================================================================
# hex_grid Tests
# =============================================================================

test_that("hex_grid creates HexGridInfo with area_km2", {
  grid <- hex_grid(area_km2 = 1000)

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@aperture, "3")
  expect_true(grid@resolution > 0)
})

test_that("hex_grid creates HexGridInfo with resolution", {
  grid <- hex_grid(resolution = 8, aperture = 3)

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@resolution, 8L)
})

test_that("hex_grid handles aperture 4", {
  grid <- hex_grid(area_km2 = 1000, aperture = 4)

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@aperture, "4")
})

test_that("hex_grid handles aperture 7", {
  grid <- hex_grid(area_km2 = 10000, aperture = 7)

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@aperture, "7")
})

test_that("hex_grid handles mixed aperture 4/3", {
  grid <- hex_grid(area_km2 = 1000, aperture = "4/3")

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@aperture, "4/3")
})

test_that("hex_grid errors without area_km2 or resolution", {
  expect_error(hex_grid(), "Exactly one")
})

test_that("hex_grid errors with both area_km2 and resolution", {
  expect_error(hex_grid(area_km2 = 1000, resolution = 5), "not both")
})

test_that("hex_grid errors with invalid aperture", {
  expect_error(hex_grid(area_km2 = 1000, aperture = 5), "Aperture must be")
})

test_that("hex_grid errors with invalid area_km2", {
  expect_error(hex_grid(area_km2 = -100), "positive number")
  expect_error(hex_grid(area_km2 = "text"), "positive number")
})

test_that("hex_grid errors with invalid resolution", {
  expect_error(hex_grid(resolution = -1), "must be between")
  expect_error(hex_grid(resolution = 100), "must be between")
})

test_that("hex_grid works with resround = 'up'", {
  grid <- hex_grid(area_km2 = 1000, resround = "up")

  expect_s4_class(grid, "HexGridInfo")
})

test_that("hex_grid works with resround = 'down'", {
  grid <- hex_grid(area_km2 = 1000, resround = "down")

  expect_s4_class(grid, "HexGridInfo")
})

test_that("hex_grid errors with invalid resround", {
  expect_error(hex_grid(area_km2 = 1000, resround = "invalid"), "resround must be")
})

test_that("hex_grid sets custom CRS", {
  grid <- hex_grid(area_km2 = 1000, crs = 3857)

  expect_equal(grid@crs, 3857L)
})

test_that("hex_grid calculates area and diagonal", {
  grid <- hex_grid(area_km2 = 1000)

  expect_true(grid@area_km2 > 0)
  expect_true(grid@diagonal_km > 0)
})

# =============================================================================
# new_hex_data Tests (internal constructor)
# =============================================================================

test_that("new_hex_data creates HexData object", {
  grid <- hex_grid(area_km2 = 1000)
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  cell_id <- c(12345, 67890)
  cell_center <- matrix(c(0, 10, 45, 50), ncol = 2, dimnames = list(NULL, c("lon", "lat")))

  result <- hexify:::new_hex_data(df, grid, cell_id, cell_center)

  expect_s4_class(result, "HexData")
  expect_equal(result@cell_id, c(12345, 67890))
})

test_that("new_hex_data errors on invalid data", {
  grid <- hex_grid(area_km2 = 1000)
  cell_id <- c(12345, 67890)
  cell_center <- matrix(c(0, 10, 45, 50), ncol = 2)

  expect_error(hexify:::new_hex_data(list(a = 1), grid, cell_id, cell_center),
               "must be a data.frame")
})

test_that("new_hex_data errors on invalid grid", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  cell_id <- c(12345, 67890)
  cell_center <- matrix(c(0, 10, 45, 50), ncol = 2)

  expect_error(hexify:::new_hex_data(df, list(a = 1), cell_id, cell_center),
               "must be a HexGridInfo")
})

test_that("new_hex_data adds column names to cell_center matrix", {
  grid <- hex_grid(area_km2 = 1000)
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  cell_id <- c(12345, 67890)
  cell_center <- matrix(c(0, 10, 45, 50), ncol = 2)  # No column names

  result <- hexify:::new_hex_data(df, grid, cell_id, cell_center)

  expect_equal(colnames(result@cell_center), c("lon", "lat"))
})

test_that("new_hex_data works with sf data", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 1000)
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  sf_data <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  cell_id <- c(12345, 67890)
  cell_center <- matrix(c(0, 10, 45, 50), ncol = 2, dimnames = list(NULL, c("lon", "lat")))

  result <- hexify:::new_hex_data(sf_data, grid, cell_id, cell_center)

  expect_s4_class(result, "HexData")
})

# tests/testthat/test-hexify-function.R
# Tests for the main hexify() convenience function
#
# The hexify() function is the primary user-facing API that converts
# a data frame with lon/lat to hexagonal grid cells.
#
# Updated for S4 HexData return type.

# =============================================================================
# BASIC FUNCTIONALITY
# =============================================================================

test_that("hexify works with data.frame and area_km2 parameter", {
  df <- data.frame(
    site = c("Vienna", "Paris", "Madrid"),
    lon = c(16.37, 2.35, -3.70),
    lat = c(48.21, 48.86, 40.42)
  )

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # Returns HexData object
  expect_s4_class(result, "HexData")

  # Original columns preserved in @data
  expect_true("site" %in% names(result@data))
  expect_true("lon" %in% names(result@data))
  expect_true("lat" %in% names(result@data))

  # Cell columns accessible via $ (virtual columns)
  expect_true("cell_id" %in% names(result))
  expect_true("cell_cen_lon" %in% names(result))
  expect_true("cell_cen_lat" %in% names(result))

  # Types
  expect_type(result$cell_id, "double")
  expect_type(result$cell_cen_lon, "double")
  expect_type(result$cell_cen_lat, "double")

  # Valid coordinates
  expect_true(all(result$cell_cen_lon >= -180 & result$cell_cen_lon <= 180))
  expect_true(all(result$cell_cen_lat >= -90 & result$cell_cen_lat <= 90))

  # Row count preserved
  expect_equal(nrow(result), 3)
})

test_that("hexify works with diagonal parameter", {
  df <- data.frame(lon = c(0, 10), lat = c(0, 45))

  result <- hexify(df, lon = "lon", lat = "lat", diagonal = 50)

  expect_s4_class(result, "HexData")
  expect_true("cell_id" %in% names(result))
  expect_equal(nrow(result), 2)
})

test_that("hexify works with grid parameter", {
  df <- data.frame(lon = c(0, 10), lat = c(0, 45))

  grid <- hex_grid(area_km2 = 1000)
  result <- hexify(df, lon = "lon", lat = "lat", grid = grid)

  expect_s4_class(result, "HexData")
  expect_identical(grid_info(result), grid)
})

# =============================================================================
# PARAMETER VALIDATION
# =============================================================================

test_that("hexify requires grid, area_km2, diagonal, or resolution", {
  df <- data.frame(lon = 0, lat = 0)

  expect_error(hexify(df, lon = "lon", lat = "lat"),
               "Either 'grid', 'area_km2', 'diagonal', or 'resolution' must be provided")

  expect_error(hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, diagonal = 50),
               "Provide either 'area_km2' or 'diagonal', not both")
})

test_that("hexify validates column names", {
  df <- data.frame(x = 0, y = 0)

  expect_error(hexify(df, lon = "lon", lat = "lat", area_km2 = 1000),
               "Column 'lon' not found")

  expect_error(hexify(df, lon = "x", lat = "lat", area_km2 = 1000),
               "Column 'lat' not found")
})

# =============================================================================
# SF INTEGRATION
# =============================================================================

test_that("hexify works with sf objects", {
  skip_if_not_installed("sf")

  df <- data.frame(
    site = c("Vienna", "Paris"),
    lon = c(16.37, 2.35),
    lat = c(48.21, 48.86)
  )
  pts <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)

  result <- hexify(pts, area_km2 = 1000)

  expect_s4_class(result, "HexData")
  # Data slot should be sf
  expect_s3_class(result@data, "sf")
  expect_true("cell_id" %in% names(result))
  expect_true("cell_cen_lon" %in% names(result))
  expect_true("cell_cen_lat" %in% names(result))
  expect_true("site" %in% names(result))
})

# =============================================================================
# APERTURE SUPPORT
# =============================================================================

test_that("hexify handles aperture 3 (ISEA3H)", {
  df <- data.frame(lon = 0, lat = 45)

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 3)

  expect_s4_class(result, "HexData")
  expect_true("cell_id" %in% names(result))
  expect_type(result$cell_id, "double")
  expect_true(result$cell_id > 0)
  expect_equal(grid_info(result)@aperture, "3")
})

test_that("hexify handles aperture 4 (ISEA4H)", {
  df <- data.frame(lon = 0, lat = 45)

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 4)

  expect_s4_class(result, "HexData")
  expect_true("cell_id" %in% names(result))
  expect_type(result$cell_id, "double")
  expect_true(result$cell_id > 0)
  expect_equal(grid_info(result)@aperture, "4")
})

test_that("hexify handles aperture 7 (ISEA7H)", {
  df <- data.frame(lon = 0, lat = 45)

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 7)

  expect_s4_class(result, "HexData")
  expect_true("cell_id" %in% names(result))
  expect_type(result$cell_id, "double")
  expect_true(result$cell_id > 0)
  expect_equal(grid_info(result)@aperture, "7")
})

test_that("hexify handles mixed aperture 4/3 (ISEA43H)", {
  df <- data.frame(lon = 0, lat = 45)

  # Test with string "4/3"
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = "4/3")

  expect_s4_class(result, "HexData")
  expect_true("cell_id" %in% names(result))
  expect_type(result$cell_id, "double")
  expect_true(result$cell_id > 0)
  expect_equal(grid_info(result)@aperture, "4/3")
})

test_that("hexify rejects unsupported apertures with clear error", {
  df <- data.frame(lon = 0, lat = 45)

  expect_error(
    hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 5),
    "Aperture must be 3, 4, 7"
  )

  expect_error(
    hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 2),
    "Aperture must be 3, 4, 7"
  )

  expect_error(
    hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = "invalid"),
    "Aperture must be 3, 4, 7"
  )
})

# =============================================================================
# OUTPUT COLUMNS
# =============================================================================

test_that("hexify returns cell_area_km2 and cell_diag_km columns", {
  df <- data.frame(lon = 0, lat = 45)

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 3)

  expect_true("cell_area_km2" %in% names(result))
  expect_true("cell_diag_km" %in% names(result))
  expect_type(result$cell_area_km2, "double")
  expect_type(result$cell_diag_km, "double")

  # Values should be reasonable
  expect_true(result$cell_area_km2 > 100 && result$cell_area_km2 < 10000)
  expect_true(result$cell_diag_km > 10 && result$cell_diag_km < 200)
})

test_that("hexify cell_area_km2 and cell_diag_km are consistent across rows", {
  df <- data.frame(lon = c(0, 10, -5), lat = c(45, 30, -20))

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 3)

  # All rows should have same area/diag
  expect_equal(length(unique(result$cell_area_km2)), 1)
  expect_equal(length(unique(result$cell_diag_km)), 1)
})

test_that("hexify cell_area_km2 and cell_diag_km work for all apertures", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 30))

  for (ap in c(3, 4, 7)) {
    result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = ap)

    expect_true("cell_area_km2" %in% names(result), info = sprintf("aperture %d", ap))
    expect_true("cell_diag_km" %in% names(result), info = sprintf("aperture %d", ap))
    expect_true(result$cell_diag_km[1] > 0)
  }
})

# =============================================================================
# NA HANDLING
# =============================================================================

test_that("hexify handles NA coordinate values", {
  df <- data.frame(
    lon = c(0, NA, 10),
    lat = c(45, 46, NA)
  )

  expect_warning(
    result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000),
    "2 coordinate pairs contain NA"
  )

  # Rows with an NA lon or lat are dropped; only row 1 (lon=0, lat=45) is valid
  expect_equal(nrow(result), 1)
  expect_equal(as.data.frame(result)$lon, 0)
  expect_equal(as.data.frame(result)$lat, 45)
})

test_that("hexify errors on all NA coordinates", {
  df <- data.frame(lon = as.numeric(c(NA, NA)), lat = as.numeric(c(NA, NA)))

  expect_error(
    hexify(df, lon = "lon", lat = "lat", area_km2 = 1000),
    "All coordinates are NA"
  )
})

test_that("hexify validates numeric coordinates", {
  df <- data.frame(lon = c("a", "b"), lat = c("c", "d"))

  expect_error(
    hexify(df, lon = "lon", lat = "lat", area_km2 = 1000),
    "Coordinates must be numeric"
  )
})

# =============================================================================
# SF INTEGRATION - ADVANCED
# =============================================================================

test_that("hexify with sf handles non-4326 CRS", {
  skip_if_not_installed("sf")

  df <- data.frame(
    site = c("Vienna", "Paris"),
    lon = c(16.37, 2.35),
    lat = c(48.21, 48.86)
  )

  pts_4326 <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  pts_3857 <- sf::st_transform(pts_4326, crs = 3857)

  result <- hexify(pts_3857, area_km2 = 1000)

  expect_s4_class(result, "HexData")
  expect_s3_class(result@data, "sf")
  expect_true("cell_id" %in% names(result))
})

test_that("hexify with sf handles NA CRS", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(16.37, 2.35),
    lat = c(48.21, 48.86)
  )

  pts <- sf::st_as_sf(df, coords = c("lon", "lat"))
  sf::st_crs(pts) <- NA

  result <- hexify(pts, area_km2 = 1000)
  expect_s4_class(result, "HexData")
})

# =============================================================================
# MIXED APERTURE DETAILED
# =============================================================================

test_that("hexify works with mixed aperture 4/3", {
  df <- data.frame(lon = 0, lat = 45)

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = "4/3")
  expect_true("cell_id" %in% names(result))
  expect_equal(grid_info(result)@aperture, "4/3")
})

# =============================================================================
# RESOLUTION ROUNDING
# =============================================================================

test_that("hexify respects resround parameter", {
  df <- data.frame(lon = 0, lat = 45)

  result_up <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000,
                      aperture = 3, resround = "up")
  result_down <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000,
                        aperture = 3, resround = "down")
  result_nearest <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000,
                           aperture = 3, resround = "nearest")

  expect_true("cell_area_km2" %in% names(result_up))
  expect_true("cell_area_km2" %in% names(result_down))
  expect_true("cell_area_km2" %in% names(result_nearest))
})

# =============================================================================
# EDGE CASES
# =============================================================================

test_that("hexify handles single point", {
  df <- data.frame(lon = 0, lat = 0)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  expect_equal(nrow(result), 1)
  expect_true(is.finite(result$cell_id))
})

test_that("hexify handles poles", {
  df <- data.frame(lon = c(0, 0), lat = c(90, -90))

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  expect_equal(nrow(result), 2)
  expect_true(all(is.finite(result$cell_id)))
})

test_that("hexify handles date line", {
  df <- data.frame(
    lon = c(179, -179, 180, -180),
    lat = c(0, 0, 45, -45)
  )

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  expect_equal(nrow(result), 4)
  expect_true(all(is.finite(result$cell_id)))
})

test_that("hexify handles many points", {
  set.seed(42)
  df <- data.frame(
    lon = runif(100, -180, 180),
    lat = runif(100, -90, 90)
  )

  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_equal(nrow(result), 100)
  expect_true(all(is.finite(result$cell_id)))
})

# =============================================================================
# HEXDATA ACCESSORS
# =============================================================================

test_that("HexData accessors work correctly", {
  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # grid() accessor
  g <- grid_info(result)
  expect_s4_class(g, "HexGridInfo")

  # cells() accessor
  c <- cells(result)
  expect_type(c, "double")
  expect_equal(length(c), n_cells(result))

  # n_cells() accessor
  expect_type(n_cells(result), "integer")

  # nrow/ncol
  expect_equal(nrow(result), 3)
  expect_true(ncol(result) >= 6)

  # names
  expect_true("cell_id" %in% names(result))
})

test_that("HexData subsetting works", {
  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55), val = c(1, 2, 3))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # Subset rows
  subset <- result[1:2, ]
  expect_s4_class(subset, "HexData")
  expect_equal(nrow(subset), 2)

  # Column access
  expect_equal(result$val, c(1, 2, 3))
  expect_equal(result[["val"]], c(1, 2, 3))
})

test_that("as.data.frame works on HexData", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  df_out <- as.data.frame(result)
  expect_s3_class(df_out, "data.frame")
  expect_true("cell_id" %in% names(df_out))
})

# =============================================================================
# HEX_GRID CONSTRUCTOR
# =============================================================================

test_that("hex_grid constructor works with area_km2", {
  grid <- hex_grid(area_km2 = 1000)

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@aperture, "3")
  expect_true(grid@resolution >= 0)
  expect_true(grid@resolution <= 30)
})

test_that("hex_grid constructor works with resolution", {
  grid <- hex_grid(resolution = 8)

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@resolution, 8L)
})

test_that("hex_grid rejects both area_km2 and resolution", {
  expect_error(
    hex_grid(area_km2 = 1000, resolution = 8),
    "Provide either 'area_km2' or 'resolution', not both"
  )
})

test_that("hex_grid supports different apertures", {
  grid3 <- hex_grid(area_km2 = 1000, aperture = 3)
  grid4 <- hex_grid(area_km2 = 1000, aperture = 4)
  grid7 <- hex_grid(area_km2 = 1000, aperture = 7)

  expect_equal(grid3@aperture, "3")
  expect_equal(grid4@aperture, "4")
  expect_equal(grid7@aperture, "7")
})

test_that("hex_grid supports mixed aperture", {
  grid <- hex_grid(area_km2 = 1000, aperture = "4/3")

  expect_equal(grid@aperture, "4/3")
})

# tests/testthat/test-coercion.R
# Tests for type coercion functions (as_tibble, as_sf)

# =============================================================================
# as_sf Tests
# =============================================================================

test_that("as_sf.HexData creates sf with point geometry by default", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  sf_obj <- as_sf(result)

  expect_s3_class(sf_obj, "sf")
  # Check geometry type is POINT
  geom_types <- unique(sf::st_geometry_type(sf_obj))
  expect_true("POINT" %in% geom_types)
})

test_that("as_sf.HexData creates sf with polygon geometry", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55), value = 1:3)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  sf_obj <- as_sf(result, geometry = "polygon")

  expect_s3_class(sf_obj, "sf")
  # Check geometry type is POLYGON
  geom_types <- unique(sf::st_geometry_type(sf_obj))
  expect_true(any(geom_types %in% c("POLYGON", "MULTIPOLYGON")))
})

test_that("as_sf.HexData preserves data columns", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = c(100, 200))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  sf_obj <- as_sf(result)

  expect_true("value" %in% names(sf_obj))
  expect_equal(sf_obj$value, c(100, 200))
})

test_that("as_sf.HexData works with sf input data", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  sf_data <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  result <- hexify(sf_data, area_km2 = 1000)

  sf_obj <- as_sf(result)

  expect_s3_class(sf_obj, "sf")
})

test_that("as_sf.HexData polygon geometry works with sf input", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
  sf_data <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  result <- hexify(sf_data, area_km2 = 10000)

  sf_obj <- as_sf(result, geometry = "polygon")

  expect_s3_class(sf_obj, "sf")
  geom_types <- unique(sf::st_geometry_type(sf_obj))
  expect_true(any(geom_types %in% c("POLYGON", "MULTIPOLYGON")))
})

test_that("as_sf.default errors for non-HexData", {
  expect_error(as_sf(data.frame()), "not defined for objects")
  expect_error(as_sf(list()), "not defined for objects")
})

test_that("as_sf requires sf package", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # This test just confirms the function runs - package is installed
  sf_obj <- as_sf(result)
  expect_s3_class(sf_obj, "sf")
})

# =============================================================================
# as_tibble Tests
# =============================================================================

test_that("as_tibble.HexData creates tibble", {
  skip_if_not_installed("tibble")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  tbl <- as_tibble.HexData(result)

  expect_s3_class(tbl, "tbl_df")
  expect_equal(nrow(tbl), 2)
})

test_that("as_tibble.HexData preserves columns", {
  skip_if_not_installed("tibble")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = c(100, 200))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  tbl <- as_tibble.HexData(result)

  expect_true("value" %in% names(tbl))
  expect_equal(tbl$value, c(100, 200))
})

test_that("as_tibble.HexData drops sf geometry", {
  skip_if_not_installed("tibble")
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  sf_data <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  result <- hexify(sf_data, area_km2 = 1000)

  tbl <- as_tibble.HexData(result)

  expect_s3_class(tbl, "tbl_df")
  expect_false(inherits(tbl, "sf"))
})

# hexify_grid.R
# Core grid construction and validation functions
#
# This file contains the fundamental grid construction functions that form
# the foundation of the hexify package.

#' @title Core Grid Construction
#' @description Core functions for hexify grid construction and validation
#' @name hexify-grid
NULL

#' Calculate resolution for target area
#'
#' Uses the 'ISEA3H'/'ISEA4H'/'ISEA7H' cell count formula
#' N = 10 * aperture^res + 2, which matches 'dggridR' resolution numbering
#' exactly.
#'
#' @param target_area_km2 Target area in square kilometers
#' @param aperture Aperture (3, 4, or 7)
#' @param radius_km Radius of the body, in kilometers
#' @return Resolution level
#' @keywords internal
calculate_resolution_for_area <- function(target_area_km2, aperture = 3,
                                          radius_km = EARTH_RADIUS_KM) {
  n_cells <- body_surface_km2(radius_km) / target_area_km2

  # Solving N = 10 * aperture^res + 2 for res given target area:
  # res = log((surface / area - 2) / 10) / log(aperture)
  resolution <- log((n_cells - 2) / 10) / log(aperture)

  return(resolution)  # Return unrounded for caller to handle rounding mode
}

#' Create a hexagonal grid specification
#'
#' Creates a discrete global grid system (DGGS) object with hexagonal cells
#' at a specified resolution. This is the main constructor for hexify grids.
#'
#' @param area Target cell area in km^2 (if metric=TRUE) or area code
#' @param topology Grid topology (only "HEXAGON" supported)
#' @param metric Whether area is in metric units (km^2)
#' @param resround How to round resolution ("nearest", "up", "down")
#' @param aperture Aperture sequence (3, 4, or 7)
#' @param projection Projection type (only 'ISEA' supported currently)
#' @param radius_km Radius of the body the grid covers, in kilometers, or a body
#'   name such as "mars" (default Earth). See \code{\link{hex_grid}}.
#'
#' @return A hexify_grid object containing:
#'   \item{area}{Target cell area}
#'   \item{resolution}{Calculated resolution level}
#'   \item{aperture}{Grid aperture (3, 4, or 7)}
#'   \item{topology}{Grid topology ("HEXAGON")}
#'   \item{projection}{Map projection ("ISEA")}
#'   \item{radius_km}{Radius of the body, in kilometers}
#'   \item{index_type}{Index encoding type ("z3", "z7", or "zorder")}
#'
#' @family hexify main
#' @seealso \code{\link{hexify}} for the main user function,
#'   \code{\link{hexify_grid_to_cell}} for coordinate conversion
#' @export
#' @examples
#' # Create a grid with ~1000 km^2 cells
#' grid <- hexify_grid(area = 1000, aperture = 3)
#' print(grid)
#'
#' # Create a finer resolution grid (~100 km^2 cells)
#' fine_grid <- hexify_grid(area = 100, aperture = 3, resround = "up")
hexify_grid <- function(area, 
                             topology = "HEXAGON", 
                             metric = TRUE,
                             resround = "nearest",
                             aperture = 3,
                             projection = "ISEA",
                             radius_km = EARTH_RADIUS_KM) {

  # Input validation
  if (topology != "HEXAGON") {
    stop("Only HEXAGON topology is supported")
  }

  if (projection != "ISEA") {
    stop("Only ISEA projection is supported")
  }

  validate_aperture(aperture)

  radius_km <- resolve_radius_km(radius_km)

  if (!resround %in% c("nearest", "up", "down")) {
    stop("resround must be 'nearest', 'up', or 'down'")
  }

  if (!is.numeric(area) || length(area) != 1 || is.na(area) || area <= 0) {
    stop("area must be a positive number")
  }

  # Calculate resolution for target area
  resolution <- calculate_resolution_for_area(area, aperture, radius_km)
  
  # Apply rounding
  if (resround == "up") {
    resolution <- ceiling(resolution)
  } else if (resround == "down") {
    resolution <- floor(resolution)
  } else {
    resolution <- round(resolution)
  }
  
  # Ensure resolution is valid
  resolution <- max(MIN_RESOLUTION, min(MAX_RESOLUTION, resolution))
  
  # Initialize icosahedron geometry (C++ function)
  cpp_build_icosa()
  
  # Determine index type based on aperture
  index_type <- if (aperture == 3) {
    "z3"
  } else if (aperture == 7) {
    "z7"
  } else {
    "zorder"
  }
  
  # Create grid specification with both hexify and dggridR-compatible fields
  grid <- list(
    # Hexify fields
    area = area,
    resolution = resolution,
    aperture = aperture,
    topology = topology,
    projection = projection,
    metric = metric,
    radius_km = radius_km,
    index_type = index_type,

    # dggridR-compatible fields (for backwards compatibility)
    res = resolution,
    topology_family = topology,
    metric_radius = if (metric) sqrt(area / pi) else NULL,
    pole_lon_deg = ISEA_VERT0_LON_DEG,
    pole_lat_deg = ISEA_VERT0_LAT_DEG,
    azimuth_deg = ISEA_AZIMUTH_DEG,
    aperture_type = "SEQUENCE",
    res_spec = resolution,
    precision = 7
  )
  
  # Set class for method dispatch
  class(grid) <- c("hexify_grid", "dggs", "list")
  
  return(grid)
}


#' Verify grid object
#'
#' Validates that a grid object has all required fields and valid values.
#' This function is called internally by most hexify functions to ensure
#' grid integrity.
#'
#' @param dggs Grid object to verify (from hexify_grid() or hex_grid())
#' @return TRUE (invisibly) if valid, otherwise throws an error
#'
#' @export
#' @examples
#' grid <- hexify_grid(area = 1000, aperture = 3)
#' dgverify(grid)  # Should pass silently
#'
#' # Modern HexGridInfo objects are accepted too
#' dgverify(hex_grid(area_km2 = 1000))
#'
#' # Invalid grid will throw error
#' bad_grid <- list(aperture = 5)
#' try(dgverify(bad_grid))  # Will error
dgverify <- function(dggs) {
  # Accept the modern HexGridInfo (S4) object by converting to the legacy
  # list representation this function checks.
  if (is_hex_grid(dggs)) {
    if (dggs@grid_type == "h3") {
      validate_resolution(dggs@resolution)
      return(invisible(TRUE))
    }
    dggs <- HexGridInfo_to_hexify_grid(dggs)
  }

  # Check object type
  if (!inherits(dggs, "hexify_grid") && !inherits(dggs, "dggs")) {
    stop("dggs must be a grid object from hexify_grid() or hex_grid()")
  }

  # Check required fields exist
  required_fields <- c("aperture", "topology", "projection")
  missing_fields <- setdiff(required_fields, names(dggs))
  if (length(missing_fields) > 0) {
    stop(sprintf("Grid object missing required field(s): %s",
                 paste(missing_fields, collapse = ", ")))
  }
  
  resolution <- get_grid_resolution(dggs, require = TRUE)
  
  # Validate resolution and aperture
  validate_resolution(resolution)
  validate_aperture(dggs$aperture)
  
  # Validate topology
  if (dggs$topology != "HEXAGON") {
    warning("Only HEXAGON topology is fully supported")
  }
  
  # Validate projection
  if (dggs$projection != "ISEA") {
    warning("Only ISEA projection is fully supported")
  }
  
  invisible(TRUE)
}

# tests/testthat/test-grid.R
# Tests for hexify_grid.R - grid construction and validation

# =============================================================================
# GRID CONSTRUCTION
# =============================================================================

test_that("hexify_grid creates valid grid object", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  expect_s3_class(grid, "hexify_grid")
  expect_s3_class(grid, "dggs")
  expect_true("aperture" %in% names(grid))
  expect_true("resolution" %in% names(grid))
  expect_true("area" %in% names(grid))
})

test_that("hexify_grid sets correct aperture", {
  for (ap in c(3, 4, 7)) {
    grid <- hexify_grid(area = 1000, aperture = ap)
    expect_equal(grid$aperture, ap, info = sprintf("aperture %d", ap))
  }
})

test_that("hexify_grid sets correct index_type based on aperture", {
  grid3 <- hexify_grid(area = 1000, aperture = 3)
  expect_equal(grid3$index_type, "z3")

  grid4 <- hexify_grid(area = 1000, aperture = 4)
  expect_equal(grid4$index_type, "zorder")

  grid7 <- hexify_grid(area = 1000, aperture = 7)
  expect_equal(grid7$index_type, "z7")
})

test_that("hexify_grid resolution rounding works", {
  area <- 1000

  grid_nearest <- hexify_grid(area = area, resround = "nearest")
  grid_up <- hexify_grid(area = area, resround = "up")
  grid_down <- hexify_grid(area = area, resround = "down")

  expect_true(grid_up$resolution >= grid_down$resolution)
  expect_true(grid_nearest$resolution >= grid_down$resolution)
  expect_true(grid_nearest$resolution <= grid_up$resolution)
})

test_that("hexify_grid validates topology", {
  expect_error(
    hexify_grid(area = 1000, topology = "TRIANGLE"),
    "HEXAGON"
  )
})

test_that("hexify_grid validates projection", {
  expect_error(
    hexify_grid(area = 1000, projection = "FULLER"),
    "ISEA"
  )
})

test_that("hexify_grid validates aperture", {
  expect_error(
    hexify_grid(area = 1000, aperture = 5),
    "Aperture"
  )
})

test_that("hexify_grid validates resround", {
  expect_error(
    hexify_grid(area = 1000, resround = "invalid"),
    "resround"
  )
})

test_that("hexify_grid resolution stays in valid range", {
  # Very small area -> high resolution (capped at MAX)
  grid_small <- hexify_grid(area = 0.001, aperture = 3)
  expect_lte(grid_small$resolution, MAX_RESOLUTION)

  # Large area -> low resolution (capped at MIN)
  # Use area that still gives valid (non-NaN) resolution
  grid_large <- hexify_grid(area = 1e8, aperture = 3)
  expect_gte(grid_large$resolution, MIN_RESOLUTION)
})

test_that("hexify_grid includes dggridR-compatible fields", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  # Check for dggridR compatibility fields
  expect_true("res" %in% names(grid))
  expect_true("topology_family" %in% names(grid))
  expect_true("pole_lon_deg" %in% names(grid))
  expect_true("pole_lat_deg" %in% names(grid))
  expect_true("azimuth_deg" %in% names(grid))
})

# =============================================================================
# GRID VERIFICATION
# =============================================================================

test_that("dgverify accepts valid grid", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  expect_true(dgverify(grid))
})

test_that("dgverify rejects non-grid objects", {
  expect_error(dgverify("not a grid"), "hexify_grid")
  expect_error(dgverify(list(x = 1)), "hexify_grid")
})

test_that("dgverify checks required fields", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  # Remove required field
  bad_grid <- grid
  bad_grid$aperture <- NULL
  class(bad_grid) <- c("hexify_grid", "dggs", "list")

  expect_error(dgverify(bad_grid), "missing")
})

test_that("dgverify warns on non-HEXAGON topology", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  grid$topology <- "OTHER"

  expect_warning(dgverify(grid), "HEXAGON")
})

test_that("dgverify warns on non-ISEA projection", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  grid$projection <- "OTHER"

  expect_warning(dgverify(grid), "ISEA")
})

# =============================================================================
# PRINT METHOD (uses default list printing now)
# =============================================================================

test_that("print outputs grid information", {
  grid <- hexify_grid(area = 1000, aperture = 3)

  # Default list print shows fields
  expect_output(print(grid), "area")
  expect_output(print(grid), "resolution")
  expect_output(print(grid), "aperture")
})

# =============================================================================
# RESOLUTION CALCULATION
# =============================================================================

test_that("calculate_resolution_for_area is monotonic", {
  areas <- c(10000, 1000, 100, 10, 1)
  resolutions <- sapply(areas, calculate_resolution_for_area, aperture = 3)

  # Smaller area -> higher resolution
  expect_true(all(diff(resolutions) > 0))
})

test_that("calculate_resolution_for_area works for all apertures", {
  area <- 1000

  for (ap in c(3, 4, 7)) {
    res <- calculate_resolution_for_area(area, aperture = ap)
    expect_true(is.finite(res))
    expect_true(res > 0)
  }
})

# R/dggrid.R - DGGRID helper functions
#
# Note: C++ bindings are auto-generated in RcppExports.R.
# This file contains only pure R helper functions.

#' Create DGGRID 43H aperture sequence
#'
#' Create an aperture sequence following DGGRID's 43H pattern:
#' first num_ap4 resolutions use aperture 4, then aperture 3.
#'
#' @param num_ap4 Number of aperture-4 resolutions
#' @param num_ap3 Number of aperture-3 resolutions
#' @return Integer vector of aperture sequence
#'
#' @family 'dggridR' compatibility
#' @keywords internal
#' @export
#' @examples
#' # DGGRID 43H with 2 ap4 resolutions, then 3 ap3 resolutions
#' seq <- dggrid_43h_sequence(2, 3)  # c(4, 4, 3, 3, 3)
dggrid_43h_sequence <- function(num_ap4, num_ap3) {
  c(rep(4L, num_ap4), rep(3L, num_ap3))
}

## ----setup, include = FALSE---------------------------------------------------
CAN_RUN <- requireNamespace("sf", quietly = TRUE)

source("_common.R")
knitr::opts_chunk$set(eval = CAN_RUN)
library(hexify)

## ----h3-grid------------------------------------------------------------------
library(sf)
library(ggplot2)

# Create an H3 grid specification
grid_h3 <- hex_grid(resolution = 5, type = "h3")
grid_h3

## ----h3-hexify----------------------------------------------------------------
# Sample cities
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome",
           "London", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50,
          -0.12, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90,
          51.51, 50.08, 52.23, 47.50, 52.37)
)

result <- hexify(cities, lon = "lon", lat = "lat", grid = grid_h3)
result

## ----h3-cell-ids--------------------------------------------------------------
# Cell IDs are hexadecimal strings
result@cell_id

# All standard accessors work
cells(result)
n_cells(result)

## ----h3-by-area---------------------------------------------------------------
grid_area <- hex_grid(area_km2 = 500, type = "h3")
grid_area

## ----h3-grid-rect, fig.width=7, fig.height=5----------------------------------
# Generate H3 hexagons over Western Europe
grid_h3 <- hex_grid(resolution = 3, type = "h3")
europe_h3 <- grid_rect(c(-10, 35, 25, 60), grid_h3)

# Basemap
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_h3, fill = NA, color = "#E6550D", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("H3 Resolution %d Grid", grid_h3@resolution),
       subtitle = sprintf("~%.0f km² average cell", grid_h3@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE)

## ----h3-grid-clip, fig.width=7, fig.height=7----------------------------------
# Clip H3 grid to France
france <- hexify_world[hexify_world$name == "France", ]
grid_h3 <- hex_grid(resolution = 4, type = "h3")
france_h3 <- grid_clip(france, grid_h3)

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_h3, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = sprintf("H3 Grid Clipped to France (res %d)", grid_h3@resolution)) +
  theme_minimal(base_size = FIG_BASE_SIZE)

## ----h3-parent----------------------------------------------------------------
# Get parent cells (one resolution coarser)
grid_h3 <- hex_grid(resolution = 5, type = "h3")
child_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40),
  lat = c(48.21, 48.86, 52.52),
  grid = grid_h3
)

parent_ids <- get_parent(child_ids, grid_h3, levels = 1)
data.frame(child = child_ids, parent = parent_ids)

## ----h3-children--------------------------------------------------------------
# Get children of a single cell (one resolution finer)
grid_coarse <- hex_grid(resolution = 3, type = "h3")
coarse_id <- lonlat_to_cell(16.37, 48.21, grid_coarse)

children <- get_children(coarse_id, grid_coarse, levels = 1)
cat(length(children[[1]]), "children at resolution", grid_coarse@resolution + 1, "\n")
head(children[[1]])

## ----h3-hierarchy-plot, fig.width=7, fig.height=5-----------------------------
# Parent cell polygon
parent_poly <- cell_to_sf(coarse_id, grid_coarse)

# Children cell polygons
grid_fine <- hex_grid(resolution = 4, type = "h3")
children_poly <- cell_to_sf(children[[1]], grid_fine)

ggplot() +
  geom_sf(data = children_poly, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.5) +
  geom_sf(data = parent_poly, fill = NA, color = "black", linewidth = 1.2) +
  labs(title = sprintf("H3 Hierarchy: 1 parent (res %d)", grid_coarse@resolution),
       subtitle = sprintf("%d children (res %d)",
                          length(children[[1]]), grid_fine@resolution)) +
  theme_minimal(base_size = FIG_BASE_SIZE)

## ----h3-workflow, fig.width=7, fig.height=5.25, message=FALSE, warning=FALSE----
set.seed(42)

# Simulate observations across Europe
obs <- data.frame(
  lon = c(rnorm(200, 10, 12), rnorm(100, 25, 8)),
  lat = c(rnorm(200, 48, 6), rnorm(100, 55, 4)),
  species = sample(c("Sp. A", "Sp. B", "Sp. C"), 300, replace = TRUE)
)
obs$lon <- pmax(-10, pmin(40, obs$lon))
obs$lat <- pmax(35, pmin(65, obs$lat))

# Hexify with H3
grid_h3 <- hex_grid(resolution = 3, type = "h3")
obs_hex <- hexify(obs, lon = "lon", lat = "lat", grid = grid_h3)

# Aggregate: species richness per cell
obs_df <- as.data.frame(obs_hex)
obs_df$cell_id <- obs_hex@cell_id

richness <- aggregate(species ~ cell_id, data = obs_df,
                      FUN = function(x) length(unique(x)))
names(richness)[2] <- "n_species"

# Map it
polys <- cell_to_sf(richness$cell_id, grid_h3)
polys <- merge(polys, richness, by = "cell_id")

europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = polys, aes(fill = n_species), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "plasma", name = "Species\nRichness") +
  coord_sf(xlim = c(-10, 40), ylim = c(35, 65)) +
  labs(title = "Species Richness on H3 Grid",
       subtitle = sprintf("H3 resolution %d (~%.0f km² avg cells)",
                          grid_h3@resolution, grid_h3@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----h3-crosswalk-------------------------------------------------------------
# Start with an ISEA grid and some cells
grid_isea <- hex_grid(resolution = 9, aperture = 3)
isea_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40, -3.70, 12.50),
  lat = c(48.21, 48.86, 52.52, 40.42, 41.90),
  grid = grid_isea
)

# Map ISEA cells to their closest H3 equivalents
xw <- h3_crosswalk(isea_ids, grid_isea)
xw[, c("isea_cell_id", "h3_cell_id", "isea_area_km2", "h3_area_km2")]

## ----h3-resolution-table------------------------------------------------------
h3_res <- hexify_compare_resolutions(type = "h3", res_range = 0:15)
h3_res$n_cells_fmt <- ifelse(
  h3_res$n_cells > 1e9,
  sprintf("%.1fB", h3_res$n_cells / 1e9),
  ifelse(h3_res$n_cells > 1e6,
         sprintf("%.1fM", h3_res$n_cells / 1e6),
         ifelse(h3_res$n_cells > 1e3,
                sprintf("%.1fK", h3_res$n_cells / 1e3),
                as.character(h3_res$n_cells)))
)
# Use scientific notation for tiny areas, fixed for larger ones
h3_res$area_fmt <- ifelse(
  h3_res$cell_area_km2 < 0.1,
  formatC(h3_res$cell_area_km2, format = "e", digits = 1),
  formatC(h3_res$cell_area_km2, format = "f", digits = 1, big.mark = ",")
)
h3_res$spacing_fmt <- ifelse(
  h3_res$cell_spacing_km < 0.1,
  formatC(h3_res$cell_spacing_km, format = "e", digits = 1),
  formatC(h3_res$cell_spacing_km, format = "f", digits = 1)
)
knitr::kable(
  h3_res[, c("resolution", "n_cells_fmt", "area_fmt", "spacing_fmt")],
  col.names = c("Resolution", "# Cells", "Avg Area (km\u00b2)", "Spacing (km)"),
  align = c("r", "r", "r", "r")
)

# R/hex_topology.R
# Pentagon detection and topological queries

#' Detect Pentagon Cells
#'
#' Identifies which cells are pentagons. Any hexagonal tiling of the sphere
#' must contain exactly 12 pentagons (at the icosahedron vertices).
#' Pentagon cells have 5 neighbors instead of 6.
#'
#' @param cell_id Cell IDs to check. Numeric for ISEA, character for H3.
#' @param grid A HexGridInfo or HexData object specifying the grid.
#'
#' @return A logical vector. `TRUE` for pentagon cells, `FALSE` for hexagons.
#'
#' @details
#' **H3 backend:** Uses the vendored H3 `isPentagon` function.
#'
#' **ISEA backend:** The 12 pentagons are located at icosahedron vertices,
#' which are always the (i, j) = (0, 0) cell of their quad. Pentagon status
#' is checked by decoding each input cell ID's own (quad, i, j) coordinates
#' via [cell_to_lonlat()]'s underlying grid math and testing whether i and j
#' are both zero, rather than by re-deriving each vertex's cell ID (the
#' forward direction has aperture-specific quirks -- e.g. aperture 7's
#' substrate/surrogate coordinate distinction -- that make a single fixed
#' formula for "the (0,0) cell ID of quad Q" unreliable across apertures).
#'
#' @seealso [get_neighbors()] for neighbor finding (pentagons have 5 neighbors)
#'
#' @export
#' @examples
#' \donttest{
#' # H3 pentagon detection
#' g <- hex_grid(resolution = 1, type = "h3")
#' cells <- grid_global(g)
#' pent <- is_pentagon(cells$cell_id, g)
#' sum(pent)  # Should be 12
#' }
is_pentagon <- function(cell_id, grid) {
  g <- extract_grid(grid)

  if (is_h3_grid(g)) {
    return(as.logical(cpp_h3_isPentagon(as.character(cell_id))))
  }

  if (g@resolution == 0L) {
    # At resolution 0, all 12 cells (the icosahedron vertices) are pentagons
    return(rep(TRUE, length(cell_id)))
  }

  # Pentagon cells are exactly the (i, j) = (0, 0) cell of each quad.
  qij <- if (is_mixed_aperture(g@aperture)) {
    mixed_cell_qij(cell_id, g@resolution, g@aperture)
  } else {
    cpp_cell_to_quad_ij(cell_id, g@resolution, aperture_to_int(g@aperture))
  }
  qij$i == 0 & qij$j == 0
}

# =============================================================================
# Hierarchical navigation for mixed aperture grids
# =============================================================================
#
# Mixed grids have no DGGRID-standard hierarchical index (DGGRID rejects
# hierarchical indexing for any non-"PURE" aperture), so there is no Z7/Z3-style
# string to port. hexify instead defines the hierarchy geometrically, which is
# the mathematically correct relationship for how these grids are built:
#
# quad_xy_to_ij_mixed() quantizes the projected point once, by a single scalar
# scale, the product of sqrt(aperture) over the levels. A cell at resolution r
# and its parent at r-1 are therefore both direct quantizations of the SAME
# continuous quad-space at different scales -- so a cell's parent is simply the
# coarser cell whose lattice point contains the cell's centre
# (center-containment), with no accumulated per-step rotation to go wrong.
# Parent/children/index all follow from re-projecting a cell centre through the
# validated forward pipeline.
#
# Every helper takes the aperture spelling alongside the resolution, since the
# coarser grid a parent lives on is the same spelling read at that resolution
# (see aperture_at_resolution()).

mixed_ap_seq <- function(aperture, resolution) {
  parse_aperture_seq(aperture_at_resolution(aperture, resolution), resolution)
}

mixed_cell_center <- function(cell_id, resolution, aperture) {
  cpp_cell_to_lonlat_seq(as.numeric(cell_id), mixed_ap_seq(aperture, resolution))
}

mixed_cell_corners <- function(cell_id, resolution, aperture) {
  cpp_cell_to_corners_seq(as.numeric(cell_id), mixed_ap_seq(aperture, resolution))
}

mixed_point_to_cell <- function(lon, lat, resolution, aperture) {
  cpp_lonlat_to_cell_seq(as.numeric(lon), as.numeric(lat),
                         mixed_ap_seq(aperture, resolution))
}

mixed_cell_qij <- function(cell_id, resolution, aperture) {
  cpp_cell_to_quad_ij_seq(as.numeric(cell_id), mixed_ap_seq(aperture, resolution))
}

mixed_qij_cell <- function(quad, i, j, resolution, aperture) {
  cpp_quad_ij_to_cell_seq(as.integer(quad), as.numeric(i), as.numeric(j),
                          mixed_ap_seq(aperture, resolution))
}

#' Maximum (i,j) substrate coordinate for a mixed grid at a resolution.
#' One less than the quad edge coordinate the C++ layer quantizes against.
#' @noRd
mixed_max_dim <- function(resolution, aperture) {
  if (resolution == 0) return(0L)
  as.integer(cpp_ap_seq_edge_dim(mixed_ap_seq(aperture, resolution))) - 1L
}

#' All valid cells within a band of the (i,j) boundary of the ten body quads,
#' plus the poles. Non-nested ISEA seams -- especially near the twelve
#' icosahedron vertices (pentagon points), where a parent's children spread into
#' a polar cap rather than onto a single edge line -- put children a few cells
#' inside the quad edge, so the band has width `w` (not just the edge line). The
#' set is O(w * sqrt(n_cells)), still cheap against the full grid.
#' @noRd
mixed_boundary_cells <- function(resolution, aperture, n_cells, w = 4L) {
  m <- mixed_max_dim(resolution, aperture)
  w <- min(w, as.integer(m) + 1L)
  lo <- 0:(w - 1L)
  hi <- (m - w + 1L):m
  band <- unique(c(lo, hi))            # (i,j) indices near either extreme
  band <- band[band >= 0 & band <= m]
  full <- 0:m
  # Two vertical strips (i in band, any j) and two horizontal strips (j in band).
  gi <- c(rep(band, times = length(full)), rep(full, times = length(band)))
  gj <- c(rep(full, each = length(band)), rep(band, each = length(full)))
  ne <- length(gi)
  q <- rep(1:10, each = ne)
  ii <- rep(gi, times = 10)
  jj <- rep(gj, times = 10)
  cells <- mixed_qij_cell(q, ii, jj, resolution, aperture)
  cells <- c(cells,
             mixed_point_to_cell(0, 90, resolution, aperture),
             mixed_point_to_cell(0, -90, resolution, aperture))
  cells <- unique(cells)
  cells <- cells[is.finite(cells) & cells >= 1 & cells <= n_cells]
  # Keep only (i,j) that round-trip to a real cell.
  if (length(cells) == 0) return(cells)
  q2 <- mixed_cell_qij(cells, resolution, aperture)
  rt <- mixed_qij_cell(q2$quad, q2$i, q2$j, resolution, aperture)
  cells[rt == cells]
}

#' Geometric parent of mixed cells (center-containment)
#' @noRd
mixed_get_parent <- function(cell_id, resolution, aperture, levels = 1L) {
  parent_res <- resolution - as.integer(levels)
  ll <- mixed_cell_center(cell_id, resolution, aperture)
  mixed_point_to_cell(ll$lon_deg, ll$lat_deg, parent_res, aperture)
}

#' Candidate neighbour cells of a set of mixed cells.
#'
#' Two sources unioned for robustness: exact axial (i,j) neighbours in each
#' cell's own quad (correct for quad interiors), plus directional lon/lat probes
#' at the cell spacing (which cross icosahedron seams into adjacent quads). Only
#' used to grow a candidate superset; correctness comes from the parent filter.
#' @noRd
mixed_neighbors <- function(cells, child_res, aperture, spacing_deg, n_cells_child) {
  qij <- mixed_cell_qij(cells, child_res, aperture)
  # 6 axial hex-neighbour offsets
  off <- rbind(c(1, 0), c(0, 1), c(-1, 1), c(-1, 0), c(0, -1), c(1, -1))
  nb <- integer(0)
  for (r in seq_len(nrow(off))) {
    nb <- c(nb, mixed_qij_cell(qij$quad, qij$i + off[r, 1], qij$j + off[r, 2],
                               child_res, aperture))
  }
  # Directional probes at the cell spacing (seam-crossing)
  cll <- mixed_cell_center(cells, child_res, aperture)
  if (is.finite(spacing_deg) && spacing_deg > 0) {
    ang <- seq(0, 2 * pi, length.out = 13)[-13]
    coslat <- cos(cll$lat_deg * pi / 180)
    for (rad in c(0.8, 1.15) * spacing_deg) {
      for (a in ang) {
        plat <- pmax(pmin(cll$lat_deg + rad * sin(a), 89.9999), -89.9999)
        dlon <- ifelse(coslat > 1e-6, rad * cos(a) / coslat, rad * cos(a))
        plon <- ((cll$lon_deg + dlon + 180) %% 360) - 180
        nb <- c(nb, mixed_point_to_cell(plon, plat, child_res, aperture))
      }
    }
  }
  nb <- unique(nb)
  nb <- nb[is.finite(nb) & nb >= 1 & nb <= n_cells_child]
  # Keep only cells whose (i,j) round-trips to a real cell
  if (length(nb) == 0) return(nb)
  q2 <- mixed_cell_qij(nb, child_res, aperture)
  rt <- mixed_qij_cell(q2$quad, q2$i, q2$j, child_res, aperture)
  nb[rt == nb]
}

#' Geometric children of a single mixed cell.
#'
#' Children are cells whose geometric parent (centre re-quantised at the coarser
#' resolution) is this cell -- an exact test. The work is producing a complete
#' candidate superset to run that test on. At coarse resolutions (few, huge
#' cells) every child cell is tested (cheap and exhaustive). Otherwise children
#' form a connected patch, so a breadth-first walk over neighbours, seeded from
#' an (i,j) box and grown only through cells that pass the parent test, visits
#' them all without relying on sampling density. Returns child IDs sorted.
#' @noRd
mixed_get_children_one <- function(cell_id, resolution, child_res, aperture,
                                   n_cells_child) {
  # Coarse levels: exhaustive filter (no locality assumptions, cheap).
  if (n_cells_child <= 2000) {
    all_child <- seq_len(n_cells_child)
    par <- mixed_get_parent(all_child, child_res, aperture, child_res - resolution)
    return(sort(all_child[par == cell_id]))
  }

  spacing_deg <- mixed_spacing_deg(child_res, aperture)
  ll <- mixed_cell_center(cell_id, resolution, aperture)

  # Seed set: (i,j) box around the central child, plus the pole cells.
  central <- mixed_point_to_cell(ll$lon_deg, ll$lat_deg, child_res, aperture)
  cq <- mixed_cell_qij(central, child_res, aperture)
  box <- 3L
  di <- rep(-box:box, times = 2 * box + 1)
  dj <- rep(-box:box, each = 2 * box + 1)
  seed <- mixed_qij_cell(cq$quad, cq$i + di, cq$j + dj, child_res, aperture)
  seed <- c(seed, central,
            mixed_point_to_cell(0, 90, child_res, aperture),
            mixed_point_to_cell(0, -90, child_res, aperture))
  seed <- unique(seed)
  seed <- seed[is.finite(seed) & seed >= 1 & seed <= n_cells_child]

  parent_of <- function(x) {
    mixed_get_parent(x, child_res, aperture, child_res - resolution)
  }

  found <- seed[parent_of(seed) == cell_id]
  if (length(found) == 0) {
    # Fall back to a broader probe if the box seed missed entirely.
    found <- central[parent_of(central) == cell_id]
  }
  visited <- unique(c(seed, found))
  frontier <- found

  # BFS over neighbours, keeping only genuine children.
  for (iter in seq_len(64)) {
    if (length(frontier) == 0) break
    nb <- mixed_neighbors(frontier, child_res, aperture, spacing_deg, n_cells_child)
    nb <- setdiff(nb, visited)
    visited <- c(visited, nb)
    if (length(nb) == 0) break
    kids <- nb[parent_of(nb) == cell_id]
    found <- c(found, kids)
    frontier <- kids
  }

  # Seam children: non-nested ISEA seams (and the twelve icosahedron vertices,
  # which are quad corners) put a few children on a distant quad boundary that
  # the interior walk cannot reach. This can only happen when the PARENT's own
  # footprint touches a seam, i.e. its (i,j) is near a quad edge or corner. For
  # such parents -- an O(sqrt(n)) minority -- sweep the O(sqrt(n)) boundary band
  # and add any cell whose exact parent is this cell. Interior parents keep all
  # children in-quad, so the fast walk above is already complete for them.
  pq <- mixed_cell_qij(cell_id, resolution, aperture)
  m_parent <- mixed_max_dim(resolution, aperture)
  gate <- 3L
  near_boundary <- min(pq$i, pq$j, m_parent - pq$i, m_parent - pq$j) < gate
  if (near_boundary) {
    bnd <- mixed_boundary_cells(child_res, aperture, n_cells_child)
    if (length(bnd) > 0) {
      found <- c(found, bnd[parent_of(bnd) == cell_id])
    }
  }

  sort(unique(found))
}

#' Approximate cell centre spacing in degrees at a resolution, for sizing the
#' child-footprint sampling disk.
#' @noRd
mixed_spacing_deg <- function(resolution, aperture) {
  # Mean cell area (km^2) from the grid's cell count over the body's surface,
  # then hexagon centre spacing s = sqrt(2 A / sqrt(3)) in degrees of arc. Area
  # goes with r^2 and both the spacing and the km per degree with r, so the
  # angle is the same on every body and Earth's figures give it.
  n <- aperture_n_cells(aperture, resolution)
  area_km2 <- EARTH_SURFACE_KM2 / n
  spacing_km <- sqrt(2 * area_km2 / sqrt(3))
  spacing_km / KM_PER_DEGREE
}

# -----------------------------------------------------------------------------
# Hierarchical index string
# -----------------------------------------------------------------------------
# The index is prefix-hierarchical: a 2-digit base cell (the resolution-0 cell,
# 1..12) followed by a 2-digit ordinal per resolution level, giving which child
# of its parent the cell is (children sorted ascending by cell ID). Two digits
# per level because near the twelve icosahedron vertices centre-containment can
# gather more than nine children into one coarse cell. Dropping the last two
# digits yields the parent's index, so the string sorts by spatial ancestry.
mixed_index_digit_width <- 2L

#' Encode a single mixed cell to its hierarchical index string.
#' @noRd
mixed_cell_to_index_one <- function(cell_id, resolution, aperture) {
  if (resolution == 0) {
    return(sprintf("%02d", as.integer(cell_id)))
  }
  # Ancestor chain a[0..resolution]; a[resolution] = cell_id.
  anc <- numeric(resolution + 1)
  anc[resolution + 1] <- cell_id
  for (k in resolution:1) {
    anc[k] <- mixed_get_parent(anc[k + 1], k, aperture, 1L)
  }
  w <- mixed_index_digit_width
  digits <- integer(resolution)
  for (k in 1:resolution) {
    kids <- mixed_get_children_one(anc[k], k - 1L, k, aperture,
                                   aperture_n_cells(aperture, k))
    pos <- match(anc[k + 1], kids)
    if (is.na(pos)) {
      stop("hexify internal error: mixed child enumeration missed a descendant ",
           "(cell ", format(cell_id, scientific = FALSE), ", level ", k, ")")
    }
    digits[k] <- pos - 1L
  }
  if (any(digits >= 10^w)) {
    stop("hexify internal error: mixed child ordinal exceeds index digit width")
  }
  paste0(sprintf("%02d", as.integer(anc[1])),
         paste(sprintf(paste0("%0", w, "d"), digits), collapse = ""))
}

#' Decode a single mixed hierarchical index string to a cell ID.
#' @noRd
mixed_index_to_cell_one <- function(index, resolution, aperture) {
  base0 <- as.integer(substr(index, 1, 2))
  cell <- base0
  if (resolution == 0) return(cell)
  w <- mixed_index_digit_width
  for (k in 1:resolution) {
    start <- 2L + (k - 1L) * w + 1L
    d <- as.integer(substr(index, start, start + w - 1L))
    kids <- mixed_get_children_one(cell, k - 1L, k, aperture,
                                   aperture_n_cells(aperture, k))
    if (d + 1L > length(kids)) {
      stop("hexify internal error: mixed index digit out of range on decode")
    }
    cell <- kids[d + 1L]
  }
  cell
}

# hexify_cell_to_sf.R
# Cell ID to sf polygon conversion and grid generation
#
# This file provides efficient polygon generation from cell IDs,
# with sf integration for modern spatial workflows.

# =============================================================================
# INTERNAL HELPERS
# =============================================================================

#' Get resolution from area (internal)
#' @noRd
.resolution_from_area <- function(area, aperture) {
  grid <- hexify_grid(area = area, aperture = as.integer(aperture),
                      resround = "nearest")
  grid$resolution
}

# =============================================================================
# CORE POLYGON GENERATION
# =============================================================================

#' Convert cell IDs to sf polygons
#'
#' Creates polygon geometries for hexagonal grid cells from their cell IDs.
#' Returns an sf object by default, or a data frame for lightweight workflows.
#'
#' @param cell_id Integer vector of cell identifiers
#' @param resolution Grid resolution level. Can be omitted if grid is provided.
#' @param aperture Grid aperture: 3, 4, or 7. Can be omitted if grid is provided.
#' @param return_sf Logical. If TRUE (default), returns sf object with polygon
#'   geometries. If FALSE, returns data frame with vertex coordinates.
#' @param grid Optional HexGridInfo object. If provided, resolution and aperture
#'   are extracted from it.
#' @param wrap_dateline Logical. If TRUE (default), calls
#'   \code{sf::st_wrap_dateline()} to split antimeridian-crossing polygons.
#'   Set to FALSE for orthographic/globe projections where wrapping creates gaps.
#'
#' @return If return_sf = TRUE: sf object with columns:
#'   \item{cell_id}{Cell identifier}
#'   \item{geometry}{POLYGON geometry (sfc_POLYGON)}
#'
#'   If return_sf = FALSE: data frame with columns:
#'   \item{cell_id}{Cell identifier}
#'   \item{lon}{Vertex longitude}
#'   \item{lat}{Vertex latitude}
#'   \item{order}{Vertex order (1-7, 7 closes the polygon)}
#'
#' @details
#' This function uses a native C++ implementation that is significantly faster
#' than 'dggridR' polygon generation, especially for large numbers of cells.
#'
#' For the recommended S4 interface, use \code{\link{cell_to_sf}} instead.
#'
#' @family sf conversion
#' @seealso \code{\link{cell_to_sf}} for the recommended S4 interface
#' @keywords internal
#' @export
#' @examples
#' library(hexify)
#'
#' # Generate some data with hex cells
#' df <- data.frame(lon = c(0, 5, 10), lat = c(45, 46, 45))
#' result <- hexify(df, lon = "lon", lat = "lat", area = 1000)
#'
#' # Get polygons as sf object (using HexData)
#' polys <- cell_to_sf(grid = result)
#'
#' # Or with explicit parameters
#' polys <- hexify_cell_to_sf(result@cell_id, resolution = 10, aperture = 3)
#'
#' # Plot with sf
#' library(sf)
#' plot(st_geometry(polys), col = "lightblue", border = "blue")
hexify_cell_to_sf <- function(cell_id, resolution = NULL, aperture = NULL,
                              return_sf = TRUE, grid = NULL,
                              wrap_dateline = TRUE) {

  # Extract from grid if provided
  if (!is.null(grid)) {
    g <- extract_grid(grid)
    # H3 grids: redirect to cell_to_sf() which handles H3 boundaries
    if (is_h3_grid(g)) {
      return(cell_to_sf(cell_id, g))
    }
    resolution <- g@resolution
    aperture <- as.integer(g@aperture)
  }

  # Input validation
  if (!is.numeric(cell_id)) {
    stop("cell_id must be numeric (integer cell IDs)")
  }
  if (is.null(resolution) || is.null(aperture)) {
    stop("resolution and aperture must be provided, or supply a grid object")
  }
  if (!aperture %in% c(3L, 4L, 7L)) {
    stop("aperture must be 3, 4, or 7")
  }
  if (resolution < 0 || resolution > 30) {
    stop("resolution must be between 0 and 30")
  }

  # Remove NA values and duplicates
  cell_id <- unique(cell_id[!is.na(cell_id)])
  if (length(cell_id) == 0) {
    stop("No valid cell_id values provided")
  }

  aperture <- as.integer(aperture)
  resolution <- as.integer(resolution)

  if (return_sf) {
    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("Package 'sf' is required for return_sf = TRUE. ",
           "Install with: install.packages('sf')")
    }

    sfc <- isea_cells_to_sfc(cell_id, resolution, aperture)
    result_sf <- sf::st_sf(cell_id = cell_id, geometry = sfc)
    if (wrap_dateline) {
      result_sf <- sf::st_wrap_dateline(result_sf,
        options = c("WRAPDATELINE=YES", "DATELINEOFFSET=180"), quiet = TRUE)
    }
    result_sf

  } else {
    # Return data frame format
    result <- cpp_cell_to_polygon(cell_id, resolution, aperture)
    names(result) <- c("cell_id", "lon", "lat", "order")
    result$cell_id <- as.integer(result$cell_id)
    result
  }
}


# =============================================================================
# GRID GENERATION
# =============================================================================

#' Generate a rectangular grid of hexagon polygons
#'
#' Creates hexagon polygons covering a rectangular geographic region.
#'
#' @param minlon,maxlon Longitude bounds
#' @param minlat,maxlat Latitude bounds
#' @param area Target cell area in km^2
#' @param aperture Grid aperture: 3, 4, or 7
#' @param resround Resolution rounding: "nearest", "up", or "down"
#' @param radius_km Radius of the body, in kilometers, or a body name such as
#'   "mars" (default Earth). See \code{\link{hex_grid}}.
#'
#' @return sf object with hexagon polygons covering the specified region
#'
#' @family sf conversion
#' @seealso \code{\link{grid_rect}} for the recommended S4 interface,
#'   \code{\link{hexify_grid_global}} for global grids
#' @keywords internal
#' @export
#' @examples
#' library(hexify)
#' library(sf)
#'
#' grid <- hexify_grid_rect(
#'   minlon = -10, maxlon = 20,
#'   minlat = 35, maxlat = 60,
#'   area = 5000
#' )
#' plot(st_geometry(grid), border = "gray")
hexify_grid_rect <- function(minlon, maxlon, minlat, maxlat,
                             area, aperture = 3L, resround = "nearest",
                             radius_km = EARTH_RADIUS_KM) {

  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required. Install with: install.packages('sf')")
  }

  radius_km <- resolve_radius_km(radius_km)

  # Create grid of sample points
  diagonal <- sqrt(area * 2 / sqrt(3))
  spacing_deg <- diagonal / km_per_degree(radius_km) * 0.8

  lons <- seq(minlon, maxlon, by = spacing_deg)
  lats <- seq(minlat, maxlat, by = spacing_deg)
  grid_pts <- expand.grid(lon = lons, lat = lats)

  # Assign to hexes and get polygons
  result <- hexify(grid_pts, lon = "lon", lat = "lat", area_km2 = area,
                   aperture = aperture, resround = resround,
                   radius_km = radius_km)

  cell_to_sf(grid = result)
}

#' Generate a global grid of hexagon polygons
#'
#' Creates hexagon polygons covering a whole body.
#'
#' @param area Target cell area in km^2
#' @param aperture Grid aperture: 3, 4, or 7
#' @param resround Resolution rounding: "nearest", "up", or "down"
#' @param radius_km Radius of the body, in kilometers, or a body name such as
#'   "mars" (default Earth). See \code{\link{hex_grid}}.
#'
#' @return sf object with hexagon polygons covering the globe
#'
#' @family sf conversion
#' @seealso \code{\link{grid_global}} for the recommended S4 interface,
#'   \code{\link{hexify_grid_rect}} for regional grids
#' @keywords internal
#' @export
#' @examples
#' library(hexify)
#' library(sf)
#'
#' # Coarse global grid (~100,000 km^2 cells)
#' global_grid <- hexify_grid_global(area = 100000)
#' plot(st_geometry(global_grid), border = "gray")
hexify_grid_global <- function(area, aperture = 3L, resround = "nearest",
                               radius_km = EARTH_RADIUS_KM) {

  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required. Install with: install.packages('sf')")
  }

  radius_km <- resolve_radius_km(radius_km)

  # Estimate cell count
  approx_cells <- body_surface_km2(radius_km) / area

  if (approx_cells > 100000) {
    warning(sprintf(
      "This will generate approximately %.0f cells. Consider larger area.",
      approx_cells
    ))
  }

  # Generate dense sample points
  diagonal <- sqrt(area * 2 / sqrt(3))
  spacing_deg <- diagonal / km_per_degree(radius_km) * 0.7

  lons <- seq(-180, 180, by = spacing_deg)
  lats <- seq(-85, 85, by = spacing_deg)
  grid_pts <- expand.grid(lon = lons, lat = lats)

  result <- hexify(grid_pts, lon = "lon", lat = "lat", area_km2 = area,
                   aperture = aperture, resround = resround,
                   radius_km = radius_km)

  cell_to_sf(grid = result)
}

# =============================================================================
# LOW-LEVEL SF HELPERS
# =============================================================================

#' Build an sf POLYGON from six (lon, lat) corner pairs
#'
#' Low-level helper to create a single hexagon polygon from corner coordinates.
#' Most users should use \code{\link{cell_to_sf}} instead.
#'
#' @param lon numeric vector of length 6 (longitude)
#' @param lat numeric vector of length 6 (latitude)
#' @param crs integer CRS (default 4326)
#' @return sf object with one POLYGON geometry
#'
#' @family sf conversion
#' @keywords internal
#' @export
hex_corners_to_sf <- function(lon, lat, crs = 4326) {
  stopifnot(length(lon) == 6L, length(lat) == 6L)
  pts <- cbind(lon, lat)
  ring <- rbind(pts, pts[1, , drop = FALSE])  # close polygon
  sf::st_sf(geometry = sf::st_sfc(sf::st_polygon(list(ring)), crs = crs))
}

# tests/testthat/test-polygons.R
# Tests for polygon generation functions
#
# Functions tested:
# - hexify_cell_to_sf()
# - cell_to_sf()
# - hexify_grid_rect()
# - hexify_grid_global()
# - hex_corners_to_sf()

# =============================================================================
# HEXIFY_CELL_TO_SF
# =============================================================================

test_that("hexify_cell_to_sf returns data frame with return_sf=FALSE", {
  hex_ids <- c(12847, 12532, 22178)

  result <- hexify_cell_to_sf(hex_ids, resolution = 10, aperture = 3, return_sf = FALSE)

  expect_s3_class(result, "data.frame")
  expect_true(all(c("cell_id", "lon", "lat", "order") %in% names(result)))
})

test_that("hexify_cell_to_sf returns 7 vertices per hexagon (closed polygon)", {
  hex_ids <- c(12847, 12532, 22178)

  result <- hexify_cell_to_sf(hex_ids, resolution = 10, aperture = 3, return_sf = FALSE)

  expect_equal(nrow(result), length(hex_ids) * 7)
  expect_equal(unique(result$order), 1:7)
})

test_that("hexify_cell_to_sf returns 6 vertices for a cell at an icosahedral vertex", {
  result <- hexify_cell_to_sf(1, resolution = 3, aperture = 3, return_sf = FALSE)

  expect_equal(nrow(result), 6)
  expect_equal(result$order, 1:6)
})

test_that("hexify_cell_to_sf returns valid coordinates", {
  hex_ids <- c(12847, 12532)

  result <- hexify_cell_to_sf(hex_ids, resolution = 10, aperture = 3, return_sf = FALSE)

  expect_true(all(result$lon >= -180 & result$lon <= 180))
  expect_true(all(result$lat >= -90 & result$lat <= 90))
})

test_that("hexify_cell_to_sf produces closed polygons", {
  hex_ids <- c(12847)

  result <- hexify_cell_to_sf(hex_ids, resolution = 10, aperture = 3, return_sf = FALSE)

  first_vertex <- result[result$order == 1, c("lon", "lat")]
  last_vertex <- result[result$order == 7, c("lon", "lat")]

  expect_equal(first_vertex$lon, last_vertex$lon)
  expect_equal(first_vertex$lat, last_vertex$lat)
})

test_that("hexify_cell_to_sf returns sf object with return_sf=TRUE", {
  skip_if_not_installed("sf")

  hex_ids <- c(12847, 12532, 22178)

  result <- hexify_cell_to_sf(hex_ids, resolution = 10, aperture = 3, return_sf = TRUE)

  expect_s3_class(result, "sf")
  expect_true("cell_id" %in% names(result))
  expect_true("geometry" %in% names(result))
  expect_equal(sf::st_crs(result)$epsg, 4326)

  geom_types <- sf::st_geometry_type(result)
  expect_true(all(geom_types == "POLYGON"))
})

test_that("hexify_cell_to_sf removes duplicates", {
  hex_ids <- c(12847, 12847, 12532)

  result <- hexify_cell_to_sf(hex_ids, resolution = 10, aperture = 3, return_sf = FALSE)

  expect_equal(length(unique(result$cell_id)), 2)
  expect_equal(nrow(result), 2 * 7)
})

test_that("hexify_cell_to_sf handles NA values", {
  hex_ids <- c(12847, NA, 12532)

  result <- hexify_cell_to_sf(hex_ids, resolution = 10, aperture = 3, return_sf = FALSE)

  expect_equal(length(unique(result$cell_id)), 2)
})

test_that("hexify_cell_to_sf validates aperture", {
  expect_error(
    hexify_cell_to_sf(c(12847), resolution = 10, aperture = 5),
    "aperture must be 3, 4, or 7"
  )
})

test_that("hexify_cell_to_sf validates resolution", {
  expect_error(
    hexify_cell_to_sf(c(12847), resolution = -1, aperture = 3),
    "resolution must be between 0 and 30"
  )

  expect_error(
    hexify_cell_to_sf(c(12847), resolution = 31, aperture = 3),
    "resolution must be between 0 and 30"
  )
})

test_that("hexify_cell_to_sf works with aperture 4", {
  hex_ids <- c(100, 200, 300)

  result <- hexify_cell_to_sf(hex_ids, resolution = 8, aperture = 4, return_sf = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 3 * 7)
})

test_that("hexify_cell_to_sf works with aperture 7", {
  hex_ids <- c(100, 200, 300)

  result <- hexify_cell_to_sf(hex_ids, resolution = 5, aperture = 7, return_sf = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 3 * 7)
})

# =============================================================================
# CELL_TO_SF (grid helper)
# =============================================================================

test_that("cell_to_sf works with HexGrid object", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 1000)
  hex_ids <- c(12847, 12532, 22178)

  result <- cell_to_sf(hex_ids, grid)

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), 3)
})

test_that("cell_to_sf works with HexData object (extracts unique cells)", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 5, 10), lat = c(45, 46, 47))
  hex_data <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  # When called with HexData and no cell_id, it uses the cells from data
  result <- cell_to_sf(grid = hex_data)

  expect_s3_class(result, "sf")
  expect_true(nrow(result) >= 1)
})

# =============================================================================
# HEXIFY_GRID_RECT
# =============================================================================

test_that("hexify_grid_rect generates grid for rectangular region", {
  skip_on_cran()  # Generates many polygons
  skip_if_not_installed("sf")

  grid <- hexify_grid_rect(
    minlon = 0, maxlon = 5,
    minlat = 45, maxlat = 48,
    area = 5000
  )

  expect_s3_class(grid, "sf")
  expect_true(nrow(grid) > 0)

  geom_types <- sf::st_geometry_type(grid)
  expect_true(all(geom_types == "POLYGON"))
})

# =============================================================================
# HEX_CORNERS_TO_SF
# =============================================================================

test_that("hex_corners_to_sf builds valid polygon", {
  skip_if_not_installed("sf")

  lon <- c(0, 1, 1, 0, -1, -1)
  lat <- c(0, 0.5, 1, 1, 0.5, 0)

  poly <- hex_corners_to_sf(lon, lat)

  expect_s3_class(poly, "sf")
  expect_true(sf::st_is_valid(poly))
  expect_equal(nrow(poly), 1L)
  expect_identical(as.character(sf::st_geometry_type(poly)), "POLYGON")
})

test_that("hex_corners_to_sf closes polygon correctly", {
  skip_if_not_installed("sf")

  lon <- c(0, 1, 1, 0, -1, -1)
  lat <- c(0, 0.5, 1, 1, 0.5, 0)

  poly <- hex_corners_to_sf(lon, lat)

  coords <- sf::st_coordinates(poly)
  expect_true(all(c("X", "Y") %in% colnames(coords)))

  # XY must match provided points + closing vertex
  xy_expected <- rbind(cbind(lon, lat), c(lon[1], lat[1]))
  actual_xy <- unname(as.matrix(coords[, c("X", "Y"), drop = FALSE]))
  expected_xy <- unname(as.matrix(xy_expected))

  expect_equal(actual_xy, expected_xy, tolerance = 0)
})

test_that("hex_corners_to_sf validates input lengths", {
  expect_error(
    hex_corners_to_sf(c(1, 2, 3), c(1, 2, 3, 4, 5, 6)),
    ""
  )
})

# =============================================================================
# ADDITIONAL INPUT VALIDATION
# =============================================================================

test_that("hexify_cell_to_sf validates cell_id is numeric", {
  expect_error(
    hexify_cell_to_sf("not_numeric", resolution = 10, aperture = 3),
    "cell_id must be numeric"
  )
})

test_that("hexify_cell_to_sf errors on empty input", {
  expect_error(
    hexify_cell_to_sf(numeric(0), resolution = 10, aperture = 3),
    "No valid cell_id values"
  )
})

test_that("hexify_cell_to_sf errors on all NA input", {
  expect_error(
    hexify_cell_to_sf(as.numeric(c()), resolution = 10, aperture = 3),
    "No valid cell_id values"
  )
})

# =============================================================================
# AS_SF METHOD
# =============================================================================

test_that("as_sf creates point geometry from HexData object", {
  skip_if_not_installed("sf")

  df <- data.frame(
    name = c("A", "B"),
    lon = c(10.5, 11.2),
    lat = c(48.5, 49.1)
  )

  hex_data <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
  result <- as_sf(hex_data, geometry = "point")

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), 2)
  expect_true(all(sf::st_geometry_type(result) == "POINT"))
})

test_that("as_sf creates polygon geometry from HexData object", {
  skip_if_not_installed("sf")

  df <- data.frame(
    name = c("A", "B"),
    lon = c(10.5, 11.2),
    lat = c(48.5, 49.1)
  )

  hex_data <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
  result <- as_sf(hex_data, geometry = "polygon")

  expect_s3_class(result, "sf")
  expect_true(all(sf::st_geometry_type(result) == "POLYGON"))
})

# =============================================================================
# RESOLUTION FROM AREA HELPER
# =============================================================================

test_that(".resolution_from_area returns valid resolution", {
  res <- hexify:::.resolution_from_area(1000, aperture = 3)
  expect_true(is.numeric(res))
  expect_true(res >= 0 && res <= 30)
})

# =============================================================================
# HEXIFY_GRID_RECT
# =============================================================================

test_that("hexify_grid_rect validates parameters", {
  skip_if_not_installed("sf")

  # Test basic functionality was already done, test different apertures
  grid_ap4 <- hexify_grid_rect(
    minlon = 0, maxlon = 5,
    minlat = 45, maxlat = 48,
    area = 5000, aperture = 4
  )

  expect_s3_class(grid_ap4, "sf")
  expect_true(nrow(grid_ap4) > 0)
})

test_that("hexify_grid_rect works with resround", {
  skip_if_not_installed("sf")

  grid_up <- hexify_grid_rect(
    minlon = 0, maxlon = 5,
    minlat = 45, maxlat = 48,
    area = 5000, resround = "up"
  )

  grid_down <- hexify_grid_rect(
    minlon = 0, maxlon = 5,
    minlat = 45, maxlat = 48,
    area = 5000, resround = "down"
  )

  expect_s3_class(grid_up, "sf")
  expect_s3_class(grid_down, "sf")
})

# =============================================================================
# HEXIFY_GRID_GLOBAL
# =============================================================================

test_that("hexify_grid_global works with large area", {
  skip_on_cran()  # Generates many cells
  skip_if_not_installed("sf")

  # Use very large area to avoid warning
  grid <- hexify_grid_global(area = 10000000)

  expect_s3_class(grid, "sf")
  expect_true(nrow(grid) > 0)
})

test_that("hexify_grid_global warns on small area", {
  skip_on_cran()  # Slow computation
  skip_if_not_installed("sf")

  expect_warning(
    hexify_grid_global(area = 1000),
    "approximately.*cells"
  )
})

# =============================================================================
# EDGE CASES FOR HEXIFY_CELL_TO_SF
# =============================================================================

test_that("hexify_cell_to_sf works with large cell IDs", {
  skip_if_not_installed("sf")

  # Higher resolution -> larger cell IDs
  hex_ids <- c(100000, 200000, 300000)

  result <- hexify_cell_to_sf(
    hex_ids, resolution = 12, aperture = 3, return_sf = TRUE
  )

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), 3)
})

test_that("hexify_cell_to_sf handles duplicate IDs correctly", {
  skip_if_not_installed("sf")

  hex_ids <- c(100, 100, 200, 200, 200)

  result <- hexify_cell_to_sf(
    hex_ids, resolution = 5, aperture = 3, return_sf = TRUE
  )

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), 2)  # Should be deduplicated
})

test_that("hexify_cell_to_sf with return_sf=FALSE produces correct structure", {
  hex_ids <- c(100, 200)

  result <- hexify_cell_to_sf(
    hex_ids, resolution = 5, aperture = 3, return_sf = FALSE
  )

  expect_s3_class(result, "data.frame")
  expect_true(all(c("cell_id", "lon", "lat", "order") %in% names(result)))
  expect_equal(nrow(result), 2 * 7)  # 7 vertices per hex
})

# =============================================================================
# AS_SF ADDITIONAL TESTS
# =============================================================================

test_that("as_sf preserves all columns", {
  skip_if_not_installed("sf")

  df <- data.frame(
    name = c("A", "B"),
    lon = c(10.5, 11.2),
    lat = c(48.5, 49.1),
    custom_col = c("x", "y")
  )

  hex_data <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
  result <- as_sf(hex_data, geometry = "point")

  expect_true("name" %in% names(result))
  expect_true("custom_col" %in% names(result))
})

test_that("as_sf polygon geometry preserves attributes", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(10.5, 11.2),
    lat = c(48.5, 49.1),
    value = c(100, 200)
  )

  hex_data <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
  result <- as_sf(hex_data, geometry = "polygon")

  expect_s3_class(result, "sf")
  expect_true("value" %in% names(result))
})

# =============================================================================
# CELL_TO_SF DEDUPLICATION
# =============================================================================

test_that("cell_to_sf deduplicates cell_ids", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 1000)
  hex_ids <- c(12847, 12847, 12532)  # duplicate

  result <- cell_to_sf(hex_ids, grid)

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), 2)  # Should deduplicate
})

# =============================================================================
# HEXIFY_CELL_TO_SF WITH GRID PARAMETER
# =============================================================================

test_that("hexify_cell_to_sf extracts parameters from grid object", {
  skip_if_not_installed("sf")

  grid <- hex_grid(area_km2 = 1000)
  hex_ids <- c(12847, 12532)

  # Pass grid instead of resolution/aperture

  result <- hexify_cell_to_sf(hex_ids, grid = grid)

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), 2)
})

test_that("hexify_cell_to_sf errors when resolution/aperture missing and no grid", {
  expect_error(
    hexify_cell_to_sf(c(12847), resolution = 10),
    "resolution and aperture must be provided"
  )

  expect_error(
    hexify_cell_to_sf(c(12847), aperture = 3),
    "resolution and aperture must be provided"
  )
})

# R/hexify_neighbors.R
# Neighbor finding for ISEA and H3 grids
#
# get_neighbors() returns the k-ring (disk) of neighboring cells
# around a given cell ID.

#' Get Neighboring Cells
#'
#' Returns the k-ring (disk) of cells neighboring the input cells.
#' For `k = 1`, returns the immediate 6 neighbors (5 for pentagons).
#' For `k > 1`, returns all cells within `k` grid hops.
#'
#' @param cell_id Cell IDs to find neighbors for. Numeric vector for ISEA
#'   grids, character vector for H3 grids.
#' @param grid A HexGridInfo or HexData object specifying the grid.
#' @param k Integer. Ring distance (default 1). `k = 1` returns immediate
#'   neighbors, `k = 2` includes neighbors-of-neighbors, etc.
#' @param include_self Logical. If `TRUE`, include the input cell in the
#'   result (default `FALSE`).
#' @param distances Logical. If `TRUE`, return a data.frame with cell IDs
#'   and their ring distance from the origin (default `FALSE`).
#'
#' @return If `distances = FALSE` (default): a list of cell ID vectors, one
#'   per input cell. If `distances = TRUE`: a list of data.frames with columns
#'   `cell_id` and `ring_distance`.
#'
#' @details
#' For **ISEA grids**, neighbors are computed using axial coordinate offsets
#' in the quad IJ space. Cells at quad boundaries are handled by reprojection
#' through lon/lat coordinates.
#'
#' For **H3 grids**, neighbors use the vendored H3 `gridDisk` /
#' `gridDiskDistances` functions.
#'
#' Pentagon cells (the 12 icosahedron vertices) have only 5 neighbors instead
#' of the usual 6.
#'
#' @seealso [hexify()] for creating HexData objects,
#'   [hex_distance()] for grid distances between cells
#'
#' @export
#' @examples
#' \donttest{
#' # ISEA grid neighbors
#' g <- hex_grid(area_km2 = 1000)
#' cell <- lonlat_to_cell(10, 50, g)
#' nbrs <- get_neighbors(cell, g)
#' nbrs[[1]]
#'
#' # H3 grid neighbors
#' g_h3 <- hex_grid(resolution = 5, type = "h3")
#' cell_h3 <- lonlat_to_cell(10, 50, g_h3)
#' get_neighbors(cell_h3, g_h3, k = 2)
#'
#' # With distances
#' get_neighbors(cell, g, k = 2, distances = TRUE)
#' }
get_neighbors <- function(cell_id, grid, k = 1L, include_self = FALSE,
                           distances = FALSE) {
  g <- extract_grid(grid)
  k <- as.integer(k)

  if (k < 0L) stop("k must be a non-negative integer")
  if (k == 0L) {
    if (include_self) {
      if (distances) {
        return(lapply(cell_id, function(cid) {
          data.frame(cell_id = cid, ring_distance = 0L)
        }))
      }
      return(as.list(cell_id))
    }
    n <- length(cell_id)
    if (distances) {
      empty <- data.frame(cell_id = cell_id[0], ring_distance = integer(0))
      return(rep(list(empty), n))
    }
    return(rep(list(cell_id[0]), n))
  }

  if (is_h3_grid(g)) {
    .get_neighbors_h3(cell_id, g, k, include_self, distances)
  } else {
    .get_neighbors_isea(cell_id, g, k, include_self, distances)
  }
}

# --- H3 backend ---
.get_neighbors_h3 <- function(cell_id, grid, k, include_self, distances) {
  if (distances) {
    result <- cpp_h3_gridDiskDistances(cell_id, k)
    if (!include_self) {
      result <- lapply(seq_along(result), function(i) {
        df <- result[[i]]
        if (nrow(df) > 0) {
          df[df$ring_distance > 0L, , drop = FALSE]
        } else {
          df
        }
      })
    }
    return(result)
  }

  result <- cpp_h3_gridDisk(cell_id, k)
  if (!include_self) {
    result <- mapply(function(nbrs, origin) {
      nbrs[nbrs != origin]
    }, result, cell_id, SIMPLIFY = FALSE, USE.NAMES = FALSE)
  }
  result
}

# --- ISEA backend ---
.get_neighbors_isea <- function(cell_id, grid, k, include_self, distances) {
  ap <- as.integer(grid@aperture)
  res <- grid@resolution

  # Use k=1 fast path helper (C++ handles all apertures including ap7)
  get_k1 <- function(cids) cpp_get_neighbors_isea(cids, res, ap)

  if (k == 1L && !distances) {
    result <- get_k1(cell_id)
    if (include_self) {
      result <- mapply(function(nbrs, origin) c(origin, nbrs),
                        result, cell_id, SIMPLIFY = FALSE, USE.NAMES = FALSE)
    }
    return(result)
  }

  # General k-ring: iterative BFS expansion
  lapply(cell_id, function(origin) {
    visited <- origin
    ring_dist <- 0L
    current_ring <- origin

    for (ring in seq_len(k)) {
      all_nbrs <- get_k1(current_ring)
      new_cells <- unique(unlist(all_nbrs))
      new_cells <- setdiff(new_cells, visited)

      if (length(new_cells) == 0) break
      visited <- c(visited, new_cells)
      ring_dist <- c(ring_dist, rep(ring, length(new_cells)))
      current_ring <- new_cells
    }

    if (!include_self) {
      keep <- ring_dist > 0L
      visited <- visited[keep]
      ring_dist <- ring_dist[keep]
    }

    if (distances) {
      data.frame(cell_id = visited, ring_distance = ring_dist)
    } else {
      visited
    }
  })
}

# R/constants.R - Package-wide constants and helpers
#
# Centralizes magic numbers to avoid duplication and improve maintainability.

# =============================================================================
# Body Geometry Constants
# =============================================================================

#' Total Earth surface area in square kilometers ('WGS84' ellipsoid)
#' @noRd
EARTH_SURFACE_KM2 <- 510065621.724078904704516

#' Mean Earth radius in kilometers ('WGS84' mean radius)
#' Calculated as (2*a + b) / 3 where a = 6378.137 km (equatorial) and b = 6356.7523142 km (polar)
#' @noRd
EARTH_RADIUS_KM <- 6371.0088

#' Approximate km per degree of latitude (at equator)
#' @noRd
KM_PER_DEGREE <- 111.0

#' Mean radii of solar system bodies, in kilometers
#'
#' 'IAU' mean radii (Archinal et al. 2018, Report of the 'IAU' Working Group on
#' Cartographic Coordinates and Rotational Elements: 2015) as tabulated by 'JPL'
#' Solar System Dynamics. "earth" carries the 'WGS84' mean radius instead, so a
#' grid built by name matches one built from the package default.
#' @noRd
BODY_RADII_KM <- c(
  mercury   = 2439.4,
  venus     = 6051.8,
  earth     = EARTH_RADIUS_KM,
  moon      = 1737.4,
  mars      = 3389.50,
  ceres     = 469.7,
  jupiter   = 69911,
  io        = 1821.49,
  europa    = 1560.80,
  ganymede  = 2631.20,
  callisto  = 2410.30,
  saturn    = 58232,
  enceladus = 252.10,
  titan     = 2574.76,
  uranus    = 25362,
  neptune   = 24622,
  pluto     = 1188.3
)

# =============================================================================
# Body Geometry Helpers
# =============================================================================

#' Radius of a body in km, from a number or a name
#'
#' @param radius_km Positive number, or a name from BODY_RADII_KM
#' @return Radius in kilometers
#' @noRd
resolve_radius_km <- function(radius_km) {
  if (is.character(radius_km)) {
    if (length(radius_km) != 1L || is.na(radius_km)) {
      stop("radius_km must be a single body name or a positive number of kilometers")
    }
    key <- tolower(trimws(radius_km))
    if (!key %in% names(BODY_RADII_KM)) {
      stop(sprintf(
        "Unknown body \"%s\". Named bodies are: %s. Any other body takes its mean radius in km.",
        radius_km, paste(names(BODY_RADII_KM), collapse = ", ")
      ))
    }
    return(unname(BODY_RADII_KM[[key]]))
  }

  if (!is.numeric(radius_km) || length(radius_km) != 1L || is.na(radius_km) ||
      !is.finite(radius_km) || radius_km <= 0) {
    stop("radius_km must be a single positive number of kilometers, or a body name such as \"mars\"")
  }

  as.numeric(radius_km)
}

#' Surface area of a body in km^2
#'
#' The area of a sphere of the given radius. Earth's radius returns the 'WGS84'
#' ellipsoid area, which is what an Earth grid is sized against; an ellipsoid
#' has a slightly different area from the sphere of its mean radius, so the two
#' part in the seventh significant figure.
#' @param radius_km Radius in kilometers
#' @noRd
body_surface_km2 <- function(radius_km) {
  if (radius_km == EARTH_RADIUS_KM) return(EARTH_SURFACE_KM2)
  4 * pi * radius_km^2
}

#' Kilometers per degree of arc at a body's radius
#' @param radius_km Radius in kilometers
#' @noRd
km_per_degree <- function(radius_km) {
  KM_PER_DEGREE * radius_km / EARTH_RADIUS_KM
}

#' Radius a grid is sized against, in km
#'
#' A grid carrying neither the slot nor the field is an Earth grid.
#' @param x HexGridInfo object or legacy hexify_grid list
#' @noRd
grid_radius_km <- function(x) {
  r <- if (isS4(x)) {
    if (.hasSlot(x, "radius_km")) x@radius_km else EARTH_RADIUS_KM
  } else {
    x$radius_km
  }
  if (is.null(r) || length(r) != 1L || is.na(r)) EARTH_RADIUS_KM else as.numeric(r)
}

#' Is this grid sized against Earth?
#' @param x HexGridInfo object or legacy hexify_grid list
#' @noRd
is_earth_grid <- function(x) {
  grid_radius_km(x) == EARTH_RADIUS_KM
}

# =============================================================================
# Body CRS Helpers
# =============================================================================

#' Longlat CRS on the sphere of a body's radius, as a 'PROJ' string
#'
#' 'EPSG' codes name Earth reference systems, so a grid on another body carries
#' a longlat CRS on the sphere of its own radius instead.
#' @param radius_km Radius in kilometers
#' @noRd
body_crs_string <- function(radius_km) {
  metres <- sub("[.]?0+$", "", sprintf("%.4f", radius_km * 1000))
  sprintf("+proj=longlat +R=%s +no_defs", metres)
}

#' A CRS specification read by sf, or NA_crs_ if sf cannot read it
#' @param crs An 'EPSG' code or a 'PROJ' or 'WKT' string
#' @noRd
parse_crs <- function(crs) {
  tryCatch(sf::st_crs(crs), error = function(e) sf::NA_crs_)
}

#' The CRS a new grid stores
#'
#' NULL takes 'WGS84' on Earth and the body's own sphere elsewhere. An 'EPSG'
#' code stores as an integer, any other CRS as the string sf reads it from.
#' @param crs NULL, an 'EPSG' code, or a 'PROJ' or 'WKT' string
#' @param radius_km Radius the grid is sized against, in kilometers
#' @noRd
resolve_crs <- function(crs, radius_km) {
  if (is.null(crs)) {
    if (radius_km == EARTH_RADIUS_KM) return(4326L)
    return(body_crs_string(radius_km))
  }

  if (is.character(crs)) {
    if (length(crs) != 1L || is.na(crs) || !nzchar(crs)) {
      stop("crs must be a single non-empty CRS string, or an EPSG code")
    }
    if (is.na(parse_crs(crs))) {
      stop(sprintf(
        "crs \"%s\" is not a coordinate reference system sf can read", crs
      ))
    }
    return(crs)
  }

  if (!is.numeric(crs) || length(crs) != 1L || is.na(crs) ||
      !is.finite(crs) || crs <= 0) {
    stop("crs must be a single positive EPSG code, or a PROJ or WKT string")
  }

  as.integer(crs)
}

#' CRS a grid's coordinates are read in
#'
#' A grid carrying no CRS is read on its own body: 'WGS84' on Earth, and a
#' longlat CRS on the sphere of the radius elsewhere.
#' @param x HexGridInfo object or legacy hexify_grid list
#' @noRd
grid_crs <- function(x) {
  crs <- if (isS4(x)) {
    if (.hasSlot(x, "crs")) x@crs else NULL
  } else {
    x$crs
  }
  if (is.null(crs) || length(crs) != 1L || is.na(crs)) {
    crs <- resolve_crs(NULL, grid_radius_km(x))
  }
  parse_crs(crs)
}

#' A grid's CRS as 'WKT', for the packages that take a string
#' @param x HexGridInfo object or legacy hexify_grid list
#' @noRd
grid_crs_wkt <- function(x) {
  grid_crs(x)$wkt
}

#' A CRS as it prints in a grid specification
#' @param crs An 'EPSG' code or a 'PROJ' or 'WKT' string
#' @noRd
format_crs <- function(crs) {
  if (is.character(crs)) crs else sprintf("EPSG:%d", as.integer(crs))
}

#' 'H3' areas read on another body
#'
#' A cell covers the same solid angle on any sphere, and 'H3' reports an area as
#' that solid angle times H3_EARTH_RADIUS_KM squared, so dividing that radius
#' back out and multiplying by the body's leaves the solid angle times the
#' body's radius squared. Earth returns H3's own figure untouched: hexify's
#' Earth radius is the mean radius and H3's the authalic one, and rescaling
#' between the two would move an Earth area no user asked to move.
#' @param area_km2 Numeric vector of 'H3' areas, in km^2
#' @param radius_km Radius in kilometers
#' @noRd
scale_area_to_body <- function(area_km2, radius_km) {
  if (radius_km == EARTH_RADIUS_KM) return(area_km2)
  area_km2 * (radius_km / H3_EARTH_RADIUS_KM)^2
}

# =============================================================================
# Internal Helper Functions
# =============================================================================

#' Get resolution from grid object (handles both field names)
#'
#' Extracts resolution from a grid object, supporting both 'resolution'
#' (hexify style) and 'res' ('dggridR' style) field names.
#'
#' @param dggs Grid specification object
#' @param require Logical; if TRUE, stops with error if resolution not found
#' @return Integer resolution value, or NULL if not found and require=FALSE
#' @noRd
get_grid_resolution <- function(dggs, require = FALSE) {
  if ("resolution" %in% names(dggs)) {
    dggs$resolution
  } else if ("res" %in% names(dggs)) {
    dggs$res
  } else if (require) {
    stop("Grid object missing resolution field (neither 'resolution' nor 'res')")
  } else {
    NULL
  }
}

# =============================================================================
# Unit Conversion Constants
# =============================================================================

#' Square miles to square kilometers conversion factor
#' @noRd
MI2_TO_KM2 <- 2.58999

#' Miles to kilometers conversion factor
#' @noRd
MI_TO_KM <- 1.60934

# =============================================================================
# ISEA Aperture-3 Calibration Constants
# =============================================================================

#' Cell area at effective resolution 10 (aperture 3) in km^2
#' Used for area-to-resolution conversions.
#' @noRd
ISEA3H_RES10_AREA_KM2 <- 863.8006

# =============================================================================
# ISEA Default Orientation Constants
# =============================================================================
# Standard ISEA orientation with vertex 0 positioned at these coordinates.
# Reference: Snyder (1992) "An Equal-Area Map Projection For Polyhedral Globes"

#' Default longitude for ISEA vertex 0 (degrees)
#' @noRd
ISEA_VERT0_LON_DEG <- 11.25

#' Default latitude for ISEA vertex 0 (degrees)
#'
#' atan(phi) in degrees, where phi = (1 + sqrt(5)) / 2.
#' @noRd
ISEA_VERT0_LAT_DEG <- 58.282525588538995

#' Default azimuth for ISEA orientation (degrees)
#' @noRd
ISEA_AZIMUTH_DEG <- 0.0

# =============================================================================
# Grid Parameter Limits
# =============================================================================

#' Valid aperture values
#' @noRd
VALID_APERTURES <- c(3L, 4L, 7L)

#' Maximum supported resolution
#' @noRd
MAX_RESOLUTION <- 30L

#' Minimum supported resolution
#' @noRd
MIN_RESOLUTION <- 0L

# =============================================================================
# H3 Grid Constants
# =============================================================================

#' Earth radius the vendored H3 library measures areas and lengths against
#'
#' The 'WGS84' authalic radius, from EARTH_RADIUS_KM in src/h3/constants.h.
#' Distinct from EARTH_RADIUS_KM here, which is the 'WGS84' mean radius.
#' @noRd
H3_EARTH_RADIUS_KM <- 6371.007180918475

#' Maximum H3 resolution
#' @noRd
H3_MAX_RESOLUTION <- 15L

#' Minimum H3 resolution
#' @noRd
H3_MIN_RESOLUTION <- 0L

#' Average cell areas from H3 documentation (km^2), indexed by resolution + 1
#' Source: https://h3geo.org/docs/core-library/restable/
#' @noRd
H3_AVG_AREA_KM2 <- c(
  4357449.416,  # res 0
  609788.442,   # res 1
  86801.780,    # res 2
  12393.435,    # res 3
  1770.348,     # res 4
  252.904,      # res 5
  36.129,       # res 6
  5.161,        # res 7
  0.737,        # res 8
  0.105,        # res 9
  0.015,        # res 10
  0.00215,      # res 11
  0.000307,     # res 12
  0.0000439,    # res 13
  0.00000627,   # res 14
  0.000000895   # res 15
)

# =============================================================================
# H3 Resolution Helpers
# =============================================================================

#' Find closest H3 resolution for a target area
#'
#' Shared by hex_grid() and h3_crosswalk() to avoid duplicating the
#' resolution-matching logic.
#'
#' @param area_km2 Target cell area in km^2
#' @param radius_km Radius of the body, in kilometers
#' @return Integer H3 resolution (0-15)
#' @noRd
closest_h3_resolution <- function(area_km2, radius_km = EARTH_RADIUS_KM) {
  diffs <- abs(scale_area_to_body(H3_AVG_AREA_KM2, radius_km) - area_km2)
  which.min(diffs) - 1L
}

#' Average area of an H3 cell at a resolution, on a given body
#' @param resolution Integer H3 resolution (0-15)
#' @param radius_km Radius of the body, in kilometers
#' @noRd
h3_avg_area_km2 <- function(resolution, radius_km = EARTH_RADIUS_KM) {
  scale_area_to_body(H3_AVG_AREA_KM2[resolution + 1L], radius_km)
}

#' Check whether a grid is H3 type
#'
#' Safe check that handles old serialized objects without grid_type slot.
#'
#' @param grid A HexGridInfo object
#' @return Logical
#' @noRd
is_h3_grid <- function(grid) {
  tryCatch(
    identical(grid@grid_type, "h3"),
    error = function(e) FALSE
  )
}

# =============================================================================
# Coordinate Validation Helpers
# =============================================================================

#' Validate longitude values
#' @param lon Numeric vector of longitudes
#' @param warn Whether to warn on out-of-range values (default TRUE)
#' @return Logical vector indicating valid values
#' @noRd
validate_lon <- function(lon, warn = TRUE) {
  if (!is.numeric(lon)) {
    stop("Longitude must be numeric")
  }
  valid <- is.na(lon) | (lon >= -180 & lon <= 180)
  if (warn && any(!valid, na.rm = TRUE)) {
    warning("Some longitude values are outside valid range [-180, 180]")
  }
  valid
}

#' Validate latitude values
#' @param lat Numeric vector of latitudes
#' @param warn Whether to warn on out-of-range values (default TRUE)
#' @return Logical vector indicating valid values
#' @noRd
validate_lat <- function(lat, warn = TRUE) {
  if (!is.numeric(lat)) {
    stop("Latitude must be numeric")
  }
  valid <- is.na(lat) | (lat >= -90 & lat <= 90)
  if (warn && any(!valid, na.rm = TRUE)) {
    warning("Some latitude values are outside valid range [-90, 90]")
  }
  valid
}

#' Validate resolution value
#' @param resolution Integer resolution value
#' @return TRUE if valid, otherwise throws error
#' @noRd
validate_resolution <- function(resolution) {
  if (!is.numeric(resolution) || length(resolution) != 1) {
    stop("Resolution must be a single numeric value")
  }
  resolution <- as.integer(resolution)
  if (is.na(resolution) ||
      resolution < MIN_RESOLUTION ||
      resolution > MAX_RESOLUTION) {
    stop(sprintf(
      "Resolution must be between %d and %d", MIN_RESOLUTION, MAX_RESOLUTION
    ))
  }
  TRUE
}

#' Validate aperture value
#' @param aperture Integer aperture value
#' @return TRUE if valid, otherwise throws error
#' @noRd
validate_aperture <- function(aperture) {
  if (!is.numeric(aperture) || length(aperture) != 1) {
    stop("Aperture must be a single numeric value")
  }
  aperture <- as.integer(aperture)
  if (!aperture %in% VALID_APERTURES) {
    stop(sprintf(
      "Aperture must be one of: %s", paste(VALID_APERTURES, collapse = ", ")
    ))
  }
  TRUE
}

#' Hierarchical index type for a grid aperture
#' @param aperture Character aperture ("3", "4", "7", or "4/3")
#' @return "z3", "z7", or "zorder"
#' @noRd
index_type_for_aperture <- function(aperture) {
  if (aperture == "3") "z3"
  else if (aperture == "7") "z7"
  else "zorder"
}

#' Integer aperture for the C++ functions that take a single one
#'
#' A mixed sequence has no single aperture, and the 3 it reports names only
#' decode_cell_id()'s substrate branch, which is the one a mixed sequence
#' stores its cells on. Anything that reads the aperture as a refinement
#' factor -- cell geometry, hierarchy, cell counts -- is wrong on a mixed grid
#' with this integer and belongs on the sequence API, which takes one aperture
#' per level.
#' @param aperture Character or numeric aperture spelling
#' @return Integer aperture (3L, 4L, or 7L)
#' @noRd
aperture_to_int <- function(aperture) {
  if (is_mixed_aperture(aperture)) 3L else as.integer(aperture)
}

#' Calculate maximum cell ID for given resolution and aperture
#' @param resolution Integer resolution value
#' @param aperture Integer aperture value
#' @return Maximum valid cell ID (numeric)
#' @noRd
max_cell_id <- function(resolution, aperture) {
  # Cell count formula: N = 10 * aperture^res + 2 (cell IDs are 1..N,
  # matching calc_grid_params() in rcpp_cell.cpp)
  10 * (aperture^resolution) + 2
}

#' Validate cell ID values
#' @param cell_id Numeric vector of cell IDs
#' @param resolution Integer resolution value
#' @param aperture Integer aperture value
#' @param warn Whether to warn on out-of-range values (default TRUE)
#' @return Logical vector indicating valid values
#' @noRd
validate_cell_id <- function(cell_id, resolution, aperture, warn = TRUE) {
  if (!is.numeric(cell_id)) {
    stop("Cell ID must be numeric")
  }
  max_id <- max_cell_id(resolution, aperture)
  valid <- is.na(cell_id) | (cell_id >= 1 & cell_id <= max_id)
  if (warn && any(!valid, na.rm = TRUE)) {
    warning(sprintf(
      "Some cell IDs are outside valid range [1, %.0f] for res %d, ap %d",
      max_id, resolution, aperture
    ))
  }
  valid
}

# tests/testthat/test-radius.R
# Tests for grids on bodies other than Earth

MARS_RADIUS_KM <- 3389.50
MARS_SURFACE_KM2 <- 4 * pi * MARS_RADIUS_KM^2

# =============================================================================
# RADIUS RESOLUTION
# =============================================================================

test_that("radius_km takes a number or a body name", {
  expect_equal(hexify:::resolve_radius_km(3389.5), 3389.5)
  expect_equal(hexify:::resolve_radius_km("mars"), MARS_RADIUS_KM)
  expect_equal(hexify:::resolve_radius_km("Mars"), MARS_RADIUS_KM)
  expect_equal(hexify:::resolve_radius_km(" mars "), MARS_RADIUS_KM)
  expect_equal(hexify:::resolve_radius_km("earth"), hexify:::EARTH_RADIUS_KM)
})

test_that("radius_km rejects invalid input", {
  expect_error(hexify:::resolve_radius_km(0), "positive")
  expect_error(hexify:::resolve_radius_km(-1), "positive")
  expect_error(hexify:::resolve_radius_km(NA), "positive")
  expect_error(hexify:::resolve_radius_km(Inf), "positive")
  expect_error(hexify:::resolve_radius_km(c(1000, 2000)), "positive")
  expect_error(hexify:::resolve_radius_km("pandora"), "Unknown body")
})

test_that("body surface area is the sphere area, Earth the ellipsoid area", {
  expect_equal(hexify:::body_surface_km2(MARS_RADIUS_KM), MARS_SURFACE_KM2)
  expect_equal(hexify:::body_surface_km2(hexify:::EARTH_RADIUS_KM),
               hexify:::EARTH_SURFACE_KM2)
})

test_that("km per degree scales with radius", {
  expect_equal(hexify:::km_per_degree(hexify:::EARTH_RADIUS_KM),
               hexify:::KM_PER_DEGREE)
  expect_equal(
    hexify:::km_per_degree(MARS_RADIUS_KM) / hexify:::km_per_degree(hexify:::EARTH_RADIUS_KM),
    MARS_RADIUS_KM / hexify:::EARTH_RADIUS_KM
  )
})

# =============================================================================
# GRID CONSTRUCTION
# =============================================================================

test_that("Earth stays the default", {
  g <- hex_grid(resolution = 6)
  expect_equal(hexify:::grid_radius_km(g), hexify:::EARTH_RADIUS_KM)
  expect_true(hexify:::is_earth_grid(g))

  named <- hex_grid(resolution = 6, radius_km = "earth")
  expect_equal(named@area_km2, g@area_km2)
  expect_equal(named@diagonal_km, g@diagonal_km)
})

test_that("a grid on another body scales its cell areas", {
  earth <- hex_grid(resolution = 6)
  mars <- hex_grid(resolution = 6, radius_km = "mars")

  expect_equal(hexify:::grid_radius_km(mars), MARS_RADIUS_KM)
  expect_equal(mars@area_km2, MARS_SURFACE_KM2 / (10 * 3^6 + 2))
  expect_equal(mars@area_km2 / earth@area_km2,
               MARS_SURFACE_KM2 / hexify:::EARTH_SURFACE_KM2)
  expect_equal(mars@diagonal_km, sqrt(mars@area_km2 * 2 / sqrt(3)))
})

test_that("a body name and its radius give the same grid", {
  by_name <- hex_grid(resolution = 4, radius_km = "mars")
  by_number <- hex_grid(resolution = 4, radius_km = MARS_RADIUS_KM)
  expect_equal(by_name@area_km2, by_number@area_km2)
  expect_equal(by_name@resolution, by_number@resolution)
})

test_that("resolution for a target area follows the body", {
  mars <- hex_grid(area_km2 = 1000, radius_km = "mars")
  n_cells <- 10 * 3^mars@resolution + 2

  # The chosen resolution is the one whose cells are closest to 1000 km^2 on
  # Mars, so a smaller body reaches that cell size at a coarser resolution.
  expect_equal(mars@area_km2, MARS_SURFACE_KM2 / n_cells)
  expect_lt(mars@resolution, hex_grid(area_km2 = 1000)@resolution)
  expect_lt(abs(log(mars@area_km2 / 1000)), log(3) / 2 + 1e-8)
})

test_that("resolution for a target area follows the body for mixed apertures", {
  mars <- hex_grid(area_km2 = 5000, aperture = "4/3", radius_km = "mars")
  n_cells <- hexify:::aperture_n_cells(mars@aperture, mars@resolution)
  expect_equal(mars@area_km2, MARS_SURFACE_KM2 / n_cells)
  expect_lt(abs(log(mars@area_km2 / 5000)), log(4) / 2 + 1e-8)
})

test_that("radius_km must be positive in a grid object", {
  expect_error(hex_grid(resolution = 4, radius_km = -100), "positive")
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = 4L, area_km2 = 100,
        diagonal_km = 10, crs = 4326L, grid_type = "isea", radius_km = -1),
    "radius_km must be positive"
  )
})

# =============================================================================
# CELL GEOMETRY IS ANGULAR, SO RADIUS-INDEPENDENT
# =============================================================================

test_that("cells and centres are the same on every body", {
  earth <- hex_grid(resolution = 5)
  mars <- hex_grid(resolution = 5, radius_km = "mars")

  lon <- c(0, 16.37, -70.5, 120)
  lat <- c(0, 48.21, -33.4, 65)

  expect_equal(lonlat_to_cell(lon, lat, mars), lonlat_to_cell(lon, lat, earth))
  expect_equal(cell_to_lonlat(lonlat_to_cell(lon, lat, mars), mars),
               cell_to_lonlat(lonlat_to_cell(lon, lat, earth), earth))
  expect_equal(get_neighbors(1000, mars), get_neighbors(1000, earth))
})

test_that("cell_area reports the body's cell area", {
  mars <- hex_grid(resolution = 5, radius_km = "mars")
  cells <- lonlat_to_cell(c(0, 30), c(0, 40), mars)
  expect_equal(unname(cell_area(cells, mars)), rep(mars@area_km2, 2))
})

# =============================================================================
# STATISTICS
# =============================================================================

test_that("dgearthstat reports the body the grid covers", {
  mars <- hex_grid(resolution = 6, radius_km = "mars")
  stats <- dgearthstat(mars)

  expect_equal(stats$area_km, MARS_SURFACE_KM2)
  expect_equal(stats$n_cells, 10 * 3^6 + 2)
  expect_equal(stats$cell_area_km2, MARS_SURFACE_KM2 / (10 * 3^6 + 2))
  expect_equal(stats$cell_spacing_km, sqrt(2 * stats$cell_area_km2 / sqrt(3)))
  expect_equal(stats$cls_km, 2 * sqrt(stats$cell_area_km2 / pi))
})

test_that("hexify_compare_resolutions takes a radius", {
  cmp <- hexify_compare_resolutions(aperture = 3, res_range = 0:5,
                                    radius_km = "mars")
  expect_equal(cmp$cell_area_km2, MARS_SURFACE_KM2 / cmp$n_cells)

  earth <- hexify_compare_resolutions(aperture = 3, res_range = 0:5)
  expect_equal(cmp$n_cells, earth$n_cells)
  expect_true(all(cmp$cell_area_km2 < earth$cell_area_km2))
})

# =============================================================================
# LEGACY GRID OBJECTS
# =============================================================================

test_that("legacy hexify_grid carries a radius", {
  g <- hexify_grid(area = 1000, aperture = 3, radius_km = "mars")
  expect_equal(g$radius_km, MARS_RADIUS_KM)

  stats <- dgearthstat(g)
  expect_equal(stats$area_km, MARS_SURFACE_KM2)
  expect_equal(stats$cell_area_km2, MARS_SURFACE_KM2 / stats$n_cells)
})

test_that("the radius survives conversion between grid representations", {
  s4 <- hex_grid(resolution = 5, radius_km = "mars")
  legacy <- hexify:::HexGridInfo_to_hexify_grid(s4)
  expect_equal(legacy$radius_km, MARS_RADIUS_KM)

  back <- hexify:::hexify_grid_to_HexGridInfo(legacy)
  expect_equal(hexify:::grid_radius_km(back), MARS_RADIUS_KM)
  expect_equal(back@area_km2, s4@area_km2)
})

test_that("a grid with no radius reads as Earth", {
  expect_equal(hexify:::grid_radius_km(list(aperture = 3, resolution = 5)),
               hexify:::EARTH_RADIUS_KM)
  expect_equal(hexify:::grid_radius_km(list(radius_km = NA_real_)),
               hexify:::EARTH_RADIUS_KM)
})

# =============================================================================
# HEXIFY AND SF EXPORT
# =============================================================================

test_that("hexify() passes a radius through to the grid", {
  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
  hd <- hexify(df, lon = "lon", lat = "lat", resolution = 5, radius_km = "mars")

  expect_equal(hexify:::grid_radius_km(hd@grid), MARS_RADIUS_KM)
  expect_equal(as.data.frame(hd)$cell_area_km2,
               rep(hd@grid@area_km2, 3))

  earth <- hexify(df, lon = "lon", lat = "lat", resolution = 5)
  expect_equal(hd@cell_id, earth@cell_id)
})

test_that("polygon export samples the same cells on every body", {
  skip_if_not_installed("sf")

  earth <- hex_grid(resolution = 3)
  mars <- hex_grid(resolution = 3, radius_km = "mars")

  bbox <- c(-10, 35, 30, 60)
  expect_equal(sort(grid_rect(bbox, mars)$cell_id),
               sort(grid_rect(bbox, earth)$cell_id))
})

# =============================================================================
# H3 ON ANOTHER BODY
# =============================================================================

# H3 measures areas against the WGS84 authalic radius, so that is the radius
# divided back out when reading a cell on another body.
MARS_AREA_SCALE <- (MARS_RADIUS_KM / hexify:::H3_EARTH_RADIUS_KM)^2

test_that("H3 areas scale with the square of the radius ratio", {
  expect_equal(hexify:::scale_area_to_body(100, hexify:::EARTH_RADIUS_KM), 100)
  expect_equal(hexify:::scale_area_to_body(100, MARS_RADIUS_KM),
               100 * MARS_AREA_SCALE)
})

test_that("an H3 area on a body is its solid angle times that radius squared", {
  earth <- hex_grid(resolution = 5, type = "h3")
  mars <- suppressMessages(hex_grid(resolution = 5, type = "h3", radius_km = "mars"))
  cells <- lonlat_to_cell(0, 45, earth)

  steradians <- unname(cell_area(cells, earth)) / hexify:::H3_EARTH_RADIUS_KM^2
  expect_equal(unname(cell_area(cells, mars)), steradians * MARS_RADIUS_KM^2)
})

test_that("an H3 grid takes a radius", {
  earth <- hex_grid(resolution = 5, type = "h3")
  mars <- suppressMessages(hex_grid(resolution = 5, type = "h3", radius_km = "mars"))

  expect_equal(hexify:::grid_radius_km(mars), MARS_RADIUS_KM)
  expect_equal(mars@area_km2, earth@area_km2 * MARS_AREA_SCALE)
  expect_equal(mars@diagonal_km, sqrt(mars@area_km2 * 2 / sqrt(3)))
  expect_equal(mars@resolution, earth@resolution)
})

test_that("H3 cell geometry is the same on every body", {
  earth <- hex_grid(resolution = 5, type = "h3")
  mars <- suppressMessages(hex_grid(resolution = 5, type = "h3", radius_km = "mars"))

  lon <- c(0, 16.37, -70.5)
  lat <- c(0, 48.21, -33.4)

  expect_equal(lonlat_to_cell(lon, lat, mars), lonlat_to_cell(lon, lat, earth))
  expect_equal(cell_to_lonlat(lonlat_to_cell(lon, lat, mars), mars),
               cell_to_lonlat(lonlat_to_cell(lon, lat, earth), earth))
})

test_that("H3 per-cell areas follow the body", {
  earth <- hex_grid(resolution = 5, type = "h3")
  mars <- suppressMessages(hex_grid(resolution = 5, type = "h3", radius_km = "mars"))
  cells <- lonlat_to_cell(c(0, 10), c(45, 50), earth)

  expect_equal(unname(cell_area(cells, mars)),
               unname(cell_area(cells, earth)) * MARS_AREA_SCALE)
})

test_that("H3 statistics and tables follow the body", {
  mars <- suppressMessages(hex_grid(resolution = 5, type = "h3", radius_km = "mars"))
  stats <- dgearthstat(mars)

  expect_equal(stats$area_km, MARS_SURFACE_KM2)
  expect_equal(stats$cell_area_km2, mars@area_km2)
  expect_equal(stats$n_cells, 2 + 120 * 7^5)

  cmp <- hexify_compare_resolutions(type = "h3", res_range = 0:5, radius_km = "mars")
  earth_cmp <- hexify_compare_resolutions(type = "h3", res_range = 0:5)
  expect_equal(cmp$cell_area_km2, earth_cmp$cell_area_km2 * MARS_AREA_SCALE)
  expect_equal(cmp$n_cells, earth_cmp$n_cells)
})

test_that("an H3 target area picks the resolution for that body", {
  mars <- suppressWarnings(suppressMessages(
    hex_grid(area_km2 = 3000, type = "h3", radius_km = "mars")
  ))
  earth <- suppressWarnings(hex_grid(area_km2 = 3000, type = "h3"))

  # A 3000 km^2 cell is a larger share of Mars, so it lands at a coarser
  # resolution there.
  expect_lt(mars@resolution, earth@resolution)
  expect_equal(mars@area_km2, hexify:::h3_avg_area_km2(mars@resolution, MARS_RADIUS_KM))
})

test_that("import_h3 takes a radius", {
  ids <- c("8528342bfffffff", "85283473fffffff")
  g <- suppressMessages(import_h3(ids, radius_km = "mars"))
  expect_equal(hexify:::grid_radius_km(g), MARS_RADIUS_KM)
  expect_true(is_h3_grid(g))
})

test_that("h3_crosswalk needs both grids on the same body", {
  mars_isea <- hex_grid(resolution = 5, radius_km = "mars")
  mars_h3 <- suppressMessages(hex_grid(resolution = 5, type = "h3", radius_km = "mars"))
  earth_isea <- hex_grid(resolution = 5)

  cells <- lonlat_to_cell(c(0, 10), c(45, 50), mars_isea)
  xw <- h3_crosswalk(cells, mars_isea, direction = "isea_to_h3")
  expect_equal(xw$isea_area_km2, rep(mars_isea@area_km2, nrow(xw)))
  expect_equal(xw$area_ratio, xw$isea_area_km2 / xw$h3_area_km2)

  h3_cells <- lonlat_to_cell(c(0, 10), c(45, 50), mars_h3)
  expect_error(
    h3_crosswalk(h3_cells, mars_h3, isea_grid = earth_isea,
                 direction = "h3_to_isea"),
    "same body"
  )
})

# =============================================================================
# COORDINATE REFERENCE SYSTEM
# =============================================================================

test_that("an Earth grid stores the WGS84 EPSG code", {
  g <- hex_grid(resolution = 5)
  expect_type(g@crs, "integer")
  expect_equal(g@crs, 4326L)
  expect_equal(sf::st_crs(hexify:::grid_crs(g)), sf::st_crs(4326))
})

test_that("a grid on another body stores a longlat CRS on its own sphere", {
  g <- hex_grid(resolution = 5, radius_km = "mars")
  expect_type(g@crs, "character")

  crs <- hexify:::grid_crs(g)
  expect_false(is.na(crs))
  expect_false(sf::st_crs(crs) == sf::st_crs(4326))
  expect_match(sf::st_crs(crs)$proj4string, "+proj=longlat", fixed = TRUE)
  expect_match(sf::st_crs(crs)$proj4string,
               sprintf("+R=%d", as.integer(MARS_RADIUS_KM * 1000)), fixed = TRUE)
})

test_that("the body CRS carries no trailing zeros", {
  expect_equal(hexify:::body_crs_string(3389.5),
               "+proj=longlat +R=3389500 +no_defs")
  expect_equal(hexify:::body_crs_string(1188.3),
               "+proj=longlat +R=1188300 +no_defs")
})

test_that("an explicit crs is kept as given", {
  expect_equal(hex_grid(resolution = 5, crs = 3035)@crs, 3035L)

  proj <- "+proj=longlat +R=3000000 +no_defs"
  g <- hex_grid(resolution = 5, radius_km = 3000, crs = proj)
  expect_equal(g@crs, proj)

  # An EPSG code on another body is the user's call to make
  expect_equal(hex_grid(resolution = 5, radius_km = "mars", crs = 4326L)@crs, 4326L)
})

test_that("a crs sf cannot read is rejected", {
  expect_error(hex_grid(resolution = 5, crs = "not a crs"), "sf can read")
  expect_error(hex_grid(resolution = 5, crs = ""), "non-empty")
  expect_error(hex_grid(resolution = 5, crs = -1), "positive EPSG code")
  expect_error(hex_grid(resolution = 5, crs = c(4326, 3035)), "positive EPSG code")
})

test_that("validity accepts both CRS forms and rejects a broken one", {
  expect_s4_class(
    new("HexGridInfo", aperture = "3", resolution = 5L,
        crs = "+proj=longlat +R=3389500 +no_defs"),
    "HexGridInfo"
  )
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = 5L, crs = "not a crs"),
    "sf can read"
  )
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = 5L, crs = 0L),
    "positive integer EPSG code"
  )
})

test_that("sf output carries the grid's own CRS", {
  mars <- hex_grid(resolution = 5, radius_km = "mars")
  earth <- hex_grid(resolution = 5)
  cells <- lonlat_to_cell(c(0, 10), c(45, 50), mars)

  mars_sf <- cell_to_sf(cells, mars)
  expect_equal(sf::st_crs(mars_sf), sf::st_crs(hexify:::grid_crs(mars)))
  expect_false(sf::st_crs(mars_sf) == sf::st_crs(cell_to_sf(cells, earth)))

  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  hd <- hexify(df, lon = "lon", lat = "lat", grid = mars)
  expect_equal(sf::st_crs(as_sf(hd)), sf::st_crs(hexify:::grid_crs(mars)))
})

test_that("a grid prints its CRS in either form", {
  expect_output(show(hex_grid(resolution = 5)), "CRS: +EPSG:4326")
  expect_output(show(hex_grid(resolution = 5, radius_km = "mars")),
                "CRS: +[+]proj=longlat")
})

test_that("a legacy grid list without a crs converts on its own body", {
  lst <- list(aperture = "3", resolution = 5L, radius_km = MARS_RADIUS_KM)
  g <- hexify:::hexify_grid_to_HexGridInfo(lst)
  expect_type(g@crs, "character")
  expect_match(g@crs, "+R=3389500", fixed = TRUE)
})

# tests/testthat/test-constants.R
# Tests for constants.R validation functions and helpers

# =============================================================================
# VALIDATE_LON
# =============================================================================

test_that("validate_lon returns TRUE for valid values", {
  result <- hexify:::validate_lon(c(0, 45, -120, 180, -180), warn = FALSE)
  expect_true(all(result))
})

test_that("validate_lon returns FALSE for out-of-range values", {
  result <- hexify:::validate_lon(c(0, 200, -200), warn = FALSE)
  expect_equal(result, c(TRUE, FALSE, FALSE))
})

test_that("validate_lon handles NA values", {
  result <- hexify:::validate_lon(c(0, NA, 45), warn = FALSE)
  expect_equal(result, c(TRUE, TRUE, TRUE))
})

test_that("validate_lon warns on invalid values", {
  expect_warning(
    hexify:::validate_lon(c(200)),
    "outside valid range"
  )
})

test_that("validate_lon errors on non-numeric input", {
  expect_error(
    hexify:::validate_lon("not numeric"),
    "must be numeric"
  )
})

# =============================================================================
# VALIDATE_LAT
# =============================================================================

test_that("validate_lat returns TRUE for valid values", {
  result <- hexify:::validate_lat(c(0, 45, -45, 90, -90), warn = FALSE)
  expect_true(all(result))
})

test_that("validate_lat returns FALSE for out-of-range values", {
  result <- hexify:::validate_lat(c(0, 100, -100), warn = FALSE)
  expect_equal(result, c(TRUE, FALSE, FALSE))
})

test_that("validate_lat handles NA values", {
  result <- hexify:::validate_lat(c(0, NA, 45), warn = FALSE)
  expect_equal(result, c(TRUE, TRUE, TRUE))
})

test_that("validate_lat warns on invalid values", {
  expect_warning(
    hexify:::validate_lat(c(100)),
    "outside valid range"
  )
})

test_that("validate_lat errors on non-numeric input", {
  expect_error(
    hexify:::validate_lat("not numeric"),
    "must be numeric"
  )
})

# =============================================================================
# VALIDATE_RESOLUTION
# =============================================================================

test_that("validate_resolution accepts valid resolutions", {
  for (res in c(0, 1, 10, 20, 30)) {
    expect_true(hexify:::validate_resolution(res))
  }
})

test_that("validate_resolution errors on negative resolution", {
  expect_error(
    hexify:::validate_resolution(-1),
    "between"
  )
})

test_that("validate_resolution errors on resolution > 30", {
  expect_error(
    hexify:::validate_resolution(31),
    "between"
  )
})

test_that("validate_resolution errors on non-numeric input", {
  expect_error(
    hexify:::validate_resolution("10"),
    "must be a single numeric"
  )
})

test_that("validate_resolution errors on vector input", {
  expect_error(
    hexify:::validate_resolution(c(1, 2)),
    "must be a single numeric"
  )
})

test_that("validate_resolution errors on NA input", {
  expect_error(
    hexify:::validate_resolution(NA),
    "single numeric"
  )
})

# =============================================================================
# VALIDATE_APERTURE
# =============================================================================

test_that("validate_aperture accepts valid apertures", {
  expect_true(hexify:::validate_aperture(3))
  expect_true(hexify:::validate_aperture(4))
  expect_true(hexify:::validate_aperture(7))
})

test_that("validate_aperture errors on invalid aperture", {
  expect_error(
    hexify:::validate_aperture(5),
    "must be one of"
  )
  expect_error(
    hexify:::validate_aperture(1),
    "must be one of"
  )
})

test_that("validate_aperture errors on non-numeric input", {
  expect_error(
    hexify:::validate_aperture("3"),
    "must be a single numeric"
  )
})

test_that("validate_aperture errors on vector input", {
  expect_error(
    hexify:::validate_aperture(c(3, 4)),
    "must be a single numeric"
  )
})

# =============================================================================
# MAX_CELL_ID
# =============================================================================

test_that("max_cell_id returns 12 for resolution 0", {
  # N = 10 * aperture^0 + 2 = 12, independent of aperture
  expect_equal(hexify:::max_cell_id(0, 3), 12)
  expect_equal(hexify:::max_cell_id(0, 4), 12)
  expect_equal(hexify:::max_cell_id(0, 7), 12)
})

test_that("max_cell_id increases with resolution", {
  max_res1 <- hexify:::max_cell_id(1, 3)
  max_res2 <- hexify:::max_cell_id(2, 3)
  max_res3 <- hexify:::max_cell_id(3, 3)

  expect_true(max_res2 > max_res1)
  expect_true(max_res3 > max_res2)
})

test_that("max_cell_id formula is correct", {
  # Formula: 10 * aperture^res + 2, one formula for every aperture
  expect_equal(hexify:::max_cell_id(1, 3), 10 * 3^1 + 2)
  expect_equal(hexify:::max_cell_id(2, 4), 10 * 4^2 + 2)
  expect_equal(hexify:::max_cell_id(1, 7), 10 * 7^1 + 2)
  expect_equal(hexify:::max_cell_id(3, 7), 10 * 7^3 + 2)
})

# =============================================================================
# VALIDATE_CELL_ID
# =============================================================================

test_that("validate_cell_id returns TRUE for valid cell IDs", {
  result <- hexify:::validate_cell_id(
    c(1, 10, 12), resolution = 0, aperture = 3, warn = FALSE
  )
  expect_true(all(result))
})

test_that("validate_cell_id returns FALSE for invalid cell IDs", {
  result <- hexify:::validate_cell_id(
    c(1, 0, 100), resolution = 0, aperture = 3, warn = FALSE
  )
  expect_equal(result, c(TRUE, FALSE, FALSE))
})

test_that("validate_cell_id handles NA values", {
  result <- hexify:::validate_cell_id(
    c(1, NA, 10), resolution = 0, aperture = 3, warn = FALSE
  )
  expect_equal(result, c(TRUE, TRUE, TRUE))
})

test_that("validate_cell_id warns on invalid values", {
  expect_warning(
    hexify:::validate_cell_id(c(0), resolution = 0, aperture = 3),
    "outside valid range"
  )
})

test_that("validate_cell_id errors on non-numeric input", {
  expect_error(
    hexify:::validate_cell_id("1", resolution = 0, aperture = 3),
    "must be numeric"
  )
})

# =============================================================================
# GET_GRID_RESOLUTION
# =============================================================================

test_that("get_grid_resolution extracts resolution from hexify_grid", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  res <- hexify:::get_grid_resolution(grid)
  expect_equal(res, grid$resolution)
})

test_that("get_grid_resolution handles res field name", {
  # Simulate dggridR-style grid with res field
  grid <- list(res = 10)
  res <- hexify:::get_grid_resolution(grid)
  expect_equal(res, 10)
})

test_that("get_grid_resolution returns NULL when field missing", {
  grid <- list(aperture = 3)
  res <- hexify:::get_grid_resolution(grid, require = FALSE)
  expect_null(res)
})

test_that("get_grid_resolution errors when require=TRUE and missing", {
  grid <- list(aperture = 3)
  expect_error(
    hexify:::get_grid_resolution(grid, require = TRUE),
    "missing resolution"
  )
})

# =============================================================================
# CONSTANTS VALUES
# =============================================================================

test_that("EARTH_SURFACE_KM2 is approximately correct", {
  # Earth surface area is approximately 510 million km²
  expect_true(hexify:::EARTH_SURFACE_KM2 > 500000000)
  expect_true(hexify:::EARTH_SURFACE_KM2 < 520000000)
})

test_that("EARTH_RADIUS_KM is approximately correct", {
  # Mean Earth radius is approximately 6371 km
  expect_equal(hexify:::EARTH_RADIUS_KM, 6371.0088, tolerance = 0.01)
})

test_that("VALID_APERTURES contains expected values", {
  expect_equal(hexify:::VALID_APERTURES, c(3L, 4L, 7L))
})

test_that("MIN_RESOLUTION and MAX_RESOLUTION are reasonable", {
  expect_equal(hexify:::MIN_RESOLUTION, 0L)
  expect_equal(hexify:::MAX_RESOLUTION, 30L)
})

# R/hex_summarize.R
# Cell-level aggregation for HexData objects

#' Summarize Data by Hex Cell
#'
#' Aggregates data within each hexagonal cell, similar to
#' `dplyr::group_by(cell_id) |> summarize(...)`. Returns a data.frame
#' with one row per unique cell, including cell center coordinates and area.
#'
#' @param hex_data A HexData object (from [hexify()]).
#' @param ... Named summary expressions (tidyeval). Each expression is
#'   evaluated per cell group with the group's columns available by name.
#'   Examples: `mean_temp = mean(temperature)`,
#'   `n_species = length(unique(species))`,
#'   `max_elev = max(elevation, na.rm = TRUE)`. Provide either `...` or
#'   `.fns`, not both.
#' @param .fns Optional named list of functions for formula-style aggregation.
#'   Example: `.fns = list(mean_temp = ~mean(temperature))`. Provide either
#'   `...` or `.fns`, not both.
#' @param geometry Logical. If `TRUE`, attach cell center points as an
#'   sf geometry column (requires sf). Default `FALSE`.
#'
#' @return A data.frame with columns:
#'   \describe{
#'     \item{cell_id}{Unique cell identifier}
#'     \item{cell_cen_lon, cell_cen_lat}{Cell center coordinates}
#'     \item{cell_area_km2}{Cell area in km^2}
#'     \item{n_points}{Number of data points in this cell}
#'     \item{...}{User-defined summary columns}
#'   }
#'   If `geometry = TRUE`, returns an sf object with POINT geometry.
#'
#' @details
#' The function works entirely in R (no C++ needed). It groups by `cell_id`
#' and evaluates the summary expressions within each group.
#'
#' If no summary expressions are provided, returns cell counts only.
#'
#' @seealso [hexify()] for creating HexData objects,
#'   [get_neighbors()] for finding neighboring cells
#'
#' @export
#' @examples
#' \donttest{
#' df <- data.frame(
#'   lon = runif(100, -10, 10),
#'   lat = runif(100, 40, 55),
#'   temperature = rnorm(100, 15, 5),
#'   species = sample(letters[1:5], 100, replace = TRUE)
#' )
#' hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 500)
#'
#' # Count points per cell
#' hex_summarize(hd)
#'
#' # Custom summaries
#' hex_summarize(hd, mean_temp = mean(temperature),
#'                   n_species = length(unique(species)))
#' }
hex_summarize <- function(hex_data, ..., .fns = NULL, geometry = FALSE) {
  if (!is_hex_data(hex_data)) {
    stop("hex_data must be a HexData object")
  }

  # Build the combined data.frame
  df <- as.data.frame(hex_data)

  # Capture user expressions
  dots <- rlang::enquos(...)

  if (length(dots) > 0 && !is.null(.fns)) {
    stop("Provide summary expressions via either '...' or '.fns', not both")
  }

  # Group by cell_id, preserving first-occurrence order so cell_ids stays
  # aligned with split_df's group order (split() would otherwise sort groups)
  cell_ids <- unique(df$cell_id)
  split_df <- split(df, factor(df$cell_id, levels = cell_ids))

  if (length(dots) == 0 && is.null(.fns)) {
    # No summary expressions: just count
    result <- data.frame(
      cell_id = cell_ids,
      n_points = vapply(split_df, nrow, integer(1)),
      stringsAsFactors = FALSE
    )
  } else if (length(dots) > 0) {
    # Tidyeval path
    summaries <- lapply(split_df, function(grp) {
      env <- rlang::as_data_mask(grp)
      vals <- lapply(dots, function(q) rlang::eval_tidy(q, data = env))
      as.data.frame(vals, stringsAsFactors = FALSE)
    })

    summary_df <- do.call(rbind, summaries)
    result <- data.frame(
      cell_id = cell_ids,
      n_points = vapply(split_df, nrow, integer(1)),
      summary_df,
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  } else {
    # Formula-style (.fns)
    if (!is.list(.fns) || is.null(names(.fns))) {
      stop(".fns must be a named list of functions or formulas")
    }

    # Convert formulas to functions
    fns <- lapply(.fns, function(f) {
      if (inherits(f, "formula")) rlang::as_function(f) else f
    })

    summaries <- lapply(split_df, function(grp) {
      vals <- lapply(fns, function(fn) fn(grp))
      as.data.frame(vals, stringsAsFactors = FALSE)
    })

    summary_df <- do.call(rbind, summaries)
    result <- data.frame(
      cell_id = cell_ids,
      n_points = vapply(split_df, nrow, integer(1)),
      summary_df,
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  }

  # Attach cell center coordinates (use first occurrence per cell)
  first_idx <- match(cell_ids, df$cell_id)
  result$cell_cen_lon <- df$cell_cen_lon[first_idx]
  result$cell_cen_lat <- df$cell_cen_lat[first_idx]
  result$cell_area_km2 <- df$cell_area_km2[first_idx]

  # Reorder columns: cell info first, then summaries
  info_cols <- c("cell_id", "cell_cen_lon", "cell_cen_lat", "cell_area_km2",
                  "n_points")
  other_cols <- setdiff(names(result), info_cols)
  result <- result[, c(info_cols, other_cols), drop = FALSE]
  rownames(result) <- NULL


  if (geometry) {
    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("Package 'sf' is required when geometry = TRUE")
    }
    result <- sf::st_as_sf(result,
                            coords = c("cell_cen_lon", "cell_cen_lat"),
                            crs = grid_crs(hex_data@grid))
  }

  result
}

# tests/testthat/test-summarize.R
# Tests for hex_summarize()

test_that("hex_summarize returns cell counts with no expressions", {
  df <- data.frame(lon = c(10, 10, 10.5, 20), lat = c(50, 50, 50.1, 55))
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
  result <- hex_summarize(hd)

  expect_true(is.data.frame(result))
  expect_true("cell_id" %in% names(result))
  expect_true("n_points" %in% names(result))
  expect_true("cell_cen_lon" %in% names(result))
  expect_true("cell_cen_lat" %in% names(result))
  expect_true("cell_area_km2" %in% names(result))
  expect_equal(sum(result$n_points), nrow(df))
})

test_that("hex_summarize works with tidyeval expressions", {
  df <- data.frame(
    lon = runif(50, 5, 15),
    lat = runif(50, 45, 55),
    temp = rnorm(50, 15, 5)
  )
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 2000)
  result <- hex_summarize(hd, mean_temp = mean(temp))

  expect_true("mean_temp" %in% names(result))
  expect_true(is.numeric(result$mean_temp))
})

test_that("hex_summarize works with .fns argument", {
  df <- data.frame(
    lon = runif(50, 5, 15),
    lat = runif(50, 45, 55),
    temp = rnorm(50, 15, 5)
  )
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 2000)
  result <- hex_summarize(hd, .fns = list(mean_temp = ~mean(.x$temp)))

  expect_true("mean_temp" %in% names(result))
})

test_that("hex_summarize with geometry=TRUE returns sf", {
  skip_if_not_installed("sf")
  df <- data.frame(
    lon = runif(20, 5, 15),
    lat = runif(20, 45, 55)
  )
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 2000)
  result <- hex_summarize(hd, geometry = TRUE)

  expect_true(inherits(result, "sf"))
})

test_that("hex_summarize rejects non-HexData input", {
  expect_error(hex_summarize(data.frame(x = 1)), "HexData")
})

# hexify_stats.R
# Grid statistics and utility functions
#
# This file contains functions for calculating grid statistics and
# other utility functions for working with hexagonal grids.

#' @title Grid Statistics
#' @description Functions for calculating grid statistics and utilities
#' @name hexify-stats
NULL

#' Get grid statistics for whole-body coverage
#'
#' Calculates statistics about the hexagonal grid at the current resolution,
#' including total number of cells, cell area, and cell spacing. Figures follow
#' the radius the grid carries, so a grid built with \code{radius_km} reports
#' that body's areas and distances.
#'
#' @param dggs Grid specification from hexify_grid()
#'
#' @return List with components:
#'   \item{area_km}{Total surface area of the body in km^2}
#'   \item{n_cells}{Total number of cells at this resolution}
#'   \item{cell_area_km2}{Average cell area in km^2}
#'   \item{cell_spacing_km}{Average distance between cell centers in km}
#'   \item{resolution}{Resolution level}
#'   \item{aperture}{Grid aperture}
#'
#' @family grid statistics
#' @export
#' @examples
#' grid <- hexify_grid(area = 1000, aperture = 3)
#' stats <- dgearthstat(grid)
#'
#' print(sprintf("Resolution %d has %.0f cells",
#'               stats$resolution, stats$n_cells))
#' print(sprintf("Average cell area: %.2f km^2",
#'               stats$cell_area_km2))
#' print(sprintf("Average cell spacing: %.2f km",
#'               stats$cell_spacing_km))
dgearthstat <- function(dggs) {
  # Accept HexGridInfo objects
  if (is_hex_grid(dggs)) {
    g <- dggs
    gt <- tryCatch(g@grid_type, error = function(e) "isea")

    if (gt == "h3") {
      h3_n_cells <- 2 + 120 * 7^g@resolution
      cell_area_km2 <- h3_avg_area_km2(g@resolution, grid_radius_km(g))
      cell_spacing_km <- sqrt(2 * cell_area_km2 / sqrt(3))
      cls_km <- 2 * sqrt(cell_area_km2 / pi)
      return(list(
        area_km = body_surface_km2(grid_radius_km(g)),
        n_cells = h3_n_cells,
        cell_area_km2 = cell_area_km2,
        cell_spacing_km = cell_spacing_km,
        cls_km = cls_km,
        resolution = g@resolution,
        aperture = 7L,
        grid_type = "h3"
      ))
    }

    # ISEA HexGridInfo
    n_cells <- aperture_n_cells(g@aperture, g@resolution)
    ap <- aperture_to_int(g@aperture)
    surface_km2 <- body_surface_km2(grid_radius_km(g))
    cell_area_km2 <- surface_km2 / n_cells
    cell_spacing_km <- sqrt(2 * cell_area_km2 / sqrt(3))
    cls_km <- 2 * sqrt(cell_area_km2 / pi)
    return(list(
      area_km = surface_km2,
      n_cells = n_cells,
      cell_area_km2 = cell_area_km2,
      cell_spacing_km = cell_spacing_km,
      cls_km = cls_km,
      resolution = g@resolution,
      aperture = ap
    ))
  }

  # Legacy hexify_grid / dggs path
  if (!inherits(dggs, "hexify_grid") && !inherits(dggs, "dggs")) {
    stop("dggs must be a hexify_grid or HexGridInfo object")
  }

  resolution <- get_grid_resolution(dggs, require = TRUE)

  # Cell count formula (matches calc_grid_params() in rcpp_cell.cpp)
  n_cells <- max_cell_id(resolution, dggs$aperture)

  # Calculate cell area
  surface_km2 <- body_surface_km2(grid_radius_km(dggs))
  cell_area_km2 <- surface_km2 / n_cells

  # Approximate cell spacing (distance between cell centers)
  # For hexagons: spacing approx  sqrt(2 * area / sqrt(3))
  cell_spacing_km <- sqrt(2 * cell_area_km2 / sqrt(3))

  # Calculate characteristic length scale (CLS)
  # CLS is the diameter of a spherical cap with same area
  # CLS = 2 * sqrt(area / pi)
  cls_km <- 2 * sqrt(cell_area_km2 / pi)

  return(list(
    area_km = surface_km2,
    n_cells = n_cells,
    cell_area_km2 = cell_area_km2,
    cell_spacing_km = cell_spacing_km,
    cls_km = cls_km,
    resolution = resolution,
    aperture = dggs$aperture
  ))
}

#' Find closest resolution for target cell area
#'
#' Finds the grid resolution that produces cells closest to the target area.
#' This is primarily used internally by \code{\link{hexify_grid}} and
#' \code{\link{hex_grid}}. Most users should use those functions directly.
#'
#' @param dggs Grid specification (aperture and topology must be set)
#' @param area Target cell area in km^2 (if metric=TRUE)
#' @param round Rounding method ("nearest", "up", "down")
#' @param metric Whether area is in metric units
#' @param show_info Print information about chosen resolution
#'
#' @return Resolution level (integer)
#'
#' @keywords internal
#' @export
#' @examples
#' # Create a temporary grid to get aperture settings
#' temp_grid <- list(aperture = 3, topology = "HEXAGON")
#' class(temp_grid) <- "hexify_grid"
#' 
#' # Find resolution for 1000 km^2 cells
#' res <- dg_closest_res_to_area(temp_grid, area = 1000, 
#'                                metric = TRUE, show_info = TRUE)
#' print(res)
#' @family grid statistics
dg_closest_res_to_area <- function(dggs, area, round = "nearest",
                                   metric = TRUE, show_info = FALSE) {
  if (!metric) {
    # Convert from square miles to square km
    area <- area * MI2_TO_KM2
  }
  
  # Calculate resolution
  resolution <- calculate_resolution_for_area(area, dggs$aperture,
                                              grid_radius_km(dggs))
  
  # Apply rounding
  if (round == "up") {
    resolution <- ceiling(resolution)
  } else if (round == "down") {
    resolution <- floor(resolution)
  } else {
    resolution <- round(resolution)
  }
  
  # Ensure valid range
  resolution <- max(0, min(30, resolution))
  
  if (show_info) {
    # Calculate actual area at this resolution
    temp_grid <- dggs
    temp_grid$resolution <- resolution
    temp_grid$res <- resolution
    stats <- dgearthstat(temp_grid)
    
    message(sprintf("Resolution %d:", resolution))
    message(sprintf("  Cell area: %.2f km^2", stats$cell_area_km2))
    message(sprintf("  Cell spacing: %.2f km", stats$cell_spacing_km))
    message(sprintf("  Total cells: %.0f", stats$n_cells))
  }
  
  return(resolution)
}


#' Compare grid resolutions
#'
#' Generates a table comparing different resolution levels for a given
#' grid configuration. Useful for choosing appropriate resolution.
#'
#' @param aperture Grid aperture (3, 4, or 7). Ignored for H3 grids.
#' @param res_range Range of resolutions to compare (e.g., 1:10)
#' @param type Grid type: "isea" (default) or "h3".
#' @param print If TRUE, prints a formatted table to console. If FALSE (default),
#'   returns a data frame.
#' @param radius_km Radius of the body, in kilometers, or a body name such as
#'   "mars" (default Earth). See \code{\link{hex_grid}}.
#'
#' @return If print=FALSE: data frame with columns resolution, n_cells,
#'   cell_area_km2, cell_spacing_km, cls_km.
#'   If print=TRUE: invisibly returns the data frame after printing.
#'
#' @family grid statistics
#' @export
#' @examples
#' # Get data frame of resolutions 0-10 for aperture 3
#' comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:10)
#' print(comparison)
#'
#' # Print formatted table directly
#' hexify_compare_resolutions(aperture = 3, res_range = 0:10, print = TRUE)
#'
#' # Find resolution with cells ~1000 km^2
#' subset(comparison, cell_area_km2 > 900 & cell_area_km2 < 1100)
#'
#' # Resolutions on another body
#' hexify_compare_resolutions(aperture = 3, res_range = 0:6, radius_km = "mars")
hexify_compare_resolutions <- function(aperture = 3, res_range = 0:15,
                                       type = c("isea", "h3"),
                                       print = FALSE,
                                       radius_km = EARTH_RADIUS_KM) {
  type <- match.arg(type)
  radius_km <- resolve_radius_km(radius_km)

  if (type == "h3") {
    # H3 resolution table from pre-computed area values
    res_range <- res_range[res_range >= H3_MIN_RESOLUTION & res_range <= H3_MAX_RESOLUTION]
    results <- lapply(res_range, function(res) {
      cell_area_km2 <- h3_avg_area_km2(res, radius_km)
      h3_n_cells <- 2 + 120 * 7^res
      cell_spacing_km <- sqrt(2 * cell_area_km2 / sqrt(3))
      cls_km <- 2 * sqrt(cell_area_km2 / pi)
      data.frame(
        resolution = res,
        n_cells = h3_n_cells,
        cell_area_km2 = cell_area_km2,
        cell_spacing_km = cell_spacing_km,
        cls_km = cls_km
      )
    })
    result_df <- do.call(rbind, results)

    if (print) {
      cat("\nGrid Resolution Comparison (H3)\n")
      cat(paste(rep("=", 70), collapse = ""), "\n")
      cat(sprintf("%-4s  %-12s  %-12s  %-12s  %-10s\n",
                  "Res", "# Cells", "Area (km^2)", "Spacing (km)", "CLS (km)"))
      cat(paste(rep("-", 70), collapse = ""), "\n")
      for (i in seq_len(nrow(result_df))) {
        row <- result_df[i, ]
        n_cells_str <- if (row$n_cells > 1e12) {
          sprintf("%.1fT", row$n_cells / 1e12)
        } else if (row$n_cells > 1e9) {
          sprintf("%.1fB", row$n_cells / 1e9)
        } else if (row$n_cells > 1e6) {
          sprintf("%.1fM", row$n_cells / 1e6)
        } else if (row$n_cells > 1e3) {
          sprintf("%.1fK", row$n_cells / 1e3)
        } else {
          sprintf("%.0f", row$n_cells)
        }
        cat(sprintf("%-4d  %-12s  %-12.4f  %-12.3f  %-10.3f\n",
                    row$resolution, n_cells_str,
                    row$cell_area_km2, row$cell_spacing_km, row$cls_km))
      }
      cat(paste(rep("=", 70), collapse = ""), "\n")
      cat("Note: H3 areas are averages; actual area varies by location\n\n")
      return(invisible(result_df))
    }
    return(result_df)
  }

  # ISEA path (original)
  # Create temporary grid
  temp_grid <- list(
    aperture = aperture,
    topology = "HEXAGON",
    projection = "ISEA",
    radius_km = radius_km
  )
  class(temp_grid) <- c("hexify_grid", "dggs", "list")

  # Calculate stats for each resolution
  results <- lapply(res_range, function(res) {
    temp_grid$resolution <- res
    temp_grid$res <- res
    stats <- dgearthstat(temp_grid)

    data.frame(
      resolution = res,
      n_cells = stats$n_cells,
      cell_area_km2 = stats$cell_area_km2,
      cell_spacing_km = stats$cell_spacing_km,
      cls_km = stats$cls_km
    )
  })

  # Combine into data frame
  result_df <- do.call(rbind, results)

  if (print) {
    .print_resolution_table(result_df, aperture)
    return(invisible(result_df))
  }

  return(result_df)
}

#' Print formatted resolution table
#' @noRd
.print_resolution_table <- function(comparison, aperture) {
  cat(sprintf("\nGrid Resolution Comparison (Aperture %d)\n", aperture))
  cat(paste(rep("=", 70), collapse = ""), "\n")
  cat(sprintf("%-4s  %-12s  %-12s  %-12s  %-10s\n",
              "Res", "# Cells", "Area (km^2)", "Spacing (km)", "CLS (km)"))
  cat(paste(rep("-", 70), collapse = ""), "\n")

  for (i in seq_len(nrow(comparison))) {
    row <- comparison[i, ]

    # Format numbers nicely
    n_cells_str <- if (row$n_cells > 1e6) {
      sprintf("%.1fM", row$n_cells / 1e6)
    } else if (row$n_cells > 1e3) {
      sprintf("%.1fK", row$n_cells / 1e3)
    } else {
      sprintf("%.0f", row$n_cells)
    }

    cat(sprintf("%-4d  %-12s  %-12.1f  %-12.1f  %-10.1f\n",
                row$resolution,
                n_cells_str,
                row$cell_area_km2,
                row$cell_spacing_km,
                row$cls_km))
  }

  cat(paste(rep("=", 70), collapse = ""), "\n\n")
}

# tests/testthat/test-stats.R
# Tests for grid statistics functions

# =============================================================================
# DGGRID HELPER FUNCTIONS
# =============================================================================

test_that("dggrid_43h_sequence creates correct aperture sequences", {
  # Basic 43H pattern
  seq1 <- dggrid_43h_sequence(2, 3)
  expect_equal(seq1, c(4L, 4L, 3L, 3L, 3L))

  # All aperture 4
  seq2 <- dggrid_43h_sequence(5, 0)
  expect_equal(seq2, c(4L, 4L, 4L, 4L, 4L))

  # All aperture 3
  seq3 <- dggrid_43h_sequence(0, 4)
  expect_equal(seq3, c(3L, 3L, 3L, 3L))

  # Single level each
  seq4 <- dggrid_43h_sequence(1, 1)
  expect_equal(seq4, c(4L, 3L))
})

# =============================================================================
# EARTH STATISTICS
# =============================================================================

test_that("dgearthstat returns correct structure", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  stats <- dgearthstat(grid)

  expect_type(stats, "list")
  expect_named(stats, c(
    "area_km", "n_cells", "cell_area_km2",
    "cell_spacing_km", "cls_km", "resolution", "aperture"
  ))

  # Check values are positive

  expect_gt(stats$area_km, 0)
  expect_gt(stats$n_cells, 0)
  expect_gt(stats$cell_area_km2, 0)
  expect_gt(stats$cell_spacing_km, 0)
  expect_gt(stats$cls_km, 0)
})

test_that("dgearthstat returns correct values for aperture 3", {
  grid <- hexify_grid(area = 10000, aperture = 3)
  stats <- dgearthstat(grid)

  # Earth surface area should be correct
  expect_equal(stats$area_km, EARTH_SURFACE_KM2)

  # Check aperture matches
  expect_equal(stats$aperture, 3)

  # n_cells should equal 10 * 3^resolution for aperture 3
  expected_cells <- 10 * (3 ^ stats$resolution) + 2
  expect_equal(stats$n_cells, expected_cells)

  # Cell area = Earth surface / n_cells
  expected_area <- EARTH_SURFACE_KM2 / expected_cells
  expect_equal(stats$cell_area_km2, expected_area)
})

test_that("dgearthstat works for different apertures", {
  for (ap in c(3, 4, 7)) {
    grid <- hexify_grid(area = 1000, aperture = ap)
    stats <- dgearthstat(grid)

    expect_equal(stats$aperture, ap)
    expect_gt(stats$n_cells, 0)
  }
})

test_that("dgearthstat validates input", {
  expect_error(dgearthstat("not a grid"), "hexify_grid")
  expect_error(dgearthstat(list(x = 1)), "hexify_grid")
})

# =============================================================================
# MAX CELL
# =============================================================================

test_that("max_cell_id returns correct number of cells", {
  grid <- hexify_grid(area = 10000, aperture = 3)
  max_cells <- max_cell_id(grid$resolution, grid$aperture)

  stats <- dgearthstat(grid)
  expect_equal(max_cells, stats$n_cells)
})

test_that("max_cell_id handles resolution 0", {
  # Resolution 0 returns 12 (10 * aperture^0 + 2 = 12 cells: 12 icosahedron vertices)
  expect_equal(max_cell_id(0, 3), 12)
  expect_equal(max_cell_id(0, 4), 12)
  expect_equal(max_cell_id(0, 7), 12)
})

# =============================================================================
# CLOSEST RESOLUTION FUNCTIONS
# =============================================================================

test_that("dg_closest_res_to_area finds correct resolution", {
  grid <- list(aperture = 3, topology = "HEXAGON")
  class(grid) <- c("hexify_grid", "dggs", "list")

  # Find resolution for ~1000 km² cells
  res <- dg_closest_res_to_area(grid, area = 1000, metric = TRUE)

  expect_type(res, "double")
  expect_gte(res, 0)
  expect_lte(res, 30)
})

test_that("dg_closest_res_to_area respects rounding", {
  grid <- list(aperture = 3, topology = "HEXAGON")
  class(grid) <- c("hexify_grid", "dggs", "list")

  res_up <- dg_closest_res_to_area(grid, area = 1000, round = "up")
  res_down <- dg_closest_res_to_area(grid, area = 1000, round = "down")
  res_nearest <- dg_closest_res_to_area(grid, area = 1000, round = "nearest")

  expect_gte(res_up, res_down)
})

test_that("dg_closest_res_to_area supports show_info", {
  grid <- list(aperture = 3, topology = "HEXAGON")
  class(grid) <- c("hexify_grid", "dggs", "list")

  expect_message(
    dg_closest_res_to_area(grid, area = 1000, show_info = TRUE),
    "Resolution"
  )
})


# =============================================================================
# RESOLUTION COMPARISON
# =============================================================================

test_that("hexify_compare_resolutions returns correct structure", {
  comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:5)

  expect_s3_class(comparison, "data.frame")
  expect_named(comparison, c(
    "resolution", "n_cells", "cell_area_km2",
    "cell_spacing_km", "cls_km"
  ))
  expect_equal(nrow(comparison), 6)
})

test_that("hexify_compare_resolutions values are monotonic", {
  comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:10)

  # Number of cells should increase with resolution
  expect_true(all(diff(comparison$n_cells) > 0))

  # Cell area should decrease with resolution
  expect_true(all(diff(comparison$cell_area_km2) < 0))
})


# =============================================================================
# ADDITIONAL COVERAGE FOR HEXIFY_STATS
# =============================================================================

test_that("dgearthstat counts aperture 7 cells", {
  grid <- hexify_grid(area = 10000, aperture = 7)
  stats <- dgearthstat(grid)

  expect_equal(stats$n_cells, 10 * 7^stats$resolution + 2)
})

test_that("dg_closest_res_to_area handles non-metric input", {
  grid <- list(aperture = 3, topology = "HEXAGON")
  class(grid) <- c("hexify_grid", "dggs", "list")

  # Square miles input
  res <- dg_closest_res_to_area(grid, area = 386, metric = FALSE)

  expect_type(res, "double")
  expect_gte(res, 0)
})


test_that("hexify_compare_resolutions works for different apertures", {
  comp_ap3 <- hexify_compare_resolutions(aperture = 3, res_range = 0:3)
  comp_ap4 <- hexify_compare_resolutions(aperture = 4, res_range = 0:3)
  comp_ap7 <- hexify_compare_resolutions(aperture = 7, res_range = 0:3)

  expect_equal(nrow(comp_ap3), 4)
  expect_equal(nrow(comp_ap4), 4)
  expect_equal(nrow(comp_ap7), 4)
})

test_that("hexify_compare_resolutions formats large cell counts", {
  # Resolution 11+ should have cells in millions
  expect_output(
    hexify_compare_resolutions(aperture = 3, res_range = 11:12, print = TRUE),
    "M"  # Million suffix
  )
})

test_that("hexify_compare_resolutions formats thousand cell counts", {
  # Resolution 3-5 should have cells in thousands
  expect_output(
    hexify_compare_resolutions(aperture = 3, res_range = 3:5, print = TRUE),
    "K"  # Thousand suffix
  )
})

# plot_globe.R
# Globe visualization for hexified data
#
# Renders hexagonized globe with orthographic projection

# =============================================================================
# CENTER PRESETS
# =============================================================================

#' Globe center presets
#'
#' Named list of lon/lat coordinates for common globe views.
#' Used by \code{\link{plot_globe}} when center is specified as a string.
#'
#' @format Named list with elements:
#' \describe{
#'   \item{europe}{c(10, 50) - Western/Central Europe}
#'   \item{north_america}{c(-100, 45) - USA and Canada}
#'   \item{south_america}{c(-60, -15) - Full continent}
#'   \item{africa}{c(20, 5) - Central Africa}
#'   \item{asia}{c(100, 35) - China, SE Asia, Japan}
#'   \item{oceania}{c(135, -25) - Australia, NZ, Indonesia}
#'   \item{middle_east}{c(45, 25) - Arabian Peninsula, Iran, Turkey}
#'   \item{south_asia}{c(80, 20) - India, Pakistan, Bangladesh}
#'   \item{pacific}{c(-160, -10) - Polynesia, Pacific islands}
#'   \item{caribbean}{c(-70, 18) - Caribbean islands}
#'   \item{arctic}{c(0, 90) - North pole view}
#'   \item{antarctic}{c(0, -90) - South pole view}
#' }
#'
#' @export
#' @examples
#' globe_centers$europe
#' globe_centers$oceania
globe_centers <- list(
  # Continental
  europe        = c(lon = 10,   lat = 50),
  north_america = c(lon = -100, lat = 45),
  south_america = c(lon = -60,  lat = -15),
  africa        = c(lon = 20,   lat = 5),
  asia          = c(lon = 100,  lat = 35),
  oceania       = c(lon = 135,  lat = -25),

  # Regional (fills gaps)
  middle_east   = c(lon = 45,   lat = 25),
  south_asia    = c(lon = 80,   lat = 20),
  pacific       = c(lon = -160, lat = -10),
  caribbean     = c(lon = -70,  lat = 18),

  # Polar
  arctic        = c(lon = 0,    lat = 90),
  antarctic     = c(lon = 0,    lat = -90)
)

# =============================================================================
# MAIN FUNCTION
# =============================================================================

#' Plot hexagonized globe
#'
#' Renders a global hexagonal grid on an orthographic projection with
#' customizable rotation, land clipping, and styling options.
#'
#' @param area Cell area in km^2 (passed to \code{\link{hex_grid}})
#' @param center Globe center: either a preset name (e.g., "europe") or
#'   numeric vector c(lon, lat). See \code{\link{globe_centers}} for presets.
#' @param clip_to_land If TRUE, clip hexagons to land boundaries
#' @param land_data Optional sf object for land boundaries. If NULL and
#'   clip_to_land is TRUE, uses rnaturalearth::ne_countries()
#' @param exclude_antarctica If TRUE, exclude Antarctica from land clipping
#' @param fill Fill color for hexagons (default "#D4B896")
#' @param border Border color for hexagons (default "grey30")
#' @param border_width Border width for hexagons (default 0.2)
#' @param ocean_fill Fill color for ocean/globe background (default "white")
#' @param ocean_border Border color for globe circle (default "grey50")
#' @param show_land If TRUE, show land boundaries (default TRUE when clipping)
#' @param land_fill Fill color for land (default NA, transparent)
#' @param land_border Border color for land boundaries (default "grey40")
#' @param land_width Border width for land boundaries (default 0.3)
#' @param use_ggplot NULL = auto-detect, TRUE = force ggplot2, FALSE = force base
#' @param return_data If TRUE, return sf objects instead of plotting
#' @param aperture Grid aperture (default 3L)
#'
#' @return If use_ggplot = TRUE: ggplot2 object (can add layers with +)
#'   If use_ggplot = FALSE: NULL invisibly (plots directly)
#'   If return_data = TRUE: list of sf objects (hexagons, land, ocean_circle, crs)
#'
#' @details
#' The function handles several technical challenges:
#' \itemize{
#'   \item Hexagons on the back side of the globe fail to transform - these are
#'     filtered out gracefully
#'   \item Invalid geometries after projection are repaired with st_buffer(0)
#'   \item Clipping is done in orthographic CRS to avoid topology errors
#' }
#'
#' @seealso \code{\link{globe_centers}} for available presets,
#'   \code{\link{grid_global}} for generating global grids without plotting
#'
#' @export
#' @examples
#' \donttest{
#' # Get data for custom plotting (no rendering)
#' data <- plot_globe(area = 100000, center = "europe", return_data = TRUE)
#' nrow(data$hexagons)
#' class(data$ocean_circle)
#'
#' # Basic usage - Europe-centered globe
#' plot_globe(area = 80000, center = "europe")
#' }
plot_globe <- function(
    area = 50000,
    center = "europe",
    clip_to_land = FALSE,
    land_data = NULL,
    exclude_antarctica = TRUE,
    fill = "#D4B896",
    border = "grey30",
    border_width = 0.2,
    ocean_fill = "white",
    ocean_border = "grey50",
    show_land = clip_to_land,
    land_fill = NA,
    land_border = "grey40",
    land_width = 0.3,
    use_ggplot = NULL,
    return_data = FALSE,
    aperture = 3L
) {
  # Check sf dependency

if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required. Install with: install.packages('sf')")
  }

  # Resolve center
  center_coords <- resolve_center(center)

  # Auto-detect ggplot2 if not specified
  if (is.null(use_ggplot)) {
    use_ggplot <- requireNamespace("ggplot2", quietly = TRUE)
  }

  # Prepare data
  data <- prepare_globe_data(
    area = area,
    center = center_coords,
    clip_to_land = clip_to_land,
    land_data = land_data,
    exclude_antarctica = exclude_antarctica,
    show_land = show_land,
    aperture = aperture
  )

  # Return data if requested
  if (return_data) {
    return(data)
  }

  # Plot with appropriate backend
  if (use_ggplot) {
    if (!requireNamespace("ggplot2", quietly = TRUE)) {
      stop("Package 'ggplot2' is required when use_ggplot = TRUE")
    }
    plot_globe_ggplot(
      data = data,
      fill = fill,
      border = border,
      border_width = border_width,
      ocean_fill = ocean_fill,
      ocean_border = ocean_border,
      show_land = show_land,
      land_fill = land_fill,
      land_border = land_border,
      land_width = land_width
    )
  } else {
    plot_globe_base(
      data = data,
      fill = fill,
      border = border,
      ocean_fill = ocean_fill,
      ocean_border = ocean_border,
      show_land = show_land,
      land_fill = land_fill,
      land_border = land_border
    )
  }
}

# =============================================================================
# HELPER FUNCTIONS
# =============================================================================

#' Resolve center preset or coordinates
#' @noRd
resolve_center <- function(center) {
  if (is.character(center)) {
    if (!center %in% names(globe_centers)) {
      valid <- paste(names(globe_centers), collapse = ", ")
      stop(sprintf("Unknown center preset '%s'. Valid options: %s", center, valid))
    }
    return(globe_centers[[center]])
  }

  if (is.numeric(center) && length(center) == 2) {
    # Handle named or unnamed numeric vector
    if (is.null(names(center))) {
      return(c(lon = center[1], lat = center[2]))
    }
    if (!setequal(names(center), c("lon", "lat"))) {
      stop(sprintf(
        "center must be named c(lon = ..., lat = ...); got names: %s",
        paste(names(center), collapse = ", ")
      ))
    }
    return(center[c("lon", "lat")])
  }

  stop("center must be a preset name or numeric c(lon, lat)")
}

#' Prepare globe data for plotting
#' @noRd
prepare_globe_data <- function(
    area,
    center,
    clip_to_land,
    land_data,
    exclude_antarctica,
    show_land,
    aperture
) {
  # Create orthographic CRS
  crs_string <- sprintf(
    "+proj=ortho +lat_0=%f +lon_0=%f +x_0=0 +y_0=0 +datum=WGS84",
    center["lat"], center["lon"]
  )

  # Disable S2 early - some polar geometries are invalid in spherical mode
  s2_state <- sf::sf_use_s2()
  sf::sf_use_s2(FALSE)
  on.exit(sf::sf_use_s2(s2_state), add = TRUE)

  # Generate global grid — skip dateline wrapping because orthographic
  # projection handles antimeridian cells natively; wrapping splits them
  # into MULTIPOLYGON halves that create visible gaps on the globe
  grid <- hex_grid(area_km2 = area, aperture = aperture)
  global_hex <- grid_global(grid, wrap_dateline = FALSE)

  # Pre-filter hexagons by angular distance (optimization)
  # Keep cells within ~95 degrees of center (visible hemisphere + margin)
  centers_ll <- suppressWarnings(sf::st_centroid(global_hex))
  coords <- sf::st_coordinates(centers_ll)

  # Angular distance calculation (spherical)
  lon1 <- center["lon"] * pi / 180
  lat1 <- center["lat"] * pi / 180
  lon2 <- coords[, 1] * pi / 180
  lat2 <- coords[, 2] * pi / 180

  # Haversine-based angular distance
  delta_lon <- lon2 - lon1
  delta_lat <- lat2 - lat1
  a <- sin(delta_lat / 2)^2 + cos(lat1) * cos(lat2) * sin(delta_lon / 2)^2
  angular_dist <- 2 * asin(sqrt(pmin(1, a))) * 180 / pi

  # Keep hexagons within visible hemisphere (with margin for edge cells)
  visible_idx <- angular_dist < 95
  global_hex <- global_hex[visible_idx, ]

  # Transform hexagons to orthographic - handle failures gracefully
  hex_ortho <- tryCatch({
    suppressWarnings(sf::st_transform(global_hex, crs_string))
  }, error = function(e) {
    # If batch transform fails, do one by one
    transform_individually(global_hex, crs_string)
  })

  # Remove empty geometries
  hex_ortho <- hex_ortho[!sf::st_is_empty(hex_ortho), ]

  # Fix invalid geometries using st_make_valid
  # MULTIPOLYGON cells (from dateline crossing) often become invalid after
  # orthographic transform due to self-intersecting parts - st_make_valid fixes this
  # Cells on the hemisphere edge collapse to rings of two points, which GEOS
  # rejects for the whole set, so drop those before repairing the rest
  hex_ortho <- hex_ortho[has_closable_rings(sf::st_geometry(hex_ortho)), ]
  hex_ortho <- tryCatch({
    sf::st_make_valid(hex_ortho)
  }, error = function(e) {
    make_valid_per_cell(hex_ortho)
  })

  # Remove empty and degenerate geometries
  # Cells at hemisphere edge get "squashed" to tiny areas - filter them out
  hex_ortho <- hex_ortho[!sf::st_is_empty(hex_ortho), ]
  areas <- tryCatch(as.numeric(sf::st_area(hex_ortho)), error = function(e) rep(NA, nrow(hex_ortho)))
  # Expected cell area is roughly area * 1e6 m² - filter cells < 1% of expected
  min_area <- area * 1e6 * 0.01  # 1% of expected area
  hex_ortho <- hex_ortho[!is.na(areas) & areas > min_area, ]

  # Create ocean circle (globe boundary)
  earth_radius <- 6371000  # meters
  ocean_circle <- create_globe_circle(earth_radius, crs_string)

  # Clip hexagons to globe circle - use st_crop for robustness
  hex_ortho <- tryCatch({
    # Get bounding box of ocean circle for cropping
    bbox <- sf::st_bbox(ocean_circle)
    cropped <- suppressWarnings(sf::st_crop(hex_ortho, bbox))
    # Then intersect with circle
    suppressWarnings(sf::st_intersection(cropped, ocean_circle))
  }, error = function(e) {
    # Fallback: just filter by centroid within circle
    tryCatch({
      centroids <- suppressWarnings(sf::st_centroid(hex_ortho))
      within <- suppressWarnings(sf::st_within(centroids, ocean_circle, sparse = FALSE))
      hex_ortho[apply(within, 1, any), ]
    }, error = function(e2) hex_ortho)
  })

  # Keep only polygon geometries
  geom_types <- sf::st_geometry_type(hex_ortho)
  hex_ortho <- hex_ortho[geom_types %in% c("POLYGON", "MULTIPOLYGON"), ]

  # Remove any empty geometries that may have resulted from clipping
  hex_ortho <- hex_ortho[!sf::st_is_empty(hex_ortho), ]

  # Handle land data
  land_ortho <- NULL
  if (clip_to_land || show_land) {
    land_ortho <- prepare_land_data(
      land_data = land_data,
      exclude_antarctica = exclude_antarctica,
      crs_string = crs_string,
      ocean_circle = ocean_circle
    )

    # Clip hexagons to land if requested
    if (clip_to_land && !is.null(land_ortho)) {
      # Filter out any malformed geometries (NA validity) before intersection
      validity <- sf::st_is_valid(hex_ortho)
      hex_ortho <- hex_ortho[!is.na(validity) & validity, ]

      hex_ortho <- tryCatch({
        result <- suppressWarnings(sf::st_intersection(hex_ortho, land_ortho))
        geom_types <- sf::st_geometry_type(result)
        result[geom_types %in% c("POLYGON", "MULTIPOLYGON"), ]
      }, error = function(e) hex_ortho)
    }
  }

  list(
    hexagons = hex_ortho,
    land = land_ortho,
    ocean_circle = ocean_circle,
    crs = crs_string,
    center = center
  )
}

#' Which geometries have rings that can close into a polygon
#'
#' A closed linear ring needs four positions, so a ring carrying fewer is a
#' cell the projection squashed onto the hemisphere edge.
#' @noRd
has_closable_rings <- function(geoms) {
  vapply(geoms, function(g) {
    rings <- if (inherits(g, "MULTIPOLYGON")) unlist(g, recursive = FALSE) else g
    length(rings) > 0L && all(vapply(rings, nrow, integer(1)) >= 4L)
  }, logical(1))
}

#' Repair geometries one cell at a time (fallback for batch failures)
#' @noRd
make_valid_per_cell <- function(sf_obj) {
  geoms <- sf::st_geometry(sf_obj)
  fixed <- vector("list", length(geoms))
  keep <- logical(length(geoms))

  for (i in seq_along(geoms)) {
    repaired <- tryCatch({
      sf::st_make_valid(geoms[i])
    }, error = function(e) {
      tryCatch(sf::st_buffer(geoms[i], 0), error = function(e2) NULL)
    })

    if (!is.null(repaired) && !sf::st_is_empty(repaired)[1]) {
      fixed[[i]] <- repaired[[1]]
      keep[i] <- TRUE
    }
  }

  out <- sf_obj[keep, ]
  sf::st_geometry(out) <- sf::st_sfc(fixed[keep], crs = sf::st_crs(geoms))
  out
}

#' Transform geometries individually (fallback for batch failures)
#' @noRd
transform_individually <- function(sf_obj, crs) {
  valid_idx <- logical(nrow(sf_obj))
  geoms <- vector("list", nrow(sf_obj))

  for (i in seq_len(nrow(sf_obj))) {
    result <- tryCatch({
      suppressWarnings(sf::st_transform(sf_obj[i, ], crs))
    }, error = function(e) NULL)

    if (!is.null(result) && !sf::st_is_empty(result)) {
      valid_idx[i] <- TRUE
      geoms[[i]] <- sf::st_geometry(result)[[1]]
    }
  }

  if (sum(valid_idx) == 0) {
    stop("No hexagons could be transformed to orthographic projection")
  }

  sf_obj <- sf_obj[valid_idx, ]
  sf::st_geometry(sf_obj) <- sf::st_sfc(geoms[valid_idx], crs = crs)
  sf_obj
}

#' Create globe boundary circle
#' @noRd
create_globe_circle <- function(radius, crs) {
  # Create circle in orthographic CRS
  angles <- seq(0, 2 * pi, length.out = 100)
  x <- radius * cos(angles)
  y <- radius * sin(angles)
  coords <- cbind(x, y)
  coords <- rbind(coords, coords[1, ])  # Close polygon

  circle <- sf::st_polygon(list(coords))
  sf::st_sfc(circle, crs = crs)
}

#' Prepare land data for globe
#' @noRd
prepare_land_data <- function(land_data, exclude_antarctica, crs_string, ocean_circle) {
  # Disable S2 for land data processing
  s2_state <- sf::sf_use_s2()
  sf::sf_use_s2(FALSE)
  on.exit(sf::sf_use_s2(s2_state), add = TRUE)

  # Get land data
  if (is.null(land_data)) {
    if (!requireNamespace("rnaturalearth", quietly = TRUE)) {
      warning("Package 'rnaturalearth' required for land boundaries. ",
              "Install with: install.packages('rnaturalearth')")
      return(NULL)
    }
    land_data <- rnaturalearth::ne_countries(scale = "medium", returnclass = "sf")
  }

  # Exclude Antarctica if requested
  if (exclude_antarctica) {
    # Try common column names for country filtering
    if ("sovereignt" %in% names(land_data)) {
      land_data <- land_data[land_data$sovereignt != "Antarctica", ]
    } else if ("name" %in% names(land_data)) {
      land_data <- land_data[land_data$name != "Antarctica", ]
    } else if ("NAME" %in% names(land_data)) {
      land_data <- land_data[land_data$NAME != "Antarctica", ]
    } else {
      warning("exclude_antarctica = TRUE but land_data has none of the ",
              "recognized country-name columns ('sovereignt', 'name', 'NAME'); ",
              "Antarctica was not excluded.")
    }
  }

  # Transform individual countries to orthographic FIRST, then union
  # (Union in lat/lon then transform creates empty geometries for global polygons)
  land_ortho <- tryCatch({
    transformed <- suppressWarnings(sf::st_transform(land_data, crs_string))
    transformed <- transformed[!sf::st_is_empty(transformed), ]

    if (nrow(transformed) == 0) return(NULL)

    # Union in orthographic space - use st_combine + st_union for robustness
    land_geom <- tryCatch({
      combined <- sf::st_combine(sf::st_geometry(transformed))
      suppressWarnings(sf::st_union(combined))
    }, error = function(e) {
      # If union fails, just return combined geometries
      sf::st_combine(sf::st_geometry(transformed))
    })

    suppressWarnings(sf::st_buffer(land_geom, 0))
  }, error = function(e) NULL)

  if (is.null(land_ortho) || length(land_ortho) == 0) return(NULL)

  # Remove empty geometries
  land_ortho <- land_ortho[!sf::st_is_empty(land_ortho)]
  if (length(land_ortho) == 0) return(NULL)

  # Clip to globe circle
  land_ortho <- tryCatch({
    result <- suppressWarnings(sf::st_intersection(land_ortho, ocean_circle))
    sf::st_buffer(result, 0)
  }, error = function(e) land_ortho)

  # Fix any degenerate geometries (e.g., rings with < 3 points)
  # Use a tiny buffer (1 meter) to clean up degenerate rings, then unbuffer
  land_ortho <- tryCatch({
    buffered <- sf::st_buffer(land_ortho, 1)  # 1 meter buffer
    unbuffered <- sf::st_buffer(buffered, -1)  # Remove the buffer
    sf::st_make_valid(unbuffered)
  }, error = function(e) {
    # Fallback: just try st_make_valid
    tryCatch(sf::st_make_valid(land_ortho), error = function(e2) land_ortho)
  })

  land_ortho
}

# =============================================================================
# PLOTTING BACKENDS
# =============================================================================

#' Plot globe with ggplot2
#' @noRd
plot_globe_ggplot <- function(
    data,
    fill,
    border,
    border_width,
    ocean_fill,
    ocean_border,
    show_land,
    land_fill,
    land_border,
    land_width
) {
  ggplot2::ggplot() +
    # Ocean/globe background
    ggplot2::geom_sf(
      data = sf::st_sf(geometry = data$ocean_circle),
      fill = ocean_fill,
      color = ocean_border,
      linewidth = 0.5
    ) +
    # Land boundaries (if showing)
    {if (show_land && !is.null(data$land))
      ggplot2::geom_sf(
        data = sf::st_sf(geometry = data$land),
        fill = land_fill,
        color = land_border,
        linewidth = land_width
      )
    } +
    # Hexagons
    ggplot2::geom_sf(
      data = data$hexagons,
      fill = fill,
      color = border,
      linewidth = border_width
    ) +
    ggplot2::coord_sf(crs = data$crs) +
    ggplot2::theme_void() +
    ggplot2::theme(
      panel.background = ggplot2::element_rect(fill = "transparent", color = NA),
      plot.background = ggplot2::element_rect(fill = "transparent", color = NA)
    )
}

#' Plot globe with base R graphics
#' @noRd
plot_globe_base <- function(
    data,
    fill,
    border,
    ocean_fill,
    ocean_border,
    show_land,
    land_fill,
    land_border
) {
  # Set up plot area
  bbox <- sf::st_bbox(data$ocean_circle)
  plot(NULL,
       xlim = c(bbox["xmin"], bbox["xmax"]),
       ylim = c(bbox["ymin"], bbox["ymax"]),
       asp = 1,
       axes = FALSE,
       xlab = "",
       ylab = "")

  # Draw ocean circle
  plot(data$ocean_circle, col = ocean_fill, border = ocean_border, add = TRUE)

  # Draw land if showing
  if (show_land && !is.null(data$land)) {
    plot(data$land, col = land_fill, border = land_border, add = TRUE)
  }

  # Draw hexagons
  plot(sf::st_geometry(data$hexagons), col = fill, border = border, add = TRUE)

  invisible(NULL)
}

# tests/testthat/test-globe.R
# Tests for globe_centers, plot_globe(), and the internal resolve_center()
# helper (R/plot_globe.R). Previously had zero test coverage.

# =============================================================================
# globe_centers
# =============================================================================

test_that("globe_centers is a named list of lon/lat presets", {
  expect_type(globe_centers, "list")
  expect_true(all(c("europe", "north_america", "south_america", "africa",
                     "asia", "oceania", "middle_east", "south_asia",
                     "pacific", "caribbean", "arctic", "antarctic")
                   %in% names(globe_centers)))

  for (nm in names(globe_centers)) {
    preset <- globe_centers[[nm]]
    expect_length(preset, 2)
    expect_equal(names(preset), c("lon", "lat"), info = nm)
    expect_true(preset["lon"] >= -180 && preset["lon"] <= 180, info = nm)
    expect_true(preset["lat"] >= -90 && preset["lat"] <= 90, info = nm)
  }
})

test_that("globe_centers$europe matches documented value", {
  expect_equal(unname(globe_centers$europe), c(10, 50))
})

# =============================================================================
# resolve_center() (internal helper used by plot_globe())
# =============================================================================

test_that("resolve_center accepts a valid preset name", {
  expect_equal(hexify:::resolve_center("europe"), globe_centers$europe)
})

test_that("resolve_center errors on an unknown preset name", {
  expect_error(hexify:::resolve_center("nowhere"), "Unknown center preset")
})

test_that("resolve_center accepts numeric c(lon, lat), named or not", {
  expect_equal(hexify:::resolve_center(c(lon = 5, lat = 40)), c(lon = 5, lat = 40))
  expect_equal(hexify:::resolve_center(c(5, 40)), c(lon = 5, lat = 40))
})

test_that("resolve_center errors on invalid input", {
  expect_error(hexify:::resolve_center(c(1, 2, 3)), "preset name or numeric")
  expect_error(hexify:::resolve_center(TRUE), "preset name or numeric")
})

# =============================================================================
# plot_globe()
# =============================================================================

test_that("plot_globe(return_data = TRUE) returns valid, back-side-filtered geometry", {
  skip_on_cran()  # Slow sf/orthographic-projection operations
  skip_if_not_installed("sf")

  data <- plot_globe(area = 500000, center = "europe", return_data = TRUE)

  expect_type(data, "list")
  expect_true("hexagons" %in% names(data))
  expect_true("ocean_circle" %in% names(data))
  expect_s3_class(data$hexagons, "sf")

  # Back-side hexagons are filtered out: fewer cells than a full global grid
  full_grid <- hex_grid(area_km2 = 500000)
  full_count <- 10 * as.integer(full_grid@aperture)^full_grid@resolution + 2
  expect_true(nrow(data$hexagons) < full_count)

  # Invalid geometries are repaired (st_buffer(0)) before being returned
  expect_true(all(sf::st_is_valid(data$hexagons)))
})

test_that("plot_globe(return_data = TRUE) works with a custom numeric center", {
  skip_on_cran()  # Slow sf/orthographic-projection operations
  skip_if_not_installed("sf")

  data <- plot_globe(area = 500000, center = c(0, 0), return_data = TRUE)
  expect_s3_class(data$hexagons, "sf")
  expect_true(nrow(data$hexagons) > 0)
})

test_that("plot_globe errors on an unknown center preset", {
  skip_if_not_installed("sf")
  expect_error(plot_globe(area = 500000, center = "nowhere"), "Unknown center preset")
})

# =============================================================================
# Aperture sequences
# =============================================================================
#
# An ISEA grid refines by one aperture per resolution level. hex_grid() takes
# that as a family name or as a per-level vector:
#
#   "3", "4", "7"    every level refines by that aperture
#   "4/3", "4/7"     the first floor(resolution / 2) levels refine by the first
#                    aperture and the rest by the second, which is how DGGRID
#                    arranges ISEA43H
#   c(4, 4, 7, 3)    one aperture per level, in order, length = resolution
#
# A grid stores the spelling as a string ("4/7", "4/4/7/3") and
# parse_aperture_seq() turns it back into the sequence the C++ layer takes:
# entry 1 names the base grid and the rest are the refinement steps, so its
# length is resolution + 1.

#' Is this aperture spelling a mixed sequence?
#' @param aperture Character aperture spelling
#' @noRd
is_mixed_aperture <- function(aperture) {
  grepl("/", as.character(aperture), fixed = TRUE)
}

#' Aperture spelling to store on a grid
#'
#' A family name passes through; a per-level vector is joined with "/".
#' @param aperture Character family name or numeric vector of apertures
#' @param resolution Integer resolution the vector spelling is given for
#' @return Single character string
#' @noRd
format_aperture <- function(aperture, resolution) {
  if (length(aperture) > 1L) {
    parts <- as.integer(aperture)
    if (anyNA(parts) || !all(parts %in% VALID_APERTURES)) {
      stop(sprintf("Aperture sequence entries must be one of: %s",
                   paste(VALID_APERTURES, collapse = ", ")))
    }
    if (length(parts) != resolution) {
      stop(sprintf(
        "An aperture sequence needs one aperture per resolution level: %d given for resolution %d",
        length(parts), as.integer(resolution)
      ))
    }
    return(paste(parts, collapse = "/"))
  }
  as.character(aperture)
}

#' Aperture of every level of a grid
#'
#' @param aperture Character aperture spelling
#' @param resolution Integer resolution
#' @return Integer vector of length resolution + 1: the base grid followed by
#'   one aperture per refinement step
#' @noRd
parse_aperture_seq <- function(aperture, resolution) {
  resolution <- as.integer(resolution)
  parts <- as.integer(strsplit(as.character(aperture), "/", fixed = TRUE)[[1]])

  if (anyNA(parts) || !all(parts %in% VALID_APERTURES)) {
    stop(sprintf("Aperture must be one of %s, a family such as \"4/3\", or one aperture per level",
                 paste(VALID_APERTURES, collapse = ", ")))
  }

  if (length(parts) == 1L) {
    steps <- rep(parts, resolution)
  } else if (length(parts) == resolution) {
    steps <- parts
  } else if (length(parts) == 2L) {
    level <- as.integer(resolution / 2)
    steps <- c(rep(parts[1], level), rep(parts[2], resolution - level))
  } else {
    stop(sprintf(
      "Aperture \"%s\" names %d levels, resolution %d needs %d",
      aperture, length(parts), resolution, resolution
    ))
  }

  base <- if (length(steps) > 0L) steps[1] else parts[1]
  as.integer(c(base, steps))
}

#' Aperture sequence of a grid object
#' @param g HexGridInfo object
#' @noRd
grid_ap_seq <- function(g) {
  parse_aperture_seq(g@aperture, g@resolution)
}

#' The same spelling read at a coarser resolution
#'
#' A family name applies at every resolution, so it passes through. A per-level
#' spelling names one aperture per level of its own resolution, so a coarser
#' grid takes the leading levels.
#' @param aperture Character aperture spelling
#' @param resolution Integer resolution to read it at
#' @noRd
aperture_at_resolution <- function(aperture, resolution) {
  parts <- strsplit(as.character(aperture), "/", fixed = TRUE)[[1]]
  if (length(parts) <= 2L || resolution >= length(parts)) return(as.character(aperture))
  paste(parts[seq_len(max(resolution, 1L))], collapse = "/")
}

#' Cell count of an aperture sequence
#'
#' N = 10 * (product of the refinement apertures) + 2, the formula
#' calc_grid_params_mixed() in src/rcpp_cell.cpp packs cell IDs for.
#' @param ap_seq Integer aperture sequence
#' @noRd
ap_seq_n_cells <- function(ap_seq) {
  10 * prod(as.numeric(ap_seq[-1])) + 2
}

#' Cell count of an aperture spelling at a resolution
#' @param aperture Character aperture spelling
#' @param resolution Integer resolution
#' @noRd
aperture_n_cells <- function(aperture, resolution) {
  if (is_mixed_aperture(aperture)) {
    ap_seq_n_cells(parse_aperture_seq(aperture, resolution))
  } else {
    max_cell_id(resolution, as.integer(aperture))
  }
}

#' Resolution whose cells have a target area, for a mixed sequence
#'
#' Cell area is the sphere over the cell count, and for a sequence that count is
#' a product rather than a power, so the resolution comes from the two levels
#' the target falls between, interpolated on the log-area scale that a pure
#' aperture's closed form gives exactly.
#' @param area_km2 Target cell area
#' @param aperture Character aperture spelling
#' @param radius_km Radius of the body, in kilometers
#' @return Numeric resolution, not rounded
#' @noRd
calculate_resolution_for_area_mixed <- function(area_km2, aperture,
                                                radius_km = EARTH_RADIUS_KM) {
  surface_km2 <- body_surface_km2(radius_km)
  res <- seq.int(MIN_RESOLUTION, MAX_RESOLUTION)
  log_area <- vapply(res, function(r) {
    log(surface_km2 / aperture_n_cells(aperture, r))
  }, numeric(1))
  target <- log(area_km2)

  if (target >= log_area[1]) return(as.numeric(MIN_RESOLUTION))
  if (target <= log_area[length(log_area)]) return(as.numeric(MAX_RESOLUTION))

  k <- max(which(log_area > target))
  frac <- (log_area[k] - target) / (log_area[k] - log_area[k + 1])
  res[k] + frac
}

# Figure defaults shared by every vignette.
#
# A figure is shown at the full width of the reader's viewport and scales to it,
# so its text arrives at the authored point size times viewport width over
# figure width. A 390 px phone showing a 7 inch figure scales by about 0.77,
# which is what puts the device default of 12 pt under the size the text stays
# legible at. Sizing against that viewport keeps every figure readable, and the
# site's stylesheet scales the text back down as the column widens.
#
# Figures are one width throughout, so a single point size serves all of them.
# A chunk that needs a different shape sets fig.height alone.

FIG_WIDTH <- 7
FIG_HEIGHT <- 5
FIG_POINTSIZE <- 16
FIG_BASE_SIZE <- 18

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = FIG_WIDTH,
  fig.height = FIG_HEIGHT,
  dev.args = list(pointsize = FIG_POINTSIZE)
)

# ggplot text is absolute points carried by the theme, so the device point size
# above does not reach it. hexify's own plots read this option for their base.
old_base_size <- options(hexify.base_size = FIG_BASE_SIZE)

# Put the option back once the article has been knit.
knitr::knit_hooks$set(document = function(x) { options(old_base_size); x })

#' Simplified World Map
#'
#' A lightweight sf object containing simplified world country borders,
#' suitable for use as a basemap when visualizing hexagonal grids.
#'
#' @format An sf object with 177 features and 15 fields:
#' \describe{
#'   \item{name}{Country short name}
#'   \item{name_long}{Country full name}
#'   \item{admin}{Administrative name}
#'   \item{sovereignt}{Sovereignty}
#'   \item{iso_a2}{ISO 3166-1 alpha-2 country code}
#'   \item{iso_a3}{ISO 3166-1 alpha-3 country code}
#'   \item{iso_n3}{ISO 3166-1 numeric code}
#'   \item{continent}{Continent name}
#'   \item{region_un}{UN region}
#'   \item{subregion}{UN subregion}
#'   \item{region_wb}{World Bank region}
#'   \item{pop_est}{Population estimate}
#'   \item{gdp_md}{GDP in millions USD}
#'   \item{income_grp}{Income group classification}
#'   \item{economy}{Economy type}
#'   \item{geometry}{MULTIPOLYGON geometry in 'WGS84' (EPSG:4326)}
#' }
#'
#' @source Simplified from Natural Earth 1:110m Cultural Vectors
#'   (\url{https://www.naturalearthdata.com/})
#'
#' @examples
#' library(sf)
#'
#' # Plot the built-in world map
#' plot(st_geometry(hexify_world), col = "lightgray", border = "white")
#'
#' # Filter by continent
#' europe <- hexify_world[hexify_world$continent == "Europe", ]
#' plot(st_geometry(europe))
"hexify_world"

# hexify.R
# Main user-facing convenience function
#
# This is the primary entry point for the hexify package.

#' Assign hexagonal DGGS cell IDs to geographic points
#'
#' Takes a data.frame or sf object with geographic coordinates and returns
#' a HexData object that stores the original data plus cell assignments.
#' The original data is preserved unchanged; cell IDs and centers are stored
#' in separate slots.
#'
#' @param data A data.frame or sf object containing coordinates
#' @param grid A HexGridInfo object from \code{hex_grid()}. If provided, overrides
#'   area_km2, resolution, aperture and radius_km parameters.
#' @param lon Column name for longitude (ignored if data is sf)
#' @param lat Column name for latitude (ignored if data is sf)
#' @param area_km2 Target cell area in km^2 (mutually exclusive with diagonal).
#' @param diagonal Target cell diagonal (long diagonal) in km
#' @param resolution Grid resolution (0-30). Alternative to area_km2.
#' @param aperture Grid aperture: 3, 4, 7, a mixed family such as "4/3" or
#'   "4/7", or one aperture per resolution level, e.g. \code{c(4, 4, 7, 3)}
#'   (default 3)
#' @param resround How to round resolution: "nearest", "up", or "down"
#' @param radius_km Radius of the body the grid covers, in kilometers, or a body
#'   name such as "mars" (default Earth). See \code{\link{hex_grid}}.
#'
#' @return A HexData object containing:
#'   \itemize{
#'     \item \code{data}: The original input data (unchanged)
#'     \item \code{grid}: The HexGridInfo specification
#'     \item \code{cell_id}: Numeric vector of cell IDs for each row
#'     \item \code{cell_center}: Matrix of cell center coordinates (lon, lat)
#'   }
#'
#'   Use \code{as.data.frame(result)} to extract the original data.
#'   Use \code{cells(result)} to get unique cell IDs.
#'   Use \code{result@@cell_id} to get all cell IDs.
#'   Use \code{result@@cell_center} to get cell center coordinates.
#'
#' @details
#' For sf objects, coordinates are automatically extracted and transformed to
#' 'WGS84' (EPSG:4326) if needed. The geometry column is preserved.
#'
#' Either \code{area_km2}, \code{diagonal}, or \code{resolution}
#' must be provided unless a \code{grid} object is supplied.
#'
#' The HexData return type (default) stores the grid specification so downstream
#' functions like \code{plot()}, \code{hexify_cell_to_sf()}, etc. don't need
#' grid parameters repeated.
#'
#' @section Grid Specification:
#' You can create a grid specification once and reuse it:
#' \preformatted{
#' grid <- hex_grid(area_km2 = 1000)
#' result1 <- hexify(df1, grid = grid)
#' result2 <- hexify(df2, grid = grid)
#' }
#'
#' @family hexify main
#' @seealso \code{\link{hex_grid}} for grid specification,
#'   \code{\link{HexData-class}} for return object details,
#'   \code{\link{as_sf}} for converting to sf
#' @export
#' @examples
#' # Simple data.frame
#' df <- data.frame(
#'   site = c("Vienna", "Paris", "Madrid"),
#'   lon = c(16.37, 2.35, -3.70),
#'   lat = c(48.21, 48.86, 40.42)
#' )
#'
#' # New recommended workflow: use grid object
#' grid <- hex_grid(area_km2 = 1000)
#' result <- hexify(df, grid = grid, lon = "lon", lat = "lat")
#' print(result)  # Shows grid info
#' plot(result)   # Plot with default styling
#'
#' # Direct area specification (grid created internally)
#' result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
#'
#' # Extract plain data.frame
#' df_result <- as.data.frame(result)
#'
#' # With sf object (any CRS)
#' library(sf)
#' pts <- st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
#' result_sf <- hexify(pts, area_km2 = 1000)
#'
#' # Different apertures
#' result_ap4 <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 4)
#'
#' # Mixed aperture (ISEA43H)
#' result_mixed <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = "4/3")
hexify <- function(data,
                   grid = NULL,
                   lon = "lon",
                   lat = "lat",
                   area_km2 = NULL,
                   diagonal = NULL,
                   resolution = NULL,
                   aperture = 3,
                   resround = "nearest",
                   radius_km = EARTH_RADIUS_KM) {

  # -------------------------------------------------------------------------
  # Extract or build grid specification
  # -------------------------------------------------------------------------
  if (!is.null(grid)) {
    # Grid object provided - extract parameters
    if (is_hex_grid(grid)) {
      hex_grid_obj <- grid
    } else if (inherits(grid, "hexify_grid")) {
      # Legacy S3 grid object
      hex_grid_obj <- hexify_grid_to_HexGridInfo(grid)
    } else {
      stop("grid must be a HexGridInfo object from hex_grid() or legacy hexify_grid")
    }
  } else {
    # Build grid from parameters
    if (is.null(area_km2) && is.null(diagonal) && is.null(resolution)) {
      stop("Either 'grid', 'area_km2', 'diagonal', or 'resolution' must be provided")
    }
    if (!is.null(area_km2) && !is.null(diagonal)) {
      stop("Provide either 'area_km2' or 'diagonal', not both")
    }

    # Convert diagonal to area if provided
    if (!is.null(diagonal)) {
      area_km2 <- diagonal^2 * sqrt(3) / 2
    }

    # Create HexGridInfo object (hex_grid handles aperture parsing)
    hex_grid_obj <- hex_grid(
      area_km2 = area_km2,
      resolution = resolution,
      aperture = aperture,
      resround = resround,
      radius_km = radius_km
    )
  }

  # -------------------------------------------------------------------------
  # Extract coordinates from data
  # -------------------------------------------------------------------------
  is_sf <- inherits(data, "sf")

  if (is_sf) {
    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("Package 'sf' is required to process sf objects")
    }

    # Get coordinates, transforming to WGS84 if needed. A missing CRS (not
    # merely one without a mapped EPSG code) can't be transformed from, so
    # treat it as already lon/lat rather than erroring; otherwise compare
    # full CRS objects (not just $epsg, which is NA for many valid CRS built
    # from WKT/proj4 strings) and transform when they differ.
    data_crs <- sf::st_crs(data)
    if (is.na(data_crs) || data_crs == sf::st_crs(4326)) {
      coords_sf <- data
    } else {
      coords_sf <- sf::st_transform(data, 4326)
    }

    coords <- sf::st_coordinates(coords_sf)
    lon_vec <- coords[, 1]
    lat_vec <- coords[, 2]
  } else {
    # Regular data.frame
    if (!lon %in% names(data)) {
      stop(sprintf("Column '%s' not found in data", lon))
    }
    if (!lat %in% names(data)) {
      stop(sprintf("Column '%s' not found in data", lat))
    }

    lon_vec <- data[[lon]]
    lat_vec <- data[[lat]]
  }

  # Validate coordinates
  if (!is.numeric(lon_vec) || !is.numeric(lat_vec)) {
    stop("Coordinates must be numeric")
  }
  validate_lon(lon_vec)
  validate_lat(lat_vec)

  na_mask <- is.na(lon_vec) | is.na(lat_vec)
  if (all(na_mask)) {
    stop("All coordinates are NA")
  }
  if (any(na_mask)) {
    warning(sprintf("%d coordinate pairs contain NA values and will be skipped",
                    sum(na_mask)))
    data <- data[!na_mask, , drop = FALSE]
    lon_vec <- lon_vec[!na_mask]
    lat_vec <- lat_vec[!na_mask]
  }

  # -------------------------------------------------------------------------
  # Perform hexification
  # -------------------------------------------------------------------------
  aperture_str <- hex_grid_obj@aperture
  res <- hex_grid_obj@resolution

  if (is_h3_grid(hex_grid_obj)) {
    # H3 path: use native C backend
    cell_ids <- cpp_h3_latLngToCell(lon_vec, lat_vec, res)
    center_df <- cpp_h3_cellToLatLng(cell_ids)
    centers <- list(lon_deg = center_df$lon, lat_deg = center_df$lat)
  } else if (is_mixed_aperture(aperture_str)) {
    ap_seq <- parse_aperture_seq(aperture_str, res)
    cell_ids <- cpp_lonlat_to_cell_seq(lon_vec, lat_vec, ap_seq)
    centers <- cpp_cell_to_lonlat_seq(cell_ids, ap_seq)
  } else {
    aperture_num <- match(aperture_str, c("3", "4", "7"))
    if (is.na(aperture_num)) {
      stop(sprintf("unexpected aperture value '%s' in grid object", aperture_str))
    }
    aperture_num <- c(3L, 4L, 7L)[aperture_num]
    cell_ids <- cpp_lonlat_to_cell(lon_vec, lat_vec, res, aperture_num)
    centers <- cpp_cell_to_lonlat(cell_ids, res, aperture_num)
  }

  # Build cell center matrix
  cell_center <- matrix(
    c(centers$lon_deg, centers$lat_deg),
    ncol = 2,
    dimnames = list(NULL, c("lon", "lat"))
  )

  # -------------------------------------------------------------------------
  # Return HexData object (original data unchanged)
  # -------------------------------------------------------------------------
  new_hex_data(
    data = data,
    grid = hex_grid_obj,
    cell_id = cell_ids,
    cell_center = cell_center
  )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cpp_hex_quantize_ap3 <- function(icosa_triangle_x, icosa_triangle_y, resolution) {
    .Call(`_hexify_cpp_hex_quantize_ap3`, icosa_triangle_x, icosa_triangle_y, resolution)
}

cpp_hex_center_ap3 <- function(i, j, resolution) {
    .Call(`_hexify_cpp_hex_center_ap3`, i, j, resolution)
}

cpp_hex_corners_ap3 <- function(i, j, resolution, hex_radius = 1.0) {
    .Call(`_hexify_cpp_hex_corners_ap3`, i, j, resolution, hex_radius)
}

cpp_lonlat_to_cell_ap3 <- function(lon_deg, lat_deg, resolution) {
    .Call(`_hexify_cpp_lonlat_to_cell_ap3`, lon_deg, lat_deg, resolution)
}

cpp_cell_to_lonlat_ap3 <- function(face, i, j, resolution) {
    .Call(`_hexify_cpp_cell_to_lonlat_ap3`, face, i, j, resolution)
}

cpp_hex_quantize_ap4 <- function(icosa_triangle_x, icosa_triangle_y, resolution) {
    .Call(`_hexify_cpp_hex_quantize_ap4`, icosa_triangle_x, icosa_triangle_y, resolution)
}

cpp_hex_center_ap4 <- function(i, j, resolution) {
    .Call(`_hexify_cpp_hex_center_ap4`, i, j, resolution)
}

cpp_hex_corners_ap4 <- function(i, j, resolution, hex_radius = 1.0) {
    .Call(`_hexify_cpp_hex_corners_ap4`, i, j, resolution, hex_radius)
}

cpp_lonlat_to_cell_ap4 <- function(lon_deg, lat_deg, resolution) {
    .Call(`_hexify_cpp_lonlat_to_cell_ap4`, lon_deg, lat_deg, resolution)
}

cpp_cell_to_lonlat_ap4 <- function(face, i, j, resolution) {
    .Call(`_hexify_cpp_cell_to_lonlat_ap4`, face, i, j, resolution)
}

cpp_hex_quantize_ap7 <- function(icosa_triangle_x, icosa_triangle_y, resolution) {
    .Call(`_hexify_cpp_hex_quantize_ap7`, icosa_triangle_x, icosa_triangle_y, resolution)
}

cpp_hex_center_ap7 <- function(i, j, resolution) {
    .Call(`_hexify_cpp_hex_center_ap7`, i, j, resolution)
}

cpp_hex_corners_ap7 <- function(i, j, resolution, hex_radius = 1.0) {
    .Call(`_hexify_cpp_hex_corners_ap7`, i, j, resolution, hex_radius)
}

cpp_lonlat_to_cell_ap7 <- function(lon_deg, lat_deg, resolution) {
    .Call(`_hexify_cpp_lonlat_to_cell_ap7`, lon_deg, lat_deg, resolution)
}

cpp_cell_to_lonlat_ap7 <- function(face, i, j, resolution) {
    .Call(`_hexify_cpp_cell_to_lonlat_ap7`, face, i, j, resolution)
}

cpp_hex_quantize_mixed <- function(icosa_triangle_x, icosa_triangle_y, ap_seq) {
    .Call(`_hexify_cpp_hex_quantize_mixed`, icosa_triangle_x, icosa_triangle_y, ap_seq)
}

cpp_hex_center_mixed <- function(i, j, ap_seq) {
    .Call(`_hexify_cpp_hex_center_mixed`, i, j, ap_seq)
}

cpp_hex_corners_mixed <- function(i, j, ap_seq, hex_radius = 1.0) {
    .Call(`_hexify_cpp_hex_corners_mixed`, i, j, ap_seq, hex_radius)
}

cpp_lonlat_to_cell_mixed <- function(lon_deg, lat_deg, ap_seq) {
    .Call(`_hexify_cpp_lonlat_to_cell_mixed`, lon_deg, lat_deg, ap_seq)
}

cpp_cell_to_lonlat_mixed <- function(face, i, j, ap_seq) {
    .Call(`_hexify_cpp_cell_to_lonlat_mixed`, face, i, j, ap_seq)
}

cpp_test_roundtrip_ap3 <- function(icosa_triangle_x, icosa_triangle_y, resolution) {
    .Call(`_hexify_cpp_test_roundtrip_ap3`, icosa_triangle_x, icosa_triangle_y, resolution)
}

cpp_test_roundtrip_ap4 <- function(icosa_triangle_x, icosa_triangle_y, resolution) {
    .Call(`_hexify_cpp_test_roundtrip_ap4`, icosa_triangle_x, icosa_triangle_y, resolution)
}

cpp_test_roundtrip_ap7 <- function(icosa_triangle_x, icosa_triangle_y, resolution) {
    .Call(`_hexify_cpp_test_roundtrip_ap7`, icosa_triangle_x, icosa_triangle_y, resolution)
}

cpp_test_roundtrip_mixed <- function(icosa_triangle_x, icosa_triangle_y, ap_seq) {
    .Call(`_hexify_cpp_test_roundtrip_mixed`, icosa_triangle_x, icosa_triangle_y, ap_seq)
}

cpp_batch_test_roundtrip_ap3 <- function(tx_vec, ty_vec, resolution) {
    .Call(`_hexify_cpp_batch_test_roundtrip_ap3`, tx_vec, ty_vec, resolution)
}

cpp_batch_test_roundtrip_ap4 <- function(tx_vec, ty_vec, resolution) {
    .Call(`_hexify_cpp_batch_test_roundtrip_ap4`, tx_vec, ty_vec, resolution)
}

cpp_batch_test_roundtrip_ap7 <- function(tx_vec, ty_vec, resolution) {
    .Call(`_hexify_cpp_batch_test_roundtrip_ap7`, tx_vec, ty_vec, resolution)
}

cpp_decode_z7 <- function(index_body, aperture) {
    .Call(`_hexify_cpp_decode_z7`, index_body, aperture)
}

cpp_icosa_tri_to_quad_ij <- function(icosa_triangle_face, icosa_triangle_x, icosa_triangle_y, aperture, resolution) {
    .Call(`_hexify_cpp_icosa_tri_to_quad_ij`, icosa_triangle_face, icosa_triangle_x, icosa_triangle_y, aperture, resolution)
}

cpp_icosa_tri_to_quad_xy <- function(icosa_triangle_face, icosa_triangle_x, icosa_triangle_y) {
    .Call(`_hexify_cpp_icosa_tri_to_quad_xy`, icosa_triangle_face, icosa_triangle_x, icosa_triangle_y)
}

cpp_quad_xy_to_icosa_tri <- function(quad, quad_x, quad_y) {
    .Call(`_hexify_cpp_quad_xy_to_icosa_tri`, quad, quad_x, quad_y)
}

cpp_quad_ij_to_xy <- function(quad, i, j, aperture, resolution) {
    .Call(`_hexify_cpp_quad_ij_to_xy`, quad, i, j, aperture, resolution)
}

cpp_lonlat_to_quad_ij <- function(lon_deg, lat_deg, aperture, resolution) {
    .Call(`_hexify_cpp_lonlat_to_quad_ij`, lon_deg, lat_deg, aperture, resolution)
}

cpp_quad_ij_to_cell <- function(quad, i, j, resolution, aperture) {
    .Call(`_hexify_cpp_quad_ij_to_cell`, quad, i, j, resolution, aperture)
}

cpp_lonlat_to_cell <- function(lon, lat, resolution, aperture) {
    .Call(`_hexify_cpp_lonlat_to_cell`, lon, lat, resolution, aperture)
}

cpp_cell_to_lonlat <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_cell_to_lonlat`, cell_id, resolution, aperture)
}

cpp_cell_to_quad_ij <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_cell_to_quad_ij`, cell_id, resolution, aperture)
}

cpp_cell_to_quad_xy <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_cell_to_quad_xy`, cell_id, resolution, aperture)
}

cpp_quad_xy_to_cell <- function(quad, quad_x, quad_y, resolution, aperture) {
    .Call(`_hexify_cpp_quad_xy_to_cell`, quad, quad_x, quad_y, resolution, aperture)
}

cpp_cell_to_icosa_tri <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_cell_to_icosa_tri`, cell_id, resolution, aperture)
}

cpp_quad_ij_to_icosa_tri <- function(quad, i, j, resolution, aperture) {
    .Call(`_hexify_cpp_quad_ij_to_icosa_tri`, quad, i, j, resolution, aperture)
}

cpp_cell_to_corners <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_cell_to_corners`, cell_id, resolution, aperture)
}

cpp_cell_to_polygon <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_cell_to_polygon`, cell_id, resolution, aperture)
}

cpp_ap_seq_edge_dim <- function(ap_seq_in) {
    .Call(`_hexify_cpp_ap_seq_edge_dim`, ap_seq_in)
}

cpp_lonlat_to_cell_seq <- function(lon, lat, ap_seq_in) {
    .Call(`_hexify_cpp_lonlat_to_cell_seq`, lon, lat, ap_seq_in)
}

cpp_cell_to_lonlat_seq <- function(cell_id, ap_seq_in) {
    .Call(`_hexify_cpp_cell_to_lonlat_seq`, cell_id, ap_seq_in)
}

cpp_cell_to_quad_ij_seq <- function(cell_id, ap_seq_in) {
    .Call(`_hexify_cpp_cell_to_quad_ij_seq`, cell_id, ap_seq_in)
}

cpp_quad_ij_to_cell_seq <- function(quad, i, j, ap_seq_in) {
    .Call(`_hexify_cpp_quad_ij_to_cell_seq`, quad, i, j, ap_seq_in)
}

cpp_cell_to_corners_seq <- function(cell_id, ap_seq_in) {
    .Call(`_hexify_cpp_cell_to_corners_seq`, cell_id, ap_seq_in)
}

cpp_cell_to_polygon_seq <- function(cell_id, ap_seq_in) {
    .Call(`_hexify_cpp_cell_to_polygon_seq`, cell_id, ap_seq_in)
}

cpp_get_neighbors_isea <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_get_neighbors_isea`, cell_id, resolution, aperture)
}

cpp_get_neighbors_z7 <- function(index_ids, resolution) {
    .Call(`_hexify_cpp_get_neighbors_z7`, index_ids, resolution)
}

cpp_icosa_tri_to_plane <- function(icosa_triangle_face, icosa_triangle_x, icosa_triangle_y) {
    .Call(`_hexify_cpp_icosa_tri_to_plane`, icosa_triangle_face, icosa_triangle_x, icosa_triangle_y)
}

cpp_cell_to_plane <- function(cell_id, resolution, aperture) {
    .Call(`_hexify_cpp_cell_to_plane`, cell_id, resolution, aperture)
}

cpp_lonlat_to_plane <- function(lon, lat) {
    .Call(`_hexify_cpp_lonlat_to_plane`, lon, lat)
}

cpp_h3_latLngToCell <- function(lon_deg, lat_deg, resolution) {
    .Call(`_hexify_cpp_h3_latLngToCell`, lon_deg, lat_deg, resolution)
}

cpp_h3_cellToLatLng <- function(cell_ids) {
    .Call(`_hexify_cpp_h3_cellToLatLng`, cell_ids)
}

cpp_h3_isValidCell <- function(cell_ids) {
    .Call(`_hexify_cpp_h3_isValidCell`, cell_ids)
}

cpp_h3_cellToParent <- function(cell_ids, parent_res) {
    .Call(`_hexify_cpp_h3_cellToParent`, cell_ids, parent_res)
}

cpp_h3_cellToChildren <- function(cell_ids, child_res) {
    .Call(`_hexify_cpp_h3_cellToChildren`, cell_ids, child_res)
}

cpp_h3_cellToBoundary <- function(cell_ids) {
    .Call(`_hexify_cpp_h3_cellToBoundary`, cell_ids)
}

cpp_h3_polygonToCells <- function(coords, resolution, holes = NULL, flags = 2L) {
    .Call(`_hexify_cpp_h3_polygonToCells`, coords, resolution, holes, flags)
}

cpp_h3_getResolution <- function(cell_ids) {
    .Call(`_hexify_cpp_h3_getResolution`, cell_ids)
}

cpp_h3_cellAreaKm2 <- function(cell_ids) {
    .Call(`_hexify_cpp_h3_cellAreaKm2`, cell_ids)
}

cpp_h3_gridDisk <- function(cell_ids, k) {
    .Call(`_hexify_cpp_h3_gridDisk`, cell_ids, k)
}

cpp_h3_gridDiskDistances <- function(cell_ids, k) {
    .Call(`_hexify_cpp_h3_gridDiskDistances`, cell_ids, k)
}

cpp_h3_gridRingUnsafe <- function(cell_ids, k) {
    .Call(`_hexify_cpp_h3_gridRingUnsafe`, cell_ids, k)
}

cpp_h3_compactCells <- function(cell_ids) {
    .Call(`_hexify_cpp_h3_compactCells`, cell_ids)
}

cpp_h3_uncompactCells <- function(cell_ids, target_res) {
    .Call(`_hexify_cpp_h3_uncompactCells`, cell_ids, target_res)
}

cpp_h3_isPentagon <- function(cell_ids) {
    .Call(`_hexify_cpp_h3_isPentagon`, cell_ids)
}

cpp_h3_gridDistance <- function(origin, destination) {
    .Call(`_hexify_cpp_h3_gridDistance`, origin, destination)
}

cpp_cell_to_index <- function(face, i, j, resolution, aperture, index_type = "auto") {
    .Call(`_hexify_cpp_cell_to_index`, face, i, j, resolution, aperture, index_type)
}

cpp_index_to_cell <- function(index, aperture, index_type = "auto") {
    .Call(`_hexify_cpp_index_to_cell`, index, aperture, index_type)
}

cpp_get_parent_index <- function(index, aperture, index_type = "auto") {
    .Call(`_hexify_cpp_get_parent_index`, index, aperture, index_type)
}

cpp_get_children_indices <- function(index, aperture, index_type = "auto") {
    .Call(`_hexify_cpp_get_children_indices`, index, aperture, index_type)
}

cpp_get_index_resolution <- function(index, aperture, index_type = "auto") {
    .Call(`_hexify_cpp_get_index_resolution`, index, aperture, index_type)
}

cpp_compare_indices <- function(idx1, idx2) {
    .Call(`_hexify_cpp_compare_indices`, idx1, idx2)
}

cpp_is_valid_index_type <- function(aperture, index_type) {
    .Call(`_hexify_cpp_is_valid_index_type`, aperture, index_type)
}

cpp_get_default_index_type <- function(aperture) {
    .Call(`_hexify_cpp_get_default_index_type`, aperture)
}

cpp_lonlat_to_index_ap3 <- function(lon_deg, lat_deg, resolution, index_type = "auto") {
    .Call(`_hexify_cpp_lonlat_to_index_ap3`, lon_deg, lat_deg, resolution, index_type)
}

cpp_lonlat_to_index_ap4 <- function(lon_deg, lat_deg, resolution, index_type = "auto") {
    .Call(`_hexify_cpp_lonlat_to_index_ap4`, lon_deg, lat_deg, resolution, index_type)
}

cpp_lonlat_to_index_ap7 <- function(lon_deg, lat_deg, resolution, index_type = "auto") {
    .Call(`_hexify_cpp_lonlat_to_index_ap7`, lon_deg, lat_deg, resolution, index_type)
}

cpp_lonlat_to_index <- function(lon_deg, lat_deg, resolution, aperture, index_type = "auto") {
    .Call(`_hexify_cpp_lonlat_to_index`, lon_deg, lat_deg, resolution, aperture, index_type)
}

cpp_index_to_lonlat <- function(index, aperture, index_type = "auto") {
    .Call(`_hexify_cpp_index_to_lonlat`, index, aperture, index_type)
}

cpp_z7_canonical_form <- function(index, max_iterations = 128L) {
    .Call(`_hexify_cpp_z7_canonical_form`, index, max_iterations)
}

cpp_build_icosa <- function(vert0_lon_deg = 11.25, vert0_lat_deg = 58.282525588538995, azimuth_deg = 0.0) {
    invisible(.Call(`_hexify_cpp_build_icosa`, vert0_lon_deg, vert0_lat_deg, azimuth_deg))
}

cpp_which_face <- function(lon_deg, lat_deg) {
    .Call(`_hexify_cpp_which_face`, lon_deg, lat_deg)
}

cpp_face_centers <- function() {
    .Call(`_hexify_cpp_face_centers`)
}

cpp_snyder_forward <- function(lon_deg, lat_deg) {
    .Call(`_hexify_cpp_snyder_forward`, lon_deg, lat_deg)
}

cpp_project_to_icosa_triangle <- function(face, lon_deg, lat_deg) {
    .Call(`_hexify_cpp_project_to_icosa_triangle`, face, lon_deg, lat_deg)
}

cpp_snyder_inv_set_precision <- function(mode = "", tol = NULL, max_iters = NULL) {
    invisible(.Call(`_hexify_cpp_snyder_inv_set_precision`, mode, tol, max_iters))
}

cpp_snyder_inv_get_precision <- function() {
    .Call(`_hexify_cpp_snyder_inv_get_precision`)
}

cpp_snyder_inv_set_verbose <- function(v = TRUE) {
    invisible(.Call(`_hexify_cpp_snyder_inv_set_verbose`, v))
}

cpp_snyder_inv_get_stats_and_reset <- function() {
    .Call(`_hexify_cpp_snyder_inv_get_stats_and_reset`)
}

cpp_face_xy_to_ll <- function(x, y, face, tol = NULL, max_iters = NULL) {
    .Call(`_hexify_cpp_face_xy_to_ll`, x, y, face, tol, max_iters)
}

cpp_icosa_face_params <- function(face) {
    .Call(`_hexify_cpp_icosa_face_params`, face)
}

cpp_hex_index_face_to_lonlat <- function(x, y, cen_lat, cen_lon, face_azimuth_offset, degrees = TRUE, tol = NULL, max_iters = NULL) {
    .Call(`_hexify_cpp_hex_index_face_to_lonlat`, x, y, cen_lat, cen_lon, face_azimuth_offset, degrees, tol, max_iters)
}

# tests/testthat/test-aperture-3.R
# Tests for aperture 3 (ISEA3H) hexagonal grid quantization
#
# Aperture 3 uses Class I (flat-top) at even resolutions and
# Class II (pointy-top, rotated 30°) at odd resolutions.

# =============================================================================
# SETUP
# =============================================================================

setup_icosa <- function() {
  cpp_build_icosa()
}

# =============================================================================
# ROUND-TRIP TESTS
# =============================================================================

test_that("aperture 3 round-trip works for Class I (even resolutions)", {
  setup_icosa()

  test_points <- list(
    c(0.5, 0.3),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0)
  )

  for (res in c(0, 2, 4, 6)) {
    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell <- cpp_hex_quantize_ap3(tx, ty, res)
      center <- cpp_hex_center_ap3(cell["i"], cell["j"], res)
      cell2 <- cpp_hex_quantize_ap3(center["cx"], center["cy"], res)

      expect_equal(cell["i"], cell2["i"],
                   info = sprintf("Class I res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
      expect_equal(cell["j"], cell2["j"],
                   info = sprintf("Class I res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
    }
  }
})

test_that("aperture 3 round-trip works for Class II (odd resolutions)", {
  setup_icosa()

  test_points <- list(
    c(0.5, 0.3),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0)
  )

  for (res in c(1, 3, 5)) {
    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell <- cpp_hex_quantize_ap3(tx, ty, res)
      center <- cpp_hex_center_ap3(cell["i"], cell["j"], res)
      cell2 <- cpp_hex_quantize_ap3(center["cx"], center["cy"], res)

      expect_equal(cell["i"], cell2["i"],
                   info = sprintf("Class II res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
      expect_equal(cell["j"], cell2["j"],
                   info = sprintf("Class II res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
    }
  }
})

test_that("aperture 3 batch round-trip succeeds", {
  setup_icosa()

  set.seed(123)
  n <- 20
  tx <- runif(n, -0.8, 0.8)
  ty <- runif(n, -0.8, 0.8)

  for (res in c(2, 3, 4, 5)) {
    result <- cpp_batch_test_roundtrip_ap3(tx, ty, res)
    n_pass <- sum(result$success)

    expect_equal(n_pass, n,
                 info = sprintf("res=%d: %d/%d passed", res, n_pass, n))
  }
})

# =============================================================================
# SCALING AND REFINEMENT
# =============================================================================

test_that("aperture 3 cell spacing decreases with resolution", {
  setup_icosa()

  # Compare spacing between same-class resolutions
  # Class I: 0, 2, 4, 6 (even)
  # Class II: 1, 3, 5, 7 (odd)

  for (res in c(2, 4)) {
    center_lo_0 <- cpp_hex_center_ap3(0, 0, res)
    center_lo_1 <- cpp_hex_center_ap3(1, 0, res)
    spacing_lo <- abs(center_lo_1["cx"] - center_lo_0["cx"])

    center_hi_0 <- cpp_hex_center_ap3(0, 0, res + 2)
    center_hi_1 <- cpp_hex_center_ap3(1, 0, res + 2)
    spacing_hi <- abs(center_hi_1["cx"] - center_hi_0["cx"])

    expect_true(spacing_hi < spacing_lo,
                info = sprintf("res=%d vs res=%d spacing", res, res + 2))
  }
})

test_that("resolution refinement decreases distance to point", {
  setup_icosa()

  tx <- 0.4
  ty <- 0.35

  cell2 <- cpp_hex_quantize_ap3(tx, ty, 2)
  cell4 <- cpp_hex_quantize_ap3(tx, ty, 4)
  cell6 <- cpp_hex_quantize_ap3(tx, ty, 6)

  center2 <- cpp_hex_center_ap3(cell2["i"], cell2["j"], 2)
  center4 <- cpp_hex_center_ap3(cell4["i"], cell4["j"], 4)
  center6 <- cpp_hex_center_ap3(cell6["i"], cell6["j"], 6)

  dist2 <- sqrt((center2["cx"] - tx)^2 + (center2["cy"] - ty)^2)
  dist4 <- sqrt((center4["cx"] - tx)^2 + (center4["cy"] - ty)^2)
  dist6 <- sqrt((center6["cx"] - tx)^2 + (center6["cy"] - ty)^2)

  expect_true(dist4 < dist2)
  expect_true(dist6 < dist4)
})

# =============================================================================
# CLASS I TO CLASS II TRANSITION
# =============================================================================

test_that("Class I to Class II transition is consistent", {
  setup_icosa()

  tx <- 0.42
  ty <- 0.41

  cell2 <- cpp_hex_quantize_ap3(tx, ty, 2)  # Class I
  cell3 <- cpp_hex_quantize_ap3(tx, ty, 3)  # Class II
  cell4 <- cpp_hex_quantize_ap3(tx, ty, 4)  # Class I

  center2 <- cpp_hex_center_ap3(cell2["i"], cell2["j"], 2)
  center3 <- cpp_hex_center_ap3(cell3["i"], cell3["j"], 3)
  center4 <- cpp_hex_center_ap3(cell4["i"], cell4["j"], 4)

  dist2 <- sqrt((center2["cx"] - tx)^2 + (center2["cy"] - ty)^2)
  dist3 <- sqrt((center3["cx"] - tx)^2 + (center3["cy"] - ty)^2)
  dist4 <- sqrt((center4["cx"] - tx)^2 + (center4["cy"] - ty)^2)

  expect_true(dist3 < dist2)
  expect_true(dist4 < dist3)
})

# =============================================================================
# HEXAGON CORNERS
# =============================================================================

test_that("aperture 3 corners form valid hexagons", {
  setup_icosa()

  for (res in c(2, 3)) {
    corners <- cpp_hex_corners_ap3(0, 0, res, 1.0)

    expect_equal(length(corners$x), 6)
    expect_equal(length(corners$y), 6)
    expect_true(all(is.finite(corners$x)))
    expect_true(all(is.finite(corners$y)))
  }
})

test_that("hexagon corners centroid matches center", {
  setup_icosa()

  for (res in c(2, 4)) {
    for (i in 0:2) {
      for (j in 0:2) {
        center <- cpp_hex_center_ap3(i, j, res)
        corners <- cpp_hex_corners_ap3(i, j, res, 1.0)

        centroid_x <- mean(corners$x)
        centroid_y <- mean(corners$y)

        expect_equal(centroid_x, as.numeric(center["cx"]), tolerance = 1e-10)
        expect_equal(centroid_y, as.numeric(center["cy"]), tolerance = 1e-10)
      }
    }
  }
})

# =============================================================================
# LON/LAT WORKFLOW
# =============================================================================

test_that("aperture 3 lon/lat workflow works", {
  setup_icosa()

  lon <- 16.37  # Vienna
  lat <- 48.21
  res <- 6

  cell <- cpp_lonlat_to_cell_ap3(lon, lat, res)

  expect_true(cell["face"] >= 0 && cell["face"] < 20)
  expect_true(is.numeric(cell["i"]))
  expect_true(is.numeric(cell["j"]))

  ll <- cpp_cell_to_lonlat_ap3(cell["face"], cell["i"], cell["j"], res)

  # Should be reasonably close (within cell diameter)
  dist <- sqrt((ll["lon"] - lon)^2 + (ll["lat"] - lat)^2)
  expect_true(dist < 10.0)
})

# =============================================================================
# SINGLE-POINT ROUNDTRIP TEST HELPER
# =============================================================================

test_that("cpp_test_roundtrip_ap3 returns TRUE for valid points", {
  setup_icosa()

  # Test various points and resolutions
  test_points <- list(
    c(0.5, 0.3),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0)
  )

  for (res in c(2, 3, 4, 5)) {
    for (pt in test_points) {
      result <- cpp_test_roundtrip_ap3(pt[1], pt[2], res)
      expect_true(result, info = sprintf("res=%d, pt=(%.2f, %.2f)", res, pt[1], pt[2]))
    }
  }
})

# test-h3-native.R
# Tests for native H3 C backend wrappers

# =============================================================================
# latLngToCell / cellToLatLng round-trip
# =============================================================================

test_that("cpp_h3_latLngToCell returns valid hex strings", {
  cells <- cpp_h3_latLngToCell(c(16.37, 2.35), c(48.21, 48.86), 5L)
  expect_true(is.character(cells))
  expect_equal(length(cells), 2)
  expect_true(all(nchar(cells) == 15))  # H3 hex strings are 15 chars
})

test_that("cpp_h3_latLngToCell handles NA inputs", {
  cells <- cpp_h3_latLngToCell(c(16.37, NA), c(48.21, 48.86), 5L)
  expect_false(is.na(cells[1]))
  expect_true(is.na(cells[2]))
})

test_that("latLngToCell → cellToLatLng round-trip is close", {
  lon <- c(16.37, -73.99, 139.69)
  lat <- c(48.21, 40.75, 35.69)
  cells <- cpp_h3_latLngToCell(lon, lat, 8L)
  centers <- cpp_h3_cellToLatLng(cells)

  for (i in seq_along(lon)) {
    expect_true(abs(centers$lon[i] - lon[i]) < 0.1)
    expect_true(abs(centers$lat[i] - lat[i]) < 0.1)
  }
})

# =============================================================================
# isValidCell
# =============================================================================

test_that("cpp_h3_isValidCell validates correctly", {
  cells <- cpp_h3_latLngToCell(c(0, 10), c(45, 50), 5L)
  valid <- cpp_h3_isValidCell(cells)
  expect_true(all(valid))

  invalid <- cpp_h3_isValidCell(c("garbage", "0000000000000000"))
  expect_false(any(invalid))
})

test_that("cpp_h3_isValidCell handles NA", {
  result <- cpp_h3_isValidCell(c("851e15b7fffffff", NA))
  expect_true(result[1])
  expect_true(is.na(result[2]))
})

# =============================================================================
# cellToParent / cellToChildren
# =============================================================================

test_that("cpp_h3_cellToParent returns coarser cell", {
  cell <- cpp_h3_latLngToCell(0, 45, 8L)
  parent <- cpp_h3_cellToParent(cell, 5L)
  expect_true(is.character(parent))
  expect_equal(length(parent), 1)
  expect_false(parent == cell)
})

test_that("cpp_h3_cellToChildren returns 7 children", {
  cell <- cpp_h3_latLngToCell(0, 45, 5L)
  children <- cpp_h3_cellToChildren(cell, 6L)
  expect_true(is.list(children))
  expect_equal(length(children[[1]]), 7)
  expect_true(all(cpp_h3_isValidCell(children[[1]])))
})

# =============================================================================
# cellToBoundary
# =============================================================================

test_that("cpp_h3_cellToBoundary returns closed rings", {
  cell <- cpp_h3_latLngToCell(0, 45, 5L)
  bndry <- cpp_h3_cellToBoundary(cell)
  ring <- bndry[[1]]
  expect_true(is.matrix(ring))
  expect_equal(ncol(ring), 2)
  expect_equal(ring[1, ], ring[nrow(ring), ])  # closed ring
  expect_true(nrow(ring) == 7)  # 6 hex verts + 1 closing
})

# =============================================================================
# polygonToCells
# =============================================================================

test_that("cpp_h3_polygonToCells fills a bounding box", {
  bbox <- matrix(c(16, 48, 17, 48, 17, 49, 16, 49, 16, 48),
                 ncol = 2, byrow = TRUE)
  cells <- cpp_h3_polygonToCells(bbox, 5L)
  expect_true(length(cells) > 0)
  expect_true(all(cpp_h3_isValidCell(cells)))
})

# =============================================================================
# cellAreaKm2
# =============================================================================

test_that("cpp_h3_cellAreaKm2 returns positive areas", {
  cells <- cpp_h3_latLngToCell(c(0, 0), c(0, 80), 5L)
  areas <- cpp_h3_cellAreaKm2(cells)
  expect_true(all(areas > 0))
  # Areas should differ (H3 is not equal-area)
  expect_false(areas[1] == areas[2])
})

test_that("cpp_h3_cellAreaKm2 magnitude is reasonable for res 5", {
  cell <- cpp_h3_latLngToCell(0, 45, 5L)
  area <- cpp_h3_cellAreaKm2(cell)
  # H3 res 5 avg area is ~252.9 km^2
  expect_true(area > 200 && area < 300)
})

# tests/testthat/test-index_zorder.R
# Comprehensive tests for Z-Order indexing (generic, all apertures)

# =============================================================================
# Helper Functions
# =============================================================================

get_max_coord <- function(aperture, resolution) {
  # Calculate maximum valid coordinate for given aperture/resolution
  if (aperture == 3) {
    # For aperture 3, effective resolution
    eff_res <- (resolution + 1) %/% 2
    return(3^eff_res - 1)
  } else if (aperture == 4) {
    # For aperture 4, uses radix-2 (binary)
    return(2^resolution - 1)
  } else if (aperture == 7) {
    # For aperture 7, uses radix-7
    return(7^resolution - 1)
  } else {
    return(aperture^resolution - 1)
  }
}

# =============================================================================
# Aperture 3 Z-Order Tests
# =============================================================================

test_that("Z-Order aperture 3: basic encode/decode", {
  test_cases <- list(
    list(face = 0, i = 0, j = 0, res = 1),
    list(face = 0, i = 1, j = 2, res = 1),  # Changed from (1,0) - i%3=1 requires j%3=2
    list(face = 0, i = 0, j = 0, res = 1),  # Changed from (0,1) - invalid for Class II
    list(face = 5, i = 1, j = 2, res = 1),  # Changed from (1,1) - i%3=1 requires j%3=2
    list(face = 10, i = 2, j = 2, res = 2),
    list(face = 19, i = 5, j = 7, res = 3)
  )
  
  for (tc in test_cases) {
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 3, "zorder")
    result <- cpp_index_to_cell(index, 3, "zorder")
    
    expect_equal(result$face, tc$face, 
                 label = sprintf("face[%d,%d,%d]", tc$i, tc$j, tc$res))
    expect_equal(result$i, tc$i,
                 label = sprintf("i[%d,%d,%d]", tc$i, tc$j, tc$res))
    expect_equal(result$j, tc$j,
                 label = sprintf("j[%d,%d,%d]", tc$i, tc$j, tc$res))
    expect_equal(result$resolution, tc$res,
                 label = sprintf("res[%d,%d,%d]", tc$i, tc$j, tc$res))
  }
})

test_that("Z-Order aperture 3: digit alternation pattern", {
  # Z-Order for aperture 3 uses simple alternation
  # For res=2 (Class I), eff_res=1: alternates i,j digits
  
  # i=1, j=2 → radix-3: "1", "2" → alternation: "12"
  index <- cpp_cell_to_index(0, 1, 2, 2, 3, "zorder")
  expect_match(index, "^0012$")
  
  # i=2, j=1 → radix-3: "2", "1" → alternation: "21"
  index <- cpp_cell_to_index(0, 2, 1, 2, 3, "zorder")
  expect_match(index, "^0021$")
})

test_that("Z-Order aperture 3: Class II j-digit inference", {
  # For Class II (odd resolution), last j digit is inferred from last i digit
  # Rule: i=0→j=0, i=1→j=2, i=2→j=1
  
  test_cases <- list(
    list(i = 0, j = 0, res = 1),  # Last i=0 → inferred j=0
    list(i = 1, j = 2, res = 1),  # Last i=1 → inferred j=2
    list(i = 2, j = 1, res = 1),  # Last i=2 → inferred j=1
    list(i = 3, j = 6, res = 3),  # i ends in 0 → j ends in 0
    list(i = 4, j = 8, res = 3)   # i ends in 1 → j ends in 2
  )
  
  for (tc in test_cases) {
    index <- cpp_cell_to_index(0, tc$i, tc$j, tc$res, 3, "zorder")
    result <- cpp_index_to_cell(index, 3, "zorder")
    
    expect_equal(result$i, tc$i,
                 label = sprintf("i inference[%d,%d]", tc$i, tc$j))
    expect_equal(result$j, tc$j,
                 label = sprintf("j inference[%d,%d]", tc$i, tc$j))
  }
})

test_that("Z-Order aperture 3: Class I vs Class II string length", {
  # Class I (even res): even number of digits after face
  # Class II (odd res): odd number of digits after face
  
  for (res in 1:6) {
    index <- cpp_cell_to_index(0, 0, 0, res, 3, "zorder")
    digits_after_face <- nchar(index) - 2
    
    if (res %% 2 == 0) {
      # Class I: even number of digits
      expect_true(digits_after_face %% 2 == 0,
                  label = sprintf("Class I res=%d", res))
    } else {
      # Class II: odd number of digits
      expect_true(digits_after_face %% 2 == 1,
                  label = sprintf("Class II res=%d", res))
    }
  }
})

test_that("Z-Order aperture 3: comprehensive roundtrip", {
  # Reduced test set for speed
  test_cases <- expand.grid(
    face = c(0, 10),
    i = c(0, 2, 5),
    j = c(0, 2, 5),
    res = 1:2
  )

  for (idx in seq_len(nrow(test_cases))) {
    tc <- test_cases[idx, ]

    # Skip invalid coordinates
    max_coord <- get_max_coord(3, tc$res)
    if (tc$i > max_coord || tc$j > max_coord) next

    # Skip invalid Class II coordinates
    if (tc$res %% 2 == 1) {  # Class II (odd resolution)
      j_required <- c(0, 2, 1)[tc$i %% 3 + 1]
      if (tc$j %% 3 != j_required) next
    }

    # Roundtrip test
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 3, "zorder")
    result <- cpp_index_to_cell(index, 3, "zorder")

    expect_equal(result$face, tc$face, label = sprintf("idx=%d", idx))
    expect_equal(result$i, tc$i, label = sprintf("idx=%d", idx))
    expect_equal(result$j, tc$j, label = sprintf("idx=%d", idx))
    expect_equal(result$resolution, tc$res, label = sprintf("idx=%d", idx))
  }
})

# =============================================================================
# Aperture 4 Z-Order Tests
# =============================================================================

test_that("Z-Order aperture 4: bit interleaving formula", {
  # DGGRID formula: digit = i_bit * 2 + j_bit
  # This is tested by checking known cases
  
  test_cases <- list(
    # i=0 (binary: 0), j=0 (binary: 0) → digit: 0
    list(i = 0, j = 0, res = 1, expected = "000"),
    
    # i=1 (binary: 1), j=0 (binary: 0) → digit: 2
    list(i = 1, j = 0, res = 1, expected = "002"),
    
    # i=0 (binary: 0), j=1 (binary: 1) → digit: 1
    list(i = 0, j = 1, res = 1, expected = "001"),
    
    # i=1 (binary: 1), j=1 (binary: 1) → digit: 3
    list(i = 1, j = 1, res = 1, expected = "003"),
    
    # i=5 (binary: 101), j=6 (binary: 110) → digits: 3,1,2
    list(i = 5, j = 6, res = 3, expected = "00312"),
    
    # i=3 (binary: 11), j=3 (binary: 11) → digits: 3,3
    list(i = 3, j = 3, res = 2, expected = "0033"),
    
    # i=15 (binary: 1111), j=0 (binary: 0000) → digits: 2,2,2,2
    list(i = 15, j = 0, res = 4, expected = "002222")
  )
  
  for (tc in test_cases) {
    index <- cpp_cell_to_index(0, tc$i, tc$j, tc$res, 4, "zorder")
    expect_equal(index, tc$expected,
                 label = sprintf("i=%d,j=%d,res=%d", tc$i, tc$j, tc$res))
  }
})

test_that("Z-Order aperture 4: basic encode/decode", {
  test_cases <- list(
    list(face = 0, i = 0, j = 0, res = 1),
    list(face = 5, i = 1, j = 2, res = 2),
    list(face = 10, i = 5, j = 6, res = 3),
    list(face = 11, i = 10, j = 15, res = 4)  # Changed from face 19 to 11
  )
  
  for (tc in test_cases) {
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 4, "zorder")
    result <- cpp_index_to_cell(index, 4, "zorder")
    
    expect_equal(result$face, tc$face)
    expect_equal(result$i, tc$i)
    expect_equal(result$j, tc$j)
    expect_equal(result$resolution, tc$res)
  }
})

test_that("Z-Order aperture 4: comprehensive roundtrip", {
  # Reduced test set for speed
  test_cases <- expand.grid(
    face = c(0, 10),
    i = c(0, 3, 7),
    j = c(0, 3, 7),
    res = 1:3
  )

  for (idx in seq_len(nrow(test_cases))) {
    tc <- test_cases[idx, ]

    # Skip invalid coordinates
    max_coord <- get_max_coord(4, tc$res)
    if (tc$i > max_coord || tc$j > max_coord) next

    # Roundtrip test
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 4, "zorder")
    result <- cpp_index_to_cell(index, 4, "zorder")

    expect_equal(result$face, tc$face, label = sprintf("idx=%d", idx))
    expect_equal(result$i, tc$i, label = sprintf("idx=%d", idx))
    expect_equal(result$j, tc$j, label = sprintf("idx=%d", idx))
    expect_equal(result$resolution, tc$res, label = sprintf("idx=%d", idx))
  }
})

# =============================================================================
# Aperture 7 Z-Order Tests
# =============================================================================

test_that("Z-Order aperture 7: basic encode/decode", {
  test_cases <- list(
    list(face = 0, i = 0, j = 0, res = 1),
    list(face = 5, i = 1, j = 2, res = 1),
    list(face = 10, i = 3, j = 5, res = 2),
    list(face = 11, i = 10, j = 20, res = 2)  # Changed from face 19 to 11
  )
  
  for (tc in test_cases) {
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 7, "zorder")
    result <- cpp_index_to_cell(index, 7, "zorder")
    
    expect_equal(result$face, tc$face)
    expect_equal(result$i, tc$i)
    expect_equal(result$j, tc$j)
    expect_equal(result$resolution, tc$res)
  }
})

test_that("Z-Order aperture 7: comprehensive roundtrip", {
  # Reduced test set for speed
  test_cases <- expand.grid(
    face = c(0, 10),
    i = c(0, 3, 6),
    j = c(0, 3, 6),
    res = 1:2
  )

  for (idx in seq_len(nrow(test_cases))) {
    tc <- test_cases[idx, ]

    # Skip invalid coordinates
    max_coord <- get_max_coord(7, tc$res)
    if (tc$i > max_coord || tc$j > max_coord) next

    # Roundtrip test
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 7, "zorder")
    result <- cpp_index_to_cell(index, 7, "zorder")

    expect_equal(result$face, tc$face, label = sprintf("idx=%d", idx))
    expect_equal(result$i, tc$i, label = sprintf("idx=%d", idx))
    expect_equal(result$j, tc$j, label = sprintf("idx=%d", idx))
    expect_equal(result$resolution, tc$res, label = sprintf("idx=%d", idx))
  }
})

# =============================================================================
# Cross-Aperture Tests
# =============================================================================

test_that("Z-Order: resolution 0 works for all apertures", {
  for (face in c(0, 5, 10, 11)) {  # Changed from c(0, 5, 10, 19) to only valid faces
    for (ap in c(3, 4, 7)) {
      index <- cpp_cell_to_index(face, 0, 0, 0, ap, "zorder")
      
      # Should be just face number
      expect_equal(nchar(index), 2)
      expect_equal(substr(index, 1, 2), sprintf("%02d", face))
      
      # Decode should work
      result <- cpp_index_to_cell(index, ap, "zorder")
      expect_equal(result$face, face)
      expect_equal(result$resolution, 0)
      expect_equal(result$i, 0)
      expect_equal(result$j, 0)
    }
  }
})

test_that("Z-Order: face encoding correct for all apertures", {
  for (ap in c(3, 4, 7)) {
    # Single digit faces
    expect_equal(substr(cpp_cell_to_index(0, 0, 0, 1, ap, "zorder"), 1, 2), "00")
    expect_equal(substr(cpp_cell_to_index(5, 0, 0, 1, ap, "zorder"), 1, 2), "05")
    expect_equal(substr(cpp_cell_to_index(9, 0, 0, 1, ap, "zorder"), 1, 2), "09")
    
    # Double digit faces
    expect_equal(substr(cpp_cell_to_index(10, 0, 0, 1, ap, "zorder"), 1, 2), "10")
    # Face 19 is only valid for aperture 3 at resolution > 0
    if (ap == 3) {
      expect_equal(substr(cpp_cell_to_index(19, 0, 0, 1, ap, "zorder"), 1, 2), "19")
    }
  }
})

test_that("Z-Order: index comparison works", {
  idx1 <- cpp_cell_to_index(0, 0, 0, 1, 4, "zorder")
  idx2 <- cpp_cell_to_index(0, 1, 0, 1, 4, "zorder")
  idx3 <- cpp_cell_to_index(1, 0, 0, 1, 4, "zorder")
  
  expect_equal(cpp_compare_indices(idx1, idx1), 0)
  expect_true(cpp_compare_indices(idx1, idx2) != 0)
  expect_true(cpp_compare_indices(idx1, idx3) != 0)
})

test_that("Z-Order: parent/child relationships", {
  # Test with valid coordinates for each aperture
  test_cases <- list(
    list(ap = 3, i = 5, j = 4, res = 3),  # Valid: i%3=2, j%3=1 ✓ (was 5,5)
    list(ap = 4, i = 5, j = 5, res = 3),  # Valid: ap4 has no Class II constraint
    list(ap = 7, i = 0, j = 0, res = 2)   # Changed to (0,0) which is always valid
  )
  
  for (tc in test_cases) {
    # Create child at higher resolution
    child_idx <- cpp_cell_to_index(0, tc$i, tc$j, tc$res, tc$ap, "zorder")
    
    # Get parent
    parent_idx <- cpp_get_parent_index(child_idx, tc$ap, "zorder")
    
    # Parent should be prefix of child
    expect_true(substr(child_idx, 1, nchar(parent_idx)) == parent_idx,
                label = sprintf("ap=%d", tc$ap))
    
    # Get children
    children <- cpp_get_children_indices(parent_idx, tc$ap, "zorder")
    
    # Child should be in children list
    expect_true(child_idx %in% children,
                label = sprintf("ap=%d child (%d,%d) in list", tc$ap, tc$i, tc$j))
    
    # Should have correct number of children
    expect_equal(length(children), tc$ap,
                 label = sprintf("ap=%d children count", tc$ap))
  }
})

test_that("Z-Order: resolution extraction works", {
  for (ap in c(3, 4, 7)) {
    for (res in 0:3) {
      index <- cpp_cell_to_index(0, 0, 0, res, ap, "zorder")
      extracted <- cpp_get_index_resolution(index, ap, "zorder")
      expect_equal(extracted, res,
                   label = sprintf("ap=%d, res=%d", ap, res))
    }
  }
})

# tests/testthat/test-index_z3.R
# Comprehensive tests for Z3 indexing (aperture 3 specialized)
# VERSION: 2024-11-03 FIXED - Valid Class II coordinates
# 
# IMPORTANT: For Class II (odd) resolutions, coordinates must satisfy:
#   j % 3 == jDigits[i % 3]
#   where jDigits = {0→0, 1→2, 2→1}
#
# Valid examples:
#   i%3=0 → j%3=0: (0,0), (0,3), (3,6), (6,9), etc.
#   i%3=1 → j%3=2: (1,2), (1,5), (4,8), (7,11), etc.
#   i%3=2 → j%3=1: (2,1), (2,4), (5,7), (8,10), etc.

# =============================================================================
# Z3 Lookup Table Tests (Core Algorithm)
# =============================================================================

test_that("Z3: lookup table matches DGGRID exactly", {
  # DGGRID Z3 lookup table (DgZ3StringRF.cpp:135-140)
  z3_table <- matrix(c(
    "00", "22", "21",   # i=0, j=0,1,2
    "01", "02", "20",   # i=1, j=0,1,2
    "12", "10", "11"    # i=2, j=0,1,2
  ), nrow = 3, byrow = TRUE)
  
  # Test all 9 lookup table entries at resolution 2 (Class I)
  for (i in 0:2) {
    for (j in 0:2) {
      index <- cpp_cell_to_index(0, i, j, 2, 3, "z3")
      expected <- paste0("00", z3_table[i+1, j+1])
      
      expect_equal(index, expected,
                   label = sprintf("Z3[%d][%d]", i, j))
    }
  }
})

test_that("Z3: specific lookup table entries", {
  # Test specific entries explicitly
  expect_equal(cpp_cell_to_index(0, 0, 0, 2, 3, "z3"), "0000")  # [0][0]→"00"
  expect_equal(cpp_cell_to_index(0, 0, 1, 2, 3, "z3"), "0022")  # [0][1]→"22"
  expect_equal(cpp_cell_to_index(0, 0, 2, 2, 3, "z3"), "0021")  # [0][2]→"21"
  
  expect_equal(cpp_cell_to_index(0, 1, 0, 2, 3, "z3"), "0001")  # [1][0]→"01"
  expect_equal(cpp_cell_to_index(0, 1, 1, 2, 3, "z3"), "0002")  # [1][1]→"02"
  expect_equal(cpp_cell_to_index(0, 1, 2, 2, 3, "z3"), "0020")  # [1][2]→"20"
  
  expect_equal(cpp_cell_to_index(0, 2, 0, 2, 3, "z3"), "0012")  # [2][0]→"12"
  expect_equal(cpp_cell_to_index(0, 2, 1, 2, 3, "z3"), "0010")  # [2][1]→"10"
  expect_equal(cpp_cell_to_index(0, 2, 2, 2, 3, "z3"), "0011")  # [2][2]→"11"
})

# =============================================================================
# Z3 Decoding Tests (Reverse Lookup)
# =============================================================================

test_that("Z3: decode all 9 lookup table entries", {
  z3_codes <- list(
    list(code = "0000", i = 0, j = 0),  # "00"→[0][0]
    list(code = "0022", i = 0, j = 1),  # "22"→[0][1]
    list(code = "0021", i = 0, j = 2),  # "21"→[0][2]
    list(code = "0001", i = 1, j = 0),  # "01"→[1][0]
    list(code = "0002", i = 1, j = 1),  # "02"→[1][1]
    list(code = "0020", i = 1, j = 2),  # "20"→[1][2]
    list(code = "0012", i = 2, j = 0),  # "12"→[2][0]
    list(code = "0010", i = 2, j = 1),  # "10"→[2][1]
    list(code = "0011", i = 2, j = 2)   # "11"→[2][2]
  )
  
  for (tc in z3_codes) {
    result <- cpp_index_to_cell(tc$code, 3, "z3")
    
    expect_equal(result$face, 0)
    expect_equal(result$i, tc$i,
                 label = sprintf("decode '%s'", tc$code))
    expect_equal(result$j, tc$j,
                 label = sprintf("decode '%s'", tc$code))
    expect_equal(result$resolution, 2)
  }
})

test_that("Z3: decode reverse lookup table correctness", {
  # Test that the reverse lookup matches DGGRID
  # DGGRID reverse table (DgZ3StringRF.cpp:196-253)
  reverse_cases <- list(
    list(z3 = "00", i = 0, j = 0),
    list(z3 = "22", i = 0, j = 1),
    list(z3 = "21", i = 0, j = 2),
    list(z3 = "01", i = 1, j = 0),
    list(z3 = "02", i = 1, j = 1),
    list(z3 = "20", i = 1, j = 2),
    list(z3 = "12", i = 2, j = 0),
    list(z3 = "10", i = 2, j = 1),
    list(z3 = "11", i = 2, j = 2)
  )
  
  for (tc in reverse_cases) {
    index <- paste0("00", tc$z3)
    result <- cpp_index_to_cell(index, 3, "z3")
    
    expect_equal(result$i, tc$i,
                 label = sprintf("z3='%s'", tc$z3))
    expect_equal(result$j, tc$j,
                 label = sprintf("z3='%s'", tc$z3))
  }
})

# =============================================================================
# Class I vs Class II Tests
# =============================================================================

test_that("Z3: Class I string length (even resolutions)", {
  # Class I resolutions: 0, 2, 4, 6, ...
  # String length: 2 (face) + resolution digits

  class_i_cases <- list(
    list(res = 0, length = 2),
    list(res = 2, length = 4),
    list(res = 4, length = 6),
    list(res = 6, length = 8)
  )
  
  for (tc in class_i_cases) {
    index <- cpp_cell_to_index(0, 0, 0, tc$res, 3, "z3")
    expect_equal(nchar(index), tc$length,
                 label = sprintf("Class I res=%d", tc$res))
  }
})

test_that("Z3: Class II string length (odd resolutions)", {
  # Class II resolutions: 1, 3, 5, 7, ...
  # String length: 2 (face) + resolution digits

  class_ii_cases <- list(
    list(res = 1, length = 3),
    list(res = 3, length = 5),
    list(res = 5, length = 7),
    list(res = 7, length = 9)
  )
  
  for (tc in class_ii_cases) {
    index <- cpp_cell_to_index(0, 0, 0, tc$res, 3, "z3")
    expect_equal(nchar(index), tc$length,
                 label = sprintf("Class II res=%d", tc$res))
  }
})

test_that("Z3: Class II trimming behavior", {
  # For Class II resolutions, last digit should be trimmed
  # Test by checking that indices differ in expected ways
  
  # Resolution 2 (Class I) vs 3 (Class II)
  idx_r2 <- cpp_cell_to_index(0, 1, 2, 2, 3, "z3")  # Valid: i%3=1, j%3=2
  idx_r3 <- cpp_cell_to_index(0, 1, 2, 3, 3, "z3")  # Valid: i%3=1, j%3=2
  
  # Class I should be 4 chars, Class II should be 5 chars
  expect_equal(nchar(idx_r2), 4)
  expect_equal(nchar(idx_r3), 5)
  
  # Resolution 4 (Class I) vs 5 (Class II)
  idx_r4 <- cpp_cell_to_index(0, 2, 1, 4, 3, "z3")  # Valid: i%3=2, j%3=1
  idx_r5 <- cpp_cell_to_index(0, 2, 1, 5, 3, "z3")  # Valid: i%3=2, j%3=1
  
  expect_equal(nchar(idx_r4), 6)
  expect_equal(nchar(idx_r5), 7)
})

test_that("Z3: Class II j-digit inference", {
  # For Class II (odd resolution), last j digit is inferred from last i digit
  # Rule from DGGRID: i%3=0→j%3=0, i%3=1→j%3=2, i%3=2→j%3=1

  # These should decode correctly even though j digit is implicit
  test_cases <- list(
    # Resolution 1 (Class II) - eff_res=1, max=2
    list(i = 0, j = 0, res = 1),  # i%3=0, j%3=0 ✓
    list(i = 1, j = 2, res = 1),  # i%3=1, j%3=2 ✓
    list(i = 2, j = 1, res = 1),  # i%3=2, j%3=1 ✓
    
    # Resolution 3 (Class II) - eff_res=2, max=8
    list(i = 3, j = 6, res = 3),  # i%3=0, j%3=0 ✓ (i="10", j="20")
    list(i = 4, j = 8, res = 3),  # i%3=1, j%3=2 ✓ (i="11", j="22")
    list(i = 5, j = 7, res = 3)   # i%3=2, j%3=1 ✓ (i="12", j="21")
  )
  
  for (tc in test_cases) {
    index <- cpp_cell_to_index(0, tc$i, tc$j, tc$res, 3, "z3")
    result <- cpp_index_to_cell(index, 3, "z3")
    
    expect_equal(result$i, tc$i,
                 label = sprintf("infer i[%d,%d,r%d]", tc$i, tc$j, tc$res))
    expect_equal(result$j, tc$j,
                 label = sprintf("infer j[%d,%d,r%d]", tc$i, tc$j, tc$res))
  }
})

# =============================================================================
# Multi-digit Z3 Encoding Tests
# =============================================================================

test_that("Z3: multi-digit encoding at resolution 4", {
  # Resolution 4 (Class I) has eff_res=2, so 2 digit pairs
  
  # i=3, j=5 in radix-3: i="10", j="12"
  # Z3 codes: [1][1]→"02", [0][2]→"21"
  # Expected: "00" + "02" + "21" = "000221"
  index <- cpp_cell_to_index(0, 3, 5, 4, 3, "z3")
  expect_equal(index, "000221")
  
  # Roundtrip
  result <- cpp_index_to_cell(index, 3, "z3")
  expect_equal(result$i, 3)
  expect_equal(result$j, 5)
  expect_equal(result$resolution, 4)
})

test_that("Z3: multi-digit encoding at resolution 6", {
  # Resolution 6 (Class I) has eff_res=3, so 3 digit pairs
  
  # i=13, j=17 in radix-3: i="111", j="122"
  # Z3 codes: [1][1]→"02", [1][2]→"20", [1][2]→"20"
  # Expected: "00" + "02" + "20" + "20" = "00022020"
  index <- cpp_cell_to_index(0, 13, 17, 6, 3, "z3")
  expect_equal(index, "00022020")
  
  # Roundtrip
  result <- cpp_index_to_cell(index, 3, "z3")
  expect_equal(result$i, 13)
  expect_equal(result$j, 17)
  expect_equal(result$resolution, 6)
})

# =============================================================================
# Roundtrip Tests
# =============================================================================

test_that("Z3: basic roundtrip for small coordinates", {
  # Generate test cases with valid Class II coordinates
  test_cases_raw <- expand.grid(
    face = c(0, 5, 10, 19),
    i = 0:8,
    j = 0:8,
    res = 1:3
  )
  
  # Filter for valid coordinates
  test_cases <- test_cases_raw[apply(test_cases_raw, 1, function(row) {
    res <- row[["res"]]
    i <- row[["i"]]
    j <- row[["j"]]
    
    # Calculate max coordinate
    eff_res <- (res + 1) %/% 2
    max_coord <- 3^eff_res - 1
    
    # Check if within bounds
    if (i > max_coord || j > max_coord) return(FALSE)
    
    # For Class II (odd res), check constraint
    if (res %% 2 == 1) {
      j_required <- c(0, 2, 1)[i %% 3 + 1]
      if (j %% 3 != j_required) return(FALSE)
    }
    
    return(TRUE)
  }), ]
  
  for (idx in seq_len(nrow(test_cases))) {
    tc <- test_cases[idx, ]
    
    # Roundtrip
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 3, "z3")
    result <- cpp_index_to_cell(index, 3, "z3")
    
    expect_equal(result$face, tc$face, label = sprintf("idx=%d", idx))
    expect_equal(result$i, tc$i, label = sprintf("idx=%d", idx))
    expect_equal(result$j, tc$j, label = sprintf("idx=%d", idx))
    expect_equal(result$resolution, tc$res, label = sprintf("idx=%d", idx))
  }
})

test_that("Z3: roundtrip for higher resolutions", {
  # All coordinates must satisfy Class II constraint for odd resolutions
  test_cases <- list(
    # Resolution 4 (Class I): eff_res=2, max=8
    list(face = 0, i = 5, j = 7, res = 4),
    
    # Resolution 5 (Class II): eff_res=3, max=26
    # Must satisfy: j%3 == jDigits[i%3]
    list(face = 5, i = 15, j = 21, res = 5),  # i%3=0, j%3=0 ✓ (was 15,20)
    
    # Resolution 6 (Class I): eff_res=3, max=26
    list(face = 10, i = 20, j = 25, res = 6),
    list(face = 19, i = 26, j = 26, res = 6)
  )
  
  for (tc in test_cases) {
    index <- cpp_cell_to_index(tc$face, tc$i, tc$j, tc$res, 3, "z3")
    result <- cpp_index_to_cell(index, 3, "z3")
    
    expect_equal(result$face, tc$face)
    expect_equal(result$i, tc$i)
    expect_equal(result$j, tc$j)
    expect_equal(result$resolution, tc$res)
  }
})

# =============================================================================
# Face Encoding Tests
# =============================================================================

test_that("Z3: face number encoding", {
  # All faces should encode with leading zero if needed
  for (face in 0:19) {
    index <- cpp_cell_to_index(face, 0, 0, 1, 3, "z3")
    expect_equal(substr(index, 1, 2), sprintf("%02d", face))
    
    # Decode should work
    result <- cpp_index_to_cell(index, 3, "z3")
    expect_equal(result$face, face)
  }
})

test_that("Z3: resolution 0 (face only)", {
  # At resolution 0, only faces 0-11 are valid (the 12 pentagons)
  # Faces 12-19 only appear at resolution 1+
  for (face in c(0, 5, 10, 11)) {  # Changed from c(0, 5, 10, 15, 19)
    index <- cpp_cell_to_index(face, 0, 0, 0, 3, "z3")
    
    # Should be exactly 2 chars
    expect_equal(nchar(index), 2)
    
    # Decode
    result <- cpp_index_to_cell(index, 3, "z3")
    expect_equal(result$face, face)
    expect_equal(result$resolution, 0)
    expect_equal(result$i, 0)
    expect_equal(result$j, 0)
  }
})

# =============================================================================
# Hierarchy Tests
# =============================================================================

test_that("Z3: parent/child relationships", {
  # Create a cell at resolution 5 (Class II) with valid coordinates
  # i=10, j=11: i%3=1, j%3=2 ✓
  child <- cpp_cell_to_index(0, 10, 11, 5, 3, "z3")
  
  # Get parent
  parent <- cpp_get_parent_index(child, 3, "z3")
  
  # Parent should be prefix of child
  expect_true(substr(child, 1, nchar(parent)) == parent)
  
  # Get children
  children <- cpp_get_children_indices(parent, 3, "z3")
  
  # Original child should be in list
  expect_true(child %in% children)
  
  # Should have 3 children (aperture 3)
  expect_equal(length(children), 3)
})

test_that("Z3: multi-level parent traversal", {
  # Start at resolution 6 with valid coordinates
  idx_r6 <- cpp_cell_to_index(0, 20, 22, 6, 3, "z3")  # i%3=2, j%3=1 ✓
  
  # Get parent at each level
  idx_r5 <- cpp_get_parent_index(idx_r6, 3, "z3")
  idx_r4 <- cpp_get_parent_index(idx_r5, 3, "z3")
  idx_r3 <- cpp_get_parent_index(idx_r4, 3, "z3")
  
  # Each should be progressively shorter
  expect_true(nchar(idx_r5) < nchar(idx_r6))
  expect_true(nchar(idx_r4) < nchar(idx_r5))
  expect_true(nchar(idx_r3) < nchar(idx_r4))
  
  # Each should be prefix of child
  expect_true(substr(idx_r6, 1, nchar(idx_r5)) == idx_r5)
  expect_true(substr(idx_r5, 1, nchar(idx_r4)) == idx_r4)
  expect_true(substr(idx_r4, 1, nchar(idx_r3)) == idx_r3)
})

# =============================================================================
# Comparison and Utility Tests
# =============================================================================

test_that("Z3: index comparison", {
  idx1 <- cpp_cell_to_index(0, 0, 0, 2, 3, "z3")
  idx2 <- cpp_cell_to_index(0, 1, 0, 2, 3, "z3")
  idx3 <- cpp_cell_to_index(1, 0, 0, 2, 3, "z3")
  
  expect_equal(cpp_compare_indices(idx1, idx1), 0)
  expect_true(cpp_compare_indices(idx1, idx2) != 0)
  expect_true(cpp_compare_indices(idx1, idx3) != 0)
})

test_that("Z3: resolution extraction", {
  # Resolution extraction works for both Class I and Class II
  # Resolution = string length (after face), regardless of class

  # Test Class I resolutions (even)
  class_i_resolutions <- c(0, 2, 4, 6)
  
  for (res in class_i_resolutions) {
    index <- cpp_cell_to_index(0, 0, 0, res, 3, "z3")
    extracted <- cpp_get_index_resolution(index, 3, "z3")
    expect_equal(extracted, res,
                 label = sprintf("Class I res=%d", res))
  }
  
  # Test Class II resolutions (odd) - these also work correctly now
  class_ii_resolutions <- c(1, 3, 5, 7)
  
  for (res in class_ii_resolutions) {
    index <- cpp_cell_to_index(0, 0, 0, res, 3, "z3")
    extracted <- cpp_get_index_resolution(index, 3, "z3")
    expect_equal(extracted, res,
                 label = sprintf("Class II res=%d", res))
  }
})

# =============================================================================
# Difference from Z-Order Tests
# =============================================================================

test_that("Z3: produces different results than Z-Order", {
  # Z3 and Z-Order should give different indices for same input
  # IMPORTANT: Use VALID coordinates for Class II!
  test_cases <- list(
    list(i = 1, j = 2, res = 2),  # Class I: any coords OK
    list(i = 2, j = 1, res = 2),  # Class I: any coords OK
    list(i = 3, j = 3, res = 3)   # Class II: i%3=0, j%3=0 ✓ (was 3,4)
  )
  
  for (tc in test_cases) {
    idx_z3 <- cpp_cell_to_index(0, tc$i, tc$j, tc$res, 3, "z3")
    idx_zo <- cpp_cell_to_index(0, tc$i, tc$j, tc$res, 3, "zorder")
    
    # Should be different
    expect_false(idx_z3 == idx_zo,
                 label = sprintf("i=%d,j=%d,res=%d", tc$i, tc$j, tc$res))
    
    # But both should decode correctly
    result_z3 <- cpp_index_to_cell(idx_z3, 3, "z3")
    result_zo <- cpp_index_to_cell(idx_zo, 3, "zorder")
    
    expect_equal(result_z3$i, tc$i,
                 label = sprintf("Z3 i for (%d,%d)", tc$i, tc$j))
    expect_equal(result_z3$j, tc$j,
                 label = sprintf("Z3 j for (%d,%d)", tc$i, tc$j))
    expect_equal(result_zo$i, tc$i,
                 label = sprintf("ZOrder i for (%d,%d)", tc$i, tc$j))
    expect_equal(result_zo$j, tc$j,
                 label = sprintf("ZOrder j for (%d,%d)", tc$i, tc$j))
  }
})

test_that("Z3: provides better spatial locality than Z-Order", {
  # This is a qualitative test - Z3 should cluster spatially close cells
  # We test that neighboring cells have similar indices
  
  # Get indices for a small neighborhood
  idx_00 <- cpp_cell_to_index(0, 0, 0, 2, 3, "z3")
  idx_01 <- cpp_cell_to_index(0, 0, 1, 2, 3, "z3")
  idx_10 <- cpp_cell_to_index(0, 1, 0, 2, 3, "z3")
  
  # All should start with same face
  expect_equal(substr(idx_00, 1, 2), "00")
  expect_equal(substr(idx_01, 1, 2), "00")
  expect_equal(substr(idx_10, 1, 2), "00")
  
  # And have same length (Class I)
  expect_equal(nchar(idx_00), nchar(idx_01))
  expect_equal(nchar(idx_00), nchar(idx_10))
})

# tests/testthat/test-aperture-7.R
# Tests for aperture 7 (ISEA7H) hexagonal grid quantization
#
# Successive aperture-7 levels rotate by kAp7RotDeg in alternating directions,
# so even resolutions are unrotated (Class I) and odd resolutions sit at
# kAp7RotDeg on a sqrt(7) substrate.
# Each resolution has 7x the cells of the previous (area ratio 7:1).

# =============================================================================
# SETUP
# =============================================================================

setup_icosa <- function() {
  cpp_build_icosa()
}

# =============================================================================
# ROUND-TRIP TESTS
# =============================================================================

test_that("aperture 7 round-trip works", {
  setup_icosa()

  test_points <- list(
    c(0.4, 0.35),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0),
    c(0.7, 0.5)
  )

  for (res in c(0, 1, 2, 3, 4, 5)) {
    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell <- cpp_hex_quantize_ap7(tx, ty, res)
      center <- cpp_hex_center_ap7(cell["i"], cell["j"], res)
      cell2 <- cpp_hex_quantize_ap7(center["cx"], center["cy"], res)

      expect_equal(cell["i"], cell2["i"],
                   info = sprintf("res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
      expect_equal(cell["j"], cell2["j"],
                   info = sprintf("res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
    }
  }
})

test_that("aperture 7 batch round-trip succeeds", {
  setup_icosa()

  set.seed(789)
  n <- 20
  tx <- runif(n, -0.8, 0.8)
  ty <- runif(n, -0.8, 0.8)

  for (res in c(2, 3, 4, 5)) {
    result <- cpp_batch_test_roundtrip_ap7(tx, ty, res)
    n_pass <- sum(result$success)

    expect_equal(n_pass, n,
                 info = sprintf("res=%d: %d/%d passed", res, n_pass, n))
  }
})

# =============================================================================
# ORIENTATION ALTERNATION
# =============================================================================

test_that("aperture 7 keeps cell (0,0) at the origin at every resolution", {
  setup_icosa()

  for (res in c(0, 1, 2, 3, 4)) {
    center <- cpp_hex_center_ap7(0, 0, res)
    expect_equal(as.numeric(center["cx"]), 0, tolerance = 1e-10)
    expect_equal(as.numeric(center["cy"]), 0, tolerance = 1e-10)
  }
})

# =============================================================================
# SCALING AND REFINEMENT
# =============================================================================

test_that("aperture 7 substrate divisor is 7^ceil(res/2)", {
  setup_icosa()

  # Cell centres are substrate coordinates divided by sqrt(7)^res times the
  # substrate factor (1 at even resolutions, sqrt(7) at odd ones), so the
  # divisor is the integer 7^ceil(res/2) that the exact aperture-7 route in
  # coordinate_transforms.cpp uses.

  for (res in 0:5) {
    center <- cpp_hex_center_ap7(1, 0, res)
    unit <- cpp_hex_center_ap7(1, 0, 0)
    divisor <- as.numeric(unit["cx"]) / as.numeric(center["cx"])

    expect_equal(divisor, 7^ceiling(res / 2), tolerance = 1e-9,
                 info = sprintf("res=%d substrate divisor", res))
  }
})

test_that("aperture 7 is centre-nested: a parent centre is also a child centre", {
  setup_icosa()

  set.seed(707)
  pts <- cbind(runif(30, -0.7, 0.7), runif(30, -0.7, 0.7))

  for (res in 0:4) {
    child_spacing <- 1 / sqrt(7)^(res + 1)
    for (k in seq_len(nrow(pts))) {
      cell <- cpp_hex_quantize_ap7(pts[k, 1], pts[k, 2], res)
      p <- cpp_hex_center_ap7(cell["i"], cell["j"], res)
      kid <- cpp_hex_quantize_ap7(p["cx"], p["cy"], res + 1L)
      q <- cpp_hex_center_ap7(kid["i"], kid["j"], res + 1L)

      d <- sqrt((p["cx"] - q["cx"])^2 + (p["cy"] - q["cy"])^2)
      expect_lt(as.numeric(d), child_spacing * 1e-6)
    }
  }
})

# =============================================================================
# HEXAGON CORNERS
# =============================================================================

test_that("aperture 7 corners form valid hexagons", {
  setup_icosa()

  for (res in c(0, 1, 2, 3)) {
    corners <- cpp_hex_corners_ap7(0, 0, res, 1.0)

    expect_equal(length(corners$x), 6)
    expect_equal(length(corners$y), 6)
    expect_true(all(is.finite(corners$x)))
    expect_true(all(is.finite(corners$y)))
  }
})

# =============================================================================
# LON/LAT WORKFLOW
# =============================================================================

test_that("aperture 7 lon/lat workflow works", {
  setup_icosa()

  lon <- 16.37  # Vienna
  lat <- 48.21
  res <- 5

  cell <- cpp_lonlat_to_cell_ap7(lon, lat, res)

  expect_true(cell["face"] >= 0 && cell["face"] < 20)
  expect_true(is.numeric(cell["i"]))
  expect_true(is.numeric(cell["j"]))

  ll <- cpp_cell_to_lonlat_ap7(cell["face"], cell["i"], cell["j"], res)

  # Aperture 7 may have larger errors due to surrogate-substrate conversion
  dist <- sqrt((ll["lon"] - lon)^2 + (ll["lat"] - lat)^2)
  expect_true(dist < 15.0)
})

# =============================================================================
# SINGLE-POINT ROUNDTRIP TEST HELPER
# =============================================================================

test_that("cpp_test_roundtrip_ap7 returns TRUE for valid points", {
  setup_icosa()

  test_points <- list(
    c(0.5, 0.3),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0)
  )

  for (res in c(2, 3, 4, 5)) {
    for (pt in test_points) {
      result <- cpp_test_roundtrip_ap7(pt[1], pt[2], res)
      expect_true(result, info = sprintf("res=%d, pt=(%.2f, %.2f)", res, pt[1], pt[2]))
    }
  }
})

# tests/testthat/test-aperture-4.R
# Tests for aperture 4 (ISEA4H) hexagonal grid quantization
#
# Aperture 4 uses Class I (flat-top) at all resolutions.
# Each resolution has 4x the cells of the previous (area ratio 4:1).

# =============================================================================
# SETUP
# =============================================================================

setup_icosa <- function() {
  cpp_build_icosa()
}

# =============================================================================
# ROUND-TRIP TESTS
# =============================================================================

test_that("aperture 4 round-trip works", {
  setup_icosa()

  test_points <- list(
    c(0.4, 0.35),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0),
    c(0.7, 0.5)
  )

  for (res in c(0, 1, 2, 3, 4, 5)) {
    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell <- cpp_hex_quantize_ap4(tx, ty, res)
      center <- cpp_hex_center_ap4(cell["i"], cell["j"], res)
      cell2 <- cpp_hex_quantize_ap4(center["cx"], center["cy"], res)

      expect_equal(cell["i"], cell2["i"],
                   info = sprintf("res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
      expect_equal(cell["j"], cell2["j"],
                   info = sprintf("res=%d, tx=%.3f, ty=%.3f", res, tx, ty))
    }
  }
})

test_that("aperture 4 batch round-trip succeeds", {
  setup_icosa()

  set.seed(456)
  n <- 20
  tx <- runif(n, -0.8, 0.8)
  ty <- runif(n, -0.8, 0.8)

  for (res in c(2, 3, 4, 5)) {
    result <- cpp_batch_test_roundtrip_ap4(tx, ty, res)
    n_pass <- sum(result$success)

    expect_equal(n_pass, n,
                 info = sprintf("res=%d: %d/%d passed", res, n_pass, n))
  }
})

# =============================================================================
# CLASS I CONSISTENCY
# =============================================================================

test_that("aperture 4 always uses Class I (no rotation)", {
  setup_icosa()

  for (res in c(0, 1, 2, 3, 4)) {
    # Cell (0,0) should always be at origin
    center <- cpp_hex_center_ap4(0, 0, res)
    expect_equal(as.numeric(center["cx"]), 0, tolerance = 1e-10)
    expect_equal(as.numeric(center["cy"]), 0, tolerance = 1e-10)

    # Cell (1,0) should follow Class I formula
    center <- cpp_hex_center_ap4(1, 0, res)
    expected_x <- 1.0 / (2^res)
    expect_equal(as.numeric(center["cx"]), expected_x, tolerance = 1e-10)
    expect_equal(as.numeric(center["cy"]), 0, tolerance = 1e-10)

    # Cell (0,1) should follow Class I formula: x = -0.5/scale
    center <- cpp_hex_center_ap4(0, 1, res)
    expected_x <- -0.5 / (2^res)
    expected_y <- sin(60 * pi / 180) / (2^res)
    expect_equal(as.numeric(center["cx"]), expected_x, tolerance = 1e-10)
    expect_equal(as.numeric(center["cy"]), expected_y, tolerance = 1e-10)
  }
})

# =============================================================================
# SCALING AND REFINEMENT
# =============================================================================

test_that("aperture 4 refines by factor of 4", {
  setup_icosa()

  # Scale should double each resolution (2^res), area quarters
  for (res in 0:4) {
    scale <- 2^res

    center <- cpp_hex_center_ap4(1, 0, res)
    expected_x <- 1.0 / scale

    expect_equal(as.numeric(center["cx"]), expected_x, tolerance = 1e-10,
                 info = sprintf("res=%d: cell (1,0) x-coord", res))
  }
})

test_that("aperture 4 cell area ratios are correct", {
  setup_icosa()

  # Adjacent resolutions should have area ratio of 4:1
  for (res in 1:4) {
    # Use cell spacing as proxy for area
    center_cur <- cpp_hex_center_ap4(1, 0, res)
    center_prev <- cpp_hex_center_ap4(1, 0, res - 1)

    spacing_cur <- as.numeric(center_cur["cx"])
    spacing_prev <- as.numeric(center_prev["cx"])

    ratio <- spacing_prev / spacing_cur
    expect_equal(ratio, 2, tolerance = 1e-10,
                 info = sprintf("res=%d spacing ratio", res))
  }
})

# =============================================================================
# HEXAGON CORNERS
# =============================================================================

test_that("aperture 4 corners form valid hexagons", {
  setup_icosa()

  for (res in c(0, 1, 2, 3)) {
    corners <- cpp_hex_corners_ap4(0, 0, res, 1.0)

    expect_equal(length(corners$x), 6)
    expect_equal(length(corners$y), 6)
    expect_true(all(is.finite(corners$x)))
    expect_true(all(is.finite(corners$y)))
  }
})

# =============================================================================
# LON/LAT WORKFLOW
# =============================================================================

test_that("aperture 4 lon/lat workflow works", {
  setup_icosa()

  lon <- 16.37  # Vienna
  lat <- 48.21
  res <- 6

  cell <- cpp_lonlat_to_cell_ap4(lon, lat, res)

  expect_true(cell["face"] >= 0 && cell["face"] < 20)
  expect_true(is.numeric(cell["i"]))
  expect_true(is.numeric(cell["j"]))

  ll <- cpp_cell_to_lonlat_ap4(cell["face"], cell["i"], cell["j"], res)

  # Should be reasonably close
  dist <- sqrt((ll["lon"] - lon)^2 + (ll["lat"] - lat)^2)
  expect_true(dist < 10.0)
})

# =============================================================================
# SINGLE-POINT ROUNDTRIP TEST HELPER
# =============================================================================

test_that("cpp_test_roundtrip_ap4 returns TRUE for valid points", {
  setup_icosa()

  test_points <- list(
    c(0.5, 0.3),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0)
  )

  for (res in c(2, 3, 4, 5)) {
    for (pt in test_points) {
      result <- cpp_test_roundtrip_ap4(pt[1], pt[2], res)
      expect_true(result, info = sprintf("res=%d, pt=(%.2f, %.2f)", res, pt[1], pt[2]))
    }
  }
})

# tests/testthat/test-compact.R
# Tests for hex_compact() and hex_uncompact()

test_that("H3 compact/uncompact roundtrip", {
  g <- hex_grid(resolution = 3, type = "h3")

  # Get children of a parent cell
  parent <- lonlat_to_cell(10, 50, g)
  parent_res2 <- cpp_h3_cellToParent(parent, 2L)
  children <- cpp_h3_cellToChildren(parent_res2, 3L)[[1]]

  # Compact children should give parent

  compacted <- hex_compact(children, g)
  expect_true(parent_res2 %in% compacted)

  # Uncompact should give back children
  uncompacted <- hex_uncompact(compacted, g, target_resolution = 3L)
  expect_equal(sort(uncompacted), sort(children))
})

test_that("ISEA Z7 compact/uncompact roundtrip", {
  g <- hex_grid(resolution = 3, aperture = 7)

  # Create 7 Z7 children of a parent
  parent <- "010"  # quad 01, digit 0 (res 1)
  children <- paste0(parent, 0:6)

  compacted <- hex_compact(children, g)
  expect_true(parent %in% compacted)

  uncompacted <- hex_uncompact(compacted, g, target_resolution = 3L)
  # Should have 7^1 = 7 cells at target res relative to the single res-2 parent
  expect_true(all(nchar(uncompacted) == 5))  # "01" + 3 digits
})

test_that("hex_compact with no full parents returns unchanged", {
  g <- hex_grid(resolution = 3, type = "h3")
  cells <- lonlat_to_cell(c(10, 20, 30), c(50, 55, 60), g)
  compacted <- hex_compact(cells, g)
  # Non-sibling cells should not be compacted
  expect_equal(length(compacted), length(cells))
})

test_that("hex_compact errors for non-aperture-7 ISEA", {
  g <- hex_grid(resolution = 3, aperture = 3)
  expect_error(hex_compact(c("0100", "0101", "0102"), g), "aperture 7")
})

# tests/testthat/test-aperture-mixed.R
# Tests for mixed aperture sequences
#
# An aperture sequence gives the aperture of every resolution level. ISEA43H is
# aperture 4 for the first few levels then aperture 3, which is DGGRID's
# standard configuration; apertures 3, 4 and 7 may appear in any order.

# =============================================================================
# SETUP
# =============================================================================

setup_icosa <- function() {
  cpp_build_icosa()
}

# The ISEA43H sequence: `level` aperture-4 refinements then aperture 3.
seq43 <- function(res, level) {
  c(if (level > 0) 4 else 3, rep(4, level), rep(3, res - level))
}

# =============================================================================
# PURE SEQUENCE CONSISTENCY
# =============================================================================

test_that("pure aperture 4 sequence matches ap4 implementation", {
  skip_on_cran()  # Detailed consistency check
  setup_icosa()

  # Pure ap4: sequence of all 4s
  for (res in c(0, 1, 2, 3)) {
    ap_seq <- rep(4, res + 1)

    test_points <- list(c(0.4, 0.35))

    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell_34 <- cpp_hex_quantize_mixed(tx, ty, ap_seq)
      cell_4 <- cpp_hex_quantize_ap4(tx, ty, res)

      expect_equal(cell_34["i"], cell_4["i"],
                   info = sprintf("res=%d, pure ap4 i mismatch", res))
      expect_equal(cell_34["j"], cell_4["j"],
                   info = sprintf("res=%d, pure ap4 j mismatch", res))
    }
  }
})

test_that("pure aperture 3 sequence matches ap3 implementation", {
  skip_on_cran()  # Detailed consistency check
  setup_icosa()

  # Pure ap3: sequence of all 3s
  for (res in c(1, 2, 3, 4)) {
    ap_seq <- rep(3, res + 1)

    test_points <- list(
      c(0.4, 0.35),
      c(-0.4, 0.2),
      c(0.1, -0.6)
    )

    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell_34 <- cpp_hex_quantize_mixed(tx, ty, ap_seq)
      cell_3 <- cpp_hex_quantize_ap3(tx, ty, res)

      expect_equal(cell_34["i"], cell_3["i"],
                   info = sprintf("res=%d, pure ap3 i mismatch", res))
      expect_equal(cell_34["j"], cell_3["j"],
                   info = sprintf("res=%d, pure ap3 j mismatch", res))

      # Compare centers
      center_34 <- cpp_hex_center_mixed(cell_34["i"], cell_34["j"], ap_seq)
      center_3 <- cpp_hex_center_ap3(cell_3["i"], cell_3["j"], res)

      expect_equal(as.numeric(center_34["cx"]), as.numeric(center_3["cx"]),
                   tolerance = 1e-10)
      expect_equal(as.numeric(center_34["cy"]), as.numeric(center_3["cy"]),
                   tolerance = 1e-10)
    }
  }
})

test_that("pure aperture 7 sequence matches ap7 implementation", {
  skip_on_cran()  # Detailed consistency check
  setup_icosa()

  set.seed(77)
  pts <- cbind(runif(40, -0.8, 0.8), runif(40, -0.8, 0.8))

  for (res in 0:5) {
    ap_seq <- rep(7, res + 1)

    for (k in seq_len(nrow(pts))) {
      tx <- pts[k, 1]
      ty <- pts[k, 2]

      cell_mixed <- cpp_hex_quantize_mixed(tx, ty, ap_seq)
      cell_7 <- cpp_hex_quantize_ap7(tx, ty, res)

      expect_equal(cell_mixed["i"], cell_7["i"],
                   info = sprintf("res=%d, pure ap7 i mismatch", res))
      expect_equal(cell_mixed["j"], cell_7["j"],
                   info = sprintf("res=%d, pure ap7 j mismatch", res))

      center_mixed <- cpp_hex_center_mixed(cell_mixed["i"], cell_mixed["j"], ap_seq)
      center_7 <- cpp_hex_center_ap7(cell_7["i"], cell_7["j"], res)

      expect_equal(as.numeric(center_mixed["cx"]), as.numeric(center_7["cx"]),
                   tolerance = 1e-10)
      expect_equal(as.numeric(center_mixed["cy"]), as.numeric(center_7["cy"]),
                   tolerance = 1e-10)
    }
  }
})

test_that("aperture sequences reject unsupported apertures", {
  setup_icosa()

  expect_error(cpp_hex_quantize_mixed(0.1, 0.2, c(3, 5)), "aperture must be 3, 4, or 7")
  expect_error(cpp_hex_quantize_mixed(0.1, 0.2, c(5, 3)), "aperture must be 3, 4, or 7")
  expect_error(cpp_hex_quantize_mixed(0.1, 0.2, integer(0)), "cannot be empty")
})

# =============================================================================
# MIXED 4/3 PATTERN (DGGRID STYLE)
# =============================================================================

test_that("mixed 43H pattern (DGGRID style) works", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  # DGGRID 43H: first 2 resolutions ap4, then ap3
  # c(4, 4, 3, 3, 3, ...)

  for (total_res in 2:5) {
    ap_seq <- c(4, 4, rep(3, max(0, total_res - 1)))

    test_points <- list(
      c(0.4, 0.35),
      c(-0.4, 0.2),
      c(0.1, -0.6)
    )

    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell <- cpp_hex_quantize_mixed(tx, ty, ap_seq)
      center <- cpp_hex_center_mixed(cell["i"], cell["j"], ap_seq)
      cell2 <- cpp_hex_quantize_mixed(center["cx"], center["cy"], ap_seq)

      expect_equal(cell["i"], cell2["i"],
                   info = sprintf("43H res=%d round-trip i", total_res))
      expect_equal(cell["j"], cell2["j"],
                   info = sprintf("43H res=%d round-trip j", total_res))
    }
  }
})

# =============================================================================
# ROUND-TRIP
# =============================================================================

test_that("mixed aperture round-trip works for various sequences", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  sequences <- list(
    c(4, 4),           # Two levels of ap4
    c(4, 4, 3),        # Standard 43H start
    c(4, 4, 3, 3),     # Longer 43H
    c(4, 4, 3, 3, 3),  # Even longer
    c(3, 4),           # Start with 3, then 4
    c(4, 3, 4)         # Alternating
  )

  test_points <- list(
    c(0.4, 0.35),
    c(-0.4, 0.2),
    c(0.0, 0.0)
  )

  for (seq in sequences) {
    for (pt in test_points) {
      tx <- pt[1]
      ty <- pt[2]

      cell <- cpp_hex_quantize_mixed(tx, ty, seq)
      center <- cpp_hex_center_mixed(cell["i"], cell["j"], seq)
      cell2 <- cpp_hex_quantize_mixed(center["cx"], center["cy"], seq)

      seq_str <- paste(seq, collapse = ",")
      expect_equal(cell["i"], cell2["i"],
                   info = sprintf("seq=[%s] round-trip i", seq_str))
      expect_equal(cell["j"], cell2["j"],
                   info = sprintf("seq=[%s] round-trip j", seq_str))
    }
  }
})

# =============================================================================
# CORNERS
# =============================================================================

test_that("mixed aperture corners form valid hexagons", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  sequences <- list(
    c(4, 4),
    c(4, 4, 3),
    c(4, 4, 3, 3),
    c(4, 7),
    c(7, 4),
    c(4, 7, 3, 7)
  )

  for (seq in sequences) {
    corners <- cpp_hex_corners_mixed(0, 0, seq, 1.0)

    expect_equal(length(corners$x), 6)
    expect_equal(length(corners$y), 6)
    expect_true(all(is.finite(corners$x)))
    expect_true(all(is.finite(corners$y)))
  }
})

# =============================================================================
# APERTURE-7 SEQUENCES (issue #55)
# =============================================================================
#
# Aperture 7 rotates the lattice by kAp7RotDeg per level rather than by a
# multiple of 30 degrees, so it composes with apertures 3 and 4 through the
# Eisenstein-generator model in grid_math.h rather than a Class I/II flag.

test_that("sequences mixing aperture 7 with 3 and 4 round-trip", {
  setup_icosa()

  sequences <- list(
    c(4, 7),
    c(7, 4),
    c(4, 4, 7),
    c(7, 7, 4),
    c(3, 7),
    c(7, 3),
    c(4, 7, 3),
    c(7, 4, 3, 7),
    c(4, 4, 7, 7, 3)
  )

  set.seed(755)
  pts <- cbind(runif(25, -0.8, 0.8), runif(25, -0.8, 0.8))

  for (seq in sequences) {
    seq_str <- paste(seq, collapse = ",")
    for (k in seq_len(nrow(pts))) {
      cell <- cpp_hex_quantize_mixed(pts[k, 1], pts[k, 2], seq)
      center <- cpp_hex_center_mixed(cell["i"], cell["j"], seq)
      cell2 <- cpp_hex_quantize_mixed(center["cx"], center["cy"], seq)

      expect_equal(cell["i"], cell2["i"], info = sprintf("seq=[%s] i", seq_str))
      expect_equal(cell["j"], cell2["j"], info = sprintf("seq=[%s] j", seq_str))
    }
  }
})

test_that("every aperture sequence is centre-nested in its refinements", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  # A cell's centre must also be a centre in the next finer grid, whatever
  # aperture that level uses. This is what breaks when the orientation of a
  # level is computed independently of the levels above it.
  parents <- list(
    c(3), c(4), c(7),
    c(3, 3), c(4, 4), c(7, 7),
    c(4, 3), c(3, 4), c(4, 7), c(7, 4), c(3, 7), c(7, 3),
    c(4, 4, 3), c(4, 4, 7), c(7, 7, 4), c(4, 7, 3),
    c(4, 4, 3, 3), c(7, 4, 3, 7)
  )

  set.seed(756)
  pts <- cbind(runif(25, -0.7, 0.7), runif(25, -0.7, 0.7))

  for (parent in parents) {
    for (nxt in c(3, 4, 7)) {
      child <- c(parent, nxt)
      child_spacing <- 1 / prod(sqrt(child[-1]))
      seq_str <- paste(child, collapse = ",")

      for (k in seq_len(nrow(pts))) {
        cell <- cpp_hex_quantize_mixed(pts[k, 1], pts[k, 2], parent)
        p <- cpp_hex_center_mixed(cell["i"], cell["j"], parent)
        kid <- cpp_hex_quantize_mixed(p["cx"], p["cy"], child)
        q <- cpp_hex_center_mixed(kid["i"], kid["j"], child)

        d <- sqrt((p["cx"] - q["cx"])^2 + (p["cy"] - q["cy"])^2)
        expect_lt(as.numeric(d), child_spacing * 1e-6,
                  label = sprintf("seq=[%s] point %d centre offset", seq_str, k))
      }
    }
  }
})

test_that("aperture order does not change the grid", {
  setup_icosa()

  # Scale is a product and orientation composes commutatively, so sequences
  # that use the same multiset of apertures describe the same grid. Each pair
  # swaps two adjacent steps, which leaves the accumulated scale bit-identical.
  set.seed(757)
  pts <- cbind(runif(25, -0.8, 0.8), runif(25, -0.8, 0.8))

  pairs <- list(
    list(c(4, 4, 3, 7), c(4, 3, 4, 7)),
    list(c(3, 3, 4, 7), c(3, 4, 3, 7)),
    list(c(7, 7, 4, 3), c(7, 4, 7, 3))
  )

  for (pr in pairs) {
    for (k in seq_len(nrow(pts))) {
      a <- cpp_hex_quantize_mixed(pts[k, 1], pts[k, 2], pr[[1]])
      b <- cpp_hex_quantize_mixed(pts[k, 1], pts[k, 2], pr[[2]])
      expect_equal(a["i"], b["i"])
      expect_equal(a["j"], b["j"])
    }
  }
})

# =============================================================================
# TRUE MIXED-RADIX SUBSTRATE (i,j) -- not a pure aperture-3 approximation
# =============================================================================
#
# The mixed cell-ID path used to quantize (i,j) as pure aperture-3 at the
# full target resolution, ignoring mixed_aperture_level entirely (only the
# cell-count/offset arithmetic accounted for the aperture-4 prefix). That
# meant distinct real cells could collide onto the same cell ID once the
# aperture-4 prefix meaningfully changed the coordinate scale. These tests
# lock in the fix: (i,j) now come from the same 2^level * sqrt(3)^(res-level)
# substrate that calc_grid_params_mixed()'s cell-count formula describes.

test_that("mixed aperture (i,j) does not collide across many sampled points", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  set.seed(43)
  lons <- runif(2000, -170, 170)
  lats <- runif(2000, -80, 80)

  for (mixed_level in c(1, 2, 3)) {
    for (res in c(3, 4, 5, 6)) {
      if (mixed_level >= res) next
      cells <- cpp_lonlat_to_cell_seq(lons, lats, seq43(res, mixed_level))
      centers <- cpp_cell_to_lonlat_seq(cells, seq43(res, mixed_level))
      cells2 <- cpp_lonlat_to_cell_seq(centers$lon_deg, centers$lat_deg, seq43(res, mixed_level))

      expect_equal(cells2, cells,
                   info = sprintf("mixed_level=%d res=%d: cell id round-trip", mixed_level, res))
    }
  }
})

test_that("mixed aperture (i,j) matches the mixed-radix grid dimension, not pure aperture 3", {
  skip_on_cran()
  setup_icosa()

  # Regression check for the specific collision from issue #31: at
  # mixed_aperture_level=2 res=5 the pure-aperture-3-at-full-resolution
  # approximation produced (i,j) an order of magnitude too small, causing
  # unrelated points to land on the same cell ID.
  set.seed(44)
  lons <- runif(500, -170, 170)
  lats <- runif(500, -80, 80)

  res <- 5
  mixed_level <- 2
  cells <- cpp_lonlat_to_cell_seq(lons, lats, seq43(res, mixed_level))
  qij <- cpp_cell_to_quad_ij_seq(cells, seq43(res, mixed_level))

  # True mixed-radix dim: 2^2 * sqrt(3)^3 (+ substrate boost if class II)
  ap3_count <- res - mixed_level
  true_scale <- 2^mixed_level * sqrt(3)^ap3_count
  if (ap3_count %% 2 == 1) true_scale <- true_scale * sqrt(3)
  pure_ap3_scale <- sqrt(3)^res

  expect_true(true_scale > pure_ap3_scale * 1.5,
              info = "sanity: mixed-radix scale is meaningfully larger than pure ap3 at this res")
  expect_true(max(qij$i) <= true_scale + 1,
              info = "i must not exceed the true mixed-radix grid dimension")
  expect_true(max(qij$j) <= true_scale + 1,
              info = "j must not exceed the true mixed-radix grid dimension")
})

# =============================================================================
# EDGE CASES - MIXED APERTURE LEVEL BOUNDARIES
# =============================================================================

test_that("mixed aperture level = 0 (all aperture 3)", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  # When mixed_aperture_level = 0, should behave like pure aperture 3
  test_lon <- c(0, 45, -120, 16.37)
  test_lat <- c(0, 30, -45, 48.21)

  for (res in c(3, 5, 7)) {
    # Mixed with level 0 = all ap3
    cell_mixed <- cpp_lonlat_to_cell_seq(test_lon, test_lat, seq43(res, 0))

    # Pure aperture 3
    cell_pure <- cpp_lonlat_to_cell(test_lon, test_lat, res, 3)

    expect_equal(cell_mixed, cell_pure,
                 info = sprintf("res=%d, mixed_level=0 should equal pure ap3", res))
  }
})

test_that("mixed aperture level = resolution produces valid cells", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  # When mixed_aperture_level = resolution, all resolutions use aperture 4
  # Note: Cell IDs may differ from pure ap4 due to different numbering schemes
  test_lon <- c(0, 45, -120, 16.37)
  test_lat <- c(0, 30, -45, 48.21)

  for (res in c(2, 4, 6)) {
    # Mixed with level = res = all ap4 subdivisions
    cell_mixed <- cpp_lonlat_to_cell_seq(test_lon, test_lat, seq43(res, res))

    # Verify cells are valid (positive integers)
    expect_true(all(cell_mixed >= 1),
                info = sprintf("res=%d, mixed_level=res should produce valid cells", res))
    expect_true(all(is.finite(cell_mixed)),
                info = sprintf("res=%d cells should be finite", res))

    # Verify round-trip works
    centers <- cpp_cell_to_lonlat_seq(cell_mixed, seq43(res, res))
    cell2 <- cpp_lonlat_to_cell_seq(centers$lon_deg, centers$lat_deg, seq43(res, res))
    expect_equal(cell_mixed, cell2,
                 info = sprintf("res=%d, mixed_level=res round-trip", res))
  }
})

test_that("mixed aperture round-trip at boundary resolutions", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  test_lon <- c(0, 45, -120)
  test_lat <- c(0, 30, -45)

  # Test at resolution boundaries
  # Skip res=1 with mixed_level=1 as it's a degenerate case
  for (res in c(2, 5, 10)) {
    for (mixed_level in c(0, 1, min(res - 1, 5))) {
      cell <- cpp_lonlat_to_cell_seq(test_lon, test_lat, seq43(res, mixed_level))
      centers <- cpp_cell_to_lonlat_seq(cell, seq43(res, mixed_level))
      cell2 <- cpp_lonlat_to_cell_seq(centers$lon_deg, centers$lat_deg, seq43(res, mixed_level))

      expect_equal(cell, cell2,
                   info = sprintf("res=%d, mixed_level=%d round-trip", res, mixed_level))
    }
  }
})

test_that("mixed aperture rejects an unsupported aperture", {
  setup_icosa()

  expect_error(
    cpp_lonlat_to_cell_seq(0, 0, c(4, 4, 5)),
    "aperture must be 3, 4, or 7"
  )

  expect_error(
    cpp_lonlat_to_cell_seq(0, 0, integer(0)),
    "ap_seq must name at least the base grid"
  )
})

# =============================================================================
# MIXED APERTURE LON/LAT WORKFLOW
# =============================================================================

test_that("cpp_lonlat_to_cell_mixed and cpp_cell_to_lonlat_mixed work", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  lon <- 16.37  # Vienna
  lat <- 48.21

  # Test with various aperture sequences
  sequences <- list(
    c(4, 3, 3, 3),      # res 4, one aperture-4 step
    c(4, 4, 3, 3),      # res 4, two aperture-4 steps
    c(3, 3, 3, 3, 3)    # res 5, all aperture-3
  )

  for (ap_seq in sequences) {
    cell <- cpp_lonlat_to_cell_mixed(lon, lat, ap_seq)

    expect_true(cell["face"] >= 0 && cell["face"] < 20)
    expect_true(is.numeric(cell["i"]))
    expect_true(is.numeric(cell["j"]))

    ll <- cpp_cell_to_lonlat_mixed(cell["face"], cell["i"], cell["j"], ap_seq)

    expect_true(ll["lon"] >= -180 && ll["lon"] <= 180)
    expect_true(ll["lat"] >= -90 && ll["lat"] <= 90)
  }
})

test_that("cpp_test_roundtrip_mixed returns TRUE for valid points", {
  skip_on_cran()  # Detailed loop test
  setup_icosa()

  test_points <- list(
    c(0.5, 0.3),
    c(-0.4, 0.2),
    c(0.1, -0.6),
    c(0.0, 0.0)
  )

  sequences <- list(
    c(4, 3, 3, 3),
    c(4, 4, 3, 3),
    c(3, 3, 3, 3, 3)
  )

  for (ap_seq in sequences) {
    for (pt in test_points) {
      result <- cpp_test_roundtrip_mixed(pt[1], pt[2], ap_seq)
      expect_true(result, info = sprintf("ap_seq length=%d, pt=(%.2f, %.2f)",
                                          length(ap_seq), pt[1], pt[2]))
    }
  }
})

# =============================================================================
# HIERARCHICAL NAVIGATION (issue #31): cell_to_index / get_parent / get_children
# =============================================================================
#
# Mixed 4/3 grids have no DGGRID-standard hierarchical index. hexify defines the
# hierarchy geometrically (a cell's parent is the coarser cell containing its
# centre; see R/aperture_mixed_hierarchy.R). These functions used to throw on
# every call for aperture "4/3"; they now navigate the hierarchy and round-trip.

test_that("mixed 4/3 hierarchical navigation no longer errors (issue #31)", {
  setup_icosa()
  grid <- hex_grid(resolution = 6, aperture = "4/3")
  cells <- lonlat_to_cell(c(0, 30, -60), c(10, 45, -30), grid)

  expect_silent(p <- get_parent(cells, grid))
  expect_silent(idx <- cell_to_index(cells, grid))
  expect_silent(kids <- get_children(cells, grid))

  expect_true(all(p >= 1))
  expect_type(idx, "character")
  expect_length(idx, length(cells))
  expect_type(kids, "list")
  expect_length(kids, length(cells))
})

test_that("mixed 4/3 get_parent(get_children()) recovers the parent", {
  setup_icosa()
  for (res in c(3, 4, 5, 6)) {
    grid  <- hex_grid(resolution = res, aperture = "4/3")
    cgrid <- hex_grid(resolution = res + 1L, aperture = "4/3")
    set.seed(res)
    cells <- unique(lonlat_to_cell(runif(40, -180, 180), runif(40, -85, 85), grid))
    kids <- get_children(cells, grid)
    for (i in seq_along(cells)) {
      k <- kids[[i]]
      expect_true(length(k) >= 1,
                  info = sprintf("res=%d cell=%.0f has at least one child", res, cells[i]))
      back <- get_parent(k, cgrid)
      expect_true(all(back == cells[i]),
                  info = sprintf("res=%d cell=%.0f: every child maps back to it", res, cells[i]))
    }
  }
})

test_that("mixed 4/3 cell_to_index round-trips to the original cell", {
  setup_icosa()
  decode <- hexify:::mixed_index_to_cell_one
  for (res in c(2, 4, 6, 7)) {
    grid <- hex_grid(resolution = res, aperture = "4/3")
    set.seed(100 + res)
    # Include high-latitude (near-pole) points so the icosahedron-vertex scatter
    # case is exercised, not just interior cells.
    lon <- c(runif(80, -180, 180), runif(20, -180, 180))
    lat <- c(runif(80, -85, 85), runif(20, 80, 89.9) * sample(c(-1, 1), 20, TRUE))
    cells <- unique(lonlat_to_cell(lon, lat, grid))
    idx <- cell_to_index(cells, grid)
    back <- vapply(idx, function(s) decode(s, res, "4/3"), numeric(1))
    expect_equal(unname(back), cells,
                 info = sprintf("res=%d: index round-trip", res))
    # Parent's index is a prefix of the child's index.
    parents <- get_parent(cells, grid)
    pidx <- cell_to_index(parents, hex_grid(resolution = res - 1L, aperture = "4/3"))
    expect_true(all(substr(idx, 1, nchar(pidx)) == pidx),
                info = sprintf("res=%d: parent index is a prefix", res))
  }
})

test_that("mixed 4/3 point lookup stays in range for a slipped cell centre", {
  setup_icosa()
  # A centre sitting on a quad corner quantizes to a tie, and the twelve
  # icosahedron vertices put such corners at the poles. A sub-microdegree slip
  # either way still has to name a cell of the grid.
  for (res in 1:4) {
    n <- hexify:::aperture_n_cells("4/3", res)
    ll <- hexify:::mixed_cell_center(seq_len(n), res, "4/3")
    for (e in c(0, 1e-11, 1e-9, 1e-7, 1e-5)) {
      for (d in list(c(e, 0), c(-e, 0), c(0, e), c(0, -e))) {
        got <- hexify:::mixed_point_to_cell(ll$lon_deg + d[1], ll$lat_deg + d[2],
                                            res, "4/3")
        expect_true(all(is.finite(got) & got >= 1 & got <= n),
                    info = sprintf("res=%d eps=%g", res, e))
      }
    }
  }
})

test_that("mixed 4/3 get_children is exactly the geometric-parent inverse", {
  skip_on_cran()  # brute-force ground-truth comparison
  setup_icosa()
  # At a small resolution, brute-force {c : parent(c) == P} and compare.
  for (res in c(2, 3)) {
    grid  <- hex_grid(resolution = res, aperture = "4/3")
    cgrid <- hex_grid(resolution = res + 1L, aperture = "4/3")
    ncc <- hexify:::aperture_n_cells("4/3", res + 1L)
    all_child <- seq_len(ncc)
    par_of_child <- get_parent(all_child, cgrid)
    parents <- sort(unique(par_of_child))
    for (p in parents) {
      truth <- sort(all_child[par_of_child == p])
      got <- get_children(p, grid)[[1]]
      expect_equal(got, truth,
                   info = sprintf("res=%d parent=%.0f exact children", res, p))
    }
  }
})

test_that("mixed 4/3 multi-level parent and children are consistent", {
  setup_icosa()
  grid <- hex_grid(resolution = 6, aperture = "4/3")
  g2   <- hex_grid(resolution = 4, aperture = "4/3")
  set.seed(7)
  cells <- unique(lonlat_to_cell(runif(30, -180, 180), runif(30, -85, 85), grid))
  gp2 <- get_parent(cells, grid, levels = 2L)
  expect_true(all(gp2 >= 1))
  desc <- get_children(gp2, g2, levels = 2L)
  for (i in seq_along(cells)) {
    expect_true(cells[i] %in% desc[[i]],
                info = sprintf("cell %.0f is a 2-level descendant of its 2-up parent", cells[i]))
  }
})

# =============================================================================
# R-LEVEL SPELLINGS (issue #57)
# =============================================================================
#
# hex_grid() takes a family name ("4/3", "4/7", "7/4") or one aperture per
# resolution level (c(4, 4, 7, 3)). Sequences containing an odd number of
# aperture-7 levels sit on a norm-7 or norm-21 substrate, which packs cell IDs
# on the same sublattice numbering as the Class II lattice of aperture 3.

test_that("hex_grid accepts family spellings and per-level sequences", {
  setup_icosa()

  g <- hex_grid(resolution = 4, aperture = "4/7")
  expect_equal(g@aperture, "4/7")
  expect_equal(hexify:::grid_ap_seq(g), c(4L, 4L, 4L, 7L, 7L))

  g2 <- hex_grid(resolution = 4, aperture = c(4, 7, 3, 7))
  expect_equal(g2@aperture, "4/7/3/7")
  expect_equal(hexify:::grid_ap_seq(g2), c(4L, 4L, 7L, 3L, 7L))

  # Cell count is the product over the sequence, plus the two poles
  expect_equal(hexify:::aperture_n_cells("4/7", 4), 10 * 4 * 4 * 7 * 7 + 2)
  expect_equal(hexify:::aperture_n_cells("4/7/3/7", 4), 10 * 4 * 7 * 3 * 7 + 2)

  expect_error(hex_grid(resolution = 4, aperture = c(4, 5, 3, 7)),
               "must be one of")
  expect_error(hex_grid(resolution = 4, aperture = c(4, 7, 3)),
               "one aperture per resolution level")
})

test_that("mixed sequence cell IDs are a bijection with (quad, i, j)", {
  skip_on_cran()  # Enumerates every cell of the grid
  setup_icosa()

  for (ap in c("4/3", "4/7", "7/4", "3/7", "7/3")) {
    for (res in 1:3) {
      g <- hex_grid(resolution = res, aperture = ap)
      ap_seq <- hexify:::grid_ap_seq(g)
      ids <- seq_len(hexify:::aperture_n_cells(ap, res))

      qij <- cpp_cell_to_quad_ij_seq(as.numeric(ids), ap_seq)
      back <- cpp_quad_ij_to_cell_seq(qij$quad, qij$i, qij$j, ap_seq)

      expect_equal(back, as.numeric(ids),
                   info = sprintf("%s res %d: cell ID round-trip", ap, res))
      expect_equal(length(unique(paste(qij$quad, qij$i, qij$j))), length(ids),
                   info = sprintf("%s res %d: distinct (quad, i, j) per cell", ap, res))

      dim <- cpp_ap_seq_edge_dim(ap_seq)
      expect_true(all(qij$i >= 0 & qij$i < dim & qij$j >= 0 & qij$j < dim),
                  info = sprintf("%s res %d: (i, j) within the quad", ap, res))
    }
  }
})

test_that("mixed sequence points land in a cell centred near them", {
  skip_on_cran()  # Sampled over the sphere
  setup_icosa()

  gc_km <- function(lon1, lat1, lon2, lat2) {
    r <- pi / 180
    6371 * acos(pmin(1, pmax(-1, sin(lat1 * r) * sin(lat2 * r) +
                               cos(lat1 * r) * cos(lat2 * r) * cos((lon2 - lon1) * r))))
  }

  set.seed(57)
  n <- 200
  lon <- 360 * runif(n) - 180
  lat <- asin(2 * runif(n) - 1) * 180 / pi

  for (ap in c("4/7", "7/4", "3/7")) {
    for (res in c(3, 4)) {
      g <- hex_grid(resolution = res, aperture = ap)
      cells <- lonlat_to_cell(lon, lat, g)
      ll <- cell_to_lonlat(cells, g)

      expect_true(all(cells >= 1 & cells <= hexify:::aperture_n_cells(ap, res)),
                  info = sprintf("%s res %d: cell IDs in range", ap, res))

      # Quantizing a cell centre returns that cell
      expect_equal(lonlat_to_cell(ll$lon_deg, ll$lat_deg, g), cells,
                   info = sprintf("%s res %d: centres quantize to their own cell", ap, res))

      # Cell centres sit within a cell spacing of the point
      spacing_km <- sqrt(2 * (EARTH_SURFACE_KM2 / hexify:::aperture_n_cells(ap, res)) / sqrt(3))
      d <- gc_km(lon, lat, ll$lon_deg, ll$lat_deg)
      expect_lt(median(d), spacing_km)
    }
  }
})

test_that("mixed sequence hierarchy navigates for aperture-7 spellings", {
  skip_on_cran()  # Walks children of sampled cells
  setup_icosa()

  for (ap in c("4/7", "7/4")) {
    grid <- hex_grid(resolution = 3, aperture = ap)
    cgrid <- hex_grid(resolution = 4, aperture = ap)
    set.seed(58)
    cells <- unique(lonlat_to_cell(runif(6, -180, 180), runif(6, -80, 80), grid))

    kids <- get_children(cells, grid)
    for (i in seq_along(cells)) {
      expect_true(length(kids[[i]]) >= 1,
                  info = sprintf("%s: cell %.0f has children", ap, cells[i]))
      expect_true(all(get_parent(kids[[i]], cgrid) == cells[i]),
                  info = sprintf("%s: every child maps back to its parent", ap))
    }
  }
})

# tests/testthat/test-quantizer-nearest.R
# Quantization returns the cell whose centre is nearest the point.
#
# The candidate cells are collected by quantizing a scan of offsets covering
# one cell spacing around the point: the nearest cell's centre lies inside that
# disc, and a point at a centre quantizes to its own cell, so every cell that
# could beat the returned one is in the set. This holds for every grid form,
# rotated or not, and needs no separate model of which (i, j) are grid cells.

scan_offsets <- function(spacing, n = 9) {
  s <- seq(-1.2 * spacing, 1.2 * spacing, length.out = n)
  expand.grid(dx = s, dy = s)
}

nearest_cell_holds <- function(quantize, center, px, py, spacing) {
  cell <- quantize(px, py)
  c0 <- center(cell[["i"]], cell[["j"]])
  best <- (c0[["cx"]] - px)^2 + (c0[["cy"]] - py)^2

  off <- scan_offsets(spacing)
  for (k in seq_len(nrow(off))) {
    q <- quantize(px + off$dx[k], py + off$dy[k])
    cc <- center(q[["i"]], q[["j"]])
    d <- (cc[["cx"]] - px)^2 + (cc[["cy"]] - py)^2
    if (d < best - 1e-12) return(FALSE)
  }
  TRUE
}

# Cell spacing of a sequence: one over the product of sqrt(aperture) per step
seq_spacing <- function(ap_seq) 1 / prod(sqrt(ap_seq[-1]))

test_that("pure aperture quantization returns the nearest cell", {
  skip_on_cran()
  cpp_build_icosa()

  set.seed(101)
  pts <- cbind(runif(12, -0.7, 0.7), runif(12, -0.7, 0.7))

  quantizers <- list(
    "3" = list(q = cpp_hex_quantize_ap3, c = cpp_hex_center_ap3),
    "4" = list(q = cpp_hex_quantize_ap4, c = cpp_hex_center_ap4),
    "7" = list(q = cpp_hex_quantize_ap7, c = cpp_hex_center_ap7)
  )

  for (ap in names(quantizers)) {
    fns <- quantizers[[ap]]
    for (res in 0:3) {
      spacing <- 1 / (sqrt(as.numeric(ap))^res)
      for (k in seq_len(nrow(pts))) {
        ok <- nearest_cell_holds(
          function(x, y) fns$q(x, y, res),
          function(i, j) fns$c(i, j, res),
          pts[k, 1], pts[k, 2], spacing
        )
        expect_true(ok, info = sprintf("aperture %s res %d, point (%.3f, %.3f)",
                                       ap, res, pts[k, 1], pts[k, 2]))
      }
    }
  }
})

test_that("mixed sequence quantization returns the nearest cell", {
  skip_on_cran()
  cpp_build_icosa()

  set.seed(202)
  pts <- cbind(runif(8, -0.7, 0.7), runif(8, -0.7, 0.7))

  sequences <- list(
    c(4, 4, 4, 3, 3),
    c(4, 4, 3, 7),
    c(3, 7, 4, 3),
    c(7, 7, 7, 3),
    c(3, 3, 4, 4, 7)
  )

  for (ap_seq in sequences) {
    spacing <- seq_spacing(ap_seq)
    for (k in seq_len(nrow(pts))) {
      ok <- nearest_cell_holds(
        function(x, y) cpp_hex_quantize_mixed(x, y, ap_seq),
        function(i, j) cpp_hex_center_mixed(i, j, ap_seq),
        pts[k, 1], pts[k, 2], spacing
      )
      expect_true(ok, info = sprintf("sequence %s, point (%.3f, %.3f)",
                                     paste(ap_seq, collapse = "/"),
                                     pts[k, 1], pts[k, 2]))
    }
  }
})

# R/hex_distance.R
# Grid distance between hex cells

#' Grid Distance Between Cells
#'
#' Computes the grid distance (minimum number of hops) between pairs
#' of hexagonal cells. This is a discrete distance measured in cell steps,
#' not a geodesic distance.
#'
#' @param cell_a,cell_b Cell IDs. Must be the same length (pairs) or one
#'   of them length 1 (broadcast). For H3 grids, character vectors.
#'   For ISEA grids, numeric vectors.
#' @param grid A HexGridInfo or HexData object specifying the grid.
#'
#' @return An integer vector of grid distances. `NA` where the distance
#'   cannot be computed (e.g., pentagon path issues in H3).
#'
#' @details
#' **H3 backend:** Uses the vendored H3 `gridDistance` function.
#' Returns the exact shortest path length in cell hops.
#'
#' **ISEA backend:** For cells on the same quad, computes the cube-coordinate
#' distance: `max(|di|, |dj|, |di + dj|)`. Cross-quad distances use BFS
#' expansion and may return `NA` for very distant cells.
#'
#' For geodesic (geographic) distances between cell centers, convert to
#' lon/lat with [cell_to_lonlat()] and use `sf::st_distance()`.
#'
#' @seealso [get_neighbors()] for finding cells within a given distance,
#'   [cell_to_lonlat()] for geographic coordinates
#'
#' @export
#' @examples
#' \donttest{
#' # H3 grid distance
#' g <- hex_grid(resolution = 5, type = "h3")
#' a <- lonlat_to_cell(10, 50, g)
#' b <- lonlat_to_cell(10.1, 50.1, g)
#' hex_distance(a, b, g)
#' }
hex_distance <- function(cell_a, cell_b, grid) {
  g <- extract_grid(grid)

  # Recycle to equal length
  n <- max(length(cell_a), length(cell_b))
  cell_a <- rep_len(cell_a, n)
  cell_b <- rep_len(cell_b, n)

  if (is_h3_grid(g)) {
    return(cpp_h3_gridDistance(as.character(cell_a), as.character(cell_b)))
  }

  # ISEA backend: same-quad cube distance
  ap <- as.integer(g@aperture)
  res <- g@resolution

  # Get quad IJ for both cells
  info_a <- hexify_cell_to_quad_ij(cell_a, res, ap)
  info_b <- hexify_cell_to_quad_ij(cell_b, res, ap)

  result <- integer(n)

  for (idx in seq_len(n)) {
    qa <- info_a$quad[idx]
    qb <- info_b$quad[idx]

    if (is.na(qa) || is.na(qb)) {
      result[idx] <- NA_integer_
      next
    }

    if (qa == qb) {
      # Same quad: cube distance
      di <- info_a$i[idx] - info_b$i[idx]
      dj <- info_a$j[idx] - info_b$j[idx]
      result[idx] <- as.integer(max(abs(di), abs(dj), abs(di + dj)))
    } else {
      # Cross-quad: BFS from cell_a, searching for cell_b
      # Limited search to avoid excessive computation
      max_search <- 100L
      found <- FALSE
      visited <- cell_a[idx]
      current <- cell_a[idx]
      target <- cell_b[idx]

      for (dist in seq_len(max_search)) {
        nbrs <- cpp_get_neighbors_isea(current, res, ap)
        new_cells <- unique(unlist(nbrs))
        new_cells <- setdiff(new_cells, visited)

        if (target %in% new_cells) {
          result[idx] <- dist
          found <- TRUE
          break
        }

        if (length(new_cells) == 0) break
        visited <- c(visited, new_cells)
        current <- new_cells
      }

      if (!found) {
        result[idx] <- NA_integer_
      }
    }
  }

  result
}

# plot_methods.R
# Plot methods for HexData objects
#
# Provides default plotting with sensible styling options.
# Users can visualize hexified data without specifying grid parameters.

# =============================================================================
# THEME HELPER
# =============================================================================

#' Clean minimal theme for pkgdown-friendly plots
#'
#' Wraps \code{ggplot2::theme_minimal()} with grid lines removed so that
#' plots render cleanly on pkgdown sites regardless of background colour.
#' The base font size follows the \code{hexify.base_size} option. A document
#' whose figures are scaled to the reader's column width sets it, so the text
#' arrives at a legible size once the figure has been scaled.
#'
#' @return A ggplot2 theme object.
#' @noRd
.theme_clean <- function() {
  ggplot2::theme_minimal(base_size = getOption("hexify.base_size", 11)) +
    ggplot2::theme(panel.grid = ggplot2::element_blank())
}

# =============================================================================
# POINT SIZE AND JITTERING HELPERS
# =============================================================================

#' Calculate point size based on hexagon size
#'
#' Calculates the cex value so that a single point covers a target
#' fraction of a hexagon cell. Points may overlap when multiple
#' points are in a cell.
#'
#' @param size Size specification: numeric (direct cex), or preset name
#' @param hex_sf sf object with cell polygons
#' @param xlim,ylim Plot limits (for coordinate-to-inches conversion)
#' @return Numeric cex value
#' @noRd
resolve_point_size <- function(size, hex_sf, xlim, ylim) {
 # Coverage targets: fraction of hex area covered by ONE point
 coverage_targets <- c(
   "tiny" = 0.02,
   "small" = 0.05,
   "normal" = 0.10,
   "large" = 0.20,
   "very large" = 0.35
 )

 # If numeric, return directly
 if (is.numeric(size)) {
   return(size)
 }

 # Resolve preset name
 size_lower <- tolower(size)
 if (size_lower == "verylarge") size_lower <- "very large"
 if (size_lower == "auto") size_lower <- "normal"

 if (!size_lower %in% names(coverage_targets)) {
   warning("Unknown point_size '", size, "', using 'normal'")
   size_lower <- "normal"
 }

 target_coverage <- coverage_targets[[size_lower]]

 # Get a representative hexagon (first one)
 hex_poly <- hex_sf[1, ]

 # Get hexagon bounding box in coordinate units
 hex_bbox <- sf::st_bbox(hex_poly)
 hex_width <- hex_bbox["xmax"] - hex_bbox["xmin"]
 hex_height <- hex_bbox["ymax"] - hex_bbox["ymin"]

 # Approximate hexagon area (regular hexagon ~0.866 of bounding box)
 hex_area_approx <- hex_width * hex_height * 0.866

 # Target area for one point
 point_area <- hex_area_approx * target_coverage

 # Point radius in coordinate units
 point_radius_coord <- sqrt(point_area / pi)

 # Convert to plot units
 plot_width <- xlim[2] - xlim[1]
 plot_height <- ylim[2] - ylim[1]

 # Get device dimensions in inches
 dev_size <- grDevices::dev.size("in")
 if (any(dev_size == 0)) dev_size <- c(7, 5)

 # cex=1 gives approximately 0.2 inches diameter
 base_point_diameter_in <- 0.2

 # Convert coordinate radius to inches
 x_scale <- dev_size[1] / plot_width
 y_scale <- dev_size[2] / plot_height
 avg_scale <- (x_scale + y_scale) / 2

 point_diameter_in <- 2 * point_radius_coord * avg_scale

 # Calculate cex
 cex <- point_diameter_in / base_point_diameter_in

 # Clamp to reasonable range
 max(0.1, min(5, cex))
}

#' Jitter points within their assigned hexagon cells
#' @param cell_ids Vector of cell IDs (one per point)
#' @param hex_sf sf object with cell polygons (must have cell_id column)
#' @param jitter If TRUE (default), randomly scatter points within their
#'   cell. If FALSE, place every point at its cell's centroid.
#' @return Data frame with lon, lat columns for jittered positions
#' @noRd
jitter_points_in_cells <- function(cell_ids, hex_sf, jitter = TRUE) {
  n_points <- length(cell_ids)

  # Pre-allocate result
  result <- data.frame(
    lon = numeric(n_points),
    lat = numeric(n_points)
  )

  # Disable S2 for sampling (simpler planar sampling in small hexagons)
  s2_state <- sf::sf_use_s2()
  suppressMessages(sf::sf_use_s2(FALSE))
  on.exit(invisible(suppressMessages(sf::sf_use_s2(s2_state))), add = TRUE)

  # Group points by cell
  cell_groups <- split(seq_along(cell_ids), cell_ids)

  # Determine if cell_ids are character (H3) or numeric (ISEA)
  h3_mode <- is.character(hex_sf$cell_id)

  for (cell_id_str in names(cell_groups)) {
    cell_id_val <- if (h3_mode) cell_id_str else as.numeric(cell_id_str)
    indices <- cell_groups[[cell_id_str]]
    n_in_cell <- length(indices)

    # Get the polygon for this cell
    poly_idx <- which(hex_sf$cell_id == cell_id_val)
    if (length(poly_idx) == 0) next

    poly <- hex_sf[poly_idx, ]

    # Sample random points inside the polygon (or none, if jitter = FALSE,
    # so every point in the cell falls back to the centroid below)
    sampled <- if (jitter) {
      suppressMessages(suppressWarnings(
        sf::st_sample(poly, size = n_in_cell, type = "random")
      ))
    } else {
      sf::st_sfc()
    }

    # If sampling failed or returned wrong count, use centroid
    if (length(sampled) == 0) {
      centroid <- suppressMessages(suppressWarnings(
        sf::st_centroid(sf::st_geometry(poly))
      ))
      coords <- sf::st_coordinates(centroid)
      result$lon[indices] <- coords[1, 1]
      result$lat[indices] <- coords[1, 2]
    } else {
      coords <- sf::st_coordinates(sampled)
      # Handle case where we got fewer points than expected
      if (nrow(coords) < n_in_cell) {
        # Pad with centroid
        centroid <- suppressMessages(suppressWarnings(
          sf::st_centroid(sf::st_geometry(poly))
        ))
        centroid_coords <- sf::st_coordinates(centroid)
        n_missing <- n_in_cell - nrow(coords)
        coords <- rbind(coords, matrix(rep(centroid_coords, n_missing),
                                        ncol = 2, byrow = TRUE))
      }
      result$lon[indices] <- coords[1:n_in_cell, 1]
      result$lat[indices] <- coords[1:n_in_cell, 2]
    }
  }

  result
}

# =============================================================================
# BASE R PLOT METHOD
# =============================================================================

#' Plot HexData objects
#'
#' Default plot method for HexData objects. Draws hexagonal grid cells
#' with an optional basemap.
#'
#' @param x A HexData object from \code{hexify()}
#' @param y Ignored (for S4 method compatibility)
#' @param basemap Basemap specification:
#'   \itemize{
#'     \item \code{TRUE} or \code{"world"}: Use built-in world map
#'     \item \code{FALSE} or \code{NULL}: No basemap
#'     \item sf object: Custom basemap
#'   }
#' @param clip_basemap Clip basemap to data extent (default TRUE). Clipping
#'   temporarily disables S2 spherical geometry to avoid edge-crossing errors.
#' @param basemap_fill Fill color for basemap (default "gray90")
#' @param basemap_border Border color for basemap (default "gray50")
#' @param basemap_lwd Line width for basemap borders (default 0.5)
#' @param grid_fill Fill color for grid cells (default "#E69F00" - amber/orange)
#' @param grid_border Border color for grid cells (default "#5D4E37" - dark brown)
#' @param grid_lwd Line width for cell borders (default 0.8)
#' @param grid_alpha Transparency for cell fill (0-1, default 0.7)
#' @param fill Column name for fill mapping (optional)
#' @param show_points Show original points on top of cells (default FALSE).
#'   Points are jittered within their assigned hexagon.
#' @param point_size Size of points. Can be:
#'   \itemize{
#'     \item A number (direct cex value)
#'     \item A preset defining what fraction of a hex cell one point covers:
#'       "tiny" (~2\%), "small" (~5\%), "normal"/"auto" (~10\%),
#'       "large" (~20\%), "very large" (~35\%)
#'   }
#' @param point_color Color of points (default "red")
#' @param point_alpha Transparency for points (0-1, default 1)
#' @param jitter If TRUE (default), points are randomly scattered within
#'   their assigned hexagon. If FALSE, all points in a cell are plotted at
#'   the cell centroid instead.
#' @param crop Crop to data extent (default TRUE)
#' @param crop_expand Expansion factor for crop (default 0.1)
#' @param main Plot title
#' @param ... Additional arguments passed to base plot()
#'
#' @return Invisibly returns the HexData object
#'
#' @details
#' This function generates polygon geometries for the cells present in
#' the data and plots them. Polygons are computed on demand, not stored,
#' to minimize memory usage.
#'
#' @seealso \code{\link{hexify_heatmap}} for ggplot2 plotting
#'
#' @export
#' @examples
#' \donttest{
#' df <- data.frame(lon = runif(50, -5, 5), lat = runif(50, 45, 50))
#' result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 2000)
#'
#' # Basic plot
#' plot(result, basemap = FALSE)
#'
#' # With basemap and custom styling
#' plot(result, grid_fill = "lightblue", grid_border = "darkblue")
#' }
setMethod("plot", signature(x = "HexData", y = "missing"),
  function(x, y,
           basemap = TRUE,
           clip_basemap = TRUE,
           basemap_fill = "gray90",
           basemap_border = "gray50",
           basemap_lwd = 0.5,
           grid_fill = "#E69F00",
           grid_border = "#5D4E37",
           grid_lwd = 0.8,
           grid_alpha = 0.7,
           fill = NULL,
           show_points = FALSE,
           point_size = "auto",
           point_color = "red",
           point_alpha = 1,
           jitter = TRUE,
           crop = TRUE,
           crop_expand = 0.1,
           main = NULL,
           ...) {

    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("Package 'sf' is required for plotting")
    }

    # Get grid spec
    g <- x@grid

    # Generate cell polygons
    unique_cells <- unique(x@cell_id)
    hex_sf <- cell_to_sf(unique_cells, g)

    # If fill column specified, merge data
    if (!is.null(fill)) {
      if (!fill %in% names(x@data)) {
        stop(sprintf("Column '%s' not found in data", fill))
      }
      # Create data frame with cell_id for merging
      data_with_id <- cbind(x@data, cell_id = x@cell_id)
      # Aggregate by cell if needed (take first value)
      agg_data <- data_with_id[!duplicated(data_with_id$cell_id), c("cell_id", fill)]
      hex_sf <- merge(hex_sf, agg_data, by = "cell_id", all.x = TRUE)
    }

    # Calculate bounding box
    hex_bbox <- sf::st_bbox(hex_sf)

    if (crop) {
      x_range <- hex_bbox["xmax"] - hex_bbox["xmin"]
      y_range <- hex_bbox["ymax"] - hex_bbox["ymin"]
      xlim <- c(hex_bbox["xmin"] - x_range * crop_expand,
                hex_bbox["xmax"] + x_range * crop_expand)
      ylim <- c(hex_bbox["ymin"] - y_range * crop_expand,
                hex_bbox["ymax"] + y_range * crop_expand)
    } else {
      xlim <- c(-180, 180)
      ylim <- c(-90, 90)
    }

    # Calculate aspect ratio
    mean_lat <- mean(ylim)
    asp <- 1 / cos(mean_lat * pi / 180)

    # Resolve basemap
    basemap_sf <- NULL
    if (isTRUE(basemap) || identical(basemap, "world")) {
      basemap_sf <- hexify_world
    } else if (inherits(basemap, c("sf", "sfc"))) {
      basemap_sf <- basemap
    }

    # Initialize plot
    plot(xlim, ylim, type = "n",
         xlim = xlim, ylim = ylim,
         xlab = "Longitude", ylab = "Latitude",
         asp = asp, main = main, ...)

    # Draw basemap
    if (!is.null(basemap_sf)) {
      if (clip_basemap) {
        # Disable S2 for clipping to avoid spherical geometry edge-crossing errors
        s2_state <- sf::sf_use_s2()
        suppressMessages(sf::sf_use_s2(FALSE))
        on.exit(sf::sf_use_s2(s2_state), add = TRUE)

        # Clip basemap to extent
        clip_box <- sf::st_bbox(c(xmin = unname(xlim[1]),
                                   xmax = unname(xlim[2]),
                                   ymin = unname(ylim[1]),
                                   ymax = unname(ylim[2])),
                                 crs = 4326)
        clip_poly <- sf::st_as_sfc(clip_box)

        basemap_clipped <- suppressMessages(suppressWarnings(
          sf::st_intersection(sf::st_make_valid(basemap_sf), clip_poly)
        ))

        if (nrow(basemap_clipped) > 0) {
          plot(sf::st_geometry(basemap_clipped),
               col = basemap_fill, border = basemap_border,
               lwd = basemap_lwd, add = TRUE)
        }
      } else {
        # No clipping - draw full basemap
        plot(sf::st_geometry(basemap_sf),
             col = basemap_fill, border = basemap_border,
             lwd = basemap_lwd, add = TRUE)
      }
    }

    # Draw hexagons
    if (is.null(fill)) {
      fill_colors <- adjustcolor(grid_fill, alpha.f = grid_alpha)
    } else {
      # Map fill column to colors
      values <- hex_sf[[fill]]
      if (is.numeric(values)) {
        # Continuous: use viridis-like palette
        n_colors <- 100
        pal <- grDevices::colorRampPalette(
          c("#440154", "#3B528B", "#21918C", "#5DC863", "#FDE725")
        )(n_colors)
        scaled <- (values - min(values, na.rm = TRUE)) /
                  (max(values, na.rm = TRUE) - min(values, na.rm = TRUE))
        scaled[is.na(scaled)] <- 0
        idx <- pmax(1, pmin(n_colors, ceiling(scaled * n_colors)))
        fill_colors <- adjustcolor(pal[idx], alpha.f = grid_alpha)
      } else {
        # Discrete: use basic palette
        levels <- unique(values)
        pal <- grDevices::rainbow(length(levels))
        fill_colors <- adjustcolor(
          pal[match(values, levels)],
          alpha.f = grid_alpha
        )
      }
    }

    plot(sf::st_geometry(hex_sf),
         col = fill_colors, border = grid_border,
         lwd = grid_lwd, add = TRUE)

    # Draw points if requested
    if (show_points) {
      # Resolve point size based on hex cell size
      cex <- resolve_point_size(point_size, hex_sf, xlim, ylim)

      # Jitter points within their hexagon (or centroid-snap if jitter = FALSE)
      jittered <- jitter_points_in_cells(x@cell_id, hex_sf, jitter = jitter)

      points(jittered$lon, jittered$lat,
             pch = 19, cex = cex,
             col = adjustcolor(point_color, alpha.f = point_alpha))
    }

    invisible(x)
  }
)

# Helper for NULL coalescing
`%||%` <- function(a, b) if (is.null(a)) b else a

# =============================================================================
# PLOT GRID CLIPPED TO BOUNDARY
# =============================================================================

#' Plot hexagonal grid clipped to a polygon boundary
#'
#' A convenience function that creates a grid, clips it to a boundary polygon,
#' and plots the result in a single call.
#'
#' @param boundary An sf/sfc polygon to clip to (e.g., country boundary)
#' @param grid A HexGridInfo object from \code{hex_grid()}
#' @param crop If TRUE (default), cells are cropped to boundary. If FALSE,
#'   only complete hexagons within boundary are shown.
#' @param boundary_fill Fill color for the boundary polygon (default "gray95")
#' @param boundary_border Border color for boundary (default "gray40")
#' @param boundary_lwd Line width for boundary (default 0.5)
#' @param grid_fill Fill color for grid cells (default "steelblue")
#' @param grid_border Border color for grid cells (default "steelblue")
#' @param grid_lwd Line width for cell borders (default 0.3)
#' @param grid_alpha Transparency for cell fill (0-1, default 0.3)
#' @param title Plot title. If NULL (default), auto-generates title with cell area.
#' @param expand Expansion factor for plot limits (default 0.05)
#'
#' @return A ggplot object that can be further customized
#'
#' @details
#' This is a convenience wrapper around \code{grid_clip()} that handles the
#' common use case of visualizing a hexagonal grid over a geographic region.
#'
#' @seealso \code{\link{grid_clip}} for the underlying clipping function,
#'   \code{\link{hex_grid}} for grid specification
#'
#' @export
#' @examples
#' \donttest{
#' # Plot grid over France
#' france <- hexify_world[hexify_world$name == "France", ]
#' grid <- hex_grid(area_km2 = 2000)
#' plot_grid(france, grid)
#'
#' # Customize colors
#' plot_grid(france, grid,
#'           grid_fill = "coral", grid_alpha = 0.5,
#'           boundary_fill = "lightyellow")
#'
#' # Keep only complete hexagons
#' plot_grid(france, grid, crop = FALSE)
#'
#' # Add ggplot2 customizations
#' library(ggplot2)
#' plot_grid(france, grid) +
#'   labs(subtitle = "ISEA3H Discrete Global Grid") +
#'   theme_void()
#' }
plot_grid <- function(boundary,
                      grid,
                      crop = TRUE,
                      boundary_fill = "gray95",
                      boundary_border = "gray40",
                      boundary_lwd = 0.5,
                      grid_fill = "steelblue",
                      grid_border = "steelblue",
                      grid_lwd = 0.3,
                      grid_alpha = 0.3,
                      title = NULL,
                      expand = 0.05) {

  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plot_grid()")
  }
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required for plot_grid()")
  }

  # Extract grid spec
  g <- extract_grid(grid)

  # Clip grid to boundary
  clipped_grid <- grid_clip(boundary, g, crop = crop)

  # Get bounding box for plot limits
  bbox <- sf::st_bbox(boundary)
  x_range <- bbox["xmax"] - bbox["xmin"]
  y_range <- bbox["ymax"] - bbox["ymin"]
  xlim <- c(bbox["xmin"] - x_range * expand, bbox["xmax"] + x_range * expand)
  ylim <- c(bbox["ymin"] - y_range * expand, bbox["ymax"] + y_range * expand)

  # Auto-generate title if not provided
  if (is.null(title)) {
    title <- sprintf("Hexagonal Grid (~%.0f km\u00b2 cells)", g@area_km2)
  }

  # Build plot
  p <- ggplot2::ggplot() +
    ggplot2::geom_sf(
      data = boundary,
      fill = boundary_fill,
      color = boundary_border,
      linewidth = boundary_lwd
    ) +
    ggplot2::geom_sf(
      data = clipped_grid,
      fill = grDevices::adjustcolor(grid_fill, alpha.f = grid_alpha),
      color = grid_border,
      linewidth = grid_lwd
    ) +
    ggplot2::coord_sf(xlim = xlim, ylim = ylim, expand = FALSE) +
    ggplot2::labs(title = title) +
    .theme_clean() +
    ggplot2::theme(
      axis.title = ggplot2::element_blank()
    )

  p
}

# tests/testthat/test-plot-methods.R
# Tests for plot methods

test_that("plot.HexData works with default settings", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  # Should not error
  expect_silent({
    pdf(NULL)  # Null device
    on.exit(dev.off())
    plot(result, basemap = FALSE)
  })
})

test_that("jitter_points_in_cells(jitter = FALSE) snaps points to the cell centroid", {
  skip_if_not_installed("sf")

  # Many points in one cell
  df <- data.frame(lon = rep(2.35, 20), lat = rep(48.86, 20))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 200000)
  hex_sf <- cell_to_sf(unique(result@cell_id), result@grid)

  jittered <- hexify:::jitter_points_in_cells(result@cell_id, hex_sf, jitter = TRUE)
  unjittered <- hexify:::jitter_points_in_cells(result@cell_id, hex_sf, jitter = FALSE)

  # With jitter, 20 random points in the same cell are extremely unlikely
  # to all land on the exact same coordinate
  expect_true(length(unique(jittered$lon)) > 1 || length(unique(jittered$lat)) > 1)

  # Without jitter, every point in the cell collapses to the same centroid
  expect_equal(length(unique(unjittered$lon)), 1)
  expect_equal(length(unique(unjittered$lat)), 1)
})

test_that("plot.HexData accepts jitter and point_alpha arguments", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_no_error({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, show_points = TRUE, jitter = FALSE,
         point_alpha = 0.5, point_color = "darkblue")
  })
})

test_that("plot.HexData works with basemap", {
  skip_on_cran()  # Slow sf operations
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  # May produce sf messages about S2, so just check it runs
  expect_no_error({
    pdf(NULL)
    on.exit(dev.off())
    suppressMessages(plot(result, basemap = TRUE))
  })
})

test_that("plot.HexData works with custom sf basemap", {
  skip_on_cran()  # Slow sf operations
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  # Use subset of world as custom basemap
  custom_basemap <- hexify_world[hexify_world$continent == "Europe", ]

  # May produce sf messages about S2, so just check it runs
  expect_no_error({
    pdf(NULL)
    on.exit(dev.off())
    suppressMessages(plot(result, basemap = custom_basemap))
  })
})

test_that("plot.HexData works with show_points", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 2.36, 2.37),
    lat = c(48.86, 48.87, 48.88)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, show_points = TRUE)
  })
})

test_that("plot.HexData works with point_size presets", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 2.36),
    lat = c(48.86, 48.87)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  for (size in c("tiny", "small", "normal", "large", "auto")) {
    expect_silent({
      pdf(NULL)
      on.exit(dev.off(), add = TRUE)
      plot(result, basemap = FALSE, show_points = TRUE, point_size = size)
    })
  }
})

test_that("plot.HexData works with numeric point_size", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 2.36),
    lat = c(48.86, 48.87)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, show_points = TRUE, point_size = 0.5)
  })
})

test_that("plot.HexData works with fill mapping", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    value = c(10, 20, 30)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, fill = "value")
  })
})

test_that("plot.HexData works with discrete fill", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90, -3.70),
    lat = c(48.86, 52.37, 40.42),
    category = c("A", "B", "A")
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, fill = "category")
  })
})

test_that("plot.HexData errors on invalid fill column", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_error({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, fill = "nonexistent")
  }, "not found")
})

test_that("plot.HexData works with crop = FALSE", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, crop = FALSE)
  })
})

test_that("plot.HexData works with clip_basemap = FALSE", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = TRUE, clip_basemap = FALSE)
  })
})

test_that("plot.HexData works with custom colors", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result,
         basemap = FALSE,
         grid_fill = "lightblue",
         grid_border = "darkblue",
         grid_alpha = 0.5,
         grid_lwd = 1.5)
  })
})

test_that("plot.HexData returns HexData invisibly", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 4.90),
    lat = c(48.86, 52.37)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  pdf(NULL)
  on.exit(dev.off())
  returned <- plot(result, basemap = FALSE)
  expect_s4_class(returned, "HexData")
})

test_that("plot.HexData works with title", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(2.35), lat = c(48.86))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, main = "Test Title")
  })
})

test_that("plot.HexData warns on unknown point_size", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(2.35), lat = c(48.86))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  expect_warning({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, show_points = TRUE, point_size = "unknown")
  }, "Unknown point_size")
})

test_that("plot_grid creates ggplot", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  grid <- hex_grid(area_km2 = 50000)
  france <- hexify_world[hexify_world$name == "France", ]

  p <- suppressMessages(plot_grid(france, grid))
  expect_s3_class(p, "ggplot")
})

test_that("plot_grid works with custom colors", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  grid <- hex_grid(area_km2 = 50000)
  france <- hexify_world[hexify_world$name == "France", ]

  p <- suppressMessages(plot_grid(france, grid,
                                   grid_fill = "coral",
                                   boundary_fill = "lightyellow"))
  expect_s3_class(p, "ggplot")
})

test_that("plot_grid works with crop = FALSE", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  grid <- hex_grid(area_km2 = 50000)
  france <- hexify_world[hexify_world$name == "France", ]

  p <- suppressMessages(plot_grid(france, grid, crop = FALSE))
  expect_s3_class(p, "ggplot")
})

test_that("plot_grid auto-generates title", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  grid <- hex_grid(area_km2 = 50000)
  france <- hexify_world[hexify_world$name == "France", ]

  p <- suppressMessages(plot_grid(france, grid, title = NULL))
  expect_s3_class(p, "ggplot")
})

test_that("plot_grid accepts custom title", {
  skip_if_not_installed("sf")
  skip_if_not_installed("ggplot2")

  grid <- hex_grid(area_km2 = 50000)
  france <- hexify_world[hexify_world$name == "France", ]

  p <- suppressMessages(plot_grid(france, grid, title = "Custom Title"))
  expect_s3_class(p, "ggplot")
})

test_that("plot.HexData works with 'very large' point_size", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 2.36),
    lat = c(48.86, 48.87)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, show_points = TRUE, point_size = "very large")
  })
})

test_that("plot.HexData works with 'verylarge' point_size variant", {
  skip_if_not_installed("sf")

  df <- data.frame(
    lon = c(2.35, 2.36),
    lat = c(48.86, 48.87)
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  expect_silent({
    pdf(NULL)
    on.exit(dev.off())
    plot(result, basemap = FALSE, show_points = TRUE, point_size = "verylarge")
  })
})

# R/hex_compact.R
# Multi-resolution compaction and uncompaction

#' Compact Hex Cells
#'
#' Merges child cells into their parent when all children are present.
#' This is a lossless compression — no spatial information is lost.
#' The compact representation uses fewer cells to cover the same area.
#'
#' @param cell_ids Cell IDs to compact. For H3 grids, a character vector.
#'   For ISEA grids, a character vector of hierarchical index strings.
#' @param grid A HexGridInfo object specifying the grid.
#'
#' @return A character vector of compacted cell IDs. Cells that could be
#'   merged into parents appear as parent IDs at coarser resolution.
#'
#' @details
#' **H3 backend:** Uses the vendored H3 `compactCells` function.
#'
#' **ISEA backend (aperture 7, Z7 index):** Groups cells by parent index
#' (dropping the last digit). If all 7 children are present, replaces them
#' with the parent. Iterates until no further compaction is possible.
#'
#' @seealso [hex_uncompact()] for the inverse operation,
#'   [get_parent()], [get_children()] for hierarchical operations
#'
#' @export
#' @examples
#' \donttest{
#' # H3 compaction
#' g <- hex_grid(resolution = 3, type = "h3")
#' parent <- "832830fffffffff"
#' children <- get_children(parent, g)[[1]]
#' compact <- hex_compact(children, g)
#' compact  # Should return the parent
#' }
hex_compact <- function(cell_ids, grid) {
  g <- extract_grid(grid)

  if (is_h3_grid(g)) {
    return(cpp_h3_compactCells(as.character(cell_ids)))
  }

  # ISEA Z7 compaction (pure R)
  ap <- as.integer(g@aperture)
  if (ap != 7) {
    stop("hex_compact() for ISEA currently only supports aperture 7 (Z7 index)")
  }

  ids <- as.character(cell_ids)
  changed <- TRUE

  while (changed) {
    changed <- FALSE
    # Group by parent (drop last digit)
    parents <- substr(ids, 1, nchar(ids) - 1L)
    # Only cells with length > 2 can be compacted (resolution > 0)
    compactable <- nchar(ids) > 2L

    if (!any(compactable)) break

    # Parents with all 7 distinct children (0-6) present. Group by distinct
    # last digit rather than raw row count, so duplicate cell IDs can't be
    # mistaken for a full sibling set.
    compactable_ids <- ids[compactable]
    compactable_parents <- parents[compactable]
    last_digit <- substr(compactable_ids, nchar(compactable_ids), nchar(compactable_ids))
    digit_sets <- split(last_digit, compactable_parents)
    full_parents <- names(digit_sets)[vapply(digit_sets, function(d) length(unique(d)) == 7L, logical(1))]

    if (length(full_parents) > 0) {
      changed <- TRUE
      # Remove children of full parents, add parent. If the input already
      # contained the parent id itself alongside all 7 children, that parent
      # entry's own `parents` value (its grandparent) never matches
      # `full_parents`, so it would survive untouched while the
      # newly-synthesized parent is also appended -- drop it explicitly to
      # avoid emitting the same parent id twice.
      is_child_of_full <- compactable & parents %in% full_parents
      is_full_parent_itself <- ids %in% full_parents
      remaining <- ids[!is_child_of_full & !is_full_parent_itself]
      ids <- c(remaining, full_parents)
    }
  }

  # Mixed-resolution input can also leave a coarser ancestor and one of its
  # own (not-fully-compactable) descendants both present in the result --
  # the ancestor's area already covers the descendant's, so the descendant
  # is redundant even though it's not a literal duplicate ID.
  if (length(ids) > 1L) {
    is_redundant <- vapply(seq_along(ids), function(k) {
      any(nchar(ids) < nchar(ids[k]) & substr(ids[k], 1, nchar(ids)) == ids)
    }, logical(1))
    ids <- ids[!is_redundant]
  }

  ids
}

#' Uncompact Hex Cells
#'
#' Expands compacted cells to a target resolution. All cells in the output
#' share the same resolution.
#'
#' @param cell_ids Character vector of (possibly mixed-resolution) cell IDs.
#' @param grid A HexGridInfo object specifying the grid.
#' @param target_resolution Integer. The resolution to expand all cells to.
#'
#' @return A character vector of cell IDs, all at `target_resolution`.
#'
#' @details
#' **H3 backend:** Uses the vendored H3 `uncompactCells` function.
#'
#' **ISEA backend (aperture 7, Z7 index):** Appends digits 0-6 to expand
#' each cell to its 7 children, repeating until the target resolution is
#' reached.
#'
#' @seealso [hex_compact()] for the inverse operation
#'
#' @export
#' @examples
#' \donttest{
#' g <- hex_grid(resolution = 3, type = "h3")
#' parent <- "832830fffffffff"
#' hex_uncompact(parent, g, target_resolution = 4L)
#' }
hex_uncompact <- function(cell_ids, grid, target_resolution) {
  g <- extract_grid(grid)
  target_resolution <- as.integer(target_resolution)

  if (is_h3_grid(g)) {
    return(cpp_h3_uncompactCells(as.character(cell_ids), target_resolution))
  }

  # ISEA Z7 uncompaction (pure R)
  ap <- as.integer(g@aperture)
  if (ap != 7) {
    stop("hex_uncompact() for ISEA currently only supports aperture 7 (Z7 index)")
  }

  ids <- as.character(cell_ids)

  initial_res <- nchar(ids) - 2L  # Z7: first 2 chars are quad
  if (any(initial_res > target_resolution)) {
    stop(sprintf(
      "target_resolution (%d) is coarser than some input cells (max resolution %d); hex_uncompact() cannot expand to a coarser resolution",
      target_resolution, max(initial_res)
    ))
  }

  repeat {
    # Check which cells need expansion
    current_res <- nchar(ids) - 2L  # Z7: first 2 chars are quad
    needs_expansion <- current_res < target_resolution

    if (!any(needs_expansion)) break

    # Expand cells that need it
    expanded <- unlist(lapply(ids[needs_expansion], function(id) {
      paste0(id, 0:6)
    }))

    ids <- c(ids[!needs_expansion], expanded)
  }

  ids
}

# R/hex_extract.R
# Raster extraction at hex cell centers

#' Extract Raster Values at Hex Cell Centers
#'
#' Samples raster values at hexagonal cell centers. Faster than
#' [hex_zonal()] because it only queries cell center points, not full
#' polygons.
#'
#' @param raster A `terra::SpatRaster` object.
#' @param grid A HexGridInfo or HexData object specifying the grid.
#' @param cells Optional cell IDs to extract. If `NULL` (default), extracts
#'   at all cell centers from `grid` (only works for HexData objects or
#'   if a `boundary` is provided).
#' @param boundary Optional sf polygon to limit extraction extent.
#'
#' @return A data.frame with columns `cell_id`, plus one column per raster
#'   layer.
#'
#' @details
#' Requires the `terra` package (in Suggests). The function:
#' 1. Generates cell centers for the raster extent
#' 2. Calls `terra::extract(raster, cell_center_matrix)`
#' 3. Attaches cell IDs
#'
#' For full zonal statistics (aggregating all pixels within each hex polygon),
#' use [hex_zonal()] instead.
#'
#' @seealso [hex_zonal()] for polygon-based zonal statistics,
#'   [hexify()] for creating HexData objects
#'
#' @export
#' @examples
#' \donttest{
#' if (requireNamespace("terra", quietly = TRUE)) {
#'   # Create a small synthetic raster
#'   r <- terra::rast(nrows = 10, ncols = 10,
#'                    xmin = -10, xmax = 10, ymin = 40, ymax = 55)
#'   terra::values(r) <- runif(100)
#'   names(r) <- "temperature"
#'
#'   # Extract at hex cell centers
#'   g <- hex_grid(area_km2 = 500)
#'   df <- data.frame(lon = c(0, 5), lat = c(45, 50))
#'   hd <- hexify(df, lon = "lon", lat = "lat", grid = g)
#'   hex_extract(r, hd)
#' }
#' }
hex_extract <- function(raster, grid, cells = NULL, boundary = NULL) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    stop("Package 'terra' is required for hex_extract()")
  }

  g <- extract_grid(grid)

  # Get cell centers. `cells`/`boundary`, when supplied, take precedence over
  # a HexData grid's own already-assigned cells (matching hex_zonal()) --
  # otherwise they'd be silently ignored whenever `grid` is a HexData object.
  if (!is.null(cells)) {
    cell_ids <- cells
    ll <- cell_to_lonlat(cells, g)
    centers <- cbind(lon = ll$lon_deg, lat = ll$lat_deg)
  } else if (!is.null(boundary)) {
    # Generate cells within boundary. grid_clip() returns polygon geometries
    # (a cell_id column, no lon/lat), so cell centers are recomputed the same
    # way as the 'cells' branch above.
    cell_data <- grid_clip(boundary, g)
    cell_ids <- cell_data$cell_id
    ll <- cell_to_lonlat(cell_ids, g)
    centers <- cbind(lon = ll$lon_deg, lat = ll$lat_deg)
  } else if (is_hex_data(grid)) {
    cell_ids <- grid@cell_id
    centers <- grid@cell_center
  } else {
    stop("Provide a HexData object, cell IDs via 'cells', or a 'boundary' polygon")
  }

  # Keep unique cells only
  uid <- !duplicated(cell_ids)
  u_cell_ids <- cell_ids[uid]
  u_centers <- centers[uid, , drop = FALSE]

  # Extract raster values at cell centers
  pts <- terra::vect(u_centers, crs = grid_crs_wkt(g))
  extracted <- terra::extract(raster, pts)

  # Build result
  result <- data.frame(cell_id = u_cell_ids, stringsAsFactors = FALSE)
  # Drop the ID column from terra::extract output
  if ("ID" %in% names(extracted)) {
    extracted$ID <- NULL
  }
  result <- cbind(result, extracted)
  result
}

# R/dggrid_compat.R
# dggridR compatibility layer
#
# Functions for converting between hexify and dggridR grid objects.
# These enable interoperability with dggridR without copying its API.

# =============================================================================
# GRID OBJECT CONVERSION
# =============================================================================

#' Convert hexify grid to 'dggridR'-compatible grid object
#'
#' Creates a 'dggridR'-compatible grid specification from a hexify_grid object.
#' The resulting object can be used with 'dggridR' functions that accept a dggs
#' object.
#'
#' @param grid A hexify_grid object from hexify_grid()
#'
#' @return A list with 'dggridR'-compatible fields:
#'   \item{pole_lon_deg}{Longitude of grid pole (default 11.25)}
#'   \item{pole_lat_deg}{Latitude of grid pole (default 58.282525588538995)}
#'   \item{azimuth_deg}{Grid azimuth rotation (default 0)}
#'   \item{aperture}{Grid aperture (3, 4, or 7)}
#'   \item{res}{Resolution level}
#'   \item{topology}{Grid topology ("HEXAGON")}
#'   \item{projection}{Map projection ('ISEA')}
#'   \item{precision}{Output decimal precision (default 7)}
#'
#' @family 'dggridR' compatibility
#' @export
as_dggrid <- function(grid) {

  # Accept the modern HexGridInfo (S4) object by converting to the legacy
  # list representation this function builds from, matching dgverify()'s
  # fix (#7) for the same gap.
  if (is_hex_grid(grid)) {
    if (grid@grid_type == "h3") {
      stop("as_dggrid() has no dggridR-compatible representation for H3 grids")
    }
    grid <- HexGridInfo_to_hexify_grid(grid)
  }

  if (!inherits(grid, "hexify_grid")) {
    stop("grid must be a hexify_grid object from hexify_grid()")
  }

  dggs <- list(
    pole_lon_deg = ISEA_VERT0_LON_DEG,
    pole_lat_deg = ISEA_VERT0_LAT_DEG,
    azimuth_deg = ISEA_AZIMUTH_DEG,
    aperture = grid$aperture,
    res = grid$resolution,
    topology = "HEXAGON",
    projection = "ISEA",
    precision = 7L
  )

  class(dggs) <- c("dggs", "list")
  dggs
}

#' Convert 'dggridR' grid object to hexify_grid
#'
#' Creates a hexify_grid object from a 'dggridR' dggs object. This allows
#' using hexify functions with grids created by 'dggridR' dgconstruct().
#'
#' @param dggs A 'dggridR' grid object from dgconstruct()
#'
#' @return A hexify_grid object
#'
#' @details
#' Only 'ISEA' projection with HEXAGON topology is fully supported.
#' Other configurations will generate warnings.
#'
#' The function validates that the 'dggridR' grid uses compatible settings:
#' - Projection must be 'ISEA' (FULLER not supported)
#' - Topology must be "HEXAGON" (DIAMOND, TRIANGLE not supported)
#' - Aperture must be 3, 4, or 7
#'
#' @family 'dggridR' compatibility
#' @export
from_dggrid <- function(dggs) {
  # Validate dggridR object

  if (!is.list(dggs)) {
    stop("dggs must be a list (dggridR grid object)")
  }

  required <- c("res", "aperture", "topology", "projection")
  missing <- setdiff(required, names(dggs))
  if (length(missing) > 0) {
    stop(sprintf("dggs missing required fields: %s", paste(missing, collapse = ", ")))
  }

  # Check compatibility

  if (!is.null(dggs$projection) && dggs$projection != "ISEA") {
    warning("Only ISEA projection is supported. Results may differ from dggridR.")
  }

  if (!is.null(dggs$topology) && dggs$topology != "HEXAGON") {
    warning("Only HEXAGON topology is supported. Results may differ from dggridR.")
  }

  if (!dggs$aperture %in% c(3L, 4L, 7L)) {
    stop(sprintf("Aperture %d not supported. Must be 3, 4, or 7.", dggs$aperture))
  }

  # Check for non-default orientation (not supported)
  if (!is.null(dggs$pole_lon_deg) && abs(dggs$pole_lon_deg - ISEA_VERT0_LON_DEG) > 1e-6) {
    warning("Non-default pole_lon_deg not supported. Using standard ISEA orientation.")
  }
  if (!is.null(dggs$pole_lat_deg) && abs(dggs$pole_lat_deg - ISEA_VERT0_LAT_DEG) > 1e-6) {
    warning("Non-default pole_lat_deg not supported. Using standard ISEA orientation.")
  }
  if (!is.null(dggs$azimuth_deg) && abs(dggs$azimuth_deg - ISEA_AZIMUTH_DEG) > 1e-6) {
    warning("Non-default azimuth_deg not supported. Using standard ISEA orientation.")
  }

  # Create hexify_grid. `area` is a throwaway placeholder -- both `resolution`
  # and `area` are overwritten below from `dggs$res` directly, but
  # hexify_grid() requires a valid positive area (see #43) to construct the
  # object at all.
  hexify_grid(
    area = 1000,
    topology = "HEXAGON",
    metric = TRUE,
    resround = "nearest",
    aperture = as.integer(dggs$aperture),
    projection = "ISEA"
  ) -> grid

  # Override resolution to match dggridR exactly
  validate_resolution(dggs$res)
  grid$resolution <- as.integer(dggs$res)
  grid$res <- as.integer(dggs$res)

  # Calculate actual area for this resolution
  n_cells <- max_cell_id(grid$resolution, grid$aperture)
  grid$area <- body_surface_km2(grid_radius_km(grid)) / n_cells

  grid
}

#' Validate 'dggridR' grid compatibility with hexify
#'
#' Checks whether a 'dggridR' grid object is compatible with hexify functions.
#' Returns TRUE if compatible, or throws an error describing incompatibilities.
#'
#' @param dggs A 'dggridR' grid object
#' @param strict If TRUE (default), throw errors for incompatibilities.
#'   If FALSE, return FALSE instead of throwing errors.
#'
#' @return TRUE if compatible, FALSE if not compatible (when strict=FALSE)
#'
#' @family 'dggridR' compatibility
#' @export
dggrid_is_compatible <- function(dggs, strict = TRUE) {
  issues <- character()

  if (!is.list(dggs)) {
    issues <- c(issues, "dggs must be a list")
  } else {
    if (is.null(dggs$projection) || dggs$projection != "ISEA") {
      issues <- c(issues, "Only ISEA projection supported (not FULLER)")
    }

    if (is.null(dggs$topology) || dggs$topology != "HEXAGON") {
      issues <- c(issues, "Only HEXAGON topology supported (not DIAMOND/TRIANGLE)")
    }

    if (is.null(dggs$aperture) || !dggs$aperture %in% c(3L, 4L, 7L)) {
      issues <- c(issues, "Aperture must be 3, 4, or 7")
    }

    # Check orientation
    if (!is.null(dggs$pole_lon_deg) && abs(dggs$pole_lon_deg - ISEA_VERT0_LON_DEG) > 1e-6) {
      issues <- c(issues, "Non-default pole_lon_deg not supported")
    }
    if (!is.null(dggs$pole_lat_deg) && abs(dggs$pole_lat_deg - ISEA_VERT0_LAT_DEG) > 1e-6) {
      issues <- c(issues, "Non-default pole_lat_deg not supported")
    }
    if (!is.null(dggs$azimuth_deg) && abs(dggs$azimuth_deg - ISEA_AZIMUTH_DEG) > 1e-6) {
      issues <- c(issues, "Non-default azimuth_deg not supported")
    }
  }

  if (length(issues) > 0) {
    if (strict) {
      stop(sprintf("dggridR grid not compatible with hexify:\n- %s",
                   paste(issues, collapse = "\n- ")))
    }
    return(FALSE)
  }

  TRUE
}

# tests/testthat/test-dggrid-compat-functions.R
# Tests for dggridR compatibility layer functions

test_that("as_dggrid converts hexify_grid to dggridR format", {
  grid <- hexify_grid(area = 1000, aperture = 3)
  dggs <- as_dggrid(grid)

  expect_type(dggs, "list")
  expect_equal(dggs$aperture, 3)
  expect_equal(dggs$res, grid$resolution)
  expect_equal(dggs$topology, "HEXAGON")
  expect_equal(dggs$projection, "ISEA")
  expect_equal(dggs$precision, 7L)
  expect_equal(dggs$pole_lon_deg, 11.25)
})

test_that("as_dggrid rejects non-hexify_grid objects", {
  expect_error(as_dggrid(list(a = 1)), "must be a hexify_grid object")
  expect_error(as_dggrid(data.frame()), "must be a hexify_grid object")
})

test_that("from_dggrid converts dggridR format to hexify_grid", {
  dggs <- list(
    res = 5L,
    aperture = 3L,
    topology = "HEXAGON",
    projection = "ISEA"
  )

  grid <- from_dggrid(dggs)

  expect_s3_class(grid, "hexify_grid")
  expect_equal(grid$resolution, 5L)
  expect_equal(grid$aperture, 3L)
})

test_that("from_dggrid rejects invalid input", {
  expect_error(from_dggrid("not a list"), "must be a list")
  expect_error(from_dggrid(list()), "missing required fields")
  expect_error(from_dggrid(list(res = 5)), "missing required fields")
})

test_that("from_dggrid warns on unsupported projection", {
  dggs <- list(
    res = 5L,
    aperture = 3L,
    topology = "HEXAGON",
    projection = "FULLER"
  )
  expect_warning(from_dggrid(dggs), "Only ISEA projection")
})

test_that("from_dggrid warns on unsupported topology", {
  dggs <- list(
    res = 5L,
    aperture = 3L,
    topology = "DIAMOND",
    projection = "ISEA"
  )
  expect_warning(from_dggrid(dggs), "Only HEXAGON topology")
})

test_that("from_dggrid rejects unsupported aperture", {
  dggs <- list(
    res = 5L,
    aperture = 5L,
    topology = "HEXAGON",
    projection = "ISEA"
  )
  expect_error(from_dggrid(dggs), "Aperture 5 not supported")
})

test_that("from_dggrid warns on non-default orientation", {
  dggs <- list(
    res = 5L,
    aperture = 3L,
    topology = "HEXAGON",
    projection = "ISEA",
    pole_lon_deg = 0
  )
  expect_warning(from_dggrid(dggs), "Non-default pole_lon_deg")

  dggs$pole_lon_deg <- 11.25
  dggs$pole_lat_deg <- 0
  expect_warning(from_dggrid(dggs), "Non-default pole_lat_deg")

  dggs$pole_lat_deg <- 58.282525588538995
  dggs$azimuth_deg <- 45
  expect_warning(from_dggrid(dggs), "Non-default azimuth_deg")
})

test_that("dggrid_is_compatible validates compatible grids", {
  dggs <- list(
    res = 5L,
    aperture = 3L,
    topology = "HEXAGON",
    projection = "ISEA"
  )
  expect_true(dggrid_is_compatible(dggs))
})

test_that("dggrid_is_compatible rejects incompatible grids (strict=TRUE)", {
  # Not a list
  expect_error(dggrid_is_compatible("not a list"), "not compatible")

  # Wrong projection
  dggs <- list(aperture = 3L, topology = "HEXAGON", projection = "FULLER")
  expect_error(dggrid_is_compatible(dggs), "ISEA projection")

  # Wrong topology
  dggs <- list(aperture = 3L, topology = "DIAMOND", projection = "ISEA")
  expect_error(dggrid_is_compatible(dggs), "HEXAGON topology")

  # Wrong aperture
  dggs <- list(aperture = 5L, topology = "HEXAGON", projection = "ISEA")
  expect_error(dggrid_is_compatible(dggs), "Aperture must be")
})

test_that("dggrid_is_compatible returns FALSE for incompatible grids (strict=FALSE)", {
  dggs <- list(aperture = 3L, topology = "HEXAGON", projection = "FULLER")
  expect_false(dggrid_is_compatible(dggs, strict = FALSE))

  dggs <- list(aperture = 3L, topology = "DIAMOND", projection = "ISEA")
  expect_false(dggrid_is_compatible(dggs, strict = FALSE))
})

test_that("dggrid_is_compatible detects non-default orientation", {
  dggs <- list(
    aperture = 3L,
    topology = "HEXAGON",
    projection = "ISEA",
    pole_lon_deg = 0
  )
  expect_error(dggrid_is_compatible(dggs), "pole_lon_deg")
})

test_that("round-trip: hexify_grid -> dggridR -> hexify_grid", {
  original <- hexify_grid(area = 1000, aperture = 3)
  dggs <- as_dggrid(original)
  recovered <- from_dggrid(dggs)

  expect_equal(original$resolution, recovered$resolution)
  expect_equal(original$aperture, recovered$aperture)
})

## ----setup, include = FALSE---------------------------------------------------
source("_common.R")
library(hexify)

## ----lambert-geometry, echo=FALSE, fig.width=7, fig.height=6------------------
# Lambert Azimuthal Equal-Area Projection Geometry
# Reference: Snyder (1987), "Map Projections: A Working Manual", p. 182-185
#
# The projection formula ρ = 2R·sin(c/2) is derived analytically from
# the equal-area constraint, not from geometric chord distance.
# This diagram shows the geometric relationship that yields the same formula.

oldpar <- par(mar = c(1, 1, 2, 1), bg = "white")
plot(NULL, xlim = c(-1.5, 2.2), ylim = c(-1.5, 1.7), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Lambert Azimuthal Equal-Area Projection Geometry")

R <- 1
theta_circle <- seq(0, 2*pi, length.out = 100)

# Draw sphere (great circle cross-section)
lines(R * cos(theta_circle), R * sin(theta_circle), lwd = 2, col = "gray30")

# Draw tangent plane at S (top of sphere)
lines(c(-1.4, 2.1), c(R, R), lwd = 2, col = "gray50")
text(-1.3, R + 0.1, "Tangent plane", adj = 0, cex = 0.9, col = "gray40")

# Mark center O
points(0, 0, pch = 19, cex = 1.2)
text(-0.12, -0.15, "O", cex = 1.1, font = 2)

# Mark tangent point S
points(0, R, pch = 19, cex = 1.2)
text(0.12, R + 0.12, "S", cex = 1.1, font = 2)

# Choose a point P on sphere at angle φ from S
phi <- 50 * pi/180  # Angular distance from S
P_x <- R * sin(phi)
P_y <- R * cos(phi)
points(P_x, P_y, pch = 19, cex = 1.2, col = "#E63946")
text(P_x + 0.1, P_y + 0.08, "P", cex = 1.1, font = 2, col = "#E63946")

# Draw radius to P
lines(c(0, P_x), c(0, P_y), lwd = 1.5, lty = 2, col = "gray50")

# Project P to tangent plane (perpendicular projection gives P')
# In Lambert: P' is at distance ρ = 2R·sin(φ/2) from S
rho <- 2 * R * sin(phi/2)
Pprime_x <- rho
Pprime_y <- R
points(Pprime_x, Pprime_y, pch = 19, cex = 1.2, col = "#457B9D")
text(Pprime_x + 0.1, Pprime_y + 0.12, "P'", cex = 1.1, font = 2, col = "#457B9D")

# Draw chord from S to P (this has length 2R·sin(φ/2))
lines(c(0, P_x), c(R, P_y), lwd = 2.5, col = "#E63946")

# Draw projected distance on tangent plane
lines(c(0, Pprime_x), c(R, R), lwd = 2.5, col = "#457B9D")

# Draw arc showing angle φ at center
arc_r <- 0.35
arc_theta <- seq(pi/2, pi/2 - phi, length.out = 30)
lines(arc_r * cos(arc_theta), arc_r * sin(arc_theta), lwd = 2, col = "#2A9D8F")
text(arc_r * 1.6, arc_r * 1.2, expression(italic(c)), cex = 1.2, col = "#2A9D8F")

# Label the chord d
mid_chord_x <- (0 + P_x)/2 - 0.12
mid_chord_y <- (R + P_y)/2 + 0.08
text(mid_chord_x, mid_chord_y, expression(italic(d)), cex = 1.2, col = "#E63946")

# Label the projected distance ρ
text(Pprime_x/2, R + 0.15, expression(rho), cex = 1.2, col = "#457B9D")

# Add formula box
rect(0.8, -1.3, 2.15, -0.7, col = "white", border = "gray70")
text(1.47, -0.85, expression(italic(d) == 2*italic(R)*sin(italic(c)/2)), cex = 1.1)
text(1.47, -1.15, expression(rho == italic(d)), cex = 1.1)

# Add reference note
text(1.47, -1.45, "Snyder (1987, p. 182)", cex = 0.9, col = "gray50")
par(oldpar)

## ----lambert-area-preservation, echo=FALSE, fig.width=7, fig.height=3.5-------
# Show equal-area property with concentric rings
oldpar <- par(mfrow = c(1, 2), mar = c(2, 1, 3, 1), bg = "white", cex = 1, cex.main = 1)

R <- 1

# Left: Sphere view (orthographic, looking from above at tangent point)
plot(NULL, xlim = c(-1.3, 1.3), ylim = c(-1.3, 1.3), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Sphere (view from above S)")

# Draw outer circle (equator as seen from S)
theta <- seq(0, 2*pi, length.out = 100)
lines(cos(theta), sin(theta), lwd = 2)

# Draw concentric latitude circles and shade bands
n_bands <- 5
cols <- c("#264653", "#2A9D8F", "#E9C46A", "#F4A261", "#E76F51")

for (i in n_bands:1) {
  phi_outer <- (i / n_bands) * (pi/2)
  phi_inner <- ((i-1) / n_bands) * (pi/2)

  r_outer <- sin(phi_outer)
  r_inner <- sin(phi_inner)

  theta_seq <- seq(0, 2*pi, length.out = 100)
  if (i == 1) {
    polygon(r_outer * cos(theta_seq), r_outer * sin(theta_seq),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  } else {
    polygon(c(r_outer * cos(theta_seq), rev(r_inner * cos(theta_seq))),
            c(r_outer * sin(theta_seq), rev(r_inner * sin(theta_seq))),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  }
}
points(0, 0, pch = 19, cex = 1.2)
text(0, -1.2, "Equal-area bands on sphere", cex = 0.9)

# Right: Lambert projection
plot(NULL, xlim = c(-1.6, 1.6), ylim = c(-1.6, 1.6), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Lambert Projection")

for (i in n_bands:1) {
  phi_outer <- (i / n_bands) * (pi/2)
  phi_inner <- ((i-1) / n_bands) * (pi/2)

  # Lambert projection: r = 2*sin(phi/2)
  r_outer <- 2 * sin(phi_outer / 2)
  r_inner <- 2 * sin(phi_inner / 2)

  theta_seq <- seq(0, 2*pi, length.out = 100)
  if (i == 1) {
    polygon(r_outer * cos(theta_seq), r_outer * sin(theta_seq),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  } else {
    polygon(c(r_outer * cos(theta_seq), rev(r_inner * cos(theta_seq))),
            c(r_outer * sin(theta_seq), rev(r_inner * sin(theta_seq))),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  }
}

r_eq <- 2 * sin(pi/4)
lines(r_eq * cos(theta), r_eq * sin(theta), lwd = 2, lty = 2, col = "gray50")

points(0, 0, pch = 19, cex = 1.2)
text(0, -1.35, "Same bands after projection\n(areas preserved)", cex = 0.9)
par(oldpar)

## ----icosahedron-projection, echo=FALSE, fig.width=7, fig.height=6------------
# Icosahedron Face Projection Geometry
# Reference: Snyder (1992), "An equal-area map projection for polyhedral globes", p. 10-12
# Reference: Coxeter (1973), "Regular Polytopes", p. 52-53
#
# Shows how points on the sphere project onto an icosahedral face tangent plane

oldpar <- par(mar = c(1, 1, 2, 1), bg = "white")
plot(NULL, xlim = c(-1.6, 1.6), ylim = c(-0.6, 1.8), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Icosahedron Face Projection")

# Sphere cross-section (arc visible above the face)
R <- 1.2
theta_arc <- seq(20*pi/180, 160*pi/180, length.out = 50)
sphere_y_offset <- 0.3
lines(R * cos(theta_arc), R * sin(theta_arc) + sphere_y_offset - R,
      lwd = 2, col = "gray40")
text(-1.3, 1.1, "Sphere surface", cex = 0.9, col = "gray50", adj = 0)

# Draw triangular face (equilateral, base at bottom)
face_scale <- 1.3
v1 <- c(-face_scale * 0.866, -0.4)   # bottom-left vertex
v2 <- c(face_scale * 0.866, -0.4)    # bottom-right vertex
v3 <- c(0, face_scale * 1.0)          # top vertex

polygon(c(v1[1], v2[1], v3[1]), c(v1[2], v2[2], v3[2]),
        col = adjustcolor("gray90", 0.6), border = "gray30", lwd = 2)

# Label vertices
text(v1[1] - 0.12, v1[2] - 0.12, "V1", cex = 0.9, font = 2)
text(v2[1] + 0.12, v2[2] - 0.12, "V2", cex = 0.9, font = 2)
text(v3[1], v3[2] + 0.12, "V3", cex = 0.9, font = 2)

# Mark face center
face_center <- c((v1[1] + v2[1] + v3[1])/3, (v1[2] + v2[2] + v3[2])/3)
points(face_center[1], face_center[2], pch = 19, cex = 1)
text(face_center[1] - 0.2, face_center[2] - 0.08, "Face center", cex = 0.9, col = "gray50")

# A point P on sphere surface
P_sphere <- c(0.5, 0.95)
points(P_sphere[1], P_sphere[2], pch = 19, cex = 1.2, col = "#E63946")
text(P_sphere[1] + 0.12, P_sphere[2] + 0.08, "P", cex = 1.1, font = 2, col = "#E63946")

# Projected point P' on face
P_proj <- c(0.5, 0.45)
points(P_proj[1], P_proj[2], pch = 19, cex = 1.2, col = "#457B9D")
text(P_proj[1] + 0.12, P_proj[2] - 0.05, "P'", cex = 1.1, font = 2, col = "#457B9D")

# Draw projection line (dashed)
lines(c(P_sphere[1], P_proj[1]), c(P_sphere[2], P_proj[2]),
      lwd = 2, lty = 2, col = "#2A9D8F")

# Add text explanation
text(0, -0.55, "Each face covers ~1/20 of the sphere", cex = 0.9)
text(1.55, 0.7, "Tangent plane\n(icosa face)", cex = 0.9, col = "gray50", adj = 1)

# Reference note
text(0, -0.75, "Snyder (1992, p. 10-12)", cex = 0.9, col = "gray50")
par(oldpar)

## ----face-centers, echo=FALSE, fig.width=7, fig.height=3.5--------------------
library(sf)
library(ggplot2)

centers <- hexify_face_centers()

center_pts <- st_as_sf(
  data.frame(face = 0:19, lon = centers$lon * 180 / pi, lat = centers$lat * 180 / pi),
  coords = c("lon", "lat"), crs = 4326
)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = center_pts, color = "#E63946", size = 3) +
  geom_sf_text(data = center_pts, aes(label = face), nudge_y = 6, size = 5) +
  labs(title = "ISEA Icosahedron Face Centers",
       subtitle = "20 triangular faces, numbered 0-19") +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----cell-counts--------------------------------------------------------------
cat("Resolution  Aperture 3    Aperture 4    Aperture 7\n")
cat("---------  ----------    ----------    ----------\n")
for (res in 0:8) {
  cells_ap3 <- 10 * 3^res + 2
  cells_ap4 <- 10 * 4^res + 2
  cells_ap7 <- 10 * 7^res + 2
  cat(sprintf("    %d      %10s    %10s    %10s\n",
              res,
              format(cells_ap3, big.mark = ","),
              format(cells_ap4, big.mark = ","),
              format(cells_ap7, big.mark = ",")))
}

## ----orientation-classes, echo=FALSE, fig.width=7, fig.height=4---------------
oldpar <- par(mfrow = c(1, 3), mar = c(1, 0.2, 3, 0.2), bg = "white", cex = 1, cex.main = 1)

hex_v <- function(cx, cy, r, rotation = 0) {
  angles <- seq(rotation, 2*pi + rotation, length.out = 7)
  list(x = cx + r * cos(angles), y = cy + r * sin(angles))
}

# Class I: Flat-top
plot(NULL, xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), asp = 1,
     axes = FALSE, xlab = "", ylab = "", main = "Class I\n(flat-top, 0°)")
h <- hex_v(0, 0, 1.2, rotation = 0)
polygon(h$x, h$y, col = adjustcolor("#457B9D", 0.4), border = "gray30", lwd = 2)
lines(h$x[1:2], h$y[1:2], col = "#E63946", lwd = 3)
text(0, -1.35, "Ap. 4 (all res)\nAp. 3 (even res)", cex = 0.9)

# Class II: Pointy-top (30° rotation)
plot(NULL, xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), asp = 1,
     axes = FALSE, xlab = "", ylab = "", main = "Class II\n(pointy-top, 30°)")
h <- hex_v(0, 0, 1.2, rotation = pi/6)
polygon(h$x, h$y, col = adjustcolor("#2A9D8F", 0.4), border = "gray30", lwd = 2)
points(h$x[1], h$y[1], pch = 19, cex = 1.5, col = "#E63946")
text(0, -1.35, "Ap. 3 (odd res)", cex = 0.9)

# Class III: Skewed
plot(NULL, xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), asp = 1,
     axes = FALSE, xlab = "", ylab = "", main = "Class III\n(skewed, ~19.1°)")
h_parent <- hex_v(0, 0, 1.2, rotation = 0)
polygon(h_parent$x, h_parent$y, col = NA, border = "gray50", lwd = 2, lty = 2)
# Correct rotation: arctan(sqrt(3/7))
rot_angle <- atan(sqrt(3/7))
h_child <- hex_v(0, 0, 0.8, rotation = rot_angle)
polygon(h_child$x, h_child$y, col = adjustcolor("#E9C46A", 0.4),
        border = "gray30", lwd = 2)
arc_r <- 0.5
arc_theta <- seq(0, rot_angle, length.out = 20)
lines(arc_r * cos(arc_theta), arc_r * sin(arc_theta), col = "#E63946", lwd = 2)
text(0.6, 0.15, "19.1°", cex = 0.9, col = "#E63946")
text(0, -1.35, "Ap. 7 (all res)\nRotation accumulates", cex = 0.9)
par(oldpar)

## ----pentagon-locations, echo=FALSE, fig.width=7, fig.height=3.5--------------
pentagon_coords <- data.frame(
  type = c("Pole", "Pole", rep("Upper ring", 5), rep("Lower ring", 5)),
  lon = c(0, 0, 0, 72, 144, 216, 288, 36, 108, 180, 252, 324),
  lat = c(90, -90, rep(26.57, 5), rep(-26.57, 5))
)

pentagon_pts <- st_as_sf(pentagon_coords, coords = c("lon", "lat"), crs = 4326)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = pentagon_pts, aes(color = type), size = 4) +
  scale_color_manual(values = c("Pole" = "#E63946",
                                 "Upper ring" = "#457B9D",
                                 "Lower ring" = "#2A9D8F")) +
  labs(title = "Pentagon Cell Locations",
       subtitle = "12 pentagonal cells at icosahedron vertices\n(area = 5/6 of hexagons)",
       color = "Location") +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())

## ----z7-example---------------------------------------------------------------
# Z7 index encoding for aperture 7
g7 <- hex_grid(resolution = 4, aperture = 7)
cell <- lonlat_to_cell(16.37, 48.21, g7)
idx <- cell_to_index(cell, g7)
cat(sprintf("Cell %d -> Z7 index: %s\n", cell, idx))
cat(sprintf("  Leading field: %s (quad %d), Digits: %s\n",
            substr(idx, 1, 2), as.integer(substr(idx, 1, 2)) %% 12L,
            substr(idx, 3, nchar(idx))))

# Hierarchical property: parent is obtained by dropping the last digit
parent_idx <- substr(idx, 1, nchar(idx) - 1)
parent_info <- hexify_index_to_cell(parent_idx, 7, "z7")
cat(sprintf("  Parent index: %s (face %d, i=%d, j=%d)\n",
            parent_idx, parent_info$face,
            as.integer(parent_info$i), as.integer(parent_info$j)))

# A valid Z7 index decodes and re-encodes without changing
decoded <- hexify_index_to_cell(idx, 7, "z7")
stopifnot(identical(
  hexify_cell_to_index(decoded$face, decoded$i, decoded$j,
                       decoded$resolution, 7, "z7"),
  idx
))

## ----z3-example---------------------------------------------------------------
# Z3 index encoding for aperture 3
g3 <- hex_grid(resolution = 8, aperture = 3)
cell <- lonlat_to_cell(16.37, 48.21, g3)
idx <- cell_to_index(cell, g3)
cat(sprintf("Cell %d -> Z3 index: %s\n", cell, idx))
cat(sprintf("  Base cell: %s, Digits: %s (%d digit pairs)\n",
            substr(idx, 1, 2), substr(idx, 3, nchar(idx)),
            (nchar(idx) - 2) / 2))

## ----zorder-example-----------------------------------------------------------
# Z-order index for aperture 4
g4 <- hex_grid(resolution = 8, aperture = 4)
cell <- lonlat_to_cell(16.37, 48.21, g4)
idx <- cell_to_index(cell, g4)
cat(sprintf("Aperture 4: Cell %d -> Z-order index: %s\n", cell, idx))

## ----h3-area-variation, fig.width=7, fig.height=4.375-------------------------
# Compare ISEA (constant area) vs H3 (variable area) at similar resolutions
lats <- seq(-85, 85, by = 5)
lons <- rep(10, length(lats))

# ISEA: aperture 7, resolution 6 (~130 km² cells)
g_isea <- hex_grid(resolution = 6, aperture = 7)
isea_cells <- lonlat_to_cell(lons, lats, g_isea)
isea_areas <- cell_area(isea_cells, g_isea)

# H3: resolution 4 (~1,770 km² cells — different scale, but shows the pattern)
g_h3 <- hex_grid(resolution = 4, type = "h3")
h3_cells <- lonlat_to_cell(lons, lats, g_h3)
h3_areas <- cell_area(h3_cells, g_h3)

# Normalize to show relative variation
isea_rel <- isea_areas / mean(isea_areas)
h3_rel <- h3_areas / mean(h3_areas)

oldpar <- par(mar = c(4, 4, 2, 1), bg = "white")
plot(lats, h3_rel, type = "l", col = "#E63946", lwd = 2.5,
     xlab = "Latitude (degrees)", ylab = "Relative cell area",
     main = "Cell Area Variation by Latitude",
     ylim = range(c(isea_rel, h3_rel)))
lines(lats, isea_rel, col = "#457B9D", lwd = 2.5)
abline(h = 1, lty = 2, col = "gray50")
legend("topright", legend = c("H3 (gnomonic)", "ISEA (equal-area)"),
       col = c("#E63946", "#457B9D"), lwd = 2.5, bty = "n")
par(oldpar)

## ----resolution-mapping-------------------------------------------------------
# H3 resolution table: compare H3 and ISEA aperture-7 cell areas
cat("H3 Res  Avg Area (km²)  ISEA Ap7 Equivalent\n")
cat("------  --------------  -------------------\n")
for (h3_res in 0:8) {
  g_h3 <- hex_grid(resolution = h3_res, type = "h3")
  h3_area <- g_h3@area_km2

  # Find closest ISEA ap7 resolution by brute force
  best_res <- 0
  best_diff <- Inf
  for (r in 0:15) {
    g_test <- hex_grid(resolution = r, aperture = 7)
    d <- abs(log(g_test@area_km2) - log(h3_area))
    if (d < best_diff) { best_diff <- d; best_res <- r }
  }
  g_isea <- hex_grid(resolution = best_res, aperture = 7)
  cat(sprintf("   %2d   %14.1f  res %d (%.1f km²)\n",
              h3_res, h3_area, best_res, g_isea@area_km2))
}

## ----round-trip---------------------------------------------------------------
original_lon <- 16.37
original_lat <- 48.21

cat(sprintf("Original: (%.4f, %.4f)\n\n", original_lon, original_lat))

for (ap in c(3, 4, 7)) {
  res <- if (ap == 7) 6 else 10
  grid <- hex_grid(resolution = res, aperture = ap)
  cell_id <- lonlat_to_cell(original_lon, original_lat, grid)
  recovered <- cell_to_lonlat(cell_id, grid)
  error_km <- sqrt((recovered$lon - original_lon)^2 +
                   (recovered$lat - original_lat)^2) * 111
  cat(sprintf("Aperture %d (res %2d): cell %d -> (%.4f, %.4f), ~%.1f km from center\n",
              ap, res, cell_id,
              recovered$lon, recovered$lat, error_km))
}

# classes.R
# S4 class definitions for HexGridInfo and HexData
#
# This file defines the core S4 classes that provide stateful return objects
# for the hexify package, enabling cleaner workflows without repeated parameters.

#' @import methods
#' @importFrom methods setClass setMethod setGeneric setValidity
#' @importFrom methods new slot slotNames validObject
NULL

# =============================================================================
# S4 CLASS: HexGridInfo
# =============================================================================
#
# HexGridInfo stores grid specification parameters (aperture, resolution, etc.)
# so downstream functions don't need them repeated.
# =============================================================================

# A grid CRS: an 'EPSG' code, or a 'PROJ' or 'WKT' string
setClassUnion("HexCRS", c("integer", "character"))

#' HexGridInfo Class
#'
#' An S4 class representing a hexagonal grid specification. Stores all
#' parameters needed for grid operations.
#'
#' @slot aperture Character. Grid aperture: "3", "4", "7", a mixed family such
#'   as "4/3" or "4/7", or one aperture per resolution level ("4/4/7/3").
#' @slot resolution Integer. Grid resolution level (0-30 for ISEA, 0-15 for H3).
#' @slot area_km2 Numeric. Cell area in square kilometers.
#' @slot diagonal_km Numeric. Cell diagonal (long diagonal) in kilometers.
#' @slot crs Integer or character. Coordinate reference system: an EPSG code,
#'   or a 'PROJ' or 'WKT' string. Defaults to 'WGS84' on Earth, and to a longlat
#'   CRS on the sphere of \code{radius_km} on any other body.
#' @slot grid_type Character. Grid system: "isea" (default) or "h3".
#' @slot radius_km Numeric. Radius of the body the grid covers, in kilometers.
#'   \code{NA} reads as Earth's mean radius.
#'
#' @details
#' Create HexGridInfo objects using the \code{\link{hex_grid}} constructor function.
#' Do not use \code{new("HexGridInfo", ...)} directly.
#'
#' The aperture can be "3", "4", "7" for grids that refine by one aperture at
#' every level; a family name such as "4/3" or "4/7", which refines by the first
#' aperture for the first floor(resolution / 2) levels and by the second for the
#' rest; or one aperture per level, "4/4/7/3".
#'
#' For H3 grids, the aperture is fixed at "7" and resolution ranges from 0 to 15.
#'
#' @seealso \code{\link{hex_grid}} for the constructor function,
#'   \code{\link{HexData-class}} for hexified data objects
#'
#' @exportClass HexGridInfo
setClass(
  "HexGridInfo",
  slots = c(
    aperture = "character",
    resolution = "integer",
    area_km2 = "numeric",
    diagonal_km = "numeric",
    crs = "HexCRS",
    grid_type = "character",
    radius_km = "numeric"
  ),
  prototype = list(
    aperture = "3",
    resolution = 0L,
    area_km2 = NA_real_,
    diagonal_km = NA_real_,
    crs = 4326L,
    grid_type = "isea",
    radius_km = NA_real_
  )
)

# =============================================================================
# S4 CLASS: HexData
# =============================================================================
#
# HexData wraps user data with cell assignments from hexification.
# Original data is preserved; cell info stored separately.
# =============================================================================

#' HexData Class
#'
#' An S4 class representing hexified data. Contains the original user data
#' plus cell assignments from the hexification process.
#'
#' @slot data Data frame or sf object. The original user data (untouched).
#' @slot grid HexGridInfo object. The grid specification used.
#' @slot cell_id Cell IDs for each row of data. Numeric for ISEA grids,
#'   character for H3 grids.
#' @slot cell_center Matrix. Two-column matrix (lon, lat) of cell centers.
#'
#' @details
#' HexData objects are created by \code{\link{hexify}}. The original data
#' is preserved in the \code{data} slot, while cell assignments are stored
#' separately in \code{cell_id} and \code{cell_center}.
#'
#' Use \code{as.data.frame()} to get a combined data frame with cell columns.
#'
#' @seealso \code{\link{hexify}} for creating HexData objects,
#'   \code{\link{HexGridInfo-class}} for grid specifications
#'
#' @exportClass HexData
setClass(
  "HexData",
  slots = c(
    data = "ANY",  # data.frame or sf
    grid = "HexGridInfo",
    cell_id = "ANY",  # numeric for ISEA, character for H3
    cell_center = "matrix"
  ),
  prototype = list(
    data = data.frame(),
    grid = new("HexGridInfo"),
    cell_id = numeric(0),
    cell_center = matrix(numeric(0), ncol = 2, dimnames = list(NULL, c("lon", "lat")))
  )
)

# =============================================================================
# VALIDITY METHODS
# =============================================================================

#' @noRd
setValidity("HexGridInfo", function(object) {

  errors <- character()

  # Validate grid_type
  gt <- object@grid_type
  if (!gt %in% c("isea", "h3")) {
    errors <- c(errors, "grid_type must be 'isea' or 'h3'")
  }

  if (gt == "h3") {
    # H3 validation: aperture fixed at "7", resolution 0-15
    if (object@aperture != "7") {
      errors <- c(errors, "H3 grids must have aperture '7'")
    }
    if (object@resolution < 0L || object@resolution > 15L) {
      errors <- c(errors, "H3 resolution must be between 0 and 15")
    }
  } else {
    # ISEA validation
    ap_ok <- tryCatch({
      parse_aperture_seq(object@aperture, object@resolution)
      TRUE
    }, error = function(e) FALSE)
    if (!ap_ok) {
      errors <- c(errors, paste0("aperture must be 3, 4, 7, a family such as \"4/3\", ",
                                 "or one aperture per resolution level"))
    }
    if (object@resolution < 0L || object@resolution > 30L) {
      errors <- c(errors, "resolution must be between 0 and 30")
    }
  }

  # Validate area_km2 (must be positive if provided)
  if (!is.na(object@area_km2) && object@area_km2 <= 0) {
    errors <- c(errors, "area_km2 must be positive")
  }

  # Validate diagonal_km (must be positive if provided)
  if (!is.na(object@diagonal_km) && object@diagonal_km <= 0) {
    errors <- c(errors, "diagonal_km must be positive")
  }

  # Validate crs (an EPSG code, or a CRS string sf can read)
  if (is.character(object@crs)) {
    if (length(object@crs) != 1L || is.na(object@crs) || !nzchar(object@crs)) {
      errors <- c(errors, "crs must be a single non-empty CRS string")
    } else if (is.na(parse_crs(object@crs))) {
      errors <- c(errors, sprintf(
        "crs \"%s\" is not a coordinate reference system sf can read",
        object@crs
      ))
    }
  } else if (length(object@crs) != 1L || is.na(object@crs) || object@crs <= 0L) {
    errors <- c(errors, "crs must be a positive integer EPSG code")
  }

  # Validate radius_km (must be a positive finite scalar if provided)
  if (length(object@radius_km) != 1L) {
    errors <- c(errors, "radius_km must be a single number")
  } else if (!is.na(object@radius_km) &&
             (!is.finite(object@radius_km) || object@radius_km <= 0)) {
    errors <- c(errors, "radius_km must be positive")
  }

  if (length(errors) == 0) TRUE else errors
})

#' @noRd
setValidity("HexData", function(object) {
  errors <- character()

  # Check data is valid type
  if (!inherits(object@data, "data.frame") && !inherits(object@data, "sf")) {
    errors <- c(errors, "data must be a data.frame or sf object")
  }

  # Check cell_id type matches grid_type
  gt <- tryCatch(object@grid@grid_type, error = function(e) "isea")
  if (gt == "h3") {
    if (!is.character(object@cell_id) && length(object@cell_id) > 0) {
      errors <- c(errors, "H3 cell_id must be character")
    }
  } else {
    if (!is.numeric(object@cell_id) && length(object@cell_id) > 0) {
      errors <- c(errors, "ISEA cell_id must be numeric")
    }
  }

  # Check cell_id length matches data rows. `length(cell_id) != n_rows` alone
  # already correctly allows the valid empty-prototype case (0 != 0 is
  # FALSE), so no extra "> 0" guard is needed -- and such a guard would hide
  # exactly the corrupt case this check exists to catch (data has rows but
  # cell_id is empty).
  n_rows <- nrow(object@data)
  if (length(object@cell_id) != n_rows) {
    errors <- c(errors, "cell_id length must match number of data rows")
  }

  # Check cell_center dimensions (same reasoning as cell_id above)
  if (nrow(object@cell_center) != n_rows) {
    errors <- c(errors, "cell_center rows must match number of data rows")
  }
  if (ncol(object@cell_center) != 2 && nrow(object@cell_center) > 0) {
    errors <- c(errors, "cell_center must have exactly 2 columns (lon, lat)")
  }

  if (length(errors) == 0) TRUE else errors
})

# =============================================================================
# GENERICS
# =============================================================================

#' Get Grid Specification
#'
#' Extract the grid specification from a HexData object.
#'
#' @param x A HexData object
#' @return A HexGridInfo object
#'
#' @export
#' @examples
#' df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
#' result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
#' grid_spec <- grid_info(result)
setGeneric("grid_info", function(x) standardGeneric("grid_info"))

#' Get Cell IDs
#'
#' Extract the unique cell IDs present in a HexData object.
#'
#' @param x A HexData object
#' @return A vector of cell IDs
#'
#' @export
setGeneric("cells", function(x) standardGeneric("cells"))

#' Get Number of Cells
#'
#' Get the number of unique cells in a HexData object.
#'
#' @param x A HexData object
#' @return Integer count of unique cells
#'
#' @export
setGeneric("n_cells", function(x) standardGeneric("n_cells"))

# =============================================================================
# ACCESSORS FOR HexGridInfo
# =============================================================================

#' HexGridInfo S4 Methods
#'
#' S4 methods for HexGridInfo objects. These provide standard R operations
#' like `$`, `names()`, `show()`, and `as.list()`.
#'
#' @name HexGridInfo-methods
#' @param x HexGridInfo object
#' @param name Slot name
#' @param object HexGridInfo object (for show)
#' @param ... Additional arguments
#' @return
#' - `$`: The value of the requested slot
#' - `names`: Character vector of slot names
#' - `show`: The object, invisibly (called for side effect of printing)
#' - `as.list`: A named list of slot values
#' @keywords internal
NULL

#' @rdname HexGridInfo-methods
#' @export
setMethod("$", "HexGridInfo", function(x, name) {
  slot(x, name)
})

#' @rdname HexGridInfo-methods
#' @keywords internal
#' @export
setMethod("names", "HexGridInfo", function(x) {
  slotNames(x)
})

# =============================================================================
# ACCESSORS FOR HexData
# =============================================================================

#' HexData S4 Methods
#'
#' S4 methods for HexData objects. These provide standard R operations
#' for accessing data, subsetting, and conversion.
#'
#' @name HexData-methods
#' @param x HexData object
#' @param name Column name
#' @param object HexData object (for show)
#' @param i,j Row/column indices
#' @param value Replacement value
#' @param drop Logical, whether to drop dimensions
#' @param row.names Optional row names
#' @param optional Logical (ignored)
#' @param ... Additional arguments
#' @return
#' - `grid_info`: HexGridInfo object containing grid specification
#' - `cells`: Numeric vector of unique cell IDs
#' - `n_cells`: Integer count of unique cells
#' - `nrow`, `ncol`, `dim`: Integer dimensions
#' - `names`: Character vector of column names (including virtual cell columns)
#' - `$`, `[[`: The requested column or cell data as a vector
#' - `$<-`, `[[<-`: The modified HexData object
#' - `[`: Subsetted HexData object or extracted data
#' - `show`: The object, invisibly (called for side effect of printing)
#' - `as.data.frame`: Data frame with original data plus cell columns
#' - `as.list`: Named list containing data, grid, cell_id, and cell_center
#' @keywords internal
NULL

#' @rdname HexData-methods
#' @export
setMethod("grid_info", "HexData", function(x) {
  x@grid
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("cells", "HexData", function(x) {
  unique(x@cell_id)
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("n_cells", "HexData", function(x) {
  length(unique(x@cell_id))
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("nrow", "HexData", function(x) {
  nrow(x@data)
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("ncol", "HexData", function(x) {
  ncol(x@data) + 5L  # +5 for cell_id, cell_cen_lon, cell_cen_lat, cell_area_km2, cell_diag_km
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("dim", "HexData", function(x) {
  dim(x@data)
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("names", "HexData", function(x) {
  c(names(x@data), "cell_id", "cell_cen_lon", "cell_cen_lat", "cell_area_km2", "cell_diag_km")
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("$", "HexData", function(x, name) {
  # Virtual cell columns
  if (name == "cell_id") {
    return(x@cell_id)
  }
  if (name == "cell_cen_lon") {
    return(x@cell_center[, "lon"])
  }
  if (name == "cell_cen_lat") {
    return(x@cell_center[, "lat"])
  }
  if (name == "cell_area_km2") {
    if (is_h3_grid(x@grid)) {
      return(as.numeric(cell_area(grid = x)))
    }
    return(rep(x@grid@area_km2, nrow(x@data)))
  }
  if (name == "cell_diag_km") {
    return(rep(x@grid@diagonal_km, nrow(x@data)))
  }
  # Regular data columns
  x@data[[name]]
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("$<-", "HexData", function(x, name, value) {
  x@data[[name]] <- value
  x
})

#' @rdname HexData-methods
#' @keywords internal
#' @details
#' Unlike \code{[.data.frame}, \code{drop} defaults to \code{FALSE}: selecting
#' a single column returns a \code{HexData} object (preserving \code{grid}/
#' \code{cell_id}/\code{cell_center}) rather than dropping to a bare vector.
#' Pass \code{drop = TRUE} explicitly to get data.frame-style dropping.
#' @export
setMethod("[", c("HexData", "ANY", "ANY"), function(x, i, j, ..., drop = FALSE) {
  # Create new HexData with subsetted data
  new_data <- x@data[i, j, ..., drop = drop]

  # If result is still a data.frame/sf, return HexData
  if (inherits(new_data, "data.frame") || inherits(new_data, "sf")) {
    # Subset cell_id and cell_center if row indices provided
    if (!missing(i)) {
      new_cell_id <- x@cell_id[i]
      new_cell_center <- x@cell_center[i, , drop = FALSE]
    } else {
      new_cell_id <- x@cell_id
      new_cell_center <- x@cell_center
    }

    new("HexData",
        data = new_data,
        grid = x@grid,
        cell_id = new_cell_id,
        cell_center = new_cell_center)
  } else {
    # If subset extracted a vector, return it directly
    new_data
  }
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("[[", c("HexData", "ANY"), function(x, i) {
  # Virtual cell columns by name
  if (is.character(i)) {
    if (i == "cell_id") return(x@cell_id)
    if (i == "cell_cen_lon") return(x@cell_center[, "lon"])
    if (i == "cell_cen_lat") return(x@cell_center[, "lat"])
    if (i == "cell_area_km2") {
      if (is_h3_grid(x@grid)) return(as.numeric(cell_area(grid = x)))
      return(rep(x@grid@area_km2, nrow(x@data)))
    }
    if (i == "cell_diag_km") return(rep(x@grid@diagonal_km, nrow(x@data)))
  }
  x@data[[i]]
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("[[<-", c("HexData", "ANY", "missing", "ANY"), function(x, i, j, value) {
  x@data[[i]] <- value
  x
})

# =============================================================================
# SHOW / PRINT METHODS
# =============================================================================

#' @rdname HexGridInfo-methods
#' @keywords internal
#' @export
setMethod("show", "HexGridInfo", function(object) {
  gt <- tryCatch(object@grid_type, error = function(e) "isea")

  if (gt == "h3") {
    cat("HexGridInfo Specification [H3]\n")
    cat("-------------------------------\n")
    cat(sprintf("Grid Type:   H3 (Uber)\n"))
    cat(sprintf("Resolution:  %d\n", object@resolution))

    if (!is.na(object@area_km2)) {
      cat(sprintf("Avg Area:    %.4f km^2 (varies by location)\n", object@area_km2))
    }
    if (!is.na(object@diagonal_km)) {
      cat(sprintf("Avg Diagonal:%.2f km\n", object@diagonal_km))
    }

    cat(sprintf("CRS:         %s\n", format_crs(object@crs)))

    if (!is_earth_grid(object)) {
      cat(sprintf("Radius:      %.2f km\n", grid_radius_km(object)))
    }

    h3_n_cells <- 2 + 120 * 7^object@resolution
    cat(sprintf("Total Cells: %.0f\n", h3_n_cells))
    cat("Note: H3 cells are NOT exactly equal-area\n")
  } else {
    cat("HexGridInfo Specification\n")
    cat("-------------------------\n")
    cat(sprintf("Aperture:    %s\n", object@aperture))
    cat(sprintf("Resolution:  %d\n", object@resolution))

    if (!is.na(object@area_km2)) {
      cat(sprintf("Area:        %.2f km^2\n", object@area_km2))
    }
    if (!is.na(object@diagonal_km)) {
      cat(sprintf("Diagonal:    %.2f km\n", object@diagonal_km))
    }

    cat(sprintf("CRS:         %s\n", format_crs(object@crs)))

    if (!is_earth_grid(object)) {
      cat(sprintf("Radius:      %.2f km\n", grid_radius_km(object)))
    }

    n_cells <- aperture_n_cells(object@aperture, object@resolution)
    cat(sprintf("Total Cells: %.0f\n", n_cells))
  }

  invisible(object)
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("show", "HexData", function(object) {
  cat("HexData Object\n")
  cat("--------------\n")
  cat(sprintf("Rows:    %d\n", nrow(object@data)))
  cat(sprintf("Columns: %d\n", ncol(object@data)))
  cat(sprintf("Cells:   %d unique\n", n_cells(object)))

  if (inherits(object@data, "sf")) {
    cat("Type:    sf (spatial features)\n")
  } else {
    cat("Type:    data.frame\n")
  }

  cat("\nGrid:\n")
  gt <- tryCatch(object@grid@grid_type, error = function(e) "isea")
  if (gt == "h3") {
    cat(sprintf("  H3 Resolution %d", object@grid@resolution))
    if (!is.na(object@grid@area_km2)) {
      cat(sprintf(" (~%.4f km^2 avg)", object@grid@area_km2))
    }
  } else {
    cat(sprintf("  Aperture %s, Resolution %d",
                object@grid@aperture, object@grid@resolution))
    if (!is.na(object@grid@area_km2)) {
      cat(sprintf(" (~%.1f km^2)", object@grid@area_km2))
    }
  }
  cat("\n")

  # Show column preview
  cat("\nColumns: ")
  col_names <- names(object@data)
  if (length(col_names) > 8) {
    cat(paste(col_names[1:8], collapse = ", "), ", ...\n")
  } else {
    cat(paste(col_names, collapse = ", "), "\n")
  }

  # Show first few rows with cell info

  if (nrow(object@data) > 0) {
    cat("\nData preview (with cell assignments):\n")
    # Combine data with cell info for preview
    preview_df <- data.frame(
      object@data[1:min(3, nrow(object@data)), 1:min(3, ncol(object@data)), drop = FALSE],
      cell_id = object@cell_id[1:min(3, length(object@cell_id))],
      check.names = FALSE
    )
    print(preview_df, row.names = FALSE)

    if (nrow(object@data) > 3) {
      cat(sprintf("... with %d more rows\n", nrow(object@data) - 3))
    }
  }

  invisible(object)
})

# =============================================================================
# COERCION METHODS
# =============================================================================

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("as.data.frame", "HexData", function(x, row.names = NULL,
                                                optional = FALSE, ...) {
  df <- x@data
  if (inherits(df, "sf")) {
    df <- as.data.frame(sf::st_drop_geometry(df))
  }

  # Add cell columns
  df$cell_id <- x@cell_id
  df$cell_cen_lon <- x@cell_center[, "lon"]
  df$cell_cen_lat <- x@cell_center[, "lat"]
  if (is_h3_grid(x@grid)) {
    df$cell_area_km2 <- as.numeric(cell_area(grid = x))
  } else {
    df$cell_area_km2 <- x@grid@area_km2
  }
  df$cell_diag_km <- x@grid@diagonal_km

  if (!is.null(row.names)) {
    rownames(df) <- row.names
  }
  df
})

#' @rdname HexGridInfo-methods
#' @keywords internal
#' @export
setMethod("as.list", "HexGridInfo", function(x, ...) {
  list(
    aperture = x@aperture,
    resolution = x@resolution,
    area_km2 = x@area_km2,
    diagonal_km = x@diagonal_km,
    crs = x@crs,
    grid_type = x@grid_type,
    radius_km = grid_radius_km(x)
  )
})

#' @rdname HexData-methods
#' @keywords internal
#' @export
setMethod("as.list", "HexData", function(x, ...) {
  list(
    data = x@data,
    grid = as.list(x@grid),
    cell_id = x@cell_id,
    cell_center = x@cell_center
  )
})

# =============================================================================
# HELPER FUNCTIONS FOR CLASS CONSTRUCTION
# =============================================================================

#' Check if object is HexGridInfo
#'
#' @param x Object to check
#' @return Logical
#' @export
is_hex_grid <- function(x) {
  inherits(x, "HexGridInfo")
}

#' Check if object is HexData
#'
#' @param x Object to check
#' @return Logical
#' @export
is_hex_data <- function(x) {
  inherits(x, "HexData")
}

#' Extract grid from various objects
#'
#' Internal function to extract a HexGridInfo from different input types.
#' Accepts HexGridInfo, HexData, or legacy hexify_grid objects.
#'
#' @param x Object containing grid info
#' @param allow_null If TRUE, return NULL when x is NULL
#' @return HexGridInfo object
#' @keywords internal
extract_grid <- function(x, allow_null = FALSE) {
  if (is.null(x)) {
    if (allow_null) return(NULL)
    stop("grid specification required")
  }

  if (is_hex_grid(x)) {
    # Handle deserialized old objects without grid_type / radius_km slots
    if (!.hasSlot(x, "grid_type")) {
      x@grid_type <- "isea"
    }
    if (!.hasSlot(x, "radius_km")) {
      x@radius_km <- EARTH_RADIUS_KM
    }
    return(x)
  }

  if (is_hex_data(x)) {
    g <- x@grid
    if (!.hasSlot(g, "grid_type")) {
      g@grid_type <- "isea"
    }
    if (!.hasSlot(g, "radius_km")) {
      g@radius_km <- EARTH_RADIUS_KM
    }
    return(g)
  }

  # Handle legacy hexify_grid objects (S3 class)
  if (inherits(x, "hexify_grid")) {
    return(hexify_grid_to_HexGridInfo(x))
  }

  stop("Cannot extract grid from object of class ", class(x)[1])
}

#' Convert legacy hexify_grid to HexGridInfo
#'
#' @param x A hexify_grid object (S3)
#' @return A HexGridInfo object (S4)
#' @keywords internal
hexify_grid_to_HexGridInfo <- function(x) {
  area <- if (!is.null(x$area)) as.numeric(x$area) else NA_real_
  diagonal <- if (!is.na(area)) sqrt(area * 2 / sqrt(3)) else NA_real_

  new("HexGridInfo",
      aperture = as.character(x$aperture),
      resolution = as.integer(x$resolution),
      area_km2 = area,
      diagonal_km = diagonal,
      crs = resolve_crs(x$crs, grid_radius_km(x)),
      radius_km = grid_radius_km(x))
}

#' Convert HexGridInfo to legacy hexify_grid
#'
#' For backwards compatibility with existing functions.
#'
#' @param x A HexGridInfo object (S4)
#' @return A hexify_grid object (S3)
#' @keywords internal
HexGridInfo_to_hexify_grid <- function(x) {
  # H3 grids cannot be converted to legacy format
 gt <- tryCatch(x@grid_type, error = function(e) "isea")
  if (gt == "h3") {
    stop("H3 grids cannot be converted to legacy hexify_grid format")
  }

  # Determine legacy index_type based on aperture
  ap <- x@aperture
  legacy_index <- if (ap == "3") {
    "z3"
  } else if (ap == "7") {
    "z7"
  } else {
    "zorder"
  }

  # Convert aperture to numeric for legacy
  aperture_num <- aperture_to_int(ap)

  grid <- list(
    area = x@area_km2,
    resolution = x@resolution,
    aperture = aperture_num,
    topology = "HEXAGON",
    projection = "ISEA",
    metric = TRUE,
    radius_km = grid_radius_km(x),
    index_type = legacy_index,
    res = x@resolution,
    topology_family = "HEXAGON",
    metric_radius = if (!is.na(x@area_km2)) sqrt(x@area_km2 / pi) else NULL,
    pole_lon_deg = ISEA_VERT0_LON_DEG,
    pole_lat_deg = ISEA_VERT0_LAT_DEG,
    azimuth_deg = ISEA_AZIMUTH_DEG,
    # MIXED43 is DGGRID's own name for the 4/3 arrangement; other sequences
    # have no DGGRID aperture type, so they carry the generic label.
    aperture_type = if (ap == "4/3") "MIXED43" else if (is_mixed_aperture(ap)) "MIXED" else "SEQUENCE",
    res_spec = x@resolution,
    precision = 7
  )

  class(grid) <- c("hexify_grid", "dggs", "list")
  grid
}

# tests/testthat/test-classes.R
# Tests for S4 classes HexGridInfo and HexData

# =============================================================================
# HexGridInfo Tests
# =============================================================================

test_that("HexGridInfo can be created via hex_grid", {
  grid <- hex_grid(area_km2 = 1000)

  expect_s4_class(grid, "HexGridInfo")
  expect_equal(grid@aperture, "3")
  expect_true(grid@resolution > 0)
  expect_true(grid@area_km2 > 0)
})

test_that("HexGridInfo $ accessor works", {
  grid <- hex_grid(area_km2 = 1000)

  expect_equal(grid$aperture, "3")
  expect_equal(grid$resolution, grid@resolution)
  expect_equal(grid$area_km2, grid@area_km2)
  expect_equal(grid$crs, 4326L)
})

test_that("HexGridInfo names() returns slot names", {
  grid <- hex_grid(area_km2 = 1000)

  expect_true("aperture" %in% names(grid))
  expect_true("resolution" %in% names(grid))
  expect_true("area_km2" %in% names(grid))
  expect_true("diagonal_km" %in% names(grid))
  expect_true("crs" %in% names(grid))
})

test_that("HexGridInfo show() prints correctly", {
  grid <- hex_grid(area_km2 = 1000)

  output <- capture.output(show(grid))
  expect_true(any(grepl("HexGridInfo", output)))
  expect_true(any(grepl("Aperture", output)))
  expect_true(any(grepl("Resolution", output)))
})

test_that("HexGridInfo show() handles mixed aperture", {
  grid <- hex_grid(area_km2 = 1000, aperture = "4/3")

  output <- capture.output(show(grid))
  expect_true(any(grepl("4/3", output)))
})

test_that("HexGridInfo as.list() works", {
  grid <- hex_grid(area_km2 = 1000)
  lst <- as.list(grid)

  expect_type(lst, "list")
  expect_equal(lst$aperture, "3")
  expect_equal(lst$resolution, grid@resolution)
  expect_equal(lst$crs, 4326L)
})

test_that("is_hex_grid identifies HexGridInfo", {
  grid <- hex_grid(area_km2 = 1000)

  expect_true(is_hex_grid(grid))
  expect_false(is_hex_grid(list()))
  expect_false(is_hex_grid(data.frame()))
})

# =============================================================================
# HexData Tests
# =============================================================================

test_that("HexData is created by hexify", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  expect_s4_class(result, "HexData")
  expect_s4_class(result@grid, "HexGridInfo")
  expect_equal(length(result@cell_id), 2)
})

test_that("HexData $ accessor works for data columns", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = c(1, 2))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  expect_equal(result$value, c(1, 2))
  expect_equal(result$lon, c(0, 10))
})

test_that("HexData $ accessor works for virtual cell columns", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  expect_equal(result$cell_id, result@cell_id)
  expect_equal(result$cell_cen_lon, result@cell_center[, "lon"])
  expect_equal(result$cell_cen_lat, result@cell_center[, "lat"])
  expect_length(result$cell_area_km2, 2)
  expect_length(result$cell_diag_km, 2)
})

test_that("HexData $<- assignment works", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  result$new_col <- c("a", "b")
  expect_equal(result$new_col, c("a", "b"))
})

test_that("HexData [[ accessor works", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = c(1, 2))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  expect_equal(result[["value"]], c(1, 2))
  expect_equal(result[["cell_id"]], result@cell_id)
  expect_equal(result[["cell_cen_lon"]], result@cell_center[, "lon"])
  expect_equal(result[["cell_area_km2"]], rep(result@grid@area_km2, 2))
})

test_that("HexData [[<- assignment works", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  result[["new_col"]] <- c(100, 200)
  expect_equal(result[["new_col"]], c(100, 200))
})

test_that("HexData [ subsetting works with row indices", {
  df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55), value = 1:3)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  subset <- result[1:2, ]
  expect_s4_class(subset, "HexData")
  expect_equal(nrow(subset), 2)
  expect_equal(length(subset@cell_id), 2)
})

test_that("HexData [ subsetting returns HexData for column subset", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # Column subsetting returns HexData (not a vector)
  subset <- result[, "value", drop = FALSE]
  expect_s4_class(subset, "HexData")
})

test_that("HexData nrow/ncol/dim work", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), a = 1:2, b = 3:4)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  expect_equal(nrow(result), 2)
  expect_equal(ncol(result), ncol(df) + 5)  # +5 for cell columns
  expect_equal(dim(result), dim(df))
})

test_that("HexData names() includes virtual columns", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  nm <- names(result)
  expect_true("cell_id" %in% nm)
  expect_true("cell_cen_lon" %in% nm)
  expect_true("cell_cen_lat" %in% nm)
  expect_true("cell_area_km2" %in% nm)
  expect_true("cell_diag_km" %in% nm)
})

test_that("HexData grid_info() returns grid", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  g <- grid_info(result)
  expect_s4_class(g, "HexGridInfo")
  expect_equal(g@aperture, "3")
})

test_that("HexData cells() returns unique cell IDs", {
  df <- data.frame(lon = c(0, 0.001, 10), lat = c(45, 45.001, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  cell_ids <- cells(result)
  expect_equal(cell_ids, unique(result@cell_id))
})

test_that("HexData n_cells() counts unique cells", {
  df <- data.frame(lon = c(0, 0.001, 10), lat = c(45, 45.001, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 10000)

  n <- n_cells(result)
  expect_equal(n, length(unique(result@cell_id)))
})

test_that("HexData show() prints correctly", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  output <- capture.output(show(result))
  expect_true(any(grepl("HexData", output)))
  expect_true(any(grepl("Rows", output)))
  expect_true(any(grepl("Cells", output)))
})

test_that("HexData show() handles sf data", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  sf_data <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  result <- hexify(sf_data, area_km2 = 1000)

  output <- capture.output(show(result))
  expect_true(any(grepl("sf", output)))
})

test_that("HexData show() handles many columns", {
  df <- data.frame(
    lon = c(0, 10), lat = c(45, 50),
    a = 1:2, b = 1:2, c = 1:2, d = 1:2, e = 1:2,
    f = 1:2, g = 1:2, h = 1:2, i = 1:2, j = 1:2
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  output <- capture.output(show(result))
  expect_true(any(grepl("\\.\\.\\.", output)))  # Truncated columns
})

test_that("HexData as.data.frame() works", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  df_out <- as.data.frame(result)
  expect_s3_class(df_out, "data.frame")
  expect_true("cell_id" %in% names(df_out))
  expect_true("cell_cen_lon" %in% names(df_out))
  expect_true("value" %in% names(df_out))
})

test_that("HexData as.data.frame() drops sf geometry", {
  skip_if_not_installed("sf")

  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  sf_data <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  result <- hexify(sf_data, area_km2 = 1000)

  df_out <- as.data.frame(result)
  expect_false(inherits(df_out, "sf"))
  expect_true("cell_id" %in% names(df_out))
})

test_that("HexData as.data.frame() accepts row.names", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  df_out <- as.data.frame(result, row.names = c("first", "second"))
  expect_equal(rownames(df_out), c("first", "second"))
})

test_that("HexData as.list() works", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  lst <- as.list(result)
  expect_type(lst, "list")
  expect_true("data" %in% names(lst))
  expect_true("grid" %in% names(lst))
  expect_true("cell_id" %in% names(lst))
})

test_that("is_hex_data identifies HexData", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  expect_true(is_hex_data(result))
  expect_false(is_hex_data(df))
  expect_false(is_hex_data(list()))
})

# =============================================================================
# Validity Tests
# =============================================================================

test_that("HexGridInfo validity rejects invalid aperture", {
  expect_error(
    new("HexGridInfo", aperture = "5", resolution = 5L, crs = 4326L),
    "aperture must be"
  )
})

test_that("HexGridInfo validity rejects invalid resolution", {
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = -1L, crs = 4326L),
    "resolution must be between"
  )
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = 50L, crs = 4326L),
    "resolution must be between"
  )
})

test_that("HexGridInfo validity rejects invalid area_km2", {
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = 5L, area_km2 = -100, crs = 4326L),
    "area_km2 must be positive"
  )
})

test_that("HexGridInfo validity rejects invalid diagonal_km", {
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = 5L, diagonal_km = -50, crs = 4326L),
    "diagonal_km must be positive"
  )
})

test_that("HexGridInfo validity rejects invalid crs", {
  expect_error(
    new("HexGridInfo", aperture = "3", resolution = 5L, crs = -1L),
    "crs must be a positive"
  )
})

# =============================================================================
# extract_grid Tests
# =============================================================================

test_that("extract_grid returns NULL when allow_null = TRUE", {
  result <- hexify:::extract_grid(NULL, allow_null = TRUE)
  expect_null(result)
})

test_that("extract_grid errors on NULL when allow_null = FALSE", {
  expect_error(hexify:::extract_grid(NULL), "grid specification required")
})

test_that("extract_grid errors on invalid object type", {
  expect_error(hexify:::extract_grid(list(a = 1)), "Cannot extract grid")
  expect_error(hexify:::extract_grid(data.frame()), "Cannot extract grid")
  expect_error(hexify:::extract_grid("string"), "Cannot extract grid")
})

# =============================================================================
# HexGridInfo_to_hexify_grid Tests
# =============================================================================

test_that("HexGridInfo_to_hexify_grid converts aperture 3", {
  grid <- hex_grid(area_km2 = 1000, aperture = 3)
  legacy <- hexify:::HexGridInfo_to_hexify_grid(grid)

  expect_s3_class(legacy, "hexify_grid")
  expect_equal(legacy$aperture, 3L)
  expect_equal(legacy$index_type, "z3")
})

test_that("HexGridInfo_to_hexify_grid converts aperture 7", {
  grid <- hex_grid(area_km2 = 10000, aperture = 7)
  legacy <- hexify:::HexGridInfo_to_hexify_grid(grid)

  expect_s3_class(legacy, "hexify_grid")
  expect_equal(legacy$aperture, 7L)
  expect_equal(legacy$index_type, "z7")
})

test_that("HexGridInfo_to_hexify_grid converts aperture 4", {
  grid <- hex_grid(area_km2 = 1000, aperture = 4)
  legacy <- hexify:::HexGridInfo_to_hexify_grid(grid)

  expect_s3_class(legacy, "hexify_grid")
  expect_equal(legacy$aperture, 4L)
  expect_equal(legacy$index_type, "zorder")
})

test_that("HexGridInfo_to_hexify_grid converts mixed aperture 4/3", {
  grid <- hex_grid(area_km2 = 1000, aperture = "4/3")
  legacy <- hexify:::HexGridInfo_to_hexify_grid(grid)

  expect_s3_class(legacy, "hexify_grid")
  expect_equal(legacy$aperture, 3L)  # Base aperture for mixed
  expect_equal(legacy$aperture_type, "MIXED43")
})

# =============================================================================
# HexData [ subset edge cases
# =============================================================================

test_that("HexData [ returns vector when extracting single column without drop=FALSE", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # Extracting single column with drop=TRUE returns vector
  vec <- result@data[, "value", drop = TRUE]
  expect_type(vec, "integer")
})

test_that("HexData [[ accessor works with numeric index", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50), value = 1:2)
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  # Numeric indexing goes to underlying data
  expect_equal(result[[1]], c(0, 10))  # First column (lon)
})

test_that("HexData [[ accessor returns cell_diag_km", {
  df <- data.frame(lon = c(0, 10), lat = c(45, 50))
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

  diag <- result[["cell_diag_km"]]
  expect_length(diag, 2)
  expect_true(all(diag > 0))
})

# =============================================================================
# HexData show() edge cases
# =============================================================================

test_that("HexData show() handles more than 3 rows", {
  df <- data.frame(
    lon = c(0, 5, 10, 15, 20),
    lat = c(45, 46, 47, 48, 49),
    value = 1:5
  )
  result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 5000)

  output <- capture.output(show(result))
  expect_true(any(grepl("more rows", output)))
})

# tests/testthat/test-import-h3.R
# Tests for import_h3() (R/h3_compat.R). Previously had zero test coverage.

h3_ids <- c("8528342bfffffff", "85283473fffffff", "85283447fffffff")

test_that("import_h3 returns a HexGridInfo when data is NULL", {
  grid <- import_h3(h3_ids)
  expect_true(is_hex_grid(grid))
  expect_equal(grid@grid_type, "h3")
  expect_equal(grid@resolution, cpp_h3_getResolution(h3_ids)[1])
})

test_that("import_h3 returns a HexData when data is supplied", {
  df <- data.frame(species = c("oak", "pine", "birch"), count = c(10, 5, 3))
  hd <- import_h3(h3_ids, data = df)

  expect_true(is_hex_data(hd))
  expect_equal(nrow(as.data.frame(hd)), 3)
  expect_equal(as.character(hd@cell_id), h3_ids)
})

test_that("import_h3 validates cell_ids type", {
  expect_error(import_h3(c(1, 2, 3)), "character vector")
})

test_that("import_h3 errors on empty input", {
  expect_error(import_h3(character(0)), "must not be empty")
})

test_that("import_h3 errors on all-NA input", {
  expect_error(import_h3(c(NA_character_, NA_character_)), "only NA")
})

test_that("import_h3 drops NA entries but keeps valid ones for resolution inference", {
  ids_with_na <- c(h3_ids[1], NA, h3_ids[2])
  grid <- import_h3(ids_with_na)
  expect_true(is_hex_grid(grid))
})

test_that("import_h3 rejects invalid H3 cell IDs when validate = TRUE", {
  expect_error(import_h3(c(h3_ids[1], "not-a-real-cell")), "invalid H3 cell")
})

test_that("import_h3 skips validation when validate = FALSE", {
  # Even a malformed ID doesn't error at import time when validation is off;
  # only resolution inference runs.
  expect_no_error(import_h3(h3_ids, validate = FALSE))
})

test_that("import_h3 errors on mixed H3 resolutions", {
  # h3_ids[1:2] are resolution 5; get a resolution 6 child of a different cell
  child <- get_children(h3_ids[1], hex_grid(resolution = 5, type = "h3"))[[1]][1]
  expect_error(import_h3(c(h3_ids[1], child)), "same H3 resolution")
})

test_that("import_h3 validates data dimensions", {
  bad_df <- data.frame(x = 1:2)  # only 2 rows for 3 cell_ids
  expect_error(import_h3(h3_ids, data = bad_df), "must match")
})

test_that("import_h3 requires data to be a data.frame", {
  expect_error(import_h3(h3_ids, data = list(x = 1:3)), "data.frame")
})

# tests/testthat/test-namespace-hygiene.R
# Guards on the generated Rcpp bindings.
#
# Rcpp::compileAttributes() writes one R function per [[Rcpp::export]] into
# R/RcppExports.R. That file sorts before most of R/, so a binding named after
# a package function would be the definition R keeps or drops depending on
# collation order alone. Prefixing every entry point cpp_ keeps the two name
# spaces apart.

rcpp_binding_names <- function() {
  dll <- .getNamespaceInfo(asNamespace("hexify"), "DLLs")[["hexify"]]
  sub("^_hexify_", "", names(getDLLRegisteredRoutines(dll)$.Call))
}

test_that("every compiled entry point is cpp_ prefixed", {
  bindings <- rcpp_binding_names()
  expect_gt(length(bindings), 0)
  expect_equal(bindings[!startsWith(bindings, "cpp_")], character(0))
})

test_that("no compiled entry point shadows a package function", {
  clashes <- intersect(rcpp_binding_names(), getNamespaceExports("hexify"))
  expect_equal(clashes, character(0))
})

test_that("no compiled entry point shadows an internal R function", {
  ns <- asNamespace("hexify")
  r_sources <- setdiff(ls(ns, all.names = TRUE), rcpp_binding_names())
  expect_equal(intersect(rcpp_binding_names(), r_sources), character(0))
})

---
title: "H3 Grid Support"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{H3 Grid Support}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
CAN_RUN <- requireNamespace("sf", quietly = TRUE)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  eval = CAN_RUN
)
library(hexify)
```

hexify v0.5.0 adds [H3](https://h3geo.org/) as a first-class grid type alongside
ISEA. Every core function --- `hexify()`, `grid_rect()`, `grid_clip()`,
`get_parent()`, `get_children()` --- works with both grid systems through the
same interface. This vignette covers what H3 is, how to use it in hexify,
and when to prefer it over ISEA.

## What is H3?

H3 is a hierarchical hexagonal grid system developed by Uber. It partitions
Earth's surface into hexagonal cells at 16 resolutions (0--15), each roughly
7$\times$ finer than the last. Cell IDs are 64-bit integers encoded as
hexadecimal strings (e.g., `"8528342bfffffff"`).

H3 has become an industry standard adopted by the FCC, Foursquare, and
numerous geospatial platforms.

**Key difference from ISEA**: H3 cells are _not_ equal-area. Cell area varies
by ~1.6$\times$ between the largest and smallest hexagons at any given
resolution, depending on latitude. For rigorous equal-area analysis, use ISEA.
For interoperability with H3 ecosystems, use `type = "h3"`.

## Getting Started

Create an H3 grid by passing `type = "h3"` to `hex_grid()`:

```{r h3-grid}
library(sf)
library(ggplot2)

# Create an H3 grid specification
grid_h3 <- hex_grid(resolution = 5, type = "h3")
grid_h3
```

Then use `hexify()` with the grid object, just like ISEA:

```{r h3-hexify}
# Sample cities
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome",
           "London", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50,
          -0.12, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90,
          51.51, 50.08, 52.23, 47.50, 52.37)
)

result <- hexify(cities, lon = "lon", lat = "lat", grid = grid_h3)
result
```

H3 cell IDs are character strings, unlike ISEA's numeric IDs:

```{r h3-cell-ids}
# Cell IDs are hexadecimal strings
result@cell_id

# All standard accessors work
cells(result)
n_cells(result)
```

### Choosing Resolution by Area

If you think in terms of cell area rather than resolution numbers, pass
`area_km2` instead of `resolution`. hexify picks the closest H3 resolution:

```{r h3-by-area}
grid_area <- hex_grid(area_km2 = 500, type = "h3")
grid_area
```

## Grid Generation

All grid generation functions work with H3 grids.

### Rectangular Region

```{r h3-grid-rect, fig.width=7, fig.height=5}
# Generate H3 hexagons over Western Europe
grid_h3 <- hex_grid(resolution = 3, type = "h3")
europe_h3 <- grid_rect(c(-10, 35, 25, 60), grid_h3)

# Basemap
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_h3, fill = NA, color = "#E6550D", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("H3 Resolution %d Grid (~%.0f km² avg cells)",
                       grid_h3@resolution, grid_h3@area_km2)) +
  theme_minimal()
```

### Clipping to a Boundary

```{r h3-grid-clip, fig.width=6, fig.height=6}
# Clip H3 grid to France
france <- hexify_world[hexify_world$name == "France", ]
grid_h3 <- hex_grid(resolution = 4, type = "h3")
france_h3 <- grid_clip(france, grid_h3)

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_h3, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = sprintf("H3 Grid Clipped to France (res %d)", grid_h3@resolution)) +
  theme_minimal()
```

## Hierarchical Navigation

H3's killer feature is its clean hierarchical structure: every cell has exactly
one parent and seven children. hexify exposes this with `get_parent()` and
`get_children()`.

### Parents

```{r h3-parent}
# Get parent cells (one resolution coarser)
grid_h3 <- hex_grid(resolution = 5, type = "h3")
child_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40),
  lat = c(48.21, 48.86, 52.52),
  grid = grid_h3
)

parent_ids <- get_parent(child_ids, grid_h3, levels = 1)
data.frame(child = child_ids, parent = parent_ids)
```

### Children

```{r h3-children}
# Get children of a single cell (one resolution finer)
grid_coarse <- hex_grid(resolution = 3, type = "h3")
coarse_id <- lonlat_to_cell(16.37, 48.21, grid_coarse)

children <- get_children(coarse_id, grid_coarse, levels = 1)
cat(length(children[[1]]), "children at resolution", grid_coarse@resolution + 1, "\n")
head(children[[1]])
```

### Visualizing the Hierarchy

```{r h3-hierarchy-plot, fig.width=7, fig.height=5}
# Parent cell polygon
parent_poly <- cell_to_sf(coarse_id, grid_coarse)

# Children cell polygons
grid_fine <- hex_grid(resolution = 4, type = "h3")
children_poly <- cell_to_sf(children[[1]], grid_fine)

ggplot() +
  geom_sf(data = children_poly, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.5) +
  geom_sf(data = parent_poly, fill = NA, color = "black", linewidth = 1.2) +
  labs(title = sprintf("H3 Hierarchy: 1 parent (res %d) → %d children (res %d)",
                       grid_coarse@resolution,
                       length(children[[1]]),
                       grid_fine@resolution)) +
  theme_minimal()
```

## Working with H3 Data

The standard hexify workflow applies to H3 grids. Here's a complete example
using simulated species observations:

```{r h3-workflow, fig.width=8, fig.height=6, message=FALSE, warning=FALSE}
set.seed(42)

# Simulate observations across Europe
obs <- data.frame(
  lon = c(rnorm(200, 10, 12), rnorm(100, 25, 8)),
  lat = c(rnorm(200, 48, 6), rnorm(100, 55, 4)),
  species = sample(c("Sp. A", "Sp. B", "Sp. C"), 300, replace = TRUE)
)
obs$lon <- pmax(-10, pmin(40, obs$lon))
obs$lat <- pmax(35, pmin(65, obs$lat))

# Hexify with H3
grid_h3 <- hex_grid(resolution = 3, type = "h3")
obs_hex <- hexify(obs, lon = "lon", lat = "lat", grid = grid_h3)

# Aggregate: species richness per cell
obs_df <- as.data.frame(obs_hex)
obs_df$cell_id <- obs_hex@cell_id

richness <- aggregate(species ~ cell_id, data = obs_df,
                      FUN = function(x) length(unique(x)))
names(richness)[2] <- "n_species"

# Map it
polys <- cell_to_sf(richness$cell_id, grid_h3)
polys <- merge(polys, richness, by = "cell_id")

europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = polys, aes(fill = n_species), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "plasma", name = "Species\nRichness") +
  coord_sf(xlim = c(-10, 40), ylim = c(35, 65)) +
  labs(title = "Species Richness on H3 Grid",
       subtitle = sprintf("H3 resolution %d (~%.0f km² avg cells)",
                          grid_h3@resolution, grid_h3@area_km2)) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## ISEA--H3 Crosswalk

hexify v0.6.0 added `h3_crosswalk()` for bidirectional mapping between ISEA
and H3 cell IDs. This is useful when you work in ISEA for analysis but need to
share results with H3 ecosystems (or vice versa).

```{r h3-crosswalk}
# Start with an ISEA grid and some cells
grid_isea <- hex_grid(resolution = 9, aperture = 3)
isea_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40, -3.70, 12.50),
  lat = c(48.21, 48.86, 52.52, 40.42, 41.90),
  grid = grid_isea
)

# Map ISEA cells to their closest H3 equivalents
xw <- h3_crosswalk(isea_ids, grid_isea)
xw[, c("isea_cell_id", "h3_cell_id", "isea_area_km2", "h3_area_km2")]
```

The `area_ratio` column shows how ISEA and H3 cell sizes compare ---
values close to 1 mean the resolutions are well-matched.

## ISEA vs H3: When to Use Which

| | ISEA | H3 |
|---|---|---|
| **Cell area** | Exactly equal | ~1.6$\times$ variation |
| **Cell IDs** | Numeric (integer) | Character (hex string) |
| **Apertures** | 3, 4, 7, 4/3 | Fixed (7) |
| **Resolutions** | 0--30 | 0--15 |
| **Hierarchy** | Approximate (aperture-dependent) | Exact (7 children per parent) |
| **Dependencies** | None (built-in C++) | None (vendored H3 C library) |
| **Industry adoption** | Scientific / government | Tech industry / commercial |

**Use ISEA when**:

- Equal-area cells are required (biodiversity surveys, density estimation,
  statistical sampling)

- You need fine control over aperture and resolution

- No external dependencies are acceptable

**Use H3 when**:

- Interoperability with H3 ecosystems (Uber, Foursquare, DuckDB, BigQuery)

- Clean hierarchical operations (parent/child traversal) are a priority

- Slight area variation across latitudes is acceptable for your analysis

## Resolution Reference

```{r h3-resolution-table}
h3_res <- hexify_compare_resolutions(type = "h3", res_range = 0:15)
h3_res$n_cells_fmt <- ifelse(
  h3_res$n_cells > 1e9,
  sprintf("%.1fB", h3_res$n_cells / 1e9),
  ifelse(h3_res$n_cells > 1e6,
         sprintf("%.1fM", h3_res$n_cells / 1e6),
         ifelse(h3_res$n_cells > 1e3,
                sprintf("%.1fK", h3_res$n_cells / 1e3),
                as.character(h3_res$n_cells)))
)
knitr::kable(
  h3_res[, c("resolution", "n_cells_fmt", "cell_area_km2", "cell_spacing_km")],
  col.names = c("Resolution", "# Cells", "Avg Area (km²)", "Spacing (km)"),
  digits = 1
)
```

Areas are averages --- actual cell area varies by latitude.

## See Also

- `vignette("quickstart")` - Getting started with hexify (ISEA-focused)

- `vignette("workflows")` - Grid generation, multi-resolution analysis, GIS export

- `vignette("visualization")` - Plotting with `plot()`, `hexify_heatmap()`

- `vignette("theory")` - Mathematical foundations of the ISEA projection

---
title: "Quick Start"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Quick Start}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
library(hexify)
```

```{r cover-image, echo=FALSE, message=FALSE, warning=FALSE, fig.width=8, fig.height=4, fig.align='center'}
# Generate cover image: multi-resolution grids over different regions
library(sf)
library(ggplot2)

# Get world basemap
countries <- hexify_world

# Generate grids at different resolutions for specific regions (coarser for speed)
# Coarse grid: South America
grid_sa <- hex_grid(area_km2 = 300000)
sa_hexes <- grid_rect(c(-80, -40, -35, 10), grid_sa)

# Medium grid: Europe/Africa
grid_eu <- hex_grid(area_km2 = 80000)
eu_hexes <- grid_rect(c(-10, -10, 40, 50), grid_eu)

# Fine grid: East Asia
grid_asia <- hex_grid(area_km2 = 50000)
asia_hexes <- grid_rect(c(90, 10, 140, 50), grid_asia)

ggplot() +
  geom_sf(data = countries, fill = "gray95", color = "gray80", linewidth = 0.15) +
  geom_sf(data = sa_hexes, fill = NA, color = "#1B9E77", linewidth = 0.5) +
  geom_sf(data = eu_hexes, fill = NA, color = "#D95F02", linewidth = 0.4) +
  geom_sf(data = asia_hexes, fill = NA, color = "#7570B3", linewidth = 0.3) +
  coord_sf(xlim = c(-100, 150), ylim = c(-50, 60)) +
  theme_void() +
  theme(panel.background = element_rect(fill = "white", color = NA))
```

# Spatial Analysis Done Right

You want to do spatial statistics, and it involves binning points into grid cells.

**The problem with rectangular grids**: A rectangular lat-lon grid introduces severe distortions. At the equator, a 1° cell covers ~12,300 km². Near the poles, the same 1° cell covers a tiny fraction of that area. This breaks any analysis that assumes equal sampling effort or comparable cell sizes.

**The solution**: Discrete global grids partition Earth's surface into cells of **equal area**, regardless of latitude. hexify implements the ISEA (Icosahedral Snyder Equal Area) projection, providing hexagonal cells that are all the same size from the equator to the Arctic.

## Basic Usage

```{r basic-usage}
# Sample data: European cities
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90)
)

# Create a grid specification
grid <- hex_grid(area_km2 = 10000)
grid

# Assign cities to hexagonal cells
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)
result
```

### Accessing HexData

```{r hexdata-access}
# Get the grid specification
grid_info(result)

# Get unique cell IDs
cells(result)

# Count unique cells
n_cells(result)

# Access all cell IDs (one per row)
result@cell_id

# Access cell centers
head(result@cell_center)

# Extract original data as data.frame
head(as.data.frame(result))
```

### With sf Objects

```{r sf-usage, eval = requireNamespace("sf", quietly = TRUE)}
library(sf)

# Create sf object (any CRS works - hexify transforms automatically)
pts <- st_as_sf(cities, coords = c("lon", "lat"), crs = 4326)

# hexify handles CRS transformation automatically
result_sf <- hexify(pts, area_km2 = 10000)
class(result_sf)
```

## Real-World Example: Species Occurrence Data

This example demonstrates the typical workflow: loading point data, assigning to grid cells, aggregating, and visualizing.

```{r bird-setup, message=FALSE, warning=FALSE}
library(sf)
library(ggplot2)

# Simulate bird observation data (reduced for faster vignette build)
set.seed(123)
n_obs <- 500

# Generate observations with realistic spatial clustering
birds <- data.frame(
  lon = c(
    rnorm(150, mean = 10, sd = 15),    # Western Europe
    rnorm(100, mean = 25, sd = 10),    # Eastern Europe
    rnorm(100, mean = 20, sd = 20),    # Mediterranean
    rnorm(80, mean = 0, sd = 15),      # West Africa
    rnorm(70, mean = 35, sd = 10)      # East Africa
  ),
  lat = c(
    rnorm(150, mean = 50, sd = 8),     # Western Europe
    rnorm(100, mean = 55, sd = 6),     # Eastern Europe
    rnorm(100, mean = 42, sd = 5),     # Mediterranean
    rnorm(80, mean = 10, sd = 10),     # West Africa
    rnorm(70, mean = -5, sd = 12)      # East Africa
  ),
  species = sample(c("Passer domesticus", "Turdus merula", "Parus major",
                     "Columba palumbus", "Sturnus vulgaris"), n_obs, replace = TRUE)
)

# Clip to valid ranges
birds$lon <- pmax(-30, pmin(60, birds$lon))
birds$lat <- pmax(-35, pmin(70, birds$lat))
```

### Assign Points to Grid Cells

```{r bird-assign}
# Create grid and assign observations (coarser grid for faster build)
grid <- hex_grid(area_km2 = 50000)
birds_hex <- hexify(birds, lon = "lon", lat = "lat", grid = grid)

# Extract data with cell IDs
birds_gridded <- as.data.frame(birds_hex)
birds_gridded$cell_id <- birds_hex@cell_id

# Count observations per cell
obs_counts <- aggregate(
  species ~ cell_id,
  data = birds_gridded,
  FUN = length
)
names(obs_counts)[2] <- "n_observations"

# Species richness per cell
richness <- aggregate(
  species ~ cell_id,
  data = birds_gridded,
  FUN = function(x) length(unique(x))
)
names(richness)[2] <- "n_species"

obs_counts <- merge(obs_counts, richness, by = "cell_id")
head(obs_counts)
```

### Generate Cell Polygons and Visualize

```{r bird-plot, fig.width=8, fig.height=6, message=FALSE, warning=FALSE}
# Generate polygons for cells with data
cell_polys <- cell_to_sf(obs_counts$cell_id, grid)
cell_polys <- merge(cell_polys, obs_counts, by = "cell_id")

# Get relevant countries
region <- hexify_world[hexify_world$continent %in% c("Europe", "Africa"), ]

ggplot() +
  geom_sf(data = region, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = n_observations), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "plasma", name = "Observations", trans = "sqrt") +
  coord_sf(xlim = c(-30, 60), ylim = c(-35, 70)) +
  labs(
    title = "Bird Observations in Equal-Area Hexagonal Cells",
    subtitle = sprintf("ISEA3H grid at resolution %d (~%.0f km² cells)",
                       grid@resolution, grid@area_km2)
  ) +
  theme_minimal() +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )
```

### Species Richness Map

```{r richness-plot, fig.width=8, fig.height=6, message=FALSE}
ggplot() +
  geom_sf(data = region, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = n_species), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "mako", name = "Species\nRichness", direction = -1) +
  coord_sf(xlim = c(-30, 60), ylim = c(-35, 70)) +
  labs(
    title = "Species Richness per Grid Cell",
    subtitle = "Number of unique species observed in each equal-area cell"
  ) +
  theme_minimal() +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )
```

## Core Concepts

hexify uses two S4 classes to make spatial workflows clean and error-free:

- **`HexGridInfo`**: A grid specification that stores all parameters (aperture, resolution, area). Define once, reuse everywhere.

- **`HexData`**: Your data + the grid that was used. Carries the grid reference so downstream operations "just work."

### HexGridInfo Slots

| Slot | Type | Description |
|------|------|-------------|
| `aperture` | character | Grid aperture ("3", "4", "7", or "4/3") |
| `resolution` | integer | Resolution level (0-30) |
| `area_km2` | numeric | Cell area in km² |
| `diagonal_km` | numeric | Cell diagonal in km |
| `crs` | integer | Coordinate reference system (EPSG code) |

### HexData Slots

| Slot | Type | Description |
|------|------|-------------|
| `data` | data.frame/sf | Your original data (unchanged) |
| `grid` | HexGridInfo | The grid specification used |
| `cell_id` | numeric | Cell ID for each row |
| `cell_center` | matrix | Cell center coordinates (lon, lat) |

## Caveats

### Pentagon Cells

At every resolution, the ISEA grid contains **12 pentagonal cells** with area 5/6 that of hexagons. These are located at the icosahedron vertices (poles + 10 evenly-spaced low-latitude points).

For most analyses, pentagons are not a concern:

1. They're a tiny minority (12 out of millions of cells at high resolutions)

2. They're in predictable locations

3. The area difference (5/6) is small and can be corrected if needed

### Multi-Scale Analysis

Hexagonal grids do **not** nest perfectly—cells at one resolution partially overlap cells at other resolutions. For hierarchical analysis, use grid-based aggregation rather than spatial containment.

### Integer Limits

Cell IDs are stored as integers. For resolutions above 15 (aperture 3), cell IDs may exceed R's integer limit (2^31-1). Use appropriate numeric types for very high resolutions.

## See Also

- `vignette("visualization")` - Plotting options: `plot()`, `hexify_heatmap()`

- `vignette("workflows")` - Grid generation, multi-resolution analysis, spatial joins, choosing resolution

- `vignette("theory")` - Mathematical foundations (ISEA projection, apertures, space-filling curves)

---
title: "Visualization"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Visualization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
library(hexify)
library(sf)
library(ggplot2)
```

This vignette covers hexify's visualization functions in detail: customizing appearance, showing points, creating heatmaps, and working with ggplot2.

## Sample Data

We'll use European cities throughout:

```{r sample-data}
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome",
           "Warsaw", "Prague", "Brussels", "Amsterdam", "Lisbon"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50,
          21.01, 14.42, 4.35, 4.90, -9.14),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90,
          52.23, 50.08, 50.85, 52.37, 38.72)
)

grid <- hex_grid(area_km2 = 10000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)
```

## Base R Plotting with plot()

The `plot()` method provides quick visualization with sensible defaults.

### Basic Plot

```{r plot-basic, fig.width=7, fig.height=5}
plot(result, main = "European Cities")
```

### Customizing Colors and Styling

```{r plot-custom, fig.width=7, fig.height=5}
plot(result,
     grid_fill = "steelblue",
     grid_border = "darkblue",
     grid_alpha = 0.6,
     basemap_fill = "ivory",
     basemap_border = "gray50",
     main = "Custom Styling")
```

### Disabling the Basemap

```{r plot-no-basemap, fig.width=7, fig.height=5}
plot(result,
     basemap = FALSE,
     grid_fill = "forestgreen",
     grid_border = "darkgreen",
     main = "No Basemap")
```

### Setting Map Extent

To control the map extent, use `crop = TRUE` with `crop_expand` to add padding:

```{r plot-extent, fig.width=7, fig.height=5}
plot(result,
     crop = TRUE,
     crop_expand = 0.2,
     main = "Custom Extent (20% padding)")
```

## Showing Points

Points can be displayed on top of hexagon cells. By default, points are jittered within their assigned cell to avoid overplotting.

### Basic Points

```{r plot-points, fig.width=7, fig.height=5}
plot(result,
     show_points = TRUE,
     point_color = "red",
     main = "Cities with Points")
```

### Point Size Presets

The `point_size` parameter accepts presets that define what fraction of a hex cell a single point covers:

| Preset | Coverage |
|--------|----------|
| `"tiny"` | ~2% of cell |
| `"small"` | ~5% of cell |
| `"normal"` / `"auto"` | ~10% of cell |
| `"large"` | ~20% of cell |
| `"very large"` | ~35% of cell |

```{r point-sizes, fig.width=7, fig.height=7}
oldpar <- par(mfrow = c(2, 2))

plot(result, show_points = TRUE, point_size = "small",
     point_color = "red", main = "small (~5%)")
plot(result, show_points = TRUE, point_size = "normal",
     point_color = "red", main = "normal (~10%)")
plot(result, show_points = TRUE, point_size = "large",
     point_color = "red", main = "large (~20%)")
plot(result, show_points = TRUE, point_size = "very large",
     point_color = "red", main = "very large (~35%)")

par(oldpar)
```

### Custom Point Styling

```{r plot-points-custom, fig.width=7, fig.height=5}
plot(result,
     show_points = TRUE,
     point_size = "small",
     point_color = "darkblue",
     point_alpha = 0.8,
     grid_fill = "lightyellow",
     grid_border = "orange",
     main = "Custom Point Styling")
```

### Disabling Jitter

For exact point locations (no randomization):

```{r plot-no-jitter, fig.width=7, fig.height=5}
plot(result,
     show_points = TRUE,
     jitter = FALSE,
     point_color = "red",
     main = "Points at Cell Centers (No Jitter)")
```

## ggplot2 with hexify_heatmap()

For ggplot2 users, `hexify_heatmap()` returns a ggplot object that can be further customized.

### Basic ggplot

```{r ggplot-basic, fig.width=7, fig.height=5}
hexify_heatmap(result, basemap = "world", title = "European Cities")
```

### Customizing with ggplot2

Since `hexify_heatmap()` returns a ggplot object, you can add layers and modify themes:

```{r ggplot-custom, fig.width=7, fig.height=5}
hexify_heatmap(result, basemap = "world") +
  labs(
    title = "Major European Cities",
    subtitle = "Assigned to ISEA hexagonal grid cells",
    caption = "Data: Sample cities"
  ) +
  theme_minimal() +
  theme(
    plot.title = element_text(face = "bold", size = 14),
    panel.grid = element_blank()
  )
```

### Adding Custom Layers

```{r ggplot-layers, fig.width=7, fig.height=5}
# Get the city coordinates
city_points <- st_as_sf(cities, coords = c("lon", "lat"), crs = 4326)

hexify_heatmap(result, basemap = "world", title = "Cities with Labels") +
  geom_sf(data = city_points, color = "red", size = 2) +
  geom_sf_text(data = city_points, aes(label = name),
               nudge_y = 0.8, size = 3, color = "darkgray") +
  coord_sf(xlim = c(-5, 25), ylim = c(45, 55))
```

## Heatmaps with Value Mapping

For choropleth-style visualizations with aggregated data, pass a `value` column to `hexify_heatmap()`.

### Creating Aggregated Data

```{r heatmap-setup}
# Simulate observation data with counts
set.seed(42)
n_obs <- 100
obs_data <- data.frame(
  lon = c(rnorm(60, 10, 8), rnorm(40, 0, 10)),
  lat = c(rnorm(60, 48, 5), rnorm(40, 52, 6)),
  count = rpois(n_obs, lambda = 50)
)

# Hexify
grid <- hex_grid(area_km2 = 10000)
obs_hex <- hexify(obs_data, lon = "lon", lat = "lat", grid = grid)
```

### Basic Heatmap

```{r heatmap-basic, fig.width=8, fig.height=6}
hexify_heatmap(
  obs_hex,
  value = "count",
  title = "Observation Counts"
)
```

### Customizing Colors

```{r heatmap-colors, fig.width=8, fig.height=6}
hexify_heatmap(
  obs_hex,
  value = "count",
  colors = "YlOrRd",
  title = "Yellow-Orange-Red Palette"
)
```

Available color palettes include any from `scale_fill_viridis_c()` or `scale_fill_distiller()`:

- Viridis options: `"viridis"`, `"magma"`, `"plasma"`, `"inferno"`, `"cividis"`, `"mako"`, `"rocket"`, `"turbo"`

- ColorBrewer sequential: `"YlOrRd"`, `"YlGnBu"`, `"Blues"`, `"Greens"`, `"Reds"`, `"Purples"`, etc.

```{r heatmap-palettes, fig.width=8, fig.height=4}
p1 <- hexify_heatmap(obs_hex, value = "count", colors = "viridis",
                     title = "viridis", xlim = c(-20, 35), ylim = c(35, 65))
p2 <- hexify_heatmap(obs_hex, value = "count", colors = "YlGnBu",
                     title = "YlGnBu", xlim = c(-20, 35), ylim = c(35, 65))

gridExtra::grid.arrange(p1, p2, ncol = 2)
```

### Setting Map Extent

```{r heatmap-extent, fig.width=8, fig.height=6}
hexify_heatmap(
  obs_hex,
  value = "count",
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "Zoomed to Region",
  legend_title = "Count"
)
```

### Basemap Options

```{r heatmap-basemap, fig.width=8, fig.height=6}
# With world basemap (default)
hexify_heatmap(
  obs_hex,
  value = "count",
  basemap = "world",
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "With World Basemap"
)
```

```{r heatmap-no-basemap, fig.width=8, fig.height=6}
# Without basemap
hexify_heatmap(
  obs_hex,
  value = "count",
  basemap = NULL,
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "No Basemap"
)
```

## World Map Helper

`plot_world()` provides a quick way to draw a world basemap:

```{r plot-world, fig.width=8, fig.height=4}
plot_world(fill = "lightgray", border = "gray50")
```

### Customizing the World Map

```{r plot-world-custom, fig.width=8, fig.height=4}
plot_world(
  fill = "antiquewhite",
  border = "sienna",
  xlim = c(-30, 50),
  ylim = c(30, 70)
)
```

## Pentagon Cell Visualization

The ISEA grid contains 12 pentagonal cells at the icosahedron vertices. Here's how to visualize them:

```{r pentagon-locations, fig.width=8, fig.height=4}
# Pentagon locations (icosahedron vertices in standard ISEA orientation)
pentagon_coords <- data.frame(
  type = c("Pole", "Pole", rep("Vertex", 10)),
  lon = c(0, 0, seq(0, 324, by = 36)),
  lat = c(90, -90, rep(c(26.57, -26.57), 5))
)

# Assign to grid and get polygons
grid <- hex_grid(area_km2 = 500000)
pentagon_cells <- lonlat_to_cell(pentagon_coords$lon, pentagon_coords$lat, grid)

pentagon_polys <- cell_to_sf(pentagon_cells, grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = pentagon_polys, fill = alpha("purple", 0.6),
          color = "purple", linewidth = 0.8) +
  labs(
    title = "Pentagon Cell Locations",
    subtitle = "12 pentagonal cells at icosahedron vertices (area = 5/6 of hexagons)"
  ) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## Random Sampling Visualization

Visualizing uniformly sampled cells across Earth:

```{r random-sample, fig.width=8, fig.height=4}
# Grid parameters (coarse for faster build)
grid <- hex_grid(area_km2 = 200000, aperture = 3)
max_cell <- 10 * (3^grid@resolution) + 2

# Sample random cell IDs
set.seed(123)
N <- 50
random_cells <- sample(1:max_cell, N, replace = FALSE)

# Generate polygons for sampled cells
sample_polys <- cell_to_sf(random_cells, grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = sample_polys, fill = alpha("forestgreen", 0.5),
          color = "darkgreen", linewidth = 0.4) +
  labs(title = sprintf("Random Sample of %d Cells (~%.0f km2 each)", N, grid@area_km2)) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## Building Custom Visualizations

For full control, generate polygons directly and use ggplot2:

```{r custom-viz, fig.width=8, fig.height=6}
# Create data with a numeric variable
set.seed(456)
stations <- data.frame(
  lon = runif(50, -10, 30),
  lat = runif(50, 35, 60),
  temperature = rnorm(50, mean = 15, sd = 5)
)

# Hexify (coarser grid for faster build)
grid <- hex_grid(area_km2 = 20000)
stations_hex <- hexify(stations, lon = "lon", lat = "lat", grid = grid)

# Aggregate temperature by cell
stations_df <- as.data.frame(stations_hex)
stations_df$cell_id <- stations_hex@cell_id
cell_temps <- aggregate(temperature ~ cell_id, data = stations_df, FUN = mean)

# Generate polygons and merge data
cell_polys <- cell_to_sf(cell_temps$cell_id, grid)
cell_polys <- merge(cell_polys, cell_temps, by = "cell_id")

# Custom visualization
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = temperature),
          color = "white", linewidth = 0.2) +
  scale_fill_gradient2(
    low = "blue", mid = "white", high = "red",
    midpoint = 15, name = "Temp (°C)"
  ) +
  coord_sf(xlim = c(-10, 30), ylim = c(35, 60)) +
  labs(
    title = "Mean Temperature by Grid Cell",
    subtitle = "Diverging color scale centered at 15°C"
  ) +
  theme_minimal() +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )
```

## Function Reference

| Function | Description |
|----------|-------------|
| `plot()` | Base R plot method for HexData objects |
| `hexify_heatmap()` | ggplot2 visualization, returns modifiable ggplot object |
| `plot_world()` | Quick world basemap |
| `cell_to_sf()` | Generate sf polygons from cell IDs |

## See Also

- `vignette("quickstart")` - Getting started with hexify

- `vignette("workflows")` - Complete analysis workflows

---
title: "Practical Workflows"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Practical Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
library(hexify)
library(sf)
library(ggplot2)
```

This vignette covers practical workflows for common spatial analysis tasks: sharing grids across datasets, generating grid polygons, choosing resolution, and exporting to GIS formats.

## One Grid, Many Datasets

The most powerful pattern in hexify is defining a grid once and reusing it across multiple datasets. This ensures spatial consistency and eliminates parameter repetition.

### The Problem

You often have:

- Several independent datasets (observations, sensors, surveys)

- All in longitude/latitude coordinates

- Collected at different times or from different sources

You want to:

- Put everything on one common global grid

- Be sure the grids actually match

- Combine results later without subtle errors

### The Solution: Shared Grid Objects

```{r one-grid-many-datasets}
# Step 1: Define the grid once
# This is your shared spatial reference system
grid <- hex_grid(area_km2 = 5000)
print(grid)

# Step 2: Create multiple datasets with different structures
set.seed(123)

# Dataset 1: Bird observations (note different column names)
bird_obs <- data.frame(
  species = sample(c("Passer domesticus", "Turdus merula", "Parus major"), 50, replace = TRUE),
  longitude = runif(50, -10, 30),
  latitude = runif(50, 35, 60)
)

# Dataset 2: Mammal records
mammal_obs <- data.frame(
  species = sample(c("Vulpes vulpes", "Meles meles", "Sciurus vulgaris"), 40, replace = TRUE),
  lon = runif(40, -10, 30),
  lat = runif(40, 35, 60)
)

# Dataset 3: Climate stations
climate_data <- data.frame(
  station_id = paste0("WS", 1:20),
  x = runif(20, -10, 30),
  y = runif(20, 35, 60),
  temp_c = rnorm(20, 12, 5)
)

# Step 3: Attach all datasets to the SAME grid
birds <- hexify(bird_obs, lon = "longitude", lat = "latitude", grid = grid)
mammals <- hexify(mammal_obs, lon = "lon", lat = "lat", grid = grid)
climate <- hexify(climate_data, lon = "x", lat = "y", grid = grid)

cat("Birds:  ", nrow(as.data.frame(birds)), "observations in", n_cells(birds), "cells\n")
cat("Mammals:", nrow(as.data.frame(mammals)), "observations in", n_cells(mammals), "cells\n")
cat("Climate:", nrow(as.data.frame(climate)), "stations in", n_cells(climate), "cells\n")
```

### Combining at the Cell Level

Once data are hexified, longitude/latitude no longer matter for analysis. The `cell_id` becomes the shared spatial key:

```{r cell-level-analysis}
# Extract data frames with cell IDs
birds_df <- as.data.frame(birds)
birds_df$cell_id <- birds@cell_id

mammals_df <- as.data.frame(mammals)
mammals_df$cell_id <- mammals@cell_id

climate_df <- as.data.frame(climate)
climate_df$cell_id <- climate@cell_id

# Aggregate each dataset by cell
bird_richness <- aggregate(
  species ~ cell_id,
  data = birds_df,
  FUN = function(x) length(unique(x))
)
names(bird_richness)[2] <- "bird_species"

mammal_richness <- aggregate(
  species ~ cell_id,
  data = mammals_df,
  FUN = function(x) length(unique(x))
)
names(mammal_richness)[2] <- "mammal_species"

mean_temp <- aggregate(
  temp_c ~ cell_id,
  data = climate_df,
  FUN = mean
)
names(mean_temp)[2] <- "mean_temp"

# Join datasets by cell_id - guaranteed to align because same grid
combined <- merge(bird_richness, mammal_richness, by = "cell_id", all = TRUE)
combined <- merge(combined, mean_temp, by = "cell_id", all = TRUE)

head(combined)
```

## Grid Generation

hexify provides functions to generate grid polygons over regions for visualization and analysis.

### Grid Over a Rectangular Region

```{r grid-rect, fig.width=7, fig.height=5}
# Create grid specification
grid <- hex_grid(area_km2 = 5000)

# Generate hexagons over Western Europe
europe_hexes <- grid_rect(c(-10, 35, 25, 60), grid)

# Get European countries for context
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_hexes, fill = NA, color = "steelblue", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal()
```

### Grid Over a Polygon (Shapefile)

Clip a grid to any sf polygon boundary:

```{r grid-polygon, fig.width=6, fig.height=6}
# Get France boundary
france <- hexify_world[hexify_world$name == "France", ]

# Generate grid covering mainland France
grid <- hex_grid(area_km2 = 2000)
france_grid <- grid_rect(c(-5, 41, 10, 52), grid)

# Clip grid to France boundary
france_grid_clipped <- st_intersection(france_grid, st_geometry(france))

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_grid_clipped, fill = alpha("steelblue", 0.3),
          color = "steelblue", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = sprintf("Hexagonal Grid Clipped to France (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal()
```

### Global Grid

```{r grid-global, fig.width=8, fig.height=4, warning=FALSE}
# Coarse global grid (be careful with fine grids - many cells!)
grid <- hex_grid(area_km2 = 500000)
global_hexes <- grid_global(grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray90", color = "gray70", linewidth = 0.2) +
  geom_sf(data = global_hexes, fill = NA, color = "darkgreen", linewidth = 0.3) +
  labs(title = sprintf("Global Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal() +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## Multi-Resolution Analysis

Analyze data at multiple spatial scales using different target areas.

```{r multi-resolution}
# Sample data
set.seed(42)
observations <- data.frame(
  species = sample(c("Species A", "Species B", "Species C"), 100, replace = TRUE),
  lon = runif(100, -10, 30),
  lat = runif(100, 35, 60)
)

# Fine resolution (~1000 km² cells)
grid_fine <- hex_grid(area_km2 = 1000)
obs_fine <- hexify(observations, lon = "lon", lat = "lat", grid = grid_fine)

# Coarse resolution (~10000 km² cells)
grid_coarse <- hex_grid(area_km2 = 10000)
obs_coarse <- hexify(observations, lon = "lon", lat = "lat", grid = grid_coarse)

cat(sprintf("Fine resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_fine), grid_fine@area_km2))
cat(sprintf("Coarse resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_coarse), grid_coarse@area_km2))
```

### Scale-Dependent Patterns

```{r multi-scale-aggregate}
# Extract data with cell IDs
fine_df <- as.data.frame(obs_fine)
fine_df$cell_id <- obs_fine@cell_id

coarse_df <- as.data.frame(obs_coarse)
coarse_df$cell_id <- obs_coarse@cell_id

# Species richness at each scale
richness_fine <- aggregate(species ~ cell_id, data = fine_df,
                           FUN = function(x) length(unique(x)))
richness_coarse <- aggregate(species ~ cell_id, data = coarse_df,
                             FUN = function(x) length(unique(x)))

cat(sprintf("Fine scale: mean %.2f species per cell\n", mean(richness_fine$species)))
cat(sprintf("Coarse scale: mean %.2f species per cell\n", mean(richness_coarse$species)))
```

## Spatial Joins

Join datasets based on shared grid cells rather than proximity.

```{r spatial-joins}
# Dataset 1: Weather stations
stations <- data.frame(
  station_id = paste0("ST", 1:50),
  lon = runif(50, -10, 30),
  lat = runif(50, 35, 60),
  temperature = rnorm(50, 15, 5)
)

# Dataset 2: Cities
cities <- data.frame(
  city = c("Vienna", "Paris", "London", "Berlin", "Rome",
           "Madrid", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -0.12, 13.40, 12.50,
          -3.70, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 51.51, 52.52, 41.90,
          40.42, 50.08, 52.23, 47.50, 52.37)
)

# Use a coarse grid for joining disparate points
grid <- hex_grid(area_km2 = 50000)

# Hexify both datasets with the same grid
stations_hex <- hexify(stations, lon = "lon", lat = "lat", grid = grid)
cities_hex <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Extract with cell IDs
stations_df <- as.data.frame(stations_hex)
stations_df$cell_id <- stations_hex@cell_id

cities_df <- as.data.frame(cities_hex)
cities_df$cell_id <- cities_hex@cell_id

# Join by cell_id
city_weather <- merge(
  cities_df[, c("city", "cell_id")],
  aggregate(temperature ~ cell_id, data = stations_df, FUN = mean),
  by = "cell_id",
  all.x = TRUE
)

city_weather
```

## Choosing Resolution

### By Target Area

Use `hex_grid()` with `area_km2` to get the closest available resolution:

```{r choosing-resolution}
# Target: 100 km² cells
grid_100 <- hex_grid(area_km2 = 100, aperture = 3)
cat(sprintf("Target ~100 km²: resolution %d (actual: %.1f km²)\n",
            grid_100@resolution, grid_100@area_km2))

# Target: 1000 km² cells
grid_1000 <- hex_grid(area_km2 = 1000, aperture = 3)
cat(sprintf("Target ~1000 km²: resolution %d (actual: %.1f km²)\n",
            grid_1000@resolution, grid_1000@area_km2))

# Target: 10000 km² cells
grid_10000 <- hex_grid(area_km2 = 10000, aperture = 3)
cat(sprintf("Target ~10000 km²: resolution %d (actual: %.1f km²)\n",
            grid_10000@resolution, grid_10000@area_km2))
```

### Resolution Table (Aperture 3)

```{r resolution-table, echo=FALSE}
comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:15)
comparison$n_cells_fmt <- ifelse(
  comparison$n_cells > 1e6,
  sprintf("%.1fM", comparison$n_cells / 1e6),
  ifelse(comparison$n_cells > 1e3,
         sprintf("%.1fK", comparison$n_cells / 1e3),
         as.character(comparison$n_cells))
)
knitr::kable(
  comparison[, c("resolution", "n_cells_fmt", "cell_area_km2", "cell_spacing_km")],
  col.names = c("Resolution", "# Cells", "Cell Area (km²)", "Spacing (km)"),
  digits = 1
)
```

### Comparing Apertures

Different apertures offer different resolution steps:

```{r compare-apertures}
target_area <- 1000  # km²

cat(sprintf("Target: ~%d km² cells\n\n", target_area))
for (ap in c(3, 4, 7)) {
  grid <- hex_grid(area_km2 = target_area, aperture = ap)
  n_cells <- 10 * (ap^grid@resolution) + 2
  cat(sprintf("Aperture %d: resolution %d -> %.1f km² (%s cells globally)\n",
              ap, grid@resolution, grid@area_km2,
              format(n_cells, big.mark = ",")))
}
```

| Aperture | Best For | Trade-offs |
|----------|----------|------------|
| 3 | Fine resolution control, dggridR compatibility | Slowest cell growth |
| 4 | Power-of-2 scaling, GIS workflows | Moderate resolution steps |
| 7 | Rapid cell count growth, coarse analysis | Largest resolution jumps |
| 4/3 | Balance of 4's fast start + 3's fine control | More complex indexing |

## Working with sf

### Convert HexData to sf

```{r sf-workflow}
# Hexify some data
grid <- hex_grid(area_km2 = 20000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Convert to sf points (uses cell centers)
sf_points <- as_sf(result, geometry = "point")
class(sf_points)

# Convert to sf polygons (for choropleth maps)
sf_polys <- as_sf(result, geometry = "polygon")
class(sf_polys)

# Or generate polygons directly from cell IDs
unique_cells <- cells(result)
cell_polys <- cell_to_sf(unique_cells, grid)
```

### Visualize with ggplot2

```{r sf-plot, fig.width=7, fig.height=5}
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "ivory", color = "gray70") +
  geom_sf(data = cell_polys, fill = "steelblue", alpha = 0.5, color = "darkblue") +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 58)) +
  labs(title = "European Cities - Hexagonal Grid") +
  theme_minimal()
```

## Export to GIS Formats

Use sf's `st_write()` to export grids for use in external GIS software:

```{r export-formats, eval=FALSE}
# Generate a grid
grid <- hex_grid(area_km2 = 10000)
europe <- grid_rect(c(-10, 35, 25, 60), grid)

# Export to various formats
st_write(europe, "europe_grid.gpkg", layer = "hexgrid")     # GeoPackage
st_write(europe, "europe_grid.shp")                         # Shapefile
st_write(europe, "europe_grid.geojson")                     # GeoJSON
st_write(europe, "europe_grid.kml", layer = "hexgrid")      # KML (Google Earth)
```

## Edge Cases

### Handling NA Coordinates

```{r edge-cases}
data_with_na <- data.frame(
  lon = c(16.37, NA, 2.35, 13.40),
  lat = c(48.21, 48.86, NA, 52.52)
)

grid <- hex_grid(area_km2 = 1000)
result <- hexify(data_with_na, lon = "lon", lat = "lat", grid = grid)

# Check which rows have valid cell assignments
cat("Cell IDs:", result@cell_id, "\n")
cat("NA indicates invalid coordinates\n")
```

### Polar Regions

Coordinates with latitude > 89° may have projection artifacts. The grid remains valid, but polygon visualization can be distorted near poles.

### Date Line

Polygons crossing the date line (lon = ±180°) are handled automatically.
`cell_to_sf()` applies `sf::st_wrap_dateline()` internally, so flat map
projections render correctly without manual intervention.

## Function Summary

| Task | Function |
|------|----------|
| Create grid specification | `hex_grid(area_km2 = ...)` |
| Assign points to cells | `hexify(df, lon, lat, grid)` |
| Get grid from HexData | `grid_info(result)` |
| Get unique cell IDs | `cells(result)` |
| Count cells | `n_cells(result)` |
| Extract data frame | `as.data.frame(result)` |
| Convert to sf | `as_sf(result, geometry = "polygon")` |
| Generate polygons | `cell_to_sf(cell_ids, grid)` |
| Grid over region | `grid_rect(bbox, grid)` |
| Global grid | `grid_global(grid)` |
| Coordinate conversion | `lonlat_to_cell()`, `cell_to_lonlat()` |
| Compare resolutions | `hexify_compare_resolutions()` |

## See Also

- `vignette("quickstart")` - Getting started with hexify

- `vignette("visualization")` - Plotting with `plot()`, `hexify_heatmap()`

- `vignette("theory")` - Mathematical foundations (ISEA projection, apertures)

---
title: "Quick Start"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Quick Start}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
source("_common.R")
library(hexify)
```

```{r cover-image, echo=FALSE, message=FALSE, warning=FALSE, fig.width=7, fig.height=3.5, fig.align='center'}
# Generate cover image: multi-resolution grids over different regions
library(sf)
library(ggplot2)

# Get world basemap
countries <- hexify_world

# Generate grids at different resolutions for specific regions (coarser for speed)
# Coarse grid: South America
grid_sa <- hex_grid(area_km2 = 300000)
sa_hexes <- grid_rect(c(-80, -40, -35, 10), grid_sa)

# Medium grid: Europe/Africa
grid_eu <- hex_grid(area_km2 = 80000)
eu_hexes <- grid_rect(c(-10, -10, 40, 50), grid_eu)

# Fine grid: East Asia
grid_asia <- hex_grid(area_km2 = 50000)
asia_hexes <- grid_rect(c(90, 10, 140, 50), grid_asia)

ggplot() +
  geom_sf(data = countries, fill = "gray95", color = "gray80", linewidth = 0.15) +
  geom_sf(data = sa_hexes, fill = NA, color = "#1B9E77", linewidth = 0.5) +
  geom_sf(data = eu_hexes, fill = NA, color = "#D95F02", linewidth = 0.4) +
  geom_sf(data = asia_hexes, fill = NA, color = "#7570B3", linewidth = 0.3) +
  coord_sf(xlim = c(-100, 150), ylim = c(-50, 60)) +
  theme_void(base_size = FIG_BASE_SIZE) +
  theme(panel.background = element_rect(fill = "white", color = NA))
```

# Spatial Analysis Done Right

You want to do spatial statistics, and it involves binning points into grid cells.

**The problem with rectangular grids**: A rectangular lat-lon grid introduces severe distortions. At the equator, a 1° cell covers ~12,300 km². Near the poles, the same 1° cell covers a tiny fraction of that area. This breaks any analysis that assumes equal sampling effort or comparable cell sizes.

**The solution**: Discrete global grids partition Earth's surface into cells of **equal area**, regardless of latitude. hexify implements the ISEA (Icosahedral Snyder Equal Area) projection, providing hexagonal cells that are all the same size from the equator to the Arctic.

## Basic Usage

```{r basic-usage}
# Sample data: European cities
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90)
)

# Create a grid specification
grid <- hex_grid(area_km2 = 10000)
grid

# Assign cities to hexagonal cells
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)
result
```

### Accessing HexData

```{r hexdata-access}
# Get the grid specification
grid_info(result)

# Get unique cell IDs
cells(result)

# Count unique cells
n_cells(result)

# Access all cell IDs (one per row)
result@cell_id

# Access cell centers
head(result@cell_center)

# Extract original data as data.frame
head(as.data.frame(result))
```

### With sf Objects

The sf package holds spatial data as simple features: a data frame with a
geometry column of points, lines or polygons, and a CRS that says how to read
their coordinates. hexify takes an sf object wherever it takes a data frame,
and transforms the coordinates itself when the CRS differs from the grid's.

```{r sf-usage, eval = requireNamespace("sf", quietly = TRUE)}
library(sf)

# Create sf object (any CRS works - hexify transforms automatically)
pts <- st_as_sf(cities, coords = c("lon", "lat"), crs = 4326)

# hexify handles CRS transformation automatically
result_sf <- hexify(pts, area_km2 = 10000)
class(result_sf)
```

## Real-World Example: Species Occurrence Data

This example demonstrates the typical workflow: loading point data, assigning to grid cells, aggregating, and visualizing.

```{r bird-setup, message=FALSE, warning=FALSE}
library(sf)
library(ggplot2)

# Simulate bird observation data (reduced for faster vignette build)
set.seed(123)
n_obs <- 500

# Generate observations with realistic spatial clustering
birds <- data.frame(
  lon = c(
    rnorm(150, mean = 10, sd = 15),    # Western Europe
    rnorm(100, mean = 25, sd = 10),    # Eastern Europe
    rnorm(100, mean = 20, sd = 20),    # Mediterranean
    rnorm(80, mean = 0, sd = 15),      # West Africa
    rnorm(70, mean = 35, sd = 10)      # East Africa
  ),
  lat = c(
    rnorm(150, mean = 50, sd = 8),     # Western Europe
    rnorm(100, mean = 55, sd = 6),     # Eastern Europe
    rnorm(100, mean = 42, sd = 5),     # Mediterranean
    rnorm(80, mean = 10, sd = 10),     # West Africa
    rnorm(70, mean = -5, sd = 12)      # East Africa
  ),
  species = sample(c("Passer domesticus", "Turdus merula", "Parus major",
                     "Columba palumbus", "Sturnus vulgaris"), n_obs, replace = TRUE)
)

# Clip to valid ranges
birds$lon <- pmax(-30, pmin(60, birds$lon))
birds$lat <- pmax(-35, pmin(70, birds$lat))
```

### Assign Points to Grid Cells

```{r bird-assign}
# Create grid and assign observations (coarser grid for faster build)
grid <- hex_grid(area_km2 = 50000)
birds_hex <- hexify(birds, lon = "lon", lat = "lat", grid = grid)

# Extract data with cell IDs
birds_gridded <- as.data.frame(birds_hex)
birds_gridded$cell_id <- birds_hex@cell_id

# Count observations per cell
obs_counts <- aggregate(
  species ~ cell_id,
  data = birds_gridded,
  FUN = length
)
names(obs_counts)[2] <- "n_observations"

# Species richness per cell
richness <- aggregate(
  species ~ cell_id,
  data = birds_gridded,
  FUN = function(x) length(unique(x))
)
names(richness)[2] <- "n_species"

obs_counts <- merge(obs_counts, richness, by = "cell_id")
head(obs_counts)
```

### Generate Cell Polygons and Visualize

```{r bird-plot, fig.width=7, fig.height=5.25, message=FALSE, warning=FALSE}
# Generate polygons for cells with data
cell_polys <- cell_to_sf(obs_counts$cell_id, grid)
cell_polys <- merge(cell_polys, obs_counts, by = "cell_id")

# Get relevant countries
region <- hexify_world[hexify_world$continent %in% c("Europe", "Africa"), ]

ggplot() +
  geom_sf(data = region, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = n_observations), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "plasma", name = "Observations", trans = "sqrt") +
  coord_sf(xlim = c(-30, 60), ylim = c(-35, 70)) +
  labs(
    title = "Bird Observations in Equal-Area Cells",
    subtitle = sprintf("ISEA3H grid at resolution %d (~%.0f km² cells)",
                       grid@resolution, grid@area_km2)
  ) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )
```

### Species Richness Map

```{r richness-plot, fig.width=7, fig.height=5.25, message=FALSE}
ggplot() +
  geom_sf(data = region, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = n_species), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "mako", name = "Species\nRichness", direction = -1) +
  coord_sf(xlim = c(-30, 60), ylim = c(-35, 70)) +
  labs(
    title = "Species Richness per Grid Cell",
    subtitle = "Unique species per equal-area cell"
  ) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )
```

## Core Concepts

hexify uses two S4 classes to make spatial workflows clean and error-free:

- **`HexGridInfo`**: A grid specification that stores all parameters (aperture, resolution, area). Define once, reuse everywhere.

- **`HexData`**: Your data + the grid that was used. Carries the grid reference so downstream operations "just work."

### HexGridInfo Slots

| Slot | Type | Description |
|------|------|-------------|
| `aperture` | character | Grid aperture ("3", "4", "7", or "4/3") |
| `resolution` | integer | Resolution level (0-30) |
| `area_km2` | numeric | Cell area in km² |
| `diagonal_km` | numeric | Cell diagonal in km |
| `crs` | integer or character | EPSG code, or a PROJ or WKT string |
| `grid_type` | character | Backend: `"isea"` or `"h3"` |

### HexData Slots

| Slot | Type | Description |
|------|------|-------------|
| `data` | data.frame/sf | Your original data (unchanged) |
| `grid` | HexGridInfo | The grid specification used |
| `cell_id` | numeric | Cell ID for each row |
| `cell_center` | matrix | Cell center coordinates (lon, lat) |

## Caveats

### Pentagon Cells

At every resolution, the ISEA grid contains **12 pentagonal cells** with area 5/6 that of hexagons. These are located at the icosahedron vertices (poles + 10 evenly-spaced low-latitude points).

For most analyses, pentagons are not a concern:

1. They're a tiny minority (12 out of millions of cells at high resolutions)

2. They're in predictable locations

3. The area difference (5/6) is small and can be corrected if needed

### Multi-Scale Analysis

Hexagonal grids do **not** nest perfectly—cells at one resolution partially overlap cells at other resolutions. For hierarchical analysis, use grid-based aggregation rather than spatial containment.

### Integer Limits

Cell IDs are stored as integers. For resolutions above 15 (aperture 3), cell IDs may exceed R's integer limit (2^31-1). Use appropriate numeric types for very high resolutions.

## See Also

- `vignette("visualization")` - Plotting options: `plot()`, `hexify_heatmap()`

- `vignette("workflows")` - Grid generation, multi-resolution analysis, spatial joins, choosing resolution

- `vignette("theory")` - Mathematical foundations (ISEA projection, apertures, space-filling curves)

---
title: "Mathematical Foundations"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Mathematical Foundations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
source("_common.R")
library(hexify)
```

## Overview

hexify implements the **ISEA (Icosahedral Snyder Equal Area)** discrete global grid system. This vignette explains the mathematical foundations: the projection geometry, coordinate systems, aperture subdivision, and cell indexing.

## The Problem: Equal-Area Grids on a Sphere

Any projection from a sphere to a plane must distort *something*. For spatial statistics, we need cells of equal area regardless of location. Standard latitude-longitude grids fail badly: a 1° cell at the equator covers ~12,300 km², while the same cell near the poles covers a tiny fraction of that.

The ISEA projection solves this by:

1. Inscribing a regular icosahedron in the sphere

2. Projecting each spherical "cap" onto its corresponding flat triangular face using a modified Lambert equal-area projection

3. Overlaying a hexagonal grid on the resulting planar triangles

## The Lambert Azimuthal Equal-Area Projection

The foundation of Snyder's projection is Lambert's azimuthal equal-area projection, developed by Johann Heinrich Lambert in 1772. The projection maps a sphere of radius $R$ to a tangent plane while **preserving area exactly** (Snyder, 1987, p. 182).

### Definition

The Lambert azimuthal equal-area projection is defined by a mathematical constraint, not a geometric construction. For a projection centered at point $S$:

- **Azimuthal property:** The direction (azimuth) from $S$ to any point $P$ on the sphere equals the direction from the origin to $P'$ on the plane

- **Equal-area property:** Any region on the sphere maps to a region of identical area on the plane

These two constraints uniquely determine the radial distance formula. For a point $P$ at angular distance $c$ from the center (measured along the sphere surface), the projected distance from the origin is:

$$\rho = 2R \sin\left(\frac{c}{2}\right)$$

This is **not** a perspective projection and has no simple geometric interpretation like "chord distance" or "ray intersection." The formula is derived analytically from the equal-area constraint (Snyder, 1987, p. 182-185).

```{r lambert-geometry, echo=FALSE, fig.width=7, fig.height=6}
# Lambert Azimuthal Equal-Area Projection Geometry
# Reference: Snyder (1987), "Map Projections: A Working Manual", p. 182-185
#
# The projection formula ρ = 2R·sin(c/2) is derived analytically from
# the equal-area constraint, not from geometric chord distance.
# This diagram shows the geometric relationship that yields the same formula.

oldpar <- par(mar = c(1, 1, 2, 1), bg = "white")
plot(NULL, xlim = c(-1.5, 2.2), ylim = c(-1.5, 1.7), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Lambert Azimuthal Equal-Area Projection Geometry")

R <- 1
theta_circle <- seq(0, 2*pi, length.out = 100)

# Draw sphere (great circle cross-section)
lines(R * cos(theta_circle), R * sin(theta_circle), lwd = 2, col = "gray30")

# Draw tangent plane at S (top of sphere)
lines(c(-1.4, 2.1), c(R, R), lwd = 2, col = "gray50")
text(-1.3, R + 0.1, "Tangent plane", adj = 0, cex = 0.9, col = "gray40")

# Mark center O
points(0, 0, pch = 19, cex = 1.2)
text(-0.12, -0.15, "O", cex = 1.1, font = 2)

# Mark tangent point S
points(0, R, pch = 19, cex = 1.2)
text(0.12, R + 0.12, "S", cex = 1.1, font = 2)

# Choose a point P on sphere at angle φ from S
phi <- 50 * pi/180  # Angular distance from S
P_x <- R * sin(phi)
P_y <- R * cos(phi)
points(P_x, P_y, pch = 19, cex = 1.2, col = "#E63946")
text(P_x + 0.1, P_y + 0.08, "P", cex = 1.1, font = 2, col = "#E63946")

# Draw radius to P
lines(c(0, P_x), c(0, P_y), lwd = 1.5, lty = 2, col = "gray50")

# Project P to tangent plane (perpendicular projection gives P')
# In Lambert: P' is at distance ρ = 2R·sin(φ/2) from S
rho <- 2 * R * sin(phi/2)
Pprime_x <- rho
Pprime_y <- R
points(Pprime_x, Pprime_y, pch = 19, cex = 1.2, col = "#457B9D")
text(Pprime_x + 0.1, Pprime_y + 0.12, "P'", cex = 1.1, font = 2, col = "#457B9D")

# Draw chord from S to P (this has length 2R·sin(φ/2))
lines(c(0, P_x), c(R, P_y), lwd = 2.5, col = "#E63946")

# Draw projected distance on tangent plane
lines(c(0, Pprime_x), c(R, R), lwd = 2.5, col = "#457B9D")

# Draw arc showing angle φ at center
arc_r <- 0.35
arc_theta <- seq(pi/2, pi/2 - phi, length.out = 30)
lines(arc_r * cos(arc_theta), arc_r * sin(arc_theta), lwd = 2, col = "#2A9D8F")
text(arc_r * 1.6, arc_r * 1.2, expression(italic(c)), cex = 1.2, col = "#2A9D8F")

# Label the chord d
mid_chord_x <- (0 + P_x)/2 - 0.12
mid_chord_y <- (R + P_y)/2 + 0.08
text(mid_chord_x, mid_chord_y, expression(italic(d)), cex = 1.2, col = "#E63946")

# Label the projected distance ρ
text(Pprime_x/2, R + 0.15, expression(rho), cex = 1.2, col = "#457B9D")

# Add formula box
rect(0.8, -1.3, 2.15, -0.7, col = "white", border = "gray70")
text(1.47, -0.85, expression(italic(d) == 2*italic(R)*sin(italic(c)/2)), cex = 1.1)
text(1.47, -1.15, expression(rho == italic(d)), cex = 1.1)

# Add reference note
text(1.47, -1.45, "Snyder (1987, p. 182)", cex = 0.9, col = "gray50")
par(oldpar)
```

### Forward Formulas

For the oblique aspect centered at $(\lambda_0, \phi_1)$, the full formulas are (Snyder, 1987, eq. 24-2 to 24-4, p. 185):

$$k' = \sqrt{\frac{2}{1 + \sin\phi_1\sin\phi + \cos\phi_1\cos\phi\cos(\lambda - \lambda_0)}}$$

$$x = R \cdot k' \cdot \cos\phi \cdot \sin(\lambda - \lambda_0)$$

$$y = R \cdot k' \cdot [\cos\phi_1\sin\phi - \sin\phi_1\cos\phi\cos(\lambda - \lambda_0)]$$

### Equal-Area Proof

The projection preserves area because the radial and tangential scale factors satisfy $h' \cdot k' = 1$ at every point. At angular distance $c$ from center (Snyder, 1987, eq. 24-22, 24-23, p. 188):

$$h' = \cos\left(\frac{c}{2}\right), \quad k' = \sec\left(\frac{c}{2}\right)$$

Therefore $h' \cdot k' = \cos(c/2) \cdot \sec(c/2) = 1$, confirming the equal-area property.

```{r lambert-area-preservation, echo=FALSE, fig.width=7, fig.height=3.5}
# Show equal-area property with concentric rings
oldpar <- par(mfrow = c(1, 2), mar = c(2, 1, 3, 1), bg = "white", cex = 1, cex.main = 1)

R <- 1

# Left: Sphere view (orthographic, looking from above at tangent point)
plot(NULL, xlim = c(-1.3, 1.3), ylim = c(-1.3, 1.3), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Sphere (view from above S)")

# Draw outer circle (equator as seen from S)
theta <- seq(0, 2*pi, length.out = 100)
lines(cos(theta), sin(theta), lwd = 2)

# Draw concentric latitude circles and shade bands
n_bands <- 5
cols <- c("#264653", "#2A9D8F", "#E9C46A", "#F4A261", "#E76F51")

for (i in n_bands:1) {
  phi_outer <- (i / n_bands) * (pi/2)
  phi_inner <- ((i-1) / n_bands) * (pi/2)

  r_outer <- sin(phi_outer)
  r_inner <- sin(phi_inner)

  theta_seq <- seq(0, 2*pi, length.out = 100)
  if (i == 1) {
    polygon(r_outer * cos(theta_seq), r_outer * sin(theta_seq),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  } else {
    polygon(c(r_outer * cos(theta_seq), rev(r_inner * cos(theta_seq))),
            c(r_outer * sin(theta_seq), rev(r_inner * sin(theta_seq))),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  }
}
points(0, 0, pch = 19, cex = 1.2)
text(0, -1.2, "Equal-area bands on sphere", cex = 0.9)

# Right: Lambert projection
plot(NULL, xlim = c(-1.6, 1.6), ylim = c(-1.6, 1.6), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Lambert Projection")

for (i in n_bands:1) {
  phi_outer <- (i / n_bands) * (pi/2)
  phi_inner <- ((i-1) / n_bands) * (pi/2)

  # Lambert projection: r = 2*sin(phi/2)
  r_outer <- 2 * sin(phi_outer / 2)
  r_inner <- 2 * sin(phi_inner / 2)

  theta_seq <- seq(0, 2*pi, length.out = 100)
  if (i == 1) {
    polygon(r_outer * cos(theta_seq), r_outer * sin(theta_seq),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  } else {
    polygon(c(r_outer * cos(theta_seq), rev(r_inner * cos(theta_seq))),
            c(r_outer * sin(theta_seq), rev(r_inner * sin(theta_seq))),
            col = adjustcolor(cols[i], 0.5), border = "gray40")
  }
}

r_eq <- 2 * sin(pi/4)
lines(r_eq * cos(theta), r_eq * sin(theta), lwd = 2, lty = 2, col = "gray50")

points(0, 0, pch = 19, cex = 1.2)
text(0, -1.35, "Same bands after projection\n(areas preserved)", cex = 0.9)
par(oldpar)
```

Each colored band has equal area on the sphere. After Lambert projection, shapes change (outer bands stretch radially, compress tangentially) but areas remain equal.

### Inverse Formulas

The inverse mapping $(x, y) \to (\lambda, \phi)$ recovers geographic coordinates from planar coordinates (Snyder, 1987, eq. 24-14 to 24-16, p. 187):

$$\rho = \sqrt{x^2 + y^2}$$

$$c = 2\arcsin\left(\frac{\rho}{2R}\right)$$

where $c$ is the angular distance from the projection center. Then:

$$\phi = \arcsin\left[\cos c \cdot \sin\phi_1 + \frac{y \cdot \sin c \cdot \cos\phi_1}{\rho}\right]$$

$$\lambda = \lambda_0 + \arctan\left[\frac{x \cdot \sin c}{\rho\cos\phi_1\cos c - y\sin\phi_1\sin c}\right]$$

If $\rho = 0$, the point is at the projection center: $\phi = \phi_1$, $\lambda = \lambda_0$.

### Limitations

**Antipode singularity:** The point diametrically opposite the projection center (angular distance 180°) maps to infinity and must be excluded from the domain.

**Shape distortion:** While area is preserved exactly, shapes distort increasingly with distance from center. The maximum angular distortion $\omega$ at distance $c$ is (Snyder, 1987, eq. 24-24, p. 188):

$$\sin\left(\frac{\omega}{2}\right) = \frac{k'^2 - 1}{k'^2 + 1}$$

At $c = 90°$, distortion reaches $\omega \approx 70.5°$. ISEA grids limit this by using icosahedral faces subtending only ~72° from face center.

**No conformality:** No projection can be both equal-area and conformal (angle-preserving)---a fundamental constraint from differential geometry (Snyder, 1987, p. 16-18).

## The Icosahedron

Lambert's projection works for a single tangent plane covering at most a hemisphere. To cover the entire globe with minimal distortion, Snyder used **20 tangent planes**---one for each face of a regular icosahedron (Snyder, 1992, p. 10).

### Geometry

A regular icosahedron has:

- **20 equilateral triangular faces** (each covering ~1/20 of Earth's surface)

- **12 vertices** (where 5 faces meet---these become pentagon cells)

- **30 edges**

```{r icosahedron-projection, echo=FALSE, fig.width=7, fig.height=6}
# Icosahedron Face Projection Geometry
# Reference: Snyder (1992), "An equal-area map projection for polyhedral globes", p. 10-12
# Reference: Coxeter (1973), "Regular Polytopes", p. 52-53
#
# Shows how points on the sphere project onto an icosahedral face tangent plane

oldpar <- par(mar = c(1, 1, 2, 1), bg = "white")
plot(NULL, xlim = c(-1.6, 1.6), ylim = c(-0.6, 1.8), asp = 1,
     axes = FALSE, xlab = "", ylab = "",
     main = "Icosahedron Face Projection")

# Sphere cross-section (arc visible above the face)
R <- 1.2
theta_arc <- seq(20*pi/180, 160*pi/180, length.out = 50)
sphere_y_offset <- 0.3
lines(R * cos(theta_arc), R * sin(theta_arc) + sphere_y_offset - R,
      lwd = 2, col = "gray40")
text(-1.3, 1.1, "Sphere surface", cex = 0.9, col = "gray50", adj = 0)

# Draw triangular face (equilateral, base at bottom)
face_scale <- 1.3
v1 <- c(-face_scale * 0.866, -0.4)   # bottom-left vertex
v2 <- c(face_scale * 0.866, -0.4)    # bottom-right vertex
v3 <- c(0, face_scale * 1.0)          # top vertex

polygon(c(v1[1], v2[1], v3[1]), c(v1[2], v2[2], v3[2]),
        col = adjustcolor("gray90", 0.6), border = "gray30", lwd = 2)

# Label vertices
text(v1[1] - 0.12, v1[2] - 0.12, "V1", cex = 0.9, font = 2)
text(v2[1] + 0.12, v2[2] - 0.12, "V2", cex = 0.9, font = 2)
text(v3[1], v3[2] + 0.12, "V3", cex = 0.9, font = 2)

# Mark face center
face_center <- c((v1[1] + v2[1] + v3[1])/3, (v1[2] + v2[2] + v3[2])/3)
points(face_center[1], face_center[2], pch = 19, cex = 1)
text(face_center[1] - 0.2, face_center[2] - 0.08, "Face center", cex = 0.9, col = "gray50")

# A point P on sphere surface
P_sphere <- c(0.5, 0.95)
points(P_sphere[1], P_sphere[2], pch = 19, cex = 1.2, col = "#E63946")
text(P_sphere[1] + 0.12, P_sphere[2] + 0.08, "P", cex = 1.1, font = 2, col = "#E63946")

# Projected point P' on face
P_proj <- c(0.5, 0.45)
points(P_proj[1], P_proj[2], pch = 19, cex = 1.2, col = "#457B9D")
text(P_proj[1] + 0.12, P_proj[2] - 0.05, "P'", cex = 1.1, font = 2, col = "#457B9D")

# Draw projection line (dashed)
lines(c(P_sphere[1], P_proj[1]), c(P_sphere[2], P_proj[2]),
      lwd = 2, lty = 2, col = "#2A9D8F")

# Add text explanation
text(0, -0.55, "Each face covers ~1/20 of the sphere", cex = 0.9)
text(1.55, 0.7, "Tangent plane\n(icosa face)", cex = 0.9, col = "gray50", adj = 1)

# Reference note
text(0, -0.75, "Snyder (1992, p. 10-12)", cex = 0.9, col = "gray50")
par(oldpar)
```

### Vertex Latitude Derivation

The 12 vertices are located at (Coxeter, 1973, p. 52-53):

| Location | Latitude | Longitudes |
|----------|----------|------------|
| North pole | +90° | 0° |
| Upper ring | $+\arctan(1/2) \approx +26.565°$ | 0°, 72°, 144°, 216°, 288° |
| Lower ring | $-\arctan(1/2) \approx -26.565°$ | 36°, 108°, 180°, 252°, 324° |
| South pole | −90° | 0° |

The latitude $\arctan(1/2)$ arises from the golden ratio geometry. An icosahedron can be constructed from three mutually perpendicular golden rectangles ($1 \times \varphi$, where $\varphi = (1+\sqrt{5})/2$). The non-polar vertices have $z$-coordinate $1/s$ where $s = \sqrt{1 + \varphi^2}$, yielding $\tan\phi = 1/2$.

### Standard ISEA Orientation

The default orientation places vertex 0 at longitude 11.25°, latitude $\arctan(\varphi) \approx 58.2825°$, with azimuth 0°. This places icosahedron vertices (pentagon cells) predominantly over oceans (Sahr et al., 2003, p. 123).

```{r face-centers, echo=FALSE, fig.width=7, fig.height=3.5}
library(sf)
library(ggplot2)

centers <- hexify_face_centers()

center_pts <- st_as_sf(
  data.frame(face = 0:19, lon = centers$lon * 180 / pi, lat = centers$lat * 180 / pi),
  coords = c("lon", "lat"), crs = 4326
)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = center_pts, color = "#E63946", size = 3) +
  geom_sf_text(data = center_pts, aes(label = face), nudge_y = 6, size = 5) +
  labs(title = "ISEA Icosahedron Face Centers",
       subtitle = "20 triangular faces, numbered 0-19") +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## Snyder's ISEA Projection

Snyder extended the Lambert projection to the icosahedron by introducing an azimuth-adjustment transformation that ensures seamless transitions between adjacent faces while maintaining the equal-area property (Snyder, 1992, p. 12).

### Key Constants

| Constant | Symbol | Value | Source |
|----------|--------|-------|--------|
| Edge-to-center angle | $E_l$ | 37.37736814° | Snyder (1992, Table 1, p. 14) |
| Geometric angle | $G$ | 36° | 360°/10 (icosahedral 5-fold symmetry) |
| Scale factor | $R_1$ | 0.9103832815 | Snyder (1992, Table 1, p. 14) |

### Forward Projection Steps

The complete algorithm comprises seven steps (Snyder, 1992, p. 13-15):

**Step 1: Compute angular distance and azimuth** from face center $(\lambda_0, \phi_0)$ to point $(\lambda, \phi)$:

$$z = \arccos(\sin\phi_0 \sin\phi + \cos\phi_0 \cos\phi \cos(\lambda - \lambda_0))$$
$$\text{Az} = \arctan2(\cos\phi \sin(\lambda - \lambda_0), \cos\phi_0 \sin\phi - \sin\phi_0 \cos\phi \cos(\lambda - \lambda_0))$$

**Step 2: Reduce azimuth** to [0°, 120°) by exploiting 3-fold symmetry.

**Step 3: Compute auxiliary angle** $\delta_z$ (Snyder, 1992, eq. 8, p. 14):
$$\delta_z = \arctan\left(\frac{\tan E_l}{\cos \text{Az} + \cot 30° \cdot \sin \text{Az}}\right)$$

**Step 4: Compute auxiliary angle** $h$ (Snyder, 1992, eq. 9, p. 14):
$$h = \arccos(\sin \text{Az} \sin G \cos E_l - \cos \text{Az} \cos G)$$

**Step 5: Compute adjusted azimuth** $\text{Az}'$ (Snyder, 1992, eq. 10-11, p. 14):
$$A_G = \text{Az} + G + h - \pi$$
$$\text{Az}' = \arctan\left(\frac{2 A_G}{R_1^2 \tan^2 E_l - 2 A_G \cot 30°}\right)$$

**Step 6: Compute radial distance** (Snyder, 1992, eq. 12-13, p. 14-15):
$$f = \frac{\tan E_l}{2(\cos \text{Az}' + \cot 30° \cdot \sin \text{Az}') \sin(\delta_z / 2)}$$
$$\rho = 2 R_1 f \sin(z / 2)$$

**Step 7: Convert to Cartesian** with sector offset restored.

### The Inverse Projection

The inverse projection cannot be solved analytically because the azimuth adjustment contains transcendental functions. A Newton-Raphson iteration finds the spherical azimuth Az from the planar azimuth Az':

$$f(\text{Az}) = \text{agh} - \text{Az} - G + (\pi - h) = 0$$

where $h = \arccos(\sin \text{Az} \sin G \cos E_l - \cos \text{Az} \cos G)$.

![Newton-Raphson iteration for inverse projection](figures/newton_raphson.png){width=80%}

The iteration exhibits **quadratic convergence**, typically reaching machine precision in 3-5 iterations. hexify provides four precision modes:

| Mode | Tolerance | Typical Iterations | Use Case |
|------|-----------|-------------------|----------|
| fast | $10^{-10}$ | 3-4 | Interactive visualization |
| default | $10^{-12}$ | 4-5 | General applications (~1 m accuracy) |
| high | $10^{-14}$ | 5-6 | High-precision geodesy |
| ultra | $10^{-15}$ | 6-7 | Research |

## Aperture and Resolution

**Aperture** defines how cells subdivide at each resolution level---it's the ratio of parent cell area to child cell area (Sahr et al., 2003, p. 124).

<div style="display: flex; justify-content: space-around; flex-wrap: wrap;">
<div style="text-align: center; flex: 1; min-width: 250px;">
![Aperture 3: Triangular (30° rotation)](figures/aperture3_subdivision.png){width=100%}
</div>
<div style="text-align: center; flex: 1; min-width: 250px;">
![Aperture 4: Rhombic (no rotation)](figures/aperture4_subdivision.png){width=100%}
</div>
<div style="text-align: center; flex: 1; min-width: 250px;">
![Aperture 7: Rosette (19.1° rotation)](figures/aperture7_subdivision.png){width=100%}
</div>
</div>

### Aperture Properties

| Aperture | Area Ratio | Linear Scale | Rotation per Level | Orientation |
|----------|------------|--------------|-------------------|-------------|
| 3 | 1:3 | $\sqrt{3} \approx 1.73$ | 30° | Alternates Class I/II |
| 4 | 1:4 | $2.0$ | 0° | Always Class I |
| 7 | 1:7 | $\sqrt{7} \approx 2.65$ | $\arctan(\sqrt{3/7}) \approx 19.1°$ | Class III |

The aperture 7 rotation angle $\arctan(\sqrt{3/7})$ arises from the geometric constraint that 7 hexagons in a rosette pattern (1 center + 6 ring) must maintain lattice consistency (DGGRID Manual, 2023).

### Cell Count Growth

```{r cell-counts}
cat("Resolution  Aperture 3    Aperture 4    Aperture 7\n")
cat("---------  ----------    ----------    ----------\n")
for (res in 0:8) {
  cells_ap3 <- 10 * 3^res + 2
  cells_ap4 <- 10 * 4^res + 2
  cells_ap7 <- 10 * 7^res + 2
  cat(sprintf("    %d      %10s    %10s    %10s\n",
              res,
              format(cells_ap3, big.mark = ","),
              format(cells_ap4, big.mark = ","),
              format(cells_ap7, big.mark = ",")))
}
```

## Orientation Classes

```{r orientation-classes, echo=FALSE, fig.width=7, fig.height=4}
oldpar <- par(mfrow = c(1, 3), mar = c(1, 0.2, 3, 0.2), bg = "white", cex = 1, cex.main = 1)

hex_v <- function(cx, cy, r, rotation = 0) {
  angles <- seq(rotation, 2*pi + rotation, length.out = 7)
  list(x = cx + r * cos(angles), y = cy + r * sin(angles))
}

# Class I: Flat-top
plot(NULL, xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), asp = 1,
     axes = FALSE, xlab = "", ylab = "", main = "Class I\n(flat-top, 0°)")
h <- hex_v(0, 0, 1.2, rotation = 0)
polygon(h$x, h$y, col = adjustcolor("#457B9D", 0.4), border = "gray30", lwd = 2)
lines(h$x[1:2], h$y[1:2], col = "#E63946", lwd = 3)
text(0, -1.35, "Ap. 4 (all res)\nAp. 3 (even res)", cex = 0.9)

# Class II: Pointy-top (30° rotation)
plot(NULL, xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), asp = 1,
     axes = FALSE, xlab = "", ylab = "", main = "Class II\n(pointy-top, 30°)")
h <- hex_v(0, 0, 1.2, rotation = pi/6)
polygon(h$x, h$y, col = adjustcolor("#2A9D8F", 0.4), border = "gray30", lwd = 2)
points(h$x[1], h$y[1], pch = 19, cex = 1.5, col = "#E63946")
text(0, -1.35, "Ap. 3 (odd res)", cex = 0.9)

# Class III: Skewed
plot(NULL, xlim = c(-1.5, 1.5), ylim = c(-1.5, 1.5), asp = 1,
     axes = FALSE, xlab = "", ylab = "", main = "Class III\n(skewed, ~19.1°)")
h_parent <- hex_v(0, 0, 1.2, rotation = 0)
polygon(h_parent$x, h_parent$y, col = NA, border = "gray50", lwd = 2, lty = 2)
# Correct rotation: arctan(sqrt(3/7))
rot_angle <- atan(sqrt(3/7))
h_child <- hex_v(0, 0, 0.8, rotation = rot_angle)
polygon(h_child$x, h_child$y, col = adjustcolor("#E9C46A", 0.4),
        border = "gray30", lwd = 2)
arc_r <- 0.5
arc_theta <- seq(0, rot_angle, length.out = 20)
lines(arc_r * cos(arc_theta), arc_r * sin(arc_theta), col = "#E63946", lwd = 2)
text(0.6, 0.15, "19.1°", cex = 0.9, col = "#E63946")
text(0, -1.35, "Ap. 7 (all res)\nRotation accumulates", cex = 0.9)
par(oldpar)
```

For **aperture 3**, orientation alternates between Class I (flat-top) and Class II (pointy-top) at each resolution.
For **aperture 7**, each level adds a rotation of $\arctan(\sqrt{3/7}) \approx 19.1°$ (Sahr, 2008, p. 176).

## Pentagon Cells

Exactly **12 cells are pentagons** at every resolution. This is a topological necessity derived from Euler's formula (Coxeter, 1973, p. 10):

$$V - E + F = 2$$

For a tiling with $h$ hexagons and $p$ pentagons on a sphere:
- $V = (6h + 5p)/3$, $E = (6h + 5p)/2$, $F = h + p$

Substituting into Euler's formula and simplifying yields $p = 12$, independent of $h$.

```{r pentagon-locations, echo=FALSE, fig.width=7, fig.height=3.5}
pentagon_coords <- data.frame(
  type = c("Pole", "Pole", rep("Upper ring", 5), rep("Lower ring", 5)),
  lon = c(0, 0, 0, 72, 144, 216, 288, 36, 108, 180, 252, 324),
  lat = c(90, -90, rep(26.57, 5), rep(-26.57, 5))
)

pentagon_pts <- st_as_sf(pentagon_coords, coords = c("lon", "lat"), crs = 4326)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = pentagon_pts, aes(color = type), size = 4) +
  scale_color_manual(values = c("Pole" = "#E63946",
                                 "Upper ring" = "#457B9D",
                                 "Lower ring" = "#2A9D8F")) +
  labs(title = "Pentagon Cell Locations",
       subtitle = "12 pentagonal cells at icosahedron vertices\n(area = 5/6 of hexagons)",
       color = "Location") +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

| Location | Latitude | Longitudes |
|----------|----------|------------|
| Poles | ±90° | 0° |
| Upper ring | $\arctan(1/2) \approx 26.57°$ | 0°, 72°, 144°, 216°, 288° |
| Lower ring | $-\arctan(1/2) \approx -26.57°$ | 36°, 108°, 180°, 252°, 324° |

Pentagon area is exactly 5/6 of hexagonal cell area at the same resolution (Sahr et al., 2003, p. 125).

## Coordinate Systems and Indexing

hexify uses a multi-stage coordinate pipeline (Sahr, 2008, p. 178):

| System | Components | Description |
|--------|------------|-------------|
| **GEO** | lon, lat | WGS84 degrees |
| **Icosa Triangle** | face (0-19), tx, ty | Snyder projection output |
| **Quad XY** | quad (0-11), qx, qy | Paired-triangle coordinates |
| **Quad IJ** | quad (0-11), i, j | Quantized grid indices |
| **Index** | string | Hierarchical cell address (Z3, Z7, or Z-order) |
| **SEQNUM** | integer | Global cell ID (1-based) |

### Triangle to Quad

The 20 triangular faces are paired into 12 "quads" (diamond-shaped regions). Each quad contains two adjacent triangular faces sharing an edge. This pairing transforms 20 triangles into 10 diamond-shaped quads plus 2 polar quads, simplifying grid indexing because a diamond admits a rectangular $(i, j)$ lattice (DGGRID Manual, 2023).

### Quad IJ: The Integer Lattice

After Snyder projection maps a point onto a triangular face, the continuous $(x, y)$ coordinates are quantized to integer $(i, j)$ indices on the hexagonal lattice. At resolution $r$ with aperture $a$, the lattice dimension along each axis is:

$$d = \left\lfloor a^{r/2} \right\rfloor$$

The $(i, j)$ pair uniquely identifies a cell within a quad. Combined with the quad number (0--11), this gives a globally unique cell address.

### Hierarchical Index Types

hexify supports three hierarchical index encodings. Each converts the $(quad, i, j)$ triple into a compact string or integer that encodes the cell's position in the subdivision hierarchy.

#### Z7 Index (Aperture 7)

The Z7 index represents each cell as a hierarchical path through the aperture-7 subdivision tree (Sahr, 2025). The format is:

$$\texttt{BB}\underbrace{\texttt{D}_1\texttt{D}_2\cdots\texttt{D}_r}_{r \text{ digits}}$$

where **BB** is the leading field and each digit $D_k \in \{0, 1, \ldots, 6\}$ selects one of 7 children at level $k$. The digit meanings correspond to positions in the IVec3D cube coordinate system:

| Digit | Direction | Meaning |
|-------|-----------|---------|
| 0 | CENTER | Center child (same position as parent) |
| 1 | K_AXES | K-axis direction |
| 2 | J_AXES | J-axis direction |
| 3 | JK_AXES | JK-axis direction |
| 4 | I_AXES | I-axis direction |
| 5 | IK_AXES | IK-axis direction |
| 6 | IJ_AXES | IJ-axis direction |

The leading field is always two characters, so the resolution is
`nchar(index) - 2`. Dropping the final digit gives the parent; appending digits
0--6 enumerates the seven positions in the next refinement level.

hexify uses a **bijective Z7 variant**. DGGRID's base-cell reassignment and
pentagon digit-skip rules can make its encoder non-injective near pentagons:
distinct cells may receive the same string. hexify keeps the geographic quad
fixed in those regions, so every cell has a distinct index and both
`cell -> index -> cell` and `index -> cell -> index` round-trip.

Holding the quad fixed is what the leading field pays for. A quad is a rhombus
while the aperture-7 parents are hexagons, so the quad boundary cuts through
the parents of the cells along it: walking such a cell up the hierarchy arrives
at one of the six neighbours of the quad's own level-0 point rather than at the
point itself. The leading field records which, as $\texttt{quad} + 12s$ with
$s \in \{0, 1, \ldots, 6\}$, and a cell whose whole ancestry stays inside its
quad has $s = 0$ and keeps the plain two-digit quad DGGRID writes. Recover the
quad with `BB %% 12`, or with `hexify_index_to_cell()`. This matters when
exchanging raw Z7 strings with DGGRID; geographic coordinates and cell geometry
remain the safest interoperability layer.

```{r z7-example}
# Z7 index encoding for aperture 7
g7 <- hex_grid(resolution = 4, aperture = 7)
cell <- lonlat_to_cell(16.37, 48.21, g7)
idx <- cell_to_index(cell, g7)
cat(sprintf("Cell %d -> Z7 index: %s\n", cell, idx))
cat(sprintf("  Leading field: %s (quad %d), Digits: %s\n",
            substr(idx, 1, 2), as.integer(substr(idx, 1, 2)) %% 12L,
            substr(idx, 3, nchar(idx))))

# Hierarchical property: parent is obtained by dropping the last digit
parent_idx <- substr(idx, 1, nchar(idx) - 1)
parent_info <- hexify_index_to_cell(parent_idx, 7, "z7")
cat(sprintf("  Parent index: %s (face %d, i=%d, j=%d)\n",
            parent_idx, parent_info$face,
            as.integer(parent_info$i), as.integer(parent_info$j)))

# A valid Z7 index decodes and re-encodes without changing
decoded <- hexify_index_to_cell(idx, 7, "z7")
stopifnot(identical(
  hexify_cell_to_index(decoded$face, decoded$i, decoded$j,
                       decoded$resolution, 7, "z7"),
  idx
))
```

#### Z3 Index (Aperture 3)

The Z3 index encodes the aperture-3 subdivision hierarchy using **digit pairs** (Sahr, 2008). Because aperture 3 alternates between Class I and Class II orientation at each level, the encoding uses two digits per resolution-level pair:

$$\texttt{BB}\underbrace{\texttt{D}_1\texttt{D}_2\texttt{D}_3\texttt{D}_4\cdots}_{2\lceil r/2 \rceil \text{ digits}}$$

Each $(i_d, j_d)$ digit pair is mapped through a lookup table to a Z3 code pair. For example, the offset $(1, 0)$ maps to digits "01", while $(0, 1)$ maps to "22". This mapping preserves the space-filling curve property, ensuring hierarchical locality.

```{r z3-example}
# Z3 index encoding for aperture 3
g3 <- hex_grid(resolution = 8, aperture = 3)
cell <- lonlat_to_cell(16.37, 48.21, g3)
idx <- cell_to_index(cell, g3)
cat(sprintf("Cell %d -> Z3 index: %s\n", cell, idx))
cat(sprintf("  Base cell: %s, Digits: %s (%d digit pairs)\n",
            substr(idx, 1, 2), substr(idx, 3, nchar(idx)),
            (nchar(idx) - 2) / 2))
```

#### Z-Order Index (Morton Curve)

The Z-order (Morton) index uses bit-interleaving of the $(i, j)$ coordinates, producing a Morton space-filling curve (Morton, 1966). The encoding depends on the aperture:

- **Aperture 4**: Binary interleaving. Each $(i, j)$ digit pair in base 2 produces one output digit: $d = 2i_b + j_b$, yielding digits in $\{0, 1, 2, 3\}$.

- **Aperture 3**: Base-3 interleaving. The $(i, j)$ coordinates are expressed in base 3, and digits alternate: $i_0, j_0, i_1, j_1, \ldots$

- **Aperture 7**: Base-7 interleaving with 2 digits per level.

```{r zorder-example}
# Z-order index for aperture 4
g4 <- hex_grid(resolution = 8, aperture = 4)
cell <- lonlat_to_cell(16.37, 48.21, g4)
idx <- cell_to_index(cell, g4)
cat(sprintf("Aperture 4: Cell %d -> Z-order index: %s\n", cell, idx))
```

### Index Type Comparison

| Property | Z7 | Z3 | Z-order |
|----------|----|----|---------|
| **Apertures** | 7 only | 3 only | 3, 4, 7 |
| **Digits per level** | 1 (range 0--6) | 2 (pairs) | 1 (ap3/4) or 2 (ap7) |
| **Encoding** | Hierarchical path | Mapping table | Bit interleaving |
| **Locality** | Excellent | Excellent | Good |
| **Parent operation** | Drop last digit | Drop last pair | Drop last digit(s) |
| **Index length** (res $r$) | $2 + r$ | $2 + 2\lceil r/2\rceil$ | $2 + r$ or $2 + 2r$ |

All three hexify encodings are bijective: each valid $(quad, i, j)$ cell maps
to exactly one index string, and vice versa. hexify stores indices as character
strings to support arbitrary precision and avoid integer overflow at high
resolutions. As noted above, hexify's bijective Z7 strings intentionally differ
from DGGRID in the pentagon regions where DGGRID's raw Z7 encoding collides.

### SEQNUM: The Flat Cell ID

The SEQNUM provides a unique integer for each cell, independent of the hierarchical index. The total cell count at resolution $r$ with aperture $a$ is:

$$N(r) = 10 \times a^r + 2$$

The "+2" accounts for the two polar pentagon cells. SEQNUMs are assigned by traversing quads in order (0--11) and cells within each quad in row-major $(i, j)$ order. This numbering maintains compatibility with dggridR.

### Compact uint64 Storage

For database-friendly storage, hexify can pack any hierarchical index into a single 64-bit unsigned integer:

$$\texttt{uint64} = (\texttt{face} \ll 60) \;|\; \texttt{packed\_digits}$$

The top 4 bits encode the face (0--11), and the remaining 60 bits pack the index digits. This supports resolutions up to $\lfloor 60 / \lceil\log_2 a\rceil \rfloor$ before overflow.

## H3 Comparison

hexify supports Uber's H3 system as a first-class alternative to the ISEA backend. H3 uses a fundamentally different design with important trade-offs.

### Architecture

| Property | ISEA (hexify native) | H3 (Uber) |
|----------|---------------------|-----------|
| **Projection** | Snyder ISEA (equal-area) | Face-centered gnomonic |
| **Polyhedron** | Icosahedron (20 faces) | Icosahedron (20 faces) |
| **Aperture** | 3, 4 and 7 in any sequence | 7 (fixed) |
| **Cell area** | Exactly equal | Varies by location |
| **Resolutions** | 0--30 | 0--15 |
| **Cell IDs** | Integer SEQNUM | 64-bit hex string |
| **Pentagon handling** | 12 per resolution | 12 per resolution |

Both systems tile the sphere with hexagons on an icosahedral framework, but the projection and subdivision choices lead to different properties.

### The Area Variation Trade-Off

The most significant difference is area uniformity. ISEA uses the Snyder equal-area projection, guaranteeing that all hexagonal cells at a given resolution have identical area (pentagons are 5/6 of hex area). H3 uses a gnomonic (central) projection per face, which is not equal-area. This causes cell areas to vary by latitude:

```{r h3-area-variation, fig.width=7, fig.height=4.375}
# Compare ISEA (constant area) vs H3 (variable area) at similar resolutions
lats <- seq(-85, 85, by = 5)
lons <- rep(10, length(lats))

# ISEA: aperture 7, resolution 6 (~130 km² cells)
g_isea <- hex_grid(resolution = 6, aperture = 7)
isea_cells <- lonlat_to_cell(lons, lats, g_isea)
isea_areas <- cell_area(isea_cells, g_isea)

# H3: resolution 4 (~1,770 km² cells — different scale, but shows the pattern)
g_h3 <- hex_grid(resolution = 4, type = "h3")
h3_cells <- lonlat_to_cell(lons, lats, g_h3)
h3_areas <- cell_area(h3_cells, g_h3)

# Normalize to show relative variation
isea_rel <- isea_areas / mean(isea_areas)
h3_rel <- h3_areas / mean(h3_areas)

oldpar <- par(mar = c(4, 4, 2, 1), bg = "white")
plot(lats, h3_rel, type = "l", col = "#E63946", lwd = 2.5,
     xlab = "Latitude (degrees)", ylab = "Relative cell area",
     main = "Cell Area Variation by Latitude",
     ylim = range(c(isea_rel, h3_rel)))
lines(lats, isea_rel, col = "#457B9D", lwd = 2.5)
abline(h = 1, lty = 2, col = "gray50")
legend("topright", legend = c("H3 (gnomonic)", "ISEA (equal-area)"),
       col = c("#E63946", "#457B9D"), lwd = 2.5, bty = "n")
par(oldpar)
```

For ecological and statistical applications where equal-area cells are important (species density estimation, spatial sampling), ISEA is the correct choice. For applications where hierarchical indexing speed matters more than area equality (ride-sharing, logistics), H3 may be preferable.

### Resolution Mapping

Because ISEA supports multiple apertures, there is no one-to-one resolution mapping to H3. The `h3_crosswalk()` function finds the closest H3 resolution for a given ISEA grid:

```{r resolution-mapping}
# H3 resolution table: compare H3 and ISEA aperture-7 cell areas
cat("H3 Res  Avg Area (km²)  ISEA Ap7 Equivalent\n")
cat("------  --------------  -------------------\n")
for (h3_res in 0:8) {
  g_h3 <- hex_grid(resolution = h3_res, type = "h3")
  h3_area <- g_h3@area_km2

  # Find closest ISEA ap7 resolution by brute force
  best_res <- 0
  best_diff <- Inf
  for (r in 0:15) {
    g_test <- hex_grid(resolution = r, aperture = 7)
    d <- abs(log(g_test@area_km2) - log(h3_area))
    if (d < best_diff) { best_diff <- d; best_res <- r }
  }
  g_isea <- hex_grid(resolution = best_res, aperture = 7)
  cat(sprintf("   %2d   %14.1f  res %d (%.1f km²)\n",
              h3_res, h3_area, best_res, g_isea@area_km2))
}
```

### When to Use Which

| Use case | Recommended | Reason |
|----------|------------|--------|
| Species distribution modeling | ISEA | Equal area eliminates sampling bias |
| Spatial statistics (Moran's I, variograms) | ISEA | Equal area ensures unbiased estimates |
| Ride-sharing / logistics | H3 | Fast hierarchical lookups |
| Visualization only | Either | Area variation invisible at map scale |
| Cross-system interoperability | Both via `h3_crosswalk()` | Bidirectional mapping |

## Round-Trip Accuracy

```{r round-trip}
original_lon <- 16.37
original_lat <- 48.21

cat(sprintf("Original: (%.4f, %.4f)\n\n", original_lon, original_lat))

for (ap in c(3, 4, 7)) {
  res <- if (ap == 7) 6 else 10
  grid <- hex_grid(resolution = res, aperture = ap)
  cell_id <- lonlat_to_cell(original_lon, original_lat, grid)
  recovered <- cell_to_lonlat(cell_id, grid)
  error_km <- sqrt((recovered$lon - original_lon)^2 +
                   (recovered$lat - original_lat)^2) * 111
  cat(sprintf("Aperture %d (res %2d): cell %d -> (%.4f, %.4f), ~%.1f km from center\n",
              ap, res, cell_id,
              recovered$lon, recovered$lat, error_km))
}
```

## Summary of ISEA Properties

| Property | Description |
|----------|-------------|
| **Equal area** | All hexagonal cells identical; pentagons = 5/6 hex area |
| **Bounded distortion** | Max angular distortion ~17.3° at face edges |
| **Uniform topology** | Hexagons have 6 neighbors; pentagons have 5 |
| **12 pentagons** | Topological necessity from Euler's formula |

## References

- Brodsky, I. (2018). H3: Uber's Hexagonal Hierarchical Spatial Index. *Uber Engineering Blog*.

- H3 documentation (2024). *H3: A Hexagonal Hierarchical Geospatial Indexing System*. https://h3geo.org/

- Coxeter, H.S.M. (1973). *Regular Polytopes* (3rd ed.). Dover Publications.

- DGGRID Manual (2023). *DGGRID Version 7.8 Documentation*. https://github.com/sahrk/DGGRID

- Morton, G.M. (1966). *A Computer Oriented Geodetic Data Base and a New Technique in File Sequencing*. IBM Technical Report.

- Sahr, K. (2008). Location coding on icosahedral aperture 3 hexagon discrete global grids. *Computers, Environment and Urban Systems*, 32(3), 174-187.

- Sahr, K. (2025). IGEO7: An equal-area hierarchical hexagonal discrete global grid system with Z7 indexing. *Cartography and Geographic Information Science*.

- Sahr, K., White, D., & Kimerling, A.J. (2003). Geodesic Discrete Global Grid Systems. *Cartography and Geographic Information Science*, 30(2), 121-134.

- Snyder, J.P. (1987). *Map Projections: A Working Manual*. U.S. Geological Survey Professional Paper 1395.

- Snyder, J.P. (1992). An equal-area map projection for polyhedral globes. *Cartographica*, 29(1), 10-21.

---
title: "H3 Grid Support"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{H3 Grid Support}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
CAN_RUN <- requireNamespace("sf", quietly = TRUE)

source("_common.R")
knitr::opts_chunk$set(eval = CAN_RUN)
library(hexify)
```

hexify v0.5.0 adds [H3](https://h3geo.org/) as a first-class grid type alongside
ISEA. Every core function (`hexify()`, `grid_rect()`, `grid_clip()`,
`get_parent()`, `get_children()`) works with both grid systems through the
same interface.

## What is H3?

H3 is a hierarchical hexagonal grid system developed by Uber. It partitions
Earth's surface into hexagonal cells at 16 resolutions (0--15), each roughly
7$\times$ finer than the last. Cell IDs are 64-bit integers encoded as
hexadecimal strings (e.g., `"8528342bfffffff"`).

H3 is widely adopted: the FCC uses it for broadband mapping, Foursquare for
POI analytics, and DuckDB/BigQuery both have native H3 extensions.

**Key difference from ISEA**: H3 cells are _not_ equal-area. Cell area varies
by ~1.6$\times$ between the largest and smallest hexagons at any given
resolution, depending on latitude. For rigorous equal-area analysis, use ISEA.
For interoperability with H3 ecosystems, use `type = "h3"`.

## Getting Started

Create an H3 grid by passing `type = "h3"` to `hex_grid()`:

```{r h3-grid}
library(sf)
library(ggplot2)

# Create an H3 grid specification
grid_h3 <- hex_grid(resolution = 5, type = "h3")
grid_h3
```

Then use `hexify()` with the grid object, just like ISEA:

```{r h3-hexify}
# Sample cities
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome",
           "London", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50,
          -0.12, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90,
          51.51, 50.08, 52.23, 47.50, 52.37)
)

result <- hexify(cities, lon = "lon", lat = "lat", grid = grid_h3)
result
```

H3 cell IDs are character strings, unlike ISEA's numeric IDs:

```{r h3-cell-ids}
# Cell IDs are hexadecimal strings
result@cell_id

# All standard accessors work
cells(result)
n_cells(result)
```

### Choosing Resolution by Area

If you think in terms of cell area rather than resolution numbers, pass
`area_km2` instead of `resolution`. hexify picks the closest H3 resolution:

```{r h3-by-area}
grid_area <- hex_grid(area_km2 = 500, type = "h3")
grid_area
```

## Grid Generation

All grid generation functions work with H3 grids.

### Rectangular Region

```{r h3-grid-rect, fig.width=7, fig.height=5}
# Generate H3 hexagons over Western Europe
grid_h3 <- hex_grid(resolution = 3, type = "h3")
europe_h3 <- grid_rect(c(-10, 35, 25, 60), grid_h3)

# Basemap
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_h3, fill = NA, color = "#E6550D", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("H3 Resolution %d Grid", grid_h3@resolution),
       subtitle = sprintf("~%.0f km² average cell", grid_h3@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE)
```

### Clipping to a Boundary

```{r h3-grid-clip, fig.width=7, fig.height=7}
# Clip H3 grid to France
france <- hexify_world[hexify_world$name == "France", ]
grid_h3 <- hex_grid(resolution = 4, type = "h3")
france_h3 <- grid_clip(france, grid_h3)

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_h3, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = sprintf("H3 Grid Clipped to France (res %d)", grid_h3@resolution)) +
  theme_minimal(base_size = FIG_BASE_SIZE)
```

## Hierarchical Navigation

H3's killer feature is its clean hierarchical structure: every cell has exactly
one parent and seven children. hexify exposes this with `get_parent()` and
`get_children()`.

### Parents

```{r h3-parent}
# Get parent cells (one resolution coarser)
grid_h3 <- hex_grid(resolution = 5, type = "h3")
child_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40),
  lat = c(48.21, 48.86, 52.52),
  grid = grid_h3
)

parent_ids <- get_parent(child_ids, grid_h3, levels = 1)
data.frame(child = child_ids, parent = parent_ids)
```

### Children

```{r h3-children}
# Get children of a single cell (one resolution finer)
grid_coarse <- hex_grid(resolution = 3, type = "h3")
coarse_id <- lonlat_to_cell(16.37, 48.21, grid_coarse)

children <- get_children(coarse_id, grid_coarse, levels = 1)
cat(length(children[[1]]), "children at resolution", grid_coarse@resolution + 1, "\n")
head(children[[1]])
```

### Visualizing the Hierarchy

```{r h3-hierarchy-plot, fig.width=7, fig.height=5}
# Parent cell polygon
parent_poly <- cell_to_sf(coarse_id, grid_coarse)

# Children cell polygons
grid_fine <- hex_grid(resolution = 4, type = "h3")
children_poly <- cell_to_sf(children[[1]], grid_fine)

ggplot() +
  geom_sf(data = children_poly, fill = alpha("#E6550D", 0.3),
          color = "#E6550D", linewidth = 0.5) +
  geom_sf(data = parent_poly, fill = NA, color = "black", linewidth = 1.2) +
  labs(title = sprintf("H3 Hierarchy: 1 parent (res %d)", grid_coarse@resolution),
       subtitle = sprintf("%d children (res %d)",
                          length(children[[1]]), grid_fine@resolution)) +
  theme_minimal(base_size = FIG_BASE_SIZE)
```

## Working with H3 Data

The same hexify workflow applies. Species observations across Europe, aggregated
into H3 cells:

```{r h3-workflow, fig.width=7, fig.height=5.25, message=FALSE, warning=FALSE}
set.seed(42)

# Simulate observations across Europe
obs <- data.frame(
  lon = c(rnorm(200, 10, 12), rnorm(100, 25, 8)),
  lat = c(rnorm(200, 48, 6), rnorm(100, 55, 4)),
  species = sample(c("Sp. A", "Sp. B", "Sp. C"), 300, replace = TRUE)
)
obs$lon <- pmax(-10, pmin(40, obs$lon))
obs$lat <- pmax(35, pmin(65, obs$lat))

# Hexify with H3
grid_h3 <- hex_grid(resolution = 3, type = "h3")
obs_hex <- hexify(obs, lon = "lon", lat = "lat", grid = grid_h3)

# Aggregate: species richness per cell
obs_df <- as.data.frame(obs_hex)
obs_df$cell_id <- obs_hex@cell_id

richness <- aggregate(species ~ cell_id, data = obs_df,
                      FUN = function(x) length(unique(x)))
names(richness)[2] <- "n_species"

# Map it
polys <- cell_to_sf(richness$cell_id, grid_h3)
polys <- merge(polys, richness, by = "cell_id")

europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = polys, aes(fill = n_species), color = "white", linewidth = 0.3) +
  scale_fill_viridis_c(option = "plasma", name = "Species\nRichness") +
  coord_sf(xlim = c(-10, 40), ylim = c(35, 65)) +
  labs(title = "Species Richness on H3 Grid",
       subtitle = sprintf("H3 resolution %d (~%.0f km² avg cells)",
                          grid_h3@resolution, grid_h3@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## ISEA--H3 Crosswalk

`h3_crosswalk()` (v0.6.0) maps between ISEA and H3 cell IDs in both directions.
If you run your analysis on ISEA but collaborators expect H3 cell IDs, this
bridges the two.

```{r h3-crosswalk}
# Start with an ISEA grid and some cells
grid_isea <- hex_grid(resolution = 9, aperture = 3)
isea_ids <- lonlat_to_cell(
  lon = c(16.37, 2.35, 13.40, -3.70, 12.50),
  lat = c(48.21, 48.86, 52.52, 40.42, 41.90),
  grid = grid_isea
)

# Map ISEA cells to their closest H3 equivalents
xw <- h3_crosswalk(isea_ids, grid_isea)
xw[, c("isea_cell_id", "h3_cell_id", "isea_area_km2", "h3_area_km2")]
```

The `area_ratio` column shows how ISEA and H3 cell sizes compare; values
close to 1 mean the resolutions are well-matched.

## ISEA vs H3: When to Use Which

| | ISEA | H3 |
|---|---|---|
| **Cell area** | Exactly equal | ~1.6$\times$ variation |
| **Cell IDs** | Numeric (integer) | Character (hex string) |
| **Apertures** | 3, 4, 7, 4/3 | Fixed (7) |
| **Resolutions** | 0--30 | 0--15 |
| **Hierarchy** | Approximate (aperture-dependent) | Exact (7 children per parent) |
| **Dependencies** | None (built-in C++) | None (vendored H3 C library) |
| **Industry adoption** | Scientific / government | Tech industry / commercial |

**Use ISEA when**:

- Equal-area cells are required (biodiversity surveys, density estimation,
  statistical sampling)

- You need fine control over aperture and resolution

- No external dependencies are acceptable

**Use H3 when**:

- Interoperability with H3 ecosystems (Uber, Foursquare, DuckDB, BigQuery)

- Clean hierarchical operations (parent/child traversal) are a priority

- Slight area variation across latitudes is acceptable for your analysis

## Resolution Reference

```{r h3-resolution-table}
h3_res <- hexify_compare_resolutions(type = "h3", res_range = 0:15)
h3_res$n_cells_fmt <- ifelse(
  h3_res$n_cells > 1e9,
  sprintf("%.1fB", h3_res$n_cells / 1e9),
  ifelse(h3_res$n_cells > 1e6,
         sprintf("%.1fM", h3_res$n_cells / 1e6),
         ifelse(h3_res$n_cells > 1e3,
                sprintf("%.1fK", h3_res$n_cells / 1e3),
                as.character(h3_res$n_cells)))
)
# Use scientific notation for tiny areas, fixed for larger ones
h3_res$area_fmt <- ifelse(
  h3_res$cell_area_km2 < 0.1,
  formatC(h3_res$cell_area_km2, format = "e", digits = 1),
  formatC(h3_res$cell_area_km2, format = "f", digits = 1, big.mark = ",")
)
h3_res$spacing_fmt <- ifelse(
  h3_res$cell_spacing_km < 0.1,
  formatC(h3_res$cell_spacing_km, format = "e", digits = 1),
  formatC(h3_res$cell_spacing_km, format = "f", digits = 1)
)
knitr::kable(
  h3_res[, c("resolution", "n_cells_fmt", "area_fmt", "spacing_fmt")],
  col.names = c("Resolution", "# Cells", "Avg Area (km\u00b2)", "Spacing (km)"),
  align = c("r", "r", "r", "r")
)
```

## See Also

- `vignette("quickstart")` - Getting started with hexify (ISEA-focused)

- `vignette("workflows")` - Grid generation, multi-resolution analysis, GIS export

- `vignette("visualization")` - Plotting with `plot()`, `hexify_heatmap()`

- `vignette("theory")` - Mathematical foundations of the ISEA projection

---
title: "Visualization"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Visualization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
source("_common.R")
library(hexify)
library(sf)
library(ggplot2)
```

Base R `plot()`, `hexify_heatmap()` for ggplot2, and building custom maps from scratch.

## Sample Data

We'll use European cities throughout:

```{r sample-data}
cities <- data.frame(
  name = c("Vienna", "Paris", "Madrid", "Berlin", "Rome",
           "Warsaw", "Prague", "Brussels", "Amsterdam", "Lisbon"),
  lon = c(16.37, 2.35, -3.70, 13.40, 12.50,
          21.01, 14.42, 4.35, 4.90, -9.14),
  lat = c(48.21, 48.86, 40.42, 52.52, 41.90,
          52.23, 50.08, 50.85, 52.37, 38.72)
)

grid <- hex_grid(area_km2 = 10000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)
```

## Base R Plotting with plot()

The `plot()` method provides quick visualization with sensible defaults.

### Basic Plot

```{r plot-basic, fig.width=7, fig.height=5}
plot(result, main = "European Cities")
```

### Customizing Colors and Styling

```{r plot-custom, fig.width=7, fig.height=5}
plot(result,
     grid_fill = "steelblue",
     grid_border = "darkblue",
     grid_alpha = 0.6,
     basemap_fill = "ivory",
     basemap_border = "gray50",
     main = "Custom Styling")
```

### Disabling the Basemap

```{r plot-no-basemap, fig.width=7, fig.height=5}
plot(result,
     basemap = FALSE,
     grid_fill = "forestgreen",
     grid_border = "darkgreen",
     main = "No Basemap")
```

### Setting Map Extent

To control the map extent, use `crop = TRUE` with `crop_expand` to add padding:

```{r plot-extent, fig.width=7, fig.height=5}
plot(result,
     crop = TRUE,
     crop_expand = 0.2,
     main = "Custom Extent (20% padding)")
```

## Showing Points

Points can be displayed on top of hexagon cells. By default, points are jittered within their assigned cell to avoid overplotting.

### Basic Points

```{r plot-points, fig.width=7, fig.height=5}
plot(result,
     show_points = TRUE,
     point_color = "red",
     main = "Cities with Points")
```

### Point Size Presets

The `point_size` parameter accepts presets that define what fraction of a hex cell a single point covers:

| Preset | Coverage |
|--------|----------|
| `"tiny"` | ~2% of cell |
| `"small"` | ~5% of cell |
| `"normal"` / `"auto"` | ~10% of cell |
| `"large"` | ~20% of cell |
| `"very large"` | ~35% of cell |

```{r point-sizes, fig.width=7, fig.height=7}
oldpar <- par(mfrow = c(2, 2), cex = 1, cex.main = 1)

plot(result, show_points = TRUE, point_size = "small",
     point_color = "red", main = "small (~5%)")
plot(result, show_points = TRUE, point_size = "normal",
     point_color = "red", main = "normal (~10%)")
plot(result, show_points = TRUE, point_size = "large",
     point_color = "red", main = "large (~20%)")
plot(result, show_points = TRUE, point_size = "very large",
     point_color = "red", main = "very large (~35%)")

par(oldpar)
```

### Custom Point Styling

```{r plot-points-custom, fig.width=7, fig.height=5}
plot(result,
     show_points = TRUE,
     point_size = "small",
     point_color = "darkblue",
     point_alpha = 0.8,
     grid_fill = "lightyellow",
     grid_border = "orange",
     main = "Custom Point Styling")
```

### Disabling Jitter

For exact point locations (no randomization):

```{r plot-no-jitter, fig.width=7, fig.height=5}
plot(result,
     show_points = TRUE,
     jitter = FALSE,
     point_color = "red",
     main = "Points at Cell Centers (No Jitter)")
```

## ggplot2 with hexify_heatmap()

For ggplot2 users, `hexify_heatmap()` returns a ggplot object that can be further customized.

### Basic ggplot

```{r ggplot-basic, fig.width=7, fig.height=5}
hexify_heatmap(result, basemap = "world", title = "European Cities")
```

### Customizing with ggplot2

Since `hexify_heatmap()` returns a ggplot object, you can add layers and modify themes:

```{r ggplot-custom, fig.width=7, fig.height=5}
hexify_heatmap(result, basemap = "world") +
  labs(
    title = "Major European Cities",
    subtitle = "Assigned to ISEA hexagonal grid cells",
    caption = "Data: Sample cities"
  ) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(
    plot.title = element_text(face = "bold", size = 14),
    panel.grid = element_blank()
  )
```

### Adding Custom Layers

```{r ggplot-layers, fig.width=7, fig.height=5}
# Get the city coordinates
city_points <- st_as_sf(cities, coords = c("lon", "lat"), crs = 4326)

hexify_heatmap(result, basemap = "world", title = "Cities with Labels") +
  geom_sf(data = city_points, color = "red", size = 2) +
  geom_sf_text(data = city_points, aes(label = name),
               nudge_y = 0.8, size = 5, color = "darkgray") +
  coord_sf(xlim = c(-5, 25), ylim = c(45, 55))
```

## Heatmaps with Value Mapping

For choropleth-style visualizations with aggregated data, pass a `value` column to `hexify_heatmap()`.

### Creating Aggregated Data

```{r heatmap-setup}
# Simulate observation data with counts
set.seed(42)
n_obs <- 100
obs_data <- data.frame(
  lon = c(rnorm(60, 10, 8), rnorm(40, 0, 10)),
  lat = c(rnorm(60, 48, 5), rnorm(40, 52, 6)),
  count = rpois(n_obs, lambda = 50)
)

# Hexify
grid <- hex_grid(area_km2 = 10000)
obs_hex <- hexify(obs_data, lon = "lon", lat = "lat", grid = grid)
```

### Basic Heatmap

```{r heatmap-basic, fig.width=7, fig.height=5.25}
hexify_heatmap(
  obs_hex,
  value = "count",
  title = "Observation Counts"
)
```

### Customizing Colors

```{r heatmap-colors, fig.width=7, fig.height=5.25}
hexify_heatmap(
  obs_hex,
  value = "count",
  colors = "YlOrRd",
  title = "Yellow-Orange-Red Palette"
)
```

Available color palettes include any from `scale_fill_viridis_c()` or `scale_fill_distiller()`:

- Viridis options: `"viridis"`, `"magma"`, `"plasma"`, `"inferno"`, `"cividis"`, `"mako"`, `"rocket"`, `"turbo"`

- ColorBrewer sequential: `"YlOrRd"`, `"YlGnBu"`, `"Blues"`, `"Greens"`, `"Reds"`, `"Purples"`, etc.

```{r heatmap-palettes, fig.width=7, fig.height=3.5, eval = requireNamespace("gridExtra", quietly = TRUE)}
p1 <- hexify_heatmap(obs_hex, value = "count", colors = "viridis",
                     title = "viridis", xlim = c(-20, 35), ylim = c(35, 65))
p2 <- hexify_heatmap(obs_hex, value = "count", colors = "YlGnBu",
                     title = "YlGnBu", xlim = c(-20, 35), ylim = c(35, 65))

gridExtra::grid.arrange(p1, p2, ncol = 2)
```

### Setting Map Extent

```{r heatmap-extent, fig.width=7, fig.height=5.25}
hexify_heatmap(
  obs_hex,
  value = "count",
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "Zoomed to Region",
  legend_title = "Count"
)
```

### Basemap Options

```{r heatmap-basemap, fig.width=7, fig.height=5.25}
# With world basemap (default)
hexify_heatmap(
  obs_hex,
  value = "count",
  basemap = "world",
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "With World Basemap"
)
```

```{r heatmap-no-basemap, fig.width=7, fig.height=5.25}
# Without basemap
hexify_heatmap(
  obs_hex,
  value = "count",
  basemap = NULL,
  xlim = c(-20, 35),
  ylim = c(35, 65),
  title = "No Basemap"
)
```

## World Map Helper

`plot_world()` provides a quick way to draw a world basemap:

```{r plot-world, fig.width=7, fig.height=3.5}
plot_world(fill = "lightgray", border = "gray50")
```

### Customizing the World Map

```{r plot-world-custom, fig.width=7, fig.height=3.5}
plot_world(
  fill = "antiquewhite",
  border = "sienna",
  xlim = c(-30, 50),
  ylim = c(30, 70)
)
```

## Pentagon Cell Visualization

The ISEA grid contains 12 pentagonal cells at the icosahedron vertices. Here's how to visualize them:

```{r pentagon-locations, fig.width=7, fig.height=3.5}
# Pentagon locations (icosahedron vertices in standard ISEA orientation)
pentagon_coords <- data.frame(
  type = c("Pole", "Pole", rep("Vertex", 10)),
  lon = c(0, 0, seq(0, 324, by = 36)),
  lat = c(90, -90, rep(c(26.57, -26.57), 5))
)

# Assign to grid and get polygons
grid <- hex_grid(area_km2 = 500000)
pentagon_cells <- lonlat_to_cell(pentagon_coords$lon, pentagon_coords$lat, grid)

pentagon_polys <- cell_to_sf(pentagon_cells, grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = pentagon_polys, fill = alpha("purple", 0.6),
          color = "purple", linewidth = 0.8) +
  labs(
    title = "Pentagon Cell Locations",
    subtitle = "12 pentagonal cells at icosahedron vertices\n(area = 5/6 of hexagons)"
  ) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## Random Sampling Visualization

Visualizing uniformly sampled cells across Earth:

```{r random-sample, fig.width=7, fig.height=3.5}
# Grid parameters (coarse for faster build)
grid <- hex_grid(area_km2 = 200000, aperture = 3)
max_cell <- 10 * (3^grid@resolution) + 2

# Sample random cell IDs
set.seed(123)
N <- 50
random_cells <- sample(1:max_cell, N, replace = FALSE)

# Generate polygons for sampled cells
sample_polys <- cell_to_sf(random_cells, grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray95", color = "gray70", linewidth = 0.2) +
  geom_sf(data = sample_polys, fill = alpha("forestgreen", 0.5),
          color = "darkgreen", linewidth = 0.4) +
  labs(title = sprintf("Random Sample of %d Cells", N),
       subtitle = sprintf("~%.0f km2 each", grid@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## Building Custom Visualizations

For full control, generate polygons directly and use ggplot2:

```{r custom-viz, fig.width=7, fig.height=5.25}
# Create data with a numeric variable
set.seed(456)
stations <- data.frame(
  lon = runif(50, -10, 30),
  lat = runif(50, 35, 60),
  temperature = rnorm(50, mean = 15, sd = 5)
)

# Hexify (coarser grid for faster build)
grid <- hex_grid(area_km2 = 20000)
stations_hex <- hexify(stations, lon = "lon", lat = "lat", grid = grid)

# Aggregate temperature by cell
stations_df <- as.data.frame(stations_hex)
stations_df$cell_id <- stations_hex@cell_id
cell_temps <- aggregate(temperature ~ cell_id, data = stations_df, FUN = mean)

# Generate polygons and merge data
cell_polys <- cell_to_sf(cell_temps$cell_id, grid)
cell_polys <- merge(cell_polys, cell_temps, by = "cell_id")

# Custom visualization
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60", linewidth = 0.2) +
  geom_sf(data = cell_polys, aes(fill = temperature),
          color = "white", linewidth = 0.2) +
  scale_fill_gradient2(
    low = "blue", mid = "white", high = "red",
    midpoint = 15, name = "Temp (°C)"
  ) +
  coord_sf(xlim = c(-10, 30), ylim = c(35, 60)) +
  labs(
    title = "Mean Temperature by Grid Cell",
    subtitle = "Diverging color scale centered at 15°C"
  ) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(
    axis.text = element_blank(),
    axis.ticks = element_blank(),
    panel.grid = element_line(color = "gray90")
  )
```

## Function Reference

| Function | Description |
|----------|-------------|
| `plot()` | Base R plot method for HexData objects |
| `hexify_heatmap()` | ggplot2 visualization, returns modifiable ggplot object |
| `plot_world()` | Quick world basemap |
| `cell_to_sf()` | Generate sf polygons from cell IDs |

## See Also

- `vignette("quickstart")` - Getting started with hexify

- `vignette("workflows")` - Complete analysis workflows

---
title: "Practical Workflows"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Practical Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
source("_common.R")
library(hexify)
library(sf)
library(ggplot2)
```

Shared grids across datasets, polygon generation, resolution selection, and GIS export.

## One Grid, Many Datasets

Define a grid once, reuse it everywhere. Same grid object = guaranteed spatial alignment across datasets.

### The Problem

You often have:

- Several independent datasets (observations, sensors, surveys)

- All in longitude/latitude coordinates

- Collected at different times or from different sources

You want to:

- Put everything on one common global grid

- Be sure the grids actually match

- Combine results later without subtle errors

### The Solution: Shared Grid Objects

```{r one-grid-many-datasets}
# Step 1: Define the grid once
# This is your shared spatial reference system
grid <- hex_grid(area_km2 = 5000)
print(grid)

# Step 2: Create multiple datasets with different structures
set.seed(123)

# Dataset 1: Bird observations (note different column names)
bird_obs <- data.frame(
  species = sample(c("Passer domesticus", "Turdus merula", "Parus major"), 50, replace = TRUE),
  longitude = runif(50, -10, 30),
  latitude = runif(50, 35, 60)
)

# Dataset 2: Mammal records
mammal_obs <- data.frame(
  species = sample(c("Vulpes vulpes", "Meles meles", "Sciurus vulgaris"), 40, replace = TRUE),
  lon = runif(40, -10, 30),
  lat = runif(40, 35, 60)
)

# Dataset 3: Climate stations
climate_data <- data.frame(
  station_id = paste0("WS", 1:20),
  x = runif(20, -10, 30),
  y = runif(20, 35, 60),
  temp_c = rnorm(20, 12, 5)
)

# Step 3: Attach all datasets to the SAME grid
birds <- hexify(bird_obs, lon = "longitude", lat = "latitude", grid = grid)
mammals <- hexify(mammal_obs, lon = "lon", lat = "lat", grid = grid)
climate <- hexify(climate_data, lon = "x", lat = "y", grid = grid)

cat("Birds:  ", nrow(as.data.frame(birds)), "observations in", n_cells(birds), "cells\n")
cat("Mammals:", nrow(as.data.frame(mammals)), "observations in", n_cells(mammals), "cells\n")
cat("Climate:", nrow(as.data.frame(climate)), "stations in", n_cells(climate), "cells\n")
```

### Combining at the Cell Level

Once data are hexified, longitude/latitude no longer matter for analysis. The `cell_id` becomes the shared spatial key:

```{r cell-level-analysis}
# Extract data frames with cell IDs
birds_df <- as.data.frame(birds)
birds_df$cell_id <- birds@cell_id

mammals_df <- as.data.frame(mammals)
mammals_df$cell_id <- mammals@cell_id

climate_df <- as.data.frame(climate)
climate_df$cell_id <- climate@cell_id

# Aggregate each dataset by cell
bird_richness <- aggregate(
  species ~ cell_id,
  data = birds_df,
  FUN = function(x) length(unique(x))
)
names(bird_richness)[2] <- "bird_species"

mammal_richness <- aggregate(
  species ~ cell_id,
  data = mammals_df,
  FUN = function(x) length(unique(x))
)
names(mammal_richness)[2] <- "mammal_species"

mean_temp <- aggregate(
  temp_c ~ cell_id,
  data = climate_df,
  FUN = mean
)
names(mean_temp)[2] <- "mean_temp"

# Join datasets by cell_id - guaranteed to align because same grid
combined <- merge(bird_richness, mammal_richness, by = "cell_id", all = TRUE)
combined <- merge(combined, mean_temp, by = "cell_id", all = TRUE)

head(combined)
```

## Grid Generation

hexify provides functions to generate grid polygons over regions for visualization and analysis.

### Grid Over a Rectangular Region

```{r grid-rect, fig.width=7, fig.height=5}
# Create grid specification
grid <- hex_grid(area_km2 = 5000)

# Generate hexagons over Western Europe
europe_hexes <- grid_rect(c(-10, 35, 25, 60), grid)

# Get European countries for context
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "gray95", color = "gray60") +
  geom_sf(data = europe_hexes, fill = NA, color = "steelblue", linewidth = 0.4) +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 60)) +
  labs(title = sprintf("Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE)
```

### Grid Over a Polygon (Shapefile)

Clip a grid to any sf polygon boundary:

```{r grid-polygon, fig.width=7, fig.height=7}
# Get France boundary
france <- hexify_world[hexify_world$name == "France", ]

# Generate grid covering mainland France
grid <- hex_grid(area_km2 = 2000)
france_grid <- grid_rect(c(-5, 41, 10, 52), grid)

# Clip grid to France boundary
france_grid_clipped <- st_intersection(france_grid, st_geometry(france))

ggplot() +
  geom_sf(data = france, fill = "gray95", color = "gray40", linewidth = 0.5) +
  geom_sf(data = france_grid_clipped, fill = alpha("steelblue", 0.3),
          color = "steelblue", linewidth = 0.3) +
  coord_sf(xlim = c(-5, 10), ylim = c(41, 52)) +
  labs(title = "Hexagonal Grid Clipped to France",
       subtitle = sprintf("~%.0f km² cells", grid@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE)
```

### Global Grid

```{r grid-global, fig.width=7, fig.height=3.5, warning=FALSE}
# Coarse global grid (be careful with fine grids - many cells!)
grid <- hex_grid(area_km2 = 500000)
global_hexes <- grid_global(grid)

ggplot() +
  geom_sf(data = hexify_world, fill = "gray90", color = "gray70", linewidth = 0.2) +
  geom_sf(data = global_hexes, fill = NA, color = "darkgreen", linewidth = 0.3) +
  labs(title = sprintf("Global Hexagonal Grid (~%.0f km² cells)", grid@area_km2)) +
  theme_minimal(base_size = FIG_BASE_SIZE) +
  theme(axis.text = element_blank(), axis.ticks = element_blank())
```

## Multi-Resolution Analysis

Analyze data at multiple spatial scales using different target areas.

```{r multi-resolution}
# Sample data
set.seed(42)
observations <- data.frame(
  species = sample(c("Species A", "Species B", "Species C"), 100, replace = TRUE),
  lon = runif(100, -10, 30),
  lat = runif(100, 35, 60)
)

# Fine resolution (~1000 km² cells)
grid_fine <- hex_grid(area_km2 = 1000)
obs_fine <- hexify(observations, lon = "lon", lat = "lat", grid = grid_fine)

# Coarse resolution (~10000 km² cells)
grid_coarse <- hex_grid(area_km2 = 10000)
obs_coarse <- hexify(observations, lon = "lon", lat = "lat", grid = grid_coarse)

cat(sprintf("Fine resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_fine), grid_fine@area_km2))
cat(sprintf("Coarse resolution: %d unique cells (area: %.1f km²)\n",
            n_cells(obs_coarse), grid_coarse@area_km2))
```

### Scale-Dependent Patterns

```{r multi-scale-aggregate}
# Extract data with cell IDs
fine_df <- as.data.frame(obs_fine)
fine_df$cell_id <- obs_fine@cell_id

coarse_df <- as.data.frame(obs_coarse)
coarse_df$cell_id <- obs_coarse@cell_id

# Species richness at each scale
richness_fine <- aggregate(species ~ cell_id, data = fine_df,
                           FUN = function(x) length(unique(x)))
richness_coarse <- aggregate(species ~ cell_id, data = coarse_df,
                             FUN = function(x) length(unique(x)))

cat(sprintf("Fine scale: mean %.2f species per cell\n", mean(richness_fine$species)))
cat(sprintf("Coarse scale: mean %.2f species per cell\n", mean(richness_coarse$species)))
```

## Spatial Joins

Join datasets based on shared grid cells rather than proximity.

```{r spatial-joins}
# Dataset 1: Weather stations
stations <- data.frame(
  station_id = paste0("ST", 1:50),
  lon = runif(50, -10, 30),
  lat = runif(50, 35, 60),
  temperature = rnorm(50, 15, 5)
)

# Dataset 2: Cities
cities <- data.frame(
  city = c("Vienna", "Paris", "London", "Berlin", "Rome",
           "Madrid", "Prague", "Warsaw", "Budapest", "Amsterdam"),
  lon = c(16.37, 2.35, -0.12, 13.40, 12.50,
          -3.70, 14.42, 21.01, 19.04, 4.90),
  lat = c(48.21, 48.86, 51.51, 52.52, 41.90,
          40.42, 50.08, 52.23, 47.50, 52.37)
)

# Use a coarse grid for joining disparate points
grid <- hex_grid(area_km2 = 50000)

# Hexify both datasets with the same grid
stations_hex <- hexify(stations, lon = "lon", lat = "lat", grid = grid)
cities_hex <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Extract with cell IDs
stations_df <- as.data.frame(stations_hex)
stations_df$cell_id <- stations_hex@cell_id

cities_df <- as.data.frame(cities_hex)
cities_df$cell_id <- cities_hex@cell_id

# Join by cell_id
city_weather <- merge(
  cities_df[, c("city", "cell_id")],
  aggregate(temperature ~ cell_id, data = stations_df, FUN = mean),
  by = "cell_id",
  all.x = TRUE
)

city_weather
```

## Cell-Level Aggregation and Neighbors

`hex_summarize()` aggregates data per cell without manually splitting and
merging, and `get_neighbors()` finds the ring of cells surrounding a given
cell — useful for smoothing, spatial lag features, or checking what
surrounds a hotspot.

```{r cell-summarize}
# Aggregate the station data straight from the HexData object
hex_summarize(stations_hex, mean_temp = mean(temperature), n_stations = length(temperature))
```

```{r cell-neighbors}
# The 6 cells surrounding a given cell (k = 1), or a wider ring with k > 1
some_cell <- stations_hex@cell_id[1]
get_neighbors(some_cell, grid)
get_neighbors(some_cell, grid, k = 2)
```

## Choosing Resolution

### By Target Area

Use `hex_grid()` with `area_km2` to get the closest available resolution:

```{r choosing-resolution}
# Target: 100 km² cells
grid_100 <- hex_grid(area_km2 = 100, aperture = 3)
cat(sprintf("Target ~100 km²: resolution %d (actual: %.1f km²)\n",
            grid_100@resolution, grid_100@area_km2))

# Target: 1000 km² cells
grid_1000 <- hex_grid(area_km2 = 1000, aperture = 3)
cat(sprintf("Target ~1000 km²: resolution %d (actual: %.1f km²)\n",
            grid_1000@resolution, grid_1000@area_km2))

# Target: 10000 km² cells
grid_10000 <- hex_grid(area_km2 = 10000, aperture = 3)
cat(sprintf("Target ~10000 km²: resolution %d (actual: %.1f km²)\n",
            grid_10000@resolution, grid_10000@area_km2))
```

### Resolution Table (Aperture 3)

```{r resolution-table, echo=FALSE}
comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:15)
comparison$n_cells_fmt <- ifelse(
  comparison$n_cells > 1e6,
  sprintf("%.1fM", comparison$n_cells / 1e6),
  ifelse(comparison$n_cells > 1e3,
         sprintf("%.1fK", comparison$n_cells / 1e3),
         as.character(comparison$n_cells))
)
comparison$area_fmt <- ifelse(
  comparison$cell_area_km2 < 0.1,
  formatC(comparison$cell_area_km2, format = "e", digits = 1),
  formatC(comparison$cell_area_km2, format = "f", digits = 1, big.mark = ",")
)
comparison$spacing_fmt <- ifelse(
  comparison$cell_spacing_km < 0.1,
  formatC(comparison$cell_spacing_km, format = "e", digits = 1),
  formatC(comparison$cell_spacing_km, format = "f", digits = 1)
)
knitr::kable(
  comparison[, c("resolution", "n_cells_fmt", "area_fmt", "spacing_fmt")],
  col.names = c("Resolution", "# Cells", "Cell Area (km\u00b2)", "Spacing (km)"),
  align = c("r", "r", "r", "r")
)
```

### Comparing Apertures

Different apertures offer different resolution steps:

```{r compare-apertures}
target_area <- 1000  # km²

cat(sprintf("Target: ~%d km² cells\n\n", target_area))
for (ap in c(3, 4, 7)) {
  grid <- hex_grid(area_km2 = target_area, aperture = ap)
  n_cells <- 10 * (ap^grid@resolution) + 2
  cat(sprintf("Aperture %d: resolution %d -> %.1f km² (%s cells globally)\n",
              ap, grid@resolution, grid@area_km2,
              format(n_cells, big.mark = ",")))
}
```

| Aperture | Best For | Trade-offs |
|----------|----------|------------|
| 3 | Fine resolution control, dggridR compatibility | Slowest cell growth |
| 4 | Power-of-2 scaling, GIS workflows | Moderate resolution steps |
| 7 | Rapid cell count growth, coarse analysis | Largest resolution jumps |
| 4/3 | Balance of 4's fast start + 3's fine control | More complex indexing |

## Grids on Other Bodies

A grid partitions a sphere, and `radius_km` sets which sphere. Pass a radius in
kilometres or the name of a body:

```{r other-bodies}
mars <- hex_grid(area_km2 = 1000, radius_km = "mars")
mars

# The same grid, by radius
identical(mars@area_km2, hex_grid(resolution = mars@resolution, radius_km = 3389.5)@area_km2)
```

Cell geometry is angular: which cell a coordinate falls in, where centres and
corners sit, the parent-child hierarchy and the neighbours are the same on
every body. The radius sets the kilometre figures — cell area, diagonal,
spacing, and the resolution that a target `area_km2` picks.

```{r other-bodies-geometry}
earth <- hex_grid(resolution = 5)
moon  <- hex_grid(resolution = 5, radius_km = "moon")

lon <- c(0, 16.37, -70.5)
lat <- c(0, 48.21, -33.4)

identical(lonlat_to_cell(lon, lat, moon), lonlat_to_cell(lon, lat, earth))
c(earth = earth@area_km2, moon = moon@area_km2)
```

Areas come from the sphere of that radius, `4 * pi * r^2`; Earth keeps the
WGS84 ellipsoid area. Named bodies carry the IAU mean radii (Archinal et al.
2018) tabulated by JPL Solar System Dynamics: `mercury`, `venus`, `earth`,
`moon`, `mars`, `ceres`, `jupiter`, `io`, `europa`, `ganymede`, `callisto`,
`saturn`, `enceladus`, `titan`, `uranus`, `neptune`, `pluto`.

`radius_km` reaches every function that reports kilometres, because the grid
object carries it: `dgearthstat()`, `cell_area()`, `hexify_compare_resolutions()`
and the sf exports all read the grid's radius.

### Coordinates on another body

EPSG codes name Earth reference systems, so a grid built on another radius
carries a longlat CRS on the sphere of that radius:

```{r other-bodies-crs}
mars@crs
```

Every sf object built from that grid carries it, so cells, centres and hexified
data come back in the body's own coordinates. A planet of your own works the
same way through a radius in kilometres:

```{r other-bodies-crs-custom}
world <- hex_grid(area_km2 = 5000, radius_km = 4200)
world@crs
```

`crs` also takes a value of your own, as an EPSG code or a PROJ or WKT string,
which is how a projected system for the body reaches the grid:

```{r other-bodies-crs-explicit}
hex_grid(area_km2 = 5000, radius_km = 4200,
         crs = "+proj=laea +lat_0=0 +lon_0=0 +R=4200000 +no_defs")@crs
```

A basemap for such a body is any file sf or terra reads: `sf::st_read()` takes a
shapefile, GeoJSON or GeoPackage of coastlines, `terra::rast()` takes a GeoTIFF.
Both go to `hexify_heatmap(basemap = )`.

Both backends take a radius. H3 reports a cell's area as its solid angle times
Earth's radius squared, so another radius scales those areas by the square of
the radius ratio:

```{r other-bodies-h3}
h3_mars <- hex_grid(resolution = 5, type = "h3", radius_km = "mars")
h3_mars@area_km2 / hex_grid(resolution = 5, type = "h3")@area_km2

# H3 measures against the WGS84 authalic radius, so that is what divides out
(3389.5 / 6371.007180918475)^2
```

One caveat rides along with H3. A cell ID names a position in H3's topology,
which Uber's H3 reads on Earth. A grid on another body reuses that topology and
its own radius for areas, so its IDs are that topology on that body, not
interchangeable with Earth H3 data. `h3_crosswalk()` needs both grids on the
same body for the same reason.

## Working with sf

### Convert HexData to sf

```{r sf-workflow}
# Hexify some data
grid <- hex_grid(area_km2 = 20000)
result <- hexify(cities, lon = "lon", lat = "lat", grid = grid)

# Convert to sf points (uses cell centers)
sf_points <- as_sf(result, geometry = "point")
class(sf_points)

# Convert to sf polygons (for choropleth maps)
sf_polys <- as_sf(result, geometry = "polygon")
class(sf_polys)

# Or generate polygons directly from cell IDs
unique_cells <- cells(result)
cell_polys <- cell_to_sf(unique_cells, grid)
```

### Visualize with ggplot2

```{r sf-plot, fig.width=7, fig.height=5}
europe <- hexify_world[hexify_world$continent == "Europe", ]

ggplot() +
  geom_sf(data = europe, fill = "ivory", color = "gray70") +
  geom_sf(data = cell_polys, fill = "steelblue", alpha = 0.5, color = "darkblue") +
  coord_sf(xlim = c(-10, 25), ylim = c(35, 58)) +
  labs(title = "European Cities - Hexagonal Grid") +
  theme_minimal(base_size = FIG_BASE_SIZE)
```

## Export to GIS Formats

Use sf's `st_write()` to export grids for use in external GIS software:

```{r export-formats, eval=FALSE}
# Generate a grid
grid <- hex_grid(area_km2 = 10000)
europe <- grid_rect(c(-10, 35, 25, 60), grid)

# Export to various formats
st_write(europe, "europe_grid.gpkg", layer = "hexgrid")     # GeoPackage
st_write(europe, "europe_grid.shp")                         # Shapefile
st_write(europe, "europe_grid.geojson")                     # GeoJSON
st_write(europe, "europe_grid.kml", layer = "hexgrid")      # KML (Google Earth)
```

## Edge Cases

### Handling NA Coordinates

```{r edge-cases}
data_with_na <- data.frame(
  lon = c(16.37, NA, 2.35, 13.40),
  lat = c(48.21, 48.86, NA, 52.52)
)

grid <- hex_grid(area_km2 = 1000)
result <- hexify(data_with_na, lon = "lon", lat = "lat", grid = grid)

# Check which rows have valid cell assignments
cat("Cell IDs:", result@cell_id, "\n")
cat("NA indicates invalid coordinates\n")
```

### Polar Regions

Coordinates with latitude > 89° may have projection artifacts. The grid remains valid, but polygon visualization can be distorted near poles.

### Date Line

Polygons crossing the date line (lon = ±180°) are handled automatically.
`cell_to_sf()` applies `sf::st_wrap_dateline()` internally, so flat map
projections render correctly without manual intervention.

## Function Summary

| Task | Function |
|------|----------|
| Create grid specification | `hex_grid(area_km2 = ...)` |
| Assign points to cells | `hexify(df, lon, lat, grid)` |
| Get grid from HexData | `grid_info(result)` |
| Get unique cell IDs | `cells(result)` |
| Count cells | `n_cells(result)` |
| Extract data frame | `as.data.frame(result)` |
| Convert to sf | `as_sf(result, geometry = "polygon")` |
| Generate polygons | `cell_to_sf(cell_ids, grid)` |
| Grid over region | `grid_rect(bbox, grid)` |
| Global grid | `grid_global(grid)` |
| Coordinate conversion | `lonlat_to_cell()`, `cell_to_lonlat()` |
| Compare resolutions | `hexify_compare_resolutions()` |
| Aggregate data per cell | `hex_summarize(result, ...)` |
| Find neighboring cells | `get_neighbors(cell_id, grid, k = ...)` |
| Merge/split multi-resolution cells | `hex_compact()`, `hex_uncompact()` |
| Distance between cells | `hex_distance(cell_id1, cell_id2, grid)` |
| Extract raster values at cell centers | `hex_extract(raster, grid)` |
| Zonal statistics over cell polygons | `hex_zonal(raster, grid)` |
| Map ISEA cells to H3 (or vice versa) | `h3_crosswalk()` |
| Per-cell area | `cell_area(cell_ids, grid)` |

## See Also

- `vignette("quickstart")` - Getting started with hexify

- `vignette("visualization")` - Plotting with `plot()`, `hexify_heatmap()`

- `vignette("theory")` - Mathematical foundations (ISEA projection, apertures)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_browse.R
\name{hex_browse}
\alias{hex_browse}
\title{Interactive Hex Map}
\usage{
hex_browse(
  hex_data,
  grid = NULL,
  value = NULL,
  palette = "viridis",
  opacity = 0.7
)
}
\arguments{
\item{hex_data}{A HexData object, or a data.frame with a \code{cell_id} column.}

\item{grid}{A HexGridInfo object. Required if \code{hex_data} is a data.frame.}

\item{value}{Optional column name (character) to color cells by.
If \code{NULL}, cells are colored uniformly.}

\item{palette}{Color palette name for continuous values. Default \code{"viridis"}.}

\item{opacity}{Fill opacity (0-1). Default 0.7.}
}
\value{
A \code{leaflet} map object (can be printed or embedded in Shiny).
}
\description{
Opens an interactive leaflet map with hexagonal cell polygons.
Cells can be colored by a data column, with popups showing cell
information on click.
}
\details{
Requires the \code{leaflet} package (in Suggests). The map is built using
\code{\link[=as_sf]{as_sf()}} to generate polygon geometries, then rendered as a leaflet
choropleth.
}
\examples{
\donttest{
if (requireNamespace("leaflet", quietly = TRUE)) {
  df <- data.frame(
    lon = runif(50, -5, 5),
    lat = runif(50, 45, 55),
    value = rnorm(50)
  )
  hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 500)
  hex_browse(hd, value = "value")
}
}
}
\seealso{
\code{\link[=hexify_heatmap]{hexify_heatmap()}} for static ggplot2 maps,
\code{\link[=as_sf]{as_sf()}} for sf conversion
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{n_cells}
\alias{n_cells}
\title{Get Number of Cells}
\usage{
n_cells(x)
}
\arguments{
\item{x}{A HexData object}
}
\value{
Integer count of unique cells
}
\description{
Get the number of unique cells in a HexData object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_projection.R
\name{hexify_face_centers}
\alias{hexify_face_centers}
\title{Get icosahedron face centers}
\usage{
hexify_face_centers()
}
\value{
Data frame with 20 rows and columns lon, lat (degrees)
}
\description{
Returns the center coordinates of all 20 icosahedral faces.
}
\examples{
centers <- hexify_face_centers()
plot(centers$lon, centers$lat)
}
\seealso{
Other projection:
\code{\link[=hexify_build_icosa]{hexify_build_icosa()}},
\code{\link[=hexify_forward]{hexify_forward()}},
\code{\link[=hexify_forward_to_face]{hexify_forward_to_face()}},
\code{\link[=hexify_get_precision]{hexify_get_precision()}},
\code{\link[=hexify_inverse]{hexify_inverse()}},
\code{\link[=hexify_projection_stats]{hexify_projection_stats()}},
\code{\link[=hexify_set_precision]{hexify_set_precision()}},
\code{\link[=hexify_set_verbose]{hexify_set_verbose()}},
\code{\link[=hexify_which_face]{hexify_which_face()}}
}
\concept{projection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{cell_to_lonlat}
\alias{cell_to_lonlat}
\title{Convert cell ID to longitude/latitude}
\usage{
cell_to_lonlat(cell_id, grid)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs}

\item{grid}{A HexGridInfo or HexData object}
}
\value{
Data frame with lon_deg and lat_deg columns
}
\description{
Converts DGGS cell IDs back to geographic coordinates (cell centers).
}
\examples{
grid <- hex_grid(area_km2 = 1000)
cells <- lonlat_to_cell(c(0, 10), c(45, 50), grid)
coords <- cell_to_lonlat(cells, grid)
}
\seealso{
\code{\link{lonlat_to_cell}} for the forward operation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_quad_ij_to_cell}
\alias{hexify_quad_ij_to_cell}
\title{Convert Quad IJ coordinates to cell ID}
\usage{
hexify_quad_ij_to_cell(quad, i, j, resolution, aperture = 3L)
}
\arguments{
\item{quad}{Quad number (0-11), integer or vector}

\item{i}{Cell index along first axis, integer or vector}

\item{j}{Cell index along second axis, integer or vector}

\item{resolution}{Grid resolution level (0-30)}

\item{aperture}{Grid aperture: 3, 4, or 7}
}
\value{
Numeric vector of cell IDs
}
\description{
Converts Quad IJ coordinates to a global cell identifier.
This is the final step in the coordinate pipeline.
}
\examples{
# Convert Quad IJ to cell ID
cell_id <- hexify_quad_ij_to_cell(quad = 1, i = 100, j = 50,
                                   resolution = 10, aperture = 3)
print(cell_id)
}
\seealso{
Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h3_compat.R
\name{import_h3}
\alias{import_h3}
\title{Import External H3 Cell IDs into hexify}
\usage{
import_h3(cell_ids, data = NULL, validate = TRUE, radius_km = EARTH_RADIUS_KM)
}
\arguments{
\item{cell_ids}{Character vector of H3 cell ID strings}

\item{data}{Optional data frame to attach. Must have the same number of rows
as \code{cell_ids}. If \code{NULL}, returns a HexGridInfo object.}

\item{validate}{If \code{TRUE} (default), checks that all cell IDs are valid
H3 cells at the same resolution before proceeding. Set to \code{FALSE} to
skip validation when cell IDs are known to be correct.}

\item{radius_km}{Radius of the body the cells cover, in kilometers, or a body
name (default Earth). See \code{\link{hex_grid}}.}
}
\value{
If \code{data = NULL}, a HexGridInfo object for the inferred H3
resolution. If \code{data} is provided, a HexData object with data
attached at the specified cells.
}
\description{
Ingests H3 cell IDs from an external source (another H3 library, a database,
or a CSV file) into hexify. Validates cell IDs, infers the H3 resolution,
and optionally attaches data to build a HexData object.
}
\details{
For converting between grid specs, use
\code{\link{hex_grid}(type = "h3")} directly. For cell-level ISEA/H3
mapping, use \code{\link{h3_crosswalk}}.

H3 cell IDs encode their resolution in the index itself, so no resolution
argument is needed. The resolution is inferred automatically. All cell IDs
must share the same resolution; mixed resolutions produce an error.
}
\examples{
\donttest{
# Import external H3 cell IDs (grid spec only)
h3_ids <- c("8528342bfffffff", "85283473fffffff", "85283447fffffff")
grid <- import_h3(h3_ids)
grid

# Import with data attached
df <- data.frame(species = c("oak", "pine", "birch"), count = c(10, 5, 3))
hd <- import_h3(h3_ids, data = df)
hd
}
}
\seealso{
\code{\link{hex_grid}} for creating grids directly,
\code{\link{h3_crosswalk}} for cell-level ISEA/H3 mapping
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{hexify_grid_to_HexGridInfo}
\alias{hexify_grid_to_HexGridInfo}
\title{Convert legacy hexify_grid to HexGridInfo}
\usage{
hexify_grid_to_HexGridInfo(x)
}
\arguments{
\item{x}{A hexify_grid object (S3)}
}
\value{
A HexGridInfo object (S4)
}
\description{
Convert legacy hexify_grid to HexGridInfo
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_compact.R
\name{hex_uncompact}
\alias{hex_uncompact}
\title{Uncompact Hex Cells}
\usage{
hex_uncompact(cell_ids, grid, target_resolution)
}
\arguments{
\item{cell_ids}{Character vector of (possibly mixed-resolution) cell IDs.}

\item{grid}{A HexGridInfo object specifying the grid.}

\item{target_resolution}{Integer. The resolution to expand all cells to.}
}
\value{
A character vector of cell IDs, all at \code{target_resolution}.
}
\description{
Expands compacted cells to a target resolution. All cells in the output
share the same resolution.
}
\details{
\strong{H3 backend:} Uses the vendored H3 \code{uncompactCells} function.

\strong{ISEA backend (aperture 7, Z7 index):} Appends digits 0-6 to expand
each cell to its 7 children, repeating until the target resolution is
reached.
}
\examples{
\donttest{
g <- hex_grid(resolution = 3, type = "h3")
parent <- "832830fffffffff"
hex_uncompact(parent, g, target_resolution = 4L)
}
}
\seealso{
\code{\link[=hex_compact]{hex_compact()}} for the inverse operation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dggrid_compat.R
\name{from_dggrid}
\alias{from_dggrid}
\title{Convert 'dggridR' grid object to hexify_grid}
\usage{
from_dggrid(dggs)
}
\arguments{
\item{dggs}{A 'dggridR' grid object from dgconstruct()}
}
\value{
A hexify_grid object
}
\description{
Creates a hexify_grid object from a 'dggridR' dggs object. This allows
using hexify functions with grids created by 'dggridR' dgconstruct().
}
\details{
Only 'ISEA' projection with HEXAGON topology is fully supported.
Other configurations will generate warnings.

The function validates that the 'dggridR' grid uses compatible settings:
\itemize{
\item Projection must be 'ISEA' (FULLER not supported)
\item Topology must be "HEXAGON" (DIAMOND, TRIANGLE not supported)
\item Aperture must be 3, 4, or 7
}
}
\seealso{
Other 'dggridR' compatibility:
\code{\link[=as_dggrid]{as_dggrid()}},
\code{\link[=dggrid_43h_sequence]{dggrid_43h_sequence()}},
\code{\link[=dggrid_is_compatible]{dggrid_is_compatible()}}
}
\concept{'dggridR' compatibility}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_map.R
\name{hexify_heatmap}
\alias{hexify_heatmap}
\title{Create a ggplot2 visualization of hexagonal grid cells}
\usage{
hexify_heatmap(
  data,
  value = NULL,
  basemap = NULL,
  crs = NULL,
  colors = NULL,
  breaks = NULL,
  labels = NULL,
  hex_border = "#5D4E37",
  hex_lwd = 0.3,
  hex_alpha = 0.7,
  basemap_fill = "gray90",
  basemap_border = "gray50",
  basemap_lwd = 0.5,
  mask_outside = FALSE,
  aperture = 3L,
  xlim = NULL,
  ylim = NULL,
  title = NULL,
  legend_title = NULL,
  na_color = "gray90",
  theme_void = TRUE
)
}
\arguments{
\item{data}{A HexData object from \code{hexify()}, a data frame with cell_id
and cell_area columns, or an sf object with hexagon polygons.}

\item{value}{Column name (as string) to use for fill color. If NULL, cells
are drawn with a uniform fill color. If not specified but data has a
'count' or 'n' column, that will be used automatically.}

\item{basemap}{Optional basemap. Can be:
\itemize{
\item \code{NULL}: No basemap (default)
\item \code{"world"}: Use built-in \code{hexify_world} map (low resolution)
\item \code{"world_hires"}: Use high-resolution map from rnaturalearth (requires package)
\item An sf object: User-supplied vector map, from any format
\code{sf::st_read()} reads, such as a shapefile, 'GeoJSON' or 'GeoPackage'
\item A SpatRaster: User-supplied raster, from any format
\code{terra::rast()} reads, such as a 'GeoTIFF', drawn in grayscale
}}

\item{crs}{Target CRS for the map projection. Can be:
\itemize{
\item A numeric EPSG code (e.g., 4326 for 'WGS84', 3035 for LAEA Europe)
\item A proj4 string
\item An sf crs object
\item NULL to use 'WGS84' (EPSG:4326)
}}

\item{colors}{Color palette for the heatmap. Can be:
\itemize{
\item A character vector of colors (for manual scale)
\item A single RColorBrewer palette name (e.g., "YlOrRd", "Greens")
\item NULL to use viridis
}}

\item{breaks}{Numeric vector of break points for binning continuous values,
or NULL for continuous scale. Use \code{Inf} and \code{-Inf} for open-ended bins.}

\item{labels}{Labels for the breaks (length should be one less than breaks).
If NULL, labels are auto-generated.}

\item{hex_border}{Border color for hexagons}

\item{hex_lwd}{Line width for hexagon borders}

\item{hex_alpha}{Transparency for hexagon fill (0-1)}

\item{basemap_fill}{Fill color for basemap polygons}

\item{basemap_border}{Border color for basemap polygons}

\item{basemap_lwd}{Line width for basemap borders}

\item{mask_outside}{Logical. If TRUE and basemap is provided, mask hexagon
portions that fall outside the basemap polygons.}

\item{aperture}{Grid aperture (default 3), used if data is from hexify()}

\item{xlim}{Optional x-axis limits (in target CRS units) as c(min, max)}

\item{ylim}{Optional y-axis limits (in target CRS units) as c(min, max)}

\item{title}{Plot title}

\item{legend_title}{Title for the color legend}

\item{na_color}{Color for NA values}

\item{theme_void}{Logical. If TRUE (default), use a minimal theme without
axes, gridlines, or background.}
}
\value{
A ggplot2 object that can be further customized or saved.
}
\description{
Creates a ggplot2-based visualization of hexagonal grid cells, optionally
colored by a value column. Supports continuous and discrete color scales,
projection transformation, and customizable styling.
}
\details{
This function provides publication-quality heatmap visualizations of
hexagonal grids using ggplot2. It returns a ggplot object that can be
further customized with standard ggplot2 functions.
}
\section{Color Scales}{

The function supports three types of color scales:
\describe{
\item{Continuous}{Set \code{breaks = NULL} for a continuous gradient}
\item{Binned}{Provide \code{breaks} vector to bin values into categories}
\item{Discrete}{If \code{value} column is a factor, discrete colors are used}
}
}

\section{Projections}{

Common projections:
\describe{
\item{4326}{'WGS84' (unprojected lat/lon)}
\item{3035}{LAEA Europe}
\item{3857}{Web Mercator}
\item{"+proj=robin"}{Robinson (world maps)}
\item{"+proj=moll"}{Mollweide (equal-area world maps)}
}
}

\examples{
library(hexify)

# Sample data with counts
cities <- data.frame(
  lon = c(16.37, 2.35, -3.70, 12.5, 4.9),
  lat = c(48.21, 48.86, 40.42, 41.9, 52.4),
  count = c(100, 250, 75, 180, 300)
)
result <- hexify(cities, lon = "lon", lat = "lat", area_km2 = 5000)

# Simple plot (uniform fill, no value mapping)
hexify_heatmap(result)

\donttest{
library(ggplot2)

# With world basemap
hexify_heatmap(result, basemap = "world")

# Heatmap with value mapping
hexify_heatmap(result, value = "count")

# With world basemap and custom colors
hexify_heatmap(result, value = "count",
               basemap = "world",
               colors = "YlOrRd",
               title = "City Density")

# Binned values with custom breaks
hexify_heatmap(result, value = "count",
               basemap = "world",
               breaks = c(-Inf, 100, 200, Inf),
               labels = c("Low", "Medium", "High"),
               colors = c("#fee8c8", "#fdbb84", "#e34a33"))

# Different projection (LAEA Europe)
hexify_heatmap(result, value = "count",
               basemap = "world",
               crs = 3035,
               xlim = c(2500000, 6500000),
               ylim = c(1500000, 5500000))

# Customize further with ggplot2
hexify_heatmap(result, value = "count", basemap = "world") +
  labs(caption = "Data source: Example") +
  theme(legend.position = "bottom")
}
}
\seealso{
\code{\link{plot_grid}} for base R plotting,
\code{\link{cell_to_sf}} to generate polygons manually

Other visualization:
\code{\link[=plot_world]{plot_world()}}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_projection.R
\name{hexify_set_precision}
\alias{hexify_set_precision}
\title{Set inverse projection precision}
\usage{
hexify_set_precision(
  mode = c("fast", "default", "high", "ultra"),
  tol = NULL,
  max_iters = NULL
)
}
\arguments{
\item{mode}{Preset mode: "fast", "default", "high", or "ultra"}

\item{tol}{Custom tolerance (overrides mode if provided)}

\item{max_iters}{Custom max iterations (overrides mode if provided)}
}
\value{
Invisible NULL
}
\description{
Controls the accuracy/speed tradeoff for inverse Snyder projection.
}
\examples{
hexify_set_precision("high")
hexify_set_precision(tol = 1e-12, max_iters = 100)
}
\seealso{
Other projection:
\code{\link[=hexify_build_icosa]{hexify_build_icosa()}},
\code{\link[=hexify_face_centers]{hexify_face_centers()}},
\code{\link[=hexify_forward]{hexify_forward()}},
\code{\link[=hexify_forward_to_face]{hexify_forward_to_face()}},
\code{\link[=hexify_get_precision]{hexify_get_precision()}},
\code{\link[=hexify_inverse]{hexify_inverse()}},
\code{\link[=hexify_projection_stats]{hexify_projection_stats()}},
\code{\link[=hexify_set_verbose]{hexify_set_verbose()}},
\code{\link[=hexify_which_face]{hexify_which_face()}}
}
\concept{projection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructors.R
\name{as_sf}
\alias{as_sf}
\title{Convert HexData to sf Object}
\usage{
as_sf(x, geometry = c("point", "polygon"), ...)
}
\arguments{
\item{x}{A HexData object}

\item{geometry}{Type of geometry: "point" (default) or "polygon"}

\item{...}{Additional arguments (ignored)}
}
\value{
An sf object
}
\description{
Converts a HexData object to an sf spatial features object. Can create
either point geometries (cell centers) or polygon geometries (cell boundaries).
}
\details{
For point geometry, cell centers (cell_cen_lon, cell_cen_lat) are used.
For polygon geometry, cell boundaries are computed using the grid specification.
}
\examples{
df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

# Get sf points
sf_pts <- as_sf(result)

# Get sf polygons
sf_poly <- as_sf(result, geometry = "polygon")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_map.R
\name{plot_world}
\alias{plot_world}
\title{Quick world map plot}
\usage{
plot_world(fill = "gray90", border = "gray50", ...)
}
\arguments{
\item{fill}{Fill color for countries}

\item{border}{Border color for countries}

\item{...}{Additional arguments passed to plot()}
}
\value{
NULL invisibly. Creates a plot as side effect.
}
\description{
Simple wrapper to plot the built-in world map.
}
\examples{
# Quick world map
plot_world()

# Custom colors
plot_world(fill = "lightblue", border = "darkblue")
}
\seealso{
Other visualization:
\code{\link[=hexify_heatmap]{hexify_heatmap()}}
}
\concept{visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_compact.R
\name{hex_compact}
\alias{hex_compact}
\title{Compact Hex Cells}
\usage{
hex_compact(cell_ids, grid)
}
\arguments{
\item{cell_ids}{Cell IDs to compact. For H3 grids, a character vector.
For ISEA grids, a character vector of hierarchical index strings.}

\item{grid}{A HexGridInfo object specifying the grid.}
}
\value{
A character vector of compacted cell IDs. Cells that could be
merged into parents appear as parent IDs at coarser resolution.
}
\description{
Merges child cells into their parent when all children are present.
This is a lossless compression — no spatial information is lost.
The compact representation uses fewer cells to cover the same area.
}
\details{
\strong{H3 backend:} Uses the vendored H3 \code{compactCells} function.

\strong{ISEA backend (aperture 7, Z7 index):} Groups cells by parent index
(dropping the last digit). If all 7 children are present, replaces them
with the parent. Iterates until no further compaction is possible.
}
\examples{
\donttest{
# H3 compaction
g <- hex_grid(resolution = 3, type = "h3")
parent <- "832830fffffffff"
children <- get_children(parent, g)[[1]]
compact <- hex_compact(children, g)
compact  # Should return the parent
}
}
\seealso{
\code{\link[=hex_uncompact]{hex_uncompact()}} for the inverse operation,
\code{\link[=get_parent]{get_parent()}}, \code{\link[=get_children]{get_children()}} for hierarchical operations
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify-package.R
\docType{package}
\name{hexify-package}
\alias{hexify-package}
\title{hexify}
\description{
Core icosahedron and 'Snyder' projection helpers.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://gillescolling.com/hexify/}
  \item Report bugs at \url{https://github.com/gcol33/hexify/issues}
}

}
\author{
\strong{Maintainer}: Gilles Colling \email{gilles.colling051@gmail.com} (\href{https://orcid.org/0000-0003-3070-6066}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Gilles Colling \email{gilles.colling051@gmail.com} (\href{https://orcid.org/0000-0003-3070-6066}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item Uber Technologies, Inc. (H3 C library, src/h3 (Apache-2.0)) [contributor, copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{grid_clip}
\alias{grid_clip}
\title{Clip hexagon grid to polygon boundary}
\usage{
grid_clip(boundary, grid, crop = TRUE)
}
\arguments{
\item{boundary}{An sf/sfc polygon to clip to. Can be a country boundary,
study area, or any polygon geometry.}

\item{grid}{A HexGridInfo object specifying the grid parameters}

\item{crop}{If TRUE (default), cells are cropped to the boundary. If FALSE,
only cells whose centroids fall within the boundary are kept (no cropping).}
}
\value{
sf object with hexagon polygons clipped to the boundary
}
\description{
Creates hexagon polygons clipped to a given polygon boundary. This is useful
for generating grids that conform to country borders, study areas, or other
irregular boundaries.
}
\details{
The function first generates cells covering the boundary polygon, then
clips or filters them. For H3 grids, all cells that overlap the boundary
are included (not just cells whose center falls inside), ensuring full
spatial coverage with no gaps along the boundary edge.

When \code{crop = TRUE}, hexagons are geometrically intersected with the
boundary, which may produce partial hexagons at the edges. When
\code{crop = FALSE}, only complete hexagons whose centroids fall within
the boundary are returned.
}
\examples{
\donttest{
# Get France boundary from built-in world map
france <- hexify_world[hexify_world$name == "France", ]

# Create grid clipped to France
grid <- hex_grid(area_km2 = 2000)
france_grid <- grid_clip(france, grid)

# Plot result
library(ggplot2)
ggplot() +
  geom_sf(data = france, fill = "gray95") +
  geom_sf(data = france_grid, fill = alpha("steelblue", 0.3),
          color = "steelblue") +
  theme_minimal()

# Keep only complete hexagons (no cropping)
france_grid_complete <- grid_clip(france, grid, crop = FALSE)
}
}
\seealso{
\code{\link{grid_rect}} for rectangular grids,
\code{\link{grid_global}} for global grids
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{extract_grid}
\alias{extract_grid}
\title{Extract grid from various objects}
\usage{
extract_grid(x, allow_null = FALSE)
}
\arguments{
\item{x}{Object containing grid info}

\item{allow_null}{If TRUE, return NULL when x is NULL}
}
\value{
HexGridInfo object
}
\description{
Internal function to extract a HexGridInfo from different input types.
Accepts HexGridInfo, HexData, or legacy hexify_grid objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{HexGridInfo-methods}
\alias{HexGridInfo-methods}
\alias{$,HexGridInfo-method}
\alias{names,HexGridInfo-method}
\alias{show,HexGridInfo-method}
\alias{as.list,HexGridInfo-method}
\title{HexGridInfo S4 Methods}
\usage{
\S4method{$}{HexGridInfo}(x, name)

\S4method{names}{HexGridInfo}(x)

\S4method{show}{HexGridInfo}(object)

\S4method{as.list}{HexGridInfo}(x, ...)
}
\arguments{
\item{x}{HexGridInfo object}

\item{name}{Slot name}

\item{object}{HexGridInfo object (for show)}

\item{...}{Additional arguments}
}
\value{
\itemize{
\item \code{$}: The value of the requested slot
\item \code{names}: Character vector of slot names
\item \code{show}: The object, invisibly (called for side effect of printing)
\item \code{as.list}: A named list of slot values
}
}
\description{
S4 methods for HexGridInfo objects. These provide standard R operations
like \code{$}, \code{names()}, \code{show()}, and \code{as.list()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_grid.R
\name{hexify-grid}
\alias{hexify-grid}
\title{Core Grid Construction}
\description{
Core functions for hexify grid construction and validation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_eff_res_to_resolution}
\alias{hexify_eff_res_to_resolution}
\title{Convert effective resolution to index resolution}
\usage{
hexify_eff_res_to_resolution(eff_res)
}
\arguments{
\item{eff_res}{Effective resolution}
}
\value{
Index resolution (integer)
}
\description{
For aperture 3: res_index = 2 * eff_res - 1 for odd index resolutions.
}
\seealso{
Other grid statistics:
\code{\link[=dg_closest_res_to_area]{dg_closest_res_to_area()}},
\code{\link[=dgearthstat]{dgearthstat()}},
\code{\link[=hexify_area_to_eff_res]{hexify_area_to_eff_res()}},
\code{\link[=hexify_compare_resolutions]{hexify_compare_resolutions()}},
\code{\link[=hexify_eff_res_to_area]{hexify_eff_res_to_area()}},
\code{\link[=hexify_resolution_to_eff_res]{hexify_resolution_to_eff_res()}}
}
\concept{grid statistics}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{hexify_world}
\alias{hexify_world}
\title{Simplified World Map}
\format{
An sf object with 177 features and 15 fields:
\describe{
\item{name}{Country short name}
\item{name_long}{Country full name}
\item{admin}{Administrative name}
\item{sovereignt}{Sovereignty}
\item{iso_a2}{ISO 3166-1 alpha-2 country code}
\item{iso_a3}{ISO 3166-1 alpha-3 country code}
\item{iso_n3}{ISO 3166-1 numeric code}
\item{continent}{Continent name}
\item{region_un}{UN region}
\item{subregion}{UN subregion}
\item{region_wb}{World Bank region}
\item{pop_est}{Population estimate}
\item{gdp_md}{GDP in millions USD}
\item{income_grp}{Income group classification}
\item{economy}{Economy type}
\item{geometry}{MULTIPOLYGON geometry in 'WGS84' (EPSG:4326)}
}
}
\source{
Simplified from Natural Earth 1:110m Cultural Vectors
(\url{https://www.naturalearthdata.com/})
}
\usage{
hexify_world
}
\description{
A lightweight sf object containing simplified world country borders,
suitable for use as a basemap when visualizing hexagonal grids.
}
\examples{
library(sf)

# Plot the built-in world map
plot(st_geometry(hexify_world), col = "lightgray", border = "white")

# Filter by continent
europe <- hexify_world[hexify_world$continent == "Europe", ]
plot(st_geometry(europe))
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{cell_area}
\alias{cell_area}
\title{Compute per-cell area in km²}
\usage{
cell_area(cell_id = NULL, grid)
}
\arguments{
\item{cell_id}{Cell IDs to compute area for. For ISEA grids, these are
numeric; for H3 grids, character strings. When \code{grid} is a HexData
object and \code{cell_id} is \code{NULL}, all cell IDs from the data are
used.}

\item{grid}{A HexGridInfo or HexData object.}
}
\value{
Named numeric vector of areas in km², one per \code{cell_id}.
}
\description{
Returns the area of each cell in square kilometers. For ISEA grids, all
cells have the same area (equal-area property). For H3 grids, each cell
has a different geodesic area depending on its location.
}
\details{
For ISEA grids the area is constant across all cells and is read directly
from the grid specification.

For H3 grids the area varies by latitude. The vendored 'H3' library computes
each cell's spherical polygon area as a solid angle, which this function
reads on the grid's body, so a grid built with \code{radius_km} reports that
body's areas.
}
\examples{
# ISEA: constant area
grid <- hex_grid(area_km2 = 1000)
cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
cell_area(cells, grid)

# H3: area varies by location
\donttest{
h3 <- hex_grid(resolution = 5, type = "h3")
h3_cells <- lonlat_to_cell(c(0, 0), c(0, 80), h3)
cell_area(h3_cells, h3)  # equator vs polar — different areas
}
}
\seealso{
\code{\link{hex_grid}} for grid specifications,
\code{\link{h3_crosswalk}} for ISEA/H3 interoperability
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_cell_to_sf.R
\name{hex_corners_to_sf}
\alias{hex_corners_to_sf}
\title{Build an sf POLYGON from six (lon, lat) corner pairs}
\usage{
hex_corners_to_sf(lon, lat, crs = 4326)
}
\arguments{
\item{lon}{numeric vector of length 6 (longitude)}

\item{lat}{numeric vector of length 6 (latitude)}

\item{crs}{integer CRS (default 4326)}
}
\value{
sf object with one POLYGON geometry
}
\description{
Low-level helper to create a single hexagon polygon from corner coordinates.
Most users should use \code{\link{cell_to_sf}} instead.
}
\seealso{
Other sf conversion:
\code{\link[=hexify_cell_to_sf]{hexify_cell_to_sf()}},
\code{\link[=hexify_grid_global]{hexify_grid_global()}},
\code{\link[=hexify_grid_rect]{hexify_grid_rect()}}
}
\concept{sf conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_zonal.R
\name{hex_zonal}
\alias{hex_zonal}
\title{Zonal Statistics for Hex Cells}
\usage{
hex_zonal(raster, grid, fun = "mean", boundary = NULL, cells = NULL)
}
\arguments{
\item{raster}{A \code{terra::SpatRaster} object.}

\item{grid}{A HexGridInfo or HexData object specifying the grid.}

\item{fun}{Summary function name: \code{"mean"} (default), \code{"sum"}, \code{"min"},
\code{"max"}, \code{"sd"}, or \code{"count"}.}

\item{boundary}{Optional sf polygon to limit the analysis extent.}

\item{cells}{Optional cell IDs. If provided, only these cells are included.}
}
\value{
A data.frame with columns \code{cell_id}, plus one column per raster
layer containing the aggregated values.
}
\description{
Computes zonal statistics by aggregating all raster pixels falling
within each hexagonal cell polygon. More accurate than \code{\link[=hex_extract]{hex_extract()}}
but slower because it requires polygon geometries.
}
\details{
Requires the \code{terra} package (in Suggests). The function:
\enumerate{
\item Generates hex polygon geometries via \code{\link[=cell_to_sf]{cell_to_sf()}}
\item Calls \code{terra::extract(raster, polygons, fun = fun)}
\item Joins results back to cell IDs
}

For point-based extraction (faster, at cell centers only), use
\code{\link[=hex_extract]{hex_extract()}} instead.
}
\examples{
\donttest{
if (requireNamespace("terra", quietly = TRUE)) {
  r <- terra::rast(nrows = 100, ncols = 100,
                   xmin = -10, xmax = 10, ymin = 40, ymax = 55)
  terra::values(r) <- runif(10000)
  names(r) <- "temperature"

  g <- hex_grid(area_km2 = 500)
  df <- data.frame(lon = c(0, 5), lat = c(45, 50))
  hd <- hexify(df, lon = "lon", lat = "lat", grid = g)
  hex_zonal(r, hd, fun = "mean")
}
}
}
\seealso{
\code{\link[=hex_extract]{hex_extract()}} for cell-center extraction,
\code{\link[=cell_to_sf]{cell_to_sf()}} for hex polygon generation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_lonlat_to_h_index}
\alias{hexify_lonlat_to_h_index}
\title{Convert longitude/latitude to hexagonal cell hierarchical index}
\usage{
hexify_lonlat_to_h_index(grid, lon, lat)
}
\arguments{
\item{grid}{Grid specification from hexify_grid()}

\item{lon}{Longitude vector in degrees (numeric, -180 to 180)}

\item{lat}{Latitude vector in degrees (numeric, -90 to 90)}
}
\value{
Data frame with columns:
\item{h_index}{Hierarchical index (character string)}
\item{face}{Icosahedron face number (integer, 0-19)}
}
\description{
Converts geographic coordinates (longitude, latitude) to hexagonal cell
hierarchical index strings. These strings encode the face, resolution, and
cell location in a Z-order (Morton code) format.
}
\details{
Most users should use \code{\link{hexify_lonlat_to_cell}} or
\code{\link{hexify_grid_to_cell}} which return DGGRID-compatible
integer cell IDs.

This function returns hierarchical index strings useful for:
\itemize{
\item Understanding the cell's position in the hierarchy
\item Prefix-based spatial queries
\item Parent/child cell relationships
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_globe.R
\name{plot_globe}
\alias{plot_globe}
\title{Plot hexagonized globe}
\usage{
plot_globe(
  area = 50000,
  center = "europe",
  clip_to_land = FALSE,
  land_data = NULL,
  exclude_antarctica = TRUE,
  fill = "#D4B896",
  border = "grey30",
  border_width = 0.2,
  ocean_fill = "white",
  ocean_border = "grey50",
  show_land = clip_to_land,
  land_fill = NA,
  land_border = "grey40",
  land_width = 0.3,
  use_ggplot = NULL,
  return_data = FALSE,
  aperture = 3L
)
}
\arguments{
\item{area}{Cell area in km^2 (passed to \code{\link{hex_grid}})}

\item{center}{Globe center: either a preset name (e.g., "europe") or
numeric vector c(lon, lat). See \code{\link{globe_centers}} for presets.}

\item{clip_to_land}{If TRUE, clip hexagons to land boundaries}

\item{land_data}{Optional sf object for land boundaries. If NULL and
clip_to_land is TRUE, uses rnaturalearth::ne_countries()}

\item{exclude_antarctica}{If TRUE, exclude Antarctica from land clipping}

\item{fill}{Fill color for hexagons (default "#D4B896")}

\item{border}{Border color for hexagons (default "grey30")}

\item{border_width}{Border width for hexagons (default 0.2)}

\item{ocean_fill}{Fill color for ocean/globe background (default "white")}

\item{ocean_border}{Border color for globe circle (default "grey50")}

\item{show_land}{If TRUE, show land boundaries (default TRUE when clipping)}

\item{land_fill}{Fill color for land (default NA, transparent)}

\item{land_border}{Border color for land boundaries (default "grey40")}

\item{land_width}{Border width for land boundaries (default 0.3)}

\item{use_ggplot}{NULL = auto-detect, TRUE = force ggplot2, FALSE = force base}

\item{return_data}{If TRUE, return sf objects instead of plotting}

\item{aperture}{Grid aperture (default 3L)}
}
\value{
If use_ggplot = TRUE: ggplot2 object (can add layers with +)
If use_ggplot = FALSE: NULL invisibly (plots directly)
If return_data = TRUE: list of sf objects (hexagons, land, ocean_circle, crs)
}
\description{
Renders a global hexagonal grid on an orthographic projection with
customizable rotation, land clipping, and styling options.
}
\details{
The function handles several technical challenges:
\itemize{
\item Hexagons on the back side of the globe fail to transform - these are
filtered out gracefully
\item Invalid geometries after projection are repaired with st_buffer(0)
\item Clipping is done in orthographic CRS to avoid topology errors
}
}
\examples{
\donttest{
# Get data for custom plotting (no rendering)
data <- plot_globe(area = 100000, center = "europe", return_data = TRUE)
nrow(data$hexagons)
class(data$ocean_circle)

# Basic usage - Europe-centered globe
plot_globe(area = 80000, center = "europe")
}
}
\seealso{
\code{\link{globe_centers}} for available presets,
\code{\link{grid_global}} for generating global grids without plotting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_globe.R
\name{globe_centers}
\alias{globe_centers}
\title{Globe center presets}
\format{
Named list with elements:
\describe{
\item{europe}{c(10, 50) - Western/Central Europe}
\item{north_america}{c(-100, 45) - USA and Canada}
\item{south_america}{c(-60, -15) - Full continent}
\item{africa}{c(20, 5) - Central Africa}
\item{asia}{c(100, 35) - China, SE Asia, Japan}
\item{oceania}{c(135, -25) - Australia, NZ, Indonesia}
\item{middle_east}{c(45, 25) - Arabian Peninsula, Iran, Turkey}
\item{south_asia}{c(80, 20) - India, Pakistan, Bangladesh}
\item{pacific}{c(-160, -10) - Polynesia, Pacific islands}
\item{caribbean}{c(-70, 18) - Caribbean islands}
\item{arctic}{c(0, 90) - North pole view}
\item{antarctic}{c(0, -90) - South pole view}
}
}
\usage{
globe_centers
}
\description{
Named list of lon/lat coordinates for common globe views.
Used by \code{\link{plot_globe}} when center is specified as a string.
}
\examples{
globe_centers$europe
globe_centers$oceania
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_projection.R
\name{hexify_set_verbose}
\alias{hexify_set_verbose}
\title{Set verbose mode for inverse projection}
\usage{
hexify_set_verbose(verbose = TRUE)
}
\arguments{
\item{verbose}{Logical}
}
\value{
Invisible NULL
}
\description{
When enabled, prints convergence information.
}
\seealso{
Other projection:
\code{\link[=hexify_build_icosa]{hexify_build_icosa()}},
\code{\link[=hexify_face_centers]{hexify_face_centers()}},
\code{\link[=hexify_forward]{hexify_forward()}},
\code{\link[=hexify_forward_to_face]{hexify_forward_to_face()}},
\code{\link[=hexify_get_precision]{hexify_get_precision()}},
\code{\link[=hexify_inverse]{hexify_inverse()}},
\code{\link[=hexify_projection_stats]{hexify_projection_stats()}},
\code{\link[=hexify_set_precision]{hexify_set_precision()}},
\code{\link[=hexify_which_face]{hexify_which_face()}}
}
\concept{projection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{get_children}
\alias{get_children}
\title{Get children cells}
\usage{
get_children(cell_id, grid, levels = 1L)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs}

\item{grid}{A HexGridInfo or HexData object}

\item{levels}{Number of levels down (default 1)}
}
\value{
List of numeric vectors containing child cell IDs
}
\description{
Returns the child cells at a finer resolution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{HexData-methods}
\alias{HexData-methods}
\alias{grid_info,HexData-method}
\alias{cells,HexData-method}
\alias{n_cells,HexData-method}
\alias{nrow,HexData-method}
\alias{ncol,HexData-method}
\alias{dim,HexData-method}
\alias{names,HexData-method}
\alias{$,HexData-method}
\alias{$<-,HexData-method}
\alias{[,HexData-method}
\alias{[[,HexData-method}
\alias{[[<-,HexData,ANY,missing-method}
\alias{show,HexData-method}
\alias{as.data.frame,HexData-method}
\alias{as.list,HexData-method}
\title{HexData S4 Methods}
\usage{
\S4method{grid_info}{HexData}(x)

\S4method{cells}{HexData}(x)

\S4method{n_cells}{HexData}(x)

\S4method{nrow}{HexData}(x)

\S4method{ncol}{HexData}(x)

\S4method{dim}{HexData}(x)

\S4method{names}{HexData}(x)

\S4method{$}{HexData}(x, name)

\S4method{$}{HexData}(x, name) <- value

\S4method{[}{HexData}(x, i, j, ..., drop = FALSE)

\S4method{[[}{HexData}(x, i)

\S4method{[[}{HexData,ANY,missing}(x, i, j) <- value

\S4method{show}{HexData}(object)

\S4method{as.data.frame}{HexData}(x, row.names = NULL, optional = FALSE, ...)

\S4method{as.list}{HexData}(x, ...)
}
\arguments{
\item{x}{HexData object}

\item{name}{Column name}

\item{value}{Replacement value}

\item{i, j}{Row/column indices}

\item{...}{Additional arguments}

\item{drop}{Logical, whether to drop dimensions}

\item{object}{HexData object (for show)}

\item{row.names}{Optional row names}

\item{optional}{Logical (ignored)}
}
\value{
\itemize{
\item \code{grid_info}: HexGridInfo object containing grid specification
\item \code{cells}: Numeric vector of unique cell IDs
\item \code{n_cells}: Integer count of unique cells
\item \code{nrow}, \code{ncol}, \code{dim}: Integer dimensions
\item \code{names}: Character vector of column names (including virtual cell columns)
\item \code{$}, \code{[[}: The requested column or cell data as a vector
\item \verb{$<-}, \verb{[[<-}: The modified HexData object
\item \code{[}: Subsetted HexData object or extracted data
\item \code{show}: The object, invisibly (called for side effect of printing)
\item \code{as.data.frame}: Data frame with original data plus cell columns
\item \code{as.list}: Named list containing data, grid, cell_id, and cell_center
}
}
\description{
S4 methods for HexData objects. These provide standard R operations
for accessing data, subsetting, and conversion.
}
\details{
Unlike \code{[.data.frame}, \code{drop} defaults to \code{FALSE}: selecting
a single column returns a \code{HexData} object (preserving \code{grid}/
\code{cell_id}/\code{cell_center}) rather than dropping to a bare vector.
Pass \code{drop = TRUE} explicitly to get data.frame-style dropping.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_projection.R
\name{hexify_inverse}
\alias{hexify_inverse}
\title{Inverse Snyder projection}
\usage{
hexify_inverse(x, y, face, tol = NULL, max_iters = NULL)
}
\arguments{
\item{x}{X coordinate on face plane}

\item{y}{Y coordinate on face plane}

\item{face}{Face index (0-19)}

\item{tol}{Convergence tolerance (NULL for default)}

\item{max_iters}{Maximum iterations (NULL for default)}
}
\value{
Named numeric vector: c(lon_deg, lat_deg)
}
\description{
Converts face plane coordinates back to geographic coordinates.
}
\examples{
coords <- hexify_inverse(0.5, 0.3, face = 2)
}
\seealso{
Other projection:
\code{\link[=hexify_build_icosa]{hexify_build_icosa()}},
\code{\link[=hexify_face_centers]{hexify_face_centers()}},
\code{\link[=hexify_forward]{hexify_forward()}},
\code{\link[=hexify_forward_to_face]{hexify_forward_to_face()}},
\code{\link[=hexify_get_precision]{hexify_get_precision()}},
\code{\link[=hexify_projection_stats]{hexify_projection_stats()}},
\code{\link[=hexify_set_precision]{hexify_set_precision()}},
\code{\link[=hexify_set_verbose]{hexify_set_verbose()}},
\code{\link[=hexify_which_face]{hexify_which_face()}}
}
\concept{projection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_grid.R
\name{dgverify}
\alias{dgverify}
\title{Verify grid object}
\usage{
dgverify(dggs)
}
\arguments{
\item{dggs}{Grid object to verify (from hexify_grid() or hex_grid())}
}
\value{
TRUE (invisibly) if valid, otherwise throws an error
}
\description{
Validates that a grid object has all required fields and valid values.
This function is called internally by most hexify functions to ensure
grid integrity.
}
\examples{
grid <- hexify_grid(area = 1000, aperture = 3)
dgverify(grid)  # Should pass silently

# Modern HexGridInfo objects are accepted too
dgverify(hex_grid(area_km2 = 1000))

# Invalid grid will throw error
bad_grid <- list(aperture = 5)
try(dgverify(bad_grid))  # Will error
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_methods.R
\name{plot_grid}
\alias{plot_grid}
\title{Plot hexagonal grid clipped to a polygon boundary}
\usage{
plot_grid(
  boundary,
  grid,
  crop = TRUE,
  boundary_fill = "gray95",
  boundary_border = "gray40",
  boundary_lwd = 0.5,
  grid_fill = "steelblue",
  grid_border = "steelblue",
  grid_lwd = 0.3,
  grid_alpha = 0.3,
  title = NULL,
  expand = 0.05
)
}
\arguments{
\item{boundary}{An sf/sfc polygon to clip to (e.g., country boundary)}

\item{grid}{A HexGridInfo object from \code{hex_grid()}}

\item{crop}{If TRUE (default), cells are cropped to boundary. If FALSE,
only complete hexagons within boundary are shown.}

\item{boundary_fill}{Fill color for the boundary polygon (default "gray95")}

\item{boundary_border}{Border color for boundary (default "gray40")}

\item{boundary_lwd}{Line width for boundary (default 0.5)}

\item{grid_fill}{Fill color for grid cells (default "steelblue")}

\item{grid_border}{Border color for grid cells (default "steelblue")}

\item{grid_lwd}{Line width for cell borders (default 0.3)}

\item{grid_alpha}{Transparency for cell fill (0-1, default 0.3)}

\item{title}{Plot title. If NULL (default), auto-generates title with cell area.}

\item{expand}{Expansion factor for plot limits (default 0.05)}
}
\value{
A ggplot object that can be further customized
}
\description{
A convenience function that creates a grid, clips it to a boundary polygon,
and plots the result in a single call.
}
\details{
This is a convenience wrapper around \code{grid_clip()} that handles the
common use case of visualizing a hexagonal grid over a geographic region.
}
\examples{
\donttest{
# Plot grid over France
france <- hexify_world[hexify_world$name == "France", ]
grid <- hex_grid(area_km2 = 2000)
plot_grid(france, grid)

# Customize colors
plot_grid(france, grid,
          grid_fill = "coral", grid_alpha = 0.5,
          boundary_fill = "lightyellow")

# Keep only complete hexagons
plot_grid(france, grid, crop = FALSE)

# Add ggplot2 customizations
library(ggplot2)
plot_grid(france, grid) +
  labs(subtitle = "ISEA3H Discrete Global Grid") +
  theme_void()
}
}
\seealso{
\code{\link{grid_clip}} for the underlying clipping function,
\code{\link{hex_grid}} for grid specification
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_assign.R
\name{hexify_assign}
\alias{hexify_assign}
\title{Assign hex cells ('ISEA3H', aperture 3) for lon/lat}
\usage{
hexify_assign(lon, lat, effective_res, make_polygons = FALSE)
}
\arguments{
\item{lon, lat}{numeric vectors (same length), degrees.}

\item{effective_res}{integer effective resolution (>= 1).}

\item{make_polygons}{logical; if TRUE, return an sf with hex polygons.}
}
\value{
data.frame with id (Z3 index string), face (quad, 0-11),
effective_res, center_lon, center_lat; if make_polygons=TRUE,
an sf with geometry column.
}
\description{
Forward projection -> quantize -> cell centre -> inverse to lon/lat, on the
aperture-3 pipeline used by \code{\link[=hexify]{hexify()}}: points fold into the non-negative
quad frame (\code{\link[=lonlat_to_cell]{lonlat_to_cell()}}), and the centre comes back from the cell
itself (\code{\link[=cell_to_lonlat]{cell_to_lonlat()}}), so the returned centre is the centre of the
assigned cell.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_extract.R
\name{hex_extract}
\alias{hex_extract}
\title{Extract Raster Values at Hex Cell Centers}
\usage{
hex_extract(raster, grid, cells = NULL, boundary = NULL)
}
\arguments{
\item{raster}{A \code{terra::SpatRaster} object.}

\item{grid}{A HexGridInfo or HexData object specifying the grid.}

\item{cells}{Optional cell IDs to extract. If \code{NULL} (default), extracts
at all cell centers from \code{grid} (only works for HexData objects or
if a \code{boundary} is provided).}

\item{boundary}{Optional sf polygon to limit extraction extent.}
}
\value{
A data.frame with columns \code{cell_id}, plus one column per raster
layer.
}
\description{
Samples raster values at hexagonal cell centers. Faster than
\code{\link[=hex_zonal]{hex_zonal()}} because it only queries cell center points, not full
polygons.
}
\details{
Requires the \code{terra} package (in Suggests). The function:
\enumerate{
\item Generates cell centers for the raster extent
\item Calls \code{terra::extract(raster, cell_center_matrix)}
\item Attaches cell IDs
}

For full zonal statistics (aggregating all pixels within each hex polygon),
use \code{\link[=hex_zonal]{hex_zonal()}} instead.
}
\examples{
\donttest{
if (requireNamespace("terra", quietly = TRUE)) {
  # Create a small synthetic raster
  r <- terra::rast(nrows = 10, ncols = 10,
                   xmin = -10, xmax = 10, ymin = 40, ymax = 55)
  terra::values(r) <- runif(100)
  names(r) <- "temperature"

  # Extract at hex cell centers
  g <- hex_grid(area_km2 = 500)
  df <- data.frame(lon = c(0, 5), lat = c(45, 50))
  hd <- hexify(df, lon = "lon", lat = "lat", grid = g)
  hex_extract(r, hd)
}
}
}
\seealso{
\code{\link[=hex_zonal]{hex_zonal()}} for polygon-based zonal statistics,
\code{\link[=hexify]{hexify()}} for creating HexData objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_icosa_tri_to_plane}
\alias{hexify_icosa_tri_to_plane}
\title{Convert Icosa Triangle coordinates to PLANE coordinates}
\usage{
hexify_icosa_tri_to_plane(
  icosa_triangle_face,
  icosa_triangle_x,
  icosa_triangle_y
)
}
\arguments{
\item{icosa_triangle_face}{Triangle face number (0-19), integer or vector}

\item{icosa_triangle_x}{X coordinate on triangle face}

\item{icosa_triangle_y}{Y coordinate on triangle face}
}
\value{
Data frame with columns:
\item{plane_x}{X coordinate in PLANE space (range ~0 to 5.5)}
\item{plane_y}{Y coordinate in PLANE space (range ~0 to 1.73)}
}
\description{
Transforms icosahedral triangle coordinates to the 2D PLANE representation
(unfolded icosahedron). Each triangle is rotated and translated to its
position in the unfolded layout.
}
\details{
Equivalent to 'dggridR' dgPROJTRI_to_PLANE().

The PLANE layout arranges all 20 icosahedral faces into a roughly
rectangular region. Faces 0-4 and 5-9 form the upper row, while
faces 10-14 and 15-19 form the lower row. Adjacent faces share
edges in this representation.
}
\examples{
# Get PLANE coordinates from triangle coordinates
fwd <- hexify_forward(lon = 2.35, lat = 48.86)
plane <- hexify_icosa_tri_to_plane(
  icosa_triangle_face = fwd["face"],
  icosa_triangle_x = fwd["icosa_triangle_x"],
  icosa_triangle_y = fwd["icosa_triangle_y"]
)
print(plane)
}
\seealso{
\code{\link{hexify_cell_to_plane}} for direct cell ID conversion,
\code{\link{hexify_lonlat_to_plane}} for lon/lat to PLANE

Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify.R
\name{hexify}
\alias{hexify}
\title{Assign hexagonal DGGS cell IDs to geographic points}
\usage{
hexify(
  data,
  grid = NULL,
  lon = "lon",
  lat = "lat",
  area_km2 = NULL,
  diagonal = NULL,
  resolution = NULL,
  aperture = 3,
  resround = "nearest",
  radius_km = EARTH_RADIUS_KM
)
}
\arguments{
\item{data}{A data.frame or sf object containing coordinates}

\item{grid}{A HexGridInfo object from \code{hex_grid()}. If provided, overrides
area_km2, resolution, aperture and radius_km parameters.}

\item{lon}{Column name for longitude (ignored if data is sf)}

\item{lat}{Column name for latitude (ignored if data is sf)}

\item{area_km2}{Target cell area in km^2 (mutually exclusive with diagonal).}

\item{diagonal}{Target cell diagonal (long diagonal) in km}

\item{resolution}{Grid resolution (0-30). Alternative to area_km2.}

\item{aperture}{Grid aperture: 3, 4, 7, a mixed family such as "4/3" or
"4/7", or one aperture per resolution level, e.g. \code{c(4, 4, 7, 3)}
(default 3)}

\item{resround}{How to round resolution: "nearest", "up", or "down"}

\item{radius_km}{Radius of the body the grid covers, in kilometers, or a body
name such as "mars" (default Earth). See \code{\link{hex_grid}}.}
}
\value{
A HexData object containing:
\itemize{
\item \code{data}: The original input data (unchanged)
\item \code{grid}: The HexGridInfo specification
\item \code{cell_id}: Numeric vector of cell IDs for each row
\item \code{cell_center}: Matrix of cell center coordinates (lon, lat)
}

Use \code{as.data.frame(result)} to extract the original data.
Use \code{cells(result)} to get unique cell IDs.
Use \code{result@cell_id} to get all cell IDs.
Use \code{result@cell_center} to get cell center coordinates.
}
\description{
Takes a data.frame or sf object with geographic coordinates and returns
a HexData object that stores the original data plus cell assignments.
The original data is preserved unchanged; cell IDs and centers are stored
in separate slots.
}
\details{
For sf objects, coordinates are automatically extracted and transformed to
'WGS84' (EPSG:4326) if needed. The geometry column is preserved.

Either \code{area_km2}, \code{diagonal}, or \code{resolution}
must be provided unless a \code{grid} object is supplied.

The HexData return type (default) stores the grid specification so downstream
functions like \code{plot()}, \code{hexify_cell_to_sf()}, etc. don't need
grid parameters repeated.
}
\section{Grid Specification}{

You can create a grid specification once and reuse it:
\preformatted{
grid <- hex_grid(area_km2 = 1000)
result1 <- hexify(df1, grid = grid)
result2 <- hexify(df2, grid = grid)
}
}

\examples{
# Simple data.frame
df <- data.frame(
  site = c("Vienna", "Paris", "Madrid"),
  lon = c(16.37, 2.35, -3.70),
  lat = c(48.21, 48.86, 40.42)
)

# New recommended workflow: use grid object
grid <- hex_grid(area_km2 = 1000)
result <- hexify(df, grid = grid, lon = "lon", lat = "lat")
print(result)  # Shows grid info
plot(result)   # Plot with default styling

# Direct area specification (grid created internally)
result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)

# Extract plain data.frame
df_result <- as.data.frame(result)

# With sf object (any CRS)
library(sf)
pts <- st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
result_sf <- hexify(pts, area_km2 = 1000)

# Different apertures
result_ap4 <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = 4)

# Mixed aperture (ISEA43H)
result_mixed <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000, aperture = "4/3")
}
\seealso{
\code{\link{hex_grid}} for grid specification,
\code{\link{HexData-class}} for return object details,
\code{\link{as_sf}} for converting to sf

Other hexify main:
\code{\link[=hexify_grid]{hexify_grid()}}
}
\concept{hexify main}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dggrid.R
\name{dggrid_43h_sequence}
\alias{dggrid_43h_sequence}
\title{Create DGGRID 43H aperture sequence}
\usage{
dggrid_43h_sequence(num_ap4, num_ap3)
}
\arguments{
\item{num_ap4}{Number of aperture-4 resolutions}

\item{num_ap3}{Number of aperture-3 resolutions}
}
\value{
Integer vector of aperture sequence
}
\description{
Create an aperture sequence following DGGRID's 43H pattern:
first num_ap4 resolutions use aperture 4, then aperture 3.
}
\examples{
# DGGRID 43H with 2 ap4 resolutions, then 3 ap3 resolutions
seq <- dggrid_43h_sequence(2, 3)  # c(4, 4, 3, 3, 3)
}
\seealso{
Other 'dggridR' compatibility:
\code{\link[=as_dggrid]{as_dggrid()}},
\code{\link[=dggrid_is_compatible]{dggrid_is_compatible()}},
\code{\link[=from_dggrid]{from_dggrid()}}
}
\concept{'dggridR' compatibility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{grid_global}
\alias{grid_global}
\title{Generate a global hexagon grid}
\usage{
grid_global(grid, wrap_dateline = TRUE)
}
\arguments{
\item{grid}{A HexGridInfo object specifying the grid parameters}

\item{wrap_dateline}{Logical. If TRUE (default), antimeridian-crossing
polygons are split at +/-180 degrees. Set to FALSE for orthographic/globe
projections where wrapping creates gaps.}
}
\value{
sf object with hexagon polygons
}
\description{
Creates hexagon polygons covering the entire Earth.
}
\details{
This function generates a complete global grid by sampling points
densely across the globe. For large grids (many small cells),
consider using \code{grid_rect()} to generate regional subsets.
}
\examples{
# Coarse global grid
grid <- hex_grid(area_km2 = 100000)
global <- grid_global(grid)
plot(global)
}
\seealso{
\code{\link{grid_rect}} for regional grids
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_stats.R
\name{hexify_compare_resolutions}
\alias{hexify_compare_resolutions}
\title{Compare grid resolutions}
\usage{
hexify_compare_resolutions(
  aperture = 3,
  res_range = 0:15,
  type = c("isea", "h3"),
  print = FALSE,
  radius_km = EARTH_RADIUS_KM
)
}
\arguments{
\item{aperture}{Grid aperture (3, 4, or 7). Ignored for H3 grids.}

\item{res_range}{Range of resolutions to compare (e.g., 1:10)}

\item{type}{Grid type: "isea" (default) or "h3".}

\item{print}{If TRUE, prints a formatted table to console. If FALSE (default),
returns a data frame.}

\item{radius_km}{Radius of the body, in kilometers, or a body name such as
"mars" (default Earth). See \code{\link{hex_grid}}.}
}
\value{
If print=FALSE: data frame with columns resolution, n_cells,
cell_area_km2, cell_spacing_km, cls_km.
If print=TRUE: invisibly returns the data frame after printing.
}
\description{
Generates a table comparing different resolution levels for a given
grid configuration. Useful for choosing appropriate resolution.
}
\examples{
# Get data frame of resolutions 0-10 for aperture 3
comparison <- hexify_compare_resolutions(aperture = 3, res_range = 0:10)
print(comparison)

# Print formatted table directly
hexify_compare_resolutions(aperture = 3, res_range = 0:10, print = TRUE)

# Find resolution with cells ~1000 km^2
subset(comparison, cell_area_km2 > 900 & cell_area_km2 < 1100)

# Resolutions on another body
hexify_compare_resolutions(aperture = 3, res_range = 0:6, radius_km = "mars")
}
\seealso{
Other grid statistics:
\code{\link[=dg_closest_res_to_area]{dg_closest_res_to_area()}},
\code{\link[=dgearthstat]{dgearthstat()}},
\code{\link[=hexify_area_to_eff_res]{hexify_area_to_eff_res()}},
\code{\link[=hexify_eff_res_to_area]{hexify_eff_res_to_area()}},
\code{\link[=hexify_eff_res_to_resolution]{hexify_eff_res_to_resolution()}},
\code{\link[=hexify_resolution_to_eff_res]{hexify_resolution_to_eff_res()}}
}
\concept{grid statistics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{lonlat_to_cell}
\alias{lonlat_to_cell}
\title{Convert longitude/latitude to cell ID}
\usage{
lonlat_to_cell(lon, lat, grid)
}
\arguments{
\item{lon}{Numeric vector of longitudes in degrees}

\item{lat}{Numeric vector of latitudes in degrees}

\item{grid}{A HexGridInfo or HexData object, or legacy hexify_grid}
}
\value{
Numeric vector of cell IDs
}
\description{
Converts geographic coordinates to DGGS cell IDs using a grid specification.
}
\details{
This function accepts either a HexGridInfo object from \code{hex_grid()} or
a HexData object from \code{hexify()}. If a HexData object is provided,
its grid specification is extracted automatically.
}
\examples{
grid <- hex_grid(area_km2 = 1000)
cells <- lonlat_to_cell(lon = c(0, 10), lat = c(45, 50), grid = grid)

# Or use HexData object
df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
cells <- lonlat_to_cell(lon = 5, lat = 48, grid = result)
}
\seealso{
\code{\link{cell_to_lonlat}} for the inverse operation,
\code{\link{hex_grid}} for creating grid specifications
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_projection.R
\name{hexify_which_face}
\alias{hexify_which_face}
\title{Determine which face contains a point}
\usage{
hexify_which_face(lon, lat)
}
\arguments{
\item{lon}{Longitude in degrees}

\item{lat}{Latitude in degrees}
}
\value{
Integer face index (0-19)
}
\description{
Returns the icosahedral face index (0-19) containing the given coordinates.
}
\examples{
face <- hexify_which_face(16.37, 48.21)
}
\seealso{
Other projection:
\code{\link[=hexify_build_icosa]{hexify_build_icosa()}},
\code{\link[=hexify_face_centers]{hexify_face_centers()}},
\code{\link[=hexify_forward]{hexify_forward()}},
\code{\link[=hexify_forward_to_face]{hexify_forward_to_face()}},
\code{\link[=hexify_get_precision]{hexify_get_precision()}},
\code{\link[=hexify_inverse]{hexify_inverse()}},
\code{\link[=hexify_projection_stats]{hexify_projection_stats()}},
\code{\link[=hexify_set_precision]{hexify_set_precision()}},
\code{\link[=hexify_set_verbose]{hexify_set_verbose()}}
}
\concept{projection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dggrid_compat.R
\name{dggrid_is_compatible}
\alias{dggrid_is_compatible}
\title{Validate 'dggridR' grid compatibility with hexify}
\usage{
dggrid_is_compatible(dggs, strict = TRUE)
}
\arguments{
\item{dggs}{A 'dggridR' grid object}

\item{strict}{If TRUE (default), throw errors for incompatibilities.
If FALSE, return FALSE instead of throwing errors.}
}
\value{
TRUE if compatible, FALSE if not compatible (when strict=FALSE)
}
\description{
Checks whether a 'dggridR' grid object is compatible with hexify functions.
Returns TRUE if compatible, or throws an error describing incompatibilities.
}
\seealso{
Other 'dggridR' compatibility:
\code{\link[=as_dggrid]{as_dggrid()}},
\code{\link[=dggrid_43h_sequence]{dggrid_43h_sequence()}},
\code{\link[=from_dggrid]{from_dggrid()}}
}
\concept{'dggridR' compatibility}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_stats.R
\name{dgearthstat}
\alias{dgearthstat}
\title{Get grid statistics for whole-body coverage}
\usage{
dgearthstat(dggs)
}
\arguments{
\item{dggs}{Grid specification from hexify_grid()}
}
\value{
List with components:
\item{area_km}{Total surface area of the body in km^2}
\item{n_cells}{Total number of cells at this resolution}
\item{cell_area_km2}{Average cell area in km^2}
\item{cell_spacing_km}{Average distance between cell centers in km}
\item{resolution}{Resolution level}
\item{aperture}{Grid aperture}
}
\description{
Calculates statistics about the hexagonal grid at the current resolution,
including total number of cells, cell area, and cell spacing. Figures follow
the radius the grid carries, so a grid built with \code{radius_km} reports
that body's areas and distances.
}
\examples{
grid <- hexify_grid(area = 1000, aperture = 3)
stats <- dgearthstat(grid)

print(sprintf("Resolution \%d has \%.0f cells",
              stats$resolution, stats$n_cells))
print(sprintf("Average cell area: \%.2f km^2",
              stats$cell_area_km2))
print(sprintf("Average cell spacing: \%.2f km",
              stats$cell_spacing_km))
}
\seealso{
Other grid statistics:
\code{\link[=dg_closest_res_to_area]{dg_closest_res_to_area()}},
\code{\link[=hexify_area_to_eff_res]{hexify_area_to_eff_res()}},
\code{\link[=hexify_compare_resolutions]{hexify_compare_resolutions()}},
\code{\link[=hexify_eff_res_to_area]{hexify_eff_res_to_area()}},
\code{\link[=hexify_eff_res_to_resolution]{hexify_eff_res_to_resolution()}},
\code{\link[=hexify_resolution_to_eff_res]{hexify_resolution_to_eff_res()}}
}
\concept{grid statistics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{cells}
\alias{cells}
\title{Get Cell IDs}
\usage{
cells(x)
}
\arguments{
\item{x}{A HexData object}
}
\value{
A vector of cell IDs
}
\description{
Extract the unique cell IDs present in a HexData object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_summarize.R
\name{hex_summarize}
\alias{hex_summarize}
\title{Summarize Data by Hex Cell}
\usage{
hex_summarize(hex_data, ..., .fns = NULL, geometry = FALSE)
}
\arguments{
\item{hex_data}{A HexData object (from \code{\link[=hexify]{hexify()}}).}

\item{...}{Named summary expressions (tidyeval). Each expression is
evaluated per cell group with the group's columns available by name.
Examples: \code{mean_temp = mean(temperature)},
\code{n_species = length(unique(species))},
\code{max_elev = max(elevation, na.rm = TRUE)}. Provide either \code{...} or
\code{.fns}, not both.}

\item{.fns}{Optional named list of functions for formula-style aggregation.
Example: \code{.fns = list(mean_temp = ~mean(temperature))}. Provide either
\code{...} or \code{.fns}, not both.}

\item{geometry}{Logical. If \code{TRUE}, attach cell center points as an
sf geometry column (requires sf). Default \code{FALSE}.}
}
\value{
A data.frame with columns:
\describe{
\item{cell_id}{Unique cell identifier}
\item{cell_cen_lon, cell_cen_lat}{Cell center coordinates}
\item{cell_area_km2}{Cell area in km^2}
\item{n_points}{Number of data points in this cell}
\item{...}{User-defined summary columns}
}
If \code{geometry = TRUE}, returns an sf object with POINT geometry.
}
\description{
Aggregates data within each hexagonal cell, similar to
\code{dplyr::group_by(cell_id) |> summarize(...)}. Returns a data.frame
with one row per unique cell, including cell center coordinates and area.
}
\details{
The function works entirely in R (no C++ needed). It groups by \code{cell_id}
and evaluates the summary expressions within each group.

If no summary expressions are provided, returns cell counts only.
}
\examples{
\donttest{
df <- data.frame(
  lon = runif(100, -10, 10),
  lat = runif(100, 40, 55),
  temperature = rnorm(100, 15, 5),
  species = sample(letters[1:5], 100, replace = TRUE)
)
hd <- hexify(df, lon = "lon", lat = "lat", area_km2 = 500)

# Count points per cell
hex_summarize(hd)

# Custom summaries
hex_summarize(hd, mean_temp = mean(temperature),
                  n_species = length(unique(species)))
}
}
\seealso{
\code{\link[=hexify]{hexify()}} for creating HexData objects,
\code{\link[=get_neighbors]{get_neighbors()}} for finding neighboring cells
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_z7_canonical}
\alias{hexify_z7_canonical}
\title{Get canonical form of Z7 index}
\usage{
hexify_z7_canonical(index, max_iterations = 128L)
}
\arguments{
\item{index}{A length-one Z7 index string.}

\item{max_iterations}{Maximum number of decode/encode iterations. This is a
safety bound for legacy indices; the default is 128.}
}
\value{
A length-one character string containing the stable index.
}
\description{
Decodes and re-encodes a Z7 index until it reaches a stable form. Current Z7
indices are bijective, so every valid index is already canonical and this
function normally returns its input unchanged. It remains available for
validating or normalizing indices created by older hexify versions.
}
\examples{
# Valid Z7 indices are stable
hexify_z7_canonical("110001")

cell <- hexify_index_to_cell("110001", aperture = 7)
identical(
  hexify_cell_to_index(cell$face, cell$i, cell$j,
    cell$resolution, aperture = 7
  ),
  "110001"
)
}
\seealso{
Other hierarchical index:
\code{\link[=hexify_cell_to_index]{hexify_cell_to_index()}},
\code{\link[=hexify_compare_indices]{hexify_compare_indices()}},
\code{\link[=hexify_default_index_type]{hexify_default_index_type()}},
\code{\link[=hexify_get_children]{hexify_get_children()}},
\code{\link[=hexify_get_parent]{hexify_get_parent()}},
\code{\link[=hexify_get_resolution]{hexify_get_resolution()}},
\code{\link[=hexify_index_to_cell]{hexify_index_to_cell()}},
\code{\link[=hexify_index_to_lonlat]{hexify_index_to_lonlat()}},
\code{\link[=hexify_is_valid_index_type]{hexify_is_valid_index_type()}},
\code{\link[=hexify_lonlat_to_index]{hexify_lonlat_to_index()}}
}
\concept{hierarchical index}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructors.R
\name{as_tibble.HexData}
\alias{as_tibble.HexData}
\title{Convert HexData to tibble}
\usage{
as_tibble.HexData(x, ...)
}
\arguments{
\item{x}{A HexData object}

\item{...}{Additional arguments (ignored)}
}
\value{
A tibble
}
\description{
Convert HexData to tibble
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\docType{class}
\name{HexData-class}
\alias{HexData-class}
\title{HexData Class}
\description{
An S4 class representing hexified data. Contains the original user data
plus cell assignments from the hexification process.
}
\details{
HexData objects are created by \code{\link{hexify}}. The original data
is preserved in the \code{data} slot, while cell assignments are stored
separately in \code{cell_id} and \code{cell_center}.

Use \code{as.data.frame()} to get a combined data frame with cell columns.
}
\section{Slots}{

\describe{
\item{\code{data}}{Data frame or sf object. The original user data (untouched).}

\item{\code{grid}}{HexGridInfo object. The grid specification used.}

\item{\code{cell_id}}{Cell IDs for each row of data. Numeric for ISEA grids,
character for H3 grids.}

\item{\code{cell_center}}{Matrix. Two-column matrix (lon, lat) of cell centers.}
}}

\seealso{
\code{\link{hexify}} for creating HexData objects,
\code{\link{HexGridInfo-class}} for grid specifications
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_distance.R
\name{hex_distance}
\alias{hex_distance}
\title{Grid Distance Between Cells}
\usage{
hex_distance(cell_a, cell_b, grid)
}
\arguments{
\item{cell_a, cell_b}{Cell IDs. Must be the same length (pairs) or one
of them length 1 (broadcast). For H3 grids, character vectors.
For ISEA grids, numeric vectors.}

\item{grid}{A HexGridInfo or HexData object specifying the grid.}
}
\value{
An integer vector of grid distances. \code{NA} where the distance
cannot be computed (e.g., pentagon path issues in H3).
}
\description{
Computes the grid distance (minimum number of hops) between pairs
of hexagonal cells. This is a discrete distance measured in cell steps,
not a geodesic distance.
}
\details{
\strong{H3 backend:} Uses the vendored H3 \code{gridDistance} function.
Returns the exact shortest path length in cell hops.

\strong{ISEA backend:} For cells on the same quad, computes the cube-coordinate
distance: \verb{max(|di|, |dj|, |di + dj|)}. Cross-quad distances use BFS
expansion and may return \code{NA} for very distant cells.

For geodesic (geographic) distances between cell centers, convert to
lon/lat with \code{\link[=cell_to_lonlat]{cell_to_lonlat()}} and use \code{sf::st_distance()}.
}
\examples{
\donttest{
# H3 grid distance
g <- hex_grid(resolution = 5, type = "h3")
a <- lonlat_to_cell(10, 50, g)
b <- lonlat_to_cell(10.1, 50.1, g)
hex_distance(a, b, g)
}
}
\seealso{
\code{\link[=get_neighbors]{get_neighbors()}} for finding cells within a given distance,
\code{\link[=cell_to_lonlat]{cell_to_lonlat()}} for geographic coordinates
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{is_hex_grid}
\alias{is_hex_grid}
\title{Check if object is HexGridInfo}
\usage{
is_hex_grid(x)
}
\arguments{
\item{x}{Object to check}
}
\value{
Logical
}
\description{
Check if object is HexGridInfo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_grid_cell_to_lonlat}
\alias{hexify_grid_cell_to_lonlat}
\title{Convert cell ID to longitude/latitude using a grid object}
\usage{
hexify_grid_cell_to_lonlat(grid, cell_id)
}
\arguments{
\item{grid}{Grid specification from hexify_grid()}

\item{cell_id}{Numeric vector of cell IDs (1-based)}
}
\value{
Data frame with lon_deg and lat_deg columns
}
\description{
Grid-based wrapper for \code{\link{hexify_cell_to_lonlat}}. Converts
DGGRID-compatible cell IDs back to cell center coordinates
using the resolution and aperture from a grid object.
}
\examples{
grid <- hexify_grid(area = 1000, aperture = 3)
cell_ids <- hexify_grid_to_cell(grid, lon = 5, lat = 45)
coords <- hexify_grid_cell_to_lonlat(grid, cell_ids)
}
\seealso{
\code{\link{cell_to_lonlat}} for the recommended S4 interface,
\code{\link{hexify_grid_to_cell}} for the forward operation

Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_icosa_tri_to_quad_ij}
\alias{hexify_icosa_tri_to_quad_ij}
\title{Convert Icosa Triangle to Quad IJ coordinates}
\usage{
hexify_icosa_tri_to_quad_ij(
  icosa_triangle_face,
  icosa_triangle_x,
  icosa_triangle_y,
  resolution,
  aperture = 3L
)
}
\arguments{
\item{icosa_triangle_face}{Triangle face number (0-19)}

\item{icosa_triangle_x}{X coordinate on triangle face}

\item{icosa_triangle_y}{Y coordinate on triangle face}

\item{resolution}{Grid resolution level}

\item{aperture}{Grid aperture: 3, 4, or 7}
}
\value{
List with components:
\item{quad}{Quad number (0-11)}
\item{i}{Integer cell index along first axis}
\item{j}{Integer cell index along second axis}
}
\description{
Converts icosahedral triangle coordinates directly to Quad IJ,
combining the transformation and quantization steps.
}
\examples{
# First get triangle coordinates from lon/lat
fwd <- hexify_forward(lon = 2.35, lat = 48.86)

# Then convert to quad IJ
quad_ij <- hexify_icosa_tri_to_quad_ij(
  icosa_triangle_face = fwd["face"],
  icosa_triangle_x = fwd["icosa_triangle_x"],
  icosa_triangle_y = fwd["icosa_triangle_y"],
  resolution = 10,
  aperture = 3
)
print(quad_ij)
}
\seealso{
Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_h_index_to_lonlat}
\alias{hexify_h_index_to_lonlat}
\title{Convert hierarchical index strings to longitude/latitude centers}
\usage{
hexify_h_index_to_lonlat(grid, h_index)
}
\arguments{
\item{grid}{Grid specification from hexify_grid()}

\item{h_index}{Hierarchical index strings (character vector)}
}
\value{
Data frame with columns:
\item{lon}{Longitude in degrees}
\item{lat}{Latitude in degrees}
}
\description{
Converts hierarchical cell index strings back to geographic coordinates,
returning the center point of each cell. This is the inverse operation
of hexify_lonlat_to_h_index().
}
\details{
Most users should use \code{\link{hexify_cell_to_lonlat}} or
\code{\link{hexify_grid_cell_to_lonlat}} which work with DGGRID-compatible
integer cell IDs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_grid.R
\name{calculate_resolution_for_area}
\alias{calculate_resolution_for_area}
\title{Calculate resolution for target area}
\usage{
calculate_resolution_for_area(
  target_area_km2,
  aperture = 3,
  radius_km = EARTH_RADIUS_KM
)
}
\arguments{
\item{target_area_km2}{Target area in square kilometers}

\item{aperture}{Aperture (3, 4, or 7)}

\item{radius_km}{Radius of the body, in kilometers}
}
\value{
Resolution level
}
\description{
Uses the 'ISEA3H'/'ISEA4H'/'ISEA7H' cell count formula
N = 10 * aperture^res + 2, which matches 'dggridR' resolution numbering
exactly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_index_to_cell}
\alias{hexify_index_to_cell}
\title{Convert index string to cell coordinates}
\usage{
hexify_index_to_cell(
  index,
  aperture = 3L,
  index_type = c("auto", "z3", "z7", "zorder")
)
}
\arguments{
\item{index}{Index string}

\item{aperture}{Aperture (3, 4, or 7)}

\item{index_type}{Index encoding used. Default "auto" infers from aperture.}
}
\value{
A list with \code{face}, \code{i}, \code{j}, and \code{resolution}.
}
\description{
Decodes a hierarchical index string back to its cell coordinates and
resolution. For Z7, valid indices round-trip exactly through
\code{\link[=hexify_cell_to_index]{hexify_cell_to_index()}}.
}
\examples{
cell <- hexify_index_to_cell("0012012", aperture = 3)

z7_cell <- hexify_index_to_cell("110001", aperture = 7)
hexify_cell_to_index(z7_cell$face, z7_cell$i, z7_cell$j,
  z7_cell$resolution, aperture = 7
)
}
\seealso{
Other hierarchical index:
\code{\link[=hexify_cell_to_index]{hexify_cell_to_index()}},
\code{\link[=hexify_compare_indices]{hexify_compare_indices()}},
\code{\link[=hexify_default_index_type]{hexify_default_index_type()}},
\code{\link[=hexify_get_children]{hexify_get_children()}},
\code{\link[=hexify_get_parent]{hexify_get_parent()}},
\code{\link[=hexify_get_resolution]{hexify_get_resolution()}},
\code{\link[=hexify_index_to_lonlat]{hexify_index_to_lonlat()}},
\code{\link[=hexify_is_valid_index_type]{hexify_is_valid_index_type()}},
\code{\link[=hexify_lonlat_to_index]{hexify_lonlat_to_index()}},
\code{\link[=hexify_z7_canonical]{hexify_z7_canonical()}}
}
\concept{hierarchical index}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_stats.R
\name{hexify-stats}
\alias{hexify-stats}
\title{Grid Statistics}
\description{
Functions for calculating grid statistics and utilities
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructors.R
\name{new_hex_data}
\alias{new_hex_data}
\title{Create a HexData Object (Internal)}
\usage{
new_hex_data(data, grid, cell_id, cell_center)
}
\arguments{
\item{data}{Data frame or sf object (original user data, untouched)}

\item{grid}{HexGridInfo object}

\item{cell_id}{Numeric vector of cell IDs for each row}

\item{cell_center}{Matrix with columns lon, lat for cell centers}
}
\value{
A HexData object
}
\description{
Internal constructor for HexData objects. Users should use \code{hexify()}
instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_neighbors.R
\name{get_neighbors}
\alias{get_neighbors}
\title{Get Neighboring Cells}
\usage{
get_neighbors(cell_id, grid, k = 1L, include_self = FALSE, distances = FALSE)
}
\arguments{
\item{cell_id}{Cell IDs to find neighbors for. Numeric vector for ISEA
grids, character vector for H3 grids.}

\item{grid}{A HexGridInfo or HexData object specifying the grid.}

\item{k}{Integer. Ring distance (default 1). \code{k = 1} returns immediate
neighbors, \code{k = 2} includes neighbors-of-neighbors, etc.}

\item{include_self}{Logical. If \code{TRUE}, include the input cell in the
result (default \code{FALSE}).}

\item{distances}{Logical. If \code{TRUE}, return a data.frame with cell IDs
and their ring distance from the origin (default \code{FALSE}).}
}
\value{
If \code{distances = FALSE} (default): a list of cell ID vectors, one
per input cell. If \code{distances = TRUE}: a list of data.frames with columns
\code{cell_id} and \code{ring_distance}.
}
\description{
Returns the k-ring (disk) of cells neighboring the input cells.
For \code{k = 1}, returns the immediate 6 neighbors (5 for pentagons).
For \code{k > 1}, returns all cells within \code{k} grid hops.
}
\details{
For \strong{ISEA grids}, neighbors are computed using axial coordinate offsets
in the quad IJ space. Cells at quad boundaries are handled by reprojection
through lon/lat coordinates.

For \strong{H3 grids}, neighbors use the vendored H3 \code{gridDisk} /
\code{gridDiskDistances} functions.

Pentagon cells (the 12 icosahedron vertices) have only 5 neighbors instead
of the usual 6.
}
\examples{
\donttest{
# ISEA grid neighbors
g <- hex_grid(area_km2 = 1000)
cell <- lonlat_to_cell(10, 50, g)
nbrs <- get_neighbors(cell, g)
nbrs[[1]]

# H3 grid neighbors
g_h3 <- hex_grid(resolution = 5, type = "h3")
cell_h3 <- lonlat_to_cell(10, 50, g_h3)
get_neighbors(cell_h3, g_h3, k = 2)

# With distances
get_neighbors(cell, g, k = 2, distances = TRUE)
}
}
\seealso{
\code{\link[=hexify]{hexify()}} for creating HexData objects,
\code{\link[=hex_distance]{hex_distance()}} for grid distances between cells
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_cell_to_sf.R
\name{hexify_cell_to_sf}
\alias{hexify_cell_to_sf}
\title{Convert cell IDs to sf polygons}
\usage{
hexify_cell_to_sf(
  cell_id,
  resolution = NULL,
  aperture = NULL,
  return_sf = TRUE,
  grid = NULL,
  wrap_dateline = TRUE
)
}
\arguments{
\item{cell_id}{Integer vector of cell identifiers}

\item{resolution}{Grid resolution level. Can be omitted if grid is provided.}

\item{aperture}{Grid aperture: 3, 4, or 7. Can be omitted if grid is provided.}

\item{return_sf}{Logical. If TRUE (default), returns sf object with polygon
geometries. If FALSE, returns data frame with vertex coordinates.}

\item{grid}{Optional HexGridInfo object. If provided, resolution and aperture
are extracted from it.}

\item{wrap_dateline}{Logical. If TRUE (default), calls
\code{sf::st_wrap_dateline()} to split antimeridian-crossing polygons.
Set to FALSE for orthographic/globe projections where wrapping creates gaps.}
}
\value{
If return_sf = TRUE: sf object with columns:
\item{cell_id}{Cell identifier}
\item{geometry}{POLYGON geometry (sfc_POLYGON)}

If return_sf = FALSE: data frame with columns:
\item{cell_id}{Cell identifier}
\item{lon}{Vertex longitude}
\item{lat}{Vertex latitude}
\item{order}{Vertex order (1-7, 7 closes the polygon)}
}
\description{
Creates polygon geometries for hexagonal grid cells from their cell IDs.
Returns an sf object by default, or a data frame for lightweight workflows.
}
\details{
This function uses a native C++ implementation that is significantly faster
than 'dggridR' polygon generation, especially for large numbers of cells.

For the recommended S4 interface, use \code{\link{cell_to_sf}} instead.
}
\examples{
library(hexify)

# Generate some data with hex cells
df <- data.frame(lon = c(0, 5, 10), lat = c(45, 46, 45))
result <- hexify(df, lon = "lon", lat = "lat", area = 1000)

# Get polygons as sf object (using HexData)
polys <- cell_to_sf(grid = result)

# Or with explicit parameters
polys <- hexify_cell_to_sf(result@cell_id, resolution = 10, aperture = 3)

# Plot with sf
library(sf)
plot(st_geometry(polys), col = "lightblue", border = "blue")
}
\seealso{
\code{\link{cell_to_sf}} for the recommended S4 interface

Other sf conversion:
\code{\link[=hex_corners_to_sf]{hex_corners_to_sf()}},
\code{\link[=hexify_grid_global]{hexify_grid_global()}},
\code{\link[=hexify_grid_rect]{hexify_grid_rect()}}
}
\concept{sf conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_lonlat_to_quad_ij}
\alias{hexify_lonlat_to_quad_ij}
\title{Convert longitude/latitude to Quad IJ coordinates}
\usage{
hexify_lonlat_to_quad_ij(lon, lat, resolution, aperture = 3L)
}
\arguments{
\item{lon}{Longitude in degrees (-180 to 180)}

\item{lat}{Latitude in degrees (-90 to 90)}

\item{resolution}{Grid resolution level (0-30)}

\item{aperture}{Grid aperture: 3, 4, or 7}
}
\value{
List with components:
\item{quad}{Quad number (0-11)}
\item{i}{Integer cell index along first axis}
\item{j}{Integer cell index along second axis}
\item{icosa_triangle_face}{Source icosahedral face (0-19)}
\item{icosa_triangle_x}{X coordinate on triangle face}
\item{icosa_triangle_y}{Y coordinate on triangle face}
}
\description{
Converts geographic coordinates to the intermediate Quad IJ representation
used internally by ISEA DGGS. Returns the quad number (0-11) and integer
cell indices (i, j) within that quad.
}
\details{
The 20 icosahedral triangle faces are grouped into 12 quads:
\itemize{
\item Quad 0: North polar region
\item Quads 1-5: Upper hemisphere rhombi
\item Quads 6-10: Lower hemisphere rhombi
\item Quad 11: South polar region
}
}
\examples{
# Get Quad IJ coordinates for Paris
result <- hexify_lonlat_to_quad_ij(lon = 2.35, lat = 48.86,
                                    resolution = 10, aperture = 3)
print(result)
}
\seealso{
Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_area_to_eff_res}
\alias{hexify_area_to_eff_res}
\title{Convert area to effective resolution}
\usage{
hexify_area_to_eff_res(area_km2)
}
\arguments{
\item{area_km2}{Target area in km^2}
}
\value{
Effective resolution (may be fractional)
}
\description{
Finds the effective resolution that gives approximately the target area.
}
\seealso{
Other grid statistics:
\code{\link[=dg_closest_res_to_area]{dg_closest_res_to_area()}},
\code{\link[=dgearthstat]{dgearthstat()}},
\code{\link[=hexify_compare_resolutions]{hexify_compare_resolutions()}},
\code{\link[=hexify_eff_res_to_area]{hexify_eff_res_to_area()}},
\code{\link[=hexify_eff_res_to_resolution]{hexify_eff_res_to_resolution()}},
\code{\link[=hexify_resolution_to_eff_res]{hexify_resolution_to_eff_res()}}
}
\concept{grid statistics}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructors.R
\name{hex_grid}
\alias{hex_grid}
\title{Create a Hexagonal Grid Specification}
\usage{
hex_grid(
  area_km2 = NULL,
  resolution = NULL,
  aperture = 3,
  type = c("isea", "h3"),
  resround = "nearest",
  crs = NULL,
  radius_km = EARTH_RADIUS_KM
)
}
\arguments{
\item{area_km2}{Target cell area in square kilometers. Mutually exclusive
with \code{resolution}.}

\item{resolution}{Grid resolution level (0-30 for ISEA, 0-15 for H3).
Mutually exclusive with \code{area_km2}. For H3, typical use cases by
resolution:
\itemize{
\item 0-3: continental/country scale
\item 4-7: regional/city scale
\item 8-10: neighborhood/block scale (FCC uses 8-9)
\item 11-15: building/sub-meter scale
}}

\item{aperture}{Grid aperture: 3 (default), 4, 7, a mixed family such as
"4/3", "4/7" or "7/4", or one aperture per resolution level as a vector,
e.g. \code{c(4, 4, 7, 3)}. A family name refines by the first aperture for
the first \code{floor(resolution / 2)} levels and by the second for the
rest, which is how DGGRID arranges ISEA43H. A per-level vector needs
\code{resolution} rather than \code{area_km2}. Ignored for H3 grids (fixed
at 7).}

\item{type}{Grid type: "isea" (default) or "h3".}

\item{resround}{Resolution rounding when using \code{area_km2}:
"nearest" (default), "up", or "down".}

\item{crs}{Coordinate reference system: an EPSG code, or a 'PROJ' or 'WKT'
string. Defaults to 'WGS84' on Earth, and to a longlat CRS on the sphere of
\code{radius_km} on any other body, which has no EPSG code to name it.}

\item{radius_km}{Radius of the body the grid covers, in kilometers, or the
name of a body: "mercury", "venus", "earth" (default), "moon", "mars",
"ceres", "jupiter", "io", "europa", "ganymede", "callisto", "saturn",
"enceladus", "titan", "uranus", "neptune", "pluto".}
}
\value{
A HexGridInfo object containing the grid specification.
}
\description{
Creates a HexGridInfo object that stores all parameters needed for hexagonal
grid operations. Use this to define the grid once and pass it to all
downstream functions.
}
\details{
Exactly one of \code{area_km2} or \code{resolution} must be provided.

When \code{area_km2} is provided, the resolution is calculated automatically
using the cell count formula: N = 10 * aperture^res + 2 (ISEA) or by
matching the closest H3 resolution.

H3 grids use the Uber H3 hierarchical hexagonal system. Unlike ISEA grids,
H3 cells are NOT exactly equal-area (area varies by ~3-5\\% depending on
location).
}
\section{Other Bodies}{


A grid is a partition of the sphere, and \code{radius_km} sets the sphere it
is measured on. Cell geometry -- which cell a coordinate lands in, where cell
centres and corners sit, the hierarchy, the neighbours -- is angular and
identical on every body; the radius sets the kilometer figures: cell area,
diagonal, spacing, and the resolution that \code{area_km2} picks. Earth's
area comes from the 'WGS84' ellipsoid, every other radius gives the sphere
area 4\emph{pi}r^2.

\preformatted{
mars <- hex_grid(area_km2 = 1000, radius_km = "mars")
hex_grid(resolution = 8, radius_km = 3389.5)   # the same grid
}

Both backends take a radius. 'H3' reports a cell's area as its solid angle
times Earth's radius squared, so another radius scales those areas by the
square of the radius ratio, exactly. One caveat carries: an 'H3' cell ID
names a position in 'H3''s topology, which 'Uber''s 'H3' reads on Earth, so
the IDs of a grid on another body are that topology on that body and are not
interchangeable with Earth 'H3' data.
}

\section{One Grid, Many Datasets}{


A HexGridInfo acts as a shared spatial reference system - like a CRS, but
discrete and equal-area. Define the grid once, then attach multiple
datasets without repeating parameters:

\preformatted{
# Step 1: Define the grid once
grid <- hex_grid(area_km2 = 1000)

# Step 2: Attach multiple datasets to the same grid
birds <- hexify(bird_obs, lon = "longitude", lat = "latitude", grid = grid)
mammals <- hexify(mammal_obs, lon = "lon", lat = "lat", grid = grid)
climate <- hexify(weather_stations, lon = "x", lat = "y", grid = grid)

# No aperture, resolution, or area needed after step 1 - the grid
# travels with the data.

# Step 3: Work at the cell level
# Once hexified, lon/lat no longer matter - cell_id is the shared key
bird_counts <- aggregate(species ~ cell_id, data = as.data.frame(birds), length)
mammal_richness <- aggregate(species ~ cell_id, data = as.data.frame(mammals),
                             function(x) length(unique(x)))

# Join datasets by cell_id - guaranteed to align because same grid
combined <- merge(bird_counts, mammal_richness, by = "cell_id")

# Step 4: Visual confirmation
# All datasets produce identical grid overlays
plot(birds)   # See the grid
plot(mammals) # Same grid, different data
}
}

\examples{
# Create grid by target area
grid <- hex_grid(area_km2 = 1000)
print(grid)

# Create grid by resolution
grid <- hex_grid(resolution = 8, aperture = 3)

# Create grid with different aperture
grid4 <- hex_grid(area_km2 = 500, aperture = 4)

# Create mixed aperture grid
grid43 <- hex_grid(area_km2 = 1000, aperture = "4/3")

# Mix in aperture 7, either as a family or level by level
grid47 <- hex_grid(area_km2 = 1000, aperture = "4/7")
grid_seq <- hex_grid(resolution = 4, aperture = c(4, 4, 7, 3))

# Grid on another body, by name or by radius
mars <- hex_grid(area_km2 = 1000, radius_km = "mars")
titan <- hex_grid(resolution = 6, radius_km = 2574.76)

# Use grid in hexify
df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
result <- hexify(df, lon = "lon", lat = "lat", grid = grid)
}
\seealso{
\code{\link{hexify}} for assigning points to cells,
\code{\link{HexGridInfo-class}} for class documentation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_compare_indices}
\alias{hexify_compare_indices}
\title{Compare two indices}
\usage{
hexify_compare_indices(idx1, idx2)
}
\arguments{
\item{idx1}{First index string}

\item{idx2}{Second index string}
}
\value{
Integer: -1 if idx1 < idx2, 0 if equal, 1 if idx1 > idx2
}
\description{
Lexicographic comparison of two index strings.
}
\seealso{
Other hierarchical index:
\code{\link[=hexify_cell_to_index]{hexify_cell_to_index()}},
\code{\link[=hexify_default_index_type]{hexify_default_index_type()}},
\code{\link[=hexify_get_children]{hexify_get_children()}},
\code{\link[=hexify_get_parent]{hexify_get_parent()}},
\code{\link[=hexify_get_resolution]{hexify_get_resolution()}},
\code{\link[=hexify_index_to_cell]{hexify_index_to_cell()}},
\code{\link[=hexify_index_to_lonlat]{hexify_index_to_lonlat()}},
\code{\link[=hexify_is_valid_index_type]{hexify_is_valid_index_type()}},
\code{\link[=hexify_lonlat_to_index]{hexify_lonlat_to_index()}},
\code{\link[=hexify_z7_canonical]{hexify_z7_canonical()}}
}
\concept{hierarchical index}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dggrid_compat.R
\name{as_dggrid}
\alias{as_dggrid}
\title{Convert hexify grid to 'dggridR'-compatible grid object}
\usage{
as_dggrid(grid)
}
\arguments{
\item{grid}{A hexify_grid object from hexify_grid()}
}
\value{
A list with 'dggridR'-compatible fields:
\item{pole_lon_deg}{Longitude of grid pole (default 11.25)}
\item{pole_lat_deg}{Latitude of grid pole (default 58.282525588538995)}
\item{azimuth_deg}{Grid azimuth rotation (default 0)}
\item{aperture}{Grid aperture (3, 4, or 7)}
\item{res}{Resolution level}
\item{topology}{Grid topology ("HEXAGON")}
\item{projection}{Map projection ('ISEA')}
\item{precision}{Output decimal precision (default 7)}
}
\description{
Creates a 'dggridR'-compatible grid specification from a hexify_grid object.
The resulting object can be used with 'dggridR' functions that accept a dggs
object.
}
\seealso{
Other 'dggridR' compatibility:
\code{\link[=dggrid_43h_sequence]{dggrid_43h_sequence()}},
\code{\link[=dggrid_is_compatible]{dggrid_is_compatible()}},
\code{\link[=from_dggrid]{from_dggrid()}}
}
\concept{'dggridR' compatibility}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_methods.R
\name{plot,HexData,missing-method}
\alias{plot,HexData,missing-method}
\title{Plot HexData objects}
\usage{
\S4method{plot}{HexData,missing}(
  x,
  y,
  basemap = TRUE,
  clip_basemap = TRUE,
  basemap_fill = "gray90",
  basemap_border = "gray50",
  basemap_lwd = 0.5,
  grid_fill = "#E69F00",
  grid_border = "#5D4E37",
  grid_lwd = 0.8,
  grid_alpha = 0.7,
  fill = NULL,
  show_points = FALSE,
  point_size = "auto",
  point_color = "red",
  point_alpha = 1,
  jitter = TRUE,
  crop = TRUE,
  crop_expand = 0.1,
  main = NULL,
  ...
)
}
\arguments{
\item{x}{A HexData object from \code{hexify()}}

\item{y}{Ignored (for S4 method compatibility)}

\item{basemap}{Basemap specification:
\itemize{
\item \code{TRUE} or \code{"world"}: Use built-in world map
\item \code{FALSE} or \code{NULL}: No basemap
\item sf object: Custom basemap
}}

\item{clip_basemap}{Clip basemap to data extent (default TRUE). Clipping
temporarily disables S2 spherical geometry to avoid edge-crossing errors.}

\item{basemap_fill}{Fill color for basemap (default "gray90")}

\item{basemap_border}{Border color for basemap (default "gray50")}

\item{basemap_lwd}{Line width for basemap borders (default 0.5)}

\item{grid_fill}{Fill color for grid cells (default "#E69F00" - amber/orange)}

\item{grid_border}{Border color for grid cells (default "#5D4E37" - dark brown)}

\item{grid_lwd}{Line width for cell borders (default 0.8)}

\item{grid_alpha}{Transparency for cell fill (0-1, default 0.7)}

\item{fill}{Column name for fill mapping (optional)}

\item{show_points}{Show original points on top of cells (default FALSE).
Points are jittered within their assigned hexagon.}

\item{point_size}{Size of points. Can be:
\itemize{
\item A number (direct cex value)
\item A preset defining what fraction of a hex cell one point covers:
"tiny" (~2\\%), "small" (~5\\%), "normal"/"auto" (~10\\%),
"large" (~20\\%), "very large" (~35\\%)
}}

\item{point_color}{Color of points (default "red")}

\item{point_alpha}{Transparency for points (0-1, default 1)}

\item{jitter}{If TRUE (default), points are randomly scattered within
their assigned hexagon. If FALSE, all points in a cell are plotted at
the cell centroid instead.}

\item{crop}{Crop to data extent (default TRUE)}

\item{crop_expand}{Expansion factor for crop (default 0.1)}

\item{main}{Plot title}

\item{...}{Additional arguments passed to base plot()}
}
\value{
Invisibly returns the HexData object
}
\description{
Default plot method for HexData objects. Draws hexagonal grid cells
with an optional basemap.
}
\details{
This function generates polygon geometries for the cells present in
the data and plots them. Polygons are computed on demand, not stored,
to minimize memory usage.
}
\examples{
\donttest{
df <- data.frame(lon = runif(50, -5, 5), lat = runif(50, 45, 50))
result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 2000)

# Basic plot
plot(result, basemap = FALSE)

# With basemap and custom styling
plot(result, grid_fill = "lightblue", grid_border = "darkblue")
}
}
\seealso{
\code{\link{hexify_heatmap}} for ggplot2 plotting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\docType{class}
\name{HexGridInfo-class}
\alias{HexGridInfo-class}
\title{HexGridInfo Class}
\description{
An S4 class representing a hexagonal grid specification. Stores all
parameters needed for grid operations.
}
\details{
Create HexGridInfo objects using the \code{\link{hex_grid}} constructor function.
Do not use \code{new("HexGridInfo", ...)} directly.

The aperture can be "3", "4", "7" for grids that refine by one aperture at
every level; a family name such as "4/3" or "4/7", which refines by the first
aperture for the first floor(resolution / 2) levels and by the second for the
rest; or one aperture per level, "4/4/7/3".

For H3 grids, the aperture is fixed at "7" and resolution ranges from 0 to 15.
}
\section{Slots}{

\describe{
\item{\code{aperture}}{Character. Grid aperture: "3", "4", "7", a mixed family such
as "4/3" or "4/7", or one aperture per resolution level ("4/4/7/3").}

\item{\code{resolution}}{Integer. Grid resolution level (0-30 for ISEA, 0-15 for H3).}

\item{\code{area_km2}}{Numeric. Cell area in square kilometers.}

\item{\code{diagonal_km}}{Numeric. Cell diagonal (long diagonal) in kilometers.}

\item{\code{crs}}{Integer or character. Coordinate reference system: an EPSG code,
or a 'PROJ' or 'WKT' string. Defaults to 'WGS84' on Earth, and to a longlat
CRS on the sphere of \code{radius_km} on any other body.}

\item{\code{grid_type}}{Character. Grid system: "isea" (default) or "h3".}

\item{\code{radius_km}}{Numeric. Radius of the body the grid covers, in kilometers.
\code{NA} reads as Earth's mean radius.}
}}

\seealso{
\code{\link{hex_grid}} for the constructor function,
\code{\link{HexData-class}} for hexified data objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_lonlat_to_index}
\alias{hexify_lonlat_to_index}
\title{Convert longitude/latitude to index string}
\usage{
hexify_lonlat_to_index(
  lon,
  lat,
  resolution,
  aperture = 3L,
  index_type = c("auto", "z3", "z7", "zorder")
)
}
\arguments{
\item{lon}{Longitude in degrees}

\item{lat}{Latitude in degrees}

\item{resolution}{Resolution level}

\item{aperture}{Aperture (3, 4, or 7)}

\item{index_type}{Index encoding: "auto" (default), "z3", "z7", or "zorder"}
}
\value{
A character vector of index strings.
}
\description{
Projects geographic coordinates to grid cells and returns their hierarchical
index strings. Inputs are vectorized over \code{lon} and \code{lat}.
}
\examples{
idx <- hexify_lonlat_to_index(16.37, 48.21, resolution = 5, aperture = 3)
idx7 <- hexify_lonlat_to_index(16.37, 48.21,
  resolution = 4, aperture = 7
)
}
\seealso{
Other hierarchical index:
\code{\link[=hexify_cell_to_index]{hexify_cell_to_index()}},
\code{\link[=hexify_compare_indices]{hexify_compare_indices()}},
\code{\link[=hexify_default_index_type]{hexify_default_index_type()}},
\code{\link[=hexify_get_children]{hexify_get_children()}},
\code{\link[=hexify_get_parent]{hexify_get_parent()}},
\code{\link[=hexify_get_resolution]{hexify_get_resolution()}},
\code{\link[=hexify_index_to_cell]{hexify_index_to_cell()}},
\code{\link[=hexify_index_to_lonlat]{hexify_index_to_lonlat()}},
\code{\link[=hexify_is_valid_index_type]{hexify_is_valid_index_type()}},
\code{\link[=hexify_z7_canonical]{hexify_z7_canonical()}}
}
\concept{hierarchical index}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_cell_to_plane}
\alias{hexify_cell_to_plane}
\title{Convert Cell ID to PLANE coordinates}
\usage{
hexify_cell_to_plane(cell_id, resolution, aperture = 3L)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs (1-based)}

\item{resolution}{Grid resolution level (0-30)}

\item{aperture}{Grid aperture: 3, 4, or 7}
}
\value{
Data frame with columns:
\item{plane_x}{X coordinate in PLANE space (range ~0 to 5.5)}
\item{plane_y}{Y coordinate in PLANE space (range ~0 to 1.73)}
}
\description{
Converts DGGRID-compatible cell IDs directly to PLANE coordinates.
Returns the cell center in the unfolded icosahedron layout.
}
\details{
Compatible with 'dggridR' dgSEQNUM_to_PLANE().
}
\examples{
# Get PLANE coordinates for cells
plane <- hexify_cell_to_plane(cell_id = c(100, 200, 300),
                               resolution = 5, aperture = 3)
plot(plane$plane_x, plane$plane_y)
}
\seealso{
\code{\link{hexify_icosa_tri_to_plane}} for triangle conversion,
\code{\link{hexify_lonlat_to_plane}} for lon/lat conversion

Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hex_topology.R
\name{is_pentagon}
\alias{is_pentagon}
\title{Detect Pentagon Cells}
\usage{
is_pentagon(cell_id, grid)
}
\arguments{
\item{cell_id}{Cell IDs to check. Numeric for ISEA, character for H3.}

\item{grid}{A HexGridInfo or HexData object specifying the grid.}
}
\value{
A logical vector. \code{TRUE} for pentagon cells, \code{FALSE} for hexagons.
}
\description{
Identifies which cells are pentagons. Any hexagonal tiling of the sphere
must contain exactly 12 pentagons (at the icosahedron vertices).
Pentagon cells have 5 neighbors instead of 6.
}
\details{
\strong{H3 backend:} Uses the vendored H3 \code{isPentagon} function.

\strong{ISEA backend:} The 12 pentagons are located at icosahedron vertices,
which are always the (i, j) = (0, 0) cell of their quad. Pentagon status
is checked by decoding each input cell ID's own (quad, i, j) coordinates
via \code{\link[=cell_to_lonlat]{cell_to_lonlat()}}'s underlying grid math and testing whether i and j
are both zero, rather than by re-deriving each vertex's cell ID (the
forward direction has aperture-specific quirks -- e.g. aperture 7's
substrate/surrogate coordinate distinction -- that make a single fixed
formula for "the (0,0) cell ID of quad Q" unreliable across apertures).
}
\examples{
\donttest{
# H3 pentagon detection
g <- hex_grid(resolution = 1, type = "h3")
cells <- grid_global(g)
pent <- is_pentagon(cells$cell_id, g)
sum(pent)  # Should be 12
}
}
\seealso{
\code{\link[=get_neighbors]{get_neighbors()}} for neighbor finding (pentagons have 5 neighbors)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crosswalk.R
\name{h3_crosswalk}
\alias{h3_crosswalk}
\title{Crosswalk Between ISEA and H3 Cell IDs}
\usage{
h3_crosswalk(
  cell_id = NULL,
  grid,
  h3_resolution = NULL,
  isea_grid = NULL,
  direction = c("isea_to_h3", "h3_to_isea")
)
}
\arguments{
\item{cell_id}{Cell IDs to translate. Numeric for ISEA, character for H3.
When \code{grid} is a HexData object and \code{cell_id} is \code{NULL},
all cell IDs from the data are used.}

\item{grid}{A HexGridInfo or HexData object. For \code{direction =
"isea_to_h3"}, this must be an ISEA grid. For \code{direction =
"h3_to_isea"}, this must be an H3 grid.}

\item{h3_resolution}{Target H3 resolution for \code{"isea_to_h3"}. When
\code{NULL} (default), the closest H3 resolution matching the ISEA cell
area is selected automatically. For \code{"h3_to_isea"}, the source H3
resolution is always inferred from \code{grid}; if \code{h3_resolution}
is supplied for that direction it is validated against \code{grid}'s
resolution rather than used to select one.}

\item{isea_grid}{A HexGridInfo for the target ISEA grid. Required when
\code{direction = "h3_to_isea"}.}

\item{direction}{One of \code{"isea_to_h3"} (default) or \code{"h3_to_isea"}.}
}
\value{
A data frame with columns:
\describe{
\item{isea_cell_id}{ISEA cell ID (numeric)}
\item{h3_cell_id}{H3 cell ID (character)}
\item{isea_area_km2}{Area of the ISEA cell in km2}
\item{h3_area_km2}{Geodesic area of the H3 cell in km2}
\item{area_ratio}{Ratio of ISEA area to H3 area}
}
}
\description{
Maps cell IDs between ISEA (equal-area) and H3 grid systems by looking up
each cell's center coordinate in the target grid. This enables workflows
where analysis is done in ISEA (exact equal-area) and reporting in H3
(industry-standard).
}
\details{
The crosswalk works by computing the center coordinate of each source cell,
then finding which cell in the target grid contains that center. This is a
many-to-one mapping: multiple ISEA cells may map to the same H3 cell (or
vice versa) depending on the relative resolutions.

When \code{h3_resolution} is \code{NULL} and \code{direction = "isea_to_h3"},
the H3 resolution whose average cell area is closest to the ISEA cell area
is chosen automatically. This gives the best 1:1 correspondence.
}
\examples{
\donttest{
# ISEA -> H3
grid <- hex_grid(area_km2 = 1000)
cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
xwalk <- h3_crosswalk(cells, grid)
head(xwalk)

# H3 -> ISEA
h3 <- hex_grid(resolution = 5, type = "h3")
h3_cells <- lonlat_to_cell(c(0, 10), c(45, 50), h3)
xwalk2 <- h3_crosswalk(h3_cells, h3, isea_grid = grid, direction = "h3_to_isea")
}
}
\seealso{
\code{\link{cell_area}} for per-cell area computation,
\code{\link{hex_grid}} for creating grids
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_cell_to_sf.R
\name{hexify_grid_rect}
\alias{hexify_grid_rect}
\title{Generate a rectangular grid of hexagon polygons}
\usage{
hexify_grid_rect(
  minlon,
  maxlon,
  minlat,
  maxlat,
  area,
  aperture = 3L,
  resround = "nearest",
  radius_km = EARTH_RADIUS_KM
)
}
\arguments{
\item{minlon, maxlon}{Longitude bounds}

\item{minlat, maxlat}{Latitude bounds}

\item{area}{Target cell area in km^2}

\item{aperture}{Grid aperture: 3, 4, or 7}

\item{resround}{Resolution rounding: "nearest", "up", or "down"}

\item{radius_km}{Radius of the body, in kilometers, or a body name such as
"mars" (default Earth). See \code{\link{hex_grid}}.}
}
\value{
sf object with hexagon polygons covering the specified region
}
\description{
Creates hexagon polygons covering a rectangular geographic region.
}
\examples{
library(hexify)
library(sf)

grid <- hexify_grid_rect(
  minlon = -10, maxlon = 20,
  minlat = 35, maxlat = 60,
  area = 5000
)
plot(st_geometry(grid), border = "gray")
}
\seealso{
\code{\link{grid_rect}} for the recommended S4 interface,
\code{\link{hexify_grid_global}} for global grids

Other sf conversion:
\code{\link[=hex_corners_to_sf]{hex_corners_to_sf()}},
\code{\link[=hexify_cell_to_sf]{hexify_cell_to_sf()}},
\code{\link[=hexify_grid_global]{hexify_grid_global()}}
}
\concept{sf conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{grid_rect}
\alias{grid_rect}
\title{Generate a rectangular grid of hexagons}
\usage{
grid_rect(bbox, grid)
}
\arguments{
\item{bbox}{Bounding box as c(xmin, ymin, xmax, ymax), or an sf/sfc object}

\item{grid}{A HexGridInfo object specifying the grid parameters}
}
\value{
sf object with hexagon polygons
}
\description{
Creates hexagon polygons covering a rectangular geographic region.
For H3 grids, all cells that overlap the bounding box are included
(not just cells whose center falls inside), ensuring full spatial coverage.
}
\examples{
grid <- hex_grid(area_km2 = 5000)
europe <- grid_rect(c(-10, 35, 30, 60), grid)
plot(europe)
}
\seealso{
\code{\link{grid_global}} for global grids
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_default_index_type}
\alias{hexify_default_index_type}
\title{Get default index type for aperture}
\usage{
hexify_default_index_type(aperture)
}
\arguments{
\item{aperture}{Aperture (3, 4, or 7)}
}
\value{
String: "z3", "z7", or "zorder"
}
\description{
Returns the recommended index type:
\itemize{
\item Aperture 3: "z3"
\item Aperture 4: "zorder"
\item Aperture 7: "z7"
}
}
\seealso{
Other hierarchical index:
\code{\link[=hexify_cell_to_index]{hexify_cell_to_index()}},
\code{\link[=hexify_compare_indices]{hexify_compare_indices()}},
\code{\link[=hexify_get_children]{hexify_get_children()}},
\code{\link[=hexify_get_parent]{hexify_get_parent()}},
\code{\link[=hexify_get_resolution]{hexify_get_resolution()}},
\code{\link[=hexify_index_to_cell]{hexify_index_to_cell()}},
\code{\link[=hexify_index_to_lonlat]{hexify_index_to_lonlat()}},
\code{\link[=hexify_is_valid_index_type]{hexify_is_valid_index_type()}},
\code{\link[=hexify_lonlat_to_index]{hexify_lonlat_to_index()}},
\code{\link[=hexify_z7_canonical]{hexify_z7_canonical()}}
}
\concept{hierarchical index}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_cell_id_to_quad_ij}
\alias{hexify_cell_id_to_quad_ij}
\title{Get cell info from cell ID}
\usage{
hexify_cell_id_to_quad_ij(cell_id, resolution, aperture)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs (1-based)}

\item{resolution}{Grid resolution (integer >= 0)}

\item{aperture}{Grid aperture (3, 4, or 7)}
}
\value{
Data frame with quad, i, j columns
}
\description{
Converts cell ID to cell components (quad, i, j).
}
\examples{
info <- hexify_cell_id_to_quad_ij(1702, resolution = 5, aperture = 3)
}
\seealso{
Other coordinate conversion:
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_quad_xy_to_icosa_tri}
\alias{hexify_quad_xy_to_icosa_tri}
\title{Convert Quad XY to Icosa Triangle coordinates}
\usage{
hexify_quad_xy_to_icosa_tri(quad, quad_x, quad_y)
}
\arguments{
\item{quad}{Quad number (0-11)}

\item{quad_x}{Continuous X coordinate in quad space}

\item{quad_y}{Continuous Y coordinate in quad space}
}
\value{
List with components:
\item{icosa_triangle_face}{Triangle face number (0-19)}
\item{icosa_triangle_x}{X coordinate on triangle face}
\item{icosa_triangle_y}{Y coordinate on triangle face}
}
\description{
Inverse transformation from quad coordinates back to icosahedral
triangle coordinates. Useful for projecting cell centers back to lon/lat.
}
\examples{
# Convert quad XY back to triangle coordinates
tri <- hexify_quad_xy_to_icosa_tri(quad = 1, quad_x = 0.5, quad_y = 0.3)
print(tri)
}
\seealso{
Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{cell_to_sf}
\alias{cell_to_sf}
\title{Convert cell IDs to sf polygons}
\usage{
cell_to_sf(cell_id = NULL, grid, wrap_dateline = TRUE)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs. If NULL and x is HexData,
uses cells from x.}

\item{grid}{A HexGridInfo or HexData object. If HexData and cell_id is NULL,
polygons are generated for all cells in the data.}

\item{wrap_dateline}{Logical. If TRUE (default), calls
\code{sf::st_wrap_dateline()} to split antimeridian-crossing polygons.
Set to FALSE for orthographic/globe projections where wrapping creates gaps.}
}
\value{
sf object with cell_id and geometry columns
}
\description{
Creates sf polygon geometries for hexagonal grid cells.
}
\details{
When called with a HexData object and no cell_id argument, this function
generates polygons for all unique cells in the data, which is useful for
plotting.
}
\examples{
# From grid specification
grid <- hex_grid(area_km2 = 1000)
cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
polys <- cell_to_sf(cells, grid)

# From HexData (all cells)
df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
polys <- cell_to_sf(grid = result)
}
\seealso{
\code{\link{hex_grid}} for grid specifications,
\code{\link{as_sf}} for converting HexData to sf
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{cell_to_index}
\alias{cell_to_index}
\title{Convert cell ID to hierarchical index string}
\usage{
cell_to_index(cell_id, grid)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs}

\item{grid}{A HexGridInfo or HexData object}
}
\value{
Character vector of hierarchical index strings
}
\description{
Advanced function for working with hierarchical index strings.
Most users don't need this - use cell IDs directly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_projection.R
\name{hexify_build_icosa}
\alias{hexify_build_icosa}
\title{Initialize icosahedron geometry}
\usage{
hexify_build_icosa(
  vert0_lon = ISEA_VERT0_LON_DEG,
  vert0_lat = ISEA_VERT0_LAT_DEG,
  azimuth = ISEA_AZIMUTH_DEG
)
}
\arguments{
\item{vert0_lon}{Vertex 0 longitude in degrees (default ISEA_VERT0_LON_DEG)}

\item{vert0_lat}{Vertex 0 latitude in degrees (default ISEA_VERT0_LAT_DEG)}

\item{azimuth}{Azimuth rotation in degrees (default ISEA_AZIMUTH_DEG)}
}
\value{
Invisible NULL. Called for side effect.
}
\description{
Sets up the icosahedron state for ISEA projection. Uses standard ISEA3H
orientation by default (vertex 0 at 11.25E, 58.28N).
}
\details{
The icosahedron is initialized lazily at the C++ level when first needed.
Manual call is only required for non-standard orientations.
}
\examples{
# Use standard ISEA3H orientation
hexify_build_icosa()

# Custom orientation
hexify_build_icosa(vert0_lon = 0, vert0_lat = 90, azimuth = 0)
}
\seealso{
Other projection:
\code{\link[=hexify_face_centers]{hexify_face_centers()}},
\code{\link[=hexify_forward]{hexify_forward()}},
\code{\link[=hexify_forward_to_face]{hexify_forward_to_face()}},
\code{\link[=hexify_get_precision]{hexify_get_precision()}},
\code{\link[=hexify_inverse]{hexify_inverse()}},
\code{\link[=hexify_projection_stats]{hexify_projection_stats()}},
\code{\link[=hexify_set_precision]{hexify_set_precision()}},
\code{\link[=hexify_set_verbose]{hexify_set_verbose()}},
\code{\link[=hexify_which_face]{hexify_which_face()}}
}
\concept{projection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_helpers.R
\name{get_parent}
\alias{get_parent}
\title{Get parent cell}
\usage{
get_parent(cell_id, grid, levels = 1L)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs}

\item{grid}{A HexGridInfo or HexData object}

\item{levels}{Number of levels up (default 1)}
}
\value{
Numeric vector of parent cell IDs
}
\description{
Returns the parent cell at a coarser resolution.
}
\examples{
grid <- hex_grid(resolution = 10)
child_cells <- lonlat_to_cell(c(0, 10), c(45, 50), grid)
parent_cells <- get_parent(child_cells, grid)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify-conversions}
\alias{hexify-conversions}
\title{Coordinate Conversions}
\description{
Functions for converting between coordinate systems
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{HexGridInfo_to_hexify_grid}
\alias{HexGridInfo_to_hexify_grid}
\title{Convert HexGridInfo to legacy hexify_grid}
\usage{
HexGridInfo_to_hexify_grid(x)
}
\arguments{
\item{x}{A HexGridInfo object (S4)}
}
\value{
A hexify_grid object (S3)
}
\description{
For backwards compatibility with existing functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_projection.R
\name{hexify_forward}
\alias{hexify_forward}
\title{Forward Snyder projection}
\usage{
hexify_forward(lon, lat)
}
\arguments{
\item{lon}{Longitude in degrees}

\item{lat}{Latitude in degrees}
}
\value{
Named numeric vector: c(face, tx, ty)
}
\description{
Projects geographic coordinates onto the icosahedron, returning
face index and planar coordinates (tx, ty).
}
\details{
tx and ty are normalized coordinates within the triangular face,
typically in range [0, 1].
}
\examples{
result <- hexify_forward(16.37, 48.21)
# result["face"], result["icosa_triangle_x"], result["icosa_triangle_y"]
}
\seealso{
Other projection:
\code{\link[=hexify_build_icosa]{hexify_build_icosa()}},
\code{\link[=hexify_face_centers]{hexify_face_centers()}},
\code{\link[=hexify_forward_to_face]{hexify_forward_to_face()}},
\code{\link[=hexify_get_precision]{hexify_get_precision()}},
\code{\link[=hexify_inverse]{hexify_inverse()}},
\code{\link[=hexify_projection_stats]{hexify_projection_stats()}},
\code{\link[=hexify_set_precision]{hexify_set_precision()}},
\code{\link[=hexify_set_verbose]{hexify_set_verbose()}},
\code{\link[=hexify_which_face]{hexify_which_face()}}
}
\concept{projection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_grid.R
\name{hexify_grid}
\alias{hexify_grid}
\title{Create a hexagonal grid specification}
\usage{
hexify_grid(
  area,
  topology = "HEXAGON",
  metric = TRUE,
  resround = "nearest",
  aperture = 3,
  projection = "ISEA",
  radius_km = EARTH_RADIUS_KM
)
}
\arguments{
\item{area}{Target cell area in km^2 (if metric=TRUE) or area code}

\item{topology}{Grid topology (only "HEXAGON" supported)}

\item{metric}{Whether area is in metric units (km^2)}

\item{resround}{How to round resolution ("nearest", "up", "down")}

\item{aperture}{Aperture sequence (3, 4, or 7)}

\item{projection}{Projection type (only 'ISEA' supported currently)}

\item{radius_km}{Radius of the body the grid covers, in kilometers, or a body
name such as "mars" (default Earth). See \code{\link{hex_grid}}.}
}
\value{
A hexify_grid object containing:
\item{area}{Target cell area}
\item{resolution}{Calculated resolution level}
\item{aperture}{Grid aperture (3, 4, or 7)}
\item{topology}{Grid topology ("HEXAGON")}
\item{projection}{Map projection ("ISEA")}
\item{radius_km}{Radius of the body, in kilometers}
\item{index_type}{Index encoding type ("z3", "z7", or "zorder")}
}
\description{
Creates a discrete global grid system (DGGS) object with hexagonal cells
at a specified resolution. This is the main constructor for hexify grids.
}
\examples{
# Create a grid with ~1000 km^2 cells
grid <- hexify_grid(area = 1000, aperture = 3)
print(grid)

# Create a finer resolution grid (~100 km^2 cells)
fine_grid <- hexify_grid(area = 100, aperture = 3, resround = "up")
}
\seealso{
\code{\link{hexify}} for the main user function,
\code{\link{hexify_grid_to_cell}} for coordinate conversion

Other hexify main:
\code{\link[=hexify]{hexify()}}
}
\concept{hexify main}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_lonlat_to_plane}
\alias{hexify_lonlat_to_plane}
\title{Convert longitude/latitude to PLANE coordinates}
\usage{
hexify_lonlat_to_plane(lon, lat)
}
\arguments{
\item{lon}{Longitude in degrees (-180 to 180)}

\item{lat}{Latitude in degrees (-90 to 90)}
}
\value{
Data frame with columns:
\item{plane_x}{X coordinate in PLANE space (range ~0 to 5.5)}
\item{plane_y}{Y coordinate in PLANE space (range ~0 to 1.73)}
}
\description{
Converts geographic coordinates directly to PLANE coordinates
(unfolded icosahedron). Combines forward 'Snyder' projection with
the PLANE transformation.
}
\details{
Equivalent to 'dggridR' dgGEO_to_PLANE().
}
\examples{
# Plot world cities in PLANE coordinates
cities <- data.frame(
  lon = c(2.35, -74.00, 139.69, 151.21),
  lat = c(48.86, 40.71, 35.69, -33.87)
)
plane <- hexify_lonlat_to_plane(cities$lon, cities$lat)
plot(plane$plane_x, plane$plane_y)
}
\seealso{
\code{\link{hexify_cell_to_plane}} for cell ID conversion,
\code{\link{hexify_icosa_tri_to_plane}} for triangle conversion

Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_stats.R
\name{dg_closest_res_to_area}
\alias{dg_closest_res_to_area}
\title{Find closest resolution for target cell area}
\usage{
dg_closest_res_to_area(
  dggs,
  area,
  round = "nearest",
  metric = TRUE,
  show_info = FALSE
)
}
\arguments{
\item{dggs}{Grid specification (aperture and topology must be set)}

\item{area}{Target cell area in km^2 (if metric=TRUE)}

\item{round}{Rounding method ("nearest", "up", "down")}

\item{metric}{Whether area is in metric units}

\item{show_info}{Print information about chosen resolution}
}
\value{
Resolution level (integer)
}
\description{
Finds the grid resolution that produces cells closest to the target area.
This is primarily used internally by \code{\link{hexify_grid}} and
\code{\link{hex_grid}}. Most users should use those functions directly.
}
\examples{
# Create a temporary grid to get aperture settings
temp_grid <- list(aperture = 3, topology = "HEXAGON")
class(temp_grid) <- "hexify_grid"

# Find resolution for 1000 km^2 cells
res <- dg_closest_res_to_area(temp_grid, area = 1000, 
                               metric = TRUE, show_info = TRUE)
print(res)
}
\seealso{
Other grid statistics:
\code{\link[=dgearthstat]{dgearthstat()}},
\code{\link[=hexify_area_to_eff_res]{hexify_area_to_eff_res()}},
\code{\link[=hexify_compare_resolutions]{hexify_compare_resolutions()}},
\code{\link[=hexify_eff_res_to_area]{hexify_eff_res_to_area()}},
\code{\link[=hexify_eff_res_to_resolution]{hexify_eff_res_to_resolution()}},
\code{\link[=hexify_resolution_to_eff_res]{hexify_resolution_to_eff_res()}}
}
\concept{grid statistics}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_cell_to_quad_xy}
\alias{hexify_cell_to_quad_xy}
\title{Convert Cell ID to Quad XY coordinates}
\usage{
hexify_cell_to_quad_xy(cell_id, resolution, aperture = 3L)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs (1-based)}

\item{resolution}{Grid resolution level (0-30)}

\item{aperture}{Grid aperture: 3, 4, or 7}
}
\value{
Data frame with columns:
\item{quad}{Quad number (0-11)}
\item{quad_x}{Continuous X coordinate in quad space}
\item{quad_y}{Continuous Y coordinate in quad space}
}
\description{
Converts DGGRID-compatible cell IDs to Quad XY coordinates
(continuous quad space). This is the cell center in quad coordinates.
}
\details{
Compatible with 'dggridR' dgSEQNUM_to_Q2DD().
}
\examples{
# Get Quad XY coordinates for a cell
result <- hexify_cell_to_quad_xy(cell_id = 1000, resolution = 10, aperture = 3)
print(result)

# Round-trip test
cell_id <- hexify_quad_xy_to_cell(result$quad, result$quad_x, result$quad_y,
                                   resolution = 10, aperture = 3)
# Should equal original cell_id
}
\seealso{
\code{\link{hexify_quad_xy_to_cell}} for the inverse operation,
\code{\link{hexify_cell_to_quad_ij}} for integer grid coordinates

Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_cell_to_icosa_tri}
\alias{hexify_cell_to_icosa_tri}
\title{Convert Cell ID to Icosa Triangle coordinates}
\usage{
hexify_cell_to_icosa_tri(cell_id, resolution, aperture = 3L)
}
\arguments{
\item{cell_id}{Numeric vector of cell IDs (1-based)}

\item{resolution}{Grid resolution level (0-30)}

\item{aperture}{Grid aperture: 3, 4, or 7}
}
\value{
Data frame with columns:
\item{icosa_triangle_face}{Triangle face number (0-19)}
\item{icosa_triangle_x}{X coordinate on triangle face}
\item{icosa_triangle_y}{Y coordinate on triangle face}
}
\description{
Converts DGGRID-compatible cell IDs to icosahedral triangle
coordinates (face, x, y). These are the coordinates produced by the
'Snyder' 'ISEA' forward projection.
}
\details{
Compatible with 'dggridR' dgSEQNUM_to_PROJTRI().
}
\examples{
# Get triangle coordinates for a cell
result <- hexify_cell_to_icosa_tri(cell_id = 1000, resolution = 10, aperture = 3)
print(result)

# Convert back to lon/lat using inverse projection
coords <- hexify_inverse(result$icosa_triangle_face,
                         result$icosa_triangle_x,
                         result$icosa_triangle_y)
}
\seealso{
\code{\link{hexify_cell_to_quad_ij}} for conversion to Quad IJ,
\code{\link{hexify_cell_to_lonlat}} for conversion to lon/lat

Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}},
\code{\link[=hexify_roundtrip_test]{hexify_roundtrip_test()}}
}
\concept{coordinate conversion}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_conversions.R
\name{hexify_roundtrip_test}
\alias{hexify_roundtrip_test}
\title{Round-trip accuracy test}
\usage{
hexify_roundtrip_test(grid, lon, lat, units = "km")
}
\arguments{
\item{grid}{Grid specification}

\item{lon}{Longitude to test}

\item{lat}{Latitude to test}

\item{units}{Distance units ("km" or "degrees")}
}
\value{
List with:
\item{original}{Original coordinates}
\item{cell}{Cell index}
\item{reconstructed}{Reconstructed coordinates}
\item{error}{Distance between original and reconstructed}
}
\description{
Tests the accuracy of the coordinate conversion functions by converting
coordinates to cells and back, measuring the distance between original
and reconstructed coordinates.
}
\seealso{
Other coordinate conversion:
\code{\link[=hexify_cell_id_to_quad_ij]{hexify_cell_id_to_quad_ij()}},
\code{\link[=hexify_cell_to_icosa_tri]{hexify_cell_to_icosa_tri()}},
\code{\link[=hexify_cell_to_lonlat]{hexify_cell_to_lonlat()}},
\code{\link[=hexify_cell_to_plane]{hexify_cell_to_plane()}},
\code{\link[=hexify_cell_to_quad_ij]{hexify_cell_to_quad_ij()}},
\code{\link[=hexify_cell_to_quad_xy]{hexify_cell_to_quad_xy()}},
\code{\link[=hexify_grid_cell_to_lonlat]{hexify_grid_cell_to_lonlat()}},
\code{\link[=hexify_grid_to_cell]{hexify_grid_to_cell()}},
\code{\link[=hexify_icosa_tri_to_plane]{hexify_icosa_tri_to_plane()}},
\code{\link[=hexify_icosa_tri_to_quad_ij]{hexify_icosa_tri_to_quad_ij()}},
\code{\link[=hexify_icosa_tri_to_quad_xy]{hexify_icosa_tri_to_quad_xy()}},
\code{\link[=hexify_lonlat_to_cell]{hexify_lonlat_to_cell()}},
\code{\link[=hexify_lonlat_to_plane]{hexify_lonlat_to_plane()}},
\code{\link[=hexify_lonlat_to_quad_ij]{hexify_lonlat_to_quad_ij()}},
\code{\link[=hexify_quad_ij_to_cell]{hexify_quad_ij_to_cell()}},
\code{\link[=hexify_quad_ij_to_icosa_tri]{hexify_quad_ij_to_icosa_tri()}},
\code{\link[=hexify_quad_ij_to_xy]{hexify_quad_ij_to_xy()}},
\code{\link[=hexify_quad_xy_to_cell]{hexify_quad_xy_to_cell()}},
\code{\link[=hexify_quad_xy_to_icosa_tri]{hexify_quad_xy_to_icosa_tri()}}
}
\concept{coordinate conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexify_index.R
\name{hexify_cell_to_index}
\alias{hexify_cell_to_index}
\title{Convert cell coordinates to index string}
\usage{
hexify_cell_to_index(
  face,
  i,
  j,
  resolution,
  aperture = 3L,
  index_type = c("auto", "z3", "z7", "zorder")
)
}
\arguments{
\item{face}{Face/quad number (0-19)}

\item{i}{I coordinate}

\item{j}{J coordinate}

\item{resolution}{Resolution level}

\item{aperture}{Aperture (3, 4, or 7)}

\item{index_type}{Index encoding: "auto" (default), "z3", "z7", or "zorder"}
}
\value{
A character vector of index strings.
}
\description{
Converts cell coordinates (\code{face}, \code{i}, \code{j}) to a hierarchical index string.
The index type is selected from \code{aperture} when \code{index_type = "auto"}.
}
\details{
Default index types by aperture:
\itemize{
\item Aperture 3: Z3 (optimized digit selection)
\item Aperture 4: Z-order (Morton curve)
\item Aperture 7: Z7 (one base-7 child digit per resolution)
}

A Z7 index has the form \code{BBd1...dr}, where \code{BB} is the two-digit base
cell (00--11), \code{r} is the resolution, and every child digit is in 0--6.
hexify's Z7 encoding is bijective: decoding and re-encoding a valid index
returns the same string. It follows DGGRID's Z7 layout for ordinary cells,
but retains distinct indices in pentagon regions where DGGRID's encoder can
map different cells to the same string.
}
\examples{
idx <- hexify_cell_to_index(5, 10, 15, resolution = 3, aperture = 3)

# Aperture-7 indices contain a two-digit base cell and one digit per level
z7 <- hexify_cell_to_index(5, 1, 1, resolution = 2, aperture = 7)
nchar(z7) == 4L
}
\seealso{
Other hierarchical index:
\code{\link[=hexify_compare_indices]{hexify_compare_indices()}},
\code{\link[=hexify_default_index_type]{hexify_default_index_type()}},
\code{\link[=hexify_get_children]{hexify_get_children()}},
\code{\link[=hexify_get_parent]{hexify_get_parent()}},
\code{\link[=hexify_get_resolution]{hexify_get_resolution()}},
\code{\link[=hexify_index_to_cell]{hexify_index_to_cell()}},
\code{\link[=hexify_index_to_lonlat]{hexify_index_to_lonlat()}},
\code{\link[=hexify_is_valid_index_type]{hexify_is_valid_index_type()}},
\code{\link[=hexify_lonlat_to_index]{hexify_lonlat_to_index()}},
\code{\link[=hexify_z7_canonical]{hexify_z7_canonical()}}
}
\concept{hierarchical index}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constructors.R
\name{st_crs.HexGridInfo}
\alias{st_crs.HexGridInfo}
\alias{st_crs.HexData}
\title{Coordinate reference system of a grid}
\usage{
\method{st_crs}{HexGridInfo}(x, ...)

\method{st_crs}{HexData}(x, ...)
}
\arguments{
\item{x}{A HexGridInfo or HexData object}

\item{...}{Passed on to \code{sf::st_crs()}}
}
\value{
An object of class \code{crs}, as \code{sf::st_crs()} returns:
a list carrying the reference system in 'PROJ' and 'WKT' form. It says how
to read the coordinates that the grid's cells, centres and sf exports come
back in.
}
\description{
Reads the CRS a grid's coordinates are in, as \code{sf::st_crs()} does for
any spatial object. An Earth grid returns 'WGS84'; a grid built on another
body returns a longlat CRS on the sphere of its radius.
}
\examples{
st_crs(hex_grid(resolution = 5))
st_crs(hex_grid(resolution = 5, radius_km = "mars"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{grid_info}
\alias{grid_info}
\title{Get Grid Specification}
\usage{
grid_info(x)
}
\arguments{
\item{x}{A HexData object}
}
\value{
A HexGridInfo object
}
\description{
Extract the grid specification from a HexData object.
}
\examples{
df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
grid_spec <- grid_info(result)
}
